-- myConsole.e
-- v1.1
-- 2012-09-09
-- console functions

-------------------------------------------------------------------------------

-- public sequence TextAttr
-- public integer  TextRows, TextCols
-- public constant KEYB = 0, SCREEN = 1, ERR = 2

-- public procedure setTextAttr(sequence col)
-- public function  editText(sequence s, sequence attr, sequence pos, integer longueur)
-- public procedure displayMsg(sequence msg, object pos)
-- public function  listBox(sequence liste, integer defaut, sequence attr, sequence pos, integer long)
-- public function  comboBox(sequence liste, integer defaut, sequence attr, sequence pos, integer long)

-------------------------------------------------------------------------------

include std/console.e
include std/math.e
include std/sequence.e
include std/graphics.e
include myLibs/myTypes.e
include myLibs/myDebug.e
include myLibs/myConv.e

public sequence TextAttr
sequence vid_conf

public integer TextRows, TextCols
TextAttr = {WHITE, BLACK}
TextRows = 24
TextCols = 80
vid_conf = video_config()
TextRows = vid_conf[VC_LINES]
TextCols = min({80,vid_conf[VC_COLUMNS]})
--VC_SCRNLINES,
--VC_SCRNCOLS

public constant KEYB = 0, SCREEN = 1, ERR = 2

public constant
  UP={27,91,65}, DOWN={27,91,66}, LEFT={27,91,68}, RIGHT={27,91,67},
  ENTER={10}, ESC={27}, SPACE={32}, PAGE_UP={27,91,53,126},
  PAGE_DOWN={27,91,54,126}, HOME={27,91,72}, END={27,91,70},
  INSERT={27,91,50,126}, DELETE={27,91,51,126}, BACKSPACE={127}

public constant F_ = {
  {27,79,80},
  {27,79,81},
  {27,79,82},
  {27,79,83},
  {27,91,49,53,126},
  {27,91,49,55,126},
  {27,91,49,56,126},
  {27,91,49,57,126},
  {27,91,50,48,126},
  {27,91,50,49,126},
  {27,91,50,51,126},
  {27,91,50,52,126}
}

integer
  listLine, listSelected, listOffset

------------------------------------------------------------------------------

public procedure setTextAttr(sequence col)
  text_color(col[1])
  bk_color(col[2])
  TextAttr = col
end procedure

------------------------------------------------------------------------------

public procedure clrEol(integer col)
-- clear end of line
  sequence CurPos, vc
  vc = video_config()
  CurPos = get_position()
  for i = CurPos[2] to vc[VC_COLUMNS] do
    put_screen_char(CurPos[1], i, {' ', col})
  end for
end procedure -- Clreol()

------------------------------------------------------------------------------

public procedure printAt(sequence pos, object obj, sequence format, sequence col)
  sequence Attr

  if sequence(obj) then
    if length(obj)=0 then return end if
  end if
  Attr = TextAttr
  setTextAttr(col)
  position(pos[1], pos[2])
  if sequence(obj) then
    printf(1, format, {obj})
  else
    printf(1, format, obj)
  end if
  setTextAttr(Attr)
end procedure

------------------------------------------------------------------------------

public function wRepeat(object c, integer n)
  sequence res

  res = {}
  for i = 1 to n do
    if atom(c) then
      res &= {c}
    else
      res &= c
    end if
  end for
  return res
end function

------------------------------------------------------------------------------

procedure drawFrame(integer x1, integer y1, integer x2, integer y2, sequence attr)
  sequence col

--puts(f_debug, "drawFrame\n")
printf(f_debug, "x1=%d, y1=%d, x2=%d, y2=%d\n", {x1, y1, x2, y2})
  col = TextAttr
  setTextAttr(attr)
  position(y1, x1)
  puts(1, "┌"&wRepeat("─", x2-x1-1)&"┐")
  for i = 1 to y2-y1-1 do
    position(y1+i, x1) puts(1, "│")
    position(y1+i, x2) puts(1, "│")
  end for
  position(y2, x1)
  puts(1, "└"&wRepeat("─", x2-x1-1)&"┘")
  setTextAttr(col)
end procedure

-------------------------------------------------------------------------------

public procedure displayMsg(sequence msg, object pos)
-- prints a message in a frame at position pos
-- if pos is an atom (mostly 0), then center the frame in the screen
  sequence saved, col, s
  integer h, w, x1, y1, x2, y2
  atom c

  col = TextAttr
  s = split(msg, '\n')
  h = length(s)
  w = 0
  for i = 1 to h do
    if length(s[i]) > w then w = length(s[i]) end if
  end for
  if atom(pos) then
    x1 = floor((TextCols-w)/2)
    y1 = floor((TextRows-h)/2)
    x2 = x1+w+1
    y2 = y1+h+1
  else
    x1 = pos[2]-1
    x2 = pos[2]+w
    y1 = pos[1]-1
    y2 = pos[1]+h
  end if
  saved = save_text_image({y1,x1}, {y2,x2})
  drawFrame(x1, y1, x2, y2, {BLACK,WHITE})
  position(y1+1,x1+1)
  setTextAttr({GRAY,WHITE})
  puts(1,  msg)
  c = wait_key()
  display_text_image({y1,x1}, saved)
  setTextAttr(col)
end procedure

-------------------------------------------------------------------------------

public function getKeys()
  sequence result
  integer code

  result = {}
  code = wait_key()
  while code != -1 do
    result = append(result,code)
    code = get_key()
  end while
  return result
end function

------------------------------------------------------------------------------

public function editText(sequence s, sequence attr, sequence pos, integer longueur)
  sequence chaine
  integer l, pointeur_chaine
  sequence ex
  sequence col
  atom fl_out, fl_ins, premier

  col = TextAttr
  fl_ins = 1
  cursor(#0E0F)
  fl_out = 0
  setTextAttr(attr)
  chaine = s  l = 0
  pointeur_chaine = 1
  premier = 1
  while fl_out = 0 do
    position(pos[1], pos[2])  puts(1, repeat(' ', longueur))
    position(pos[1], pos[2])  puts(1, mid(chaine, 1, longueur))
    position(pos[1], pos[2] + pointeur_chaine - 1)
    ex = getKeys()
    if equal(ex, HOME) then
       pointeur_chaine = 1
    elsif equal(ex, END) then
      if length(chaine) < longueur then
        pointeur_chaine = length(chaine) + 1
      else
        pointeur_chaine = length(chaine)
      end if
    elsif equal(ex, LEFT) then
      if pointeur_chaine > 1 then
        pointeur_chaine = pointeur_chaine - 1
      end if
    elsif equal(ex, RIGHT) then
      if (pointeur_chaine < longueur) and (pointeur_chaine < length(chaine)+1) then
        pointeur_chaine = pointeur_chaine + 1
      end if
    elsif equal(ex, DELETE) then
      l = length(chaine)
      if pointeur_chaine <= l then
        if pointeur_chaine = 1 then
          chaine = chaine[2..l]
        elsif pointeur_chaine = l then
          chaine = chaine[1..l-1]
          pointeur_chaine = l-1
        else
          chaine = remove(chaine, pointeur_chaine)
        end if
      end if
    elsif equal(ex, INSERT) then
      fl_ins = not fl_ins
      if fl_ins then
        cursor(#0E0F)
      else
        cursor(#000F)
      end if
    elsif find(ex, F_) or equal(ex, UP) or equal(ex, DOWN) or
          equal(ex, ENTER) or equal(ex, ESC) then
      fl_out = 1
    elsif equal(ex, BACKSPACE) then
      l = length(chaine)
      if pointeur_chaine > 1 then
        chaine = remove(chaine, pointeur_chaine-1)
        pointeur_chaine = pointeur_chaine-1
      end if
    else
      if premier then
        chaine = append("", ex[1])
        pointeur_chaine = 2
      elsif fl_ins and (length(chaine) < longueur) then
        chaine = insert(chaine, ex[1], pointeur_chaine)
        if (pointeur_chaine < longueur) and (pointeur_chaine < length(chaine)+1) then
          pointeur_chaine = pointeur_chaine + 1
        end if
      elsif (not fl_ins) and (pointeur_chaine <= length(chaine)) then
        chaine[pointeur_chaine] = ex[1]
        if (pointeur_chaine < longueur) and (pointeur_chaine < length(chaine)+1) then
          pointeur_chaine = pointeur_chaine + 1
        end if
      end if
    end if
    premier = 0
  end while
  setTextAttr(col)
  cursor(#2000)
  return {chaine, ex}
end function

-------------------------------------------------------------------------------

procedure listGoFirst(sequence currentList)
  if length(currentList)=0 then return end if
  listSelected = 1
  listLine = 1
  listOffset = 0
end procedure

-------------------------------------------------------------------------------

procedure listGoPrevious(sequence currentList)
  if length(currentList)=0 then return end if
  if listSelected=1 then return end if
  listSelected -= 1
  if listLine=1 then
    listOffset -= 1
  else
    listLine -= 1
  end if
end procedure

-------------------------------------------------------------------------------

procedure listGoNext(sequence currentList)
  if length(currentList)=0 then return end if
  if listSelected=length(currentList) then return end if
  listSelected += 1
  if listLine=3 then
    listOffset += 1
  else
    listLine += 1
  end if
end procedure

-------------------------------------------------------------------------------

procedure listGoLast(sequence currentList)
  integer lg

  lg = length(currentList)
  if lg=0 then
    listLine = 0
    listSelected = 0
    listOffset = 0
    return
  end if
  if lg >= 3 then
    listLine = 3
    listOffset = lg-3
  else
    listGoFirst(currentList)
    listLine = lg
  end if
  listSelected = lg
end procedure

-------------------------------------------------------------------------------

procedure listGo(sequence currentList, integer pos)
  if pos = 0 then
    listLine = 0
    listSelected = 0
    listOffset = 0
  elsif pos = 1 then
    listGoFirst(currentList)
  elsif pos = length(currentList) then
    listGoLast(currentList)
  else
    listLine = 2
    listSelected = pos
    listOffset = pos-2
  end if
end procedure

-------------------------------------------------------------------------------

procedure refreshListBox(sequence liste, sequence pos, integer long)
  sequence s

  for i = 1 to min({3,length(liste)}) do
    position(pos[1]+i, pos[2]+1)
    if i = listLine then
      setTextAttr({WHITE,BLUE})
    else
      setTextAttr({BLACK,WHITE})
    end if
    s = liste[listOffset+i]
    puts(1,  pad_tail(s[1..min({long, length(s)})], long))
  end for
end procedure

-------------------------------------------------------------------------------

function drawListBox(sequence liste, integer defaut, sequence pos, integer long)
  sequence saved, col
  integer m, x1, y1, x2, y2

  col = TextAttr
  x1 = pos[2]-1
  x2 = pos[2]+long
  m = min({3,length(liste)})
  y1 = pos[1]-1
  y2 = pos[1]+m
  saved = save_text_image({y1,x1}, {y2,x2})
  drawFrame(x1, y1, x2, y2, {BLACK,WHITE})
  listGo(liste, defaut)
  refreshListBox(liste, {y1,x1}, long)
  setTextAttr(col)
  return {{y1,x1},saved}
end function

-------------------------------------------------------------------------------

public function listBox(sequence liste, integer defaut, sequence attr, sequence pos, integer long)
  sequence saved, s, box, col
  sequence resultat

  col = TextAttr
  setTextAttr(attr)
  s = drawListBox(liste, defaut, pos, long)
  saved = s[2]
  box = s[1]
  resultat = {0, {}}
  listSelected = defaut
  while not find(resultat[2], {ENTER,ESC}) do
    position(pos[1],pos[2])
    resultat[2] = ' '
    while not find(resultat[2], {ENTER,ESC,UP,DOWN}) do
      resultat[2] = getKeys()
    end while
    if equal(resultat[2],ENTER) then
      resultat[1] = liste[listSelected]
    elsif equal(resultat[2],UP) then
      listGoPrevious(liste)
    elsif equal(resultat[2],DOWN) then
      listGoNext(liste)
    elsif equal(resultat[2],ESC) then
      resultat[1] = ""
    end if
    refreshListBox(liste, box, long)
  end while
  setTextAttr(col)
  return resultat
end function

-------------------------------------------------------------------------------

function drawComboListBox(sequence liste, integer defaut, sequence pos, integer long)
  sequence saved, col
  integer m, x1, y1, x2, y2

  col = TextAttr
  x1 = pos[2]-1
  x2 = pos[2]+long
  m = min({3,length(liste)})
  if pos[1] <= TextRows-(m+2) then
    y1 = pos[1]+1
    y2 = pos[1]+m+2
  else
    y1 = pos[1]-m-2
    y2 = pos[1]-1
  end if
  saved = save_text_image({y1,x1}, {y2,x2})
  drawFrame(x1, y1, x2, y2, {BLACK,WHITE})
  listGo(liste, defaut)
  refreshListBox(liste, {y1,x1}, long)
  setTextAttr(col)
  return {{y1,x1},saved}
end function

-------------------------------------------------------------------------------

public function comboBox(sequence liste, integer defaut, sequence attr, sequence pos, integer long)
  sequence saved, s, box, col
  sequence resultat

  col = TextAttr
  setTextAttr(attr)
  s = drawComboListBox(liste, defaut, pos, long)
  saved = s[2]
  box = s[1]
  resultat = {0, {}}
  listSelected = defaut
  if (defaut = 0) or (length(liste)=0) then
    s = ""
  else
    s = liste[listSelected]
  end if
  while not find(resultat[2], {ENTER,ESC}) do
    resultat = editText(s, attr, pos, long)
    if equal(resultat[2],ENTER) then
    elsif equal(resultat[2],UP) then
      listGoPrevious(liste)
      s = liste[listSelected]
    elsif equal(resultat[2],DOWN) then
      listGoNext(liste)
      s = liste[listSelected]
    elsif equal(resultat[2],ESC) then
      resultat[1] = ""
    end if
    refreshListBox(liste, box, long)
  end while
  display_text_image(box, saved)
  setTextAttr(col)
  position(pos[1],pos[2])
  puts(1,  pad_tail(resultat[1], long))
  return resultat
end function
