-- myDebug.e
-- v1.1
-- 2012-09-09
-- debug controls
-- adapted to Euphoria 4.1

include std/graphics.e
include std/console.e
include std/io.e
include std/sequence.e
include std/types.e
include std/text.e
include std/wildcard.e
include myLibs/myTypes.e

global integer  f_debug = 1    -- debug file
global integer  debugLevel = 0   -- 0=no debug, 1=debug file only, 2=debug_file and screen output

------------------------------------------------------------------------------

public function dateStamp(sequence msg)
-- prefixes the message by date and time for logging
  sequence cur_date

  cur_date = date()
  return sprintf("%d-%02d-%02d %02d:%02d:%02d -> %s\n", {
                             (cur_date[1] + 1900), cur_date[2], cur_date[3],
                              cur_date[4], cur_date[5], cur_date[6], msg
                            })
end function

------------------------------------------------------------------------------

public procedure logMsg(sequence msg)
-- records logs
  if debugLevel>0 then
    puts(f_debug, dateStamp(msg))
    flush(f_debug)
  end if
  if debugLevel>1 then
    puts(1, dateStamp(msg))
  end if
end procedure

------------------------------------------------------------------------------

public procedure warnError(sequence msg, atom quit)
-- logs an error and displays a warning
-- if quit is set then abort
  logMsg(msg)
  maybe_any_key()
  if quit then
    abort(1)
  end if
end procedure

------------------------------------------------------------------------------

public function showPrintable(sequence s)
-- returns only printable characters of a sequence
-- non printable characters are replace by a dot
  sequence res

  res = ""
  for i = 1 to length(s) do
    if integer(s[i]) then
      if (s[i] > 31) and (s[i] < 127) then
        res &= s[i]
      elsif s[i] = 9 then
        res &= "\\t"
      elsif s[i] = 13 then
        res &= "\\r"
      elsif s[i] = 10 then
        res &= "\\n"
      else
        res &= "."
      end if
    else
      res &= "."
    end if
  end for
  return res
end function

------------------------------------------------------------------------------

public function sequenceDump(sequence s, integer detailed=0)
-- prints a sequence structure in a human readable way as one string
  integer subSequence
  sequence result = ""

  if isString(s) then
    result = sprintf("\"%s\"", {showPrintable(s)})
  else
    subSequence = 0
    for i=1 to length(s) do
      if sequence(s[i]) then
        subSequence = 1
        exit
      end if
    end for
    if subSequence then
      result = sequenceDump(s[1])
      for i=2 to length(s) do
        result &= ", " & sequenceDump(s[i])
      end for
      result = "{" & result & "}"
    else
      result = sprint(s) & "\t'" & showPrintable(s) & "'"
    end if
  end if
  return result
end function

------------------------------------------------------------------------------

public procedure objectDump(object x, sequence variable, integer output=1, integer detailed=0)
-- prints an object in a human readable way
  integer subSequence

--puts(output, variable&" = "&sprint(x)&"\n")
  if integer(x) then
    printf(output, "%s = %d\n", {variable, x})
  elsif atom(x) then
    if (x >= 0) and (x = floor(x)) then
      printf(output, "%s = %.0f\n", {variable, x})
    elsif (x < 0) and (x = floor(x+1)) then
      printf(output, "%s = %.0f\n", {variable, x})
    else
      printf(output, "%s = %f\n", {variable, x})
    end if
  elsif isString(x) then
    if detailed then
      printf(output, "%s = \"%s\"\t%s\n", {variable, x, sprint(x)})
    else
      printf(output, "%s = \"%s\"\n", {variable, x})
    end if
  else
    subSequence = 0
    for i=1 to length(x) do
      if sequence(x[i]) then
        subSequence = 1
        exit
      end if
    end for
    if subSequence then
      for i=1 to length(x) do
        objectDump(x[i], variable&sprintf("[%d]", i), output)
      end for
    else
      if detailed then
        printf(output, "%s = %s\t\"%s\"\n", {variable, sprint(x), showPrintable(x)})
      else
        printf(output, "%s = %s\n", {variable, sprint(x)})
      end if
    end if
  end if
  if output=f_debug then flush(f_debug) end if
end procedure

------------------------------------------------------------------------------

public procedure analyzeObject(object x, sequence Name, integer Output=1, integer translator=0, sequence path={}, integer level=0, sequence prefix="")
-- prints a sequence structure in a human readable way
  integer subSequence
  sequence s, offset

  if level=0 then printf(Output, "%s = ", {Name}) end if
  offset = ""
  for i = 1 to level do offset &= ".  " end for
  s = ""
  if integer(x) then
    s = sprintf(" %d", {x})
    puts(Output, offset&prefix&s&"\n")
  elsif atom(x) then
    if (x >= 0) and (x = floor(x)) then
      s = sprintf(" %.0f", {x})
    elsif (x < 0) and (x = floor(x+1)) then
      s = sprintf(" %.0f", {x})
    else
      s = sprintf(" %f", {x})
    end if
    puts(Output, offset&prefix&s&"\n")
  elsif string(x) then
    if length(x) = 0 then
      s = " \"\""
    else
      s = sprintf(" \"%s\"", {showPrintable(x)})
    end if
    puts(Output, offset&prefix&s&"\n")
  else
    subSequence = 0
    for i=1 to length(x) do
      if sequence(x[i]) then
        subSequence = 1
        exit
      end if
    end for
    if subSequence = 0 then
      s = sprintf(" %s", {sprint(x)})
      puts(Output, offset&prefix&s&"\n")
    else
      if length(s) = 0 then
        puts(Output, offset&prefix&"\n")
      end if
      for i=1 to length(x) do
        if translator > 0 then
          prefix = sprintf("[%s]", { call_func(translator, {path, level, i}) })
        else
          prefix = sprintf("[%d]", i)
        end if
        analyzeObject(x[i], Name, Output, translator, path&sprintf("[%d]", i), level+1, prefix)
      end for
    end if
  end if
  if Output=f_debug then flush(f_debug) end if
end procedure
