-- mySeq.e
-- v1.1
-- 2012-09-09
-- sequence functions
-- adapted to Euphoria 4.1

include std/types.e
include std/sequence.e
include std/text.e
include myTypes.e
include myDebug.e
include myConv.e

------------------------------------------------------------------------------

public function splitString(sequence str, atom sep)
-- splits str string in tokens according to separator sep
  integer slen
  sequence result, s
  integer tmpStr

  result = {}
  slen = length(str)
  if slen = 0 then return result end if
  
  s = {}
  tmpStr = 0      -- si 1, on est entre deux guillemets

  for i = 1 to slen do
    if (str[i] = '"') and (sep != '"') then
      if tmpStr = 0 then
        tmpStr = 1
      else
        tmpStr = 0
      end if
    else
      if tmpStr = 0 and str[i] = sep then
        result = append(result, trim(s))
        s = {}
          else
            s = s & str[i]
          end if
    end if
  end for
  result = append(result, trim(s))
  return result
end function

------------------------------------------------------------------------------

public function replaceString(sequence str, sequence old, sequence new)
-- replaces sub-sequence old by new in sequence str
  sequence s
  integer n, lg

  s = str
  lg = length(old)
  n = -1
  while n != 0 do
    n = match(old, s)
    if n > 1 then
      s = s[1..n-1]&new&s[n+lg..$]
    elsif n = 1 then
      s = new&s[n+lg..$]
    end if
  end while
  return s
end function

------------------------------------------------------------------------------

public function strtok(sequence str, atom sep)
-- splits str string in tokens according to separator sep
-- converts segments to atoms when needed
  sequence s

  s = split(str, sep)
  for i = 1 to length(s) do
    if isNumber(s[i]) or isHex(s[i]) then
      s[i] = val(s[i])
    elsif string(s[i]) then
      s[i] = dequote(s[i])
    end if
  end for
  return s
end function

------------------------------------------------------------------------------

public procedure hexDump(sequence s, integer output=1)
-- dumps a sequence in hex form to the specified output
  puts(output, "\nAddr  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F\t0123456789ABCDEF\n")
  sequence address = "0000 "
  sequence binary = ""
  sequence ascii  = "\t"
  integer lg = length(s)
  for i = 1 to lg do
    binary &= sprintf(" %02x", s[i])
    if (s[i] > 31) and (s[i] < 127) then
      ascii &= s[i]
    elsif s[i] = 9 then
      ascii &= "\\t"
    elsif s[i] = 13 then
      ascii &= "\\r"
    elsif s[i] = 10 then
      ascii &= "\\n"
    else
      ascii &= "."
    end if
    if (remainder(i-1, 16) = 15) then
      puts(output, address&binary&ascii&"\n")
      address = sprintf("%04x ", i)
      binary = ""
      ascii  = "\t"
    end if
  end for
  if length(ascii) then
    puts(output, address&pad_tail(binary,48)&ascii&"\n")
  else
    puts(output, "\n")
  end if
end procedure

------------------------------------------------------------------------------

public function appendIfNew(sequence s, object o)
  if find(o, s) then
    return s
  else
    return append(s, o)
  end if
end function

-------------------------------------------------------------------------------

public function nbOccurences(sequence s, sequence item)
  integer nb, pos, lg

  nb = 0
  pos = 1
  lg = length(item)
  while pos > 0 do
    if match(item, s) then
      nb += 1
      s = remove(s, 1, pos+lg-1)
    end if
  end while
  return nb
end function

