include common.e
include myFile.e
include wxeud.e

public constant
  wxSTRING=0, wxINTEGER=1, wxFLOAT=2, wxBOOL=3

constant
  FIELD_NAME=1, FIELD_SIZE=2, FIELD_ALIGNMENT=3, FIELD_FORMAT=4, FIELD_WIDTH=5, FIELD_PRECISION=6

-- procedure setupGrid(integer grid, sequence fields)
-- procedure cleanGrid(integer grid)
-- procedure loadGrid(atom grid, sequence filename)
-- procedure saveGrid(atom grid, sequence filename)
-- procedure setEditText(atom grid)

------------------------------------------------------------------------------

global procedure setupGrid(integer grid, sequence fields)
  integer lg = length(fields)
  set_grid_size ( grid, 1, lg )
  for i = 1 to lg  do
    set_col_label( grid, i-1, fields[i][FIELD_NAME] )
    set_col_size( grid, i-1, fields[i][FIELD_SIZE] )
    set_col_alignment( grid, i-1, fields[i][FIELD_ALIGNMENT][1], fields[i][FIELD_ALIGNMENT][2] )
    switch fields[i][FIELD_FORMAT] do
      case wxBOOL then set_col_format_bool( grid, i-1 )
      case wxFLOAT then set_col_format_float( grid, i-1, fields[i][FIELD_WIDTH], fields[i][FIELD_PRECISION] )
      case wxINTEGER then set_col_format_number( grid, i-1 )
    end switch
  end for
  force_refresh_grid ( grid )
end procedure

------------------------------------------------------------------------------

global procedure cleanGrid(integer grid)
  integer lg = get_number_cols ( grid )
  sequence colLabel = {}
  for i = 1 to lg  do
    colLabel = append( colLabel, get_col_label(grid, i-1) )
  end for
  clear_grid ( grid )
  set_grid_size ( grid, 1, lg )
  for i = 1 to lg  do
    set_col_label( grid, i-1, colLabel[i] )
  end for
end procedure

------------------------------------------------------------------------------

global procedure loadGrid(atom grid, sequence filename)
  sequence s = readCSV(filename)
--  analyzeSequence(s, "s", f_debug)
  integer nbRows = length(s)
  integer nbCols = length(s[1])
  set_grid_size ( grid, nbRows+2, nbCols )
  for c = 1 to nbCols do
    set_col_label ( grid, s[1][c], c-1 )
  end for
  for r = 2 to nbRows do
    for c = 1 to nbCols do
      set_cell_value ( grid, s[r][c], r-2, c-2 )
    end for
  end for
  force_refresh_grid ( grid )
  set_focus(grid)
end procedure

------------------------------------------------------------------------------

global procedure saveGrid(atom grid, sequence filename)
  integer nbRows = get_number_rows ( grid )
  integer nbCols = get_number_cols ( grid )
  sequence s = {}
  sequence colLabel = {}
  for i = 1 to nbCols  do
    colLabel = append( colLabel, get_col_label(grid, i-1) )
  end for
  s = append(s, colLabel)
  for r = 1 to nbRows do
    sequence t = {}
    for c = 1 to nbCols do
      t = append(t, get_cell_value ( grid, r-1, c-1 ) )
    end for
    s = append(s, t)
  end for
--  AnalyseSequence(s, "s", f_debug)
  saveCSV(filename, s)
end procedure

------------------------------------------------------------------------------

global procedure setEditText(atom grid)
  if grid_pos ( grid ) = get_number_rows ( grid ) then     -- clic en dernire ligne, on en rajoute une
    append_rows ( grid, 1 )
  end if
end procedure

