-- use objects as keys in a hash table

include std/machine.e
include std/convert.e

include stringhash.e as hash

-- Use a different header for each object type. This helps prevent
-- collisions of atoms and sequences, such as "ABC" and 4407873,
-- which are the same when passed through int_to_bytes().
constant OBJ_INT = #FC,
         OBJ_F32 = #FD,
         OBJ_F64 = #FE,
         OBJ_SEQ = #FF

-- range for 32-bit floating-point numbers
constant MIN_F32 = power( 10, 308 ) * (-1),
         MAX_F32 = power( 10, 308 )

-- make a valid key for each object type
function make_key( object key )

    if integer( key ) then
        key = OBJ_INT & int_to_bytes( key )
        
    elsif atom( key ) then
        if key <= MIN_F32 or key >= MAX_F32 then
            key = OBJ_F32 & atom_to_float32( key )
        else
            key = OBJ_F64 & atom_to_float64( key )
        end if
        
    else
        key = OBJ_SEQ & key
    end if
    
    return key
end function

-- wrap stringhash functions

global function new( integer seed )
    return hash:hash( seed )
end function

global function set( sequence table, object key, object data )
    return hash:set( table, make_key(key), data )
end function

global function get( sequence table, object key )
    return hash:get( table, make_key(key) )
end function

global function has( sequence table, object key )
    return hash:has( table, make_key(key) )
end function

