#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef EWINDOWS
#include <windows.h>
#define LIBMAIN int __stdcall LibMain(int hDLL, int Reason, void *Reserved)
 #else
#define LIBMAIN void __attribute__ ((constructor)) my_init(void)
#endif
#include "alldefs.h"


/* convert atom to char. *must avoid side effects in elem* */
#define Char(elem) ((IS_ATOM_INT(elem)) ? ((char)INT_VAL(elem)) : doChar(elem)) 

#ifdef EWINDOWS
int winInstance;
HANDLE default_heap;
#endif
static char doChar(object elem)
/* convert to char (int done in-line) */
{
    if (IS_ATOM_INT(elem)) 
	return (char)elem;
    if (IS_ATOM(elem)) 
	return (char)(DBL_PTR(elem)->dbl);
    else {
	//RTFatal("sequence found inside character string");
	return NULL;
    }
}

LIBMAIN
{
#ifdef EWINDOWS
    if (Reason == 1){
        winInstance = hDLL;

        default_heap = GetProcessHeap();
    }
    return 1;
#endif
}

#ifndef EWINDOWS
void __attribute__ ((destructor)) my_fini(void){}
#endif

object NewDouble(double d)
/* allocate space for a new double value */
{
	register d_ptr new_dptr;
	new_dptr = (d_ptr)malloc((long)D_SIZE);
	new_dptr->ref = 1;
	new_dptr->dbl = d;
	new_dptr->cleanup = 0;
	return MAKE_DBL(new_dptr);
}

s1_ptr NewS1(long size)
/* make a new s1 sequence block with a single reference count */
/* size is number of elements, NOVALUE is added as an end marker */
{
    register s1_ptr s1;

    s1 = (s1_ptr) malloc(sizeof(struct s1) + (size+1) * sizeof(object));
    s1->ref = 1;
    s1->base = (object_ptr)(s1 + 1);
    s1->length = size;
    s1->postfill = 0; /* there may be some available but don't waste time */
		      /* prepend assumes this is set to 0 */
	s1->cleanup = 0;
    *(s1->base + size) = NOVALUE;
    s1->base--;  // point to "0th" element
    return(s1);
}

object NewString(const unsigned char *s)
/* create a new string sequence */
{
    int len;
    object_ptr obj_ptr;
    s1_ptr c1;

    len = strlen((char *)s);
    c1 = NewS1((long)len);
    obj_ptr = (object_ptr)c1->base;
    if (len > 0) {
	do {
	    *(++obj_ptr) = (unsigned char)*s++;
	} while (--len > 0);
    }
    return MAKE_SEQ(c1);
}

object NewStringConst(const char *s)
/* create a new string sequence */
{
    int len;
    object_ptr obj_ptr;
    s1_ptr c1;

    len = strlen(s);
    c1 = NewS1((long)len);
    obj_ptr = (object_ptr)c1->base;
    if (len > 0) {
	do {
	    *(++obj_ptr) = (unsigned char)*s++;
	} while (--len > 0);
    }
    return MAKE_SEQ(c1);
}

void MakeCString(char *s, object obj)
/* make an atom or sequence into a C string */
/* N.B. caller must allow one extra for the null terminator */
{
    register object_ptr elem;
    object x;

    if (IS_ATOM(obj)) 
	*s++ = Char(obj);
    else {
	obj = (object)SEQ_PTR(obj);
	elem = ((s1_ptr)obj)->base;
	while (TRUE) { 
	    x = *(++elem);
	    if (IS_ATOM_INT(x)) {
		*s++ = (char)x;
	    }
	    else {
		if (x == NOVALUE)
		    break;
		*s++ = doChar(x);
	    }
	}
    }
    *s = '\0';
}

