// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "alldefs.h"
#include "wxeu.h"
#include "wx/config.h"


extern "C"{
object WXEUAPI new_wxConfig(object params)
{
	
	int len = LENGTH( params );
	wxString appName;
	wxString vendorName = wxEmptyString;
	wxString localFilename = wxEmptyString;
	wxString globalFilename = wxEmptyString;
	intptr_t style = 0;
	
	wxDeRefDS( params );
	
	switch (len)
	{
		case 5:
			style = (intptr_t)get_int( params, 5 );
		case 4:
			globalFilename = get_string( params, 4 );
		case 3:
			localFilename = get_string( params, 3 );
		case 2:
			vendorName = get_string( params, 2 );
		default:
			appName = get_string( params, 1 );
	}
	
	wxConfig * config = new wxConfig( appName, vendorName, localFilename, globalFilename, style );
	
	return BOX_INT(config);
	
}

void WXEUAPI config_delete_all(intptr_t config)
{
	((wxConfigBase*)config)->DeleteAll();
}

void WXEUAPI config_delete_entry(intptr_t config, object key, int delete_group_if_empty)
{
	wxDeRefDS( key );
	((wxConfigBase*)config)->DeleteEntry( get_string(key), (bool)delete_group_if_empty );
}

void WXEUAPI config_delete_group(intptr_t config, object key)
{
	wxDeRefDS( key );
	((wxConfigBase*)config)->DeleteGroup( get_string(key) );
}

void WXEUAPI config_exists(intptr_t config, object key)
{
	wxDeRefDS( key );
	((wxConfigBase*)config)->Exists( get_string(key) );
}

void WXEUAPI config_flush(intptr_t config, int current_only)
{
	((wxConfigBase*)config)->Flush( (bool)current_only );
}

object WXEUAPI get_config_app_name(intptr_t config)
{
	return get_sequence( ((wxConfigBase*)config)->GetAppName() );
}

object WXEUAPI get_first_config_group(intptr_t config, int index)
{
	wxString str;
	long dummy = (long)index;
	
	if ( !((wxConfigBase*)config)->GetFirstGroup(str, dummy) )
		str = wxEmptyString;
	
	return get_sequence( str );
}

object WXEUAPI get_first_config_entry(intptr_t config, int index)
{
	wxString str;
	long dummy = (long)index;
	
	if ( !((wxConfigBase*)config)->GetFirstEntry(str, dummy) )
		str = wxEmptyString;
	
	return get_sequence( str );
}

object WXEUAPI get_next_config_group(intptr_t config, int index)
{
	wxString str;
	long dummy = (long)index;
	
	if ( !((wxConfigBase*)config)->GetNextGroup(str, dummy) )
		str = wxEmptyString;
	
	return get_sequence( str );
}

object WXEUAPI get_next_config_entry(intptr_t config, int index)
{
	wxString str;
	long dummy = (long)index;
	
	if ( !((wxConfigBase*)config)->GetNextEntry(str, dummy) )
		str = wxEmptyString;
	
	return get_sequence( str );
}

object WXEUAPI config_read(intptr_t config, object key, object defaultValue)
{

	wxDeRefDS( key );
	
	if ( IS_DBL_OR_SEQUENCE(defaultValue) ) {
	
		wxDeRefDS( defaultValue );
		
		if ( IS_SEQUENCE(defaultValue) ) {
			wxString value;
			((wxConfigBase*)config)->Read( get_string(key), &value, get_string(defaultValue) );
			return get_sequence( value );
			
		} else {
			double value;
			((wxConfigBase*)config)->Read( get_string(key), &value, get_double(defaultValue) );
			return NewDouble( value );
			
		}
		
	} else {
		int value;
		((wxConfigBase*)config)->Read( get_string(key), &value, (int)defaultValue );
		return BOX_INT( value );
		
	}
	
}

void WXEUAPI config_write(intptr_t config, object key, object value)
{

	wxDeRefDS( key );
	
	if ( IS_DBL_OR_SEQUENCE(value) ) {
	
		wxDeRefDS( value );
		
		if ( IS_SEQUENCE(value) )
			((wxConfigBase*)config)->Write( get_string(key), get_string(value) );
			
		else
			((wxConfigBase*)config)->Write( get_string(key), get_double(value) );
		
	} else
		((wxConfigBase*)config)->Write( get_string(key), (int)value );
	
}

void WXEUAPI delete_config( object obj )
{
	DELETE_INSTANCE( wxConfig, obj );
}

}
