// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "alldefs.h"
#include "wxeu.h"
#include "wx/progdlg.h"
#include "wx/fontdlg.h"
#include "wx/dirdlg.h"
#include "wx/printdlg.h"
#include "wx/colordlg.h"
#include "wx/aboutdlg.h"
#include "wx/icon.h"
#include <wx/richtext/richtextformatdlg.h>
#include <wx/choicdlg.h>
#include <wx/textdlg.h>
#include <wx/popupwin.h>

extern "C"{
object WXEUAPI  show_modal(intptr_t dialog)
{

	wxObject * obj = (wxObject*) dialog;
	int show;
	if( obj->IsKindOf( CLASSINFO(wxPageSetupDialog) ) ){
		show = ((wxPageSetupDialog*)dialog)->ShowModal();
	}
	else if( obj->IsKindOf( CLASSINFO(wxPrintDialog) ) ){
		show = ((wxPrintDialog*)dialog)->ShowModal();
	}
	else if( obj->IsKindOf( CLASSINFO(wxMultiChoiceDialog) ) ){
		show = ((wxMultiChoiceDialog*)dialog)->ShowModal();
	}
	else if( obj->IsKindOf( CLASSINFO(wxTextEntryDialog) ) ){
		show = ((wxTextEntryDialog*)dialog)->ShowModal();
	
	}
	else{
		show = ((wxDialog*)dialog)->ShowModal();
	}
	return BOX_INT( show );
}

void WXEUAPI  end_modal(intptr_t dialog, int ret)
{
	((wxDialog*)dialog)->EndModal( ret );
}


class EuDialog : public wxDialog
{
public:
	EuDialog(const wxString& title)
       : wxDialog(NULL, wxID_ANY, title)
       {}
	public:
	void OnCloseWindow( wxCommandEvent & event )
	{
		EndModal(0);
	}
};

object WXEUAPI  new_wxDialog(object params)
{
	
	NEW_CONTROL_STYLE( wxDialog, dialog, wxDEFAULT_DIALOG_STYLE );

}

object WXEUAPI  new_wxProgressDialog(object params)
{
/*
--/li /b title
--/li /b message
--/li /b maximum [= 100]
--/li /b parent [= NULL]
--/li /b style [= wxPD_AUTO_HIDE | wxPD_APP_MODAL]
wxProgressDialog(const wxString& title, const wxString& message, 
int maximum = 100, wxWindow * parent = NULL, int style = wxPD_AUTO_HIDE | wxPD_APP_MODAL)
*/
	int len = ((s1_ptr)SEQ_PTR(params))->length;
	int style = wxPD_AUTO_HIDE | wxPD_APP_MODAL;
	int maximum = 100;
	wxWindow * parent = NULL;
	wxProgressDialog * pd;
	wxDeRefDS( params );
	switch( len ){
		case 5:
			style = get_int( params, 5 );
		case 4:
			parent = (wxWindow *) get_int( params, 4 );
		case 3:
			maximum = get_int( params, 3 );
		default:
			pd = new wxProgressDialog( get_string( params, 1 ),
				get_string( params, 2 ), maximum, parent, style );
			return BOX_INT( pd );
	}
	return 0;
}

object WXEUAPI  progress_update(intptr_t dialog, int val, object message)
{

	bool update = ((wxProgressDialog*)dialog)->Update( val, get_string( message ) );
	wxDeRefDS( message );
	return (object) update;
}

void WXEUAPI  progress_resume(intptr_t dialog)
{
	((wxProgressDialog*)dialog)->Resume();
}

void WXEUAPI  set_initial_font(intptr_t fdata, intptr_t font)
{
	((wxFontData*)fdata)->SetInitialFont( *(wxFont*)font );
}

object WXEUAPI  new_wxFontData(object params)
{
	wxDeRefDS( params );
	return BOX_INT( new wxFontData() );
}

object WXEUAPI  get_font_data(intptr_t dialog)
{
	wxFontData * data = new wxFontData ( ((wxFontDialog*)dialog)->GetFontData() );
	return BOX_INT( data );
}

object WXEUAPI  get_chosen_font(intptr_t fdata)
{
	wxFont* font = new wxFont( ((wxFontData*)fdata)->GetChosenFont() );
	return BOX_INT( font );
}

object WXEUAPI  font_selector(int parent, intptr_t initial)
{
	wxFontData data = wxFontData();
	
	if ( initial != 0 ){
		wxFont initial_font = *( (wxFont *) initial);
		data.SetInitialFont( initial_font );
	}
	
	wxFontDialog dialog( (wxWindow *) parent, data );
	bool result = dialog.ShowModal();
	if( result ){
		wxFont font = dialog.GetFontData().GetChosenFont();
		wxFont * f = new wxFont( font );
		return BOX_INT( f );
	}
	else{
		return 0;
	}
}

object WXEUAPI  new_wxDirDialog(object params)
{
/*
--/li /b parent 
--/li /b message [= "Choose a directory"]
--/li /b defaultPath [= ""]
--/li /b x [= -1]
--/li /b y [= -1]
--/li /b cx [= -1]
--/li /b cy [= -1]
--/li /b style [= 0]
wxDirDialog(wxWindow* parent, const wxString& message = "Choose a directory", 
const wxString& defaultPath = "", intptr_t style = wxDD_DEFAULT_STYLE, 
const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxDefaultSize, 
const wxString& name = "wxDirCtrl")
*/
	int len = ((s1_ptr)SEQ_PTR(params))->length;
	wxString message = wxString(wxT("Choose a directory"));
	wxString defaultPath = wxString(wxT(""));
	int style = wxDD_DEFAULT_STYLE;
	wxPoint pos = wxDefaultPosition;
	wxSize size = wxDefaultSize;

	switch( len ){
	    case 8:
			style = get_int( params, 8 );
		case 7:
			size = wxSize( get_int(params, 6), get_int(params, 7) );
		case 6:
		case 5:
			pos = wxPoint( get_int(params, 4), get_int(params, 5) );
		case 4:
		case 3:
			defaultPath = get_string( params, 3 );
		case 2:
			message = get_string( params, 2 );
		default:
			break;
	}
	wxDirDialog * dialog = new wxDirDialog( (wxWindow *) get_int( params, 1 ),
		message, defaultPath, style, pos, size );
	wxDeRefDS( params );
	return BOX_INT( dialog );
}

object WXEUAPI  get_dir_path(intptr_t dirdlg)
{
	wxString path = ((wxDirDialog *)dirdlg)->GetPath();
	return get_sequence( path );
}

object WXEUAPI  new_wxPageSetupDialog(object params)
{
/*
--/li /b parent [= 0]
--/li /wxPageSetupDialogData [= 0]
*/
	wxWindow * parent = NULL;
	wxPageSetupDialogData * data = NULL;
	int len = ((s1_ptr)SEQ_PTR(params))->length;

	switch( len ){
		case 2:
			data = (wxPageSetupDialogData *) get_int( params, 2 );
		case 1:
			parent = (wxWindow *) get_int( params, 1 );
		default:
			break;
	}
	wxPageSetupDialog * dialog = new wxPageSetupDialog( parent, data );
	wxDeRefDS( params );
	return BOX_INT( dialog );
}

object WXEUAPI  get_pagesetup_data(intptr_t ps)
{
	wxPageSetupDialog * dialog = (wxPageSetupDialog *) ps;
	wxPageSetupDialogData * data = new wxPageSetupDialogData( dialog->GetPageSetupData() );
	return BOX_INT( data );
}

object WXEUAPI  new_wxPrintDialog(object params)
{
	wxWindow * parent = NULL;
	wxPrintDialogData * data = NULL;
	int len = ((s1_ptr)SEQ_PTR(params))->length;

	switch( len ){
		case 2:
			data = (wxPrintDialogData *) get_int( params, 2 );
		case 1:
			parent = (wxWindow *) get_int( params, 1 );
		default:
			break;
	}
	wxPrintDialog * dialog = new wxPrintDialog( parent, data );
	wxDeRefDS( params );
	return BOX_INT( dialog );
}

object WXEUAPI  get_print_data(intptr_t prdlg)
{
	wxPrintDialog * dialog = (wxPrintDialog *) prdlg;
	wxPrintDialogData * data = new wxPrintDialogData( dialog->GetPrintDialogData() );
	return BOX_INT( data );
}

object WXEUAPI  new_wxColourDialog(object params)
{
	wxColourData * data;
	data = NULL;
	if( LENGTH(params) == 2 ) data = (wxColourData*) get_int( params, 2 );
	
	wxColourDialog * cd = new wxColourDialog( (wxWindow*)get_int( params, 1 ), data );
	wxDeRefDS( params );
	return BOX_INT( cd );
}

object WXEUAPI  get_color_data(intptr_t dialog)
{
	return BOX_INT( new wxColourData(((wxColourDialog*)dialog)->GetColourData()));
}

object WXEUAPI  get_color(intptr_t colordata)
{
	return BOX_INT( new wxColour( ((wxColourData*)colordata)->GetColour() ) );
}

object WXEUAPI  color_selector(intptr_t parent, intptr_t initial)
{
	wxColourData * data = new wxColourData();
	if( initial ){
		data->SetColour(*((wxColour *) initial) );
	}
	wxColourDialog dialog( (wxWindow *) parent, data );
	int result = dialog.ShowModal();
	delete data;
	if( result == wxID_OK ){
		wxColourData cd = dialog.GetColourData();
		wxColour * color = new wxColour( cd.GetColour() );
		return BOX_INT( color );
	}
	return 0;
}

object WXEUAPI new_wxRichTextFormattingDialog( object params )
{
	//NEW_CONTROL_STYLE( wxRichTextFormattingDialog, richformat, wxDEFAULT_DIALOG_STYLE );
// 	wxRichTextFormattingDialog(
// 	intptr_t flags, 
// 	wxWindow* parent, 
// 	const wxString& title = _("Formatting"),
// 	 wxWindowID id = wxID_ANY, 
// 	 const wxPoint& pos = wxDefaultPosition, 
// 	 const wxSize& sz = wxDefaultSize, 
// 	 intptr_t style = wxDEFAULT_DIALOG_STYLE)
	s1_ptr p = SEQ_PTR( params );
	wxString text(wxT("Formatting"));
	wxWindowID id = -1;
	wxPoint pos = wxDefaultPosition;
	wxSize size = wxDefaultSize;
	intptr_t style = wxDEFAULT_DIALOG_STYLE;
	switch( p->length ){
		case 10:
		case 9:
			style = get_int( params, 8 );
		case 8:
			size = wxSize( get_int( params, 6 ), get_int( params, 7 ) );
		case 7:
		case 6:
			pos = wxPoint( get_int( params, 4 ), get_int( params, 5) );
		case 5:
		case 4:
			id = get_int( params, 2 );
		case 3:
			text = get_string( params, 3 );
		default:
			break;
	}
	wxDeRefDS(params);
	wxRichTextFormattingDialog * name = new wxRichTextFormattingDialog( get_int(params, 1),
		(wxWindow *)get_int(params, 2),text, id, pos, size, style );
	return BOX_INT( name );
}

void WXEUAPI apply_rich_style_from_dialog( intptr_t dialog, intptr_t rich, int start_range, int end_range, int flags )
{
	((wxRichTextFormattingDialog*)dialog)->ApplyStyle( (wxRichTextCtrl*) rich, 
		wxRichTextRange( start_range, end_range ), flags );
}

object WXEUAPI get_rich_dialog_attributes( intptr_t dialog )
{
	return BOX_INT( new wxTextAttrEx( ((wxRichTextFormattingDialog*)dialog)->GetAttributes() ) );
}

object WXEUAPI get_rich_dialog_image_list( intptr_t dialog )
{
	return BOX_INT(  ((wxRichTextFormattingDialog*)dialog)->GetImageList() );
	
}

void WXEUAPI get_rich_dialog_style( intptr_t dialog, intptr_t rich, int start_range, int end_range )
{
	((wxRichTextFormattingDialog*)dialog)->GetStyle( (wxRichTextCtrl*)rich,
		wxRichTextRange( start_range, end_range ) );
}

object WXEUAPI  new_wxAboutDialogInfo(object params)
{
	wxDeRefDS( params );
	wxAboutDialogInfo * aboutInfo = new wxAboutDialogInfo();
	
	return BOX_INT( aboutInfo );
}

void WXEUAPI  about_dlg_add_artist(intptr_t info, object person)
{

	((wxAboutDialogInfo*)info)->AddArtist( get_string( person ) );
	wxDeRefDS( person );
}

void WXEUAPI  about_dlg_add_developer(intptr_t info, object person)
{

	((wxAboutDialogInfo*)info)->AddDeveloper( get_string( person ) );
	wxDeRefDS( person );
}

void WXEUAPI  about_dlg_add_doc_writer(intptr_t info, object person)
{

	((wxAboutDialogInfo*)info)->AddDocWriter( get_string( person ) );
	wxDeRefDS( person );
}

void WXEUAPI  about_dlg_add_translator(intptr_t info, object person)
{

	((wxAboutDialogInfo*)info)->AddTranslator( get_string( person ) );
	wxDeRefDS( person );
}


void WXEUAPI  about_dlg_set_copyright(intptr_t info, object copyright)
{

	((wxAboutDialogInfo*)info)->SetCopyright( get_string( copyright ) );
	wxDeRefDS( copyright );
}

void WXEUAPI  about_dlg_set_description(intptr_t info, object desc)
{

	((wxAboutDialogInfo*)info)->SetDescription( get_string( desc ) );
	wxDeRefDS( desc );
}

void WXEUAPI  about_dlg_set_icon(intptr_t info, int icon)
{
	wxIcon ico = *((wxIcon*)icon);

	((wxAboutDialogInfo*)info)->SetIcon( ico);
}


void WXEUAPI  about_dlg_set_license(intptr_t info, object license)
{

	((wxAboutDialogInfo*)info)->SetLicense( get_string( license ) );
	wxDeRefDS( license );
}

void WXEUAPI  about_dlg_set_name(intptr_t info, object name)
{

	((wxAboutDialogInfo*)info)->SetLicense( get_string( name ) );
	wxDeRefDS( name );
}

void WXEUAPI  about_dlg_set_version(intptr_t info, object version)
{

	((wxAboutDialogInfo*)info)->SetVersion( get_string( version ) );
	wxDeRefDS( version );
}

void WXEUAPI  about_dlg_set_web_site(intptr_t info, object params)
{
/*
--/li /b url
--/li /b desc = wxEmptyString
SetWebSite(const wxString& url, const wxString& desc = wxEmptyString)
*/
	int len = ((s1_ptr)SEQ_PTR(params))->length;
	wxString desc = wxEmptyString;

	if (len == 2) {
		desc = get_string( params, 2);
	} 

	((wxAboutDialogInfo*)info)->SetWebSite( get_string( params, 1), desc );
	wxDeRefDS( params );
}

void WXEUAPI  about_dlg_show(intptr_t info)
{
	wxAboutBox( *((wxAboutDialogInfo*)info) );
}

object WXEUAPI  new_wxMultiChoiceDialog(object params)
{
/*
wxMultiChoiceDialog(
--1  1 wxWindow* parent, 
--2  2 const wxString& message, 
--3  3 const wxString& caption, 
--   4 int n, 
--4  5 const wxString* choices,	
--5  6 intptr_t style = wxCHOICEDLG_STYLE, 
--67 7 const wxPoint& pos = wxDefaultPosition)
*/
	s1_ptr p = SEQ_PTR( params );
	wxPoint pos = wxDefaultPosition;
	intptr_t style = wxCHOICEDLG_STYLE;
	int n = 0;
	wxString * choices = NULL;
	object c;
	
	switch( p->length ){
		case 7:
			pos = wxPoint( get_int( params, 6 ), get_int( params, 7) );
			
		case 6:
		case 5:
			style = get_int( params, 5 );
		
		case 4:
		
			c = p->base[4];
			n = ((s1_ptr)SEQ_PTR(c))->length;
			choices = new wxString[n];
			
			for( int i = 0; i < n; i++ ){
				choices[i] = get_string( c, i + 1 );
			}

			
		default:
			break;
	}
	wxDeRefDS(params);
	
	wxMultiChoiceDialog * dlg = new wxMultiChoiceDialog( 
		(wxWindow *)get_int(params, 1), get_string( params, 2 ), get_string( params, 3 ), 
		n, choices, style, pos );
		
	return BOX_INT( dlg );
}

object WXEUAPI  new_wxSingleChoiceDialog(object params)
{
/*
wxSingleChoiceDialog(
--1  1 wxWindow* parent,
--2  2 const wxString& message,
--3  3 const wxString& caption,
--   4 int n,
--4  5 const wxString* choices,
--   6 void** clientData = NULL,
--5  7 intptr_t style = wxCHOICEDLG_STYLE,
--67 8 const wxPoint& pos = wxDefaultPosition)
*/
	s1_ptr p = SEQ_PTR( params );
	wxPoint pos = wxDefaultPosition;
	intptr_t style = wxCHOICEDLG_STYLE;
	int n = 0;
	wxString * choices = NULL;
	object c;
	
	switch( p->length ){
		case 7:
			pos = wxPoint( get_int( params, 6 ), get_int( params, 7) );
			
		case 6:
		case 5:
			style = get_int( params, 5 );
		
		case 4:
		
			c = p->base[4];
			n = ((s1_ptr)SEQ_PTR(c))->length;
			choices = new wxString[n];
			
			for( int i = 0; i < n; i++ ){
				choices[i] = get_string( c, i + 1 );
			}

			
		default:
			break;
	}
	wxDeRefDS(params);
	
	wxSingleChoiceDialog * dlg = new wxSingleChoiceDialog( 
		(wxWindow *)get_int(params, 1), get_string( params, 2 ), get_string( params, 3 ), 
		n, choices, NULL, style, pos );
		
	return BOX_INT( dlg );
}

object WXEUAPI  new_wxPasswordEntryDialog(object params)
{
/*
new_wxPasswordEntryDialog(
--1  1 wxWindow* parent, 
--2  2 const wxString& message, 
--3  3 const wxString& caption = "Enter password", 
--4  4 const wxString& defaultValue = "",	
--5  5 intptr_t style = wxOK | wxCANCEL | wxCENTRE, 
--67  6 const wxPoint& pos = wxDefaultPosition)
*/
	s1_ptr p = SEQ_PTR( params );
	wxPoint pos = wxDefaultPosition;
	intptr_t style = wxOK | wxCANCEL | wxCENTRE;
	wxString caption = wxString(wxT("Enter password"));
	wxString defaultValue = wxString(wxT(""));
	
	switch( p->length ){
		case 7:
			pos = wxPoint( get_int( params, 6 ), get_int( params, 7) );
			
		case 6:
		case 5:
			style = get_int( params, 5 );
		
		case 4:
			defaultValue = get_string(params, 4);

		case 3:
			caption = get_string(params, 3);
			
		default:
			break;
	}
	wxDeRefDS(params);
	
	wxPasswordEntryDialog * dlg = new wxPasswordEntryDialog( 
		(wxWindow *)get_int(params, 1), get_string( params, 2 ), caption, 
		defaultValue, style, pos );
		
	return BOX_INT( dlg );
}

object WXEUAPI  new_wxTextEntryDialog(object params)
{
/*
new wxTextEntryDialog(
--1  1 wxWindow* parent, 
--2  2 const wxString& message, 
--3  3 const wxString& caption = "Please enter text", 
--4  4 const wxString& defaultValue = "",	
--5  5 intptr_t style = wxOK | wxCANCEL | wxCENTRE, 
--67  6 const wxPoint& pos = wxDefaultPosition)
*/
	s1_ptr p = SEQ_PTR( params );
	wxPoint pos = wxDefaultPosition;
	intptr_t style = wxOK | wxCANCEL | wxCENTRE;
	wxString caption = wxString(wxT("Please enter text"));
	wxString defaultValue = wxString(wxT(""));
	
	switch( p->length ){
		case 7:
			pos = wxPoint( get_int( params, 6 ), get_int( params, 7) );
			
		case 6:
		case 5:
			style = get_int( params, 5 );
		
		case 4:
			defaultValue = get_string(params, 4);

		case 3:
			caption = get_string(params, 3);
			
		default:
			break;
	}
	wxDeRefDS(params);
	
	wxTextEntryDialog * dlg = new wxTextEntryDialog( 
		(wxWindow *)get_int(params, 1), get_string( params, 2 ), caption, 
		defaultValue, style, pos );
		
	return BOX_INT( dlg );
}

object WXEUAPI create_button_sizer( intptr_t dlg, int flags )
{
	return BOX_INT( ((wxDialog*)dlg)->CreateButtonSizer((intptr_t)flags) );
}

object WXEUAPI new_wxPopupWindow( object params )
{
	s1_ptr p = SEQ_PTR( params );
	wxPopupWindow * popup = NULL;
	wxDeRefDS( params );
	
	if ( p->length == 2 ) {
		popup = new wxPopupWindow(
			(wxWindow *)get_int(params, 1),
			get_int(params, 2) );
	} else {
		popup = new wxPopupWindow(
			(wxWindow *)get_int(params, 1) );
	}
	
	return BOX_INT( popup );
}

object WXEUAPI new_wxPopupTransientWindow( object params )
{
	s1_ptr p = SEQ_PTR( params );
	wxPopupWindow * popup = NULL;
	wxDeRefDS( params );
	
	if ( p->length == 2 ) {
		popup = new wxPopupTransientWindow(
			(wxWindow *)get_int(params, 1),
			get_int(params, 2) );
	} else {
		popup = new wxPopupTransientWindow(
			(wxWindow *)get_int(params, 1) );
	}
	
	return BOX_INT( popup );
}

void WXEUAPI position_popup( intptr_t popup, object origin, object size ){

	wxDeRefDS( origin );
	wxDeRefDS( size );
	
	((wxPopupWindow*)popup)->Position(
		wxPoint(get_int(origin, 1), get_int(origin, 2)),
		wxSize(get_int(size, 1), get_int(size, 2)) );

}

} // extern "C"
