// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "alldefs.h"
#include "wxeu.h"
#include "wx/dir.h"
#include "wx/dirctrl.h"

extern "C"{
object WXEUAPI  new_wxDir(object params)
{

	wxDir * dir  = new wxDir( get_string( params ) );
	wxDeRefDS( params );
	return BOX_INT( dir );
}

object WXEUAPI  is_opened(intptr_t dir)
{
	return ((wxDir*)dir)->IsOpened();
}

class euDirTraverser : public wxDirTraverser{

    public:
        euDirTraverser(wxArrayString& files) : m_files(files) { }

        virtual wxDirTraverseResult OnFile(const wxString& filename)
        {
            m_files.Add(filename);
            return wxDIR_CONTINUE;
        }

        virtual wxDirTraverseResult OnDir(const wxString& WXUNUSED(dirname))
        {
            return wxDIR_CONTINUE;
        }
		
		wxArrayString& getFiles(){ return m_files; };
    private:
        wxArrayString& m_files;

};

object WXEUAPI  new_wxDirTraverser(object params)
{
	euDirTraverser * edt = new euDirTraverser( *((wxArrayString *) get_int( params, 1)) );
	wxDeRefDS( params );
	return BOX_INT( edt );
}

object WXEUAPI  get_all_files(object dirname, object filespec, int flags, intptr_t traverser)
{
	wxDir * dir;
	euDirTraverser * dt;
	wxArrayString files;
	
	if( IS_SEQUENCE(dirname) ) dir = new wxDir( get_string( dirname ) );
	else  dir = (wxDir *) dirname;
	
	wxString fspec = get_string( filespec );
	
	if( dir->IsOpened() ){
		if( traverser ){
			dt = (euDirTraverser *) traverser;
			files = dt->getFiles();
		}
		else{
			dt = new euDirTraverser(files);
		}
		
	}
	else{
		return MAKE_SEQ(NewS1(0));
	}
	
	intptr_t count = (intptr_t) dir->Traverse(*dt, fspec, flags);
	s1_ptr s = NewS1( count );
	for( int i = 0; i < count; i++){
		s->base[i+1] = get_sequence( files[i] );
	}
	
	if( !traverser ) delete dt;
	if(!IS_SEQUENCE(dirname)) delete dir;
	wxDeRef( dirname );
	wxDeRefDS( filespec );
	return MAKE_SEQ( s );
}

object WXEUAPI  has_files(intptr_t dir, object spec)
{
	wxString filespec = get_string( spec );
	wxDeRefDS( spec );
	return ((wxDir*)dir)->HasFiles( filespec );
}

object WXEUAPI  has_subdirs(intptr_t dir, object spec)
{
	wxString filespec = get_string( spec );
	wxDeRefDS( spec );
	return ((wxDir*)dir)->HasSubDirs( filespec );
}

object WXEUAPI  open_dir(intptr_t dir, object dname)
{
	wxString dirname = get_string( dname );
	wxDeRefDS( dname );
	return ((wxDir*)dir)->Open( dirname );
}

object WXEUAPI  dir_exists(object directory)
{
	wxString dirname = get_string( directory );
	wxDeRefDS( directory );
	return wxDir::Exists( dirname );
}

object WXEUAPI  new_wxDirCtrl(object params)
{
	NEW_CONTROL_STYLE( wxGenericDirCtrl, dir, wxDIRCTRL_3D_INTERNAL | wxSUNKEN_BORDER );
}

void WXEUAPI  collapse_dirctrl_tree(intptr_t ctrl)
{
	((wxGenericDirCtrl*)ctrl)->CollapseTree();
}

object WXEUAPI  expand_dirctrl_path(intptr_t ctrl, object path)
{
	wxDeRefDS( path );
	return ((wxGenericDirCtrl*)ctrl)->ExpandPath( get_string(path) );
}

object WXEUAPI  collapse_dirctrl_path(intptr_t ctrl, object path)
{
	wxDeRefDS( path );
	return ((wxGenericDirCtrl*)ctrl)->CollapsePath( get_string(path) );
}

object WXEUAPI  get_dirctrl_default_path(intptr_t ctrl)
{
	return get_sequence( ((wxGenericDirCtrl*)ctrl)->GetDefaultPath() );
}

object WXEUAPI  get_dirctrl_path(intptr_t ctrl)
{
	return get_sequence( ((wxGenericDirCtrl*)ctrl)->GetPath() );
}

object WXEUAPI  get_dirctrl_file_path(intptr_t ctrl)
{
	return get_sequence( ((wxGenericDirCtrl*)ctrl)->GetFilePath() );
}

object WXEUAPI  get_dirctrl_filter(intptr_t ctrl)
{
	return get_sequence( ((wxGenericDirCtrl*)ctrl)->GetFilter() );
}

object WXEUAPI  get_dirctrl_filter_index(intptr_t ctrl)
{
	return ((wxGenericDirCtrl*)ctrl)->GetFilterIndex();
}

object WXEUAPI  get_dirctrl_filter_listctrl(intptr_t ctrl)
{
	return BOX_INT( ((wxGenericDirCtrl*)ctrl)->GetFilterListCtrl() );
}

object WXEUAPI  get_dirctrl_root_id(intptr_t ctrl)
{
	return BOX_INT( ((wxGenericDirCtrl*)ctrl)->GetRootId() );
}

object WXEUAPI  get_dirctrl_treectrl(intptr_t ctrl)
{
	return BOX_INT( ((wxGenericDirCtrl*)ctrl)->GetTreeCtrl() );
}

void WXEUAPI  recreate_dirctrl_tree(intptr_t ctrl)
{
	((wxGenericDirCtrl*)ctrl)->ReCreateTree();
}

void WXEUAPI  set_dirctrl_default_path(intptr_t ctrl, object path)
{
	((wxGenericDirCtrl*)ctrl)->SetDefaultPath( get_string(path) );
	wxDeRefDS( path );
}

void WXEUAPI  set_dirctrl_filter(intptr_t ctrl, object filter)
{
	((wxGenericDirCtrl*)ctrl)->SetFilter( get_string(filter) );
	wxDeRefDS( filter );
}

void WXEUAPI  set_dirctrl_filter_index(intptr_t ctrl, int index)
{
	((wxGenericDirCtrl*)ctrl)->SetFilterIndex( index );
}

void WXEUAPI  set_dirctrl_path(intptr_t ctrl, object path)
{
	((wxGenericDirCtrl*)ctrl)->SetPath( get_string(path) );
	wxDeRefDS( path );
}

void WXEUAPI  show_dirctrl_hidden_files(intptr_t ctrl, int show)
{
	((wxGenericDirCtrl*)ctrl)->ShowHidden( (bool)show );
}

}
