// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "wx/image.h"
#include "wx/frame.h"
#include "alldefs.h"
#include "wxeu.h"
#include "wx/dcclient.h"
#include "wx/listctrl.h"
#include "wx/treectrl.h"
#include "wx/tooltip.h"
#include "wx/utils.h"
#include "wx/window.h"
#include "wx/socket.h"

#include <map>
using namespace std;
EuApp* EuApp::wxEuApp;
object EuApp::the_object;
EuCallFunc EuApp::callFunc;
EuCallProc EuApp::callProc;
intptr_t EuApp::error_rid = -1;

void EuApp::proc( intptr_t rid, object obj )
{
	the_object = obj;
	callProc( rid );
}

void EuApp::RTFatal( const char *msg )
{
	unsigned char *emsg;
	emsg = (unsigned char*) malloc( strlen( msg ) + 2);
	strncpy( (char*)emsg, msg, strlen( msg ) + 1 );
	
	object obj_msg = NewString( emsg );
	s1_ptr s = NewS1( LENGTH( obj_msg ) );
	s->base[1] = obj_msg;
	the_object = MAKE_SEQ( s );
	
	callProc( error_rid );
}

object EuApp::func( intptr_t rid, object obj )
{
	the_object = obj;
	return callFunc( obj );
}

extern "C"{

void WXEUAPI delete_instance( object obj )
{
	DELETE_INSTANCE( wxObject, obj )
}

object WXEUAPI get_event_type( wxEvent & event )
{
	return BOX_INT( event.GetEventType() );
}

object WXEUAPI get_event_id( wxEvent & event )
{
	return BOX_INT( event.GetId() );
}

object WXEUAPI get_event_object( intptr_t event )
{
	return BOX_INT(((wxEvent*)event)->GetEventObject());
}

static void unbox_int1( intptr_t x )
{
	if(IS_DBL_OR_SEQUENCE(x)){
		free( DBL_PTR( x ) );
	}
}

struct wxObjRid{
	wxObjRid(){};
	wxObjRid( intptr_t _rid, intptr_t _event_object ):rid(_rid),event_object(_event_object){};
	intptr_t rid;
	intptr_t event_object;
};


typedef map<intptr_t,wxObjRid> wxId2RidObj;
typedef map<intptr_t,wxId2RidObj> wxEvtType2Id;

class wxEuClientData{
	// used for event handling
	// event_type->id->(handler,rid)
	public:
		wxEuClientData(){};
		virtual ~wxEuClientData(){};
		void add( intptr_t handler, intptr_t event_type, intptr_t id, intptr_t rid );
		void get( intptr_t event_type, intptr_t id, intptr_t* handler, intptr_t* rid );
	private:
		wxEvtType2Id m_evtType2Id;
};

void wxEuClientData::get( intptr_t event_type, intptr_t id, intptr_t* handler, intptr_t* rid )
{
	wxEvtType2Id::iterator evIt = m_evtType2Id.find( event_type );
	if( evIt == m_evtType2Id.end() ){
		*rid = -1;
	}
	else{
		wxId2RidObj::iterator rIt = evIt->second.find(id);
		if( rIt == evIt->second.end() ) rIt = evIt->second.find(-1); // try the generic id
		if( rIt == evIt->second.end() ){
			*rid = -1;
		}
		else{
			*handler = rIt->second.event_object;
			*rid = rIt->second.rid;
		}
	}
}

void wxEuClientData::add( intptr_t handler, intptr_t event_type, intptr_t id, intptr_t rid )
{
	m_evtType2Id[event_type][id] = wxObjRid(rid,handler);
}

class EuEvent : public wxEvtHandler{
	public:
	
	void Handler( wxEvent & event )
	{
		
		wxEvtHandler * evtObj = 0;
		wxEuClientData * euClient = 0;
		
		intptr_t eventType;
		if( event.GetEventType() == wxEVT_SOCKET ){
			evtObj = this;
			wxSocketEvent* se = (wxSocketEvent*)( &event );
			eventType = se->GetSocketEvent();
		}
		else{
			evtObj = (wxEvtHandler*) event.GetEventObject();
			eventType = event.GetEventType();
		}
		euClient = (wxEuClientData*) evtObj->GetClientData();
		if( euClient == 0 ){
			/*
			 *  wxWidgets sets the wxTimer as the eventObject regardless of calls to
			 *  wxTimer::SetOwner().  So we'll go through the owner's client data 
			 *  the first time, and set up a wxEuClientData on the wxTimer for
			 *  subsequent events to avoid the extra checks later.
			 */
			if( evtObj->IsKindOf(CLASSINFO(wxTimer)) ){
				wxEuClientData * timerClient = (wxEuClientData*) ((wxTimer*)evtObj)->GetOwner()->GetClientData();
				if( timerClient != 0 ){
					intptr_t handler, rid;
					timerClient->get( event.GetEventType(), event.GetId(), &handler, &rid );
					euClient = new wxEuClientData();
					euClient->add( handler, event.GetEventType(), event.GetId(), rid );
					evtObj->SetClientData( (void*)euClient );
				}
			}
			
			if( euClient == 0 ){
				return;
			}
		}
		intptr_t handler, rid;
		euClient->get( event.GetEventType(), event.GetId(), &handler, &rid );
		if( rid == -1 ){
			return;
		}
		s1_ptr s = NewS1( 4 );
		s->base[1] = BOX_INT( handler );
		s->base[2] = BOX_INT( eventType );
		s->base[3] = BOX_INT( event.GetId() );
		s->base[4] = BOX_INT( &event );
		
		EuApp::wxEuApp->proc( rid, MAKE_SEQ( s ) );
	}

};


void WXEUAPI  set_event_handler(intptr_t handler,intptr_t id, wxEventType eventType, intptr_t rid )
{
	wxEvtHandler * evtHandler = (wxEvtHandler*) handler;

	if( evtHandler->IsKindOf(CLASSINFO(wxMenu)) ){
		// Menus in wxGTK that are on a wxMenuBar need to have the parent window
		// handle its events.
		wxMenuBar * bar = ((wxMenu*)evtHandler)->GetMenuBar();
		wxEvtHandler * parent = 0;
		if( bar != 0 ){
			parent = bar->GetFrame();
		}
		evtHandler = (parent ? parent : (bar ? bar : evtHandler));
		
		wxEuClientData* euClient = (wxEuClientData*) ((wxEvtHandler*)handler)->GetClientData();
		if( euClient == 0 ){
			euClient = new wxEuClientData();
			((wxEvtHandler*)handler)->SetClientData( euClient );
		}
		euClient->add( handler, eventType, id, rid );
	}
	
	wxEuClientData* euClient = (wxEuClientData*) evtHandler->GetClientData();
	if( euClient == 0 ){
		euClient = new wxEuClientData();
		evtHandler->SetClientData( euClient );
	}
	euClient->add( handler, eventType, id, rid );
	evtHandler->Connect(id, eventType,  wxEventHandler(EuEvent::Handler) );
}


object WXEUAPI  wx_or_all(object a)
{
	s1_ptr s = SEQ_PTR( a );
	int val = 0;
	for( int i = 1; i <= s->length; i++)
		val |= s->base[i];
	wxDeRefDS( a );
	return val;
}

object WXEUAPI  get_sys_metric(int metric)
{
//return call_cdecl( wxSystemSettingsNative_GetMetric, {metric})
	return wxSystemSettings::GetMetric( (wxSystemMetric) metric );
}


object WXEUAPI  new_id()
{
	intptr_t id = wxNewId();
	return BOX_INT( id );
}

void WXEUAPI  set_back_color(intptr_t window, intptr_t color)
{
	((wxWindow*)window)->SetBackgroundColour( *((wxColour*)color) );
}

void WXEUAPI  set_fore_color(intptr_t window, intptr_t color)
{
	((wxWindow*)window)->SetForegroundColour( *((wxColour*)color) );
}

object WXEUAPI  get_id(intptr_t win)
{
	return BOX_INT( ((wxWindow*)win)->GetId());
}

void WXEUAPI  enable(object control, intptr_t on )
{
	((wxWindow*)control)->Enable( (bool) on );
}

void WXEUAPI  show_fullscreen(intptr_t frame, intptr_t show, intptr_t flags)
{
	((wxFrame*)frame)->ShowFullScreen( show, flags );
}

void WXEUAPI  show_window(intptr_t window, intptr_t show)
{
	((wxWindow*)window)->Show( (bool) show );
}

object WXEUAPI  is_visible(intptr_t window)
{
	return (object) ((wxWindow*)window)->IsShown();
}

void WXEUAPI  destroy(intptr_t window)
{
	((wxWindow*)window)->Destroy();
}

void WXEUAPI  raise_window(intptr_t window)
{
	((wxWindow*)window)->Raise();
}

void WXEUAPI  exit_main()
{
	EuApp::wxEuApp->ExitMainLoop();
}


object WXEUAPI  message_box(object s1, object s2, intptr_t ok)
{
	intptr_t ret = wxMessageBox( get_string(s1), get_string(s2), ok );

	wxDeRefDS( s1 );
	wxDeRefDS( s2 );
	return ret;
}

object WXEUAPI  mouse_event_shiftdown(intptr_t event)
{
	return (object) ((wxMouseEvent*)event)->ShiftDown();
}

object WXEUAPI  mouse_event_position(intptr_t event)
{
	s1_ptr pos = NewS1(2);
	wxMouseEvent * e = (wxMouseEvent*) event;
	pos->base[1] = e->GetX();
	pos->base[2] = e->GetY();
	return MAKE_SEQ( pos );
}

void WXEUAPI  release_mouse(intptr_t window)
{
	((wxWindow*)window)->ReleaseMouse();
}

void WXEUAPI  capture_mouse(intptr_t window)
{
	((wxWindow*)window)->CaptureMouse();
}

void WXEUAPI  skip(intptr_t event)
{
	((wxEvent*)event)->Skip();
}


class EuFrame : public wxFrame
{
public:
	EuFrame(const wxString& title)
       : wxFrame(NULL, wxID_ANY, title)
       {}
	public:
	void OnCloseWindow( wxCommandEvent & event )
	{
		if( this == EuApp::wxEuApp->GetTopWindow() )
		{
				EuApp::wxEuApp->ExitMainLoop();
		}
		else{
			Show( false );
		}
	}
};

object WXEUAPI new_wxFrame(object params)
{
	s1_ptr p = SEQ_PTR(params);
	intptr_t len = p->length;
	wxWindow * parent = (wxWindow *) get_int( params, 1 );
	intptr_t id = get_int( params, 2 );
	wxString caption = get_string( params, 3 );
	wxFrame *frame;
	do {
	if( len == 3 ){
		frame = new wxFrame();
		frame->Create( parent, id, caption );
		break;
	}
	wxPoint xy( get_int( params, 4 ), get_int( params, 5 ) );

	if( len == 5 ) {
		frame = new wxFrame( parent, id, caption, xy ) ;
		break;
	}	
	
	wxSize size( get_int( params, 6 ), get_int( params, 7 ) );

	if( len == 7 ) {
		frame = new wxFrame( parent, id, caption, xy, size );
		break;
	}

	frame = new wxFrame( parent, id, caption, xy, size, get_int( params, 8 ) );

	} while (0);

	frame->Connect(frame->GetId(), frame->GetId(), wxEVT_CLOSE_WINDOW, 
			wxCommandEventHandler(EuFrame::OnCloseWindow ));
	wxDeRefDS( params );
	return BOX_INT( frame );
}

void WXEUAPI close_window( intptr_t win, intptr_t force )
{
	((wxWindow*)win)->Close( force );
}

object WXEUAPI can_veto( intptr_t event )
{
	return 1 & ((wxCloseEvent*)event)->CanVeto();
}

object WXEUAPI  new_wxWindow(object params)
{
	NEW_SIMPLE_CONTROL( wxWindow );
}

object WXEUAPI  new_wxPanel(object params)
{
	NEW_SIMPLE_CONTROL_STYLE( wxPanel, wxTAB_TRAVERSAL );
}

object WXEUAPI  get_handle(intptr_t win)
{
	intptr_t handle = (intptr_t) ((wxWindow*)win)->GetHandle();
	return BOX_INT( handle );
}

void WXEUAPI  set_tip(intptr_t win, object tip)
{
	((wxWindow*)win)->SetToolTip( get_string( tip ) );
	wxDeRefDS( tip );
}

void WXEUAPI  set_title(intptr_t window, object title)
{
	wxObject * obj = (wxObject*)window;
	if( obj->IsKindOf( CLASSINFO( wxFrame ) )){
		((wxFrame*)window)->SetTitle( get_string( title ) );
	}
	else{
		((wxWindow*)window)->SetLabel( get_string( title ) );
	}
	
	wxDeRefDS( title );
}

object WXEUAPI  get_label(intptr_t control)
{
	if( ((wxObject*)control)->IsKindOf( CLASSINFO(wxStaticText) ) ){
		return get_sequence( ((wxStaticText*) control)->GetLabel() );
	}
	return get_sequence( ((wxWindow*) control)->GetLabel() );
}

void WXEUAPI  refresh_window(object params)
{
	if( IS_SEQUENCE(params) ){
		s1_ptr p = SEQ_PTR( params );
		if( p->length == 6 ){
			wxRect rect( get_int( params, 3 ), get_int( params, 4 ),
					get_int(params, 5), get_int(params, 6));
			((wxWindow *) get_int(params, 1))->RefreshRect( rect, (bool)get_int(params, 1));
		}
		else{
			((wxWindow *) get_int(params, 1))->Refresh( (bool)get_int(params, 2));
		}
	}
	else{
		((wxWindow *) get_int( params ))->Refresh();

	}
	wxDeRef( params );
}

void WXEUAPI  set_icon(intptr_t frame, intptr_t icon)
{
	((wxTopLevelWindow *)frame)->SetIcon( *(wxIcon *)icon );
}

void WXEUAPI  set_default_item(intptr_t panel, intptr_t button)
{
	((wxTopLevelWindow*)panel)->SetDefaultItem( (wxButton*)button);
}

void WXEUAPI  request_user_attention( intptr_t frame, intptr_t flags )
{
	((wxTopLevelWindow*)frame)->RequestUserAttention( flags );
}

object WXEUAPI  get_client_size(intptr_t win )
{
	wxSize size = ((wxWindow*)win)->GetClientSize();
	s1_ptr seq = NewS1( 2 );
	seq->base[1] = size.GetWidth();
	seq->base[2] = size.GetHeight();
	
	return MAKE_SEQ( seq );
}

void WXEUAPI  set_size(intptr_t win, intptr_t cx, intptr_t cy)
{
	((wxWindow*)win)->SetSize( cx, cy );
}

void WXEUAPI  move(intptr_t win, intptr_t x, intptr_t y)
{
	((wxWindow*)win)->Move( x, y );
}

void WXEUAPI  set_focus(intptr_t window)
{
	if( ((wxWindow*)window)->AcceptsFocus() )
		((wxWindow*)window)->SetFocus();
}

object WXEUAPI  get_text_extent(intptr_t win, object text, intptr_t font)
{
	int x = 0;
	int y = 0;
	
	((wxWindow *)win)->GetTextExtent( get_string( text ), &x, &y, NULL, NULL,
		(wxFont *) font);
	
	s1_ptr ret = NewS1(2);
	ret->base[1] = x;
	ret->base[2] = y;
	wxDeRefDS( text )
	return MAKE_SEQ(ret);
}

object WXEUAPI  get_key_code(intptr_t event)
{
	return ((wxKeyEvent *) event)->GetKeyCode();
}

object WXEUAPI get_unicode_key( intptr_t event )
{
#if wxUSE_UNICODE
	return BOX_INT( reinterpret_cast<wxKeyEvent*>( event )->GetUnicodeKey() );
#else
	return get_key_code( event );
#endif
}

object WXEUAPI  key_event_shiftdown(intptr_t event)
{
	return ((wxKeyEvent *)event)->ShiftDown();
}

object WXEUAPI  key_event_controldown(intptr_t event)
{
	return ((wxKeyEvent *)event)->ControlDown();
}

object WXEUAPI  key_event_altdown(intptr_t event)
{
	return ((wxKeyEvent *)event)->AltDown();
}

object WXEUAPI  get_key_state(intptr_t keycode)
{
	return wxGetKeyState( (wxKeyCode) keycode );
}

object WXEUAPI  get_rect(intptr_t win)
{
	s1_ptr rect = NewS1(4);
	wxPoint pos = ((wxWindow*)win)->GetPosition();
	wxSize size = ((wxWindow*)win)->GetSize();
	
	rect->base[1] = pos.x;
	rect->base[2] = pos.y;
	rect->base[3] = size.GetWidth();
	rect->base[4] = size.GetHeight();
	
	return MAKE_SEQ( rect );
}

void WXEUAPI  set_cursor(intptr_t window, intptr_t cursor)
{
	((wxWindow*)window)->SetCursor( *((wxCursor*)cursor) );
}

object WXEUAPI  new_wxCursor(object params)
{
	intptr_t c = get_int( params, 1 );
	wxCursor * cursor;
	if( c < 26 ){
		cursor = new wxCursor( c );
	}
	else{
		cursor = new wxCursor( *((wxImage*)c) );
	}

	return BOX_INT( cursor );
}

object WXEUAPI  copy_file(object file1, object file2)
{
	object ret = wxCopyFile( get_string( file1 ), get_string( file2 ) );

	wxDeRefDS( file1 );
	wxDeRefDS( file2 );
	return ret;
}

object WXEUAPI  remove_file(object file)
{
	object ret = wxRemoveFile( get_string( file ) );
	wxDeRefDS( file );
	return ret;
}

object WXEUAPI  file_selector(intptr_t parent, object message, object info, intptr_t flag)
{
	int len = LENGTH( info );
	wxString path(wxT(""));
	wxString filename(wxT(""));
	wxString extension(wxT(""));
	wxString wildcard(wxT(""));
	switch( len ){
		case 4:
			wildcard = get_string( info, 4 );
		case 3:
			extension = get_string( info, 3 );
		case 2:
			extension = get_string( info, 2 );
		case 1:
			path = get_string( info, 1 );
		default:
			break;
	}
	wxString file = wxFileSelector( get_string( message ), path, filename, extension,
		wildcard, flag, (wxWindow*)parent);

	wxDeRefDS( message );
	wxDeRefDS( info );
	return get_sequence( file );
}

void WXEUAPI  cleanup()
{
	EuApp::wxEuApp->CleanUp();
}

object WXEUAPI find_window(intptr_t id, intptr_t parent)
{
	intptr_t win = (intptr_t) ((wxWindow*)parent)->FindWindow( id );
	return BOX_INT(win);
}

void WXEUAPI  wxMain(intptr_t topWindow)
{
	EuApp::wxEuApp->SetTopWindow( (wxWindow *) topWindow );
	((wxFrame *)topWindow)->Show(true);
	EuApp::wxEuApp->MainLoop();
}

object WXEUAPI app_yield( intptr_t onlyIfNeeded ){
	return EuApp::wxEuApp->Yield( onlyIfNeeded );
}

void WXEUAPI set_app_name( object name )
{
	wxDeRefDS( name );
	EuApp::wxEuApp->SetAppName( get_string(name) );
}

object WXEUAPI get_app_name()
{
	return get_sequence( EuApp::wxEuApp->GetAppName() );
}

void WXEUAPI set_vendor_name( object name )
{
	wxDeRefDS( name );
	EuApp::wxEuApp->SetVendorName( get_string(name) );
}

object WXEUAPI get_vendor_name()
{
	return get_sequence( EuApp::wxEuApp->GetVendorName() );
}

object WXEUAPI is_active()
{
	return 1 & EuApp::wxEuApp->IsActive();
}

object WXEUAPI  initialize( EuCallFunc func, EuCallProc proc, intptr_t error )
{
		//wxEntry( NULL, NULL, NULL, 0 );
		wxInitialize();
		EuApp::wxEuApp = new EuApp();
		EuApp::SetInstance( EuApp::wxEuApp );
		int wx = 0;
		EuApp::wxEuApp->Initialize(wx,NULL);
		EuApp::the_object = 0;
		EuApp::callFunc = func;
		EuApp::callProc = proc;
		EuApp::error_rid = error;
		wxInitAllImageHandlers();
		return (object)WX_LIB_BUILD;
}

object WXEUAPI get_desktop_window()
{
#ifdef __WXMSW__
	return BOX_INT( GetDesktopWindow() );
#else
    return 0;
#endif
}

object WXEUAPI get_top_level_parent( intptr_t window )
{
	return BOX_INT( wxGetTopLevelParent( (wxWindow *)window ));
}

object WXEUAPI get_parent( intptr_t window )
{
	return BOX_INT( ((wxWindow*)window)->GetParent() );
}

object WXEUAPI get_children( intptr_t window )
{
	wxWindowList list = ((wxWindow*)window)->GetChildren();
	intptr_t count = list.GetCount();
	s1_ptr sel = NewS1( count );
	
	for( intptr_t i = 1; i <= count; i++ ){
		sel->base[i] = BOX_INT( list[i-1] );
	}
	
	return MAKE_SEQ( sel );
}

object WXEUAPI client_to_screen( intptr_t window, int x, int y )
{
	s1_ptr pt = NewS1( 2 );
	
	((wxWindow*)window)->ClientToScreen( &x, &y );
	pt->base[1] = x;
	pt->base[2] = y;
	
	return MAKE_SEQ( pt );
}

object WXEUAPI screen_to_client( intptr_t window, int x, int y )
{
	s1_ptr pt = NewS1( 2 );
	
	((wxWindow*)window)->ScreenToClient( &x, &y );
	pt->base[1] = x;
	pt->base[2] = y;
	
	return MAKE_SEQ( pt );
}

void WXEUAPI popup_menu( intptr_t window, intptr_t menu, intptr_t x, intptr_t y )
{
	((wxWindow*)window)->PopupMenu( (wxMenu*)menu, x, y );
}

object WXEUAPI hit_test( intptr_t window, int x, int y, int flags )
{
	wxObject * obj = (wxObject*)window;
	
	if ( obj->IsKindOf( CLASSINFO(wxListCtrl) ) ) {
		return BOX_INT( ((wxListCtrl*)window)->HitTest( wxPoint( x, y ), flags ) );
		
	} else if ( obj->IsKindOf( CLASSINFO(wxTreeCtrl) ) ) {
		return BOX_INT( ((wxTreeCtrl*)window)->HitTest( wxPoint( x, y ), flags ) );
		
	} else if ( obj->IsKindOf( CLASSINFO(wxTextCtrl) ) ) {
		wxTextCoord col, row;
		s1_ptr seq = NewS1( 2 );
		
		((wxTextCtrlBase*)window)->HitTest( wxPoint( x, y ), &col, &row );
		seq->base[1] = col;
		seq->base[2] = row;
		
		return MAKE_SEQ( seq );
		
	}
	
	return 0;
}

object WXEUAPI get_wheel_rotation( intptr_t event )
{
	return ((wxMouseEvent*)event)->GetWheelRotation();
}

object WXEUAPI get_the_object()
{
	return EuApp::the_object;
}

object WXEUAPI get_window_style( intptr_t win )
{
	return BOX_INT( ((wxWindow*)win)->GetWindowStyle() );
}

void WXEUAPI set_window_style( intptr_t win, intptr_t style )
{
	((wxWindow*)win)->SetWindowStyle( style );
}

void WXEUAPI center_on_parent( intptr_t win, intptr_t direction )
{
	((wxWindow*)win)->CentreOnParent(direction);
}

void WXEUAPI enable_tooltips( intptr_t enable )
{
	wxToolTip::Enable( enable );
}

void WXEUAPI set_tooltip_delay( intptr_t msec )
{
	wxToolTip::SetDelay( msec );
}

object WXEUAPI get_menu_bar(intptr_t frame)
{
	intptr_t id = (intptr_t) ((wxFrame*)frame)->GetMenuBar();
	return BOX_INT(id);
}

object WXEUAPI get_status_bar(intptr_t frame)
{
	intptr_t id = (intptr_t) ((wxFrame*)frame)->GetStatusBar();
	return BOX_INT(id);
}

object WXEUAPI get_status_bar_pane(intptr_t frame)
{
	intptr_t id = (intptr_t) ((wxFrame*)frame)->GetStatusBarPane();
	return BOX_INT(id);
}

object WXEUAPI get_tool_bar(intptr_t frame)
{
	intptr_t id = (intptr_t) ((wxFrame*)frame)->GetToolBar();
	return BOX_INT(id);
}

object WXEUAPI get_display_size()
{
	wxSize size = wxGetDisplaySize();
	s1_ptr s = NewS1( 2 );
	
	s->base[1] = size.x;
	s->base[2] = size.y;
	
	return MAKE_SEQ( s );
}

}

