// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "alldefs.h"
#include "wxeu.h"
#include <wx/filename.h>
#include <wx/mimetype.h>

extern "C"{
object WXEUAPI  new_wxFileName(object params)
{
/*
--/topic wxFileName
--/parent Classes
--/info
--include file: wxFile.e
--Creation parameters:
--<ul>
--/li /b path
--/li /b pathformat [=wxPATH_NATIVE]
--/li     wxPATH_NATIVE
--/li     wxPATH_UNIX
--/li     wxPATH_MAC
--/li     wxPATH_DOS
--/li     wxPATH_VMS
--/li     wxPATH_BEOS = wxPATH_UNIX,
--/li     wxPATH_WIN = wxPATH_DOS,
--/li     wxPATH_OS2 = wxPATH_DOS
--</ul></ul>
*/
	wxPathFormat format = wxPATH_NATIVE;
	wxFileName * fn;
	switch( LENGTH( params ) ){
		case 2:
			format = (wxPathFormat) get_int( params, 2 );
			
		default:
			fn = new wxFileName( get_string( params, 1 ), format );
			
	}
	wxDeRefDS( params );
	return BOX_INT( fn );
}

object WXEUAPI  wx_mkdir(object directory, object perm, int flags)
{
//static bool Mkdir(const wxString& dir, int perm = 0777, int flags = 0)
	int permission = 0;
	if( IS_SEQUENCE( perm ) ){
		permission = get_int( perm, 1 ) + get_int( perm, 2 ) * 8 + 
			get_int( perm, 3 ) * 64;
	}
	else{
		permission = get_int( perm );
	}
	
	bool ok = wxFileName::Mkdir( get_string( directory ), permission, flags );
	
	wxDeRef(perm);
	
	return ok;
}

object WXEUAPI  wx_rmdir( object directory )
{
	wxDeRefDS( directory );
	return wxFileName::Rmdir( get_string( directory ) );
}


#define WXFILENAME_CONVSEQ( in, out ) \
	wxFileName * out;\
	if ( IS_SEQUENCE(in) ) {\
		out = new wxFileName( get_string(in) );\
		wxDeRefDS( in );\
	} else {\
		out = (wxFileName*) in;\
	}

#define WXFILENAME_RETSEQ( fn, func ) \
	return get_sequence( fn->func() );

#define WXFILENAME_FUNC( filename, c_func )\
	WXFILENAME_CONVSEQ( filename, fn );\
	WXFILENAME_RETSEQ( fn, c_func );


object WXEUAPI  get_ext(object filename)
{
	WXFILENAME_FUNC( filename, GetExt );
}

object WXEUAPI  get_forbidden_chars(object filename)
{
	WXFILENAME_FUNC( filename, GetForbiddenChars );
}

object WXEUAPI  get_full_name(object filename)
{
	WXFILENAME_FUNC( filename, GetFullName );
}

object WXEUAPI  get_full_path(object filename)
{
WXFILENAME_FUNC( filename, GetFullPath );
}

object WXEUAPI  get_home_dir(object filename)
{
	WXFILENAME_FUNC( filename, GetHomeDir );
}

object WXEUAPI  get_long_path(object filename)
{
	WXFILENAME_FUNC( filename, GetLongPath );
}

object WXEUAPI  get_name(object filename)
{
	WXFILENAME_FUNC( filename, GetName );
}

object WXEUAPI  get_short_path(object filename)
{
	WXFILENAME_FUNC( filename, GetShortPath );
}

object WXEUAPI  get_modified_time(object filename)
{
	wxDateTime lastmod;
	s1_ptr seq;
	
	WXFILENAME_CONVSEQ( filename, fn );
	lastmod = fn->GetModificationTime();
	
	seq = NewS1( 6 );
	seq->base[1] = lastmod.GetYear();
	seq->base[2] = lastmod.GetMonth();
	seq->base[3] = lastmod.GetDay();
	seq->base[4] = lastmod.GetHour();
	seq->base[5] = lastmod.GetMinute();
	seq->base[6] = lastmod.GetSecond();
	
	return MAKE_SEQ( seq );
}

object WXEUAPI  get_size(object filename)
{
	wxULongLong size;
	
	WXFILENAME_CONVSEQ( filename, fn );
	size = fn->GetSize();
	
	return BOX_INT( size.ToDouble() );
}

object WXEUAPI  get_type(object filename)
{
	wxFileType *ft;
	wxString ext, desc;

	WXFILENAME_CONVSEQ( filename, fn );
	ext = fn->GetExt();

	ft = wxTheMimeTypesManager->GetFileTypeFromExtension( *ext );
	ft->GetDescription( &desc );

	return get_sequence( desc );
}

}
