// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "alldefs.h"
#include "wxeu.h"
#include "wx/grid.h"
#include "wx/generic/gridctrl.h"

extern "C"{


object WXEUAPI  new_wxGrid(object params)
{
	NEW_SIMPLE_CONTROL_NO_RETURN( wxGrid, grid );

	if( ((s1_ptr)SEQ_PTR(params))->length > 7 ){
		grid->CreateGrid( get_int( params, 8 ), get_int( params, 9 ) );
	}
	else{
		grid->CreateGrid( 10, 10 );
	}
	return BOX_INT( grid );
}

object WXEUAPI  get_cell_value(intptr_t grid, int row, int col)
{
	wxString val = ((wxGrid *)grid)->GetCellValue( row, col );
	
	return get_sequence( val );
}

void WXEUAPI  append_cols(intptr_t grid, int cols)
{
	((wxGrid*)grid)->AppendCols( cols, false );
}

void WXEUAPI  append_rows(intptr_t grid, int rows)
{
	((wxGrid*)grid)->AppendRows( rows );
}

void WXEUAPI  set_cell_value(intptr_t grid, object val, int row, int col)
{
	((wxGrid*)grid)->SetCellValue( row, col, get_string( val ) );
	wxDeRefDS( val );
}

void WXEUAPI  delete_rows(intptr_t grid, int pos, int n, int relabel)
{
	((wxGrid*)grid)->DeleteRows( pos, n, relabel );
}

void WXEUAPI  delete_cols(intptr_t grid, int pos, int n, int relabel)
{
	((wxGrid*)grid)->DeleteCols( pos, n, relabel );
}

object WXEUAPI  grid_size(intptr_t grid)
{
	s1_ptr size = NewS1( 2 );
	size->base[1] = ((wxGrid*)grid)->GetNumberRows();
	size->base[2] = ((wxGrid*)grid)->GetNumberCols();
	
	return MAKE_SEQ( size );
}

void WXEUAPI  set_col_label(intptr_t grid, int pos, object text)
{
	((wxGrid*)grid)->SetColLabelValue(pos, get_string( text ) );
	wxDeRefDS( text );
}

void WXEUAPI  set_row_label(intptr_t grid, int pos, object text)
{
	((wxGrid*)grid)->SetRowLabelValue( pos, get_string( text ) );
	wxDeRefDS(text);
}

void WXEUAPI set_row_label_size( intptr_t grid, int size )
{
	((wxGrid*)grid)->SetRowLabelSize( size );
}

void WXEUAPI  begin_batch(intptr_t grid)
{
	((wxGrid*)grid)->BeginBatch();
}

void WXEUAPI  end_batch(intptr_t grid)
{
	((wxGrid*)grid)->EndBatch();
}

object WXEUAPI  batch_count(intptr_t grid)
{
	return ((wxGrid*)grid)->GetBatchCount();
}

object WXEUAPI  last_cell_visible(intptr_t grid)
{
	wxGrid * g = (wxGrid*) grid;
	int rows = g->GetNumberRows();
	if( !rows ) return -1;
	
	int lo = 0;
	int hi = rows;
	int mid = 0;
	while( lo <= hi ){
		mid = (lo+hi)/2;
		if( g->IsVisible( 0, mid, true) )
			lo = mid;
		else
			hi = mid;
	}
	return mid;
}

void WXEUAPI  clear_grid(intptr_t grid)
{
	((wxGrid*)grid)->ClearGrid();
}

object WXEUAPI  grid_event_cell(intptr_t event)
{
	s1_ptr cell = NewS1(2);
	cell->base[1] = ((wxGridEvent*)event)->GetRow();
	cell->base[2] = ((wxGridEvent*)event)->GetCol();
	return MAKE_SEQ( cell );
}

void WXEUAPI  select_grid_cell(intptr_t grid, int row, int col)
{
	((wxGrid*)grid)->SetGridCursor( row, col );
	((wxGrid*)grid)->SelectBlock( row, col, row, col, false );
}

object WXEUAPI  grid_pos(intptr_t grid)
{
	s1_ptr pos = NewS1( 2 );
	pos->base[1] = ((wxGrid *)grid)->GetGridCursorRow();
	pos->base[2] = ((wxGrid *)grid)->GetGridCursorCol();
	
	return MAKE_SEQ( pos );
}

void WXEUAPI  set_cell(intptr_t grid, int row, int col)
{
	((wxGrid*)grid)->SetGridCursor( row, col );
}

void WXEUAPI  make_cell_visible(intptr_t grid, int row, int col)
{
	((wxGrid*)grid)->MakeCellVisible( row, col );
}

void WXEUAPI  set_grid_size(intptr_t grid, int rows, int cols)
{
	wxGrid * g = (wxGrid*) grid;
	
	int size = g->GetNumberRows();
	if( size < rows ){
		g->AppendRows( rows - size );
	}
	else if( size > rows ){
		g->DeleteRows( rows ? rows-1 : rows, size - rows );
	}
	
	size = g->GetNumberCols();
	if( size < cols ){
		g->AppendCols( cols - size );
	}
	else if( size > cols ){
		g->DeleteCols( cols ? cols-1 : 0, size - cols );
	}
}

object WXEUAPI  new_wxGridCellBoolEditor(object params)
{
	wxDeRefDS( params );
	return BOX_INT( new wxGridCellBoolEditor() );
}

object WXEUAPI  new_wxGridCellAttr(object params)
{
	wxGridCellAttr * attr = new wxGridCellAttr( 
		*(wxColour*)get_int( params, 1 ),
		*(wxColour*)get_int( params, 2 ),
		*(wxFont*)  get_int( params, 3 ),
		            get_int( params, 4 ),
		            get_int( params, 5 ));
	wxDeRefDS( params );
	return BOX_INT( attr );
}


object WXEUAPI new_wxGridCellAutoWrapStringEditor(object params )
{
	wxDeRefDS( params );
	return BOX_INT( new wxGridCellAutoWrapStringEditor() );
}


// These are defined in wxeud.e:
#define eu_wxGridCellBoolEditor     37
#define eu_wxGridCellChoiceEditor   81
#define eu_wxGridCellAutoWrapStringEditor 101
void WXEUAPI  set_col_editor(intptr_t grid, int col, int ed_type, object data)
{
	wxGridCellAttr * attr = new wxGridCellAttr();
	wxDeRef( data );
	switch( ed_type ){
	case eu_wxGridCellBoolEditor:
		attr->SetEditor( new wxGridCellBoolEditor() );
		attr->SetRenderer( new wxGridCellBoolRenderer() );
		break;
	
	case eu_wxGridCellChoiceEditor:
		{
			bool allow_others = get_int( data, 2 );
			data = SEQ_PTR(data)->base[1];
			int count = LENGTH( data );
			wxString * choices = new wxString[count];
			for( int i = 0; i < count; i++){
				choices[i] = get_string(data, i+1);
			}
			wxGridCellChoiceEditor * ed = new wxGridCellChoiceEditor( count, choices, allow_others);
			attr->SetEditor( ed );
			break;
			
		}
	case eu_wxGridCellAutoWrapStringEditor:
		attr->SetEditor( new wxGridCellAutoWrapStringEditor() );
		attr->SetRenderer( new wxGridCellAutoWrapStringRenderer() );
		break;
	default: 

		return;
	}
	
	((wxGrid*)grid)->SetColAttr( col, attr );

}

object WXEUAPI  get_cell_editor(intptr_t grid, object cell)
{
	wxGridCellEditor * editor = ((wxGrid*)grid)->GetCellEditor( get_int( cell, 1),
		get_int( cell, 2 ));
	wxDeRefDS( cell );
	return BOX_INT( editor );
}

object WXEUAPI  get_cell_editor_control(intptr_t grid, object editor)
{
	wxGridCellEditor * ed;
	if( IS_SEQUENCE( editor ) ){
		ed = ((wxGrid*)grid)->GetCellEditor( get_int( editor, 1 ), get_int( editor, 2 ) );

	}
	else{
		ed = (wxGridCellEditor*) get_int( editor );

	}
	wxDeRef(editor);
	return BOX_INT( ed->GetControl() );
}

void WXEUAPI  set_editor_choices(intptr_t grid, object cell, object choices)
{	
	wxGridCellChoiceEditor * editor;
	if( IS_SEQUENCE( cell ) ){
	
		editor = (wxGridCellChoiceEditor*) ((wxGrid*)grid)->GetCellEditor( 
			get_int( cell, 1 ),	get_int( cell, 2 ) );
	}
	else{
		editor = (wxGridCellChoiceEditor*) ((wxGrid*)grid)->GetCellEditor( 0, cell );
	}
	wxString * params = new wxString[LENGTH(choices)];
	for( int i = 0; i < LENGTH(choices); i++){
		params[i] = get_string( choices, i+1 );
	}
	
	editor->SetParameters( *params );
	
	delete[] params;
	
	wxDeRef(cell);
	wxDeRef(choices);

}

void WXEUAPI  set_cell_editor(intptr_t grid, int row, int col, int ed_type, object data)
{	
	if( ed_type == eu_wxGridCellBoolEditor ){
		((wxGrid*)grid)->SetCellEditor( row, col, new wxGridCellBoolEditor() );
		((wxGrid*)grid)->SetCellRenderer( row, col, new wxGridCellBoolRenderer() );
	}
	else if( ed_type == eu_wxGridCellChoiceEditor ){
		bool allow_others = get_int( data, 2 );
		data = SEQ_PTR(data)->base[1];
		int count = LENGTH( data );
		wxString * choices = new wxString[count];
		for( int i = 0; i < count; i++){
			choices[i] = get_string(data, i+1);
		}
		
		((wxGrid*)grid)->SetCellEditor( row, col, new wxGridCellChoiceEditor( count, choices, allow_others) );
		
	}
	else if( ed_type == eu_wxGridCellAutoWrapStringEditor ){
		((wxGrid*)grid)->SetCellEditor( row, col, new wxGridCellAutoWrapStringEditor() );
		((wxGrid*)grid)->SetCellRenderer( row, col, new wxGridCellAutoWrapStringRenderer() );
	}
	wxDeRef(data);

}

void WXEUAPI  set_default_grid_editor(intptr_t grid, int editor)
{
	((wxGrid*)grid)->SetDefaultEditor( (wxGridCellEditor*)editor );
}

void WXEUAPI  set_cell_readonly(intptr_t grid, int row, int col, int readonly)
{
	((wxGrid*)grid)->SetReadOnly( row, col, readonly );
}

void WXEUAPI  set_grid_editable(intptr_t grid, int editable)
{
	((wxGrid*)grid)->EnableEditing( editable );
}

void WXEUAPI  autosize_columns(intptr_t grid)
{
	((wxGrid*)grid)->AutoSizeColumns();
}

void WXEUAPI  set_col_label_align(intptr_t grid, int horiz, int vert)
{
	((wxGrid*)grid)->SetColLabelAlignment(horiz, vert);
}

void WXEUAPI  set_row_label_align(intptr_t grid, int horiz, int vert)
{
	((wxGrid*)grid)->SetRowLabelAlignment(horiz, vert);
}

void WXEUAPI  set_col_label_size(intptr_t grid, int height)
{
	((wxGrid*)grid)->SetColLabelSize( height );
}


object WXEUAPI  get_col_label(intptr_t grid, int col)
{
	return get_sequence( ((wxGrid*)grid)->GetColLabelValue( col ) );
}

void WXEUAPI  set_grid_line_color(intptr_t grid, int colour)
{
	wxColour color = * ((wxColour *)colour);
	((wxGrid *)grid)->SetGridLineColour( color );
}

void WXEUAPI  force_refresh_grid(intptr_t grid)
{
	((wxGrid*)grid)->ForceRefresh();
}

void WXEUAPI  set_col_size(intptr_t grid, int col, int size)
{
	((wxGrid*)grid)->SetColSize( col, size );
}

void WXEUAPI  set_cell_alignment(intptr_t grid, int row, int col, int horz, int vert)
{
	((wxGrid*)grid)->SetCellAlignment( row, col, horz, vert );
}

void WXEUAPI  set_col_alignment(intptr_t grid, int col, int horz, int vert)
{
/*
	attr = create( wxGridCellAttr, {grid, 0, col})
	void = cpp:call_member( wxGridCellAttr_SetAlignment, attr, {horz, vert})
	void = cpp:call_member( wxGrid_SetColAttr, grid, {col, attr})
*/
	wxGridCellAttr * attr = ((wxGrid*)grid)->GetOrCreateCellAttr( 0, col );
	attr->SetAlignment( horz, vert );
	((wxGrid*)grid)->SetColAttr( col, attr );
	
}

void WXEUAPI  set_col_readonly(intptr_t grid, int col, int readonly)
{
	wxGridCellAttr * attr = new wxGridCellAttr( );
	attr->SetReadOnly(readonly);
	((wxGrid*)grid)->SetColAttr( col, attr );
}

object WXEUAPI  get_row_label_size(intptr_t grid)
{
	return ((wxGrid*)grid)->GetRowLabelSize();
}

object WXEUAPI  get_col_label_size(intptr_t grid)
{
	return ((wxGrid*)grid)->GetColLabelSize();
}

object WXEUAPI  get_col_size(intptr_t grid, int col)
{
	return ((wxGrid*)grid)->GetColSize(col);
}

object WXEUAPI  get_row_size(intptr_t grid, int row)
{
	return ((wxGrid*)grid)->GetRowSize(row);
}

void WXEUAPI save_grid_edit_control_value( intptr_t grid )
{
	((wxGrid*)grid)->SaveEditControlValue();
}

object WXEUAPI is_cell_visible( intptr_t grid, int row, int col, int wholly_visible )
{
	return ((wxGrid*)grid)->IsVisible( row, col, wholly_visible );
}

void WXEUAPI set_grid_selection_mode( intptr_t grid, int mode )
{
	((wxGrid*)grid)->SetSelectionMode( (wxGrid::wxGridSelectionModes)mode );
}

void WXEUAPI enable_drag_grid_size( intptr_t grid, int enable )
{
	((wxGrid*)grid)->EnableDragGridSize( (bool) enable );
}

void WXEUAPI select_grid_block( intptr_t grid, int top_row, int left_col, int bottom_row, int right_col, int add_to_selected )
{
	((wxGrid*)grid)->SelectBlock( top_row, left_col, bottom_row, right_col, (bool) add_to_selected );
}

void WXEUAPI select_grid_row( intptr_t grid, int row, int add_to_selected )
{
	((wxGrid*)grid)->SelectRow( row, (bool) add_to_selected );
}

void WXEUAPI select_grid_col( intptr_t grid, int col, int add_to_selected )
{
	((wxGrid*)grid)->SelectCol( col, (bool) add_to_selected );
}

void WXEUAPI select_grid_all( intptr_t grid )
{
	((wxGrid*)grid)->SelectAll();
}

object WXEUAPI is_in_grid_selection( intptr_t grid, int row, int col )
{
	return 1 & ((wxGrid*)grid)->IsInSelection( row, col );
}

void WXEUAPI autosize_grid( intptr_t grid )
{
	((wxGrid*)grid)->AutoSize();
}

void WXEUAPI autosize_rows( intptr_t grid )
{
	((wxGrid*)grid)->AutoSizeRows();
}

void WXEUAPI auotosize_row( intptr_t grid, int row )
{
	((wxGrid*)grid)->AutoSizeRow( row );
}

object WXEUAPI can_drag_grid_col_move( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->CanDragColMove();
}

object WXEUAPI can_drag_grid_col_size( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->CanDragColSize();
}

object WXEUAPI can_drag_grid_row_size( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->CanDragRowSize();
}

object WXEUAPI can_drag_grid_size( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->CanDragGridSize();
}

object WXEUAPI can_enable_cell_control( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->CanEnableCellControl();
}

object WXEUAPI grid_cell_to_rect( intptr_t grid, int row, int col )
{
	wxRect r = ((wxGrid*)grid)->CellToRect( row, col );
	s1_ptr s = NewS1( 4 );
	s->base[1] = r.x;
	s->base[2] = r.y;
	s->base[3] = r.width;
	s->base[4] = r.height;
	
	return MAKE_SEQ( s );
}

void WXEUAPI clear_grid_selection( intptr_t grid )
{
	((wxGrid*)grid)->ClearSelection();
}

void WXEUAPI enable_grid_cell_edit_control( intptr_t grid, int enable )
{
	if( enable )
		((wxGrid*)grid)->EnableCellEditControl();
	else
		((wxGrid*)grid)->DisableCellEditControl();
}

void WXEUAPI enable_drag_grid_col_move( intptr_t grid, int enable )
{
	if( enable )
		((wxGrid*)grid)->EnableDragColMove();
	else
		((wxGrid*)grid)->DisableDragColMove();
}

void WXEUAPI enable_drag_grid_col_size( intptr_t grid, int enable )
{
	if( enable )
		((wxGrid*)grid)->EnableDragColSize();
	else
		((wxGrid*)grid)->DisableDragColSize();
}

void WXEUAPI enable_drag_grid_row_size( intptr_t grid, int enable )
{
	if( enable )
		((wxGrid*)grid)->EnableDragRowSize();
	else
		((wxGrid*)grid)->DisableDragRowSize();
}

void WXEUAPI enable_grid_editing( intptr_t grid, int edit )
{
	((wxGrid*)grid)->EnableEditing( edit );
}

void WXEUAPI enable_grid_lines( intptr_t grid, int enable )
{
	((wxGrid*)grid)->EnableGridLines( enable );
}

object WXEUAPI get_grid_batch_count( intptr_t grid )
{
	return BOX_INT( ((wxGrid*)grid)->GetBatchCount() );
}

object WXEUAPI get_cell_alignment( intptr_t grid, int row, int col )
{
	s1_ptr s = NewS1( 2 );
	((wxGrid*)grid)->GetCellAlignment( row, col, (int*)s->base + 1, (int*)s->base + 2 );
	return MAKE_SEQ( s );
}

object WXEUAPI get_cell_background_color( intptr_t grid, int row, int col )
{
	return BOX_INT( new wxColour( ((wxGrid*)grid)->GetCellBackgroundColour( row, col ) ) );
}

object WXEUAPI get_cell_font( intptr_t grid, int row, int col )
{
	return BOX_INT( new wxFont( ((wxGrid*)grid)->GetCellFont( row, col ) ) );
}

object WXEUAPI get_cell_text_color( intptr_t grid, int row, int col )
{
	return BOX_INT( new wxColour( ((wxGrid*)grid)->GetCellTextColour( row, col ) ) );
}

object WXEUAPI get_default_cell_background_color( intptr_t grid )
{
	return BOX_INT( new wxColour( ((wxGrid*)grid)->GetDefaultCellBackgroundColour() ) );
}

object WXEUAPI get_default_cell_font( intptr_t grid )
{
	return BOX_INT( new wxFont( ((wxGrid*)grid)->GetDefaultCellFont() ) );
}

object WXEUAPI get_default_cell_text_colour( intptr_t grid )
{
	return BOX_INT( new wxColour( ((wxGrid*)grid)->GetDefaultCellTextColour() ) );
}

object WXEUAPI get_default_col_label_size( intptr_t grid )
{
	return BOX_INT( ((wxGrid*)grid)->GetDefaultColLabelSize() );
}

object WXEUAPI get_default_col_size( intptr_t grid )
{
	return BOX_INT( ((wxGrid*)grid)->GetDefaultColSize() );
}

object WXEUAPI get_default_grid_editor( intptr_t grid )
{
	return BOX_INT( ((wxGrid*)grid)->GetDefaultEditor() );
}

object WXEUAPI get_default_editor_for_cell( intptr_t grid, int row, int col )
{
	return BOX_INT( ((wxGrid*)grid)->GetDefaultEditorForCell( row, col ) );
}

object WXEUAPI get_default_row_label_size( intptr_t grid )
{
	return BOX_INT( ((wxGrid*)grid)->GetDefaultRowLabelSize() );
}

object WXEUAPI get_default_row_size( intptr_t grid )
{
	return BOX_INT( ((wxGrid*)grid)->GetDefaultRowSize() );
}

object WXEUAPI get_grid_line_color( intptr_t grid )
{
	return BOX_INT( new wxColour( ((wxGrid*)grid)->GetGridLineColour() ) );
}

object WXEUAPI get_grid_lines_enabled( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->GridLinesEnabled();
}

object WXEUAPI get_grid_label_background_color( intptr_t grid )
{
	return BOX_INT( new wxColour( ((wxGrid*)grid)->GetLabelBackgroundColour() ) );
}

object WXEUAPI get_grid_label_font( intptr_t grid )
{
	return BOX_INT( new wxFont( ((wxGrid*)grid)->GetLabelFont() ) );
}

object WXEUAPI get_grid_label_text_color( intptr_t grid )
{
	return BOX_INT( new wxFont( ((wxGrid*)grid)->GetLabelFont() ) );
}

object WXEUAPI get_grid_row_label_align( intptr_t grid ){
	s1_ptr s = NewS1( 2 );
	((wxGrid*)grid)->GetRowLabelAlignment( (int*) s->base + 1, (int*) s->base + 2 );
	return MAKE_SEQ( s );
}

object WXEUAPI get_grid_scroll_line( intptr_t grid )
{
	s1_ptr s = NewS1( 2 );
	s->base[1] = ((wxGrid*)grid)->GetScrollLineX();
	s->base[2] = ((wxGrid*)grid)->GetScrollLineY();
	return MAKE_SEQ( s );
}

object WXEUAPI get_grid_selection_background( intptr_t grid )
{
	return BOX_INT( new wxColour( ((wxGrid*)grid)->GetSelectionBackground() ) );
}

object WXEUAPI get_grid_selection_foreground( intptr_t grid )
{
	return BOX_INT( new wxColour( ((wxGrid*)grid)->GetSelectionForeground() ) );
}

object WXEUAPI get_grid_view_width( intptr_t grid )
{
	return BOX_INT( ((wxGrid*)grid)->GetViewWidth() );
}

void WXEUAPI hide_cell_edit_control( intptr_t grid )
{
	((wxGrid*)grid)->HideCellEditControl();
}

object WXEUAPI insert_grid_cols( intptr_t grid, int pos, int num_cols )
{
	return 1 & ((wxGrid*)grid)->InsertCols( pos, num_cols );
}

object WXEUAPI insert_grid_rows( intptr_t grid, int pos, int num_rows )
{
	return 1 & ((wxGrid*)grid)->InsertRows( pos, num_rows );
}

object WXEUAPI is_cell_edit_control_enabled( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->IsCellEditControlEnabled();
}

object WXEUAPI is_current_cell_read_only( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->IsCurrentCellReadOnly();
}

object WXEUAPI is_grid_editable( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->IsEditable();
}

object WXEUAPI is_cell_read_only( intptr_t grid, int row, int col )
{
	return 1 & ((wxGrid*)grid)->IsReadOnly( row, col );
}

object WXEUAPI is_selection_for_grid( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->IsSelection();
}

object WXEUAPI move_grid_cursor_down( intptr_t grid, int expand_selection )
{
	return 1 & ((wxGrid*)grid)->MoveCursorDown( expand_selection );
}

object WXEUAPI move_grid_cursor_left( intptr_t grid, int expand_selection )
{
	return 1 & ((wxGrid*)grid)->MoveCursorLeft( expand_selection );
}

object WXEUAPI move_grid_cursor_right( intptr_t grid, int expand_selection )
{
	return 1 & ((wxGrid*)grid)->MoveCursorLeft( expand_selection );
}

object WXEUAPI move_grid_cursor_up( intptr_t grid, int expand_selection )
{
	return 1 & ((wxGrid*)grid)->MoveCursorUp( expand_selection );
}

object WXEUAPI move_grid_cursor_down_block( intptr_t grid, int expand_selection )
{
	return 1 & ((wxGrid*)grid)->MoveCursorDownBlock( expand_selection );
}

object WXEUAPI move_grid_cursor_left_block( intptr_t grid, int expand_selection )
{
	return 1 & ((wxGrid*)grid)->MoveCursorLeftBlock( expand_selection );
}

object WXEUAPI move_grid_cursor_right_block( intptr_t grid, int expand_selection )
{
	return 1 & ((wxGrid*)grid)->MoveCursorRight( expand_selection );
}

object WXEUAPI move_grid_cursor_up_block( intptr_t grid, int expand_selection )
{
	return 1 & ((wxGrid*)grid)->MoveCursorUpBlock( expand_selection );
}

object WXEUAPI move_grid_page_down( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->MovePageDown();
}

object WXEUAPI move_grid_page_up( intptr_t grid )
{
	return 1 & ((wxGrid*)grid)->MovePageUp();
}

void WXEUAPI set_cell_background_color( intptr_t grid, int row, int col, intptr_t color )
{
	((wxGrid*)grid)->SetCellBackgroundColour( row, col, *(wxColour*)color);
}

void WXEUAPI set_cell_font( intptr_t grid, int row, int col, int font )
{
	((wxGrid*)grid)->SetCellFont( row, col, *(wxFont*)font );
}

void WXEUAPI set_cell_text_color( intptr_t grid, int row, int col, intptr_t color )
{
	((wxGrid*)grid)->SetCellTextColour( row, col, *(wxColour*)color);
}

void WXEUAPI set_col_format_bool( intptr_t grid, int col )
{
	((wxGrid*)grid)->SetColFormatBool( col );
}

void WXEUAPI set_col_format_number( intptr_t grid, int col )
{
	((wxGrid*)grid)->SetColFormatNumber( col );
}

void WXEUAPI set_col_format_float( intptr_t grid, int col, int width, int precision )
{
	((wxGrid*)grid)->SetColFormatFloat( col, width, precision );
}

void WXEUAPI set_col_pos( intptr_t grid, int col_id, int new_pos )
{
	((wxGrid*)grid)->SetColPos( col_id, new_pos );
}

void WXEUAPI set_default_cell_align( intptr_t grid, int horiz, int vert )
{
	((wxGrid*)grid)->SetDefaultCellAlignment( horiz, vert );
}

void WXEUAPI set_default_cell_background_color( intptr_t grid, intptr_t color )
{
	((wxGrid*)grid)->SetDefaultCellBackgroundColour( *(wxColour*)color );
}

void WXEUAPI set_default_cell_font( intptr_t grid, int font )
{
	((wxGrid*)grid)->SetDefaultCellFont( *(wxFont*)font );
}

void WXEUAPI set_default_cell_text_color( intptr_t grid, intptr_t color )
{
	((wxGrid*)grid)->SetDefaultCellTextColour( *(wxColour*)color);
}

void WXEUAPI set_default_col_size( intptr_t grid, int width, int resize_existing )
{
	((wxGrid*)grid)->SetDefaultColSize( width, resize_existing );
}

void WXEUAPI set_default_row_size( intptr_t grid, int width, int resize_existing )
{
	((wxGrid*)grid)->SetDefaultRowSize( width, resize_existing );
}

void WXEUAPI set_grid_label_background_color( intptr_t grid, intptr_t color )
{
	((wxGrid*)grid)->SetLabelBackgroundColour( *(wxColour*)color );
}

void WXEUAPI set_grid_label_font( intptr_t grid, int font )
{
	((wxGrid*)grid)->SetLabelFont( *(wxFont*)font);
}

void WXEUAPI set_grid_label_text_color( intptr_t grid, intptr_t color )
{
	((wxGrid*)grid)->SetLabelTextColour( *(wxColour*)color );
}

void WXEUAPI set_grid_margins( intptr_t grid, int extra_width, int extra_height )
{
	((wxGrid*)grid)->SetMargins( extra_width, extra_height );
}

void WXEUAPI set_row_size( intptr_t grid, int row, int height )
{
	((wxGrid*)grid)->SetRowSize( row, height );
}

void WXEUAPI set_grid_scroll_line( intptr_t grid, object xy )
{
	s1_ptr x = SEQ_PTR( xy );
	((wxGrid*)grid)->SetScrollLineX( x->base[1] );
	((wxGrid*)grid)->SetScrollLineY( x->base[2] );
	wxDeRefDS( xy );
}

void WXEUAPI set_grid_selection_background( intptr_t grid, intptr_t color )
{
	((wxGrid*)grid)->SetSelectionBackground( *(wxColour*)color );
}

void WXEUAPI set_grid_selection_foreground( intptr_t grid, intptr_t color )
{
	((wxGrid*)grid)->SetSelectionForeground( *(wxColour*) color );
}

void WXEUAPI show_cell_edit_control( intptr_t grid )
{
	((wxGrid*)grid)->ShowCellEditControl();
}

object WXEUAPI get_number_cols( intptr_t grid )
{
	return ((wxGrid*)grid)->GetNumberCols();
}

object WXEUAPI get_number_rows( intptr_t grid )
{
	return ((wxGrid*)grid)->GetNumberRows();
}

object WXEUAPI get_selected_cells( intptr_t grid )
{
	wxGridCellCoordsArray cells = ((wxGrid*)grid)->GetSelectedCells();
	int count = cells.GetCount();
	s1_ptr s = NewS1( count );
	
	for (int i=0; i<count; i++)
	{
		wxGridCellCoords coord = cells.Item(i);
		
		s1_ptr c = NewS1( 2 );
		c->base[1] = coord.GetRow();
		c->base[2] = coord.GetCol();
		
		s->base[i+1] = MAKE_SEQ( c );
	}
	
	return MAKE_SEQ( s );
}


object WXEUAPI get_selected_cols( intptr_t grid )
{
	wxArrayInt cells = ((wxGrid*)grid)->GetSelectedCols();
	int count = cells.GetCount();
	s1_ptr s = NewS1( count );
	
	for (int i=0; i<count; i++)
	{
		s->base[i+1] = cells.Item(i);
	}
	
	return MAKE_SEQ( s );
}

object WXEUAPI get_selected_rows( intptr_t grid )
{
	wxArrayInt cells = ((wxGrid*)grid)->GetSelectedRows();
	int count = cells.GetCount();
	s1_ptr s = NewS1( count );
	
	for (int i=0; i<count; i++)
	{
		s->base[i+1] = cells.Item(i);
	}
	
	return MAKE_SEQ( s );
}

}
