// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "alldefs.h"
#include "wxeu.h"

#include "wx/socket.h"
#include "wx/buffer.h"
#include <wx/url.h>
#include <wx/protocol/ftp.h>


extern "C"{
object WXEUAPI  new_wxSocketClient(object params)
{

	wxSocketClient * client = new wxSocketClient( get_int( params, 1 ));
	wxDeRefDS(params);
	return BOX_INT( client );
}

object WXEUAPI  new_wxSocketServer(object params)
{
	int len = LENGTH( params );
	int flags = wxSOCKET_NONE;
	if ( len > 1 ) flags = get_int( params, 2 );
	wxSocketServer * server = new wxSocketServer( *((wxSockAddress*)get_int(params, 1)),
		flags );
	wxDeRefDS( params );
	return BOX_INT( server );
}

object WXEUAPI  set_ip_host(intptr_t ip, object name)
{
	wxDeRefDS(name);
	return (object) ((wxIPV4address*)ip)->Hostname( get_string( name ) );
}

object WXEUAPI  set_ip_service(intptr_t ip, object service)
{
	wxDeRef( service );
	if( IS_SEQUENCE( service ) ){
		return (object) ((wxIPV4address*)ip)->Service( get_string( service ) );
	}
	else {
		return (object) ((wxIPV4address*)ip)->Service( (unsigned short) service );
	}
}

object WXEUAPI  socket_connect(intptr_t client, intptr_t address, int wait)
{
	return (object) ((wxSocketClient*)client)->Connect( *((wxSockAddress*)address),
		(bool) wait );
}

void WXEUAPI  socket_write(intptr_t socket, object bytes)
{

	int len = LENGTH( bytes );
	char * buffer = new char[len];
	s1_ptr b = SEQ_PTR( bytes );
	for( int i = 0; i < len; i++ ){
		buffer[i] = (char) b->base[i+1];
	}
	
	((wxSocketBase*)socket)->Write( (void *)buffer, len );
	wxDeRefDS(bytes);
	delete buffer;
}

object WXEUAPI  socket_read(intptr_t socket)
{

	wxMemoryBuffer data(8192);
	
	wxSocketBase * s = ((wxSocketBase*)socket);
	s->Read( data.GetData(), 8192 );
	int last = s->LastCount();
	int size = last;
	while( last == 8192 ){
		s->Read( data.GetAppendBuf( 8192 ) , 8192);
		last = s->LastCount();
		size += last;
	}
	
	//int size = data.GetDataLen();
	s1_ptr seq = NewS1( size );
	char * buffer = (char *) data.GetData();
	for( int i = 0; i < size; i++ ){
		seq->base[i+1] = (int) buffer[i];
	}
	
	return MAKE_SEQ( seq );
	
}

void WXEUAPI  stream_write(intptr_t stream, object bytes)
{
	int len = LENGTH( bytes );
	char * buffer = new char[len];
	s1_ptr b = SEQ_PTR( bytes );
	for ( int i = 0; i < len; i++ ){
		buffer[i] = (char) b->base[i+1];
	}
	
	((wxOutputStream*)stream)->Write( (void *)buffer, len );
	wxDeRefDS( bytes );
	delete buffer;
}

object WXEUAPI  stream_read(intptr_t stream)
{
	wxMemoryBuffer data(8192);
	wxInputStream * s = ((wxInputStream*)stream);

	//s->Read( data.GetData(), 8192 );
	s->Read( data.GetAppendBuf(8192), 8192 );
	
	int last = s->LastRead();
	data.UngetAppendBuf( last );
	
	int size = last;
	
	while( last == 8192 ){
		s->Read( data.GetAppendBuf(8192), 8192 );
		last = s->LastRead();
		data.UngetAppendBuf( last );
		size += last;
	}
	
	//int size = data.GetDataLen();
	//size = data.GetDataLen();
	
	s1_ptr seq = NewS1( size );
	char * buffer = (char *) data.GetData();
	
	for( int i = 0; i < size; i++ ){
		seq->base[i+1] = (intptr_t) buffer[i];
	}
	
	return MAKE_SEQ( seq );	
}

void WXEUAPI  set_socket_notify(intptr_t socket, object notify)
{


	if( IS_SEQUENCE( notify )){
		int flags = 0;
		s1_ptr n = SEQ_PTR( notify );
		for( int i = 1; i <= n->length; i++ ){
			flags |= n->base[i];
		}
		((wxSocketBase*)socket)->SetNotify( flags );
		((wxSocketBase*)socket)->Notify( 1 );
	}
	else{
		((wxSocketBase*)socket)->Notify( notify );
	}
	wxDeRef(notify);

}

object WXEUAPI  is_connected(intptr_t socket)
{
	return ((wxSocketBase*)socket)->IsConnected();
}

void WXEUAPI  socket_event_handler(intptr_t socket, intptr_t evt_handler, intptr_t id)
{
//void SetEventHandler(wxEvtHandler& handler, int id = -1)
	((wxSocketBase*)socket)->SetEventHandler( *((wxEvtHandler*)evt_handler), id );

}

object WXEUAPI  get_socket_event(intptr_t event)
{
	return ((wxSocketEvent*)event)->GetSocketEvent();
}

object WXEUAPI  socket_from_event(intptr_t event)
{
	intptr_t socket = (intptr_t) ((wxSocketEvent*)event)->GetSocket();
	return BOX_INT( socket );
}

void WXEUAPI  set_flags(intptr_t socket, int flags)
{
	((wxSocketBase*)socket)->SetFlags( flags );
}

object WXEUAPI  new_wxURL(object params)
{
	object url;
	if( LENGTH( params ) ){
		url =  BOX_INT( new wxURL( get_string( params, 1 ) ) );
		
	}
	else{
		
		url = BOX_INT( new wxURL() );
	}
	wxDeRefDS( params );
	return url;
}

object WXEUAPI  get_url_input(intptr_t url)
{
	return BOX_INT( ((wxURL*)url)->GetInputStream() );
}

object WXEUAPI  new_wxHTTP(object params)
{
	wxDeRefDS( params );
	return BOX_INT( new wxHTTP() );
}

object WXEUAPI  http_connect(intptr_t http, object host, int port)
{
	wxDeRefDS( host );
	return ((wxHTTP*)http)->Connect(get_string( host ), (unsigned) port);
}

object WXEUAPI  get_http_content_type(intptr_t http)
{
	return get_sequence( ((wxHTTP*)http)->GetContentType() );
}

object WXEUAPI  get_http_header(intptr_t http, object header)
{
	wxDeRefDS( header );
	return get_sequence( ((wxHTTP*)http)->GetHeader( get_string(header) ) );
}

void WXEUAPI  set_http_header(intptr_t http, object header, object data)
{
	wxDeRefDS( header );
	wxDeRefDS( data );

	((wxHTTP*)http)->SetHeader( get_string(header), get_string( data ) );
}

void WXEUAPI  set_http_post_buffer(intptr_t http, object buffer)
{
	wxDeRefDS( buffer );
	((wxHTTP*)http)->SetPostBuffer( get_string( buffer ) );
}

void WXEUAPI  set_http_proxy_mode(intptr_t http, int on)
{
	((wxHTTP*)http)->SetProxyMode( (bool) on );
}

object WXEUAPI  get_http_input(intptr_t http, object path)
{
	wxDeRefDS( path );
	return BOX_INT( ((wxHTTP*)http)->GetInputStream( get_string( path ) ) );
}

object WXEUAPI new_wxFTP( object params )
{
	wxDeRefDS( params );
	return BOX_INT( new wxFTP );
}

object WXEUAPI  ftp_abort(intptr_t ftp)
{
	return ((wxFTP*)ftp)->Abort();
}

object WXEUAPI  ftp_connect(intptr_t ftp, object host)
{
	wxDeRefDS( host );
	return ((wxFTP*)ftp)->Connect( get_string( host ) );
}

object WXEUAPI  ftp_close(intptr_t ftp)
{
	return ((wxFTP*)ftp)->Close();
}

void WXEUAPI  ftp_user(intptr_t ftp, object user)
{
	((wxFTP*)ftp)->SetUser( get_string( user ) );
	wxDeRefDS( user );
}

void WXEUAPI  ftp_password(intptr_t ftp, object pwd)
{
	((wxFTP*)ftp)->SetPassword( get_string( pwd ) );
	wxDeRefDS(pwd);
}

void WXEUAPI  ftp_mode(intptr_t ftp, int mode)
{
	((wxFTP*)ftp)->SetTransferMode( (wxFTP::TransferMode) mode );
}

void WXEUAPI  ftp_passive(intptr_t ftp, int passive)
{
	((wxFTP*)ftp)->SetPassive( (bool) passive );
}

object WXEUAPI  ftp_input_stream(intptr_t ftp, object path)
{
	wxInputStream * is = ((wxFTP*)ftp)->GetInputStream( get_string( path ) );
	wxDeRefDS( path );
	return BOX_INT( is );
}

object WXEUAPI  ftp_output_stream(intptr_t ftp, object path)
{
	wxOutputStream * os = ((wxFTP*)ftp)->GetOutputStream( get_string( path ) );
	wxDeRefDS( path );
	return BOX_INT( os );
}

object WXEUAPI  ftp_check_command(intptr_t ftp, object command, int ret)
{
	wxDeRefDS( command );
	return ((wxFTP*)ftp)->CheckCommand( get_string( command ), ret );
}

object WXEUAPI  ftp_send_command(intptr_t ftp, object command)
{
	wxDeRefDS( command );
	return ((wxFTP*)ftp)->SendCommand( get_string( command ) );
}

object WXEUAPI  ftp_file_exists(intptr_t ftp, object filename)
{
	wxDeRefDS( filename );
	return ((wxFTP*)ftp)->FileExists( get_string( filename ) );

}

object WXEUAPI  ftp_file_size(intptr_t ftp, object filename)
{
	wxDeRefDS( filename );
	return BOX_INT( ((wxFTP*)ftp)->GetFileSize( get_string( filename ) ) );
}


object WXEUAPI  ftp_rm(intptr_t ftp, object path)
{
	bool success =  ((wxFTP*)ftp)->RmFile( get_string( path ) );
	wxDeRefDS( path );
	return success;
}

object WXEUAPI  ftp_chdir(intptr_t ftp, object path)
{
	bool success = ((wxFTP*)ftp)->ChDir( get_string( path ) );
	wxDeRef( path );
	return success;
}

object WXEUAPI  ftp_mkdir(intptr_t ftp, object path)
{
	bool success = ((wxFTP*)ftp)->MkDir( get_string( path ) );
	wxDeRef( path );
	return success;
}

object WXEUAPI  ftp_rmdir(intptr_t ftp, object path)
{
	bool success = ((wxFTP*)ftp)->RmDir( get_string( path ) );
	wxDeRefDS( path );
	return success;
}

object WXEUAPI  ftp_rename(intptr_t ftp, object source, object dest)
{
	bool success = ((wxFTP*)ftp)->Rename( get_string( source ), get_string( dest ) );

	wxDeRefDS( dest );
	return success;
}

object WXEUAPI  ftp_pwd(intptr_t ftp)
{
	return get_sequence( ((wxFTP*)ftp)->Pwd());
}

object WXEUAPI  ftp_files(intptr_t ftp, object wildcard)
{

	wxArrayString files;
	wxString wcard = get_string( wildcard );
	s1_ptr seq;
	intptr_t count;

	if ( ((wxFTP*)ftp)->GetFilesList(files, wcard) ) {

		count = (intptr_t) files.GetCount();

		seq = NewS1( count );
		for( int i = 0; i < count; i++){
			seq->base[i+1] = get_sequence( files.Item(i) );
		}

	} else {
		// return empty sequence
		seq = NewS1( 0 );

	}

	wxDeRefDS( wildcard );
	return MAKE_SEQ( seq );
}

object WXEUAPI  ftp_dir(intptr_t ftp, object wildcard)
{
	wxArrayString files;
	wxString wcard = get_string( wildcard );
	s1_ptr seq;
	intptr_t count;
	
	if ( ((wxFTP*)ftp)->GetDirList(files, wcard) ) {
	
		count = (intptr_t) files.GetCount();
		
		seq = NewS1( count );
		for( int i = 0; i < count; i++){
			seq->base[i+1] = get_sequence( files.Item(i) );
		}
	
	} else {
		seq = NewS1( 0 );
	}

	wxDeRefDS( wildcard );
	return MAKE_SEQ( seq );
}

object WXEUAPI  ftp_last_result(intptr_t ftp)
{
	wxString result = ((wxFTP*)ftp)->GetLastResult();
	return get_sequence( result );
}

object WXEUAPI  socket_accept(intptr_t socket, int wait)
{
	wxSocketBase * server = ((wxSocketServer*)socket)->Accept( wait );
	return BOX_INT( server );
}

void WXEUAPI  any_address(intptr_t addr)
{
	((wxIPV4address*)addr)->AnyAddress();
}

object WXEUAPI new_wxIPV4address( object params )
{
	wxDeRefDS( params );
	wxIPV4address * addr = new wxIPV4address();
	return BOX_INT( addr );
}

object WXEUAPI wait_on_connect( int client, int seconds, int milliseconds )
{
	return 1 & ((wxSocketClient*)client)->WaitOnConnect( seconds, milliseconds );
}

}
