/*****************************************************************************/
/*      (c) Copyright 2007 Rapid Deployment Software - See License.txt       */
/*****************************************************************************/
/*                                                                           */
/*                       SCREEN OUTPUT HANDLER                               */
/*                                                                           */
/*****************************************************************************/

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "wx/string.h"

/******************/
/* Included files */
/******************/
#include <stdio.h>
#include <time.h>
#include <wchar.h>
#include <wctype.h>
#include <stdlib.h>
#include <stdio.h>
#ifndef WXEUMSW
#include <sys/stat.h>
#include <unistd.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#endif
#ifdef WXEUMSW
#define UNICODE
#include <windows.h>
#endif
#include "alldefs.h"
#include "wxeu.h"

/**********************/
/* Imported functions */
/**********************/
extern unsigned int default_heap;
extern s1_ptr NewS1(long);

/******************/
/* Local defines  */
/******************/
#define MAX_SCREEN_WIDTH 200 /* what if resolutions get really high? >1280 */
#define TAB_WIDTH 8    /* power of 2 assumed */
#define NUM_SIZE 30
#define TEMP_SIZE 1040

/* convert atom to char. *must avoid side effects in elem* */
#define wchar_t(elem) ((IS_ATOM_INT(elem)) ? ((wchar_t)INT_VAL(elem)) : dowchar_t(elem)) 

#define IsDigit(x) (x >= '0' && x <= '9')

#define EMalloc(size) malloc(size)
#define EFree(ptr) free(ptr)
#define ERealloc(orig, newsize) realloc(orig, newsize)

#define LOCAL_SPACE 100 /* some local space */

/**********************/
/* Exported variables */
/**********************/

wxString collect;    /* to collect sprintf/sprint output */


/*******************/
/* Local variables */
/*******************/

//wxChar * expanded_string[MAX_SCREEN_WIDTH];
//wxChar * expanded_ptr = expanded_string;
static int collect_next;   /* place to store next collect output */
static int collect_free;   /* number of chars of empty space remaining */

/**********************/
/* Declared functions */
/**********************/


/*********************/
/* Defined functions */
/*********************/

object NewWString(wxChar *s)
/* create a new string sequence */
{
    int len;
    object_ptr obj_ptr;
    s1_ptr c1;

    len = wcslen( (wchar_t*)s);
    c1 = NewS1((long)len);
    obj_ptr = (object_ptr)c1->base;
    if (len > 0) {
	do {
	    *(++obj_ptr) = BOX_INT(*s++);
	} while (--len > 0);
    }
    return MAKE_SEQ(c1);
}

static wxChar dowxChar(object elem)
/* convert to char (int done in-line) */
{
    if (IS_ATOM_INT(elem)) 
	return (wxChar)elem;
    if (IS_ATOM(elem)) 
	return (wxChar)(DBL_PTR(elem)->dbl);
    else {
	// RTFatal("sequence found inside character string");
	return 0;
    }
}

void MakeWString(wxChar *s, object obj)
/* make an atom or sequence into a wide string */
/* N.B. caller must allow one extra for the null terminator */
{
    object_ptr elem;
    object x;

    if (IS_ATOM(obj)) 
	*s++ = dowxChar(obj);
    else {
	obj = (object)SEQ_PTR(obj);
	elem = ((s1_ptr)obj)->base;
	while (TRUE) { 
	    x = *(++elem);
	    if (IS_ATOM_INT(x)) {
		*s++ = (wxChar)x;
	    }
	    else {
		if (x == NOVALUE)
		    break;
		*s++ = dowxChar(x);
	    }
	}
    }
    *s = 0;
}


void wide_output(wxChar *out_string)
/* out_string is null-terminated string of characters to write out */
{
    collect += wxString( out_string );
}

void wide_output( wxString out_string )
{
    collect += out_string;
}

static object_ptr FormatItemW(wxChar *cstring, object_ptr f_elem, object_ptr f_last, object_ptr v_elem)
/* print one format item from printf */
{
    int flen;
    wxChar c;
    long dval;
    unsigned long uval;
    double gval;
    wxChar *sval;
    long slength;
    wxChar quick_alloc1[LOCAL_SPACE];
    int free_sv;
    int free_sb;
    
    c = wxT('%');
    flen = 0;
    do {
	cstring[flen++] = c;
	if (++f_elem > f_last) { 
	    cstring[flen] = wxT('\0');
	    //sprintf(TempBuff, "format specifier is incomplete (%s)", cstring);
	    //RTFatal(TempBuff);
	} 
	c = dowxChar(*f_elem);
    } while (IsDigit(c) || c == wxT('.') || c == wxT('-') || c == wxT('+'));

    free_sb = FALSE;
    if (c == L's') {

	cstring[flen++] = c;
	cstring[flen] = wxT('\0'); 
	free_sv = FALSE;
	if (IS_SEQUENCE(*v_elem)) {
	    slength = (SEQ_PTR(*v_elem))->length + 1;
	    
	    if (slength > LOCAL_SPACE) {
		sval = new wxChar[slength];
		free_sv = TRUE;
	    }
	    else 
		sval = quick_alloc1;
	    
	    MakeWString(sval, *v_elem);
	}
	else {
	    slength = 4L;
	    sval = quick_alloc1;
	    sval[0] =wxChar(*v_elem);
	    sval[1] = wxT('\0');
	}    
	wide_output( wxString::Format( cstring, sval ) );

	if (free_sv)
	    delete[] sval;
    }
    else if (c == wxT('d') || c == wxT('x') || c == wxT('o')) {
	cstring[flen++] = L'l';
	if (c == wxT('x'))
	    c = wxT('X');
	if (IS_ATOM_INT(*v_elem)) {
	    dval = INT_VAL(*v_elem);
	}
	    
	else {
	    gval = DBL_PTR(*v_elem)->dbl;
	    if (gval > (long)0x7FFFFFFF || gval < (long)0x80000000) {
		/* can't convert to long integer */
		if (c == wxT('d')) {
		    /* use .0f instead */
		    cstring[flen-1] = wxT('.');
		    cstring[flen++] = wxT('0');
		    c = wxT('f');
		}
		else if (gval >= 0.0 &&
			 gval <= (unsigned long)0xFFFFFFFF) {
		    /* need conversion to unsigned */
		    uval = (unsigned int)gval;
		    dval = (long)uval;
		}
		else {
		    //RTFatal("number is too big for %x or %o format");
		}
	    }
	    else {
		/* convert to positive or negative long integer */
		dval = (int)gval;
	    }
	}
	cstring[flen++] = c;
	cstring[flen] = wxT('\0'); 
	if (c == wxT('f'))
	    wide_output( wxString::Format( cstring, gval) );
	else{
	    wide_output( wxString::Format( cstring, dval) );
	}
    }
    else if (c == wxT('e') || c == wxT('f') || c == wxT('g')) {
	cstring[flen++] = c;
	cstring[flen] = wxT('\0'); 
	if (IS_ATOM_INT(*v_elem))
	    gval = (double)INT_VAL(*v_elem);
	else
	    gval = DBL_PTR(*v_elem)->dbl;
	
	wide_output( wxString::Format( cstring, gval ) );
    }
    else {
	cstring[flen++] = c;
	cstring[flen] = wxT('\0'); 
	//sprintf(TempBuff, "Unknown printf format (%s)", cstring);
	//RTFatal(TempBuff);
    }
    
    return(f_elem);
}

object ESprintfW( object format_obj, object values_obj)
/* formatted print */
{
    object_ptr f_elem, f_last;
    wxChar c; /* avoid peep bug - no register decl */
    object_ptr v_elem, v_last;
    wxChar *cstring;
    wxChar quick_alloc[LOCAL_SPACE]; // don't use TempBuff - FormatItem uses it
    int free_cs;
    wxChar out_string[LOCAL_SPACE];
    long flen;
    int s;
    s1_ptr format, values;

    free_cs = FALSE;
    collect = wxEmptyString;
    
    if (IS_ATOM(format_obj)) { 
	out_string[0] = dowxChar(format_obj); 
	out_string[1] = wxT('\0');
	wide_output(out_string);
    }
    else {
	format = SEQ_PTR(format_obj);
	flen = format->length;
	if (flen == 0) {
	    wide_output(wxT(""));
	}
	else {
	    f_elem = format->base;
	    f_last = f_elem + flen;
	    f_elem++;
	    if (flen > LOCAL_SPACE) {
		cstring = new wxChar[flen+1];
		free_cs = TRUE;
	    }
	    else
		cstring = quick_alloc;
	    if (IS_ATOM(values_obj)) 
		v_elem = (object_ptr)values;
	    else {
	        values = SEQ_PTR(values_obj);
		v_elem = values->base;
		v_last = v_elem + values->length;
		v_elem++;
	    }
	    out_string[0] = wxT('\0');
	    s = 0;
	    while (f_elem <= f_last) {
		c = wxChar((object)*f_elem);
		if (c == wxT('%')) {
		    if (f_elem < f_last && wxChar((object)*(f_elem + 1)) == wxT('%')) {
			/* %% */
			if (s >= LOCAL_SPACE-1) {
			    out_string[s] = wxT('\0');
			    wide_output(out_string);
			    s = 0;
			}
			out_string[s++] = wxT('%');
			f_elem++;
		    }
		    else {
			if (IS_SEQUENCE(values_obj) && v_elem > v_last) {
				puts("not enough values!");
			    //if (file_no == DOING_SPRINTF) {
				//RTFatal("not enough values to print in sprintf()");
			    //}
			    //else {
				//RTFatal("not enough values to print in printf()");
			    //}
			}
			if (s != 0) {
			    out_string[s] = wxT('\0');
			    wide_output(out_string);
			    s = 0;
			}
			f_elem = FormatItemW(cstring, f_elem, f_last, v_elem);
			if (IS_SEQUENCE(values_obj))  
			    v_elem++;
		    }
		}
		else {
		    if (s >= LOCAL_SPACE-1) {
			out_string[s] = wxT('\0');
			wide_output(out_string);
			s = 0;
		    }
		    out_string[s++] = c;
		}
		f_elem++; 
	    } /* end while */
	    
	    if (s != 0) {
		out_string[s] = 0;
		wide_output(out_string);
	    }
	    if (free_cs)
		delete[] cstring;
	}
    }
	/* sprintf */
	return get_sequence(collect);
	
}

extern "C" object WXEUAPI w_sprintf( object format, object values )
{
    return ESprintfW( format, values );
}

extern "C" object WXEUAPI to_utf8( object string )
{
	wxCharBuffer cb = get_string( string ).utf8_str();
	int len = strlen(cb);
	s1_ptr s = NewS1( len );
	object_ptr elem = s->base;
	for( size_t i = 0; i < len; i++ ){
		wxChar ch = cb[i];
		s->base[i+1] = static_cast<unsigned char>( ch );
	}
	
	return MAKE_SEQ( s );
}

extern "C" object WXEUAPI from_utf8( object utf8 ) 
{
	char * str = new char[LENGTH(utf8)+1];
	MakeCString( str, utf8 );
	
	object string = get_sequence( wxString::FromUTF8( str, LENGTH(utf8) ) );
	delete[] str;
	return string;
}

