// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "alldefs.h"
#include "wxeu.h"

#include "wx/spinctrl.h"
#include "wx/richtext/richtextctrl.h"
#include "wx/richtext/richtextbuffer.h"

extern "C"{
object WXEUAPI  new_wxTextCtrl(object params)
{
        NEW_CONTROL(wxTextCtrl)
}

object WXEUAPI xy_to_position( intptr_t ctrl, int x, int y )
{
        return NewDouble( ((wxTextCtrl*)ctrl)->XYToPosition( (intptr_t)x, (intptr_t)y ) );
}

object WXEUAPI position_to_xy( intptr_t ctrl, int position )
{
        long x, y;
        s1_ptr xy;
        
        if ( ((wxTextCtrlBase*)ctrl)->PositionToXY( position, &x, &y ) ) {
                xy = NewS1( 2 );
                xy->base[1] = x;
                xy->base[2] = y;
                
        } else {
                xy = NewS1( 0 );
        }
        
        return MAKE_SEQ( xy );
}

object WXEUAPI text_hit_test( intptr_t ctrl, int x, int y )
{
        wxTextCoord col, row;
        s1_ptr seq = NewS1( 2 );
        
        ((wxTextCtrlBase*)ctrl)->HitTest(wxPoint( x, y ), &col, &row);
        seq->base[1] = col;
        seq->base[2] = row;
        
        return MAKE_SEQ( seq );
}

object WXEUAPI  get_text_value(intptr_t ctrl)
{
        wxObject * obj = (wxObject*) ctrl;
        wxString s;
        if( obj->IsKindOf( CLASSINFO(wxTextCtrlBase) ) ){
                s = ((wxTextCtrlBase*)ctrl)->GetValue();
        }
        else if ( obj->IsKindOf( CLASSINFO(wxTextEntryDialog) ) ){
                s = ((wxTextEntryDialog*)ctrl)->GetValue();
        }

        return get_sequence( s );
}

void WXEUAPI  write_text(intptr_t ctrl, object text)
{
        ((wxTextCtrlBase*)ctrl)->WriteText( get_string( text ) );
        wxDeRefDS( text );
}

void WXEUAPI  set_text(intptr_t text, object val)
{

        wxObject * obj = (wxObject*) text;
        if( obj->IsKindOf( CLASSINFO(wxTextCtrlBase) ) ){
                ((wxTextCtrlBase*)text)->SetValue( get_string( val ) );
        }
        else if ( obj->IsKindOf( CLASSINFO(wxTextEntryDialog) ) ){
                ((wxTextEntryDialog*)text)->SetValue( get_string( val ) );
        }

        wxDeRefDS( val );

}

void WXEUAPI  set_text_selection(intptr_t text, int first, int last)
{
        ((wxTextCtrlBase*)text)->SetSelection( first, last );
}

object WXEUAPI  get_text_length(intptr_t text)
{
        return BOX_INT(((wxTextCtrlBase*)text)->GetLastPosition());
}

void WXEUAPI  select_all_text(intptr_t text)
{
        ((wxTextCtrlBase*)text)->SetSelection( 0, ((wxTextCtrlBase*)text)->GetLastPosition() );
}

void WXEUAPI  append_text(intptr_t text, object val)
{
        ((wxTextCtrlBase*)text)->AppendText( get_string( val ) );
        wxDeRefDS( val );

}

void WXEUAPI  wrap(intptr_t text, int width)
{
        ((wxStaticText*)text)->Wrap( width );
}

void WXEUAPI  set_label(int label, object text)
{
        ((wxStaticText*)label)->SetLabel( get_string( text ) );
        wxDeRefDS( text );
}

object WXEUAPI  new_wxSpinCtrl(object params)
{

        wxWindowID id = -1;
        wxString value = wxEmptyString;
        wxPoint pos = wxDefaultPosition;
        wxSize size = wxDefaultSize;
        intptr_t style = wxSP_ARROW_KEYS;
        int min = 0;
        int max = 100;
        int initial = 0;
        wxDeRefDS( params );
        int len = ((s1_ptr)SEQ_PTR(params))->length;
        switch( len ){
                case 11:
                        initial = get_int( params, 11 );
                case 10:
                        max = get_int( params, 10 );
                case 9:
                        min = get_int( params, 9 );
                case 8:
                        style = get_int( params, 8 );
                case 7:
                        size = wxSize( get_int( params, 6 ), get_int( params, 7 ) );
                case 6:
                case 5:
                        pos = wxPoint( get_int( params, 4 ), get_int( params, 5 ) );
                case 4:
                case 3:
                        value = get_string( params, 3 );
                case 2:
                        id = (wxWindowID) get_int( params, 2 );
                default:
                        wxSpinCtrl * spin = new wxSpinCtrl( (wxWindow *) get_int( params, 1 ),
                                id, value, pos, size, style, min, max, initial );

                        return BOX_INT( spin );
        }
}

object WXEUAPI  get_spin_value(intptr_t spin)
{
        int val = ((wxSpinCtrl*)spin)->GetValue();
        return BOX_INT( val );
}

void WXEUAPI  set_spin_value(intptr_t spin, int val)
{
        ((wxSpinCtrl*)spin)->SetValue( val );
}

object WXEUAPI new_wxTextAttr( object params )
{
        wxDeRefDS( params );
        return BOX_INT( new wxTextAttr() );
}

object WXEUAPI  get_text_attr_color(intptr_t attr)
{
        wxColour * color = new wxColour( ((wxTextAttr*)attr)->GetTextColour());
        return BOX_INT( color );
}

object WXEUAPI  get_text_attr_back_color(intptr_t attr)
{
        wxColour * color = new wxColour(((wxTextAttr*)attr)->GetBackgroundColour());
        return BOX_INT( color );
}

object WXEUAPI  get_text_attr_font(intptr_t attr)
{
        wxFont * font;
        font = new wxFont(((wxTextAttr*)attr)->GetFont());
        return BOX_INT( font );
}

void WXEUAPI  set_text_attr_color(intptr_t attr, intptr_t color)
{
        ((wxTextAttr*)attr)->SetTextColour( *(wxColour*)color );
}

void WXEUAPI  set_text_attr_back_color(intptr_t attr, intptr_t color)
{
        ((wxTextAttr*)attr)->SetBackgroundColour( *(wxColour*)color);
}

void WXEUAPI  set_text_attr_font(intptr_t attr, intptr_t font)
{
        ((wxTextAttr*)attr)->SetFont( *(wxFont*)font);
}

void WXEUAPI  set_text_style(intptr_t text, int first, int last, intptr_t attr)
{
        ((wxTextCtrlBase*)text)->SetStyle( first, last, *(wxTextAttr*)attr );
}

void WXEUAPI  set_default_text_style(intptr_t text, intptr_t attr)
{
        ((wxTextCtrlBase*)text)->SetDefaultStyle( *(wxTextAttr*)attr );
}

void WXEUAPI  copy_text(intptr_t text)
{
        ((wxTextCtrlBase*)text)->Copy();
}

void WXEUAPI  cut_text(intptr_t text)
{
        ((wxTextCtrlBase*)text)->Cut();
}

void WXEUAPI  paste_text(intptr_t text)
{
        ((wxTextCtrlBase*)text)->Paste();
}

void WXEUAPI  undo_text(intptr_t text)
{
        ((wxTextCtrlBase*)text)->Undo();
}

void WXEUAPI  redo_text(intptr_t text)
{
        ((wxTextCtrlBase*)text)->Redo();
}

void WXEUAPI  clear_text(intptr_t text)
{
        ((wxTextCtrlBase*)text)->Clear();
}

void WXEUAPI  set_maximum_text_length(intptr_t text, int max_text)
{
        ((wxTextCtrlBase*)text)->SetMaxLength( max_text );
}

object WXEUAPI can_copy_text( intptr_t text )
{
        return 1 & ((wxTextCtrlBase*)text)->CanCopy();
}

object WXEUAPI can_cut_text( intptr_t text )
{
        return 1 & ((wxTextCtrlBase*)text)->CanCut();
}

object WXEUAPI can_paste_text( intptr_t text )
{
        return 1 & ((wxTextCtrlBase*)text)->CanPaste();
}

object WXEUAPI can_redo_text( intptr_t text )
{
        return 1 & ((wxTextCtrlBase*)text)->CanRedo();
}

object WXEUAPI can_undo_text( intptr_t text )
{
        return 1 & ((wxTextCtrlBase*)text)->CanUndo();
}

void WXEUAPI discard_text_edits( intptr_t text )
{
        ((wxTextCtrlBase*)text)->DiscardEdits();
}

object WXEUAPI get_default_text_style( intptr_t text )
{
        return BOX_INT( new wxTextAttr( ((wxTextCtrlBase*)text)->GetDefaultStyle()) );
}

object WXEUAPI get_text_insertion_point( intptr_t text )
{
        return BOX_INT( ((wxTextCtrlBase*)text)->GetInsertionPoint() );
}

object WXEUAPI get_text_line_length( intptr_t text, int line )
{
        return BOX_INT( ((wxTextCtrlBase*)text)->GetLineLength( line ) );
}

object WXEUAPI get_text_lines( intptr_t text )
{
        return BOX_INT( ((wxTextCtrlBase*)text)->GetNumberOfLines() );
}

object WXEUAPI get_text_line( intptr_t text, int line )
{
        return get_sequence( ((wxTextCtrlBase*)text)->GetLineText( line ) );
}

object WXEUAPI get_text_range( intptr_t text, intptr_t text_from, intptr_t text_to )
{
        return get_sequence( ((wxTextCtrlBase*)text)->GetRange( text_from, text_to ) );
}

object WXEUAPI get_text_selection_range( intptr_t text )
{
        long from, to;
        ((wxTextCtrlBase*)text)->GetSelection( &from, &to );
        
        s1_ptr s = NewS1( 2 );
        s->base[1] = BOX_INT(from);
        s->base[2] = BOX_INT(to);
        return MAKE_SEQ( s );
}

object WXEUAPI get_text_selection( intptr_t text )
{
        return get_sequence( ((wxTextCtrlBase*)text)->GetStringSelection() );
}

object WXEUAPI get_text_style( intptr_t text, int position )
{
        wxTextAttr attr;
        
        s1_ptr s = NewS1( 2 );
        s->base[1] = 1 & ((wxTextCtrlBase*)text)->GetStyle( position, attr );
        s->base[2] = BOX_INT( new wxTextAttr( attr ) );
        return MAKE_SEQ( s );
}

object WXEUAPI is_text_editable( intptr_t text )
{
        return 1 & ((wxTextCtrlBase*)text)->IsEditable();
}

object WXEUAPI is_text_empty( intptr_t text )
{
        return 1 & ((wxTextCtrlBase*)text)->IsEmpty();
}

object WXEUAPI is_text_modified( intptr_t text )
{
        return 1 & ((wxTextCtrlBase*)text)->IsModified();
}

object WXEUAPI is_text_multi_line( intptr_t text )
{
        return 1 & ((wxTextCtrlBase*)text)->IsMultiLine();
}

object WXEUAPI is_text_single_line( intptr_t text )
{
        return 1 & ((wxTextCtrlBase*)text)->IsSingleLine();
}

object WXEUAPI load_text_file( intptr_t text, object file )
{
        wxDeRefDS( file );
        return 1 & ((wxTextCtrlBase*)text)->LoadFile( get_string( file ) );
}

void WXEUAPI mark_text_dirty( intptr_t text )
{
        ((wxTextCtrlBase*)text)->MarkDirty();
}

void WXEUAPI replace_text( intptr_t text, int pos_from, int pos_to, object str )
{
        ((wxTextCtrlBase*)text)->Replace( pos_from, pos_to, get_string( str ) );
        wxDeRefDS( str );
}

object WXEUAPI save_text_file( intptr_t text, object file )
{
        wxDeRefDS( file );
        return 1 & ((wxTextCtrlBase*)text)->SaveFile( get_string( file ) );
}

void WXEUAPI set_text_editable( intptr_t text, int editable )
{
        ((wxTextCtrlBase*)text)->SetEditable( editable );
}

void WXEUAPI set_text_insertion_point( intptr_t text, int pos )
{
        ((wxTextCtrlBase*)text)->SetInsertionPoint( pos );
}

void WXEUAPI set_text_insertion_point_end( intptr_t text )
{
        ((wxTextCtrlBase*)text)->SetInsertionPointEnd();
}

void WXEUAPI set_text_modified( intptr_t text, int modified )
{
        ((wxTextCtrlBase*)text)->SetModified( modified );
}

void WXEUAPI change_text_value( intptr_t text, object str )
{
        ((wxTextCtrlBase*)text)->ChangeValue( get_string( str ) );
        wxDeRefDS( str );
}

void WXEUAPI show_text_position( intptr_t text, int pos )
{
        ((wxTextCtrlBase*)text)->ShowPosition( pos );
}

object WXEUAPI new_wxTextAttrEx( object params )
{
        wxDeRefDS( params );
        return BOX_INT( new wxTextAttrEx() );
}

void WXEUAPI set_textattrex_char_style_name( intptr_t attr, object name )
{
        ((wxTextAttrEx*)attr)->SetCharacterStyleName( get_string( name ) );
}

void WXEUAPI set_textattrex_para_style_name( intptr_t attr, object name )
{
        ((wxTextAttrEx*)attr)->SetParagraphStyleName( get_string( name ) );
}

void WXEUAPI set_textattrex_list_style_name( intptr_t attr, object name )
{
        ((wxTextAttrEx*)attr)->SetListStyleName( get_string( name ) );
}

void WXEUAPI set_textattrex_para_spacing_after( intptr_t attr, int after )
{
        ((wxTextAttrEx*)attr)->SetParagraphSpacingAfter( after );
}

void WXEUAPI set_textattrex_para_spacing_before( intptr_t attr, int before )
{
        ((wxTextAttrEx*)attr)->SetParagraphSpacingBefore( before );
}

void WXEUAPI set_textattrex_line_spacing( intptr_t attr, int spacing )
{
        ((wxTextAttrEx*)attr)->SetLineSpacing( spacing );
}

void WXEUAPI set_textattrex_bullet_style( intptr_t attr, intptr_t style )
{
        ((wxTextAttrEx*)attr)->SetBulletStyle( style );
}

void WXEUAPI set_textattrex_bullet_number( intptr_t attr, int num )
{
        ((wxTextAttrEx*)attr)->SetBulletNumber( num );
}

void WXEUAPI set_textattrex_bullet_text( intptr_t attr, object text )
{
        wxDeRefDS( text );
        ((wxTextAttrEx*)attr)->SetBulletText( get_string( text ) );
}

void WXEUAPI set_textattrex_bullet_name( intptr_t attr, object name )
{
        wxDeRefDS( name );
        ((wxTextAttrEx*)attr)->SetBulletName( get_string( name ) );
}

void WXEUAPI set_textattrex_url_name( intptr_t attr, object url )
{
        wxDeRefDS( url );
        ((wxTextAttrEx*)attr)->SetURL( get_string( url ) );
}

void WXEUAPI set_textattrex_page_break( intptr_t attr, int page_break )
{
        ((wxTextAttrEx*)attr)->SetPageBreak( page_break );
}

void WXEUAPI set_textattrex_bullet_font( intptr_t attr, object font_name )
{
        wxDeRefDS( font_name );
        ((wxTextAttrEx*)attr)->SetBulletFont( get_string( font_name ) );
}

void WXEUAPI set_textattrex_effects( intptr_t attr, int effects )
{
        ((wxTextAttrEx*)attr)->SetTextEffects( effects );
}

void WXEUAPI set_textattrex_effect_flags( intptr_t attr, int flags )
{
        ((wxTextAttrEx*)attr)->SetTextEffectFlags( flags );
}

void WXEUAPI set_textattrex_outline_level( intptr_t attr, int level )
{
        ((wxTextAttrEx*)attr)->SetOutlineLevel( level );
}

object WXEUAPI get_textattrex_char_style( intptr_t attr )
{
        return get_sequence( ((wxTextAttrEx*)attr)->GetCharacterStyleName() );
}

object WXEUAPI get_textattrex_para_style( intptr_t attr )
{
        return get_sequence( ((wxTextAttrEx*)attr)->GetParagraphStyleName() );
}

object WXEUAPI get_textattrex_list_style( intptr_t attr )
{
        return get_sequence( ((wxTextAttrEx*)attr)->GetListStyleName() );
}

object WXEUAPI get_textattrex_para_spacing_before( intptr_t attr )
{
        return BOX_INT( ((wxTextAttrEx*)attr)->GetParagraphSpacingBefore() );
}

object WXEUAPI get_textattrex_para_spacing_after( intptr_t attr )
{
        return BOX_INT( ((wxTextAttrEx*)attr)->GetParagraphSpacingAfter() );
}

object WXEUAPI get_textattrex_line_spacing( intptr_t attr )
{
        return BOX_INT( ((wxTextAttrEx*)attr)->GetLineSpacing() );
}

object WXEUAPI get_textattrex_bullet_spacing( intptr_t attr )
{
        return BOX_INT( ((wxTextAttrEx*)attr)->GetBulletStyle() );
}

object WXEUAPI get_textattrex_bullet_number( intptr_t attr )
{
        return BOX_INT( ((wxTextAttrEx*)attr)->GetBulletNumber() );
}

object WXEUAPI get_textattrex_bullet_text( intptr_t attr )
{
        return get_sequence( ((wxTextAttrEx*)attr)->GetBulletText() );
}

object WXEUAPI get_textattrex_bullet_name( intptr_t attr )
{
        return get_sequence( ((wxTextAttrEx*)attr)->GetBulletName() );
}

object WXEUAPI get_textattrex_bullet_font( intptr_t attr )
{
        return get_sequence( ((wxTextAttrEx*)attr)->GetBulletFont() );
}

object WXEUAPI get_textattrex_url( intptr_t attr )
{
        return get_sequence( ((wxTextAttrEx*)attr)->GetURL() );
}

object WXEUAPI get_textattrex_effects( intptr_t attr )
{
        return BOX_INT( ((wxTextAttrEx*)attr)->GetTextEffects() );
}

object WXEUAPI get_textattrex_effect_flags( intptr_t attr )
{
        return BOX_INT( ((wxTextAttrEx*)attr)->GetTextEffectFlags() );
}

object WXEUAPI get_textattrex_outline_level( intptr_t attr )
{
        return BOX_INT( ((wxTextAttrEx*)attr)->GetOutlineLevel() );
}

object WXEUAPI has_font_weight_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasFontWeight();
}

object WXEUAPI has_font_size_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasFontSize();
}

object WXEUAPI has_font_italic_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasFontItalic();
}

object WXEUAPI has_font_underlined_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasFontUnderlined();
}

object WXEUAPI has_font_name_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasFontFaceName();
}

object WXEUAPI has_para_spacing_after_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasParagraphSpacingAfter();
}

object WXEUAPI has_para_spacing_before_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasParagraphSpacingBefore();
}

object WXEUAPI has_line_spacing_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasLineSpacing();
}

object WXEUAPI has_char_style_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasCharacterStyleName();
}

object WXEUAPI has_para_style_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasParagraphStyleName();
}

object WXEUAPI has_list_style_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasListStyleName();
}

object WXEUAPI has_bullet_style_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasBulletStyle();
}

object WXEUAPI has_bullet_number_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasBulletNumber();
}

object WXEUAPI has_bullet_text_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasBulletText( );
}

object WXEUAPI has_bullet_name_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasBulletName();
}

object WXEUAPI has_url_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasURL();
}

object WXEUAPI has_page_break_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasPageBreak();
}

object WXEUAPI has_text_effects_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasTextEffects();
}

object WXEUAPI has_text_effect_textattrex( intptr_t attr, int effect )
{
        return 1 & ((wxTextAttrEx*)attr)->HasTextEffect( effect );
}

object WXEUAPI has_outline_level_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->HasOutlineLevel();
}

object WXEUAPI is_char_style_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->IsCharacterStyle();
}

object WXEUAPI is_para_style_textattrex( intptr_t attr )
{
        return 1 & ((wxTextAttrEx*)attr)->IsParagraphStyle();
}

object WXEUAPI is_default_textattrex( intptr_t attr )
{
        return ((wxTextAttrEx*)attr)->IsDefault();
}

}
