// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "alldefs.h"
#include "wxeu.h"
#include "wx/treectrl.h"

#define ITEM( item ) wxTreeItemId( (void*) item )
#define ITEMVAL( itemval ) (void*) itemval

class euTreeItemId : public wxTreeItemId{
	public: 
		//euTreeItemId( wxTreeItemId & id ){ m_pItem = id.m_pItem; }
		euTreeItemId( wxTreeItemId id){ m_pItem = id.m_pItem; }
		intptr_t GetValue(){ return (intptr_t) m_pItem; }
		void SetValue( intptr_t item ){ m_pItem = (void *)item; }
};

extern "C"{
object WXEUAPI  new_wxTreeCtrl(object params)
{
	NEW_SIMPLE_CONTROL_STYLE( wxTreeCtrl, wxTR_HAS_BUTTONS | wxTR_LINES_AT_ROOT );

}

object WXEUAPI  add_tree_root(intptr_t tree, object text, int image, int selimage)
{
//wxTreeItemId AddRoot(const wxString& text, int image = -1, 
//int selImage = -1, wxTreeItemData* data = NULL)

	wxTreeItemId id = ((wxTreeCtrl*)tree)->AddRoot( get_string( text ),
		image, selimage );
	wxDeRefDS( text );
	return BOX_INT( ((euTreeItemId)id).GetValue() );
}

object WXEUAPI  add_tree_item(intptr_t tree, intptr_t parent, object text, int image, int selimage)
{
//wxTreeItemId AppendItem(const wxTreeItemId& parent, const wxString& text, 
//int image = -1, int selImage = -1, wxTreeItemData* data = NULL)
	wxTreeItemId parent_id( (void *) parent);

	wxTreeItemId id = (euTreeItemId) ((wxTreeCtrl*)tree)->AppendItem( parent_id,
		get_string( text ), image, selimage );
	wxDeRefDS( text );
	return BOX_INT( ((euTreeItemId)id).GetValue() );
}

void WXEUAPI  delete_tree_item(intptr_t tree, object item)
{

	//void Delete(const wxTreeItemId& item)
	if( IS_SEQUENCE(item) ){
		int len = LENGTH( item );
		for( int i = 1; i <= len; i++)
			((wxTreeCtrl *)tree)->Delete( wxTreeItemId( (void *) get_int( item, i )));
	}
	else{
		((wxTreeCtrl *)tree)->Delete( wxTreeItemId( (void *) get_int( item )));
	}
	wxDeRef( item );
}

void WXEUAPI  tree_imagelist(intptr_t tree, intptr_t il)
{
	((wxTreeCtrl*)tree)->SetImageList( (wxImageList*)il);
	((wxTreeCtrl*)tree)->SetStateImageList( (wxImageList*)il);
}

object WXEUAPI  get_tree_selection(intptr_t tree)
{
	euTreeItemId id = ((wxTreeCtrl*)tree)->GetSelection();
	return BOX_INT( id.GetValue() );
}

object WXEUAPI  get_tree_text(intptr_t tree, intptr_t id)
{
	return get_sequence(((wxTreeCtrl*)tree)->GetItemText( wxTreeItemId((void*)id)));
	
}

object WXEUAPI  get_tree_selected_text(intptr_t tree)
{
	wxTreeItemId id = ((wxTreeCtrl*)tree)->GetSelection();
	
	return get_sequence( ((wxTreeCtrl*)tree)->GetItemText(id) );
		
}

void WXEUAPI  expand_tree_item(intptr_t tree, intptr_t item)
{
	((wxTreeCtrl*)tree)->Expand( wxTreeItemId( (void*) item ) );
}

void WXEUAPI  set_tree_item_text(intptr_t tree, intptr_t item, object text)
{
	((wxTreeCtrl*)tree)->SetItemText( wxTreeItemId((void*)item), get_string( text ) );
	wxDeRefDS( text );
}

object WXEUAPI  get_tree_parent(intptr_t tree, intptr_t item)
{
	euTreeItemId id = ((wxTreeCtrl*)tree)->GetItemParent( wxTreeItemId( (void*)item) );
	return BOX_INT( id.GetValue() );
}

object WXEUAPI  get_tree_root(intptr_t tree)
{
	euTreeItemId root = ((wxTreeCtrl*)tree)->GetRootItem();
	return BOX_INT( root.GetValue() );
}

object WXEUAPI  valid_tree_item(intptr_t item)
{
	return wxTreeItemId( (void*)item).IsOk();
}

object WXEUAPI  get_tree_event_item(intptr_t event)
{
	euTreeItemId id = ((wxTreeEvent*)event)->GetItem();
	return BOX_INT( id.GetValue() );
}

void WXEUAPI  select_tree_item(intptr_t tree, intptr_t treeitem)
{
	((wxTreeCtrl*)tree)->SelectItem( wxTreeItemId( (void*) treeitem ) );
}

void WXEUAPI  delete_tree_children(intptr_t tree, intptr_t treeitem)
{
	((wxTreeCtrl*)tree)->DeleteChildren( wxTreeItemId( (void*)treeitem) );
}

void WXEUAPI collapse_tree_item( intptr_t tree, intptr_t item )
{
	((wxTreeCtrl*)tree)->Collapse( wxTreeItemId( (void*)item ));
}

void WXEUAPI collapse_all_tree_items( intptr_t tree )
{
	((wxTreeCtrl*)tree)->CollapseAll();
}

void WXEUAPI collapse_all_tree_children( intptr_t tree, intptr_t item )
{
	((wxTreeCtrl*)tree)->CollapseAllChildren( wxTreeItemId( (void*)item ) );
}

void WXEUAPI collapse_and_reset_tree_item( intptr_t tree, intptr_t item )
{
	((wxTreeCtrl*)tree)->CollapseAndReset( wxTreeItemId( (void*)item ) );
}

void WXEUAPI edit_tree_item( intptr_t tree, intptr_t item )
{
	((wxTreeCtrl*)tree)->EditLabel( wxTreeItemId( (void*)item ) );
}

void WXEUAPI ensure_tree_item_visual( intptr_t tree, intptr_t item )
{
	((wxTreeCtrl*)tree)->EnsureVisible( wxTreeItemId( (void*)item) );
}

void WXEUAPI expand_tree( intptr_t tree )
{
	((wxTreeCtrl*)tree)->ExpandAll();
}

void WXEUAPI expand_all_tree_children( intptr_t tree, intptr_t item )
{
	((wxTreeCtrl*)tree)->ExpandAllChildren( wxTreeItemId( (void*) item ) );
}

object WXEUAPI get_tree_item_bounding_rect( intptr_t tree, intptr_t item, int text_only )
{
	wxRect rect;
	s1_ptr s = NewS1( 2 );
	s1_ptr r = NewS1( 4 );
	
	bool ok = ((wxTreeCtrl*)tree)->GetBoundingRect( wxTreeItemId( (void*) item ), rect, text_only );
	s->base[1] = ok;
	s->base[2] = MAKE_SEQ( r );
	if( ok ){
		r->base[1] = rect.x;
		r->base[2] = rect.y;
		r->base[3] = rect.width;
		r->base[4] = rect.height;
	}
	
	return MAKE_SEQ( s );
}

object WXEUAPI get_tree_item_children_count( intptr_t tree, intptr_t item, int recursively )
{
	return BOX_INT( ((wxTreeCtrl*)tree)->GetChildrenCount( wxTreeItemId( (void*)item), recursively ) );
}

object WXEUAPI get_tree_item_count( intptr_t tree )
{
	return BOX_INT( ((wxTreeCtrl*)tree)->GetCount() );
}

object WXEUAPI get_tree_edit_control( intptr_t tree )
{
	return BOX_INT( ((wxTreeCtrl*)tree)->GetEditControl() );
}

object WXEUAPI get_first_visible_tree_item( intptr_t tree )
{
	return BOX_INT( new wxTreeItemId( ((wxTreeCtrl*)tree)->GetFirstVisibleItem() ));
}

object WXEUAPI get_tree_imagelist( intptr_t tree )
{
	return BOX_INT( ((wxTreeCtrl*)tree)->GetImageList() );
}

object WXEUAPI get_tree_indent( intptr_t tree )
{
	return BOX_INT( ((wxTreeCtrl*)tree)->GetIndent() );
}


object WXEUAPI get_tree_item_background_color( intptr_t tree, intptr_t item )
{
	return BOX_INT( new wxColour( 
		((wxTreeCtrl*)tree)->GetItemBackgroundColour( ITEM( item ) ) ) );
}

object WXEUAPI get_tree_item_font( intptr_t tree, intptr_t item )
{
	return BOX_INT( new wxFont( ((wxTreeCtrl*)tree)->GetItemFont( ITEM( item ) ) ) );
}

object WXEUAPI get_tree_item_image( intptr_t tree, intptr_t item, int which )
{
	return BOX_INT( ((wxTreeCtrl*)tree)->GetItemImage( ITEM(item), (wxTreeItemIcon) which ) );
}

object WXEUAPI get_tree_item_text_color( intptr_t tree, intptr_t item )
{
	return BOX_INT( new wxColour( ((wxTreeCtrl*)tree)->GetItemTextColour( ITEM(item) ) ) );
}

object WXEUAPI get_first_tree_child( intptr_t tree, intptr_t item )
{
	wxTreeItemIdValue * cookie = new wxTreeItemIdValue();
	
	euTreeItemId id = ((wxTreeCtrl*)tree)->GetFirstChild( ITEM(item), * cookie );
	
	s1_ptr s = NewS1( 2 );
	s->base[1] = BOX_INT( id.GetValue() );
	s->base[2] = BOX_INT( cookie );
	
	return MAKE_SEQ( s );
}

object WXEUAPI get_next_tree_child( intptr_t tree, intptr_t item, intptr_t cookie )
{
	wxTreeItemIdValue c = (wxTreeItemIdValue) cookie;
	wxTreeItemId id = ((wxTreeCtrl*)tree)->GetNextChild( ITEM( item ), c );
	
	return BOX_INT( id );
}

object WXEUAPI get_next_tree_sibling( intptr_t tree, intptr_t item )
{
	return BOX_INT(  ((wxTreeCtrl*)tree)->GetNextSibling( ITEM( item ) ) );
}

object WXEUAPI get_prev_tree_sibling( intptr_t tree, intptr_t item )
{
	return BOX_INT( ((wxTreeCtrl*)tree)->GetPrevSibling( ITEM( item ) ) );
}

object WXEUAPI get_next_visible_tree_item( intptr_t tree, intptr_t item )
{
	return BOX_INT( ((wxTreeCtrl*)tree)->GetNextVisible( ITEM( item ) ) );
}

object WXEUAPI get_prev_visible_tree_item( intptr_t tree, intptr_t item )
{
	return BOX_INT( ((wxTreeCtrl*)tree)->GetPrevVisible( ITEM( item ) ) );
}

object WXEUAPI get_tree_selections( intptr_t tree )
{
	wxArrayTreeItemIds selection;
	int count = ((wxTreeCtrl*)tree)->GetSelections( selection );
	
	s1_ptr s = NewS1( count );
	
	for( int i = 0; i < selection.GetCount(); i++){
		s->base[i+1] = BOX_INT( selection.Item( i ) );
	}
	
	return MAKE_SEQ( s );
}

object WXEUAPI tree_hit_test( intptr_t tree, int x, int y, int flags )
{
	return BOX_INT( ((wxTreeCtrl*)tree)->HitTest( wxPoint( x, y ), flags ) );
}

object WXEUAPI insert_tree_item_after( intptr_t tree, intptr_t parent, intptr_t previous, object text, int image, int sel_image )
{
	wxDeRefDS( text );
	return BOX_INT( ((wxTreeCtrl*)tree)->InsertItem( ITEM( parent ), ITEM(previous), get_string( text ),
		image, sel_image ) );
}

object WXEUAPI insert_tree_item_before( intptr_t tree, intptr_t parent, intptr_t before, object text, int image, int sel_image )
{
	wxDeRefDS( text );
	return BOX_INT( ((wxTreeCtrl*)tree)->InsertItem( ITEM(parent), before, get_string( text ), image, sel_image ) );
}

object WXEUAPI is_tree_item_bold( intptr_t tree, intptr_t item )
{
	return 1 & ((wxTreeCtrl*)tree)->IsBold( ITEM( item ) );
}

object WXEUAPI is_tree_empty( intptr_t tree )
{
	return 1 & ((wxTreeCtrl*)tree)->IsEmpty();
}

object WXEUAPI is_tree_item_expanded( intptr_t tree, intptr_t item )
{
	return 1 & ((wxTreeCtrl*)tree)->IsExpanded( ITEM( item ) );
}

object WXEUAPI is_tree_item_selected( intptr_t tree, intptr_t item )
{
	return 1 & ((wxTreeCtrl*)tree)->IsSelected( ITEM( item ) );
}

object WXEUAPI is_tree_item_visible( intptr_t tree, intptr_t item )
{
	return 1 & ((wxTreeCtrl*)tree)->IsVisible( ITEM(item) );
}

object WXEUAPI tree_item_has_children( intptr_t tree, intptr_t item )
{
	return 1 & ((wxTreeCtrl*)tree)->ItemHasChildren( ITEM( item ) );
}

object WXEUAPI prepend_tree_item( intptr_t tree, intptr_t parent, object text, int image, int sel_image )
{
	wxDeRefDS( text );
	return BOX_INT( ((wxTreeCtrl*)tree)->PrependItem( ITEM(parent), get_string(text), image, sel_image ) );
}

void WXEUAPI scroll_to_tree_item( intptr_t tree, intptr_t item )
{
	((wxTreeCtrl*)tree)->ScrollTo( ITEM( item ) );
}

void WXEUAPI set_tree_indent( intptr_t tree, int indent )
{
	((wxTreeCtrl*)tree)->SetIndent( indent );
}

void WXEUAPI set_tree_item_background( intptr_t tree, intptr_t item, int color )
{
	((wxTreeCtrl*)tree)->SetItemBackgroundColour( ITEM(item), *(wxColour*)color );
}

void WXEUAPI set_tree_item_bold( intptr_t tree, intptr_t item, int bold )
{
	((wxTreeCtrl*)tree)->SetItemBold( ITEM(item), bold );
}

void WXEUAPI set_tree_item_drop_highlight( intptr_t tree, intptr_t item, int highlight )
{
	((wxTreeCtrl*)tree)->SetItemDropHighlight( ITEM(item), highlight );
}

void WXEUAPI set_tree_item_font( intptr_t tree, intptr_t item, int font )
{
	((wxTreeCtrl*)tree)->SetItemFont( ITEM(item), *(wxFont*)font );
}

void WXEUAPI set_tree_item_has_children( intptr_t tree, intptr_t item, int has_children )
{
	((wxTreeCtrl*)tree)->SetItemHasChildren( ITEM( item ), has_children );
}

void WXEUAPI set_tree_item_image( intptr_t tree, intptr_t item, int image, int which )
{
	((wxTreeCtrl*)tree)->SetItemImage( ITEM(item), image, (wxTreeItemIcon) which );
}

void WXEUAPI set_tree_item_text_color( intptr_t tree, intptr_t item, intptr_t color )
{
	((wxTreeCtrl*)tree)->SetItemTextColour( ITEM(item), *(wxColour*)color );
}

void WXEUAPI sort_tree_children( intptr_t tree, intptr_t item )
{
	((wxTreeCtrl*)tree)->SortChildren( ITEM(item) );
}

void WXEUAPI toggle_tree_item( intptr_t tree, intptr_t item )
{
	((wxTreeCtrl*)tree)->Toggle( ITEM(item) );
}

void WXEUAPI toggle_tree_item_selection( intptr_t tree, intptr_t item )
{
	((wxTreeCtrl*)tree)->ToggleItemSelection( ITEM(item) );
}

void WXEUAPI unselect_tree_item( intptr_t tree, object items )
{
	if( IS_SEQUENCE( items ) && LENGTH( items ) ){
		for( int i = 1; i <= LENGTH( items ); i++){
			((wxTreeCtrl*)tree)->UnselectItem( ITEM( get_int( items, i ) ) );
		}
	}
	else
		((wxTreeCtrl*)tree)->UnselectAll();
	
	wxDeRef( items );
}

}

