// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "alldefs.h"
#include "wxeu.h"
#include <wx/xml/xml.h>

extern "C"{

object WXEUAPI new_wxXmlDocument( object params ){
	wxDeRefDS( params );
	int len = LENGTH( params );
	wxString encoding( wxT("UTF-8") );
	
	switch( len ){
		case 0:
			return BOX_INT( new wxXmlDocument() );
			
		case 2:
			encoding = get_string( params, 2 );
		
		default:
			return BOX_INT( new wxXmlDocument( get_string( params, 1 ), encoding ) );
	}
}

object WXEUAPI detatch_xml_root( intptr_t xmldoc )
{
	return BOX_INT( ((wxXmlDocument*)xmldoc)->DetachRoot() );
}

object WXEUAPI get_xml_encoding( intptr_t xmldoc )
{
#if !wxUSE_UNICODE
	return get_sequence( ((wxXmlDocument*)xmldoc)->GetEncoding() );
#else
	return get_sequence( wxEmptyString );
#endif
}

object WXEUAPI get_xml_file_encoding( intptr_t xmldoc )
{
	return get_sequence( ((wxXmlDocument*)xmldoc)->GetFileEncoding() );
}

object WXEUAPI get_xml_root( intptr_t xmldoc )
{
	return BOX_INT( ((wxXmlDocument*)xmldoc)->GetRoot());
}

object WXEUAPI get_xml_version( intptr_t xmldoc )
{
	return get_sequence( ((wxXmlDocument*)xmldoc)->GetVersion() );
}

object WXEUAPI xml_is_ok( intptr_t xmldoc )
{
	return 1 & ((wxXmlDocument*)xmldoc)->IsOk();
}

object WXEUAPI load_xml( intptr_t xmldoc, object filename, object options )
{
	wxDeRefDS( filename );
	wxDeRef( options );
	
	wxString encoding = wxT("UTF-8");
	int flags = wxXMLDOC_NONE;
	
	if( IS_SEQUENCE( options ) ){
		switch( LENGTH( options ) ){
			case 2:
				flags = get_int( options, 2 );
			
			case 1:
				encoding = get_string( options, 1 );
			
			default:
				break;
		}
	}
	
	return 1 & ((wxXmlDocument*)xmldoc)->Load( get_string(filename), encoding, flags );
}

object WXEUAPI save_xml( intptr_t xmldoc, object filename, int indentstep )
{
	wxDeRefDS( filename );
	return 1 & ((wxXmlDocument*)xmldoc)->Save( get_string( filename ), indentstep );
}

void WXEUAPI set_xml_encoding( intptr_t xmldoc, object enc )
{
	wxDeRefDS( enc );
#if !wxUSE_UNICODE
	((wxXmlDocument*)xmldoc)->SetEncoding( get_string( enc ) );
#endif
}

void WXEUAPI set_xml_file_encoding( intptr_t xmldoc, object enc )
{
	wxDeRefDS( enc );
	((wxXmlDocument*)xmldoc)->SetFileEncoding( get_string( enc ) );
}

void WXEUAPI set_xml_root( intptr_t xmldoc, intptr_t node )
{
	((wxXmlDocument*)xmldoc)->SetRoot( (wxXmlNode*) node );
}

void WXEUAPI set_xml_version( intptr_t xmldoc, object version )
{
	wxDeRefDS( version );
	((wxXmlDocument*)xmldoc)->SetVersion( get_string( version ) );
}

object WXEUAPI new_wxXmlNode( object params )
{
	wxDeRefDS( params );
	wxString content = wxEmptyString;
	wxXmlProperty * props = NULL;
	wxXmlNode * next = NULL;
	
	switch( LENGTH( params ) ){
		case 6:
			next = (wxXmlNode*) get_int( params, 6 );
		
		case 5:
			props = (wxXmlProperty*) get_int( params, 5 );
			
		case 4:
			content = get_string( params, 4 );
			
		default:
			return BOX_INT( new wxXmlNode(
				(wxXmlNode*) get_int( params, 1 ),    // parent
				(wxXmlNodeType) get_int( params, 2 ), // type
				get_string( params, 3 ),              // name
				content,
				props,
				next ) );
	}
}

void WXEUAPI delete_wxXmlNode( object obj ){
	DELETE_INSTANCE( wxXmlNode, obj )
}

void WXEUAPI add_xml_node_child( intptr_t node, intptr_t child )
{
	((wxXmlNode*)node)->AddChild( (wxXmlNode*) child );
}

void WXEUAPI add_xml_node_property( intptr_t node, object name, object val )
{
	((wxXmlNode*)node)->AddProperty( get_string( name ), get_string( val ));
	wxDeRefDS( name );
	wxDeRefDS( val );
}

object WXEUAPI delete_xml_node_property( intptr_t node, object name )
{
	wxDeRefDS( name );
	return 1 & ((wxXmlNode*)node)->DeleteProperty( get_string( name ) );
}

object WXEUAPI get_xml_node_children( intptr_t node )
{
	return BOX_INT( ((wxXmlNode*)node)->GetChildren() );
}

object WXEUAPI get_xml_content( intptr_t node )
{
	return get_sequence( ((wxXmlNode*)node)->GetContent() );
}

object WXEUAPI get_xml_node_depth( intptr_t node, int grandparent )
{
	return BOX_INT( ((wxXmlNode*)node)->GetDepth( (wxXmlNode*) grandparent ) );
}

object WXEUAPI get_xml_node_content( intptr_t node )
{
	return get_sequence( ((wxXmlNode*)node)->GetNodeContent() );
}

object WXEUAPI get_xml_node_name( intptr_t node )
{
	return get_sequence( ((wxXmlNode*)node)->GetName() );
}

object WXEUAPI get_next_xml_node( intptr_t node )
{
	return BOX_INT( ((wxXmlNode*)node)->GetNext() );
}

object WXEUAPI get_parent_xml_node( intptr_t node )
{
	return BOX_INT( ((wxXmlNode*)node)->GetParent() );
}

object WXEUAPI get_xml_node_prop_val( intptr_t node, object prop_name, object default_val )
{
	wxDeRefDS( prop_name );
	wxDeRefDS( default_val );
	
	return get_sequence( ((wxXmlNode*)node)->GetPropVal( get_string( prop_name ), get_string( default_val ) ));
}

object WXEUAPI get_xml_node_properties( intptr_t node )
{
	return BOX_INT( ((wxXmlNode*)node)->GetProperties() );
}

object WXEUAPI get_xml_node_type( intptr_t node )
{
	return BOX_INT( ((wxXmlNode*)node)->GetType() );
}

object WXEUAPI xml_node_has_prop( intptr_t node, object prop_name )
{
	wxDeRefDS( prop_name );
	return 1 & ((wxXmlNode*)node)->HasProp( get_string( prop_name ) );
}

object WXEUAPI insert_xml_child( intptr_t node, intptr_t child, int before_node ){
	return 1 & ((wxXmlNode*)node)->InsertChild( (wxXmlNode*)child, (wxXmlNode*)before_node );
}

object WXEUAPI is_xml_whitespace_only( intptr_t node )
{
	return 1 & ((wxXmlNode*)node)->IsWhitespaceOnly();
}

object WXEUAPI remove_xml_node_child( intptr_t node, intptr_t child )
{
	return 1 & ((wxXmlNode*)node)->RemoveChild( (wxXmlNode*) child );
}

void WXEUAPI set_xml_node_children( intptr_t node, intptr_t child )
{
	((wxXmlNode*)node)->SetChildren( (wxXmlNode*) child );
}

void WXEUAPI set_xml_node_content( intptr_t node, object con )
{
	wxDeRefDS( con );
	((wxXmlNode*)node)->SetContent( get_string( con ) );
}

void WXEUAPI set_xml_node_name( intptr_t node, object name )
{
	wxDeRefDS( name );
	((wxXmlNode*)node)->SetName( get_string( name ) );
}

void WXEUAPI set_next_xml_node( intptr_t node, intptr_t next )
{
	((wxXmlNode*)node)->SetNext( (wxXmlNode*)next );
}

void WXEUAPI set_parent_xml_mode( intptr_t node, intptr_t parent )
{
	((wxXmlNode*)node)->SetParent( (wxXmlNode*) parent );
}

void WXEUAPI set_xml_node_properties( intptr_t node, intptr_t prop )
{
	((wxXmlNode*)node)->SetProperties( (wxXmlProperty*) prop );
}

void WXEUAPI set_xml_node_type( intptr_t node, int type )
{
	((wxXmlNode*)node)->SetType( (wxXmlNodeType) type );
}

object WXEUAPI new_wxXmlProperty( object params )
{
	wxDeRefDS( params );
	wxXmlProperty * next = LENGTH( params ) == 3 ? (wxXmlProperty*) get_int( params, 3 ) : NULL;
	
	return BOX_INT( new wxXmlProperty( get_string( params, 1 ), get_string( params, 2 ), next ) );
}

void WXEUAPI delete_wxXmlProperty( intptr_t obj )
{
	DELETE_INSTANCE( wxXmlProperty, obj );
}

object WXEUAPI get_xml_prop_name( intptr_t prop )
{
	return get_sequence( ((wxXmlProperty*)prop)->GetName() );
}

object WXEUAPI get_next_xml_prop( intptr_t prop )
{
	return BOX_INT( ((wxXmlProperty*)prop)->GetNext() );
}

object WXEUAPI get_xml_prop_value( intptr_t prop )
{
	return get_sequence( ((wxXmlProperty*)prop)->GetValue() );
}

void WXEUAPI set_xml_prop_name( intptr_t prop, object name )
{
	wxDeRefDS( name );
	((wxXmlProperty*)prop)->SetName( get_string( name ) );
}

void WXEUAPI set_next_xml_prop( intptr_t prop, intptr_t next )
{
	((wxXmlProperty*)prop)->SetNext( (wxXmlProperty*) next );
}

void WXEUAPI set_xml_prop_value( intptr_t prop, object val )
{
	((wxXmlProperty*)prop)->SetValue( get_string( val ) );
}


} // extern "C"
