--/makedoc title wxEuphoria Documentation

--Site Header
------------------------------
--/topic Site Header
--/info
--&nbsp;
--<div id="site_header">
--<div id="site_heading"><a href="http://wxeuphoria.sourceforge.net/">wxEuphoria v0.17.0</a></div>
--<div>Wrapper for <a href="http://www.wxwidgets.org/">wxWidgets</a> by
-- <a href="mailto:mattlewis@users.sourceforge.net">Matt Lewis</a>
-- and <a href="mailto:ghaberek@users.sourceforge.net">Greg Haberek</a><br/>
-- for use with the <a href="http://www.rapideuphoria.com/">Euphoria Programming Language</a></div>
--</div>

--Introduction
------------------------------
--/topic Introduction
--/info
--
--wxEuphoria is an implementation of <a href="http://www.wxwidgets.org">wxWidgets</a> by Julian
--Smart, et. al., for the Euphoria programming language.  wxWidgets is a cross platform GUI library.
--The wxWidgets library is compiled into a DLL/SO, and is linked to by Euphoria code.
--
--/b "This version of wxEuphoria requires library build #17."
--
--wxEuphoria is licensed under the wxWidgets license (see the license folder).<br><br>
--/hilite
--DISCLAIMER:
--THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
--INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
--AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
--AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
--OR CONSEQUENTIAL DAMGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
--SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFIT; OR BUSINESS
--INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
--CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
--ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
--POSSIBILITY OF SUCH DAMAGE.
--/lolite
--<h3>Other Licensing Issues</h3>
--<ul>
--/li Please see /b "license\tiff-copyright.txt" for libtiff license details
--/li This software is based in part on the work of the Independent JPEG Group. See /b "license\jpeg-readme.txt"
--/li Regarding the implementation of /b wxRegEx:
--
--<p>Copyright 1992, 1993, 1994, 1997 Henry Spencer.  All rights reserved.
--This software is not subject to any license of the American Telephone
--and Telegraph Company or of the Regents of the University of California.</p>
--
--<p>Permission is granted to anyone to use this software for any purpose on
--any computer system, and to alter it and redistribute it, subject
--to the following restrictions:</p>
--
--<ol>
--/li The author is not responsible for the consequences of use of this
--   software, no matter how awful, even if they arise from flaws in it.
--/li The origin of this software must not be misrepresented, either by
--explicit claim or by omission.  Since few users ever read sources,
--credits must appear in the documentation.
--/li Altered versions must be plainly marked as such, and must not be
--misrepresented as being the original software.  Since few users
--ever read sources, credits must appear in the documentation.
--/li This notice may not be removed or altered.
--</ol></ul>
--
--<h3>Files Included</h3>
--<ul>
--/li /b Documentation<ul>
--/li see wxEuphoria/docs
--/li install.txt
--/li readme.txt
--</ul>
--/li /b Licenses<ul>
--/li gpl
--/li jpeg-readme.txt
--/li lgpl.txt
--/li licence.txt
--/li tiff-copyright.txt
--</ul>
--/li /b Library<ul>
--/li wxeud.e
--</ul>
--/li /b Demos<ul>
--/li see the demo subdirectory
--</ul>
--/li /b "Files not included" (in source only distribution)<ul>
--/li libwxeu.dll
--/li libwxeu.so.$build
--/li wxWidgets source code and documentation<br/>(go to <a href="http://www.wxwidgets.org">www.wxwidgets.org</a>)
--/li Euphoria v4.0 or later
--</ul>
--</ul>

--/topic Versioning
--/info
-- Determining which version of wxEuphoria you are using
--
-- At parse / compile time, there are several words defined for use with ifdefs.
-- These follow the form:
--
-- /li WXEU_$MAJOR
-- /li WXEU_$MAJOR_$MINOR
-- /li WXEU_$MAJOR_$MINOR_$PATCH
--
-- In addition, at run time, there is a public constant:
-- <eucode>
-- wxEuphoria_Version = { MAJOR, MINOR, PATCH }
-- </eucode>
-- The MAJOR, MINOR and PATCH elements are integers representing the current version
-- of wxEuphoria.
--
-- You can also use the public constant wxNEED_BUILD, which is used to
-- ensure that an older .dll or .so is not used.

with define WXEU_0
with define WXEU_0_16
with define WXEU_0_16_1
with define WXEU_0_17

without trace
without type_check
include std/dll.e
include std/error.e
include std/get.e
include std/sequence.e
include std/text.e

--/topic Miscellaneous Routines
--/func wx_or_all( sequence a )
--
--Executes a bitwise /b or operation on a sequence of atoms.  Use this to convert
--from a C/C++ style.
--
--Example:
--/code
--
--    THING1|THING2|THING3
--
--    wx_or_all( THING1 & THING2 & THING3 )
--
--/endcode
public function wx_or_all( sequence a )
    atom x
    x = a[1]
    for i = 2 to length( a ) do
        x = or_bits( x, a[i] )
    end for
    return x
end function



public constant
wxEuphoria_Version = {0,17,0},
wxEuphoria_Platform = platform(),
wxNEED_BUILD = 17

atom wxWidgets

ifdef WINDOWS then
        wxWidgets = dll:open_dll( "libwxeu.2.9.dll" )
elsedef
        -- This is the one we really want
    wxWidgets = dll:open_dll( sprintf("libwxeu.so.2.9.%d", wxNEED_BUILD ) )
    if wxWidgets = 0 then
        -- But a later one should be fine, too
        wxWidgets = dll:open_dll( "libwxeu.so" )
    end if
end ifdef

if wxWidgets = 0 then
    puts(2, "Could not open wxEuphoria library.  Press enter to abort.\n" )
    if object(gets(0)) then
        abort(1)
    end if
end if

-- These functions were trimmed down to remove the redundant "dll"
-- object. Doing so has dropped a about 15 KB from the library.

function wx_define_c_proc( object proc, object args )

    atom id

    ifdef WINDOWS then
        proc = '+' & proc
    end ifdef

    id = dll:define_c_proc( wxWidgets, proc, args )
    ifdef DEBUG then
                if id = -1 then
                        printf( 2, "Failed to link procedure: %s\n", {proc} )
                end if
    end ifdef
    return id
end function

function wx_define_c_func( object func, object args, object ret )

    atom id

    ifdef WINDOWS then
        func = '+' & func
    end ifdef

    id = dll:define_c_func( wxWidgets, func, args, ret )
    ifdef DEBUG then
                if id = -1 then
                        printf( 2, "Failed to link function: %s\n", {func} )
                end if
    end ifdef

    return id
end function

function wx_define_c_var( object var )

    atom id

    id = dll:define_c_var( wxWidgets, var )
        ifdef DEBUG then
                if id = -1 then
                        printf( 2, "Failed to link variable: %s\n", {var} )
                end if
    end ifdef

    return id
end function

public constant
	WX_ABOUT_DLG_ADD_ARTIST                = wx_define_c_proc( "about_dlg_add_artist", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_ABOUT_DLG_ADD_DEVELOPER             = wx_define_c_proc( "about_dlg_add_developer", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_ABOUT_DLG_ADD_DOC_WRITER            = wx_define_c_proc( "about_dlg_add_doc_writer", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_ABOUT_DLG_ADD_TRANSLATOR            = wx_define_c_proc( "about_dlg_add_translator", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_ABOUT_DLG_SET_COPYRIGHT             = wx_define_c_proc( "about_dlg_set_copyright", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_ABOUT_DLG_SET_DESCRIPTION           = wx_define_c_proc( "about_dlg_set_description", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_ABOUT_DLG_SET_ICON                  = wx_define_c_proc( "about_dlg_set_icon", {dll:C_POINTER,dll:C_POINTER} ),
	WX_ABOUT_DLG_SET_LICENSE               = wx_define_c_proc( "about_dlg_set_license", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_ABOUT_DLG_SET_NAME                  = wx_define_c_proc( "about_dlg_set_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_ABOUT_DLG_SET_VERSION               = wx_define_c_proc( "about_dlg_set_version", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_ABOUT_DLG_SET_WEB_SITE              = wx_define_c_proc( "about_dlg_set_web_site", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_ABOUT_DLG_SHOW                      = wx_define_c_proc( "about_dlg_show", {dll:C_POINTER} ),
	WX_ACCEPT_FILES                        = wx_define_c_proc( "accept_files", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_ACTIVATE_CHILD                      = wx_define_c_proc( "activate_child", {dll:C_POINTER} ),
	WX_ACTIVATE_NEXT_MDI                   = wx_define_c_proc( "activate_next_mdi", {dll:C_POINTER} ),
	WX_ACTIVATE_PREV_MDI                   = wx_define_c_proc( "activate_prev_mdi", {dll:C_POINTER} ),
	WX_ADD_BUTTON                          = wx_define_c_proc( "add_button", {dll:C_POINTER,dll:C_POINTER} ),
	WX_ADD_IMAGE                           = wx_define_c_func( "add_image", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_ADD_ITEM                            = wx_define_c_proc( "add_item", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_ADD_MENU                            = wx_define_c_proc( "add_menu", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT} ),
	WX_ADD_PAGE                            = wx_define_c_proc( "add_page", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT} ),
	WX_ADD_PANE_SIMPLE                     = wx_define_c_proc( "add_pane_simple", {dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:E_OBJECT} ),
	WX_ADD_PANE_WITH_INFO                  = wx_define_c_proc( "add_pane_with_info", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_ADD_PANE_WITH_POS                   = wx_define_c_proc( "add_pane_with_pos", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_ADD_PARAGRAPH                       = wx_define_c_func( "add_paragraph", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_ADD_RICH_IMAGE                      = wx_define_c_func( "add_rich_image", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_ADD_SEPARATOR                       = wx_define_c_proc( "add_separator", {dll:C_POINTER} ),
	WX_ADD_SIZER_TO_SIZER                  = wx_define_c_proc( "add_sizer_to_sizer", {dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_ADD_TOOL                            = wx_define_c_func( "add_tool", {dll:C_POINTER,dll:C_INT,dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_ADD_TREE_ITEM                       = wx_define_c_func( "add_tree_item", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_ADD_TREE_ROOT                       = wx_define_c_func( "add_tree_root", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_ADD_WINDOW_TO_SIZER                 = wx_define_c_proc( "add_window_to_sizer", {dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_ADD_XML_NODE_CHILD                  = wx_define_c_proc( "add_xml_node_child", {dll:C_POINTER,dll:C_POINTER} ),
	WX_ADD_XML_NODE_PROPERTY               = wx_define_c_proc( "add_xml_node_property", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT} ),
	WX_ADVANCE_SELECTION                   = wx_define_c_proc( "advance_selection", {dll:C_POINTER,dll:C_INT} ),
	WX_ANY_ADDRESS                         = wx_define_c_proc( "any_address", {dll:C_POINTER} ),
	WX_APPEND_COLS                         = wx_define_c_proc( "append_cols", {dll:C_POINTER,dll:C_INT} ),
	WX_APPEND_ROWS                         = wx_define_c_proc( "append_rows", {dll:C_POINTER,dll:C_INT} ),
	WX_APPEND_SEPARATOR                    = wx_define_c_proc( "append_separator", {dll:C_POINTER} ),
	WX_APPEND_TEXT                         = wx_define_c_proc( "append_text", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_APPLY_ALIGNMENT_TO_SELECTION        = wx_define_c_proc( "apply_alignment_to_selection", {dll:C_POINTER,dll:C_INT} ),
	WX_APPLY_BOLD_TO_SELECTION             = wx_define_c_proc( "apply_bold_to_selection", {dll:C_POINTER} ),
	WX_APPLY_ITALIC_TO_SELECTION           = wx_define_c_proc( "apply_italic_to_selection", {dll:C_POINTER} ),
	WX_APPLY_RICH_ATTR                     = wx_define_c_proc( "apply_rich_attr", {dll:C_POINTER,dll:C_POINTER} ),
	WX_APPLY_RICH_STYLE_FROM_DIALOG        = wx_define_c_proc( "apply_rich_style_from_dialog", {dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_APPLY_RICH_STYLE_SHEET              = wx_define_c_proc( "apply_rich_style_sheet", {dll:C_POINTER,dll:C_POINTER} ),
	WX_APPLY_RICH_STYLE                    = wx_define_c_proc( "apply_rich_style", {dll:C_POINTER,dll:C_POINTER} ),
	WX_APPLY_UNDERLINE_TO_SELECTION        = wx_define_c_proc( "apply_underline_to_selection", {dll:C_POINTER} ),
	WX_APP_YIELD                           = wx_define_c_func( "app_yield", {dll:C_POINTER}, E_OBJECT ),
	WX_ARRANGE_MDI_ICONS                   = wx_define_c_proc( "arrange_mdi_icons", {dll:C_POINTER} ),
	WX_AUI_BEST_SIZE                       = wx_define_c_proc( "aui_best_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_AUI_BOTTOM_DOCKABLE                 = wx_define_c_proc( "aui_bottom_dockable", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_BOTTOM                          = wx_define_c_proc( "aui_bottom", {dll:C_POINTER} ),
	WX_AUI_CAPTION_VISIBLE                 = wx_define_c_proc( "aui_caption_visible", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_CAPTION                         = wx_define_c_proc( "aui_caption", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_AUI_CENTER_PANE                     = wx_define_c_proc( "aui_center_pane", {dll:C_POINTER} ),
	WX_AUI_CENTER                          = wx_define_c_proc( "aui_center", {dll:C_POINTER} ),
	WX_AUI_CLOSE_BUTTON                    = wx_define_c_proc( "aui_close_button", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_DEFAULT_PANE                    = wx_define_c_proc( "aui_default_pane", {dll:C_POINTER} ),
	WX_AUI_DESTROY_ON_CLOSE                = wx_define_c_proc( "aui_destroy_on_close", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_DIRECTION                       = wx_define_c_proc( "aui_direction", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_DOCKABLE                        = wx_define_c_proc( "aui_dockable", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_DOCK                            = wx_define_c_proc( "aui_dock", {dll:C_POINTER} ),
	WX_AUI_FIXED                           = wx_define_c_proc( "aui_fixed", {dll:C_POINTER} ),
	WX_AUI_FLOATABLE                       = wx_define_c_proc( "aui_floatable", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_FLOATING_POSITION               = wx_define_c_proc( "aui_floating_position", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_AUI_FLOATING_SIZE                   = wx_define_c_proc( "aui_floating_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_AUI_FLOAT                           = wx_define_c_proc( "aui_float", {dll:C_POINTER} ),
	WX_AUI_GET_SELECTION                   = wx_define_c_func( "aui_get_selection", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_GRIPPER_TOP                     = wx_define_c_proc( "aui_gripper_top", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_GRIPPER                         = wx_define_c_proc( "aui_gripper", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_HAS_BORDER                      = wx_define_c_func( "aui_has_border", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_HAS_CAPTION                     = wx_define_c_func( "aui_has_caption", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_HAS_CLOSE_BUTTON                = wx_define_c_func( "aui_has_close_button", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_HAS_FLAG                        = wx_define_c_func( "aui_has_flag", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_AUI_HAS_GRIPPER_TOP                 = wx_define_c_func( "aui_has_gripper_top", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_HAS_GRIPPER                     = wx_define_c_func( "aui_has_gripper", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_HAS_MAXIMIZE_BUTTON             = wx_define_c_func( "aui_has_maximize_button", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_HAS_MINIMIZE_BUTTON             = wx_define_c_func( "aui_has_minimize_button", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_HAS_PIN_BUTTON                  = wx_define_c_func( "aui_has_pin_button", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_HIDE                            = wx_define_c_proc( "aui_hide", {dll:C_POINTER} ),
	WX_AUI_IS_BOTTOM_DOCKABLE              = wx_define_c_func( "aui_is_bottom_dockable", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_DOCKED                       = wx_define_c_func( "aui_is_docked", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_FIXED                        = wx_define_c_func( "aui_is_fixed", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_FLOATABLE                    = wx_define_c_func( "aui_is_floatable", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_FLOATING                     = wx_define_c_func( "aui_is_floating", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_LEFT_DOCKABLE                = wx_define_c_func( "aui_is_left_dockable", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_MOVABLE                      = wx_define_c_func( "aui_is_movable", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_OK                           = wx_define_c_func( "aui_is_ok", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_RESIZABLE                    = wx_define_c_func( "aui_is_resizable", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_RIGHT_DOCKABLE               = wx_define_c_func( "aui_is_right_dockable", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_SHOWN                        = wx_define_c_func( "aui_is_shown", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_TOOLBAR                      = wx_define_c_func( "aui_is_toolbar", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_IS_TOP_DOCKABLE                 = wx_define_c_func( "aui_is_top_dockable", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_LAYER                           = wx_define_c_proc( "aui_layer", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_LEFT_DOCKABLE                   = wx_define_c_proc( "aui_left_dockable", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_LEFT                            = wx_define_c_proc( "aui_left", {dll:C_POINTER} ),
	WX_AUI_MAXIMIZE_BUTTON                 = wx_define_c_proc( "aui_maximize_button", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_MAX_SIZE                        = wx_define_c_proc( "aui_max_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_AUI_MINIMIZE_BUTTON                 = wx_define_c_proc( "aui_minimize_button", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_MIN_SIZE                        = wx_define_c_proc( "aui_min_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_AUI_MOVABLE                         = wx_define_c_proc( "aui_movable", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_NAME                            = wx_define_c_proc( "aui_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_AUI_NOTEBOOK_ADD_PAGE               = wx_define_c_func( "aui_notebook_add_page", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_POINTER}, E_OBJECT ),
	WX_AUI_NOTEBOOK_DELETE_PAGE            = wx_define_c_func( "aui_notebook_delete_page", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_AUI_NOTEBOOK_GET_ART_PROVIDER       = wx_define_c_func( "aui_notebook_get_art_provider", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_NOTEBOOK_GET_PAGE_BITMAP        = wx_define_c_func( "aui_notebook_get_page_bitmap", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_AUI_NOTEBOOK_GET_PAGE_COUNT         = wx_define_c_func( "aui_notebook_get_page_count", {dll:C_POINTER}, E_OBJECT ),
	WX_AUI_NOTEBOOK_GET_PAGE_INDEX         = wx_define_c_func( "aui_notebook_get_page_index", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_AUI_NOTEBOOK_GET_PAGE               = wx_define_c_func( "aui_notebook_get_page", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_AUI_NOTEBOOK_INSERT_PAGE            = wx_define_c_func( "aui_notebook_insert_page", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_POINTER}, E_OBJECT ),
	WX_AUI_NOTEBOOK_REMOVE_PAGE            = wx_define_c_func( "aui_notebook_remove_page", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_AUI_NOTEBOOK_SET_ART_PROVIDER       = wx_define_c_proc( "aui_notebook_set_art_provider", {dll:C_POINTER,dll:C_POINTER} ),
	WX_AUI_NOTEBOOK_SET_PAGE_BITMAP        = wx_define_c_func( "aui_notebook_set_page_bitmap", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_AUI_NOTEBOOK_SET_PAGE_TEXT          = wx_define_c_func( "aui_notebook_set_page_text", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_AUI_NOTEBOOK_SET_SELECTION          = wx_define_c_func( "aui_notebook_set_selection", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_AUI_NOTEBOOK_SET_TAB_CTRL_HEIGHT    = wx_define_c_proc( "aui_notebook_set_tab_ctrl_height", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_NOTEBOOK_SET_UNIFORM_BITMAP_SIZE= wx_define_c_proc( "aui_notebook_set_uniform_bitmap_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_AUI_NOTEBOOK_SPLIT                  = wx_define_c_proc( "aui_notebook_split", {dll:C_POINTER,dll:C_POINTER,dll:C_INT} ),
	WX_AUI_PANE_BORDER                     = wx_define_c_proc( "aui_pane_border", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_PIN_BUTTON                      = wx_define_c_proc( "aui_pin_button", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_POSITION                        = wx_define_c_proc( "aui_position", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_RESIZABLE                       = wx_define_c_proc( "aui_resizable", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_RIGHT_DOCKABLE                  = wx_define_c_proc( "aui_right_dockable", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_RIGHT                           = wx_define_c_proc( "aui_right", {dll:C_POINTER} ),
	WX_AUI_ROW                             = wx_define_c_proc( "aui_row", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_SAFE_SET                        = wx_define_c_proc( "aui_safe_set", {dll:C_POINTER,dll:C_POINTER} ),
	WX_AUI_SET_FLAG                        = wx_define_c_proc( "aui_set_flag", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_AUI_SHOW                            = wx_define_c_proc( "aui_show", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_TOOLBAR_PANE                    = wx_define_c_proc( "aui_toolbar_pane", {dll:C_POINTER} ),
	WX_AUI_TOP_DOCKABLE                    = wx_define_c_proc( "aui_top_dockable", {dll:C_POINTER,dll:C_INT} ),
	WX_AUI_TOP                             = wx_define_c_proc( "aui_top", {dll:C_POINTER} ),
	WX_AUI_WINDOW                          = wx_define_c_proc( "aui_window", {dll:C_POINTER,dll:C_POINTER} ),
	WX_AUOTOSIZE_ROW                       = wx_define_c_proc( "auotosize_row", {dll:C_POINTER,dll:C_INT} ),
	WX_AUTOSIZE_COLUMNS                    = wx_define_c_proc( "autosize_columns", {dll:C_POINTER} ),
	WX_AUTOSIZE_GRID                       = wx_define_c_proc( "autosize_grid", {dll:C_POINTER} ),
	WX_AUTOSIZE_ROWS                       = wx_define_c_proc( "autosize_rows", {dll:C_POINTER} ),
	WX_BATCH_COUNT                         = wx_define_c_func( "batch_count", {dll:C_POINTER}, E_OBJECT ),
	WX_BATCHING_UNDO                       = wx_define_c_func( "batching_undo", {dll:C_POINTER}, E_OBJECT ),
	WX_BEGIN_ALIGNMENT                     = wx_define_c_proc( "begin_alignment", {dll:C_POINTER,dll:C_INT} ),
	WX_BEGIN_BATCH_UNDO                    = wx_define_c_proc( "begin_batch_undo", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_BEGIN_BATCH                         = wx_define_c_proc( "begin_batch", {dll:C_POINTER} ),
	WX_BEGIN_BOLD                          = wx_define_c_proc( "begin_bold", {dll:C_POINTER} ),
	WX_BEGIN_CHAR_STYLE                    = wx_define_c_proc( "begin_char_style", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_BEGIN_DRAWING                       = wx_define_c_proc( "begin_drawing", {dll:C_POINTER} ),
	WX_BEGIN_FONT_SIZE                     = wx_define_c_proc( "begin_font_size", {dll:C_POINTER,dll:C_INT} ),
	WX_BEGIN_FONT                          = wx_define_c_proc( "begin_font", {dll:C_POINTER,dll:C_POINTER} ),
	WX_BEGIN_ITALIC                        = wx_define_c_proc( "begin_italic", {dll:C_POINTER} ),
	WX_BEGIN_LEFT_INDENT                   = wx_define_c_proc( "begin_left_indent", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_BEGIN_LINE_SPACING                  = wx_define_c_proc( "begin_line_spacing", {dll:C_POINTER,dll:C_INT} ),
	WX_BEGIN_LIST_STYLE                    = wx_define_c_proc( "begin_list_style", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT} ),
	WX_BEGIN_NUMBERED_BULLET               = wx_define_c_proc( "begin_numbered_bullet", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_POINTER} ),
	WX_BEGIN_PARAGRAPH_SPACING             = wx_define_c_proc( "begin_paragraph_spacing", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_BEGIN_PARAGRAPH_STYLE               = wx_define_c_proc( "begin_paragraph_style", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_BEGIN_RICH_STYLE                    = wx_define_c_proc( "begin_rich_style", {dll:C_POINTER,dll:C_POINTER} ),
	WX_BEGIN_RIGHT_INDENT                  = wx_define_c_proc( "begin_right_indent", {dll:C_POINTER,dll:C_INT} ),
	WX_BEGIN_STANDARD_BULLET               = wx_define_c_proc( "begin_standard_bullet", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT,dll:C_POINTER} ),
	WX_BEGIN_SUPPRESS_UNDO                 = wx_define_c_proc( "begin_suppress_undo", {dll:C_POINTER} ),
	WX_BEGIN_SYMBOL_BULLET                 = wx_define_c_proc( "begin_symbol_bullet", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT,dll:C_POINTER} ),
	WX_BEGIN_TEXT_COLOR                    = wx_define_c_proc( "begin_text_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_BEGIN_UNDERLINE                     = wx_define_c_proc( "begin_underline", {dll:C_POINTER} ),
	WX_BEGIN_URL                           = wx_define_c_proc( "begin_url", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT} ),
	WX_BEST_SIZE_PANE                      = wx_define_c_proc( "best_size_pane", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_BLIT                                = wx_define_c_proc( "blit", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_BLUR_HORIZONTAL                     = wx_define_c_func( "blur_horizontal", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_BLUR_VERTICAL                       = wx_define_c_func( "blur_vertical", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_BLUR                                = wx_define_c_func( "blur", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_CALC_SCROLLED_POSITION              = wx_define_c_func( "calc_scrolled_position", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_CALC_UNSCROLLED_POSITION            = wx_define_c_func( "calc_unscrolled_position", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_CAN_COPY_TEXT                       = wx_define_c_func( "can_copy_text", {dll:C_POINTER}, E_OBJECT ),
	WX_CAN_CUT_TEXT                        = wx_define_c_func( "can_cut_text", {dll:C_POINTER}, E_OBJECT ),
	WX_CAN_DRAG_GRID_COL_MOVE              = wx_define_c_func( "can_drag_grid_col_move", {dll:C_POINTER}, E_OBJECT ),
	WX_CAN_DRAG_GRID_COL_SIZE              = wx_define_c_func( "can_drag_grid_col_size", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_CAN_DRAG_GRID_ROW_SIZE              = wx_define_c_func( "can_drag_grid_row_size", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_CAN_DRAG_GRID_SIZE                  = wx_define_c_func( "can_drag_grid_size", {dll:C_POINTER}, E_OBJECT ),
	WX_CAN_ENABLE_CELL_CONTROL             = wx_define_c_func( "can_enable_cell_control", {dll:C_POINTER}, E_OBJECT ),
	WX_CAN_PASTE_TEXT                      = wx_define_c_func( "can_paste_text", {dll:C_POINTER}, E_OBJECT ),
	WX_CAN_REDO_TEXT                       = wx_define_c_func( "can_redo_text", {dll:C_POINTER}, E_OBJECT ),
	WX_CAN_UNDO_TEXT                       = wx_define_c_func( "can_undo_text", {dll:C_POINTER}, E_OBJECT ),
	WX_CAN_UPDATE                          = wx_define_c_func( "can_update", {dll:C_POINTER}, E_OBJECT ),
	WX_CAN_VETO                            = wx_define_c_func( "can_veto", {dll:C_POINTER}, E_OBJECT ),
	WX_CAPTURE_MOUSE                       = wx_define_c_proc( "capture_mouse", {dll:C_POINTER} ),
	WX_CARET_ISOK                          = wx_define_c_func( "caret_isok", {dll:C_POINTER}, E_OBJECT ),
	WX_CARET_ISVISIBLE                     = wx_define_c_func( "caret_isvisible", {dll:C_POINTER}, E_OBJECT ),
	WX_CARET_KILL_FOCUS                    = wx_define_c_proc( "caret_kill_focus", {dll:C_POINTER} ),
	WX_CARET_SET_FOCUS                     = wx_define_c_proc( "caret_set_focus", {dll:C_POINTER} ),
	WX_CASCADE_MDI                         = wx_define_c_proc( "cascade_mdi", {dll:C_POINTER} ),
	WX_CENTER_ON_PARENT                    = wx_define_c_proc( "center_on_parent", {dll:C_POINTER,dll:C_POINTER} ),
	WX_CHANGE_TEXT_VALUE                   = wx_define_c_proc( "change_text_value", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_CHECK_MENUITEM                      = wx_define_c_proc( "check_menuitem", {dll:C_POINTER,dll:C_INT} ),
	WX_CHECK_OS_VERSION                    = wx_define_c_func( "check_os_version", {dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_CHECK_TOOLKIT_VERSION               = wx_define_c_func( "check_toolkit_version", {dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_CHECK_UI                            = wx_define_c_proc( "check_ui", {dll:C_POINTER,dll:C_INT} ),
	WX_CHOICE_COUNT                        = wx_define_c_func( "choice_count", {dll:C_POINTER}, E_OBJECT ),
	WX_CLEANUP                             = wx_define_c_proc( "cleanup", {} ),
	WX_CLEAR_CHOICE                        = wx_define_c_proc( "clear_choice", {dll:C_POINTER} ),
	WX_CLEAR_COMBO                         = wx_define_c_proc( "clear_combo", {dll:C_POINTER} ),
	WX_CLEAR_DC                            = wx_define_c_proc( "clear_dc", {dll:C_POINTER} ),
	WX_CLEAR_GRID_SELECTION                = wx_define_c_proc( "clear_grid_selection", {dll:C_POINTER} ),
	WX_CLEAR_GRID                          = wx_define_c_proc( "clear_grid", {dll:C_POINTER} ),
	WX_CLEAR_HTML_HISTORY                  = wx_define_c_proc( "clear_html_history", {dll:C_POINTER} ),
	WX_CLEAR_ITEMS                         = wx_define_c_proc( "clear_items", {dll:C_POINTER} ),
	WX_CLEAR_LIST                          = wx_define_c_proc( "clear_list", {dll:C_POINTER} ),
	WX_CLEAR_TEXT                          = wx_define_c_proc( "clear_text", {dll:C_POINTER} ),
	WX_CLIENT_TO_SCREEN                    = wx_define_c_func( "client_to_screen", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_CLOSE_PROCESS_OUTPUT                = wx_define_c_proc( "close_process_output", {dll:C_POINTER} ),
	WX_CLOSE_WINDOW                        = wx_define_c_proc( "close_window", {dll:C_POINTER,dll:C_POINTER} ),
	WX_COLLAPSE_ALL_TREE_CHILDREN          = wx_define_c_proc( "collapse_all_tree_children", {dll:C_POINTER,dll:C_POINTER} ),
	WX_COLLAPSE_ALL_TREE_ITEMS             = wx_define_c_proc( "collapse_all_tree_items", {dll:C_POINTER} ),
	WX_COLLAPSE_AND_RESET_TREE_ITEM        = wx_define_c_proc( "collapse_and_reset_tree_item", {dll:C_POINTER,dll:C_POINTER} ),
	WX_COLLAPSE_DIRCTRL_PATH               = wx_define_c_func( "collapse_dirctrl_path", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_COLLAPSE_DIRCTRL_TREE               = wx_define_c_proc( "collapse_dirctrl_tree", {dll:C_POINTER} ),
	WX_COLLAPSE_TREE_ITEM                  = wx_define_c_proc( "collapse_tree_item", {dll:C_POINTER,dll:C_POINTER} ),
	WX_COLOR_SELECTOR                      = wx_define_c_func( "color_selector", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_COMBOBOX_COUNT                      = wx_define_c_func( "combobox_count", {dll:C_POINTER}, E_OBJECT ),
	WX_COMPILE_REGEX                       = wx_define_c_func( "compile_regex", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_CONFIG_DELETE_ALL                   = wx_define_c_proc( "config_delete_all", {dll:C_POINTER} ),
	WX_CONFIG_DELETE_ENTRY                 = wx_define_c_proc( "config_delete_entry", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT} ),
	WX_CONFIG_DELETE_GROUP                 = wx_define_c_proc( "config_delete_group", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_CONFIG_EXISTS                       = wx_define_c_proc( "config_exists", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_CONFIG_FLUSH                        = wx_define_c_proc( "config_flush", {dll:C_POINTER,dll:C_INT} ),
	WX_CONFIG_READ                         = wx_define_c_func( "config_read", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT}, E_OBJECT ),
	WX_CONFIG_WRITE                        = wx_define_c_proc( "config_write", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT} ),
	WX_CONVERT_TO_GREYSCALE                = wx_define_c_func( "convert_to_greyscale", {dll:C_POINTER}, E_OBJECT ),
	WX_CONVERT_TO_IMAGE                    = wx_define_c_func( "convert_to_image", {dll:C_POINTER}, E_OBJECT ),
	WX_CONVERT_TO_MONO                     = wx_define_c_func( "convert_to_mono", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_COPY_FILE                           = wx_define_c_func( "copy_file", {dll:E_OBJECT,dll:E_OBJECT}, E_OBJECT ),
	WX_COPY_TEXT                           = wx_define_c_proc( "copy_text", {dll:C_POINTER} ),
	WX_CREATE_BUTTON_SIZER                 = wx_define_c_func( "create_button_sizer", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_CREATE_FONT_FROM_RICH_ATTR          = wx_define_c_func( "create_font_from_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_CUT_TEXT                            = wx_define_c_proc( "cut_text", {dll:C_POINTER} ),
	WX_DELETE_ALL_PAGES                    = wx_define_c_proc( "delete_all_pages", {dll:C_POINTER} ),
	WX_DELETE_COLS                         = wx_define_c_proc( "delete_cols", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_DELETE_COMBO_ITEM                   = wx_define_c_proc( "delete_combo_item", {dll:C_POINTER,dll:C_POINTER} ),
	WX_DELETE_CONFIG                       = wx_define_c_proc( "delete_config", {dll:E_OBJECT} ),
	WX_DELETE_INSTANCE                     = wx_define_c_proc( "delete_instance", {dll:E_OBJECT} ),
	WX_DELETE_ITEM                         = wx_define_c_proc( "delete_item", {dll:C_POINTER,dll:C_POINTER} ),
	WX_DELETE_LIST_ITEM                    = wx_define_c_proc( "delete_list_item", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_DELETE_PAGE                         = wx_define_c_proc( "delete_page", {dll:C_POINTER,dll:C_POINTER} ),
	WX_DELETE_ROWS                         = wx_define_c_proc( "delete_rows", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_DELETE_TREE_CHILDREN                = wx_define_c_proc( "delete_tree_children", {dll:C_POINTER,dll:C_POINTER} ),
	WX_DELETE_TREE_ITEM                    = wx_define_c_proc( "delete_tree_item", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_DELETE_WXAUIPANEINFO                = wx_define_c_proc( "delete_wxAuiPaneInfo", {dll:E_OBJECT} ),
	WX_DELETE_WXREGEX                      = wx_define_c_proc( "delete_wxRegEx", {dll:E_OBJECT} ),
	WX_DELETE_WXXMLNODE                    = wx_define_c_proc( "delete_wxXmlNode", {dll:E_OBJECT} ),
	WX_DELETE_WXXMLPROPERTY                = wx_define_c_proc( "delete_wxXmlProperty", {dll:C_POINTER} ),
	WX_DELETE_XML_NODE_PROPERTY            = wx_define_c_func( "delete_xml_node_property", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_DESTROY                             = wx_define_c_proc( "destroy", {dll:C_POINTER} ),
	WX_DETACH_PANE                         = wx_define_c_func( "detach_pane", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_DETACH_PROCESS                      = wx_define_c_proc( "detach_process", {dll:C_POINTER} ),
	WX_DETATCH_SIZER                       = wx_define_c_proc( "detatch_sizer", {dll:C_POINTER,dll:C_POINTER} ),
	WX_DETATCH_XML_ROOT                    = wx_define_c_func( "detatch_xml_root", {dll:C_POINTER}, E_OBJECT ),
	WX_DIR_EXISTS                          = wx_define_c_func( "dir_exists", {dll:E_OBJECT}, E_OBJECT ),
	WX_DISCARD_TEXT_EDITS                  = wx_define_c_proc( "discard_text_edits", {dll:C_POINTER} ),
	WX_DISPLAY_CHANGE_MODE                 = wx_define_c_func( "display_change_mode", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_DISPLAY_GET_CLIENT_AREA             = wx_define_c_func( "display_get_client_area", {dll:C_POINTER}, E_OBJECT ),
	WX_DISPLAY_GET_COUNT                   = wx_define_c_func( "display_get_count", {dll:C_POINTER}, E_OBJECT ),
	WX_DISPLAY_GET_CURRENT_MODE            = wx_define_c_func( "display_get_current_mode", {dll:C_POINTER}, E_OBJECT ),
	WX_DISPLAY_GET_FROM_POINT              = wx_define_c_func( "display_get_from_point", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_DISPLAY_GET_FROM_WINDOW             = wx_define_c_func( "display_get_from_window", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_DISPLAY_GET_GEOMETRY                = wx_define_c_func( "display_get_geometry", {dll:C_POINTER}, E_OBJECT ),
	WX_DISPLAY_GET_MODES                   = wx_define_c_func( "display_get_modes", {dll:C_POINTER}, E_OBJECT ),
	WX_DISPLAY_GET_NAME                    = wx_define_c_func( "display_get_name", {dll:C_POINTER}, E_OBJECT ),
	WX_DISPLAY_IS_PRIMARY                  = wx_define_c_func( "display_is_primary", {dll:C_POINTER}, E_OBJECT ),
	WX_DRAW_BITMAP2                        = wx_define_c_proc( "draw_bitmap2", {dll:E_OBJECT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_DRAW_BITMAP                         = wx_define_c_proc( "draw_bitmap", {dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_DRAW_ICON                           = wx_define_c_proc( "draw_icon", {dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_DRAW_LABEL                          = wx_define_c_proc( "draw_label", {dll:C_POINTER,dll:E_OBJECT,dll:C_POINTER,dll:E_OBJECT,dll:C_INT} ),
	WX_DRAW_LINE                           = wx_define_c_proc( "draw_line", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_DRAW_POLYGON                        = wx_define_c_proc( "draw_polygon", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_EDIT_TREE_ITEM                      = wx_define_c_proc( "edit_tree_item", {dll:C_POINTER,dll:C_POINTER} ),
	WX_ENABLE_DRAG_GRID_COL_MOVE           = wx_define_c_proc( "enable_drag_grid_col_move", {dll:C_POINTER,dll:C_INT} ),
	WX_ENABLE_DRAG_GRID_COL_SIZE           = wx_define_c_proc( "enable_drag_grid_col_size", {dll:C_POINTER,dll:C_INT} ),
	WX_ENABLE_DRAG_GRID_ROW_SIZE           = wx_define_c_proc( "enable_drag_grid_row_size", {dll:C_POINTER,dll:C_INT} ),
	WX_ENABLE_DRAG_GRID_SIZE               = wx_define_c_proc( "enable_drag_grid_size", {dll:C_POINTER,dll:C_INT} ),
	WX_ENABLE_GRID_CELL_EDIT_CONTROL       = wx_define_c_proc( "enable_grid_cell_edit_control", {dll:C_POINTER,dll:C_INT} ),
	WX_ENABLE_GRID_EDITING                 = wx_define_c_proc( "enable_grid_editing", {dll:C_POINTER,dll:C_INT} ),
	WX_ENABLE_GRID_LINES                   = wx_define_c_proc( "enable_grid_lines", {dll:C_POINTER,dll:C_INT} ),
	WX_ENABLE_MENUITEM                     = wx_define_c_proc( "enable_menuitem", {dll:C_POINTER,dll:C_INT} ),
	WX_ENABLE_SCROLLING                    = wx_define_c_proc( "enable_scrolling", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_ENABLE_TOOLTIPS                     = wx_define_c_proc( "enable_tooltips", {dll:C_POINTER} ),
	WX_ENABLE_UI                           = wx_define_c_proc( "enable_ui", {dll:C_POINTER,dll:C_INT} ),
	WX_ENABLE                              = wx_define_c_proc( "enable", {dll:E_OBJECT,dll:C_POINTER} ),
	WX_END_ALIGNMENT                       = wx_define_c_proc( "end_alignment", {dll:C_POINTER} ),
	WX_END_ALL_RICH_STYLES                 = wx_define_c_proc( "end_all_rich_styles", {dll:C_POINTER} ),
	WX_END_BATCH_UNDO                      = wx_define_c_proc( "end_batch_undo", {dll:C_POINTER} ),
	WX_END_BATCH                           = wx_define_c_proc( "end_batch", {dll:C_POINTER} ),
	WX_END_BOLD                            = wx_define_c_proc( "end_bold", {dll:C_POINTER} ),
	WX_END_CHAR_STYLE                      = wx_define_c_proc( "end_char_style", {dll:C_POINTER} ),
	WX_END_DRAWING                         = wx_define_c_proc( "end_drawing", {dll:C_POINTER} ),
	WX_END_FONT_SIZE                       = wx_define_c_proc( "end_font_size", {dll:C_POINTER} ),
	WX_END_FONT                            = wx_define_c_proc( "end_font", {dll:C_POINTER} ),
	WX_END_ITALIC                          = wx_define_c_proc( "end_italic", {dll:C_POINTER} ),
	WX_END_LEFT_INDENT                     = wx_define_c_proc( "end_left_indent", {dll:C_POINTER} ),
	WX_END_LINE_SPACING                    = wx_define_c_proc( "end_line_spacing", {dll:C_POINTER} ),
	WX_END_LIST_STYLE                      = wx_define_c_proc( "end_list_style", {dll:C_POINTER} ),
	WX_END_MODAL                           = wx_define_c_proc( "end_modal", {dll:C_POINTER,dll:C_INT} ),
	WX_END_NUMBERED_BULLET                 = wx_define_c_proc( "end_numbered_bullet", {dll:C_POINTER} ),
	WX_END_PARAGRAPH_SPACING               = wx_define_c_proc( "end_paragraph_spacing", {dll:C_POINTER} ),
	WX_END_PARAGRAPH_STYLE                 = wx_define_c_proc( "end_paragraph_style", {dll:C_POINTER} ),
	WX_END_RICH_STYLE                      = wx_define_c_proc( "end_rich_style", {dll:C_POINTER} ),
	WX_END_RIGHT_INDENT                    = wx_define_c_proc( "end_right_indent", {dll:C_POINTER} ),
	WX_END_STANDARD_BULLET                 = wx_define_c_proc( "end_standard_bullet", {dll:C_POINTER} ),
	WX_END_SUPPRESS_UNDO                   = wx_define_c_proc( "end_suppress_undo", {dll:C_POINTER} ),
	WX_END_SYMBOL_BULLET                   = wx_define_c_proc( "end_symbol_bullet", {dll:C_POINTER} ),
	WX_END_TEXT_COLOR                      = wx_define_c_proc( "end_text_color", {dll:C_POINTER} ),
	WX_END_UNDERLINE                       = wx_define_c_proc( "end_underline", {dll:C_POINTER} ),
	WX_END_URL                             = wx_define_c_proc( "end_url", {dll:C_POINTER} ),
	WX_ENSURE_TREE_ITEM_VISUAL             = wx_define_c_proc( "ensure_tree_item_visual", {dll:C_POINTER,dll:C_POINTER} ),
	WX_EXIT_MAIN                           = wx_define_c_proc( "exit_main", {} ),
	WX_EXPAND_ALL_TREE_CHILDREN            = wx_define_c_proc( "expand_all_tree_children", {dll:C_POINTER,dll:C_POINTER} ),
	WX_EXPAND_DIRCTRL_PATH                 = wx_define_c_func( "expand_dirctrl_path", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_EXPAND_TREE_ITEM                    = wx_define_c_proc( "expand_tree_item", {dll:C_POINTER,dll:C_POINTER} ),
	WX_EXPAND_TREE                         = wx_define_c_proc( "expand_tree", {dll:C_POINTER} ),
	WX_FILE_SELECTOR                       = wx_define_c_func( "file_selector", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT,dll:C_POINTER}, E_OBJECT ),
	WX_FIND_CHOICE_STRING                  = wx_define_c_func( "find_choice_string", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FIND_COMBO_STRING                   = wx_define_c_func( "find_combo_string", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FIND_LIST_STRING                    = wx_define_c_func( "find_list_string", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FIND_MENUITEM_BY_POSITION           = wx_define_c_func( "find_menuitem_by_position", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_FIND_MENUITEM_ID                    = wx_define_c_func( "find_menuitem_id", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FIND_STRING                         = wx_define_c_func( "find_string", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FIND_WINDOW                         = wx_define_c_func( "find_window", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_FIT_INSIDE                          = wx_define_c_proc( "fit_inside", {dll:C_POINTER} ),
	WX_FIT_WINDOW                          = wx_define_c_func( "fit_window", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_FONT_SELECTOR                       = wx_define_c_func( "font_selector", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_FORCE_REFRESH_GRID                  = wx_define_c_proc( "force_refresh_grid", {dll:C_POINTER} ),
	WX_FROM_UTF8                           = wx_define_c_func( "from_utf8", {dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_ABORT                           = wx_define_c_func( "ftp_abort", {dll:C_POINTER}, E_OBJECT ),
	WX_FTP_CHDIR                           = wx_define_c_func( "ftp_chdir", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_CHECK_COMMAND                   = wx_define_c_func( "ftp_check_command", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_FTP_CLOSE                           = wx_define_c_func( "ftp_close", {dll:C_POINTER}, E_OBJECT ),
	WX_FTP_CONNECT                         = wx_define_c_func( "ftp_connect", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_DIR                             = wx_define_c_func( "ftp_dir", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_FILE_EXISTS                     = wx_define_c_func( "ftp_file_exists", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_FILE_SIZE                       = wx_define_c_func( "ftp_file_size", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_FILES                           = wx_define_c_func( "ftp_files", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_INPUT_STREAM                    = wx_define_c_func( "ftp_input_stream", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_LAST_RESULT                     = wx_define_c_func( "ftp_last_result", {dll:C_POINTER}, E_OBJECT ),
	WX_FTP_MKDIR                           = wx_define_c_func( "ftp_mkdir", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_MODE                            = wx_define_c_proc( "ftp_mode", {dll:C_POINTER,dll:C_INT} ),
	WX_FTP_OUTPUT_STREAM                   = wx_define_c_func( "ftp_output_stream", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_PASSIVE                         = wx_define_c_proc( "ftp_passive", {dll:C_POINTER,dll:C_INT} ),
	WX_FTP_PASSWORD                        = wx_define_c_proc( "ftp_password", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_FTP_PWD                             = wx_define_c_func( "ftp_pwd", {dll:C_POINTER}, E_OBJECT ),
	WX_FTP_RENAME                          = wx_define_c_func( "ftp_rename", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_RMDIR                           = wx_define_c_func( "ftp_rmdir", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_RM                              = wx_define_c_func( "ftp_rm", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_SEND_COMMAND                    = wx_define_c_func( "ftp_send_command", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_FTP_USER                            = wx_define_c_proc( "ftp_user", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_GET_ACTIVE_MDI_CHILD                = wx_define_c_func( "get_active_mdi_child", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_ALL_FILES                       = wx_define_c_func( "get_all_files", {dll:E_OBJECT,dll:E_OBJECT,dll:C_INT,dll:C_POINTER}, E_OBJECT ),
	WX_GET_ALPHA                           = wx_define_c_func( "get_alpha", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_ARCH_FROM_NAME                  = wx_define_c_func( "get_arch_from_name", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_ARCH_NAME_FROM_ENUM             = wx_define_c_func( "get_arch_name_from_enum", {dll:C_INT}, E_OBJECT ),
	WX_GET_ARCH_NAME                       = wx_define_c_func( "get_arch_name", {}, E_OBJECT ),
	WX_GET_ARCH                            = wx_define_c_func( "get_arch", {}, E_OBJECT ),
	WX_GET_BASIC_RICH_STYLE                = wx_define_c_func( "get_basic_rich_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_BEZEL_FACE                      = wx_define_c_func( "get_bezel_face", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_BITMAP_SIZE                     = wx_define_c_func( "get_bitmap_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_BLUE                            = wx_define_c_func( "get_blue", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_CAL_DAYS_IN_MONTH               = wx_define_c_func( "get_cal_days_in_month", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CARET_BLINK_TIME                = wx_define_c_func( "get_caret_blink_time", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CARET_POSITION                  = wx_define_c_func( "get_caret_position", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CARET_SIZE                      = wx_define_c_func( "get_caret_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CARET_WINDOW                    = wx_define_c_func( "get_caret_window", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CELL_ALIGNMENT                  = wx_define_c_func( "get_cell_alignment", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_GET_CELL_BACKGROUND_COLOR           = wx_define_c_func( "get_cell_background_color", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_GET_CELL_EDITOR_CONTROL             = wx_define_c_func( "get_cell_editor_control", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_CELL_EDITOR                     = wx_define_c_func( "get_cell_editor", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_CELL_FONT                       = wx_define_c_func( "get_cell_font", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_GET_CELL_TEXT_COLOR                 = wx_define_c_func( "get_cell_text_color", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_GET_CELL_VALUE                      = wx_define_c_func( "get_cell_value", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_GET_CHECKED_UI                      = wx_define_c_func( "get_checked_ui", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CHECKED                         = wx_define_c_func( "get_checked", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CHILDREN                        = wx_define_c_func( "get_children", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CHOSEN_FONT                     = wx_define_c_func( "get_chosen_font", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CLIENT_SIZE                     = wx_define_c_func( "get_client_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CLIENT_WINDOW                   = wx_define_c_func( "get_client_window", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CLIP_TEXT                       = wx_define_c_func( "get_clip_text", {}, E_OBJECT ),
	WX_GET_COL_LABEL_SIZE                  = wx_define_c_func( "get_col_label_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_COL_LABEL                       = wx_define_c_func( "get_col_label", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_COLOR_DATA                      = wx_define_c_func( "get_color_data", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_COLOR                           = wx_define_c_func( "get_color", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_COL_SIZE                        = wx_define_c_func( "get_col_size", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_COMBO_BITMAP_SIZE               = wx_define_c_func( "get_combo_bitmap_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_COMBO_ITEM_BITMAP               = wx_define_c_func( "get_combo_item_bitmap", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_COMBO_ITEM_STRING               = wx_define_c_func( "get_combo_item_string", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_COMBO_STRING                    = wx_define_c_func( "get_combo_string", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CONFIG_APP_NAME                 = wx_define_c_func( "get_config_app_name", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CURRENT_PAGE                    = wx_define_c_func( "get_current_page", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DATE_TIME                       = wx_define_c_func( "get_date_time", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DATE                            = wx_define_c_func( "get_date", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DAYS_IN_MONTH                   = wx_define_c_func( "get_days_in_month", {dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_GET_DC_SIZE                         = wx_define_c_func( "get_dc_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DC_TEXT_EXTENT                  = wx_define_c_func( "get_dc_text_extent", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_DEFAULT_CELL_BACKGROUND_COLOR   = wx_define_c_func( "get_default_cell_background_color", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DEFAULT_CELL_FONT               = wx_define_c_func( "get_default_cell_font", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DEFAULT_CELL_TEXT_COLOUR        = wx_define_c_func( "get_default_cell_text_colour", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DEFAULT_COL_LABEL_SIZE          = wx_define_c_func( "get_default_col_label_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DEFAULT_COL_SIZE                = wx_define_c_func( "get_default_col_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DEFAULT_EDITOR_FOR_CELL         = wx_define_c_func( "get_default_editor_for_cell", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_GET_DEFAULT_GRID_EDITOR             = wx_define_c_func( "get_default_grid_editor", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DEFAULT_ROW_LABEL_SIZE          = wx_define_c_func( "get_default_row_label_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DEFAULT_ROW_SIZE                = wx_define_c_func( "get_default_row_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DEFAULT_TEXT_STYLE              = wx_define_c_func( "get_default_text_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DESKTOP_WINDOW                  = wx_define_c_func( "get_desktop_window", {}, E_OBJECT ),
	WX_GET_DIRCTRL_DEFAULT_PATH            = wx_define_c_func( "get_dirctrl_default_path", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DIRCTRL_FILE_PATH               = wx_define_c_func( "get_dirctrl_file_path", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DIRCTRL_FILTER_INDEX            = wx_define_c_func( "get_dirctrl_filter_index", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DIRCTRL_FILTER_LISTCTRL         = wx_define_c_func( "get_dirctrl_filter_listctrl", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DIRCTRL_FILTER                  = wx_define_c_func( "get_dirctrl_filter", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DIRCTRL_PATH                    = wx_define_c_func( "get_dirctrl_path", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DIRCTRL_ROOT_ID                 = wx_define_c_func( "get_dirctrl_root_id", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DIRCTRL_TREECTRL                = wx_define_c_func( "get_dirctrl_treectrl", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DIR_PATH                        = wx_define_c_func( "get_dir_path", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DROP_FILES                      = wx_define_c_func( "get_drop_files", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_ENABLED_UI                      = wx_define_c_func( "get_enabled_ui", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_ENCODINGS                       = wx_define_c_func( "get_encodings", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_ENDIANNESS_FROM_NAME            = wx_define_c_func( "get_endianness_from_name", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_ENDIANNESS_NAME_FROM_ENUM       = wx_define_c_func( "get_endianness_name_from_enum", {dll:C_INT}, E_OBJECT ),
	WX_GET_ENDIANNESS_NAME                 = wx_define_c_func( "get_endianness_name", {}, E_OBJECT ),
	WX_GET_ENDIANNESS                      = wx_define_c_func( "get_endianness", {}, E_OBJECT ),
	WX_GET_EVENT_ID                        = wx_define_c_func( "get_event_id", {}, E_OBJECT ),
	WX_GET_EVENT_OBJECT                    = wx_define_c_func( "get_event_object", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_EVENT_TYPE                      = wx_define_c_func( "get_event_type", {}, E_OBJECT ),
	WX_GET_EXT                             = wx_define_c_func( "get_ext", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_FACENAMES                       = wx_define_c_func( "get_facenames", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_FACESIZES                       = wx_define_c_func( "get_facesizes", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_FIELD_RECT                      = wx_define_c_func( "get_field_rect", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_FIRST_CONFIG_ENTRY              = wx_define_c_func( "get_first_config_entry", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_FIRST_CONFIG_GROUP              = wx_define_c_func( "get_first_config_group", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_FIRST_REGION                    = wx_define_c_func( "get_first_region", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FIRST_TREE_CHILD                = wx_define_c_func( "get_first_tree_child", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_FIRST_VISIBLE_TREE_ITEM         = wx_define_c_func( "get_first_visible_tree_item", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FONT_DATA                       = wx_define_c_func( "get_font_data", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FONT_ENCODING                   = wx_define_c_func( "get_font_encoding", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FONT_FAMILY                     = wx_define_c_func( "get_font_family", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FONT_MEMBERS                    = wx_define_c_func( "get_font_members", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FONT_NAME                       = wx_define_c_func( "get_font_name", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FONT_POINTSIZE                  = wx_define_c_func( "get_font_pointsize", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FONT_STYLE                      = wx_define_c_func( "get_font_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FONT_UNDERLINED                 = wx_define_c_func( "get_font_underlined", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FONT_WEIGHT                     = wx_define_c_func( "get_font_weight", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FONT                            = wx_define_c_func( "get_font", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FORBIDDEN_CHARS                 = wx_define_c_func( "get_forbidden_chars", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_FULL_NAME                       = wx_define_c_func( "get_full_name", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_FULL_PATH                       = wx_define_c_func( "get_full_path", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_GREEN                           = wx_define_c_func( "get_green", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_GRID_BATCH_COUNT                = wx_define_c_func( "get_grid_batch_count", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_GRID_LABEL_BACKGROUND_COLOR     = wx_define_c_func( "get_grid_label_background_color", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_GRID_LABEL_FONT                 = wx_define_c_func( "get_grid_label_font", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_GRID_LABEL_TEXT_COLOR           = wx_define_c_func( "get_grid_label_text_color", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_GRID_LINE_COLOR                 = wx_define_c_func( "get_grid_line_color", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_GRID_LINES_ENABLED              = wx_define_c_func( "get_grid_lines_enabled", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_GRID_ROW_LABEL_ALIGN            = wx_define_c_func( "get_grid_row_label_align", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_GRID_SCROLL_LINE                = wx_define_c_func( "get_grid_scroll_line", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_GRID_SELECTION_BACKGROUND       = wx_define_c_func( "get_grid_selection_background", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_GRID_SELECTION_FOREGROUND       = wx_define_c_func( "get_grid_selection_foreground", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_GRID_VIEW_WIDTH                 = wx_define_c_func( "get_grid_view_width", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_HANDLE                          = wx_define_c_func( "get_handle", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_HOME_DIR                        = wx_define_c_func( "get_home_dir", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_HTML_ANCHOR                     = wx_define_c_func( "get_html_anchor", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_HTML_EASY_PRINT_DATA            = wx_define_c_func( "get_html_easy_print_data", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_HTML_EASY_PRINTING_PARENT       = wx_define_c_func( "get_html_easy_printing_parent", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_HTML_EASY_PRINT_PREVIEW_FILE    = wx_define_c_func( "get_html_easy_print_preview_file", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_HTML_EASY_PRINT_PREVIEW_TEXT    = wx_define_c_func( "get_html_easy_print_preview_text", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_HTML_EASY_PRINT_SETUP           = wx_define_c_func( "get_html_easy_print_setup", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_HTML_PAGE_TITLE                 = wx_define_c_func( "get_html_page_title", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_HTML_PAGE                       = wx_define_c_func( "get_html_page", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_HTTP_CONTENT_TYPE               = wx_define_c_func( "get_http_content_type", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_HTTP_HEADER                     = wx_define_c_func( "get_http_header", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_HTTP_INPUT                      = wx_define_c_func( "get_http_input", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_ID                              = wx_define_c_func( "get_id", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_KEY_CODE                        = wx_define_c_func( "get_key_code", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_KEY_STATE                       = wx_define_c_func( "get_key_state", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_LABEL                           = wx_define_c_func( "get_label", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_LINK_HREF                       = wx_define_c_func( "get_link_href", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_LINK_TARGET                     = wx_define_c_func( "get_link_target", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_LIST_EVENT_COLUMN               = wx_define_c_func( "get_list_event_column", {dll:C_INT}, E_OBJECT ),
	WX_GET_LIST_STRING                     = wx_define_c_func( "get_list_string", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_LONG_PATH                       = wx_define_c_func( "get_long_path", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_MANAGED_WINDOW                  = wx_define_c_func( "get_managed_window", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MANAGER                         = wx_define_c_func( "get_manager", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MATCH_COUNT                     = wx_define_c_func( "get_match_count", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MATCH                           = wx_define_c_func( "get_match", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_GET_MDI_WINDOW_MENU                 = wx_define_c_func( "get_mdi_window_menu", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MEDIA_LENGTH                    = wx_define_c_func( "get_media_length", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MEDIA_POSITION                  = wx_define_c_func( "get_media_position", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MEDIA_STATE                     = wx_define_c_func( "get_media_state", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MEDIA_VOLUME                    = wx_define_c_func( "get_media_volume", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MENU_BAR                        = wx_define_c_func( "get_menu_bar", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MENU_BY_INDEX                   = wx_define_c_func( "get_menu_by_index", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_MENU_BY_LABEL                   = wx_define_c_func( "get_menu_by_label", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_MENUEVENT_ID                    = wx_define_c_func( "get_menuevent_id", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MENUITEM_COUNT                  = wx_define_c_func( "get_menuitem_count", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MENUITEM_ID                     = wx_define_c_func( "get_menuitem_id", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MODE_UI                         = wx_define_c_func( "get_mode_ui", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_MODIFIED_TIME                   = wx_define_c_func( "get_modified_time", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_NAME                            = wx_define_c_func( "get_name", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_NATIVE_FONT_INFO_USER_DESC      = wx_define_c_func( "get_native_font_info_user_desc", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_NATIVE_FONT_INFO                = wx_define_c_func( "get_native_font_info", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_NEXT_CONFIG_ENTRY               = wx_define_c_func( "get_next_config_entry", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_NEXT_CONFIG_GROUP               = wx_define_c_func( "get_next_config_group", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_NEXT_PAGE                       = wx_define_c_func( "get_next_page", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_NEXT_REGION                     = wx_define_c_func( "get_next_region", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_NEXT_TREE_CHILD                 = wx_define_c_func( "get_next_tree_child", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_NEXT_TREE_SIBLING               = wx_define_c_func( "get_next_tree_sibling", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_NEXT_VISIBLE_TREE_ITEM          = wx_define_c_func( "get_next_visible_tree_item", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_NEXT_XML_NODE                   = wx_define_c_func( "get_next_xml_node", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_NEXT_XML_PROP                   = wx_define_c_func( "get_next_xml_prop", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_NUMBER_COLS                     = wx_define_c_func( "get_number_cols", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_NUMBER_ROWS                     = wx_define_c_func( "get_number_rows", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_OPTION_INT                      = wx_define_c_func( "get_option_int", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_OPTION                          = wx_define_c_func( "get_option", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_OS_FAMILY_NAME_FROM_ENUM        = wx_define_c_func( "get_os_family_name_from_enum", {dll:C_INT}, E_OBJECT ),
	WX_GET_OS_FAMILY_NAME                  = wx_define_c_func( "get_os_family_name", {}, E_OBJECT ),
	WX_GET_OS_ID_FROM_NAME                 = wx_define_c_func( "get_os_id_from_name", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_OS_ID_NAME_FROM_ENUM            = wx_define_c_func( "get_os_id_name_from_enum", {dll:C_INT}, E_OBJECT ),
	WX_GET_OS_ID_NAME                      = wx_define_c_func( "get_os_id_name", {}, E_OBJECT ),
	WX_GET_OS_ID                           = wx_define_c_func( "get_os_id", {}, E_OBJECT ),
	WX_GET_OS_MAJOR_VERSION                = wx_define_c_func( "get_os_major_version", {}, E_OBJECT ),
	WX_GET_OS_MINOR_VERSION                = wx_define_c_func( "get_os_minor_version", {}, E_OBJECT ),
	WX_GET_PAGE_COUNT                      = wx_define_c_func( "get_page_count", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PAGE_IMAGE                      = wx_define_c_func( "get_page_image", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_PAGE_MM                         = wx_define_c_func( "get_page_mm", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PAGE_PIXELS                     = wx_define_c_func( "get_page_pixels", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PAGESETUP_DATA                  = wx_define_c_func( "get_pagesetup_data", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PAGE_TEXT                       = wx_define_c_func( "get_page_text", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_PALETTE                         = wx_define_c_func( "get_palette", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PANE                            = wx_define_c_func( "get_pane", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_PARENT                          = wx_define_c_func( "get_parent", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PARENT_XML_NODE                 = wx_define_c_func( "get_parent_xml_node", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PID                             = wx_define_c_func( "get_pid", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_POINT_SIZE                      = wx_define_c_func( "get_point_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PORT_ID_FROM_NAME               = wx_define_c_func( "get_port_id_from_name", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_PORT_ID_NAME_FROM_ENUM          = wx_define_c_func( "get_port_id_name_from_enum", {dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_GET_PORT_ID_NAME                    = wx_define_c_func( "get_port_id_name", {}, E_OBJECT ),
	WX_GET_PORT_ID_SHORT_NAME_FROM_ENUM    = wx_define_c_func( "get_port_id_short_name_from_enum", {dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_GET_PORT_ID_SHORT_NAME              = wx_define_c_func( "get_port_id_short_name", {}, E_OBJECT ),
	WX_GET_PORT_ID                         = wx_define_c_func( "get_port_id", {}, E_OBJECT ),
	WX_GET_POSITION_RICH_EVENT             = wx_define_c_func( "get_position_rich_event", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PPI_PRINTER                     = wx_define_c_func( "get_ppi_printer", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PPI_SCREEN                      = wx_define_c_func( "get_ppi_screen", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PREVIEW_FRAME                   = wx_define_c_func( "get_preview_frame", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PREV_TREE_SIBLING               = wx_define_c_func( "get_prev_tree_sibling", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_PREV_VISIBLE_TREE_ITEM          = wx_define_c_func( "get_prev_visible_tree_item", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_PRINT_DATA                      = wx_define_c_func( "get_print_data", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PRINTOUT_DC                     = wx_define_c_func( "get_printout_dc", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RADIO_SELECTION                 = wx_define_c_func( "get_radio_selection", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RANGE                           = wx_define_c_func( "get_range", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RECT                            = wx_define_c_func( "get_rect", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RED                             = wx_define_c_func( "get_red", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_REGIONS                         = wx_define_c_func( "get_regions", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_REGION                          = wx_define_c_func( "get_region", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_ALIGNMENT             = wx_define_c_func( "get_rich_attr_alignment", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_BACK_COLOR            = wx_define_c_func( "get_rich_attr_back_color", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_BULLET_FONT           = wx_define_c_func( "get_rich_attr_bullet_font", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_BULLET_NAME           = wx_define_c_func( "get_rich_attr_bullet_name", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_BULLET_NUMBER         = wx_define_c_func( "get_rich_attr_bullet_number", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_BULLET_STYLE          = wx_define_c_func( "get_rich_attr_bullet_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_BULLET_TEXT           = wx_define_c_func( "get_rich_attr_bullet_text", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_CHAR_STYLE            = wx_define_c_func( "get_rich_attr_char_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_COLOR                 = wx_define_c_func( "get_rich_attr_color", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_EFFECT_FLAGS          = wx_define_c_func( "get_rich_attr_effect_flags", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_EFFECTS               = wx_define_c_func( "get_rich_attr_effects", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_FLAGS                 = wx_define_c_func( "get_rich_attr_flags", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_FONT_ATTRIBUTES       = wx_define_c_func( "get_rich_attr_font_attributes", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_FONT_NAME             = wx_define_c_func( "get_rich_attr_font_name", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_FONT_SIZE             = wx_define_c_func( "get_rich_attr_font_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_FONT_STYLE            = wx_define_c_func( "get_rich_attr_font_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_FONT_WEIGHT           = wx_define_c_func( "get_rich_attr_font_weight", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_LEFT_INDENT           = wx_define_c_func( "get_rich_attr_left_indent", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_LEFT_SUB_INDENT       = wx_define_c_func( "get_rich_attr_left_sub_indent", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_LINE_SPACING          = wx_define_c_func( "get_rich_attr_line_spacing", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_LIST_STYLE            = wx_define_c_func( "get_rich_attr_list_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_OUTLINE_LEVEL         = wx_define_c_func( "get_rich_attr_outline_level", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_PARA_SPACING_AFTER    = wx_define_c_func( "get_rich_attr_para_spacing_after", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_PARA_SPACING_BEFORE   = wx_define_c_func( "get_rich_attr_para_spacing_before", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_PARA_STYLE            = wx_define_c_func( "get_rich_attr_para_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_RIGHT_INDENT          = wx_define_c_func( "get_rich_attr_right_indent", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_TABS                  = wx_define_c_func( "get_rich_attr_tabs", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_UNDERLINED            = wx_define_c_func( "get_rich_attr_underlined", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_ATTR_URL                   = wx_define_c_func( "get_rich_attr_url", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_DIALOG_ATTRIBUTES          = wx_define_c_func( "get_rich_dialog_attributes", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_DIALOG_IMAGE_LIST          = wx_define_c_func( "get_rich_dialog_image_list", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_DIALOG_STYLE               = wx_define_c_proc( "get_rich_dialog_style", {dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_GET_RICH_EVENT_CHAR                 = wx_define_c_func( "get_rich_event_char", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_EVENT_FLAGS                = wx_define_c_func( "get_rich_event_flags", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_EVENT_NEW_STYLE_SHEET      = wx_define_c_func( "get_rich_event_new_style_sheet", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_EVENT_OLD_STYLE_SHEET      = wx_define_c_func( "get_rich_event_old_style_sheet", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_EVENT_RANGE                = wx_define_c_func( "get_rich_event_range", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_INTERNAL_SELECTION_RANGE   = wx_define_c_func( "get_rich_internal_selection_range", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_SELECTION_RANGE            = wx_define_c_func( "get_rich_selection_range", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICH_STYLE_SHEET                = wx_define_c_func( "get_rich_style_sheet", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_ROW_LABEL_SIZE                  = wx_define_c_func( "get_row_label_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_ROW_SIZE                        = wx_define_c_func( "get_row_size", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_SASH_GRAVITY                    = wx_define_c_func( "get_sash_gravity", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SCROLL_PAGE_SIZE                = wx_define_c_func( "get_scroll_page_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SCROLL_PIXELS_PER_UNIT          = wx_define_c_func( "get_scroll_pixels_per_unit", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SCROLL_POSITION                 = wx_define_c_func( "get_scroll_position", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SCROLL_RANGE                    = wx_define_c_func( "get_scroll_range", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SCROLL_THUMB_SIZE               = wx_define_c_func( "get_scroll_thumb_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SELECTION                       = wx_define_c_func( "get_selection", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SHADOW_WIDTH                    = wx_define_c_func( "get_shadow_width", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SHORT_PATH                      = wx_define_c_func( "get_short_path", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_SHOWN_UI                        = wx_define_c_func( "get_shown_ui", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SIZER_MINSIZE                   = wx_define_c_func( "get_sizer_minsize", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SIZER_POSITION                  = wx_define_c_func( "get_sizer_position", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SIZER_SIZE                      = wx_define_c_func( "get_sizer_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SIZER_WINDOW                    = wx_define_c_func( "get_sizer_window", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SIZE                            = wx_define_c_func( "get_size", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_SLIDE_VALUE                     = wx_define_c_func( "get_slide_value", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SOCKET_EVENT                    = wx_define_c_func( "get_socket_event", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SPIN_VALUE                      = wx_define_c_func( "get_spin_value", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SPLITTER_SASH_POSITION          = wx_define_c_func( "get_splitter_sash_position", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_STATICBITMAP                    = wx_define_c_func( "get_staticbitmap", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_STATUS_BAR_PANE                 = wx_define_c_func( "get_status_bar_pane", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_STATUS_BAR                      = wx_define_c_func( "get_status_bar", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_STATUS_FIELDS_COUNT             = wx_define_c_func( "get_status_fields_count", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_STATUS_TEXT                     = wx_define_c_func( "get_status_text", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_STRING_SELECTION                = wx_define_c_func( "get_string_selection", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SYS_COLOR                       = wx_define_c_func( "get_sys_color", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SYS_METRIC                      = wx_define_c_func( "get_sys_metric", {dll:C_INT}, E_OBJECT ),
	WX_GET_TEXT_ATTR_BACK_COLOR            = wx_define_c_func( "get_text_attr_back_color", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXT_ATTR_COLOR                 = wx_define_c_func( "get_text_attr_color", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_BULLET_FONT          = wx_define_c_func( "get_textattrex_bullet_font", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_BULLET_NAME          = wx_define_c_func( "get_textattrex_bullet_name", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_BULLET_NUMBER        = wx_define_c_func( "get_textattrex_bullet_number", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_BULLET_SPACING       = wx_define_c_func( "get_textattrex_bullet_spacing", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_BULLET_TEXT          = wx_define_c_func( "get_textattrex_bullet_text", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_CHAR_STYLE           = wx_define_c_func( "get_textattrex_char_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_EFFECT_FLAGS         = wx_define_c_func( "get_textattrex_effect_flags", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_EFFECTS              = wx_define_c_func( "get_textattrex_effects", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_LINE_SPACING         = wx_define_c_func( "get_textattrex_line_spacing", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_LIST_STYLE           = wx_define_c_func( "get_textattrex_list_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_OUTLINE_LEVEL        = wx_define_c_func( "get_textattrex_outline_level", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_PARA_SPACING_AFTER   = wx_define_c_func( "get_textattrex_para_spacing_after", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_PARA_SPACING_BEFORE  = wx_define_c_func( "get_textattrex_para_spacing_before", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_PARA_STYLE           = wx_define_c_func( "get_textattrex_para_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXTATTREX_URL                  = wx_define_c_func( "get_textattrex_url", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXT_ATTR_FONT                  = wx_define_c_func( "get_text_attr_font", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXT_EXTENT                     = wx_define_c_func( "get_text_extent", {dll:C_POINTER,dll:E_OBJECT,dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXT_INSERTION_POINT            = wx_define_c_func( "get_text_insertion_point", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXT_LENGTH                     = wx_define_c_func( "get_text_length", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXT_LINE_LENGTH                = wx_define_c_func( "get_text_line_length", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_TEXT_LINES                      = wx_define_c_func( "get_text_lines", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXT_LINE                       = wx_define_c_func( "get_text_line", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_TEXT_RANGE                      = wx_define_c_func( "get_text_range", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXT_SELECTION_RANGE            = wx_define_c_func( "get_text_selection_range", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXT_SELECTION                  = wx_define_c_func( "get_text_selection", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXT_STYLE                      = wx_define_c_func( "get_text_style", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_TEXT_UI                         = wx_define_c_func( "get_text_ui", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TEXT_VALUE                      = wx_define_c_func( "get_text_value", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_THE_OBJECT                      = wx_define_c_func( "get_the_object", {}, E_OBJECT ),
	WX_GET_TOOL_BAR                        = wx_define_c_func( "get_tool_bar", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TOOLKIT_MAJOR_VERSION           = wx_define_c_func( "get_toolkit_major_version", {}, E_OBJECT ),
	WX_GET_TOOLKIT_MINOR_VERSION           = wx_define_c_func( "get_toolkit_minor_version", {}, E_OBJECT ),
	WX_GET_TOOL_STATE                      = wx_define_c_func( "get_tool_state", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_TOP_LEVEL_PARENT                = wx_define_c_func( "get_top_level_parent", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_EDIT_CONTROL               = wx_define_c_func( "get_tree_edit_control", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_EVENT_ITEM                 = wx_define_c_func( "get_tree_event_item", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_IMAGELIST                  = wx_define_c_func( "get_tree_imagelist", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_INDENT                     = wx_define_c_func( "get_tree_indent", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_ITEM_BACKGROUND_COLOR      = wx_define_c_func( "get_tree_item_background_color", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_ITEM_BOUNDING_RECT         = wx_define_c_func( "get_tree_item_bounding_rect", {dll:C_POINTER,dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_TREE_ITEM_CHILDREN_COUNT        = wx_define_c_func( "get_tree_item_children_count", {dll:C_POINTER,dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_TREE_ITEM_COUNT                 = wx_define_c_func( "get_tree_item_count", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_ITEM_FONT                  = wx_define_c_func( "get_tree_item_font", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_ITEM_IMAGE                 = wx_define_c_func( "get_tree_item_image", {dll:C_POINTER,dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_TREE_ITEM_TEXT_COLOR            = wx_define_c_func( "get_tree_item_text_color", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_PARENT                     = wx_define_c_func( "get_tree_parent", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_ROOT                       = wx_define_c_func( "get_tree_root", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_SELECTED_TEXT              = wx_define_c_func( "get_tree_selected_text", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_SELECTIONS                 = wx_define_c_func( "get_tree_selections", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_SELECTION                  = wx_define_c_func( "get_tree_selection", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TREE_TEXT                       = wx_define_c_func( "get_tree_text", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_UPDATE_INTERVAL_UI              = wx_define_c_func( "get_update_interval_ui", {}, E_OBJECT ),
	WX_GET_URL_INPUT                       = wx_define_c_func( "get_url_input", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_USER_SCALE                      = wx_define_c_func( "get_user_scale", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_VALUE                           = wx_define_c_func( "get_value", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_VIEW_START                      = wx_define_c_func( "get_view_start", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_VIRTUAL_SIZE                    = wx_define_c_func( "get_virtual_size", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_WHEEL_ROTATION                  = wx_define_c_func( "get_wheel_rotation", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_WINDOW_SCROLL_POSITION          = wx_define_c_func( "get_window_scroll_position", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_WINDOW_SIZER                    = wx_define_c_func( "get_window_sizer", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_WINDOW_STYLE                    = wx_define_c_func( "get_window_style", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_CONTENT                     = wx_define_c_func( "get_xml_content", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_ENCODING                    = wx_define_c_func( "get_xml_encoding", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_FILE_ENCODING               = wx_define_c_func( "get_xml_file_encoding", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_NODE_CHILDREN               = wx_define_c_func( "get_xml_node_children", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_NODE_CONTENT                = wx_define_c_func( "get_xml_node_content", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_NODE_DEPTH                  = wx_define_c_func( "get_xml_node_depth", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_NODE_NAME                   = wx_define_c_func( "get_xml_node_name", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_NODE_PROPERTIES             = wx_define_c_func( "get_xml_node_properties", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_NODE_PROP_VAL               = wx_define_c_func( "get_xml_node_prop_val", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_XML_NODE_TYPE                   = wx_define_c_func( "get_xml_node_type", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_PROP_NAME                   = wx_define_c_func( "get_xml_prop_name", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_PROP_VALUE                  = wx_define_c_func( "get_xml_prop_value", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_ROOT                        = wx_define_c_func( "get_xml_root", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_XML_VERSION                     = wx_define_c_func( "get_xml_version", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_ZOOM                            = wx_define_c_func( "get_zoom", {dll:C_POINTER}, E_OBJECT ),
	WX_GRID_CELL_TO_RECT                   = wx_define_c_func( "grid_cell_to_rect", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_GRID_EVENT_CELL                     = wx_define_c_func( "grid_event_cell", {dll:C_POINTER}, E_OBJECT ),
	WX_GRID_POS                            = wx_define_c_func( "grid_pos", {dll:C_POINTER}, E_OBJECT ),
	WX_GRID_SIZE                           = wx_define_c_func( "grid_size", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_ALIGNMENT_RICH_ATTR             = wx_define_c_func( "has_alignment_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_BACK_COLOR_RICH_ATTR            = wx_define_c_func( "has_back_color_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_BULLET_NAME_RICH_ATTR           = wx_define_c_func( "has_bullet_name_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_BULLET_NAME_TEXTATTREX          = wx_define_c_func( "has_bullet_name_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_BULLET_NUMBER_RICH_ATTR         = wx_define_c_func( "has_bullet_number_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_BULLET_NUMBER_TEXTATTREX        = wx_define_c_func( "has_bullet_number_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_BULLET_STYLE_RICH_ATTR          = wx_define_c_func( "has_bullet_style_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_BULLET_STYLE_TEXTATTREX         = wx_define_c_func( "has_bullet_style_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_BULLET_TEXT_RICH_ATTR           = wx_define_c_func( "has_bullet_text_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_BULLET_TEXT_TEXTATTREX          = wx_define_c_func( "has_bullet_text_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_CHAR_RICH_ATTR                  = wx_define_c_func( "has_char_rich_attr", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_POINTER}, E_OBJECT ),
	WX_HAS_CHAR_STYLE_RICH_ATTR            = wx_define_c_func( "has_char_style_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_CHAR_STYLE_TEXTATTREX           = wx_define_c_func( "has_char_style_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_CHAR_TEXT_ATTR                  = wx_define_c_func( "has_char_text_attr", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_POINTER}, E_OBJECT ),
	WX_HAS_EFFECT_RICH_ATTR                = wx_define_c_func( "has_effect_rich_attr", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_HAS_EFFECTS_RICH_ATTR               = wx_define_c_func( "has_effects_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_FILES                           = wx_define_c_func( "has_files", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_HAS_FLAG_RICH_ATTR                  = wx_define_c_func( "has_flag_rich_attr", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_HAS_FONT_ITALIC_TEXTATTREX          = wx_define_c_func( "has_font_italic_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_FONT_NAME_RICH_ATTR             = wx_define_c_func( "has_font_name_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_FONT_NAME_TEXTATTREX            = wx_define_c_func( "has_font_name_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_FONT_RICH_ATTR                  = wx_define_c_func( "has_font_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_FONT_SIZE_RICH_ATTR             = wx_define_c_func( "has_font_size_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_FONT_SIZE_TEXTATTREX            = wx_define_c_func( "has_font_size_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_FONT_UNDERLINED_TEXTATTREX      = wx_define_c_func( "has_font_underlined_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_FONT_WEIGHT_RICH_ATTR           = wx_define_c_func( "has_font_weight_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_FONT_WEIGHT_TEXTATTREX          = wx_define_c_func( "has_font_weight_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_ITALIC_RICH_ATTR                = wx_define_c_func( "has_italic_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_LEFT_INDENT_RICH_ATTR           = wx_define_c_func( "has_left_indent_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_LINE_SPACING_RICH_ATTR          = wx_define_c_func( "has_line_spacing_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_LINE_SPACING_TEXTATTREX         = wx_define_c_func( "has_line_spacing_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_LIST_STYLE_RICH_ATTR            = wx_define_c_func( "has_list_style_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_LIST_STYLE_TEXTATTREX           = wx_define_c_func( "has_list_style_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_OPTION                          = wx_define_c_func( "has_option", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_HAS_OUTLINE_LEVEL_RICH_ATTR         = wx_define_c_func( "has_outline_level_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_OUTLINE_LEVEL_TEXTATTREX        = wx_define_c_func( "has_outline_level_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_PAGE_BREAK_RICH_ATTR            = wx_define_c_func( "has_page_break_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_PAGE_BREAK_TEXTATTREX           = wx_define_c_func( "has_page_break_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_PARA_CHAR_ATTR                  = wx_define_c_func( "has_para_char_attr", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_POINTER}, E_OBJECT ),
	WX_HAS_PARA_SPACING_AFTER_RICH_ATTR    = wx_define_c_func( "has_para_spacing_after_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_PARA_SPACING_AFTER_TEXTATTREX   = wx_define_c_func( "has_para_spacing_after_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_PARA_SPACING_BEFORE_RICH_ATTR   = wx_define_c_func( "has_para_spacing_before_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_PARA_SPACING_BEFORE_TEXTATTREX  = wx_define_c_func( "has_para_spacing_before_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_PARA_STYLE_RICH_ATTR            = wx_define_c_func( "has_para_style_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_PARA_STYLE_TEXTATTREX           = wx_define_c_func( "has_para_style_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_PARA_TEXT_ATTR                  = wx_define_c_func( "has_para_text_attr", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_POINTER}, E_OBJECT ),
	WX_HAS_RIGHT_INDENT_RICH_ATTR          = wx_define_c_func( "has_right_indent_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_SUBDIRS                         = wx_define_c_func( "has_subdirs", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_HAS_TABS_RICH_ATTR                  = wx_define_c_func( "has_tabs_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_TEXT_COLOR_RICH_ATTR            = wx_define_c_func( "has_text_color_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_TEXT_EFFECTS_TEXTATTREX         = wx_define_c_func( "has_text_effects_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_TEXT_EFFECT_TEXTATTREX          = wx_define_c_func( "has_text_effect_textattrex", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_HAS_UNDERLINED_RICH_ATTR            = wx_define_c_func( "has_underlined_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_URL_RICH_ATTR                   = wx_define_c_func( "has_url_rich_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_HAS_URL_TEXTATTREX                  = wx_define_c_func( "has_url_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_HIDE_CARET                          = wx_define_c_proc( "hide_caret", {dll:C_POINTER} ),
	WX_HIDE_CELL_EDIT_CONTROL              = wx_define_c_proc( "hide_cell_edit_control", {dll:C_POINTER} ),
	WX_HIDE_HINT                           = wx_define_c_proc( "hide_hint", {dll:C_POINTER} ),
	WX_HIT_TEST                            = wx_define_c_func( "hit_test", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_HTML_APPEND                         = wx_define_c_proc( "html_append", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_HTML_BACK                           = wx_define_c_proc( "html_back", {dll:C_POINTER} ),
	WX_HTML_EASY_FONTS                     = wx_define_c_proc( "html_easy_fonts", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT,dll:E_OBJECT} ),
	WX_HTML_EASY_FOOTER                    = wx_define_c_proc( "html_easy_footer", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT} ),
	WX_HTML_EASY_HEADER                    = wx_define_c_proc( "html_easy_header", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT} ),
	WX_HTML_EASY_PAGE_SETUP                = wx_define_c_proc( "html_easy_page_setup", {dll:C_POINTER} ),
	WX_HTML_EASY_PARENT                    = wx_define_c_proc( "html_easy_parent", {dll:C_POINTER,dll:C_POINTER} ),
	WX_HTML_FORWARD                        = wx_define_c_proc( "html_forward", {dll:C_POINTER} ),
	WX_HTTP_CONNECT                        = wx_define_c_func( "http_connect", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_INITIALIZE_SPLIT                    = wx_define_c_proc( "initialize_split", {dll:C_POINTER,dll:C_POINTER} ),
	WX_INITIALIZE                          = wx_define_c_func( "initialize", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_INSERT_GRID_COLS                    = wx_define_c_func( "insert_grid_cols", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_INSERT_GRID_ROWS                    = wx_define_c_func( "insert_grid_rows", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_INSERT_ITEM                         = wx_define_c_proc( "insert_item", {dll:C_POINTER,dll:C_INT,dll:E_OBJECT} ),
	WX_INSERT_LIST_COLUMN                  = wx_define_c_func( "insert_list_column", {dll:C_POINTER,dll:C_INT,dll:E_OBJECT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_INSERT_LISTCTRL_ITEM                = wx_define_c_func( "insert_listctrl_item", {dll:C_POINTER,dll:C_INT,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_INSERT_MENUITEM                     = wx_define_c_func( "insert_menuitem", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:E_OBJECT,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_INSERT_MENU                         = wx_define_c_proc( "insert_menu", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT,dll:C_INT} ),
	WX_INSERT_PAGE                         = wx_define_c_proc( "insert_page", {dll:C_INT,dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT} ),
	WX_INSERT_PANE                         = wx_define_c_func( "insert_pane", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_INSERT_SEPARATOR                    = wx_define_c_proc( "insert_separator", {dll:C_POINTER,dll:C_INT} ),
	WX_INSERT_TREE_ITEM_AFTER              = wx_define_c_func( "insert_tree_item_after", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_INSERT_TREE_ITEM_BEFORE             = wx_define_c_func( "insert_tree_item_before", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_INSERT_XML_CHILD                    = wx_define_c_func( "insert_xml_child", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_IS_CELL_EDIT_CONTROL_ENABLED        = wx_define_c_func( "is_cell_edit_control_enabled", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_CELL_READ_ONLY                   = wx_define_c_func( "is_cell_read_only", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_IS_CELL_VISIBLE                     = wx_define_c_func( "is_cell_visible", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_IS_CHAR_STYLE_TEXTATTREX            = wx_define_c_func( "is_char_style_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_CONNECTED                        = wx_define_c_func( "is_connected", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_CURRENT_CELL_READ_ONLY           = wx_define_c_func( "is_current_cell_read_only", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_DEFAULT_TEXTATTREX               = wx_define_c_func( "is_default_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_GAUGE_VERTICAL                   = wx_define_c_func( "is_gauge_vertical", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_GRID_EDITABLE                    = wx_define_c_func( "is_grid_editable", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_IN_GRID_SELECTION                = wx_define_c_func( "is_in_grid_selection", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_IS_MENUITEM_CHECKED                 = wx_define_c_func( "is_menuitem_checked", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_OPENED                           = wx_define_c_func( "is_opened", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_PARA_STYLE_TEXTATTREX            = wx_define_c_func( "is_para_style_textattrex", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_PREVIEW                          = wx_define_c_func( "is_preview", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_RETAINED                         = wx_define_c_func( "is_retained", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_RICH_ATTR_CHAR_STYLE             = wx_define_c_func( "is_rich_attr_char_style", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_RICH_ATTR_DEFAULT                = wx_define_c_func( "is_rich_attr_default", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_RICH_ATTR_PARA_STYLE             = wx_define_c_func( "is_rich_attr_para_style", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_SELECTION_ALIGNED                = wx_define_c_func( "is_selection_aligned", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_IS_SELECTION_BOLD                   = wx_define_c_func( "is_selection_bold", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_SELECTION_FOR_GRID               = wx_define_c_func( "is_selection_for_grid", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_SELECTION_ITALICS                = wx_define_c_func( "is_selection_italics", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_SELECTION_UNDERLINED             = wx_define_c_func( "is_selection_underlined", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_TEXT_EDITABLE                    = wx_define_c_func( "is_text_editable", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_TEXT_EMPTY                       = wx_define_c_func( "is_text_empty", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_TEXT_MODIFIED                    = wx_define_c_func( "is_text_modified", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_TEXT_MULTI_LINE                  = wx_define_c_func( "is_text_multi_line", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_TEXT_SINGLE_LINE                 = wx_define_c_func( "is_text_single_line", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_TREE_EMPTY                       = wx_define_c_func( "is_tree_empty", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_TREE_ITEM_BOLD                   = wx_define_c_func( "is_tree_item_bold", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_IS_TREE_ITEM_EXPANDED               = wx_define_c_func( "is_tree_item_expanded", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_IS_TREE_ITEM_SELECTED               = wx_define_c_func( "is_tree_item_selected", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_IS_TREE_ITEM_VISIBLE                = wx_define_c_func( "is_tree_item_visible", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_IS_VALID_REGEX                      = wx_define_c_func( "is_valid_regex", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_VISIBLE                          = wx_define_c_func( "is_visible", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_XML_WHITESPACE_ONLY              = wx_define_c_func( "is_xml_whitespace_only", {dll:C_POINTER}, E_OBJECT ),
	WX_KEY_EVENT_ALTDOWN                   = wx_define_c_func( "key_event_altdown", {dll:C_POINTER}, E_OBJECT ),
	WX_KEY_EVENT_CONTROLDOWN               = wx_define_c_func( "key_event_controldown", {dll:C_POINTER}, E_OBJECT ),
	WX_KEY_EVENT_SHIFTDOWN                 = wx_define_c_func( "key_event_shiftdown", {dll:C_POINTER}, E_OBJECT ),
	WX_KILL_PROCESS                        = wx_define_c_func( "kill_process", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_LAST_CELL_VISIBLE                   = wx_define_c_func( "last_cell_visible", {dll:C_POINTER}, E_OBJECT ),
	WX_LAUNCH_DEFAULT_BROWSER              = wx_define_c_proc( "launch_default_browser", {dll:E_OBJECT} ),
	WX_LAYOUT_RICH_CONTENT                 = wx_define_c_proc( "layout_rich_content", {dll:C_POINTER,dll:C_INT} ),
	WX_LAYOUT_SIZER                        = wx_define_c_proc( "layout_sizer", {dll:C_POINTER} ),
	WX_LINE_BREAK                          = wx_define_c_proc( "line_break", {dll:C_POINTER} ),
	WX_LIST_COUNT                          = wx_define_c_func( "list_count", {dll:C_POINTER}, E_OBJECT ),
	WX_LIST_ENSURE_VISIBLE                 = wx_define_c_proc( "list_ensure_visible", {dll:C_POINTER,dll:C_POINTER} ),
	WX_LIST_HIT_TEST                       = wx_define_c_func( "list_hit_test", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_LIST_IMAGELIST                      = wx_define_c_proc( "list_imagelist", {dll:C_POINTER,dll:C_POINTER,dll:C_INT} ),
	WX_LOAD_HTML_FILE                      = wx_define_c_proc( "load_html_file", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_LOAD_HTML_PAGE                      = wx_define_c_proc( "load_html_page", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_LOAD_ICON                           = wx_define_c_func( "load_icon", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_LOAD_MEDIA_URI                      = wx_define_c_func( "load_media_uri", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_LOAD_MEDIA                          = wx_define_c_func( "load_media", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_LOAD_PANE_INFO                      = wx_define_c_proc( "load_pane_info", {dll:C_POINTER,dll:E_OBJECT,dll:C_POINTER} ),
	WX_LOAD_PERSPECTIVE                    = wx_define_c_func( "load_perspective", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_LOAD_TEXT_FILE                      = wx_define_c_func( "load_text_file", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_LOAD_XML                            = wx_define_c_func( "load_xml", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT}, E_OBJECT ),
	WX_MAKE_CELL_VISIBLE                   = wx_define_c_proc( "make_cell_visible", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_MARK_TEXT_DIRTY                     = wx_define_c_proc( "mark_text_dirty", {dll:C_POINTER} ),
	WX_MAXIMIZE_CHILD                      = wx_define_c_proc( "maximize_child", {dll:C_POINTER,dll:C_INT} ),
	WX_MESSAGE_BOX                         = wx_define_c_func( "message_box", {dll:E_OBJECT,dll:E_OBJECT,dll:C_POINTER}, E_OBJECT ),
	WX_MIRROR                              = wx_define_c_func( "mirror", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_WX_MKDIR                            = wx_define_c_func( "wx_mkdir", {dll:E_OBJECT,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_MOUSE_EVENT_POSITION                = wx_define_c_func( "mouse_event_position", {dll:C_POINTER}, E_OBJECT ),
	WX_MOUSE_EVENT_SHIFTDOWN               = wx_define_c_func( "mouse_event_shiftdown", {dll:C_POINTER}, E_OBJECT ),
	WX_MOVE_CARET                          = wx_define_c_proc( "move_caret", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_MOVE_DOWN                           = wx_define_c_proc( "move_down", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_MOVE_END                            = wx_define_c_proc( "move_end", {dll:C_POINTER,dll:C_INT} ),
	WX_MOVE_GRID_CURSOR_DOWN_BLOCK         = wx_define_c_func( "move_grid_cursor_down_block", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_MOVE_GRID_CURSOR_DOWN               = wx_define_c_func( "move_grid_cursor_down", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_MOVE_GRID_CURSOR_LEFT_BLOCK         = wx_define_c_func( "move_grid_cursor_left_block", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_MOVE_GRID_CURSOR_LEFT               = wx_define_c_func( "move_grid_cursor_left", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_MOVE_GRID_CURSOR_RIGHT_BLOCK        = wx_define_c_func( "move_grid_cursor_right_block", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_MOVE_GRID_CURSOR_RIGHT              = wx_define_c_func( "move_grid_cursor_right", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_MOVE_GRID_CURSOR_UP_BLOCK           = wx_define_c_func( "move_grid_cursor_up_block", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_MOVE_GRID_CURSOR_UP                 = wx_define_c_func( "move_grid_cursor_up", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_MOVE_GRID_PAGE_DOWN                 = wx_define_c_func( "move_grid_page_down", {dll:C_POINTER}, E_OBJECT ),
	WX_MOVE_GRID_PAGE_UP                   = wx_define_c_func( "move_grid_page_up", {dll:C_POINTER}, E_OBJECT ),
	WX_MOVE_HOME                           = wx_define_c_proc( "move_home", {dll:C_POINTER,dll:C_INT} ),
	WX_MOVE_LEFT                           = wx_define_c_proc( "move_left", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_MOVE_RICH_CARET                     = wx_define_c_proc( "move_rich_caret", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_MOVE_RIGHT                          = wx_define_c_proc( "move_right", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_MOVE_TO_LINE_END                    = wx_define_c_proc( "move_to_line_end", {dll:C_POINTER,dll:C_INT} ),
	WX_MOVE_TO_LINE_START                  = wx_define_c_proc( "move_to_line_start", {dll:C_POINTER,dll:C_INT} ),
	WX_MOVE_TO_PARAGRAPH_END               = wx_define_c_proc( "move_to_paragraph_end", {dll:C_POINTER,dll:C_INT} ),
	WX_MOVE_TO_PARAGRAPH_START             = wx_define_c_proc( "move_to_paragraph_start", {dll:C_POINTER,dll:C_INT} ),
	WX_MOVE_UP                             = wx_define_c_proc( "move_up", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_MOVE                                = wx_define_c_proc( "move", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_NEW_ID                              = wx_define_c_func( "new_id", {}, E_OBJECT ),
	WX_NEW_LINE                            = wx_define_c_proc( "new_line", {dll:C_POINTER} ),
	WX_NEW_WXABOUTDIALOGINFO               = wx_define_c_func( "new_wxAboutDialogInfo", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXAUIMANAGER                    = wx_define_c_func( "new_wxAuiManager", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXAUINOTEBOOK                   = wx_define_c_func( "new_wxAuiNotebook", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXAUIPANEINFO                   = wx_define_c_func( "new_wxAuiPaneInfo", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXBITMAPBUTTON                  = wx_define_c_func( "new_wxBitmapButton", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXBITMAPCOMBOBOX                = wx_define_c_func( "new_wxBitmapComboBox", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXBITMAP                        = wx_define_c_func( "new_wxBitmap", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXBOXSIZER                      = wx_define_c_func( "new_wxBoxSizer", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXBRUSH                         = wx_define_c_func( "new_wxBrush", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXBUTTON                        = wx_define_c_func( "new_wxButton", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCALENDARCTRL                  = wx_define_c_func( "new_wxCalendarCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCARET                         = wx_define_c_func( "new_wxCaret", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCHECKBOX                      = wx_define_c_func( "new_wxCheckBox", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCHOICEBOOK                    = wx_define_c_func( "new_wxChoicebook", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCHOICE                        = wx_define_c_func( "new_wxChoice", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCLIENTDC                      = wx_define_c_func( "new_wxClientDC", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCOLOURDIALOG                  = wx_define_c_func( "new_wxColourDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCOLOUR                        = wx_define_c_func( "new_wxColour", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCOMBOBOX                      = wx_define_c_func( "new_wxComboBox", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCONFIG                        = wx_define_c_func( "new_wxConfig", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCURSOR                        = wx_define_c_func( "new_wxCursor", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXDEFBUTTON                     = wx_define_c_func( "new_wxDefButton", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXDIALOG                        = wx_define_c_func( "new_wxDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXDIRCTRL                       = wx_define_c_func( "new_wxDirCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXDIRDIALOG                     = wx_define_c_func( "new_wxDirDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXDIRTRAVERSER                  = wx_define_c_func( "new_wxDirTraverser", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXDIR                           = wx_define_c_func( "new_wxDir", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXDISPLAY                       = wx_define_c_func( "new_wxDisplay", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXFILENAME                      = wx_define_c_func( "new_wxFileName", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXFLEXGRIDSIZER                 = wx_define_c_func( "new_wxFlexGridSizer", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXFONTDATA                      = wx_define_c_func( "new_wxFontData", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXFONT                          = wx_define_c_func( "new_wxFont", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXFRAME                         = wx_define_c_func( "new_wxFrame", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXFTP                           = wx_define_c_func( "new_wxFTP", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXGAUGE                         = wx_define_c_func( "new_wxGauge", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXGRIDCELLATTR                  = wx_define_c_func( "new_wxGridCellAttr", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXGRIDCELLAUTOWRAPSTRINGEDITOR  = wx_define_c_func( "new_wxGridCellAutoWrapStringEditor", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXGRIDCELLBOOLEDITOR            = wx_define_c_func( "new_wxGridCellBoolEditor", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXGRID                          = wx_define_c_func( "new_wxGrid", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXHTMLEASYPRINTING              = wx_define_c_func( "new_wxHtmlEasyPrinting", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXHTMLWINDOW                    = wx_define_c_func( "new_wxHtmlWindow", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXHTTP                          = wx_define_c_func( "new_wxHTTP", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXICON                          = wx_define_c_func( "new_wxIcon", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXIMAGELIST                     = wx_define_c_func( "new_wxImageList", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXIMAGE                         = wx_define_c_func( "new_wxImage", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXIPV4ADDRESS                   = wx_define_c_func( "new_wxIPV4address", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXLISTBOOK                      = wx_define_c_func( "new_wxListbook", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXLISTBOX                       = wx_define_c_func( "new_wxListBox", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXLISTCTRL                      = wx_define_c_func( "new_wxListCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXMASK                          = wx_define_c_func( "new_wxMask", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXMDICHILDFRAME                 = wx_define_c_func( "new_wxMDIChildFrame", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXMDICLIENTWINDOW               = wx_define_c_func( "new_wxMDIClientWindow", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXMDIPARENTFRAME                = wx_define_c_func( "new_wxMDIParentFrame", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXMEDIACTRL                     = wx_define_c_func( "new_wxMediaCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXMEMORYDC                      = wx_define_c_func( "new_wxMemoryDC", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXMENUBAR                       = wx_define_c_func( "new_wxMenuBar", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXMENUITEM                      = wx_define_c_func( "new_wxMenuItem", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXMENU                          = wx_define_c_func( "new_wxMenu", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXMULTICHOICEDIALOG             = wx_define_c_func( "new_wxMultiChoiceDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXNOTEBOOK                      = wx_define_c_func( "new_wxNotebook", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPAGESETUPDIALOG               = wx_define_c_func( "new_wxPageSetupDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPAINTDC                       = wx_define_c_func( "new_wxPaintDC", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPANEL                         = wx_define_c_func( "new_wxPanel", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPASSWORDENTRYDIALOG           = wx_define_c_func( "new_wxPasswordEntryDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPEN                           = wx_define_c_func( "new_wxPen", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPOPUPTRANSIENTWINDOW          = wx_define_c_func( "new_wxPopupTransientWindow", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPOPUPWINDOW                   = wx_define_c_func( "new_wxPopupWindow", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPREVIEWFRAME                  = wx_define_c_func( "new_wxPreviewFrame", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPRINTDIALOG                   = wx_define_c_func( "new_wxPrintDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPRINTER                       = wx_define_c_func( "new_wxPrinter", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPRINTOUT                      = wx_define_c_func( "new_wxPrintout", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPRINTPREVIEW                  = wx_define_c_func( "new_wxPrintPreview", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPROCESS                       = wx_define_c_func( "new_wxProcess", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPROGRESSDIALOG                = wx_define_c_func( "new_wxProgressDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXRADIOBOX                      = wx_define_c_func( "new_wxRadioBox", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXRADIOBUTTON                   = wx_define_c_func( "new_wxRadioButton", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXREGEX                         = wx_define_c_func( "new_wxRegEx", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXRICHTEXTATTR                  = wx_define_c_func( "new_wxRichTextAttr", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXRICHTEXTCTRL                  = wx_define_c_func( "new_wxRichTextCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXRICHTEXTFORMATTINGDIALOG      = wx_define_c_func( "new_wxRichTextFormattingDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSCROLLBAR                     = wx_define_c_func( "new_wxScrollBar", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSCROLLEDWINDOW                = wx_define_c_func( "new_wxScrolledWindow", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSLIDER                        = wx_define_c_func( "new_wxSlider", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSOCKETCLIENT                  = wx_define_c_func( "new_wxSocketClient", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSOCKETSERVER                  = wx_define_c_func( "new_wxSocketServer", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSOUND                         = wx_define_c_func( "new_wxSound", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSPINCTRL                      = wx_define_c_func( "new_wxSpinCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSPLASHSCREEN                  = wx_define_c_func( "new_wxSplashScreen", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSPLITTERWINDOW                = wx_define_c_func( "new_wxSplitterWindow", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSTATICBITMAP                  = wx_define_c_func( "new_wxStaticBitmap", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSTATICBOXSIZER                = wx_define_c_func( "new_wxStaticBoxSizer", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSTATICBOX                     = wx_define_c_func( "new_wxStaticBox", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSTATICLINE                    = wx_define_c_func( "new_wxStaticLine", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSTATICTEXT                    = wx_define_c_func( "new_wxStaticText", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSTATUSBAR                     = wx_define_c_func( "new_wxStatusBar", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSTDDIALOGBUTTONSIZER          = wx_define_c_func( "new_wxStdDialogButtonSizer", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXTEXTATTREX                    = wx_define_c_func( "new_wxTextAttrEx", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXTEXTATTR                      = wx_define_c_func( "new_wxTextAttr", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXTEXTCTRL                      = wx_define_c_func( "new_wxTextCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXTEXTENTRYDIALOG               = wx_define_c_func( "new_wxTextEntryDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXTIMER                         = wx_define_c_func( "new_wxTimer", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXTOGGLEBUTTON                  = wx_define_c_func( "new_wxToggleButton", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXTOOLBAR                       = wx_define_c_func( "new_wxToolBar", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXTOOLBOOK                      = wx_define_c_func( "new_wxToolbook", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXTREEBOOK                      = wx_define_c_func( "new_wxTreebook", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXTREECTRL                      = wx_define_c_func( "new_wxTreeCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXURL                           = wx_define_c_func( "new_wxURL", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXWINDOW                        = wx_define_c_func( "new_wxWindow", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXWIZARDPAGE                    = wx_define_c_func( "new_wxWizardPage", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXWIZARD                        = wx_define_c_func( "new_wxWizard", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXXMLDOCUMENT                   = wx_define_c_func( "new_wxXmlDocument", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXXMLNODE                       = wx_define_c_func( "new_wxXmlNode", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXXMLPROPERTY                   = wx_define_c_func( "new_wxXmlProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXXMLRESOURCE                   = wx_define_c_func( "new_wxXmlResource", {dll:E_OBJECT}, E_OBJECT ),
	WX_NOTEBOOK_EVENT_SELECTION            = wx_define_c_func( "notebook_event_selection", {dll:C_POINTER}, E_OBJECT ),
	WX_NOTEBOOK_IMAGELIST                  = wx_define_c_proc( "notebook_imagelist", {dll:C_POINTER,dll:C_POINTER} ),
	WX_OPEN_DIR                            = wx_define_c_func( "open_dir", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_PAGE_DOWN                           = wx_define_c_proc( "page_down", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_PAGE_UP                             = wx_define_c_proc( "page_up", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_PASTE_TEXT                          = wx_define_c_proc( "paste_text", {dll:C_POINTER} ),
	WX_PAUSE_MEDIA                         = wx_define_c_func( "pause_media", {dll:C_POINTER}, E_OBJECT ),
	WX_PLATFORM_INFO_OK                    = wx_define_c_func( "platform_info_ok", {}, E_OBJECT ),
	WX_PLAY_MEDIA                          = wx_define_c_func( "play_media", {dll:C_POINTER}, E_OBJECT ),
	WX_POP_RICH_STYLE_SHEET                = wx_define_c_func( "pop_rich_style_sheet", {dll:C_POINTER}, E_OBJECT ),
	WX_POP_STATUS_TEXT                     = wx_define_c_proc( "pop_status_text", {dll:C_POINTER,dll:C_INT} ),
	WX_POPUP_MENU                          = wx_define_c_proc( "popup_menu", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_POSITION_POPUP                      = wx_define_c_proc( "position_popup", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT} ),
	WX_POSITION_TO_XY                      = wx_define_c_func( "position_to_xy", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_PREPARE_DC                          = wx_define_c_proc( "prepare_dc", {dll:C_POINTER,dll:C_POINTER} ),
	WX_PREPEND_TREE_ITEM                   = wx_define_c_func( "prepend_tree_item", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_PRINT_HTML_FILE                     = wx_define_c_func( "print_html_file", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_PRINT_HTML_TEXT                     = wx_define_c_func( "print_html_text", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT}, E_OBJECT ),
	WX_PROCESS_ERROR_AVAILABLE             = wx_define_c_func( "process_error_available", {dll:C_POINTER}, E_OBJECT ),
	WX_PROCESS_ERROR_STREAM                = wx_define_c_func( "process_error_stream", {dll:C_POINTER}, E_OBJECT ),
	WX_PROCESS_EVENT_EXITCODE              = wx_define_c_func( "process_event_exitcode", {dll:C_POINTER}, E_OBJECT ),
	WX_PROCESS_EVENT_PID                   = wx_define_c_func( "process_event_pid", {dll:C_POINTER}, E_OBJECT ),
	WX_PROCESS_EXISTS                      = wx_define_c_func( "process_exists", {dll:C_POINTER}, E_OBJECT ),
	WX_PROCESS_INPUT_AVAILABLE             = wx_define_c_func( "process_input_available", {dll:C_POINTER}, E_OBJECT ),
	WX_PROCESS_INPUT_OPENED                = wx_define_c_func( "process_input_opened", {dll:C_POINTER}, E_OBJECT ),
	WX_PROCESS_INPUT_STREAM                = wx_define_c_func( "process_input_stream", {dll:C_POINTER}, E_OBJECT ),
	WX_PROCESS_OUTPUT_STREAM               = wx_define_c_func( "process_output_stream", {dll:C_POINTER}, E_OBJECT ),
	WX_PROCESS_REDIRECT                    = wx_define_c_proc( "process_redirect", {dll:C_POINTER} ),
	WX_PROGRESS_RESUME                     = wx_define_c_proc( "progress_resume", {dll:C_POINTER} ),
	WX_PROGRESS_UPDATE                     = wx_define_c_func( "progress_update", {dll:C_POINTER,dll:C_INT,dll:E_OBJECT}, E_OBJECT ),
	WX_PULSE                               = wx_define_c_proc( "pulse", {dll:C_POINTER} ),
	WX_PUSH_RICH_STYLE_SHEET               = wx_define_c_proc( "push_rich_style_sheet", {dll:C_POINTER,dll:C_POINTER} ),
	WX_PUSH_STATUS_TEXT                    = wx_define_c_proc( "push_status_text", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT} ),
	WX_RAISE_WINDOW                        = wx_define_c_proc( "raise_window", {dll:C_POINTER} ),
	WX_REALIZE                             = wx_define_c_proc( "realize", {dll:C_POINTER} ),
	WX_RECREATE_DIRCTRL_TREE               = wx_define_c_proc( "recreate_dirctrl_tree", {dll:C_POINTER} ),
	WX_REDO_TEXT                           = wx_define_c_proc( "redo_text", {dll:C_POINTER} ),
	WX_REFRESH_WINDOW                      = wx_define_c_proc( "refresh_window", {dll:E_OBJECT} ),
	WX_REGEX_MATCHES                       = wx_define_c_func( "regex_matches", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_RELEASE_MOUSE                       = wx_define_c_proc( "release_mouse", {dll:C_POINTER} ),
	WX_REMOVE_FILE                         = wx_define_c_func( "remove_file", {dll:E_OBJECT}, E_OBJECT ),
	WX_REMOVE_MENUITEM                     = wx_define_c_proc( "remove_menuitem", {dll:C_POINTER,dll:C_POINTER} ),
	WX_REMOVE_PAGE                         = wx_define_c_proc( "remove_page", {dll:C_POINTER,dll:C_POINTER} ),
	WX_REMOVE_XML_NODE_CHILD               = wx_define_c_func( "remove_xml_node_child", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_REPLACE_REGEX                       = wx_define_c_func( "replace_regex", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_REPLACE_SIZER_ITEM                  = wx_define_c_func( "replace_sizer_item", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_REPLACE_TEXT                        = wx_define_c_proc( "replace_text", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:E_OBJECT} ),
	WX_RESCALE                             = wx_define_c_func( "rescale", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_RESIZE                              = wx_define_c_func( "resize", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT}, E_OBJECT ),
	WX_RESTORE_CHILD                       = wx_define_c_proc( "restore_child", {dll:C_POINTER} ),
	WX_RICH_TO_TEXTEX_ATTR                 = wx_define_c_func( "rich_to_textex_attr", {dll:C_POINTER}, E_OBJECT ),
	WX_ROTATE_90                           = wx_define_c_func( "rotate_90", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_ROTATE                              = wx_define_c_func( "rotate", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT,dll:C_INT,dll:E_OBJECT}, E_OBJECT ),
	WX_RUN_WIZARD                          = wx_define_c_func( "run_wizard", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_SAVE_BITMAP                         = wx_define_c_func( "save_bitmap", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_POINTER}, E_OBJECT ),
	WX_SAVE_GRID_EDIT_CONTROL_VALUE        = wx_define_c_proc( "save_grid_edit_control_value", {dll:C_POINTER} ),
	WX_SAVE_PANE_INFO                      = wx_define_c_func( "save_pane_info", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_SAVE_PERSPECTIVE                    = wx_define_c_func( "save_perspective", {dll:C_POINTER}, E_OBJECT ),
	WX_SAVE_TEXT_FILE                      = wx_define_c_func( "save_text_file", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_SAVE_XML                            = wx_define_c_func( "save_xml", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_SCREEN_TO_CLIENT                    = wx_define_c_func( "screen_to_client", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_SCROLL_TO_TREE_ITEM                 = wx_define_c_proc( "scroll_to_tree_item", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SCROLL_WINDOW                       = wx_define_c_proc( "scroll_window", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SEEK_MEDIA                          = wx_define_c_proc( "seek_media", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SELECT_ALL_TEXT                     = wx_define_c_proc( "select_all_text", {dll:C_POINTER} ),
	WX_SELECT_GRID_ALL                     = wx_define_c_proc( "select_grid_all", {dll:C_POINTER} ),
	WX_SELECT_GRID_BLOCK                   = wx_define_c_proc( "select_grid_block", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_SELECT_GRID_CELL                    = wx_define_c_proc( "select_grid_cell", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SELECT_GRID_COL                     = wx_define_c_proc( "select_grid_col", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SELECT_GRID_ROW                     = wx_define_c_proc( "select_grid_row", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SELECT_NONE                         = wx_define_c_proc( "select_none", {dll:C_POINTER} ),
	WX_SELECT_OBJECT                       = wx_define_c_proc( "select_object", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SELECT_TREE_ITEM                    = wx_define_c_proc( "select_tree_item", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SELECT_WORD                         = wx_define_c_proc( "select_word", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_ART_PROVIDER                    = wx_define_c_proc( "set_art_provider", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_AUI_FLAGS                       = wx_define_c_proc( "set_aui_flags", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_BACK_BRUSH                      = wx_define_c_proc( "set_back_brush", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_BACK_COLOR                      = wx_define_c_proc( "set_back_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_BACKGROUND_MODE                 = wx_define_c_proc( "set_background_mode", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_BASIC_RICH_STYLE                = wx_define_c_proc( "set_basic_rich_style", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_BASIC_RICH_TEXT_STYLE           = wx_define_c_proc( "set_basic_rich_text_style", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_BEZEL_FACE                      = wx_define_c_proc( "set_bezel_face", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_BITMAP_DISABLED                 = wx_define_c_proc( "set_bitmap_disabled", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_BITMAP_FOCUS                    = wx_define_c_proc( "set_bitmap_focus", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_BITMAP_HOVER                    = wx_define_c_proc( "set_bitmap_hover", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_BITMAP_LABEL                    = wx_define_c_proc( "set_bitmap_label", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_BITMAP_SELECTED                 = wx_define_c_proc( "set_bitmap_selected", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_BITMAP_SIZE                     = wx_define_c_proc( "set_bitmap_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_BMPBUTTON                       = wx_define_c_proc( "set_bmpbutton", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_BRUSH                           = wx_define_c_proc( "set_brush", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_BUTTON_LABEL                    = wx_define_c_proc( "set_button_label", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_CARET_BLINK_TIME                = wx_define_c_proc( "set_caret_blink_time", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_CARET_SIZE                      = wx_define_c_proc( "set_caret_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_CELL_ALIGNMENT                  = wx_define_c_proc( "set_cell_alignment", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_SET_CELL_BACKGROUND_COLOR           = wx_define_c_proc( "set_cell_background_color", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_POINTER} ),
	WX_SET_CELL_EDITOR                     = wx_define_c_proc( "set_cell_editor", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:E_OBJECT} ),
	WX_SET_CELL_FONT                       = wx_define_c_proc( "set_cell_font", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_POINTER} ),
	WX_SET_CELL_READONLY                   = wx_define_c_proc( "set_cell_readonly", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_SET_CELL_TEXT_COLOR                 = wx_define_c_proc( "set_cell_text_color", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_POINTER} ),
	WX_SET_CELL_VALUE                      = wx_define_c_proc( "set_cell_value", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT} ),
	WX_SET_CELL                            = wx_define_c_proc( "set_cell", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_CHECKED                         = wx_define_c_proc( "set_checked", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_CHOICE                          = wx_define_c_proc( "set_choice", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_CLIP_TEXT                       = wx_define_c_proc( "set_clip_text", {dll:E_OBJECT} ),
	WX_SET_COL_ALIGNMENT                   = wx_define_c_proc( "set_col_alignment", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_SET_COL_EDITOR                      = wx_define_c_proc( "set_col_editor", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:E_OBJECT} ),
	WX_SET_COL_FORMAT_BOOL                 = wx_define_c_proc( "set_col_format_bool", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_COL_FORMAT_FLOAT                = wx_define_c_proc( "set_col_format_float", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_SET_COL_FORMAT_NUMBER               = wx_define_c_proc( "set_col_format_number", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_COL_LABEL_ALIGN                 = wx_define_c_proc( "set_col_label_align", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_COL_LABEL_SIZE                  = wx_define_c_proc( "set_col_label_size", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_COL_LABEL                       = wx_define_c_proc( "set_col_label", {dll:C_POINTER,dll:C_INT,dll:E_OBJECT} ),
	WX_SET_COL_POS                         = wx_define_c_proc( "set_col_pos", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_COL_READONLY                    = wx_define_c_proc( "set_col_readonly", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_COL_SIZE                        = wx_define_c_proc( "set_col_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_COLUMN_LABEL                    = wx_define_c_proc( "set_column_label", {dll:C_POINTER,dll:C_INT,dll:E_OBJECT} ),
	WX_SET_COMBO_ITEM_BITMAP               = wx_define_c_proc( "set_combo_item_bitmap", {dll:C_POINTER,dll:C_INT,dll:C_POINTER} ),
	WX_SET_COMBO_VALUE                     = wx_define_c_proc( "set_combo_value", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_CURSOR                          = wx_define_c_proc( "set_cursor", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_DATE_BACK_COLOR                 = wx_define_c_proc( "set_date_back_color", {dll:C_POINTER,dll:E_OBJECT,dll:C_POINTER} ),
	WX_SET_DATE_BORDER_COLOR               = wx_define_c_proc( "set_date_border_color", {dll:C_POINTER,dll:E_OBJECT,dll:C_POINTER} ),
	WX_SET_DATE_BORDER                     = wx_define_c_proc( "set_date_border", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT} ),
	WX_SET_DATE_FONT                       = wx_define_c_proc( "set_date_font", {dll:C_POINTER,dll:E_OBJECT,dll:C_POINTER} ),
	WX_SET_DATE_HOLIDAY                    = wx_define_c_proc( "set_date_holiday", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT} ),
	WX_SET_DATE_TEXT_COLOR                 = wx_define_c_proc( "set_date_text_color", {dll:C_POINTER,dll:E_OBJECT,dll:C_POINTER} ),
	WX_SET_DATE                            = wx_define_c_proc( "set_date", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_DEFAULT_CELL_ALIGN              = wx_define_c_proc( "set_default_cell_align", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_DEFAULT_CELL_BACKGROUND_COLOR   = wx_define_c_proc( "set_default_cell_background_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_DEFAULT_CELL_FONT               = wx_define_c_proc( "set_default_cell_font", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_DEFAULT_CELL_TEXT_COLOR         = wx_define_c_proc( "set_default_cell_text_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_DEFAULT_COL_SIZE                = wx_define_c_proc( "set_default_col_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_DEFAULT_FONT_ENCODING           = wx_define_c_proc( "set_default_font_encoding", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_DEFAULT_FONT                    = wx_define_c_proc( "set_default_font", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_DEFAULT_GRID_EDITOR             = wx_define_c_proc( "set_default_grid_editor", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_DEFAULT_ITEM                    = wx_define_c_proc( "set_default_item", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_DEFAULT_ROW_SIZE                = wx_define_c_proc( "set_default_row_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_DEFAULT_STYLE_TO_CURSOR_STYLE   = wx_define_c_proc( "set_default_style_to_cursor_style", {dll:C_POINTER} ),
	WX_SET_DEFAULT_TEXT_STYLE              = wx_define_c_proc( "set_default_text_style", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_DIRCTRL_DEFAULT_PATH            = wx_define_c_proc( "set_dirctrl_default_path", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_DIRCTRL_FILTER_INDEX            = wx_define_c_proc( "set_dirctrl_filter_index", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_DIRCTRL_FILTER                  = wx_define_c_proc( "set_dirctrl_filter", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_DIRCTRL_PATH                    = wx_define_c_proc( "set_dirctrl_path", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_DOCK_SIZE_CONSTRAINT            = wx_define_c_proc( "set_dock_size_constraint", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_SET_EDITOR_CHOICES                  = wx_define_c_proc( "set_editor_choices", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT} ),
	WX_SET_EVENT_HANDLER                   = wx_define_c_proc( "set_event_handler", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_FIRST_LIST_ITEM                 = wx_define_c_proc( "set_first_list_item", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_FLAGS                           = wx_define_c_proc( "set_flags", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_FOCUS                           = wx_define_c_proc( "set_focus", {dll:C_POINTER} ),
	WX_SET_FONT_FACE_NAME                  = wx_define_c_proc( "set_font_face_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_FONT_FAMILTY                    = wx_define_c_proc( "set_font_familty", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_FONT_STYLE                      = wx_define_c_proc( "set_font_style", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_FONT_UNDERLINED                 = wx_define_c_proc( "set_font_underlined", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_FONT_WEIGHT                     = wx_define_c_proc( "set_font_weight", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_FONT                            = wx_define_c_proc( "set_font", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_FORE_COLOR                      = wx_define_c_proc( "set_fore_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_GRID_EDITABLE                   = wx_define_c_proc( "set_grid_editable", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_GRID_LABEL_BACKGROUND_COLOR     = wx_define_c_proc( "set_grid_label_background_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_GRID_LABEL_FONT                 = wx_define_c_proc( "set_grid_label_font", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_GRID_LABEL_TEXT_COLOR           = wx_define_c_proc( "set_grid_label_text_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_GRID_LINE_COLOR                 = wx_define_c_proc( "set_grid_line_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_GRID_MARGINS                    = wx_define_c_proc( "set_grid_margins", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_GRID_SCROLL_LINE                = wx_define_c_proc( "set_grid_scroll_line", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_GRID_SELECTION_BACKGROUND       = wx_define_c_proc( "set_grid_selection_background", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_GRID_SELECTION_FOREGROUND       = wx_define_c_proc( "set_grid_selection_foreground", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_GRID_SELECTION_MODE             = wx_define_c_proc( "set_grid_selection_mode", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_GRID_SIZE                       = wx_define_c_proc( "set_grid_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_HTML_PAGE                       = wx_define_c_proc( "set_html_page", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_HTTP_HEADER                     = wx_define_c_proc( "set_http_header", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT} ),
	WX_SET_HTTP_POST_BUFFER                = wx_define_c_proc( "set_http_post_buffer", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_HTTP_PROXY_MODE                 = wx_define_c_proc( "set_http_proxy_mode", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_ICON                            = wx_define_c_proc( "set_icon", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_INITIAL_FONT                    = wx_define_c_proc( "set_initial_font", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_IP_HOST                         = wx_define_c_func( "set_ip_host", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_SET_IP_SERVICE                      = wx_define_c_func( "set_ip_service", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_SET_LABEL                           = wx_define_c_proc( "set_label", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_LINK_EVENT                      = wx_define_c_proc( "set_link_event", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_LIST_ITEM_BACK_COLOR            = wx_define_c_proc( "set_list_item_back_color", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_LIST_ITEM_IMAGE                 = wx_define_c_proc( "set_list_item_image", {dll:C_POINTER,dll:C_POINTER,dll:C_INT} ),
	WX_SET_LIST_ITEM                       = wx_define_c_proc( "set_list_item", {dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:E_OBJECT,dll:C_INT} ),
	WX_SET_LIST_SELECTION                  = wx_define_c_proc( "set_list_selection", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_MANAGED_WINDOW                  = wx_define_c_proc( "set_managed_window", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_MASK                            = wx_define_c_proc( "set_mask", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_MAXIMUM_TEXT_LENGTH             = wx_define_c_proc( "set_maximum_text_length", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_MDI_CHILD_TITLE                 = wx_define_c_proc( "set_mdi_child_title", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_MEDIA_VOLUME                    = wx_define_c_func( "set_media_volume", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_SET_MENU_LABEL                      = wx_define_c_proc( "set_menu_label", {dll:C_POINTER,dll:C_INT,dll:E_OBJECT} ),
	WX_SET_MIN_PANE_SIZE                   = wx_define_c_proc( "set_min_pane_size", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_MIN_STATUS_HEIGHT               = wx_define_c_proc( "set_min_status_height", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_NATIVE_FONT_INFO_USER_DESC      = wx_define_c_func( "set_native_font_info_user_desc", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_SET_NATIVE_FONT_INFO                = wx_define_c_func( "set_native_font_info", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_SET_NEXT_WIZARD_PAGE                = wx_define_c_proc( "set_next_wizard_page", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_NEXT_XML_NODE                   = wx_define_c_proc( "set_next_xml_node", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_NEXT_XML_PROP                   = wx_define_c_proc( "set_next_xml_prop", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_OPEN_URL_EVENT                  = wx_define_c_proc( "set_open_url_event", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_OPTION                          = wx_define_c_proc( "set_option", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT} ),
	WX_SET_PADDING                         = wx_define_c_proc( "set_padding", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_PAGE_IMAGE                      = wx_define_c_proc( "set_page_image", {dll:C_POINTER,dll:C_POINTER,dll:C_INT} ),
	WX_SET_PAGE_INFO                       = wx_define_c_proc( "set_page_info", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_PAGE_SELECTION                  = wx_define_c_proc( "set_page_selection", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_PAGE_TEXT                       = wx_define_c_proc( "set_page_text", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_PARENT_XML_MODE                 = wx_define_c_proc( "set_parent_xml_mode", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_PEN_COLOR                       = wx_define_c_proc( "set_pen_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_PEN_SIZE                        = wx_define_c_proc( "set_pen_size", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_PEN                             = wx_define_c_proc( "set_pen", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_POINT_SIZE                      = wx_define_c_proc( "set_point_size", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_PREVIEW_FRAME                   = wx_define_c_proc( "set_preview_frame", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_PREV_WIZARD_PAGE                = wx_define_c_proc( "set_prev_wizard_page", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RADIOBOX_SELECTION              = wx_define_c_proc( "set_radiobox_selection", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RANGE                           = wx_define_c_proc( "set_range", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RELATED_FRAME                   = wx_define_c_proc( "set_related_frame", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT,dll:C_INT} ),
	WX_SET_RICH_ATTR_ALIGNMENT             = wx_define_c_proc( "set_rich_attr_alignment", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_BACK_COLOR            = wx_define_c_proc( "set_rich_attr_back_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICH_ATTR_BULLET_FONT           = wx_define_c_proc( "set_rich_attr_bullet_font", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_RICH_ATTR_BULLET_NAME           = wx_define_c_proc( "set_rich_attr_bullet_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_RICH_ATTR_BULLET_NUMBER         = wx_define_c_proc( "set_rich_attr_bullet_number", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_BULLET_STYLE          = wx_define_c_proc( "set_rich_attr_bullet_style", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICH_ATTR_CHAR_STYLE            = wx_define_c_proc( "set_rich_attr_char_style", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_RICH_ATTR_EFFECT_FLAGS          = wx_define_c_proc( "set_rich_attr_effect_flags", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_EFFECTS               = wx_define_c_proc( "set_rich_attr_effects", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_FLAGS                 = wx_define_c_proc( "set_rich_attr_flags", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_FONT_NAME             = wx_define_c_proc( "set_rich_attr_font_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_RICH_ATTR_FONT_SIZE             = wx_define_c_proc( "set_rich_attr_font_size", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_FONT_STYLE            = wx_define_c_proc( "set_rich_attr_font_style", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICH_ATTR_LEFT_INDENT           = wx_define_c_proc( "set_rich_attr_left_indent", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_RICH_ATTR_LINE_SPACING          = wx_define_c_proc( "set_rich_attr_line_spacing", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_LIST_STYLE            = wx_define_c_proc( "set_rich_attr_list_style", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_RICH_ATTR_OUTLINE_LEVEL         = wx_define_c_proc( "set_rich_attr_outline_level", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_PAGE_BREAK            = wx_define_c_proc( "set_rich_attr_page_break", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_PARA_SPACING_AFTER    = wx_define_c_proc( "set_rich_attr_para_spacing_after", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_PARA_SPACING_BEFORE   = wx_define_c_proc( "set_rich_attr_para_spacing_before", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_PARA_STYLE            = wx_define_c_proc( "set_rich_attr_para_style", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_RICH_ATTR_RIGHT_INDENT          = wx_define_c_proc( "set_rich_attr_right_indent", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_TABS                  = wx_define_c_proc( "set_rich_attr_tabs", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_RICH_ATTR_TEXT_COLOR            = wx_define_c_proc( "set_rich_attr_text_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICH_ATTR_UNDERLINED            = wx_define_c_proc( "set_rich_attr_underlined", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_RICH_ATTR_URL                   = wx_define_c_proc( "set_rich_attr_url", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_RICH_INTERNAL_SELECTION_RANGE   = wx_define_c_proc( "set_rich_internal_selection_range", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_RICH_SELECTION_RANGE            = wx_define_c_proc( "set_rich_selection_range", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_RICH_STYLE_SHEET                = wx_define_c_proc( "set_rich_style_sheet", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_ROW_LABEL_ALIGN                 = wx_define_c_proc( "set_row_label_align", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_ROW_LABEL_SIZE                  = wx_define_c_proc( "set_row_label_size", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_ROW_LABEL                       = wx_define_c_proc( "set_row_label", {dll:C_POINTER,dll:C_INT,dll:E_OBJECT} ),
	WX_SET_ROW_SIZE                        = wx_define_c_proc( "set_row_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_SASH_GRAVITY                    = wx_define_c_proc( "set_sash_gravity", {dll:C_POINTER,dll:C_DOUBLE} ),
	WX_SET_SCROLLBARS                      = wx_define_c_proc( "set_scrollbars", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_SET_SCROLLBAR                       = wx_define_c_proc( "set_scrollbar", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_SET_SCROLL_POSITION                 = wx_define_c_proc( "set_scroll_position", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_SCROLL_RATE                     = wx_define_c_proc( "set_scroll_rate", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_SELECTION                       = wx_define_c_proc( "set_selection", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_SHADOW_WIDTH                    = wx_define_c_proc( "set_shadow_width", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_SIZER                           = wx_define_c_proc( "set_sizer", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_SIZE                            = wx_define_c_proc( "set_size", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_SLIDE_VALUE                     = wx_define_c_proc( "set_slide_value", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_SOCKET_NOTIFY                   = wx_define_c_proc( "set_socket_notify", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_SPIN_VALUE                      = wx_define_c_proc( "set_spin_value", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_SPLITTER_SASH_POSITION          = wx_define_c_proc( "set_splitter_sash_position", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_STATICBITMAP                    = wx_define_c_proc( "set_staticbitmap", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_STATUS_FIELDS                   = wx_define_c_proc( "set_status_fields", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_STATUS_STYLES                   = wx_define_c_proc( "set_status_styles", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_STATUS_TEXT                     = wx_define_c_proc( "set_status_text", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT} ),
	WX_SET_STRING                          = wx_define_c_proc( "set_string", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TAB_SIZE                        = wx_define_c_proc( "set_tab_size", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TARGET_WINDOW                   = wx_define_c_proc( "set_target_window", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_TEXT_ATTR_BACK_COLOR            = wx_define_c_proc( "set_text_attr_back_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_TEXT_ATTR_COLOR                 = wx_define_c_proc( "set_text_attr_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_TEXTATTREX_BULLET_FONT          = wx_define_c_proc( "set_textattrex_bullet_font", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TEXTATTREX_BULLET_NAME          = wx_define_c_proc( "set_textattrex_bullet_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TEXTATTREX_BULLET_NUMBER        = wx_define_c_proc( "set_textattrex_bullet_number", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TEXTATTREX_BULLET_STYLE         = wx_define_c_proc( "set_textattrex_bullet_style", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_TEXTATTREX_BULLET_TEXT          = wx_define_c_proc( "set_textattrex_bullet_text", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TEXTATTREX_CHAR_STYLE_NAME      = wx_define_c_proc( "set_textattrex_char_style_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TEXTATTREX_EFFECT_FLAGS         = wx_define_c_proc( "set_textattrex_effect_flags", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TEXTATTREX_EFFECTS              = wx_define_c_proc( "set_textattrex_effects", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TEXTATTREX_LINE_SPACING         = wx_define_c_proc( "set_textattrex_line_spacing", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TEXTATTREX_LIST_STYLE_NAME      = wx_define_c_proc( "set_textattrex_list_style_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TEXTATTREX_OUTLINE_LEVEL        = wx_define_c_proc( "set_textattrex_outline_level", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TEXTATTREX_PAGE_BREAK           = wx_define_c_proc( "set_textattrex_page_break", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TEXTATTREX_PARA_SPACING_AFTER   = wx_define_c_proc( "set_textattrex_para_spacing_after", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TEXTATTREX_PARA_SPACING_BEFORE  = wx_define_c_proc( "set_textattrex_para_spacing_before", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TEXTATTREX_PARA_STYLE_NAME      = wx_define_c_proc( "set_textattrex_para_style_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TEXTATTREX_URL_NAME             = wx_define_c_proc( "set_textattrex_url_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TEXT_ATTR_FONT                  = wx_define_c_proc( "set_text_attr_font", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_TEXT_BACK_COLOR                 = wx_define_c_proc( "set_text_back_color", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TEXT_COLOR                      = wx_define_c_proc( "set_text_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_TEXT_EDITABLE                   = wx_define_c_proc( "set_text_editable", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TEXT_INSERTION_POINT_END        = wx_define_c_proc( "set_text_insertion_point_end", {dll:C_POINTER} ),
	WX_SET_TEXT_INSERTION_POINT            = wx_define_c_proc( "set_text_insertion_point", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TEXT_MODIFIED                   = wx_define_c_proc( "set_text_modified", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TEXT_SELECTION                  = wx_define_c_proc( "set_text_selection", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_TEXT_STYLE                      = wx_define_c_proc( "set_text_style", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_POINTER} ),
	WX_SET_TEXT_UI                         = wx_define_c_proc( "set_text_ui", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TEXT                            = wx_define_c_proc( "set_text", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TIP                             = wx_define_c_proc( "set_tip", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TITLE                           = wx_define_c_proc( "set_title", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_TOOL_BITMAP_SIZE                = wx_define_c_proc( "set_tool_bitmap_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_TOOL_BITMAP                     = wx_define_c_proc( "set_tool_bitmap", {dll:C_POINTER,dll:C_INT,dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_TOOL_HELP                       = wx_define_c_proc( "set_tool_help", {dll:C_POINTER,dll:C_INT,dll:E_OBJECT,dll:E_OBJECT} ),
	WX_SET_TOOL_LABEL                      = wx_define_c_proc( "set_tool_label", {dll:C_POINTER,dll:C_INT,dll:E_OBJECT} ),
	WX_SET_TOOL_MARGINS                    = wx_define_c_proc( "set_tool_margins", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_TOOLTIP_DELAY                   = wx_define_c_proc( "set_tooltip_delay", {dll:C_POINTER} ),
	WX_SET_TREE_INDENT                     = wx_define_c_proc( "set_tree_indent", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_TREE_ITEM_BACKGROUND            = wx_define_c_proc( "set_tree_item_background", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_TREE_ITEM_BOLD                  = wx_define_c_proc( "set_tree_item_bold", {dll:C_POINTER,dll:C_POINTER,dll:C_INT} ),
	WX_SET_TREE_ITEM_DROP_HIGHLIGHT        = wx_define_c_proc( "set_tree_item_drop_highlight", {dll:C_POINTER,dll:C_POINTER,dll:C_INT} ),
	WX_SET_TREE_ITEM_FONT                  = wx_define_c_proc( "set_tree_item_font", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_TREE_ITEM_HAS_CHILDREN          = wx_define_c_proc( "set_tree_item_has_children", {dll:C_POINTER,dll:C_POINTER,dll:C_INT} ),
	WX_SET_TREE_ITEM_IMAGE                 = wx_define_c_proc( "set_tree_item_image", {dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_TREE_ITEM_TEXT_COLOR            = wx_define_c_proc( "set_tree_item_text_color", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_TREE_ITEM_TEXT                  = wx_define_c_proc( "set_tree_item_text", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_UI_UPDATE_INTERVAL              = wx_define_c_proc( "set_ui_update_interval", {dll:C_INT} ),
	WX_SET_UI_UPDATE_MODE                  = wx_define_c_proc( "set_ui_update_mode", {dll:C_INT} ),
	WX_SET_USER_SCALE                      = wx_define_c_proc( "set_user_scale", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_SET_VALUE                           = wx_define_c_proc( "set_value", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_VIRTUAL_SIZE                    = wx_define_c_proc( "set_virtual_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_WINDOW_STYLE                    = wx_define_c_proc( "set_window_style", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_XML_ENCODING                    = wx_define_c_proc( "set_xml_encoding", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_XML_FILE_ENCODING               = wx_define_c_proc( "set_xml_file_encoding", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_XML_NODE_CHILDREN               = wx_define_c_proc( "set_xml_node_children", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_XML_NODE_CONTENT                = wx_define_c_proc( "set_xml_node_content", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_XML_NODE_NAME                   = wx_define_c_proc( "set_xml_node_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_XML_NODE_PROPERTIES             = wx_define_c_proc( "set_xml_node_properties", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_XML_NODE_TYPE                   = wx_define_c_proc( "set_xml_node_type", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_XML_PROP_NAME                   = wx_define_c_proc( "set_xml_prop_name", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_XML_PROP_VALUE                  = wx_define_c_proc( "set_xml_prop_value", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_XML_ROOT                        = wx_define_c_proc( "set_xml_root", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_XML_VERSION                     = wx_define_c_proc( "set_xml_version", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SHOW_CARET                          = wx_define_c_proc( "show_caret", {dll:C_POINTER} ),
	WX_SHOW_CELL_EDIT_CONTROL              = wx_define_c_proc( "show_cell_edit_control", {dll:C_POINTER} ),
	WX_SHOW_DIRCTRL_HIDDEN_FILES           = wx_define_c_proc( "show_dirctrl_hidden_files", {dll:C_POINTER,dll:C_INT} ),
	WX_SHOW_FULLSCREEN                     = wx_define_c_proc( "show_fullscreen", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_SHOW_MODAL                          = wx_define_c_func( "show_modal", {dll:C_POINTER}, E_OBJECT ),
	WX_SHOW_PAGE                           = wx_define_c_proc( "show_page", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SHOW_PLAYER_CONTROLS                = wx_define_c_func( "show_player_controls", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_SHOW_SIZER_ITEM_SIZER               = wx_define_c_proc( "show_sizer_item_sizer", {dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SHOW_SIZER_ITEM_WIN                 = wx_define_c_proc( "show_sizer_item_win", {dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SHOW_TEXT_POSITION                  = wx_define_c_proc( "show_text_position", {dll:C_POINTER,dll:C_INT} ),
	WX_SHOW_TOOLBAR                        = wx_define_c_proc( "show_toolbar", {dll:C_POINTER} ),
	WX_SHOW_UI                             = wx_define_c_proc( "show_ui", {dll:C_POINTER,dll:C_INT} ),
	WX_SHOW_WINDOW                         = wx_define_c_proc( "show_window", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SIZER_CALC_MIN                      = wx_define_c_func( "sizer_calc_min", {dll:C_POINTER}, E_OBJECT ),
	WX_SIZER_ITEM_SHOWN                    = wx_define_c_func( "sizer_item_shown", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_SKIP                                = wx_define_c_proc( "skip", {dll:C_POINTER} ),
	WX_SOCKET_ACCEPT                       = wx_define_c_func( "socket_accept", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_SOCKET_CONNECT                      = wx_define_c_func( "socket_connect", {dll:C_POINTER,dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_SOCKET_EVENT_HANDLER                = wx_define_c_proc( "socket_event_handler", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_SOCKET_FROM_EVENT                   = wx_define_c_func( "socket_from_event", {dll:C_POINTER}, E_OBJECT ),
	WX_SOCKET_READ                         = wx_define_c_func( "socket_read", {dll:C_POINTER}, E_OBJECT ),
	WX_SOCKET_WRITE                        = wx_define_c_proc( "socket_write", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SORT_TREE_CHILDREN                  = wx_define_c_proc( "sort_tree_children", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SOUND_CREATE                        = wx_define_c_func( "sound_create", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_SOUND_IS_OK                         = wx_define_c_func( "sound_is_ok", {dll:C_POINTER}, E_OBJECT ),
	WX_SOUND_PLAY                          = wx_define_c_func( "sound_play", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_SOUND_STOP                          = wx_define_c_proc( "sound_stop", {} ),
	WX_SPACE_SIZER                         = wx_define_c_proc( "space_sizer", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_SPLIT_EVENT_GET_SASH_POS            = wx_define_c_func( "split_event_get_sash_pos", {dll:C_POINTER}, E_OBJECT ),
	WX_SPLIT_EVENT_SET_SASH_POS            = wx_define_c_proc( "split_event_set_sash_pos", {dll:C_POINTER,dll:C_INT} ),
	WX_SPLIT_EVENT_WIN_REMOVED             = wx_define_c_func( "split_event_win_removed", {dll:C_POINTER}, E_OBJECT ),
	WX_SPLIT_EVENT_X                       = wx_define_c_func( "split_event_x", {dll:C_POINTER}, E_OBJECT ),
	WX_SPLIT_EVENT_Y                       = wx_define_c_func( "split_event_y", {dll:C_POINTER}, E_OBJECT ),
	WX_SPLIT_WINDOW                        = wx_define_c_proc( "split_window", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_START_PRINT                         = wx_define_c_func( "start_print", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_START_PROCESS                       = wx_define_c_func( "start_process", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_START_TIMER                         = wx_define_c_proc( "start_timer", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_STOP_MEDIA                          = wx_define_c_func( "stop_media", {dll:C_POINTER}, E_OBJECT ),
	WX_STOP_TIMER                          = wx_define_c_proc( "stop_timer", {dll:C_POINTER} ),
	WX_STREAM_READ                         = wx_define_c_func( "stream_read", {dll:C_POINTER}, E_OBJECT ),
	WX_SUPPRESSING_UNDO                    = wx_define_c_func( "suppressing_undo", {dll:C_POINTER}, E_OBJECT ),
	WX_TEXT_HIT_TEST                       = wx_define_c_func( "text_hit_test", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_TILE_MDI                            = wx_define_c_proc( "tile_mdi", {dll:C_POINTER,dll:C_INT} ),
	WX_TOGGLE_TOOL                         = wx_define_c_proc( "toggle_tool", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_TOGGLE_TREE_ITEM_SELECTION          = wx_define_c_proc( "toggle_tree_item_selection", {dll:C_POINTER,dll:C_POINTER} ),
	WX_TOGGLE_TREE_ITEM                    = wx_define_c_proc( "toggle_tree_item", {dll:C_POINTER,dll:C_POINTER} ),
	WX_TO_UTF8                             = wx_define_c_func( "to_utf8", {dll:E_OBJECT}, E_OBJECT ),
	WX_TREE_HIT_TEST                       = wx_define_c_func( "tree_hit_test", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_TREE_IMAGELIST                      = wx_define_c_proc( "tree_imagelist", {dll:C_POINTER,dll:C_POINTER} ),
	WX_TREE_ITEM_HAS_CHILDREN              = wx_define_c_func( "tree_item_has_children", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_UNDO_TEXT                           = wx_define_c_proc( "undo_text", {dll:C_POINTER} ),
	WX_UNINIT_AUI                          = wx_define_c_proc( "uninit_aui", {dll:C_POINTER} ),
	WX_UNSELECT_TREE_ITEM                  = wx_define_c_proc( "unselect_tree_item", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_UNSPLIT_WINDOW                      = wx_define_c_proc( "unsplit_window", {dll:C_POINTER,dll:C_POINTER} ),
	WX_UPDATE_AUI                          = wx_define_c_proc( "update_aui", {dll:C_POINTER} ),
	WX_USING_UNIVERSAL_WIDGETS             = wx_define_c_func( "using_universal_widgets", {}, E_OBJECT ),
	WX_VALID_TREE_ITEM                     = wx_define_c_func( "valid_tree_item", {dll:C_POINTER}, E_OBJECT ),
	WX_WAIT_ON_CONNECT                     = wx_define_c_func( "wait_on_connect", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_WIZARD_EVENT_DIRECTION              = wx_define_c_func( "wizard_event_direction", {dll:C_POINTER}, E_OBJECT ),
	WX_WIZARD_EVENT_PAGE                   = wx_define_c_func( "wizard_event_page", {dll:C_POINTER}, E_OBJECT ),
	WX_WORD_LEFT                           = wx_define_c_proc( "word_left", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_WORD_RIGHT                          = wx_define_c_proc( "word_right", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_WRAP                                = wx_define_c_proc( "wrap", {dll:C_POINTER,dll:C_INT} ),
	WX_WRITE_IMAGE_BLOCK                   = wx_define_c_func( "write_image_block", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_WRITE_IMAGE_FROM_FILE               = wx_define_c_func( "write_image_from_file", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_WRITE_IMAGE                         = wx_define_c_func( "write_image", {dll:C_POINTER,dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_WRITE_TEXT                          = wx_define_c_proc( "write_text", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_W_SPRINTF                           = wx_define_c_func( "w_sprintf", {dll:E_OBJECT,dll:E_OBJECT}, E_OBJECT ),
	WX_WX_EXECUTE                          = wx_define_c_func( "wx_execute", {dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_WXMAIN                              = wx_define_c_proc( "wxMain", {dll:C_POINTER} ),
	WX_WX_OR_ALL                           = wx_define_c_func( "wx_or_all", {dll:E_OBJECT}, E_OBJECT ),
	WX_WX_PUTS                             = wx_define_c_proc( "wx_puts", {dll:E_OBJECT,dll:E_OBJECT} ),
	WX_WX_RMDIR                            = wx_define_c_func( "wx_rmdir", {dll:E_OBJECT}, E_OBJECT ),
	WX_WX_SHELL                            = wx_define_c_proc( "wx_shell", {dll:E_OBJECT} ),
	WX_XML_IS_OK                           = wx_define_c_func( "xml_is_ok", {dll:C_POINTER}, E_OBJECT ),
	WX_XML_NODE_HAS_PROP                   = wx_define_c_func( "xml_node_has_prop", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_XRC_GET_CONTROL                     = wx_define_c_func( "xrc_get_control", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_XRC_GET_XRCID                       = wx_define_c_func( "xrc_get_xrcid", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_XRC_LOAD_BITMAP                     = wx_define_c_func( "xrc_load_bitmap", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_XRC_LOAD_DIALOG                     = wx_define_c_func( "xrc_load_dialog", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_XRC_LOAD_FRAME                      = wx_define_c_func( "xrc_load_frame", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_XRC_LOAD_ICON                       = wx_define_c_func( "xrc_load_icon", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_XRC_LOAD_MENU_BAR                   = wx_define_c_func( "xrc_load_menu_bar", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_XRC_LOAD_MENU                       = wx_define_c_func( "xrc_load_menu", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_XRC_LOAD_PANEL                      = wx_define_c_func( "xrc_load_panel", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_XRC_LOAD_TOOL_BAR                   = wx_define_c_func( "xrc_load_tool_bar", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GET_LINK_EVENT_INFO                 = wx_define_c_func( "get_link_event_info", {dll:C_POINTER}, E_OBJECT ),
	WX_NEW_WXTASKBARICON                   = wx_define_c_func( "new_wxTaskBarIcon", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_TASKBAR_MENU                    = wx_define_c_func( "get_taskbar_menu", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_TASKBAR_ICON                    = wx_define_c_proc( "set_taskbar_icon", {dll:C_POINTER,dll:C_POINTER,dll:E_OBJECT} ),
	WX_NEW_WXSINGLECHOICEDIALOG            = wx_define_c_func( "new_wxSingleChoiceDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_XY_TO_POSITION                      = wx_define_c_func( "xy_to_position", {dll:C_POINTER,dll:C_INT,dll:C_INT}, E_OBJECT ),
	WX_STREAM_WRITE                        = wx_define_c_proc( "stream_write", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_GET_SELECTED_COLS                   = wx_define_c_func( "get_selected_cols", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SELECTED_ROWS                   = wx_define_c_func( "get_selected_rows", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SELECTED_CELLS                  = wx_define_c_func( "get_selected_cells", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DISPLAY_SIZE                    = wx_define_c_func( "get_display_size", {}, E_OBJECT ),
	WX_GET_TOOL_POS                        = wx_define_c_func( "get_tool_pos", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_TOOL_SIZE                       = wx_define_c_func( "get_tool_size", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_MENUITEM_BITMAP                 = wx_define_c_proc( "set_menuitem_bitmap", {dll:C_POINTER,dll:C_POINTER} ),
	WX_IS_ICON_INSTALLED                   = wx_define_c_func( "is_icon_installed", {dll:C_POINTER}, E_OBJECT ),
	WX_REMOVE_TASKBAR_ICON                 = wx_define_c_proc( "remove_taskbar_icon", {dll:C_POINTER} ),
	WX_SET_MDI_TOOLBAR                     = wx_define_c_proc( "set_mdi_toolbar", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_MDI_WINDOW_MENU                 = wx_define_c_proc( "set_mdi_window_menu", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_APP_NAME                        = wx_define_c_proc( "set_app_name", {dll:E_OBJECT} ),
	WX_GET_APP_NAME                        = wx_define_c_func( "get_app_name", {}, E_OBJECT ),
	WX_SET_VENDOR_NAME                     = wx_define_c_proc( "set_vendor_name", {dll:E_OBJECT} ),
	WX_GET_VENDOR_NAME                     = wx_define_c_func( "get_vendor_name", {}, E_OBJECT ),
	WX_REQUEST_USER_ATTENTION              = wx_define_c_proc( "request_user_attention", {dll:C_POINTER,dll:C_POINTER} ),
	WX_IS_ACTIVE                           = wx_define_c_func( "is_active", {}, E_OBJECT ),
	WX_FIND_COLOR                          = wx_define_c_func( "find_color", {dll:E_OBJECT}, E_OBJECT ),
	WX_DRAW_ARC                            = wx_define_c_proc( "draw_arc", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_DRAW_CHECK_MARK                     = wx_define_c_proc( "draw_check_mark", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_DRAW_CIRCLE                         = wx_define_c_proc( "draw_circle", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_DRAW_ELLIPSE                        = wx_define_c_proc( "draw_ellipse", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_DRAW_ELLIPTIC_ARC                   = wx_define_c_proc( "draw_elliptic_arc", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:E_OBJECT,dll:E_OBJECT} ),
	WX_DRAW_POINT                          = wx_define_c_proc( "draw_point", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_DRAW_RECTANGLE                      = wx_define_c_proc( "draw_rectangle", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_DRAW_ROUNDED_RECTANGLE              = wx_define_c_proc( "draw_rounded_rectangle", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT,dll:E_OBJECT} ),
	WX_FLOOD_FILL                          = wx_define_c_proc( "flood_fill", {dll:C_POINTER,dll:C_INT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_GRADIENT_FILL                       = wx_define_c_proc( "gradient_fill", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT,dll:C_INT,dll:C_INT} ),
	WX_ENABLE_TOOL                         = wx_define_c_proc( "enable_tool", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SELECTION_TO_TEXT                   = wx_define_c_func( "selection_to_text", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_TYPE                            = wx_define_c_func( "get_type", {dll:E_OBJECT}, E_OBJECT ),
	WX_ADD_GROWABLE_COL                    = wx_define_c_proc( "add_growable_col", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_ADD_GROWABLE_ROW                    = wx_define_c_proc( "add_growable_row", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_REMOVE_ALL_IMAGES                   = wx_define_c_func( "remove_all_images", {dll:C_POINTER}, E_OBJECT ),
	WX_REMOVE_IMAGE                        = wx_define_c_func( "remove_image", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_REPLACE_IMAGE                       = wx_define_c_func( "replace_image", {dll:C_POINTER,dll:C_INT,dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_NEW_WXFILEDIALOG                    = wx_define_c_func( "new_wxFileDialog", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_FD_FILENAME                     = wx_define_c_func( "get_fd_filename", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FD_FILENAMES                    = wx_define_c_func( "get_fd_filenames", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FD_FILTER_INDEX                 = wx_define_c_func( "get_fd_filter_index", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FD_MESSAGE                      = wx_define_c_func( "get_fd_message", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FD_PATH                         = wx_define_c_func( "get_fd_path", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FD_PATHS                        = wx_define_c_func( "get_fd_paths", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FD_WILDCARD                     = wx_define_c_func( "get_fd_wildcard", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_FD_DIRECTORY                    = wx_define_c_proc( "set_fd_directory", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_FD_FILENAME                     = wx_define_c_proc( "set_fd_filename", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_FD_FILTERINDEX                  = wx_define_c_proc( "set_fd_filterindex", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_FD_MESSAGE                      = wx_define_c_proc( "set_fd_message", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_FD_PATH                         = wx_define_c_proc( "set_fd_path", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_FD_WILDCARD                     = wx_define_c_proc( "set_fd_wildcard", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_GET_SPLASH_WINDOW                   = wx_define_c_func( "get_splash_window", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DC_GDK_WINDOW                   = wx_define_c_func( "get_dc_gdk_window", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_PICKER_INTERNAL_MARGIN          = wx_define_c_proc( "set_picker_internal_margin", {dll:C_POINTER,dll:C_INT} ),
	WX_GET_PICKER_INTERNAL_MARGIN          = wx_define_c_func( "get_picker_internal_margin", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_PICKER_TEXTCTRL_PROPORTION      = wx_define_c_proc( "set_picker_textctrl_proportion", {dll:C_POINTER,dll:C_INT} ),
	WX_GET_PICKER_TEXTCTRL_PROPORTION      = wx_define_c_func( "get_picker_textctrl_proportion", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_PICKERCTRL_PROPORTION           = wx_define_c_proc( "set_pickerctrl_proportion", {dll:C_POINTER,dll:C_INT} ),
	WX_GET_PICKERCTRL_PROPORTION           = wx_define_c_func( "get_pickerctrl_proportion", {dll:C_POINTER}, E_OBJECT ),
	WX_PICKER_HAS_TEXTCTRL                 = wx_define_c_func( "picker_has_textctrl", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PICKER_TEXTCTRL                 = wx_define_c_func( "get_picker_textctrl", {dll:C_POINTER}, E_OBJECT ),
	WX_PICKER_IS_TEXTCTRL_GROWABLE         = wx_define_c_func( "picker_is_textctrl_growable", {dll:C_POINTER}, E_OBJECT ),
	WX_PICKER_IS_PICKERCTRL_GROWABLE       = wx_define_c_func( "picker_is_pickerctrl_growable", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_PICKERCTRL_GROWABLE             = wx_define_c_proc( "set_pickerctrl_growable", {dll:C_POINTER,dll:C_INT} ),
	WX_SET_PICKER_TEXTCTRL_GROWABLE        = wx_define_c_proc( "set_picker_textctrl_growable", {dll:C_POINTER,dll:C_INT} ),
	WX_NEW_WXDIRPICKERCTRL                 = wx_define_c_func( "new_wxDirPickerCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_DIRPICKER_PATH                  = wx_define_c_func( "get_dirpicker_path", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_DIRPICKER_PATH                  = wx_define_c_proc( "set_dirpicker_path", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_NEW_WXCOLOURPICKERCTRL              = wx_define_c_func( "new_wxColourPickerCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_PICKER_COLOR                    = wx_define_c_func( "get_picker_color", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_PICKER_COLOR                    = wx_define_c_proc( "set_picker_color", {dll:C_POINTER,dll:C_POINTER} ),
	WX_NEW_WXDATEPICKERCTRL                = wx_define_c_func( "new_wxDatePickerCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_PICKER_RANGE                    = wx_define_c_func( "get_picker_range", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_PICKER_DATE                     = wx_define_c_func( "get_picker_date", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_PICKER_RANGE                    = wx_define_c_proc( "set_picker_range", {dll:C_POINTER,dll:E_OBJECT,dll:E_OBJECT} ),
	WX_SET_PICKER_DATE                     = wx_define_c_proc( "set_picker_date", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_NEW_WXFILEPICKERCTRL                = wx_define_c_func( "new_wxFilePickerCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_FILEPICKER_PATH                 = wx_define_c_func( "get_filepicker_path", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_FILEPICKER_PATH                 = wx_define_c_proc( "set_filepicker_path", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_NEW_WXFONTPICKERCTRL                = wx_define_c_func( "new_wxFontPickerCtrl", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_PICKER_FONT                     = wx_define_c_func( "get_picker_font", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_PICKER_FONT                     = wx_define_c_proc( "set_picker_font", {dll:C_POINTER,dll:C_POINTER} ),
	WX_GET_PICKER_MAX_POINT                = wx_define_c_func( "get_picker_max_point", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_PICKER_MAX_POINT                = wx_define_c_proc( "set_picker_max_point", {dll:C_POINTER,dll:C_INT} ),
	WX_GET_COLORPICKEREVENT_COLOR          = wx_define_c_func( "get_colorpickerevent_color", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_DATEEVENT_DATE                  = wx_define_c_func( "get_dateevent_date", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FILEDIRPICKER_EVENT_PATH        = wx_define_c_func( "get_filedirpicker_event_path", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_FONTPICKER_EVENT_FONT           = wx_define_c_func( "get_fontpicker_event_font", {dll:C_POINTER}, E_OBJECT ),
	WX_NEW_WXCOLLAPSIBLEPANE               = wx_define_c_func( "new_wxCollapsiblePane", {dll:E_OBJECT}, E_OBJECT ),
	WX_IS_COLLPANE_COLLAPSED               = wx_define_c_func( "is_collpane_collapsed", {dll:C_POINTER}, E_OBJECT ),
	WX_IS_COLLPANE_EXPANDED                = wx_define_c_func( "is_collpane_expanded", {dll:C_POINTER}, E_OBJECT ),
	WX_COLLAPSE_COLLPANE                   = wx_define_c_proc( "collapse_collpane", {dll:C_POINTER,dll:C_INT} ),
	WX_EXPAND_COLLPANE                     = wx_define_c_proc( "expand_collpane", {dll:C_POINTER} ),
	WX_GET_COLLPANE_WIN                    = wx_define_c_func( "get_collpane_win", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_COLLPAN_EVENT_COLLAPSED         = wx_define_c_func( "get_collpan_event_collapsed", {dll:C_POINTER}, E_OBJECT ),
	WX_USE_PRIMARY_SELECTION               = wx_define_c_proc( "use_primary_selection", {dll:C_INT} ),
	WX_NEW_WXLOCALE                        = wx_define_c_func( "new_wxLocale", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_LANGUAGE                        = wx_define_c_func( "get_language", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_CANONICAL_LANGUAGE_NAME         = wx_define_c_func( "get_canonical_language_name", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_LANGUAGE_NAME                   = wx_define_c_func( "get_language_name", {dll:C_INT}, E_OBJECT ),
	WX_GET_LOCALE                          = wx_define_c_func( "get_locale", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_LOCALE_NAME                     = wx_define_c_func( "get_locale_name", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_LOCALE_SYS_NAME                 = wx_define_c_func( "get_locale_sys_name", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_SYSTEM_ENCODING                 = wx_define_c_func( "get_system_encoding", {}, E_OBJECT ),
	WX_GET_SYSTEM_ENCODING_NAME            = wx_define_c_func( "get_system_encoding_name", {}, E_OBJECT ),
	WX_GET_SYSTEM_LANGUAGE                 = wx_define_c_func( "get_system_language", {}, E_OBJECT ),
	WX_NEW_WXCSCONV                        = wx_define_c_func( "new_wxCSConv", {dll:E_OBJECT}, E_OBJECT ),
	WX_CONV_OK                             = wx_define_c_func( "conv_ok", {dll:C_POINTER}, E_OBJECT ),
	WX_CONV_MB2WC                          = wx_define_c_func( "conv_mb2wc", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_CONV_WC2MB                          = wx_define_c_func( "conv_wc2mb", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_IS_LOCALE_AVAILABLE                 = wx_define_c_func( "is_locale_available", {dll:C_INT}, E_OBJECT ),
	WX_GET_DEFAULT_FONT_ENCODING           = wx_define_c_func( "get_default_font_encoding", {}, E_OBJECT ),
	WX_NEW_WXAUIDOCKART                    = wx_define_c_func( "new_wxAuiDockArt", {dll:E_OBJECT}, E_OBJECT ),
	WX_DOCKART_DRAWBACKGROUND_ID           = wx_define_c_proc( "dockart_drawbackground_id", {dll:C_POINTER,dll:C_INT} ),
	WX_DOCKART_DRAWBORDER_ID               = wx_define_c_proc( "dockart_drawborder_id", {dll:C_POINTER,dll:C_INT} ),
	WX_DOCKART_DRAWCAPTION_ID              = wx_define_c_proc( "dockart_drawcaption_id", {dll:C_POINTER,dll:C_INT} ),
	WX_DOCKART_DRAWGRIPPER_ID              = wx_define_c_proc( "dockart_drawgripper_id", {dll:C_POINTER,dll:C_INT} ),
	WX_DOCKART_DRAWPANEBUTTON_ID           = wx_define_c_proc( "dockart_drawpanebutton_id", {dll:C_POINTER,dll:C_INT} ),
	WX_DOCKART_DRAWSASH_ID                 = wx_define_c_proc( "dockart_drawsash_id", {dll:C_POINTER,dll:C_INT} ),
	WX_NEW_WXRICHTEXTPRINTING              = wx_define_c_func( "new_wxRichTextPrinting", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_DOCKART_COLOUR                  = wx_define_c_func( "get_dockart_colour", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_DOCKART_FONT                    = wx_define_c_func( "get_dockart_font", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_GET_DOCKART_METRIC                  = wx_define_c_func( "get_dockart_metric", {dll:C_POINTER,dll:C_INT}, E_OBJECT ),
	WX_SET_DOCKART_COLOUR                  = wx_define_c_proc( "set_dockart_colour", {dll:C_POINTER,dll:C_INT,dll:C_POINTER} ),
	WX_SET_DOCKART_FONT                    = wx_define_c_proc( "set_dockart_font", {dll:C_POINTER,dll:C_INT,dll:C_POINTER} ),
	WX_SET_DOCKART_METRIC                  = wx_define_c_proc( "set_dockart_metric", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_GET_BUFFER                          = wx_define_c_func( "get_buffer", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICHPRINT_FOOTER_TEXT           = wx_define_c_func( "get_richprint_footer_text", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICHPRINT_HEADER_FOOTER_DATA    = wx_define_c_func( "get_richprint_header_footer_data", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICHPRINT_HEADER_TEXT           = wx_define_c_func( "get_richprint_header_text", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICHPRINT_PAGE_SETUP_DATA       = wx_define_c_func( "get_richprint_page_setup_data", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICHPRINT_PARENT_WINDOW         = wx_define_c_func( "get_richprint_parent_window", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICHPRINT_PREVIEW_RECT          = wx_define_c_func( "get_richprint_preview_rect", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICHPRINT_PRINT_DATA            = wx_define_c_func( "get_richprint_print_data", {dll:C_POINTER}, E_OBJECT ),
	WX_GET_RICHPRINT_TITLE                 = wx_define_c_func( "get_richprint_title", {dll:C_POINTER}, E_OBJECT ),
	WX_SHOW_RICHPRINT_PAGE_SETUP           = wx_define_c_proc( "show_richprint_page_setup", {dll:C_POINTER} ),
	WX_SHOW_RICHPRINT_PREVIEW_BUFFER       = wx_define_c_proc( "show_richprint_preview_buffer", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SHOW_RICHPRINT_PREVIEW_FILE         = wx_define_c_proc( "show_richprint_preview_file", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SHOW_RICHPRINT_PRINT_BUFFER         = wx_define_c_proc( "show_richprint_print_buffer", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SHOW_RICHPRINT_PRINT_FILE           = wx_define_c_proc( "show_richprint_print_file", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_RICHPRINT_FOOTER_TEXT           = wx_define_c_proc( "set_richprint_footer_text", {dll:C_POINTER,dll:E_OBJECT,dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICHPRINT_HEADER_FOOTER_DATA    = wx_define_c_proc( "set_richprint_header_footer_data", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICHPRINT_HEADER_FOOTER_FONT    = wx_define_c_proc( "set_richprint_header_footer_font", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICHPRINT_HEADER_FOOTER_TEXT_COLOUR= wx_define_c_proc( "set_richprint_header_footer_text_colour", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICHPRINT_HEADER_TEXT           = wx_define_c_proc( "set_richprint_header_text", {dll:C_POINTER,dll:E_OBJECT,dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICHPRINT_PAGE_SETUP_DATA       = wx_define_c_proc( "set_richprint_page_setup_data", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICHPRINT_PARENT_WINDOW         = wx_define_c_proc( "set_richprint_parent_window", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICHPRINT_PREVIEW_RECT          = wx_define_c_proc( "set_richprint_preview_rect", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_SET_RICHPRINT_PRINT_DATA            = wx_define_c_proc( "set_richprint_print_data", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SHOW_RICHPRINT_ON_FIRST_PAGE        = wx_define_c_proc( "show_richprint_on_first_page", {dll:C_POINTER,dll:C_POINTER} ),
	WX_SET_RICHPRINT_TITLE                 = wx_define_c_proc( "set_richprint_title", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_GET_UNICODE_KEY                     = wx_define_c_func( "get_unicode_key", {dll:C_POINTER}, E_OBJECT ),
	WX_NEW_WXTOOLBARITEM                   = wx_define_c_func( "new_wxToolBarItem", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXGRAPHICSMATRIX                = wx_define_c_func( "new_wxGraphicsMatrix", {dll:E_OBJECT}, E_OBJECT ),
	WX_GRAPHICS_MATRIX_CONCAT              = wx_define_c_proc( "graphics_matrix_concat", {dll:C_POINTER,dll:C_POINTER} ),
	WX_GRAPHICS_MATRIX_GET                 = wx_define_c_func( "graphics_matrix_get", {dll:C_POINTER}, E_OBJECT ),
	WX_GRAPHICS_MATRIX_GET_NATIVE          = wx_define_c_func( "graphics_matrix_get_native", {dll:C_POINTER}, E_OBJECT ),
	WX_GRAPHICS_MATRIX_INVERT              = wx_define_c_proc( "graphics_matrix_invert", {dll:C_POINTER} ),
	WX_GRAPHICS_MATRIX_EQUAL               = wx_define_c_func( "graphics_matrix_equal", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GRAPHICS_MATRIX_IDENTITY            = wx_define_c_func( "graphics_matrix_identity", {dll:C_POINTER}, E_OBJECT ),
	WX_GRAPHICS_MATRIX_ROTATE              = wx_define_c_proc( "graphics_matrix_rotate", {dll:C_POINTER,dll:C_DOUBLE} ),
	WX_GRAPHICS_MATRIX_SCALE               = wx_define_c_proc( "graphics_matrix_scale", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GRAPHICS_MATRIX_TRANSLATE           = wx_define_c_proc( "graphics_matrix_translate", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GRAPHICS_MATRIX_SET                 = wx_define_c_proc( "graphics_matrix_set", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GRAPHICS_MATRIX_TRANSFORM_POINT     = wx_define_c_func( "graphics_matrix_transform_point", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE}, E_OBJECT ),
	WX_GRAPHICS_MATRIX_TRANSFORM_DISTANCE  = wx_define_c_func( "graphics_matrix_transform_distance", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE}, E_OBJECT ),
	WX_GRAPHICS_MATRIX_RENDERER            = wx_define_c_func( "graphics_matrix_renderer", {dll:C_POINTER}, E_OBJECT ),
	WX_GRAPHICS_MATRIX_ISNULL              = wx_define_c_func( "graphics_matrix_isnull", {dll:C_POINTER}, E_OBJECT ),
	WX_FIND_MENUITEM_BY_ID                 = wx_define_c_func( "find_menuitem_by_id", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GET_GRAPHICS_OBJECT_RENDERER        = wx_define_c_func( "get_graphics_object_renderer", {dll:C_POINTER}, E_OBJECT ),
	WX_GRAPHICS_OBJECT_ISNULL              = wx_define_c_func( "graphics_object_isnull", {dll:C_POINTER}, E_OBJECT ),
	WX_NEW_WXGRAPHICSCONTEXT               = wx_define_c_func( "new_wxGraphicsContext", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXGRAPHICSPEN                   = wx_define_c_func( "new_wxGraphicsPen", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXGRAPHICSBRUSH                 = wx_define_c_func( "new_wxGraphicsBrush", {dll:E_OBJECT}, E_OBJECT ),
	WX_GET_DEFAULT_GRAPHICS_RENDERER       = wx_define_c_func( "get_default_graphics_renderer", {}, E_OBJECT ),
	WX_CREATE_LINEAR_GRADIENT_BRUSH        = wx_define_c_func( "create_linear_gradient_brush", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_CREATE_RADIAL_GRADIENT_BRUSH        = wx_define_c_func( "create_radial_gradient_brush", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_CREATE_GRAPHICS_FONT                = wx_define_c_func( "create_graphics_font", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_NEW_WXGRAPHICSPATH                  = wx_define_c_func( "new_wxGraphicsPath", {dll:E_OBJECT}, E_OBJECT ),
	WX_PATH_MOVE_TO                        = wx_define_c_proc( "path_move_to", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_PATH_ADD_ARC                        = wx_define_c_proc( "path_add_arc", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_INT} ),
	WX_PATH_ADD_ARC_TO_POINT               = wx_define_c_proc( "path_add_arc_to_point", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_PATH_ADD_CIRCLE                     = wx_define_c_proc( "path_add_circle", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_PATH_ADD_CURVE_TO_POINT             = wx_define_c_proc( "path_add_curve_to_point", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_PATH_ADD_ELLIPSE                    = wx_define_c_proc( "path_add_ellipse", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_PATH_ADD_LINE_TO_POINT              = wx_define_c_proc( "path_add_line_to_point", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_PATH_ADD                            = wx_define_c_proc( "path_add", {dll:C_POINTER,dll:C_POINTER} ),
	WX_PATH_ADD_QUAD_CURVE_TO_POINT        = wx_define_c_proc( "path_add_quad_curve_to_point", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_PATH_ADD_RECTANGLE                  = wx_define_c_proc( "path_add_rectangle", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_PATH_ADD_ROUNDED_RECTANGLE          = wx_define_c_proc( "path_add_rounded_rectangle", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_PATH_CLOSE_SUBPATH                  = wx_define_c_proc( "path_close_subpath", {dll:C_POINTER} ),
	WX_PATH_CONTAINS                       = wx_define_c_func( "path_contains", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_INT}, E_OBJECT ),
	WX_PATH_GET_BOX                        = wx_define_c_func( "path_get_box", {dll:C_POINTER}, E_OBJECT ),
	WX_PATH_GET_CURRENT_POINT              = wx_define_c_func( "path_get_current_point", {dll:C_POINTER}, E_OBJECT ),
	WX_PATH_TRANSFORM                      = wx_define_c_func( "path_transform", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_PATH_GET_NATIVE                     = wx_define_c_func( "path_get_native", {dll:C_POINTER}, E_OBJECT ),
	WX_PATH_UNGET_NATIVE_PATH              = wx_define_c_func( "path_unget_native_path", {dll:C_POINTER,dll:C_POINTER}, E_OBJECT ),
	WX_GC_CLIP                             = wx_define_c_proc( "gc_clip", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GC_RESET_CLIP                       = wx_define_c_proc( "gc_reset_clip", {dll:C_POINTER} ),
	WX_GC_DRAW_BITMAP                      = wx_define_c_proc( "gc_draw_bitmap", {dll:C_POINTER,dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GC_DRAW_ELLIPSE                     = wx_define_c_proc( "gc_draw_ellipse", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GC_DRAW_ICON                        = wx_define_c_proc( "gc_draw_icon", {dll:C_POINTER,dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GC_DRAW_PATH                        = wx_define_c_proc( "gc_draw_path", {dll:C_POINTER,dll:C_POINTER,dll:C_INT} ),
	WX_GC_DRAW_RECTANGLE                   = wx_define_c_proc( "gc_draw_rectangle", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GC_DRAW_ROUNDED_RECTANGLE           = wx_define_c_proc( "gc_draw_rounded_rectangle", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GC_DRAW_TEXT                        = wx_define_c_proc( "gc_draw_text", {dll:C_POINTER,dll:E_OBJECT,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GC_FILL_PATH                        = wx_define_c_proc( "gc_fill_path", {dll:C_POINTER,dll:C_POINTER,dll:C_INT} ),
	WX_GC_STROKE_PATH                      = wx_define_c_proc( "gc_stroke_path", {dll:C_POINTER,dll:C_POINTER} ),
	WX_GC_GET_NATIVE_CONTEXT               = wx_define_c_func( "gc_get_native_context", {dll:C_POINTER}, E_OBJECT ),
	WX_GC_GET_PARTIAL_TEXT_EXTENTS         = wx_define_c_func( "gc_get_partial_text_extents", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GC_GET_TEXT_EXTENT                  = wx_define_c_func( "gc_get_text_extent", {dll:C_POINTER,dll:E_OBJECT}, E_OBJECT ),
	WX_GC_ROTATE                           = wx_define_c_proc( "gc_rotate", {dll:C_POINTER,dll:C_DOUBLE} ),
	WX_GC_SCALE                            = wx_define_c_proc( "gc_scale", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GC_TRANSLATE                        = wx_define_c_proc( "gc_translate", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GC_GET_TRANSFORM                    = wx_define_c_func( "gc_get_transform", {dll:C_POINTER}, E_OBJECT ),
	WX_GC_SET_TRANSFORM                    = wx_define_c_proc( "gc_set_transform", {dll:C_POINTER,dll:C_POINTER} ),
	WX_GC_CONCAT_TRANSFORM                 = wx_define_c_proc( "gc_concat_transform", {dll:C_POINTER,dll:C_POINTER} ),
	WX_GC_SET_BRUSH                        = wx_define_c_proc( "gc_set_brush", {dll:C_POINTER,dll:C_POINTER} ),
	WX_GC_SET_GRAPHICS_BRUSH               = wx_define_c_proc( "gc_set_graphics_brush", {dll:C_POINTER,dll:C_POINTER} ),
	WX_GC_SET_FONT                         = wx_define_c_proc( "gc_set_font", {dll:C_POINTER,dll:C_POINTER,dll:C_POINTER} ),
	WX_GC_SET_GRAPHICS_FONT                = wx_define_c_proc( "gc_set_graphics_font", {dll:C_POINTER,dll:C_POINTER} ),
	WX_GC_SET_PEN                          = wx_define_c_proc( "gc_set_pen", {dll:C_POINTER,dll:C_POINTER} ),
	WX_GC_SET_GRAPHICS_PEN                 = wx_define_c_proc( "gc_set_graphics_pen", {dll:C_POINTER,dll:C_POINTER} ),
	WX_GC_STROKE_LINE                      = wx_define_c_proc( "gc_stroke_line", {dll:C_POINTER,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE,dll:C_DOUBLE} ),
	WX_GC_STROKE_LINES                     = wx_define_c_proc( "gc_stroke_lines", {dll:C_POINTER,dll:E_OBJECT} ),
	WX_NEW_WXARRAYSTRINGPROPERTY           = wx_define_c_func( "new_wxArrayStringProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXBOOLPROPERTY                  = wx_define_c_func( "new_wxBoolProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCOLOURPROPERTY                = wx_define_c_func( "new_wxColourProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXCURSORPROPERTY                = wx_define_c_func( "new_wxCursorProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXDATEPROPERTY                  = wx_define_c_func( "new_wxDateProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXDIRPROPERTY                   = wx_define_c_func( "new_wxDirProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXEDITENUMPROPERTY              = wx_define_c_func( "new_wxEditEnumProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXENUMPROPERTY                  = wx_define_c_func( "new_wxEnumProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXFILEPROPERTY                  = wx_define_c_func( "new_wxFileProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXFLOATPROPERTY                 = wx_define_c_func( "new_wxFloatProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXFONTPROPERTY                  = wx_define_c_func( "new_wxFontProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXIMAGEFILEPROPERTY             = wx_define_c_func( "new_wxImageFileProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXINTPROPERTY                   = wx_define_c_func( "new_wxIntProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXLONGSTRINGPROPERTY            = wx_define_c_func( "new_wxLongStringProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXMULTICHOICEPROPERTY           = wx_define_c_func( "new_wxMultiChoiceProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPROPERTYCATEGORY              = wx_define_c_func( "new_wxPropertyCategory", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSTRINGPROPERTY                = wx_define_c_func( "new_wxStringProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXSYSTEMCOLOURPROPERTY          = wx_define_c_func( "new_wxSystemColourProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXUINTPROPERTY                  = wx_define_c_func( "new_wxUIntProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_NEW_WXPROPERTYGRID                  = wx_define_c_func( "new_wxPropertyGrid", {dll:E_OBJECT}, E_OBJECT ),
	WX_PROPGRID_APPEND                     = wx_define_c_proc( "propgrid_append", {dll:C_POINTER,dll:C_POINTER} ),
	WX_NEW_WXFLAGSPROPERTY                 = wx_define_c_func( "new_wxFlagsProperty", {dll:E_OBJECT}, E_OBJECT ),
	WX_PGPROP_ADD_CHOICE                   = wx_define_c_func( "pgprop_add_choice", {dll:C_POINTER,dll:E_OBJECT,dll:C_INT}, E_OBJECT ),
	WX_PGPROP_GET_VALUE                    = wx_define_c_func( "pgprop_get_value", {dll:C_POINTER}, E_OBJECT ),
	WX_SET_MIN_SIZE                        = wx_define_c_proc( "set_min_size", {dll:C_POINTER,dll:C_INT,dll:C_INT} ),
	WX_SET_TREE_COMPARE_RID                = wx_define_c_proc( "set_tree_compare_rid", {dll:C_POINTER,dll:C_INT} ),
    --@define
-- 1 End of imports
------------------------------

-- define c vars
----------------------------
    wxNullPen                              = wx_define_c_var( "null_pen" ),
    wxNullBrush                            = wx_define_c_var( "null_brush" )

-- Imports end
---------------------------

public constant
    wxFalse = 0,
    wxTrue = not wxFalse

-- Create definitions
----------------------------
public constant
    wxButton                       = 1,
    wxDefButton                    = 2,
    wxBitmapButton                 = 3,
    wxCalendarCtrl                 = 4,
    wxPen                          = 5,
    wxCheckBox                     = 6,
    wxRadioButton                  = 7,
    wxRadioBox                     = 8,
    wxToggleButton                 = 9,
    wxFTP                          = 10,
    wxDialog                       = 11,
    wxProgressDialog               = 12,
    wxFontData                     = 13,
    wxDirDialog                    = 14,
    wxPageSetupDialog              = 15,
    wxPrintDialog                  = 16,
    wxColourDialog                 = 17,
    wxDir                          = 18,
    wxDirTraverser                 = 19,
    wxString                       = 20,
    wxArrayString                  = 21,
    wxPoint                        = 22,
    wxFrame                        = 23,
    wxWindow                       = 24,
    wxScrolledWindow               = 25,
    wxPanel                        = 26,
    wxCursor                       = 27,
    wxFileName                     = 28,
    wxFont                         = 29,
    wxColour                       = 30,
    wxBitmap                       = 31,
    wxIcon                         = 32,
    wxMemoryDC                     = 33,
    wxImageList                    = 34,
    wxBrush                        = 35,
    wxGrid                         = 36,
    wxGridCellBoolEditor           = 37,
    wxGridCellAttr                 = 38,
    wxHtmlWindow                   = 39,
    wxListBox                      = 40,
    wxComboBox                     = 41,
    wxChoice                       = 42,
    wxListCtrl                     = 43,
    wxMDIParentFrame               = 44,
    wxMDIClientWindow              = 45,
    wxMDIChildFrame                = 46,
    wxMenuBar                      = 47,
    wxMenuItem                     = 48,
    wxMenu                         = 49,
    wxSocketClient                 = 50,
    wxSocketServer                 = 51,
    wxURL                          = 52,
    wxHTTP                         = 53,
    wxNotebook                     = 54,
    wxPreviewFrame                 = 55,
    wxPrintout                     = 56,
    wxPrintPreview                 = 57,
    wxScrollBar                    = 58,
    wxSlider                       = 59,
    UNDEFINED_wxNotebookSizer      = 60,
    wxSplitterWindow               = 61,
    wxFlexGridSizer                = 62,
    wxSplashScreen                 = 63,
    wxStaticLine                   = 64,
    wxStaticBox                    = 65,
    wxStaticBitmap                 = 66,
    wxStatusBar                    = 67,
    wxTextCtrl                     = 68,
    wxStaticText                   = 69,
    wxSpinCtrl                     = 70,
    wxTimer                        = 71,
    wxToolBar                      = 72,
    wxTreeCtrl                     = 73,
    wxWizard                       = 74,
    wxWizardPage                   = 75,
    wxPaintDC                      = 76,
    wxBoxSizer                     = 77,
    wxClientDC                     = 78,
    wxIPV4address                  = 79,
    wxMask                         = 80,
    wxGridCellChoiceEditor         = 81,
    wxRegEx                        = 82,
    wxTextAttr                     = 83,
    wxMediaCtrl                    = 84,
    wxAuiManager                   = 85,
    wxAuiPaneInfo                  = 86,
    wxAuiNotebook                  = 87,
    wxXmlDocument                  = 88,
    wxXmlNode                      = 89,
    wxXmlProperty                  = 90,
    wxRichTextCtrl                 = 91,
    wxTextAttrEx                   = 92,
    wxRichTextAttr                 = 93,
    wxPrinter                      = 94,
    wxHtmlEasyPrinting             = 95,
    wxChoicebook                   = 96,
    wxListbook                     = 97,
    wxTreebook                     = 98,
    wxToolbook                     = 99,
    wxRichTextFormattingDialog     = 100,
    wxGridCellAutoWrapStringEditor = 101,
    wxStaticBoxSizer               = 102,
    wxImage                        = 103,
    wxAboutDialogInfo              = 104,
    wxCaret                        = 105,
    wxXmlResource                  = 106,
    wxDisplay                      = 107,
    wxMultiChoiceDialog            = 108,
    wxPasswordEntryDialog          = 109,
    wxTextEntryDialog              = 110,
    wxSound                        = 111,
    wxBitmapComboBox               = 112,
    wxDirCtrl                      = 113,
    wxProcess                      = 114,
    wxStdDialogButtonSizer         = 115,
    wxGauge                        = 116,
    wxPopupWindow                  = 117,
    wxPopupTransientWindow         = 118,
    wxConfig                       = 119,
    wxTaskBarIcon                  = 120,
    wxSingleChoiceDialog           = 121,
        wxFileDialog                   = 122,
        wxDirPickerCtrl                = 123,
        wxColourPickerCtrl             = 124,
        wxDatePickerCtrl               = 125,
        wxFilePickerCtrl               = 126,
        wxFontPickerCtrl               = 127,
        wxCollapsiblePane              = 128,
        wxLocale                       = 129,
        wxCSConv                       = 130,
        wxAuiDockArt                   = 131,
        wxRichTextPrinting             = 132,
	wxToolBarItem                  = 133,
	wxGraphicsMatrix               = 134,
	wxGraphicsContext              = 135,
	wxGraphicsPen                  = 136,
	wxGraphicsBrush                = 137,
	wxGraphicsPath                 = 138,
	wxArrayStringProperty          = 139,
	wxBoolProperty                 = 140,
	wxColourProperty               = 141,
	wxCursorProperty               = 142,
	wxDateProperty                 = 143,
	wxDirProperty                  = 144,
	wxEditEnumProperty             = 145,
	wxEnumProperty                 = 146,
	wxFileProperty                 = 147,
	wxFloatProperty                = 148,
	wxFontProperty                 = 149,
	wxImageFileProperty            = 150,
	wxIntProperty                  = 151,
	wxLongStringProperty           = 152,
	wxMultiChoiceProperty          = 153,
	wxPropertyCategory             = 154,
	wxStringProperty               = 155,
	wxSystemColourProperty         = 156,
	wxUIntProperty                 = 157,
	wxPropertyGrid                 = 158,
	wxFlagsProperty                = 159,
--@object
    WX_NEW_WXGRIDCELLCHOICEEDITOR  = -1    -- not implemented

sequence WX_CREATE_ARRAY

    WX_CREATE_ARRAY = {
        WX_NEW_WXBUTTON
        ,WX_NEW_WXDEFBUTTON
        ,WX_NEW_WXBITMAPBUTTON
        ,WX_NEW_WXCALENDARCTRL
        ,WX_NEW_WXPEN
        ,WX_NEW_WXCHECKBOX
        ,WX_NEW_WXRADIOBUTTON
        ,WX_NEW_WXRADIOBOX
        ,WX_NEW_WXTOGGLEBUTTON
        ,WX_NEW_WXFTP
        ,WX_NEW_WXDIALOG
        ,WX_NEW_WXPROGRESSDIALOG
        ,WX_NEW_WXFONTDATA
        ,WX_NEW_WXDIRDIALOG
        ,WX_NEW_WXPAGESETUPDIALOG
        ,WX_NEW_WXPRINTDIALOG
        ,WX_NEW_WXCOLOURDIALOG
        ,WX_NEW_WXDIR
        ,WX_NEW_WXDIRTRAVERSER
        ,-1 --WX_NEW_WXSTRING
        ,-1 --WX_NEW_WXARRAYSTRING
        ,-1 --WX_NEW_WXPOINT
        ,WX_NEW_WXFRAME
        ,WX_NEW_WXWINDOW
        ,WX_NEW_WXSCROLLEDWINDOW
        ,WX_NEW_WXPANEL
        ,WX_NEW_WXCURSOR
        ,WX_NEW_WXFILENAME
        ,WX_NEW_WXFONT
        ,WX_NEW_WXCOLOUR
        ,WX_NEW_WXBITMAP
        ,WX_NEW_WXICON
        ,WX_NEW_WXMEMORYDC
        ,WX_NEW_WXIMAGELIST
        ,WX_NEW_WXBRUSH
        ,WX_NEW_WXGRID
        ,WX_NEW_WXGRIDCELLBOOLEDITOR
        ,WX_NEW_WXGRIDCELLATTR
        ,WX_NEW_WXHTMLWINDOW
        ,WX_NEW_WXLISTBOX
        ,WX_NEW_WXCOMBOBOX
        ,WX_NEW_WXCHOICE
        ,WX_NEW_WXLISTCTRL
        ,WX_NEW_WXMDIPARENTFRAME
        ,WX_NEW_WXMDICLIENTWINDOW
        ,WX_NEW_WXMDICHILDFRAME
        ,WX_NEW_WXMENUBAR
        ,WX_NEW_WXMENUITEM
        ,WX_NEW_WXMENU
        ,WX_NEW_WXSOCKETCLIENT
        ,WX_NEW_WXSOCKETSERVER
        ,WX_NEW_WXURL
        ,WX_NEW_WXHTTP
        ,WX_NEW_WXNOTEBOOK
        ,WX_NEW_WXPREVIEWFRAME
        ,WX_NEW_WXPRINTOUT
        ,WX_NEW_WXPRINTPREVIEW
        ,WX_NEW_WXSCROLLBAR
        ,WX_NEW_WXSLIDER
        ,-1 --WX_NEW_WXNOTEBOOKSIZER
        ,WX_NEW_WXSPLITTERWINDOW
        ,WX_NEW_WXFLEXGRIDSIZER
        ,WX_NEW_WXSPLASHSCREEN
        ,WX_NEW_WXSTATICLINE
        ,WX_NEW_WXSTATICBOX
        ,WX_NEW_WXSTATICBITMAP
        ,WX_NEW_WXSTATUSBAR
        ,WX_NEW_WXTEXTCTRL
        ,WX_NEW_WXSTATICTEXT
        ,WX_NEW_WXSPINCTRL
        ,WX_NEW_WXTIMER
        ,WX_NEW_WXTOOLBAR
        ,WX_NEW_WXTREECTRL
        ,WX_NEW_WXWIZARD
        ,WX_NEW_WXWIZARDPAGE
        ,WX_NEW_WXPAINTDC
        ,WX_NEW_WXBOXSIZER
        ,WX_NEW_WXCLIENTDC
        ,WX_NEW_WXIPV4ADDRESS
        ,WX_NEW_WXMASK
        ,WX_NEW_WXGRIDCELLCHOICEEDITOR   -- not implemented
        ,WX_NEW_WXREGEX
        ,WX_NEW_WXTEXTATTR
        ,WX_NEW_WXMEDIACTRL
        ,WX_NEW_WXAUIMANAGER
        ,WX_NEW_WXAUIPANEINFO
        ,WX_NEW_WXAUINOTEBOOK
        ,WX_NEW_WXXMLDOCUMENT
        ,WX_NEW_WXXMLNODE
        ,WX_NEW_WXXMLPROPERTY
        ,WX_NEW_WXRICHTEXTCTRL
        ,WX_NEW_WXTEXTATTREX
        ,WX_NEW_WXRICHTEXTATTR
        ,WX_NEW_WXPRINTER
        ,WX_NEW_WXHTMLEASYPRINTING
        ,WX_NEW_WXCHOICEBOOK
        ,WX_NEW_WXLISTBOOK
        ,WX_NEW_WXTREEBOOK
        ,WX_NEW_WXTOOLBOOK
        ,WX_NEW_WXRICHTEXTFORMATTINGDIALOG
        ,WX_NEW_WXGRIDCELLAUTOWRAPSTRINGEDITOR
        ,WX_NEW_WXSTATICBOXSIZER
        ,WX_NEW_WXIMAGE
        ,WX_NEW_WXABOUTDIALOGINFO
        ,WX_NEW_WXCARET
        ,WX_NEW_WXXMLRESOURCE
        ,WX_NEW_WXDISPLAY
        ,WX_NEW_WXMULTICHOICEDIALOG
        ,WX_NEW_WXPASSWORDENTRYDIALOG
        ,WX_NEW_WXTEXTENTRYDIALOG
        ,WX_NEW_WXSOUND
        ,WX_NEW_WXBITMAPCOMBOBOX
        ,WX_NEW_WXDIRCTRL
        ,WX_NEW_WXPROCESS
        ,WX_NEW_WXSTDDIALOGBUTTONSIZER
        ,WX_NEW_WXGAUGE
        ,WX_NEW_WXPOPUPWINDOW
        ,WX_NEW_WXPOPUPTRANSIENTWINDOW
        ,WX_NEW_WXCONFIG
        ,WX_NEW_WXTASKBARICON
        ,WX_NEW_WXSINGLECHOICEDIALOG
                ,WX_NEW_WXFILEDIALOG
                ,WX_NEW_WXDIRPICKERCTRL
                ,WX_NEW_WXCOLOURPICKERCTRL
                ,WX_NEW_WXDATEPICKERCTRL
                ,WX_NEW_WXFILEPICKERCTRL
                ,WX_NEW_WXFONTPICKERCTRL
                ,WX_NEW_WXCOLLAPSIBLEPANE
                ,WX_NEW_WXLOCALE
                ,WX_NEW_WXCSCONV
                ,WX_NEW_WXAUIDOCKART
                ,WX_NEW_WXRICHTEXTPRINTING
		,WX_NEW_WXTOOLBARITEM
		,WX_NEW_WXGRAPHICSMATRIX
		,WX_NEW_WXGRAPHICSCONTEXT
		,WX_NEW_WXGRAPHICSPEN
		,WX_NEW_WXGRAPHICSBRUSH
		,WX_NEW_WXGRAPHICSPATH
		,WX_NEW_WXARRAYSTRINGPROPERTY
		,WX_NEW_WXBOOLPROPERTY
		,WX_NEW_WXCOLOURPROPERTY
		,WX_NEW_WXCURSORPROPERTY
		,WX_NEW_WXDATEPROPERTY
		,WX_NEW_WXDIRPROPERTY
		,WX_NEW_WXEDITENUMPROPERTY
		,WX_NEW_WXENUMPROPERTY
		,WX_NEW_WXFILEPROPERTY
		,WX_NEW_WXFLOATPROPERTY
		,WX_NEW_WXFONTPROPERTY
		,WX_NEW_WXIMAGEFILEPROPERTY
		,WX_NEW_WXINTPROPERTY
		,WX_NEW_WXLONGSTRINGPROPERTY
		,WX_NEW_WXMULTICHOICEPROPERTY
		,WX_NEW_WXPROPERTYCATEGORY
		,WX_NEW_WXSTRINGPROPERTY
		,WX_NEW_WXSYSTEMCOLOURPROPERTY
		,WX_NEW_WXUINTPROPERTY
		,WX_NEW_WXPROPERTYGRID
		,WX_NEW_WXFLAGSPROPERTY
--@WX_NEW
    }

-- Event definitions
-------------------------------------
atom wx_event_type
wx_event_type = wx_define_c_var( "EuEventTypes" )
--wx_event_type = define_c_var( wxWidgets, "EuEventTypes" )

function next_event()
    --integer event_type
    atom event_type
    event_type = peek4u( wx_event_type )
    wx_event_type += 4
    return event_type
end function

public constant
wxSPLASH_CENTRE_ON_PARENT   = #01,
wxSPLASH_CENTRE_ON_SCREEN   = #02,
wxSPLASH_NO_CENTRE          = #00,
wxSPLASH_TIMEOUT            = #04,
wxSPLASH_NO_TIMEOUT         = #00

public constant

--  wxEVT_COMMAND_BUTTON_CLICKED        = next_event(), -- 32, -- 10000 + 1,
--  wxEVT_COMMAND_CHECKBOX_CLICKED      = next_event(), -- 33, --10000 + 2,
--  wxEVT_COMMAND_CHOICE_SELECTED       = next_event(), -- 10000 + 3,
--  wxEVT_COMMAND_LISTBOX_SELECTED      = next_event(), -- 10000 + 4,
--  wxEVT_COMMAND_LISTBOX_DOUBLECLICKED = next_event(), -- 10000 + 5,
--  wxEVT_COMMAND_CHECKLISTBOX_TOGGLED  = next_event(), -- 37, -- 10000 + 6, -- 37?
----// now they are in wx/textctrl.h
----#if WXWIN_COMPATIBILITY_EVENT_TYPES
--  wxEVT_COMMAND_TEXT_UPDATED          = next_event(), -- 10000 + 7,
--  wxEVT_COMMAND_TEXT_ENTER            = next_event(), -- 10000 + 8,
--  wxEVT_COMMAND_TEXT_URL              = next_event(), -- 10000 + 13,
--  wxEVT_COMMAND_TEXT_MAXLEN           = next_event(), -- 10000 + 14,
----#endif // WXWIN_COMPATIBILITY_EVENT_TYPES
--  wxEVT_COMMAND_MENU_SELECTED         = next_event(), -- 10000 + 9,
--  wxEVT_COMMAND_SLIDER_UPDATED        = next_event(), -- 10000 + 10,
--  wxEVT_COMMAND_RADIOBOX_SELECTED     = next_event(), -- 10000 + 11,
--  wxEVT_COMMAND_RADIOBUTTON_SELECTED  = next_event(), -- 41, --10000 + 12,
--
----// wxEVT_COMMAND_SCROLLBAR_UPDATED is now obsolete since we use
----// wxEVT_SCROLL... events
--
--  wxEVT_COMMAND_SCROLLBAR_UPDATED     = next_event(), -- 10000 + 13, -- 42?
--  wxEVT_COMMAND_VLBOX_SELECTED        = next_event(), -- 10000 + 14,
--  wxEVT_COMMAND_COMBOBOX_SELECTED     = next_event(), -- 10000 + 15,
--  wxEVT_COMMAND_TOOL_RCLICKED         = next_event(), -- 10000 + 16,
--  wxEVT_COMMAND_TOOL_ENTER            = next_event(), -- 10000 + 17,
--  wxEVT_COMMAND_SPINCTRL_UPDATED      = next_event(), -- 10000 + 18, -- 47?
--
----// Sockets and timers send events, too
--  wxEVT_SOCKET                        = next_event(), -- 12000 + 50,
--  wxEVT_TIMER                         = next_event(), -- 10000 + 80,
--
----// Mouse event types
--  wxEVT_LEFT_DOWN                     = next_event(), -- 10000 + 100,
--  wxEVT_LEFT_UP                       = next_event(), -- 10000 + 101,
--  wxEVT_MIDDLE_DOWN                   = next_event(), -- 10000 + 102,
--  wxEVT_MIDDLE_UP                     = next_event(), -- 10000 + 103,
--  wxEVT_RIGHT_DOWN                    = next_event(), -- 10000 + 104,
--  wxEVT_RIGHT_UP                      = next_event(), -- 10000 + 105,
--  wxEVT_MOTION                        = next_event(), -- 10000 + 106,
--  wxEVT_ENTER_WINDOW                  = next_event(), -- 10000 + 107,
--  wxEVT_LEAVE_WINDOW                  = next_event(), -- 10000 + 108,
--  wxEVT_LEFT_DCLICK                   = next_event(), -- 10000 + 109,
--  wxEVT_MIDDLE_DCLICK                 = next_event(), -- 10000 + 110,
--  wxEVT_RIGHT_DCLICK                  = next_event(), -- 10000 + 111,
--  wxEVT_SET_FOCUS                     = next_event(), -- 10000 + 112,
--  wxEVT_KILL_FOCUS                    = next_event(), -- 10000 + 113,
--  wxEVT_CHILD_FOCUS                   = next_event(), -- 10000 + 114,
--  wxEVT_MOUSEWHEEL                    = next_event(), -- 10000 + 115, -- 65? -> 64
--
----// Non-client mouse events
--  wxEVT_NC_LEFT_DOWN                  = next_event(), -- 10000 + 200,
--  wxEVT_NC_LEFT_UP                    = next_event(), -- 10000 + 201,
--  wxEVT_NC_MIDDLE_DOWN                = next_event(), -- 10000 + 202,
--  wxEVT_NC_MIDDLE_UP                  = next_event(), -- 10000 + 203,
--  wxEVT_NC_RIGHT_DOWN                 = next_event(), -- 10000 + 204,
--  wxEVT_NC_RIGHT_UP                   = next_event(), -- 10000 + 205,
--  wxEVT_NC_MOTION                     = next_event(), -- 10000 + 206,
--  wxEVT_NC_ENTER_WINDOW               = next_event(), -- 10000 + 207,
--  wxEVT_NC_LEAVE_WINDOW               = next_event(), -- 10000 + 208,
--  wxEVT_NC_LEFT_DCLICK                = next_event(), -- 10000 + 209,
--  wxEVT_NC_MIDDLE_DCLICK              = next_event(), -- 10000 + 210,
--  wxEVT_NC_RIGHT_DCLICK               = next_event(), -- 77, --10000 + 211, -- 77
--
----// Character input event type
--  wxEVT_CHAR                          = next_event(), -- 10000 + 212, -- 78?
--  wxEVT_CHAR_HOOK                     = next_event(), -- 10000 + 213,
--  wxEVT_NAVIGATION_KEY                = next_event(), -- 10000 + 214,
--  wxEVT_KEY_DOWN                      = next_event(), -- 10000 + 215,
--  wxEVT_KEY_UP                        = next_event(), -- 81, --10000 + 216,
----#if wxUSE_HOTKEY
--  wxEVT_HOTKEY                        = next_event(), -- 10000 + 217, -- 82 ?
----#endif
----// Set cursor event
--  wxEVT_SET_CURSOR                    = next_event(), -- 10000 + 230,
--
----// wxScrollBar and wxSlider event identifiers
--  wxEVT_SCROLL_TOP                    = next_event(), -- 10000 + 300,
--  wxEVT_SCROLL_BOTTOM                 = next_event(), -- 10000 + 301,
--  wxEVT_SCROLL_LINEUP                 = next_event(), -- 10000 + 302,
--  wxEVT_SCROLL_LINEDOWN               = next_event(), -- 10000 + 303,
--  wxEVT_SCROLL_PAGEUP                 = next_event(), -- 10000 + 304,
--  wxEVT_SCROLL_PAGEDOWN               = next_event(), -- 10000 + 305,
--  wxEVT_SCROLL_THUMBTRACK             = next_event(), -- 10000 + 306,
--  wxEVT_SCROLL_THUMBRELEASE           = next_event(), -- 10000 + 307,
----#if wxABI_VERSION >= next_event(), -- 20601
--  wxEVT_SCROLL_CHANGED                = next_event(), -- 10000 + 308,
----#endif
--
----// Scroll events from wxWindow
--  wxEVT_SCROLLWIN_TOP                 = next_event(), -- 10000 + 320,
--  wxEVT_SCROLLWIN_BOTTOM              = next_event(), -- 10000 + 321,
--  wxEVT_SCROLLWIN_LINEUP              = next_event(), -- 10000 + 322,
--  wxEVT_SCROLLWIN_LINEDOWN            = next_event(), -- 10000 + 323,
--  wxEVT_SCROLLWIN_PAGEUP              = next_event(), -- 10000 + 324,
--  wxEVT_SCROLLWIN_PAGEDOWN            = next_event(), -- 10000 + 325,
--  wxEVT_SCROLLWIN_THUMBTRACK          = next_event(), -- 10000 + 326,
--  wxEVT_SCROLLWIN_THUMBRELEASE        = next_event(), -- 10000 + 327, -- 100
--
----// System events
--  wxEVT_SIZE                          = next_event(), -- 10000 + 400,
--  wxEVT_MOVE                          = next_event(), -- 10000 + 401,
--  wxEVT_CLOSE_WINDOW                  = next_event(), -- 10000 + 402,
--  wxEVT_END_SESSION                   = next_event(), -- 10000 + 403,
--  wxEVT_QUERY_END_SESSION             = next_event(), -- 10000 + 404,
--  wxEVT_ACTIVATE_APP                  = next_event(), -- 10000 + 405,
----    wxEVT_POWER                         = next_event(), -- 10000 + 406,
--  wxEVT_ACTIVATE                      = next_event(), -- 10000 + 409,
--  wxEVT_CREATE                        = next_event(), -- 10000 + 410,
--  wxEVT_DESTROY                       = next_event(), -- 10000 + 411,
--  wxEVT_SHOW                          = next_event(), -- 10000 + 412,
--  wxEVT_ICONIZE                       = next_event(), -- 10000 + 413,
--  wxEVT_MAXIMIZE                      = next_event(), -- 10000 + 414,
--  wxEVT_MOUSE_CAPTURE_CHANGED         = next_event(), -- 10000 + 415,
--  wxEVT_PAINT                         = next_event(), -- 10000 + 416, -- 116
--  wxEVT_ERASE_BACKGROUND              = next_event(), -- 10000 + 417,
--  wxEVT_NC_PAINT                      = next_event(), -- 10000 + 418,
--  wxEVT_PAINT_ICON                    = next_event(), -- 10000 + 419,
--  wxEVT_MENU_OPEN                     = next_event(), -- 10000 + 420,
--  wxEVT_MENU_CLOSE                    = next_event(), -- 10000 + 421,
--  wxEVT_MENU_HIGHLIGHT                = next_event(), -- 10000 + 422,
----    wxEVT_POPUP_MENU_INIT               = next_event(), -- 10000 + 423,
--  wxEVT_CONTEXT_MENU                  = next_event(), -- 10000 + 424,
--  wxEVT_SYS_COLOUR_CHANGED            = next_event(), -- 10000 + 425,
--  wxEVT_DISPLAY_CHANGED               = next_event(), -- 10000 + 426,
--  wxEVT_SETTING_CHANGED               = next_event(), -- 10000 + 427,
--  wxEVT_QUERY_NEW_PALETTE             = next_event(), -- 10000 + 428,
--  wxEVT_PALETTE_CHANGED               = next_event(), -- 10000 + 429,
--  wxEVT_JOY_BUTTON_DOWN               = next_event(), -- 10000 + 430,
--  wxEVT_JOY_BUTTON_UP                 = next_event(), -- 10000 + 431,
--  wxEVT_JOY_MOVE                      = next_event(), -- 10000 + 432,
--  wxEVT_JOY_ZMOVE                     = next_event(), -- 10000 + 433,
--  wxEVT_DROP_FILES                    = next_event(), -- 10000 + 434,
--  wxEVT_DRAW_ITEM                     = next_event(), -- 10000 + 435,
--  wxEVT_MEASURE_ITEM                  = next_event(), -- 10000 + 436,
--  wxEVT_COMPARE_ITEM                  = next_event(), -- 10000 + 437,
--  wxEVT_INIT_DIALOG                   = next_event(), -- 10000 + 438,
--  wxEVT_IDLE                          = next_event(), -- 12000 + 439,
--  wxEVT_UPDATE_UI                     = next_event(), -- 10000 + 440,
--  wxEVT_SIZING                        = next_event(), -- 10000 + 441,
--  wxEVT_MOVING                        = next_event(), -- 10000 + 442,
--  wxEVT_HIBERNATE                     = next_event(), -- 10000 + 443,
--
----// Generic command events
----// Note: a click is a higher-level event than button down/up
--  wxEVT_COMMAND_LEFT_CLICK            = next_event(), -- 10000 + 500,
--  wxEVT_COMMAND_LEFT_DCLICK           = next_event(), -- 10000 + 501,
--  wxEVT_COMMAND_RIGHT_CLICK           = next_event(), -- 10000 + 502,
--  wxEVT_COMMAND_RIGHT_DCLICK          = next_event(), -- 10000 + 503,
--  wxEVT_COMMAND_SET_FOCUS             = next_event(), -- 10000 + 504,
--  wxEVT_COMMAND_KILL_FOCUS            = next_event(), -- 10000 + 505,
--  wxEVT_COMMAND_ENTER                 = next_event(), -- 10000 + 506,
--
----// Help events
--  wxEVT_HELP                          = next_event(), -- 10000 + 1050,
--  wxEVT_DETAILED_HELP                 = next_event(), -- 10000 + 1051,
--
--  wxEVT_COMMAND_TOOL_CLICKED          = next_event(),
--
--  wxEVT_WIZARD_CANCEL                 = next_event(),
--  wxEVT_WIZARD_FINISHED               = next_event(),
--  wxEVT_WIZARD_HELP                   = next_event(),
--  wxEVT_WIZARD_PAGE_CHANGED           = next_event(),
--  wxEVT_WIZARD_PAGE_CHANGING          = next_event(),
--
--  wxEVT_COMMAND_TOGGLEBUTTON_CLICKED  = next_event(),
--
--  wxEVT_GRID_CELL_CHANGE              = next_event(),
--  wxEVT_GRID_CELL_LEFT_CLICK          = next_event(),
--  wxEVT_GRID_CELL_LEFT_DCLICK         = next_event(),
--  wxEVT_GRID_CELL_RIGHT_CLICK         = next_event(),
--  wxEVT_GRID_CELL_RIGHT_DCLICK        = next_event(),
--  wxEVT_GRID_COL_SIZE                 = next_event(),
--  wxEVT_GRID_EDITOR_CREATED           = next_event(),
--  wxEVT_GRID_EDITOR_HIDDEN            = next_event(),
--  wxEVT_GRID_EDITOR_SHOWN             = next_event(),
--  wxEVT_GRID_LABEL_LEFT_CLICK         = next_event(),
--  wxEVT_GRID_LABEL_LEFT_DCLICK        = next_event(),
--  wxEVT_GRID_LABEL_RIGHT_CLICK        = next_event(),
--  wxEVT_GRID_LABEL_RIGHT_DCLICK       = next_event(),
--  wxEVT_GRID_RANGE_SELECT             = next_event(),
--  wxEVT_GRID_ROW_SIZE                 = next_event(),
--  wxEVT_GRID_SELECT_CELL              = next_event(),
--  wxEVT_COMMAND_TREE_SEL_CHANGED      = next_event(),
--  wxEVT_COMMAND_LIST_ITEM_SELECTED    = next_event(),
--  wxEVT_COMMAND_HTML_LINK_CLICKED     = next_event(),
--  wxEVT_COMMAND_LIST_ITEM_RIGHT_CLICK = next_event(),
--
--  wxEVT_COMMAND_RICHTEXT_CHARACTER            = next_event(),
--  wxEVT_COMMAND_RICHTEXT_DELETE               = next_event(),
--  wxEVT_COMMAND_RICHTEXT_RETURN               = next_event(),
--  wxEVT_COMMAND_RICHTEXT_STYLE_CHANGED        = next_event(),
--  wxEVT_COMMAND_RICHTEXT_STYLESHEET_CHANGING  = next_event(),
--  wxEVT_COMMAND_RICHTEXT_STYLESHEET_REPLACING = next_event(),
--  wxEVT_COMMAND_RICHTEXT_STYLESHEET_REPLACED  = next_event(),
--  wxEVT_COMMAND_RICHTEXT_CONTENT_INSERTED     = next_event(),
--  wxEVT_COMMAND_RICHTEXT_CONTENT_DELETED      = next_event(),
--  wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGING    = next_event(),
--  wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGED     = next_event(),
--  wxEVT_COMMAND_SPLITTER_UNSPLIT              = next_event(),
--  wxEVT_COMMAND_SPLITTER_DOUBLECLICKED        = next_event(),
--  wxEVT_COMMAND_NOTEBOOK_PAGE_CHANGED         = next_event(),
--  wxEVT_COMMAND_NOTEBOOK_PAGE_CHANGING        = next_event(),
--
--  wxEVT_END_PROCESS                           = next_event(),
--
--  wxEVT_MEDIA_LOADED                          = next_event(),
--  wxEVT_MEDIA_STOP                            = next_event(),
--  wxEVT_MEDIA_FINISHED                        = next_event(),
--  wxEVT_MEDIA_STATECHANGED                    = next_event(),
--  wxEVT_MEDIA_PLAY                            = next_event(),
--  wxEVT_MEDIA_PAUSE                           = next_event()


    wxEVT_COMMAND_BUTTON_CLICKED    = next_event(), --// 32 = next_event(), -- 10000 + 1    = next_event(),
    wxEVT_COMMAND_CHECKBOX_CLICKED  = next_event(), --// 33 = next_event(), --10000 + 2 = next_event(),
    wxEVT_COMMAND_CHOICE_SELECTED   = next_event(), --// 10000 + 3  = next_event(),
    wxEVT_COMMAND_LISTBOX_SELECTED  = next_event(), --// 10000 + 4  = next_event(),
    wxEVT_COMMAND_LISTBOX_DOUBLECLICKED = next_event(), --// 10000 + 5  = next_event(),
    wxEVT_COMMAND_CHECKLISTBOX_TOGGLED  = next_event(), --// 37 = next_event(), -- 10000 + 6    = next_event(), -- 37?
    --// now they are in wx/textctrl.h
--//#if WXWIN_COMPATIBILITY_EVENT_TYPES
    wxEVT_COMMAND_TEXT_UPDATED  = next_event(), --// 10000 + 7  = next_event(),
    wxEVT_COMMAND_TEXT_ENTER    = next_event(), --// 10000 + 8  = next_event(),
    wxEVT_COMMAND_TEXT_URL  = next_event(), --// 10000 + 13 = next_event(),
    wxEVT_COMMAND_TEXT_MAXLEN   = next_event(), --// 10000 + 14 = next_event(),
--//#endif --// WXWIN_COMPATIBILITY_EVENT_TYPES
    wxEVT_COMMAND_MENU_SELECTED = next_event(), --// 10000 + 9  = next_event(),
    wxEVT_COMMAND_SLIDER_UPDATED    = next_event(), --// 10000 + 10 = next_event(),
    wxEVT_COMMAND_RADIOBOX_SELECTED = next_event(), --// 10000 + 11 = next_event(),
    wxEVT_COMMAND_RADIOBUTTON_SELECTED  = next_event(), --// 41 = next_event(), --10000 + 12    = next_event(),

    --// wxEVT_COMMAND_SCROLLBAR_UPDATED is now obsolete since we use
    --// wxEVT_SCROLL... events

    wxEVT_COMMAND_SCROLLBAR_UPDATED = next_event(), --// 10000 + 13 = next_event(), -- 42?
    wxEVT_COMMAND_VLBOX_SELECTED    = next_event(), --// 10000 + 14 = next_event(),
    wxEVT_COMMAND_COMBOBOX_SELECTED = next_event(), --// 10000 + 15 = next_event(),
    wxEVT_COMMAND_TOOL_RCLICKED = next_event(), --// 10000 + 16 = next_event(),
    wxEVT_COMMAND_TOOL_ENTER    = next_event(), --// 10000 + 17 = next_event(),
    wxEVT_COMMAND_SPINCTRL_UPDATED  = next_event(), --// 10000 + 18 = next_event(), -- 47?

    --// Sockets and timers send events = next_event(), too
    wxEVT_SOCKET    = next_event(), --// 12000 + 50 = next_event(),
    wxEVT_TIMER = next_event(), --// 10000 + 80 = next_event(),

    --// Mouse event types
    wxEVT_LEFT_DOWN = next_event(), --// 10000 + 100    = next_event(),
    wxEVT_LEFT_UP   = next_event(), --// 10000 + 101    = next_event(),
    wxEVT_MIDDLE_DOWN   = next_event(), --// 10000 + 102    = next_event(),
    wxEVT_MIDDLE_UP = next_event(), --// 10000 + 103    = next_event(),
    wxEVT_RIGHT_DOWN    = next_event(), --// 10000 + 104    = next_event(),
    wxEVT_RIGHT_UP  = next_event(), --// 10000 + 105    = next_event(),
    wxEVT_MOTION    = next_event(), --// 10000 + 106    = next_event(),
    wxEVT_ENTER_WINDOW  = next_event(), --// 10000 + 107    = next_event(),
    wxEVT_LEAVE_WINDOW  = next_event(), --// 10000 + 108    = next_event(),
    wxEVT_LEFT_DCLICK   = next_event(), --// 10000 + 109    = next_event(),
    wxEVT_MIDDLE_DCLICK = next_event(), --// 10000 + 110    = next_event(),
    wxEVT_RIGHT_DCLICK  = next_event(), --// 10000 + 111    = next_event(),
    wxEVT_SET_FOCUS = next_event(), --// 10000 + 112    = next_event(),
    wxEVT_KILL_FOCUS    = next_event(), --// 10000 + 113    = next_event(),
    wxEVT_CHILD_FOCUS   = next_event(), --// 10000 + 114    = next_event(),
    wxEVT_MOUSEWHEEL    = next_event(), --// 10000 + 115    = next_event(), -- 65? -> 64

    --// Non-client mouse events
    wxEVT_NC_LEFT_DOWN  = next_event(), --// 10000 + 200    = next_event(),
    wxEVT_NC_LEFT_UP    = next_event(), --// 10000 + 201    = next_event(),
    wxEVT_NC_MIDDLE_DOWN    = next_event(), --// 10000 + 202    = next_event(),
    wxEVT_NC_MIDDLE_UP  = next_event(), --// 10000 + 203    = next_event(),
    wxEVT_NC_RIGHT_DOWN = next_event(), --// 10000 + 204    = next_event(),
    wxEVT_NC_RIGHT_UP   = next_event(), --// 10000 + 205    = next_event(),
    wxEVT_NC_MOTION = next_event(), --// 10000 + 206    = next_event(),
    wxEVT_NC_ENTER_WINDOW   = next_event(), --// 10000 + 207    = next_event(),
    wxEVT_NC_LEAVE_WINDOW   = next_event(), --// 10000 + 208    = next_event(),
    wxEVT_NC_LEFT_DCLICK    = next_event(), --// 10000 + 209    = next_event(),
    wxEVT_NC_MIDDLE_DCLICK  = next_event(), --// 10000 + 210    = next_event(),
    wxEVT_NC_RIGHT_DCLICK   = next_event(), --// 77 = next_event(), --10000 + 211   = next_event(), -- 77

    --// Character input event type
    wxEVT_CHAR  = next_event(), --// 10000 + 212    = next_event(), -- 78?
    wxEVT_CHAR_HOOK = next_event(), --// 10000 + 213    = next_event(),
    wxEVT_NAVIGATION_KEY    = next_event(), --// 10000 + 214    = next_event(),
    wxEVT_KEY_DOWN  = next_event(), --// 10000 + 215    = next_event(),
    wxEVT_KEY_UP    = next_event(), --// 81 = next_event(), --10000 + 216   = next_event(),

    wxEVT_HOTKEY    = next_event(), --// 10000 + 217    = next_event(), -- 82 ?
    --// Set cursor event
    wxEVT_SET_CURSOR    = next_event(), --// 10000 + 230    = next_event(),

    --// wxScrollBar and wxSlider event identifiers
    wxEVT_SCROLL_TOP    = next_event(), --// 10000 + 300    = next_event(),
    wxEVT_SCROLL_BOTTOM = next_event(), --// 10000 + 301    = next_event(),
    wxEVT_SCROLL_LINEUP = next_event(), --// 10000 + 302    = next_event(),
    wxEVT_SCROLL_LINEDOWN   = next_event(), --// 10000 + 303    = next_event(),
    wxEVT_SCROLL_PAGEUP = next_event(), --// 10000 + 304    = next_event(),
    wxEVT_SCROLL_PAGEDOWN   = next_event(), --// 10000 + 305    = next_event(),
    wxEVT_SCROLL_THUMBTRACK = next_event(), --// 10000 + 306    = next_event(),
    wxEVT_SCROLL_THUMBRELEASE   = next_event(), --// 10000 + 307    = next_event(),
    wxEVT_SCROLL_CHANGED    = next_event(), --// 10000 + 308    = next_event(),

    --// Scroll events from wxWindow
    wxEVT_SCROLLWIN_TOP = next_event(), --// 10000 + 320    = next_event(),
    wxEVT_SCROLLWIN_BOTTOM  = next_event(), --// 10000 + 321    = next_event(),
    wxEVT_SCROLLWIN_LINEUP  = next_event(), --// 10000 + 322    = next_event(),
    wxEVT_SCROLLWIN_LINEDOWN    = next_event(), --// 10000 + 323    = next_event(),
    wxEVT_SCROLLWIN_PAGEUP  = next_event(), --// 10000 + 324    = next_event(),
    wxEVT_SCROLLWIN_PAGEDOWN    = next_event(), --// 10000 + 325    = next_event(),
    wxEVT_SCROLLWIN_THUMBTRACK  = next_event(), --// 10000 + 326    = next_event(),
    wxEVT_SCROLLWIN_THUMBRELEASE    = next_event(), --// 10000 + 327    = next_event(), -- 100

    --// System events
    wxEVT_SIZE  = next_event(), --// 10000 + 400    = next_event(),
    wxEVT_MOVE  = next_event(), --// 10000 + 401    = next_event(),
    wxEVT_CLOSE_WINDOW  = next_event(), --// 10000 + 402    = next_event(),
    wxEVT_END_SESSION   = next_event(), --// 10000 + 403    = next_event(),
    wxEVT_QUERY_END_SESSION = next_event(), --// 10000 + 404    = next_event(),
    wxEVT_ACTIVATE_APP  = next_event(), --// 10000 + 405    = next_event(),
--//    wxEVT_POWER = next_event(), --// 10000 + 406    = next_event(),
    wxEVT_ACTIVATE  = next_event(), --// 10000 + 409    = next_event(),
    wxEVT_CREATE    = next_event(), --// 10000 + 410    = next_event(),
    wxEVT_DESTROY   = next_event(), --// 10000 + 411    = next_event(),
    wxEVT_SHOW  = next_event(), --// 10000 + 412    = next_event(),
    wxEVT_ICONIZE   = next_event(), --// 10000 + 413    = next_event(),
    wxEVT_MAXIMIZE  = next_event(), --// 10000 + 414    = next_event(),
    wxEVT_MOUSE_CAPTURE_CHANGED = next_event(), --// 10000 + 415    = next_event(),
    wxEVT_PAINT = next_event(), --// 10000 + 416    = next_event(), -- 116
    wxEVT_ERASE_BACKGROUND  = next_event(), --// 10000 + 417    = next_event(),
    wxEVT_NC_PAINT  = next_event(), --// 10000 + 418    = next_event(),
    wxEVT_PAINT_ICON    = next_event(), --// 10000 + 419    = next_event(),
    wxEVT_MENU_OPEN = next_event(), --// 10000 + 420    = next_event(),
    wxEVT_MENU_CLOSE    = next_event(), --// 10000 + 421    = next_event(),
    wxEVT_MENU_HIGHLIGHT    = next_event(), --// 10000 + 422    = next_event(),
    --// wxEVT_POPUP_MENU_INIT  = next_event(), --// 10000 + 423    = next_event(),
    wxEVT_CONTEXT_MENU  = next_event(), --// 10000 + 424    = next_event(),
    wxEVT_SYS_COLOUR_CHANGED    = next_event(), --// 10000 + 425    = next_event(),
    wxEVT_DISPLAY_CHANGED   = next_event(), --// 10000 + 426    = next_event(),
    wxEVT_SETTING_CHANGED   = next_event(), --// 10000 + 427    = next_event(),
    wxEVT_QUERY_NEW_PALETTE = next_event(), --// 10000 + 428    = next_event(),
    wxEVT_PALETTE_CHANGED   = next_event(), --// 10000 + 429    = next_event(),
    wxEVT_JOY_BUTTON_DOWN   = next_event(), --// 10000 + 430    = next_event(),
    wxEVT_JOY_BUTTON_UP = next_event(), --// 10000 + 431    = next_event(),
    wxEVT_JOY_MOVE  = next_event(), --// 10000 + 432    = next_event(),
    wxEVT_JOY_ZMOVE = next_event(), --// 10000 + 433    = next_event(),
    wxEVT_DROP_FILES    = next_event(), --// 10000 + 434    = next_event(),
    wxEVT_DRAW_ITEM = next_event(), --// 10000 + 435    = next_event(),
    wxEVT_MEASURE_ITEM  = next_event(), --// 10000 + 436    = next_event(),
    wxEVT_COMPARE_ITEM  = next_event(), --// 10000 + 437    = next_event(),
    wxEVT_INIT_DIALOG   = next_event(), --// 10000 + 438    = next_event(),
    wxEVT_IDLE  = next_event(), --// 12000 + 439    = next_event(),
    wxEVT_UPDATE_UI = next_event(), --// 10000 + 440    = next_event(),
    wxEVT_SIZING    = next_event(), --// 10000 + 441    = next_event(),
    wxEVT_MOVING    = next_event(), --// 10000 + 442    = next_event(),
    wxEVT_HIBERNATE = next_event(), --// 10000 + 443    = next_event(),

    --// Generic command events
    --// Note: a click is a higher-level event than button down/up
    wxEVT_COMMAND_LEFT_CLICK    = next_event(), --// 10000 + 500    = next_event(),
    wxEVT_COMMAND_LEFT_DCLICK   = next_event(), --// 10000 + 501    = next_event(),
    wxEVT_COMMAND_RIGHT_CLICK   = next_event(), --// 10000 + 502    = next_event(),
    wxEVT_COMMAND_RIGHT_DCLICK  = next_event(), --// 10000 + 503    = next_event(),
    wxEVT_COMMAND_SET_FOCUS = next_event(), --// 10000 + 504    = next_event(),
    wxEVT_COMMAND_KILL_FOCUS    = next_event(), --// 10000 + 505    = next_event(),
    wxEVT_COMMAND_ENTER = next_event(), --// 10000 + 506    = next_event(),

    wxEVT_COMMAND_TREE_BEGIN_DRAG = next_event(), -- 600)
    wxEVT_COMMAND_TREE_BEGIN_RDRAG = next_event(), -- 601)
    wxEVT_COMMAND_TREE_BEGIN_LABEL_EDIT = next_event(), -- 602)
    wxEVT_COMMAND_TREE_END_LABEL_EDIT = next_event(), -- 603)
    wxEVT_COMMAND_TREE_DELETE_ITEM = next_event(), -- 604)
    wxEVT_COMMAND_TREE_GET_INFO = next_event(), -- 605)
    wxEVT_COMMAND_TREE_SET_INFO = next_event(), -- 606)
    wxEVT_COMMAND_TREE_ITEM_EXPANDED = next_event(), -- 607)
    wxEVT_COMMAND_TREE_ITEM_EXPANDING = next_event(), -- 608)
    wxEVT_COMMAND_TREE_ITEM_COLLAPSED = next_event(), -- 609)
    wxEVT_COMMAND_TREE_ITEM_COLLAPSING = next_event(), -- 610)
    wxEVT_COMMAND_TREE_SEL_CHANGED = next_event(), -- 611)
    wxEVT_COMMAND_TREE_SEL_CHANGING = next_event(), -- 612)
    wxEVT_COMMAND_TREE_KEY_DOWN = next_event(), -- 613)
    wxEVT_COMMAND_TREE_ITEM_ACTIVATED = next_event(), -- 614)
    wxEVT_COMMAND_TREE_ITEM_RIGHT_CLICK = next_event(), -- 615)
    wxEVT_COMMAND_TREE_ITEM_MIDDLE_CLICK = next_event(), -- 616)
    wxEVT_COMMAND_TREE_END_DRAG = next_event(), -- 617)
    wxEVT_COMMAND_TREE_STATE_IMAGE_CLICK = next_event(), -- 618)
    wxEVT_COMMAND_TREE_ITEM_GETTOOLTIP = next_event(), -- 619)
    wxEVT_COMMAND_TREE_ITEM_MENU = next_event(), -- 620)
    --// Help events
    wxEVT_HELP  = next_event(), --// 10000 + 1050   = next_event(),
    wxEVT_DETAILED_HELP = next_event(), --// 10000 + 1051   = next_event(),

    wxEVT_COMMAND_TOOL_CLICKED  = next_event(), --// wxEVT_COMMAND_MENU_SELECTED

    wxEVT_WIZARD_CANCEL = next_event(),
    wxEVT_WIZARD_FINISHED   = next_event(),
    wxEVT_WIZARD_HELP   = next_event(),
    wxEVT_WIZARD_PAGE_CHANGED   = next_event(),
    wxEVT_WIZARD_PAGE_CHANGING  = next_event(),

    wxEVT_COMMAND_TOGGLEBUTTON_CLICKED  = next_event(),

    wxEVT_GRID_CELL_CHANGE  = next_event(),
    wxEVT_GRID_CELL_LEFT_CLICK  = next_event(),
    wxEVT_GRID_CELL_LEFT_DCLICK = next_event(),
    wxEVT_GRID_CELL_RIGHT_CLICK = next_event(),
    wxEVT_GRID_CELL_RIGHT_DCLICK    = next_event(),
    wxEVT_GRID_COL_SIZE = next_event(),
    wxEVT_GRID_EDITOR_CREATED   = next_event(),
    wxEVT_GRID_EDITOR_HIDDEN    = next_event(),
    wxEVT_GRID_EDITOR_SHOWN = next_event(),
    wxEVT_GRID_LABEL_LEFT_CLICK = next_event(),
    wxEVT_GRID_LABEL_LEFT_DCLICK    = next_event(),
    wxEVT_GRID_LABEL_RIGHT_CLICK    = next_event(),
    wxEVT_GRID_LABEL_RIGHT_DCLICK   = next_event(),
    wxEVT_GRID_RANGE_SELECT = next_event(),
    wxEVT_GRID_ROW_SIZE = next_event(),
    wxEVT_GRID_SELECT_CELL  = next_event(),
    wxEVT_COMMAND_LIST_ITEM_SELECTED    = next_event(),
    wxEVT_COMMAND_HTML_LINK_CLICKED = next_event(),
    wxEVT_COMMAND_LIST_ITEM_RIGHT_CLICK = next_event(),

    wxEVT_COMMAND_RICHTEXT_CHARACTER    = next_event(),
    wxEVT_COMMAND_RICHTEXT_DELETE   = next_event(),
    wxEVT_COMMAND_RICHTEXT_RETURN   = next_event(),
    wxEVT_COMMAND_RICHTEXT_STYLE_CHANGED    = next_event(),
    wxEVT_COMMAND_RICHTEXT_STYLESHEET_CHANGING  = next_event(),
    wxEVT_COMMAND_RICHTEXT_STYLESHEET_REPLACING = next_event(),
    wxEVT_COMMAND_RICHTEXT_STYLESHEET_REPLACED  = next_event(),
    wxEVT_COMMAND_RICHTEXT_CONTENT_INSERTED = next_event(),
    wxEVT_COMMAND_RICHTEXT_CONTENT_DELETED  = next_event(),
    wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGING    = next_event(),
    wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGED = next_event(),
    wxEVT_COMMAND_SPLITTER_UNSPLIT  = next_event(),
    wxEVT_COMMAND_SPLITTER_DOUBLECLICKED    = next_event(),

    wxEVT_COMMAND_NOTEBOOK_PAGE_CHANGED = next_event(),
    wxEVT_COMMAND_NOTEBOOK_PAGE_CHANGING    = next_event(),

    wxEVT_END_PROCESS   = next_event(),

    wxEVT_MEDIA_LOADED  = next_event(),
    wxEVT_MEDIA_STOP    = next_event(),
    wxEVT_MEDIA_FINISHED    = next_event(),
    wxEVT_MEDIA_STATECHANGED    = next_event(),
    wxEVT_MEDIA_PLAY    = next_event(),
    wxEVT_MEDIA_PAUSE   = next_event(),

    wxEVT_AUI_PANE_BUTTON = next_event(),
    wxEVT_AUI_PANE_CLOSE = next_event(),
    wxEVT_AUI_PANE_MAXIMIZE = next_event(),
    wxEVT_AUI_PANE_RESTORE = next_event(),
    wxEVT_AUI_RENDER = next_event(),
    wxEVT_AUI_FIND_MANAGER = next_event(),

    wxEVT_COMMAND_AUINOTEBOOK_PAGE_CLOSE = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_PAGE_CHANGED = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_PAGE_CHANGING = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_BUTTON = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_BEGIN_DRAG = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_END_DRAG = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_DRAG_MOTION = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_ALLOW_DND = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_TAB_MIDDLE_DOWN = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_TAB_MIDDLE_UP = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_TAB_RIGHT_DOWN = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_TAB_RIGHT_UP = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_PAGE_CLOSED = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_DRAG_DONE = next_event(),
    wxEVT_COMMAND_AUINOTEBOOK_BG_DCLICK = next_event(),

    wxEVT_COMMAND_LIST_BEGIN_DRAG = next_event(),
    wxEVT_COMMAND_COLOURPICKER_CHANGED = next_event(),
        wxEVT_DATE_CHANGED = next_event(),
        wxEVT_COMMAND_DIRPICKER_CHANGED = next_event(),
        wxEVT_COMMAND_FILEPICKER_CHANGED = next_event(),
        wxEVT_COMMAND_FONTPICKER_CHANGED = next_event(),
        wxEVT_COMMAND_COLLPANE_CHANGED = next_event(),
        $

-- Constants
----------------------------
public constant
    wxOS_UNKNOWN = 0,                 --// returned on error

    wxOS_MAC_OS                 = power( 2,  0 ),   --// Apple Mac OS 8/9/X with Mac paths
    wxOS_MAC_OSX_DARWIN         = power( 2,  1 ),   --// Apple Mac OS X with Unix paths
    wxOS_MAC = wx_or_all(
        wxOS_MAC_OS         &
        wxOS_MAC_OSX_DARWIN ),

    wxOS_WINDOWS_9X             = power( 2,  2 ),   --// Windows 9x family (95/98/ME)
    wxOS_WINDOWS_NT             = power( 2,  3 ),   --// Windows NT family (NT/2000/XP)
    wxOS_WINDOWS_MICRO          = power( 2,  4 ),   --// MicroWindows
    wxOS_WINDOWS_CE             = power( 2,  5 ),   --// Windows CE (Window Mobile)
    wxOS_WINDOWS = wx_or_all(
        wxOS_WINDOWS_9X    &
        wxOS_WINDOWS_NT    &
        wxOS_WINDOWS_MICRO &
        wxOS_WINDOWS_CE    ),

    wxOS_UNIX_LINUX             = power( 2,  6 ),   --// Linux
    wxOS_UNIX_FREEBSD           = power( 2,  7 ),   --// FreeBSD
    wxOS_UNIX_OPENBSD           = power( 2,  8 ),   --// OpenBSD
    wxOS_UNIX_NETBSD            = power( 2,  9 ),   --// NetBSD
    wxOS_UNIX_SOLARIS           = power( 2, 10 ),   --// SunOS
    wxOS_UNIX_AIX               = power( 2, 11 ),   --// AIX
    wxOS_UNIX_HPUX              = power( 2, 12 ),   --// HP/UX
    wxOS_UNIX = wx_or_all(
        wxOS_UNIX_LINUX   &
        wxOS_UNIX_FREEBSD &
        wxOS_UNIX_OPENBSD &
        wxOS_UNIX_NETBSD  &
        wxOS_UNIX_SOLARIS &
        wxOS_UNIX_AIX     &
        wxOS_UNIX_HPUX    ),

    wxOS_DOS                    = power( 2, 15 ),   --// Microsoft DOS
    wxOS_OS2                    = power( 2, 16 ),   --// OS/2

    wxPORT_UNKNOWN  = 0,                --// returned on error
    wxPORT_BASE     = 1,                --// wxBase, no native toolkit used
    wxPORT_MSW      = power( 2, 1 ),    --// wxMSW, native toolkit is Windows API
    wxPORT_MOTIF    = power( 2, 2 ),    --// wxMotif, using [Open]Motif or Lesstif
    wxPORT_GTK      = power( 2, 3 ),    --// wxGTK, using GTK+ 1.x, 2.x, GPE or Maemo
    wxPORT_MGL      = power( 2, 4 ),    --// wxMGL, using wxUniversal
    wxPORT_X11      = power( 2, 5 ),    --// wxX11, using wxUniversal
    wxPORT_OS2      = power( 2, 6 ),    --// wxOS2, using OS/2 Presentation Manager
    wxPORT_MAC      = power( 2, 7 ),    --// wxMac, using Carbon or Classic Mac API
    wxPORT_COCOA    = power( 2, 8 ),    --// wxCocoa, using Cocoa NextStep/Mac API
    wxPORT_WINCE    = power( 2, 9 ),    --// wxWinCE, toolkit is WinCE SDK API
    wxPORT_PALMOS   = power( 2, 10),    --// wxPalmOS, toolkit is PalmOS API
    wxPORT_DFB      = power( 2, 11),    --// wxDFB, using wxUniversal

    wxARCH_INVALID  = -1,   --// returned on error
    wxARCH_32       =  0,   --// 32 bit
    wxARCH_64       =  1,   --// 64 bit
    wxARCH_MAX      =  2

public constant
    wxPAGE_ODD      = 0,
    wxPAGE_EVEN     = 1,
    wxPAGE_ALL      = 2

public constant
    wxRICHTEXT_TYPE_ANY     =  0,
    wxRICHTEXT_TYPE_TEXT    =  1,
    wxRICHTEXT_TYPE_XML     =  2,
    wxRICHTEXT_TYPE_HTML    =  3,
    wxRICHTEXT_TYPE_RTF     =  4,
    wxRICHTEXT_TYPE_PDF     =  5,

-- * Flags determining the available space, passed to Layout
    wxRICHTEXT_FIXED_WIDTH      = #01,
    wxRICHTEXT_FIXED_HEIGHT     = #02,
    wxRICHTEXT_VARIABLE_WIDTH   = #04,
    wxRICHTEXT_VARIABLE_HEIGHT  = #08,

    wxRICHTEXT_LAYOUT_SPECIFIED_RECT = #10, --// Only lay out the part of the buffer that lies within
                                            --// the rect passed to Layout.

-- * Flags to pass to Draw
    wxRICHTEXT_DRAW_IGNORE_CACHE    = #01,  --// Ignore paragraph cache optimization, e.g. for printing purposes
                                            --// where one line may be drawn higher (on the next page) compared
                                            --// with the previous line

-- * Flags returned from hit-testing
    wxRICHTEXT_HITTEST_NONE     = #01,  --// The point was not on this object
    wxRICHTEXT_HITTEST_BEFORE   = #02,  --// The point was before the position returned from HitTest
    wxRICHTEXT_HITTEST_AFTER    = #04,  --// The point was after the position returned from HitTest
    wxRICHTEXT_HITTEST_ON       = #08,  --// The point was on the position returned from HitTest
    wxRICHTEXT_HITTEST_OUTSIDE  = #10,  --// The point was on space outside content

-- * Flags for GetRangeSize
    wxRICHTEXT_FORMATTED        = #01,
    wxRICHTEXT_UNFORMATTED      = #02,

-- * Flags for SetStyle/SetListStyle
    wxRICHTEXT_SETSTYLE_NONE            = #00,
    wxRICHTEXT_SETSTYLE_WITH_UNDO       = #01,  --// Specifies that this operation should be undoable
    wxRICHTEXT_SETSTYLE_OPTIMIZE        = #02,  --// Specifies that the style should not be applied if the
                                                --// combined style at this point is already the style in question.
    wxRICHTEXT_SETSTYLE_PARAGRAPHS_ONLY = #04,  --// Specifies that the style should only be applied to paragraphs,
                                                --// and not the content. This allows content styling to be
                                                --// preserved independently from that of e.g. a named paragraph style.
    wxRICHTEXT_SETSTYLE_CHARACTERS_ONLY = #08,  --// Specifies that the style should only be applied to characters,
                                                --// and not the paragraph. This allows content styling to be
                                                --// preserved independently from that of e.g. a named paragraph style.
    wxRICHTEXT_SETSTYLE_RENUMBER        = #10,  --// For SetListStyle only: specifies starting from the given number, otherwise
                                                --// deduces number from existing attributes
    wxRICHTEXT_SETSTYLE_SPECIFY_LEVEL   = #20,  --// For SetListStyle only: specifies the list level for all paragraphs, otherwise
                                                --// the current indentation will be used
    wxRICHTEXT_SETSTYLE_RESET           = #40,  --// Resets the existing style before applying the new style
    wxRICHTEXT_SETSTYLE_REMOVE          = #80,  --// Removes the given style instead of applying it

-- * Flags for text insertion
    wxRICHTEXT_INSERT_NONE                              = #00,
    wxRICHTEXT_INSERT_WITH_PREVIOUS_PARAGRAPH_STYLE     = #01,

-- * Extra formatting flags not in wxTextAttr
    wxTEXT_ATTR_PARA_SPACING_AFTER      = #00000800,
    wxTEXT_ATTR_PARA_SPACING_BEFORE     = #00001000,
    wxTEXT_ATTR_LINE_SPACING            = #00002000,
    wxTEXT_ATTR_CHARACTER_STYLE_NAME    = #00004000,
    wxTEXT_ATTR_PARAGRAPH_STYLE_NAME    = #00008000,
    wxTEXT_ATTR_LIST_STYLE_NAME         = #00010000,
    wxTEXT_ATTR_BULLET_STYLE            = #00020000,
    wxTEXT_ATTR_BULLET_NUMBER           = #00040000,
    wxTEXT_ATTR_BULLET_TEXT             = #00080000,
    wxTEXT_ATTR_BULLET_NAME             = #00100000,
    wxTEXT_ATTR_URL                     = #00200000,
    wxTEXT_ATTR_PAGE_BREAK              = #00400000,
    wxTEXT_ATTR_EFFECTS                 = #00800000,
    wxTEXT_ATTR_OUTLINE_LEVEL           = #01000000,

-- * Styles for wxTextAttrEx::SetBulletStyle
    wxTEXT_ATTR_BULLET_STYLE_NONE               = #00000000,
    wxTEXT_ATTR_BULLET_STYLE_ARABIC             = #00000001,
    wxTEXT_ATTR_BULLET_STYLE_LETTERS_UPPER      = #00000002,
    wxTEXT_ATTR_BULLET_STYLE_LETTERS_LOWER      = #00000004,
    wxTEXT_ATTR_BULLET_STYLE_ROMAN_UPPER        = #00000008,
    wxTEXT_ATTR_BULLET_STYLE_ROMAN_LOWER        = #00000010,
    wxTEXT_ATTR_BULLET_STYLE_SYMBOL             = #00000020,
    wxTEXT_ATTR_BULLET_STYLE_BITMAP             = #00000040,
    wxTEXT_ATTR_BULLET_STYLE_PARENTHESES        = #00000080,
    wxTEXT_ATTR_BULLET_STYLE_PERIOD             = #00000100,
    wxTEXT_ATTR_BULLET_STYLE_STANDARD           = #00000200,
    wxTEXT_ATTR_BULLET_STYLE_RIGHT_PARENTHESIS  = #00000400,
    wxTEXT_ATTR_BULLET_STYLE_OUTLINE            = #00000800,
    wxTEXT_ATTR_BULLET_STYLE_ALIGN_LEFT         = #00000000,
    wxTEXT_ATTR_BULLET_STYLE_ALIGN_RIGHT        = #00001000,
    wxTEXT_ATTR_BULLET_STYLE_ALIGN_CENTRE       = #00002000,

-- * Styles for wxTextAttrEx::SetTextEffects
    wxTEXT_ATTR_EFFECT_NONE                     = #00000000,
    wxTEXT_ATTR_EFFECT_CAPITALS                 = #00000001,
    wxTEXT_ATTR_EFFECT_SMALL_CAPITALS           = #00000002,
    wxTEXT_ATTR_EFFECT_STRIKETHROUGH            = #00000004,
    wxTEXT_ATTR_EFFECT_DOUBLE_STRIKETHROUGH     = #00000008,
    wxTEXT_ATTR_EFFECT_SHADOW                   = #00000010,
    wxTEXT_ATTR_EFFECT_EMBOSS                   = #00000020,
    wxTEXT_ATTR_EFFECT_OUTLINE                  = #00000040,
    wxTEXT_ATTR_EFFECT_ENGRAVE                  = #00000080,
    wxTEXT_ATTR_EFFECT_SUPERSCRIPT              = #00000100,
    wxTEXT_ATTR_EFFECT_SUBSCRIPT                = #00000200,

-- * Line spacing values
    wxTEXT_ATTR_LINE_SPACING_NORMAL        = 10,
    wxTEXT_ATTR_LINE_SPACING_HALF          = 15,
    wxTEXT_ATTR_LINE_SPACING_TWICE         = 20

public constant
    wxRE_READONLY          = #0010,
    wxRE_MULTILINE         = #0020,

    wxRICHTEXT_SHIFT_DOWN  = #01,
    wxRICHTEXT_CTRL_DOWN   = #02,
    wxRICHTEXT_ALT_DOWN    = #04,

    wxRICHTEXT_SELECTED    = #01,
    wxRICHTEXT_TAGGED      = #02,
    wxRICHTEXT_FOCUSSED    = #04,
    wxRICHTEXT_IS_FOCUS    = #08

public constant
    wxXML_ELEMENT_NODE       =  1,
    wxXML_ATTRIBUTE_NODE     =  2,
    wxXML_TEXT_NODE          =  3,
    wxXML_CDATA_SECTION_NODE =  4,
    wxXML_ENTITY_REF_NODE    =  5,
    wxXML_ENTITY_NODE        =  6,
    wxXML_PI_NODE            =  7,
    wxXML_COMMENT_NODE       =  8,
    wxXML_DOCUMENT_NODE      =  9,
    wxXML_DOCUMENT_TYPE_NODE = 10,
    wxXML_DOCUMENT_FRAG_NODE = 11,
    wxXML_NOTATION_NODE      = 12,
    wxXML_HTML_DOCUMENT_NODE = 13,

    wxXML_NO_INDENTATION    = (-1),

    wxXMLDOC_NONE = 0,
    wxXMLDOC_KEEP_WHITESPACE_NODES = 1,

    wxAUI_INSERT_PANE = 0,
    wxAUI_INSERT_ROW = 1,
    wxAUI_INSERT_DOCK = 2,

    wxAUI_NB_TOP                 = #0001,   --1 << 0,
    wxAUI_NB_LEFT                = #0002,   --1 << 1,  // not implemented yet
    wxAUI_NB_RIGHT               = #0004,   --1 << 2,  // not implemented yet
    wxAUI_NB_BOTTOM              = #0008,   --1 << 3,  // not implemented yet
    wxAUI_NB_TAB_SPLIT           = #0010,   --1 << 4,
    wxAUI_NB_TAB_MOVE            = #0020,   --1 << 5,
    wxAUI_NB_TAB_EXTERNAL_MOVE   = #0040,   --1 << 6,
    wxAUI_NB_TAB_FIXED_WIDTH     = #0080,   --1 << 7,
    wxAUI_NB_SCROLL_BUTTONS      = #0100,   --1 << 8,
    wxAUI_NB_WINDOWLIST_BUTTON   = #0200,   --1 << 9,
    wxAUI_NB_CLOSE_BUTTON        = #0400,   --1 << 10,
    wxAUI_NB_CLOSE_ON_ACTIVE_TAB = #0800,   --1 << 11,
    wxAUI_NB_CLOSE_ON_ALL_TABS   = #1000,   --1 << 12,

    wxAUI_NB_DEFAULT_STYLE = wx_or_all( wxAUI_NB_TOP &
                             wxAUI_NB_TAB_SPLIT &
                             wxAUI_NB_TAB_MOVE &
                             wxAUI_NB_SCROLL_BUTTONS &
                             wxAUI_NB_CLOSE_ON_ACTIVE_TAB ),
    wxAUI_DOCK_NONE = 0,
    wxAUI_DOCK_TOP = 1,
    wxAUI_DOCK_RIGHT = 2,
    wxAUI_DOCK_BOTTOM = 3,
    wxAUI_DOCK_LEFT = 4,
    wxAUI_DOCK_CENTER = 5,
    wxAUI_DOCK_CENTRE = wxAUI_DOCK_CENTER,

    wxAUI_MGR_ALLOW_FLOATING           = #01,   --1 << 0,
    wxAUI_MGR_ALLOW_ACTIVE_PANE        = #02,   --1 << 1,
    wxAUI_MGR_TRANSPARENT_DRAG         = #04,   --1 << 2,
    wxAUI_MGR_TRANSPARENT_HINT         = #08,   --1 << 3,
    wxAUI_MGR_VENETIAN_BLINDS_HINT     = #10,   --1 << 4,
    wxAUI_MGR_RECTANGLE_HINT           = #20,   --1 << 5,
    wxAUI_MGR_HINT_FADE                = #40,   --1 << 6,
    wxAUI_MGR_NO_VENETIAN_BLINDS_FADE  = #80,   --1 << 7,

    wxAUI_MGR_DEFAULT = wx_or_all( wxAUI_MGR_ALLOW_FLOATING &
                        wxAUI_MGR_TRANSPARENT_HINT &
                        wxAUI_MGR_HINT_FADE &
                        wxAUI_MGR_NO_VENETIAN_BLINDS_FADE ),

    wxAUI_DOCKART_SASH_SIZE                         =  0,
    wxAUI_DOCKART_CAPTION_SIZE                      =  1,
    wxAUI_DOCKART_GRIPPER_SIZE                      =  2,
    wxAUI_DOCKART_PANE_BORDER_SIZE                  =  3,
    wxAUI_DOCKART_PANE_BUTTON_SIZE                  =  4,
    wxAUI_DOCKART_BACKGROUND_COLOUR                 =  5,
    wxAUI_DOCKART_SASH_COLOUR                       =  6,
    wxAUI_DOCKART_ACTIVE_CAPTION_COLOUR             =  7,
    wxAUI_DOCKART_ACTIVE_CAPTION_GRADIENT_COLOUR    =  8,
    wxAUI_DOCKART_INACTIVE_CAPTION_COLOUR           =  9,
    wxAUI_DOCKART_INACTIVE_CAPTION_GRADIENT_COLOUR  = 10,
    wxAUI_DOCKART_ACTIVE_CAPTION_TEXT_COLOUR        = 11,
    wxAUI_DOCKART_INACTIVE_CAPTION_TEXT_COLOUR      = 12,
    wxAUI_DOCKART_BORDER_COLOUR                     = 13,
    wxAUI_DOCKART_GRIPPER_COLOUR                    = 14,
    wxAUI_DOCKART_CAPTION_FONT                      = 15,
    wxAUI_DOCKART_GRADIENT_TYPE                     = 16,

    wxAUI_GRADIENT_NONE         = 0,
    wxAUI_GRADIENT_VERTICAL     = 1,
    wxAUI_GRADIENT_HORIZONTAL   = 2,

    wxAUI_BUTTON_STATE_NORMAL   = #00,
    wxAUI_BUTTON_STATE_HOVER    = #01,  --1 << 1,
    wxAUI_BUTTON_STATE_PRESSED  = #02,  --1 << 2,
    wxAUI_BUTTON_STATE_DISABLED = #04,  --1 << 3,
    wxAUI_BUTTON_STATE_HIDDEN   = #08,  --1 << 4,
    wxAUI_BUTTON_STATE_CHECKED  = #10,  --1 << 5

    wxAUI_BUTTON_CLOSE              = 101,
    wxAUI_BUTTON_MAXIMIZE_RESTORE   = 102,
    wxAUI_BUTTON_MINIMIZE           = 103,
    wxAUI_BUTTON_PIN                = 104,
    wxAUI_BUTTON_OPTIONS            = 105,
    wxAUI_BUTTON_WINDOWLIST         = 106,
    wxAUI_BUTTON_LEFT               = 107,
    wxAUI_BUTTON_RIGHT              = 108,
    wxAUI_BUTTON_UP                 = 109,
    wxAUI_BUTTON_DOWN               = 110,
    wxAUI_BUTTON_CUSTOM1            = 201,
    wxAUI_BUTTON_CUSTOM2            = 202,
    wxAUI_BUTTON_CUSTOM3            = 203

public constant
    wxCURSOR_NONE               = 0,
    wxCURSOR_ARROW              = 1,
    wxCURSOR_RIGHT_ARROW        = 2,
    wxCURSOR_BULLSEYE           = 3,
    wxCURSOR_CHAR               = 4,
    wxCURSOR_CROSS              = 5,
    wxCURSOR_HAND               = 6,
    wxCURSOR_IBEAM              = 7,
    wxCURSOR_LEFT_BUTTON        = 8,
    wxCURSOR_MAGNIFIER          = 9,
    wxCURSOR_MIDDLE_BUTTON      = 10,
    wxCURSOR_NO_ENTRY           = 11,
    wxCURSOR_PAINT_BRUSH        = 12,
    wxCURSOR_PENCIL             = 13,
    wxCURSOR_POINT_LEFT         = 14,
    wxCURSOR_POINT_RIGHT        = 15,
    wxCURSOR_QUESTION_ARROW     = 16,
    wxCURSOR_RIGHT_BUTTON       = 17,
    wxCURSOR_SIZENESW           = 18,
    wxCURSOR_SIZENS             = 19,
    wxCURSOR_SIZENWSE           = 20,
    wxCURSOR_SIZEWE             = 21,
    wxCURSOR_SIZING             = 22,
    wxCURSOR_SPRAYCAN           = 23,
    wxCURSOR_WAIT               = 24,
    wxCURSOR_WATCH              = 25,
    wxCURSOR_BLANK              = 26

public constant
    wxEXEC_ASYNC                = 0,    -- execute the process asynchronously
    wxEXEC_SYNC                 = 1,    -- execute it synchronously, i.e. wait until it finishes
    wxEXEC_NOHIDE               = 2,    -- under Windows, don't hide the child even if it's IO is redirected (this
                                        -- is done by default)
    wxEXEC_MAKE_GROUP_LEADER    = 4     -- under Unix, if the process is the group leader then killing -pid kills
                                        -- all children as well as pid

public constant
    wxSIGNONE   =  0,   -- verify if the process exists under Unix
    wxSIGHUP    =  1,
    wxSIGINT    =  2,
    wxSIGQUIT   =  3,
    wxSIGILL    =  4,
    wxSIGTRAP   =  5,
    wxSIGABRT   =  6,
    wxSIGEMT    =  7,
    wxSIGFPE    =  8,
    wxSIGKILL   =  9,   -- forcefully kill, dangerous!
    wxSIGBUS    = 10,
    wxSIGSEGV   = 11,
    wxSIGSYS    = 12,
    wxSIGPIPE   = 13,
    wxSIGALRM   = 14,
    wxSIGTERM   = 15    -- terminate the process gently

public constant
    wxKILL_OK               = 1,    -- no error
    wxKILL_BAD_SIGNAL       = 2,    -- no such signal
    wxKILL_ACCESS_DENIED    = 3,    -- permission denied
    wxKILL_NO_PROCESS       = 4,    -- no such process
    wxKILL_ERROR            = 5     -- another, unspecified error

public constant
    wxFULLSCREEN_NOMENUBAR   = #0001,
    wxFULLSCREEN_NOTOOLBAR   = #0002,
    wxFULLSCREEN_NOSTATUSBAR = #0004,
    wxFULLSCREEN_NOBORDER    = #0008,
    wxFULLSCREEN_NOCAPTION   = #0010,

    wxFULLSCREEN_ALL = wx_or_all(
        wxFULLSCREEN_NOMENUBAR   &
        wxFULLSCREEN_NOTOOLBAR   &
        wxFULLSCREEN_NOSTATUSBAR &
        wxFULLSCREEN_NOBORDER    &
        wxFULLSCREEN_NOCAPTION   )

public constant
    wxEVT_SCROLL = {wxEVT_SCROLL_TOP, wxEVT_SCROLL_BOTTOM, wxEVT_SCROLL_LINEUP, wxEVT_SCROLL_LINEDOWN,
    wxEVT_SCROLL_PAGEUP, wxEVT_SCROLL_PAGEDOWN, wxEVT_SCROLL_THUMBTRACK, wxEVT_SCROLL_THUMBRELEASE}

public constant
    wxHW_SCROLLBAR_NEVER = #0002,
    wxHW_SCROLLBAR_AUTO  = #0004

public constant
--// from GSocket
    GSOCK_NOFAMILY      = 0,
    GSOCK_INET          = 1,
    GSOCK_INET6         = 2,
    GSOCK_UNIX          = 3,

    GSOCK_STREAMED      = 0,
    GSOCK_UNSTREAMED    = 1,

    GSOCK_NOERROR       = 0,
    GSOCK_INVOP         = 1,
    GSOCK_IOERR         = 2,
    GSOCK_INVADDR       = 3,
    GSOCK_INVSOCK       = 4,
    GSOCK_NOHOST        = 5,
    GSOCK_INVPORT       = 6,
    GSOCK_WOULDBLOCK    = 7,
    GSOCK_TIMEDOUT      = 8,
    GSOCK_MEMERR        = 9,

    GSOCK_INPUT         = 0,
    GSOCK_OUTPUT        = 1,
    GSOCK_CONNECTION    = 2,
    GSOCK_LOST          = 3,
    GSOCK_MAX_EVENT     = 4,

    GSOCK_INPUT_FLAG        = power( 2, GSOCK_INPUT ),
    GSOCK_OUTPUT_FLAG       = power( 2, GSOCK_OUTPUT ),
    GSOCK_CONNECTION_FLAG   = power( 2, GSOCK_CONNECTION ),
    GSOCK_LOST_FLAG         = power( 2, GSOCK_LOST ),

--enum wxSocketNotify
    wxSOCKET_INPUT              = GSOCK_INPUT,
    wxSOCKET_OUTPUT             = GSOCK_OUTPUT,
    wxSOCKET_CONNECTION         = GSOCK_CONNECTION,
    wxSOCKET_LOST               = GSOCK_LOST,

    wxSOCKET_INPUT_FLAG         = GSOCK_INPUT_FLAG,
    wxSOCKET_OUTPUT_FLAG        = GSOCK_OUTPUT_FLAG,
    wxSOCKET_CONNECTION_FLAG    = GSOCK_CONNECTION_FLAG,
    wxSOCKET_LOST_FLAG          = GSOCK_LOST_FLAG,

--enum wxSocketError
    wxSOCKET_NOERROR            = GSOCK_NOERROR,
    wxSOCKET_INVOP              = GSOCK_INVOP,
    wxSOCKET_IOERR              = GSOCK_IOERR,
    wxSOCKET_INVADDR            = GSOCK_INVADDR,
    wxSOCKET_INVSOCK            = GSOCK_INVSOCK,
    wxSOCKET_NOHOST             = GSOCK_NOHOST,
    wxSOCKET_INVPORT            = GSOCK_INVPORT,
    wxSOCKET_WOULDBLOCK         = GSOCK_WOULDBLOCK,
    wxSOCKET_TIMEDOUT           = GSOCK_TIMEDOUT,
    wxSOCKET_MEMERR             = GSOCK_MEMERR,

--// wxSocket-specific (not yet implemented)
    wxSOCKET_DUMMY              = 0,

    wxSOCKET_NONE               = 0,
    wxSOCKET_NOWAIT             = 1,
    wxSOCKET_WAITALL            = 2,
    wxSOCKET_BLOCK              = 4,

--enum wxSocketType
    wxSOCKET_UNINIT             = 0,
    wxSOCKET_CLIENT             = 1,
    wxSOCKET_SERVER             = 2,
    wxSOCKET_BASE               = 3,
    wxSOCKET_DATAGRAM           = 4

public constant wxNullBitmap = 0

public constant

-- * wxProgressDialog flags
    wxPD_CAN_ABORT          = #0001,
    wxPD_APP_MODAL          = #0002,
    wxPD_AUTO_HIDE          = #0004,
    wxPD_ELAPSED_TIME       = #0008,
    wxPD_ESTIMATED_TIME     = #0010,
    wxPD_REMAINING_TIME     = #0040,

-- * wxDirDialog styles
    wxDD_NEW_DIR_BUTTON     = #0080

public constant
    wxIMAGE_LIST_NORMAL = 0,  -- Normal icons
    wxIMAGE_LIST_SMALL  = 1,  -- Small icons
    wxIMAGE_LIST_STATE  = 2   -- State icons: unimplemented (see WIN32 documentation)

public constant
    BM_FROM_FILE = 1,
    BM_IN_MEMORY = 2,
    BM_XPM = 3,
    BM_FROM_IMAGE = 4,

    wxBITMAP_TYPE_INVALID       =  0,   --// should be == 0 for compatibility!
    wxBITMAP_TYPE_BMP           =  1,
    wxBITMAP_TYPE_BMP_RESOURCE  =  2,
    wxBITMAP_TYPE_RESOURCE      =  wxBITMAP_TYPE_BMP_RESOURCE,
    wxBITMAP_TYPE_ICO           =  3,
    wxBITMAP_TYPE_ICO_RESOURCE  =  4,
    wxBITMAP_TYPE_CUR           =  5,
    wxBITMAP_TYPE_CUR_RESOURCE  =  6,
    wxBITMAP_TYPE_XBM           =  7,
    wxBITMAP_TYPE_XBM_DATA      =  8,
    wxBITMAP_TYPE_XPM           =  9,
    wxBITMAP_TYPE_XPM_DATA      = 10,
    wxBITMAP_TYPE_TIF           = 11,
    wxBITMAP_TYPE_TIF_RESOURCE  = 12,
    wxBITMAP_TYPE_GIF           = 13,
    wxBITMAP_TYPE_GIF_RESOURCE  = 14,
    wxBITMAP_TYPE_PNG           = 15,
    wxBITMAP_TYPE_PNG_RESOURCE  = 16,
    wxBITMAP_TYPE_JPEG          = 17,
    wxBITMAP_TYPE_JPEG_RESOURCE = 18,
    wxBITMAP_TYPE_PNM           = 19,
    wxBITMAP_TYPE_PNM_RESOURCE  = 20,
    wxBITMAP_TYPE_PCX           = 21,
    wxBITMAP_TYPE_PCX_RESOURCE  = 22,
    wxBITMAP_TYPE_PICT          = 23,
    wxBITMAP_TYPE_PICT_RESOURCE = 24,
    wxBITMAP_TYPE_ICON          = 25,
    wxBITMAP_TYPE_ICON_RESOURCE = 26,
    wxBITMAP_TYPE_MACCURSOR     = 27,
    wxBITMAP_TYPE_MACCURSOR_RESOURCE = 28,
    wxBITMAP_TYPE_ANY           = 50

public constant
    wxITEM_SEPARATOR    = -1,
    wxITEM_NORMAL       =  0,
    wxITEM_CHECK        =  1,
    wxITEM_RADIO        =  2,
    wxITEM_MAX          =  3

public constant
    wxID_SEPARATOR  = (-1),

-- Standard menu IDs
    wxID_LOWEST                      = 4999,

    wxID_OPEN                        = 5000,
    wxID_CLOSE                       = 5001,
    wxID_NEW                         = 5002,
    wxID_SAVE                        = 5003,
    wxID_SAVEAS                      = 5004,
    wxID_REVERT                      = 5005,
    wxID_EXIT                        = 5006,
    wxID_UNDO                        = 5007,
    wxID_REDO                        = 5008,
    wxID_HELP                        = 5009,
    wxID_PRINT                       = 5010,
    wxID_PRINT_SETUP                 = 5011,
    wxID_PREVIEW                     = 5012,
    wxID_ABOUT                       = 5013,
    wxID_HELP_CONTENTS               = 5014,
    wxID_HELP_COMMANDS               = 5015,
    wxID_HELP_PROCEDURES             = 5016,
    wxID_HELP_CONTEXT                = 5017,

    wxID_CUT                         = 5030,
    wxID_COPY                        = 5031,
    wxID_PASTE                       = 5032,
    wxID_CLEAR                       = 5033,
    wxID_FIND                        = 5034,
    wxID_DUPLICATE                   = 5035,
    wxID_SELECTALL                   = 5036,

    wxID_FILE1                       = 5050,
    wxID_FILE2                       = 5051,
    wxID_FILE3                       = 5052,
    wxID_FILE4                       = 5053,
    wxID_FILE5                       = 5054,
    wxID_FILE6                       = 5055,
    wxID_FILE7                       = 5056,
    wxID_FILE8                       = 5057,
    wxID_FILE9                       = 5058,

-- Standard button IDs
    wxID_OK                          = 5100,
    wxID_CANCEL                      = 5101,
    wxID_APPLY                       = 5102,
    wxID_YES                         = 5103,
    wxID_NO                          = 5104,
    wxID_STATIC                      = 5105,
    wxID_FORWARD                     = 5106,
    wxID_BACKWARD                    = 5107,
    wxID_DEFAULT                     = 5108,
    wxID_MORE                        = 5109,
    wxID_SETUP                       = 5110,
    wxID_RESET                       = 5111,

-- IDs used by generic file dialog (11 consecutive starting from this value)
    wxID_FILEDLGG                    = 5900,
    wxID_HIGHEST                     = 5999,

    wxVSCROLL                       = #80000000,
    wxHSCROLL                       = #40000000,
    wxCAPTION                       = #20000000,

-- New styles
    wxDOUBLE_BORDER                 = #10000000,
    wxSUNKEN_BORDER                 = #08000000,
    wxRAISED_BORDER                 = #04000000,
    wxBORDER                        = #02000000,
    wxSIMPLE_BORDER                 = wxBORDER,
    wxSTATIC_BORDER                 = #01000000,
    wxTRANSPARENT_WINDOW            = #00100000,
    wxNO_BORDER                     = #00200000,

    wxNO_3D                         = #00800000,    -- Override CTL3D etc. control colour processing to allow own background colour.
                                                    -- Override CTL3D or native 3D styles for children
    wxCLIP_CHILDREN                 = #00400000,    -- Clip children when painting, which reduces flicker in e.g. frames and
                                                    -- splitter windows, but can't be used in a panel where a static box must be
                                                    -- 'transparent' (panel paints the background for it)
    wxTAB_TRAVERSAL                 = #00080000,    -- Add this style to a panel to get tab traversal working outside of dialogs
                                                    -- (on by default for wxPanel, wxDialog, wxScrolledWindow)
    wxWANTS_CHARS                   = #00040000,    -- Add this style if the control wants to get all keyboard messages (under
                                                    -- Windows, it won't normally get the dialog navigation key events)
    wxRETAINED                      = #00020000,    -- Make window retained (mostly Motif, I think)
    wxBACKINGSTORE                  = wxRETAINED,
    wxNO_FULL_REPAINT_ON_RESIZE     = #00010000,    -- Don't invalidate the whole window (resulting in a PAINT event) when the
                                                    -- window is resized (currently, makes sense for wxMSW only)
    wxWS_EX_VALIDATE_RECURSIVELY    = #00000001,    -- * Extra window style flags (use wxWS_EX prefix to make it clear that they
                                                    -- * should be passed to wxWindow::SetExtraStyle(), not SetWindowStyle())
                                                    -- by default, TransferDataTo/FromWindow() only work on direct children of the
                                                    -- window (compatible behaviour), set this flag to make them recursively
                                                    -- descend into all subwindows
-- * wxFrame/wxDialog style flags
    wxSTAY_ON_TOP                   = #8000,
    wxICONIZE                       = #4000,
    wxMINIMIZE                      = wxICONIZE,
    wxMAXIMIZE                      = #2000,
    wxCLOSE_BOX                     = #1000,
    wxSYSTEM_MENU                   = #0800,
    wxMINIMIZE_BOX                  = #0400,
    wxMAXIMIZE_BOX                  = #0200,
    wxTINY_CAPTION_HORIZ            = #0100,
    wxTINY_CAPTION_VERT             = #0080,
    wxRESIZE_BORDER                 = #0040,

-- Add for normal Windows frame behaviour
    wxDIALOG_NO_PARENT      = #0001,  --// Don't make owned by apps top window
    wxFRAME_NO_TASKBAR      = #0002,  --// No taskbar button (MSW only)
    wxFRAME_TOOL_WINDOW     = #0004,  --// No taskbar button, no system menu
    wxFRAME_FLOAT_ON_PARENT = #0008,  --// Always above its parent
    wxFRAME_SHAPED          = #0010,  --// Create a window that is able to be shaped

    wxFRAME_NO_WINDOW_MENU          = #0100,    -- * MDI parent frame style flags
                                                -- * Can overlap with some of the above.

    wxDEFAULT_FRAME_STYLE = wx_or_all(
        wxSYSTEM_MENU   &
        wxRESIZE_BORDER &
        wxMINIMIZE_BOX  &
        wxMAXIMIZE_BOX  &
        wxCAPTION       &
        wxCLOSE_BOX     &
        wxCLIP_CHILDREN ),

--  Under Unix, the dialogs don't have a system menu. Specifying wxSYSTEM_MENU
--  here will make a close button appear.
--#   define wxDEFAULT_DIALOG_STYLE  wxCAPTION
    wxDEFAULT_DIALOG_STYLE  = wx_or_all({wxSYSTEM_MENU, wxCAPTION}),

-- * wxExtDialog style flags
    wxED_CLIENT_MARGIN              = #0004,
    wxED_BUTTONS_BOTTOM             = #0000,    -- has no effect
    wxED_BUTTONS_RIGHT              = #0002,
    wxED_STATIC_LINE                = #0001,

--#if defined(__WXMSW__) || defined(__WXMAC__)
    wxEXT_DIALOG_STYLE = wx_or_all( wxDEFAULT_DIALOG_STYLE & wxED_CLIENT_MARGIN ),
--#else
--#   define wxEXT_DIALOG_STYLE  (wxDEFAULT_DIALOG_STYLE|wxED_CLIENT_MARGIN|wxED_STATIC_LINE)

    wxOK                            = #00000004,
    wxYES_NO                        = #0000000A,
    wxCANCEL                        = #00000010,
    wxYES                           = #00000002,
    wxNO                            = #00000008,
    wxNO_DEFAULT                    = #00000080,
    wxYES_DEFAULT                   = #00000000,  -- has no effect
    wxICON_EXCLAMATION              = #00000100,
    wxICON_HAND                     = #00000200,
    wxICON_WARNING                  = wxICON_EXCLAMATION,
    wxICON_ERROR                    = wxICON_HAND,
    wxICON_QUESTION                 = #00000400,
    wxICON_INFORMATION              = #00000800,
    wxICON_STOP                     = wxICON_HAND,
    wxICON_ASTERISK                 = wxICON_INFORMATION,
    wxICON_MASK                     = wx_or_all( #00000100 & #00000200 & #00000400 & #00000800 ),

    wxLEFT                          = #0010,
    wxRIGHT                         = #0020,
    wxUP                            = #0040,
    wxDOWN                          = #0080,

    wxTOP                           = wxUP,
    wxBOTTOM                        = wxDOWN,

    wxNORTH                         = wxUP,
    wxSOUTH                         = wxDOWN,
    wxWEST                          = wxLEFT,
    wxEAST                          = wxRIGHT,

    wxALL                           = (wxUP + wxDOWN + wxRIGHT + wxLEFT),

    wxALIGN_NOT                     = #0000,
    wxALIGN_CENTER_HORIZONTAL       = #0100,
    wxALIGN_CENTRE_HORIZONTAL       = wxALIGN_CENTER_HORIZONTAL,
    wxALIGN_LEFT                    = wxALIGN_NOT,
    wxALIGN_TOP                     = wxALIGN_NOT,
    wxALIGN_RIGHT                   = #0200,
    wxALIGN_BOTTOM                  = #0400,
    wxALIGN_CENTER_VERTICAL         = #0800,
    wxALIGN_CENTRE_VERTICAL         = wxALIGN_CENTER_VERTICAL,
    wxALIGN_CENTER                  = wx_or_all( {wxALIGN_CENTER_HORIZONTAL, wxALIGN_CENTER_VERTICAL}),

    wxHORIZONTAL                    = #0004,
    wxVERTICAL                      = #0008,
    wxBOTH                          = wxVERTICAL + wxHORIZONTAL,

    wxSTRETCH_NOT                   = #0000,
    wxSHRINK                        = #1000,
    wxGROW                          = #2000,
    wxEXPAND                        = wxGROW,
    wxSHAPED                        = #4000,
    wxFIXED_MINSIZE                 = #8000,
    wxTILE                          = #C000,

    wxSYS_WHITE_BRUSH               =  0,
    wxSYS_LTGRAY_BRUSH              =  1,
    wxSYS_GRAY_BRUSH                =  2,
    wxSYS_DKGRAY_BRUSH              =  3,
    wxSYS_BLACK_BRUSH               =  4,
    wxSYS_NULL_BRUSH                =  5,
    wxSYS_HOLLOW_BRUSH              = wxSYS_NULL_BRUSH,
    wxSYS_WHITE_PEN                 =  6,
    wxSYS_BLACK_PEN                 =  7,
    wxSYS_NULL_PEN                  =  8,
    wxSYS_OEM_FIXED_FONT            = 10,
    wxSYS_ANSI_FIXED_FONT           = 11,
    wxSYS_ANSI_VAR_FONT             = 12,
    wxSYS_SYSTEM_FONT               = 13,
    wxSYS_DEVICE_DEFAULT_FONT       = 14,
    wxSYS_DEFAULT_PALETTE           = 15,
    wxSYS_SYSTEM_FIXED_FONT         = 16,
    wxSYS_DEFAULT_GUI_FONT          = 17,

    wxSYS_COLOUR_SCROLLBAR          =  0,
    wxSYS_COLOUR_BACKGROUND         =  1,
    wxSYS_COLOUR_ACTIVECAPTION      =  2,
    wxSYS_COLOUR_INACTIVECAPTION    =  3,
    wxSYS_COLOUR_MENU               =  4,
    wxSYS_COLOUR_WINDOW             =  5,
    wxSYS_COLOUR_WINDOWFRAME        =  6,
    wxSYS_COLOUR_MENUTEXT           =  7,
    wxSYS_COLOUR_WINDOWTEXT         =  8,
    wxSYS_COLOUR_CAPTIONTEXT        =  9,
    wxSYS_COLOUR_ACTIVEBORDER       = 10,
    wxSYS_COLOUR_INACTIVEBORDER     = 11,
    wxSYS_COLOUR_APPWORKSPACE       = 12,
    wxSYS_COLOUR_HIGHLIGHT          = 13,
    wxSYS_COLOUR_HIGHLIGHTTEXT      = 14,
    wxSYS_COLOUR_BTNFACE            = 15,
    wxSYS_COLOUR_BTNSHADOW          = 16,
    wxSYS_COLOUR_GRAYTEXT           = 17,
    wxSYS_COLOUR_BTNTEXT            = 18,
    wxSYS_COLOUR_INACTIVECAPTIONTEXT= 19,
    wxSYS_COLOUR_BTNHIGHLIGHT       = 20,
    wxSYS_COLOUR_3DDKSHADOW         = 21,
    wxSYS_COLOUR_3DLIGHT            = 22,
    wxSYS_COLOUR_INFOTEXT           = 23,
    wxSYS_COLOUR_INFOBK             = 24,
    wxSYS_COLOUR_LISTBOX            = 25,
    wxSYS_COLOUR_DESKTOP            = wxSYS_COLOUR_BACKGROUND,
    wxSYS_COLOUR_3DFACE             = wxSYS_COLOUR_BTNFACE,
    wxSYS_COLOUR_3DSHADOW           = wxSYS_COLOUR_BTNSHADOW,
    wxSYS_COLOUR_3DHIGHLIGHT        = wxSYS_COLOUR_BTNHIGHLIGHT,
    wxSYS_COLOUR_3DHILIGHT          = wxSYS_COLOUR_BTNHIGHLIGHT,
    wxSYS_COLOUR_BTNHILIGHT         = wxSYS_COLOUR_BTNHIGHLIGHT,

-- Metrics
    wxSYS_MOUSE_BUTTONS             =  1,
    wxSYS_BORDER_X                  =  2,
    wxSYS_BORDER_Y                  =  3,
    wxSYS_CURSOR_X                  =  4,
    wxSYS_CURSOR_Y                  =  5,
    wxSYS_DCLICK_X                  =  6,
    wxSYS_DCLICK_Y                  =  7,
    wxSYS_DRAG_X                    =  8,
    wxSYS_DRAG_Y                    =  9,
    wxSYS_EDGE_X                    = 10,
    wxSYS_EDGE_Y                    = 11,
    wxSYS_HSCROLL_ARROW_X           = 12,
    wxSYS_HSCROLL_ARROW_Y           = 13,
    wxSYS_HTHUMB_X                  = 14,
    wxSYS_ICON_X                    = 15,
    wxSYS_ICON_Y                    = 16,
    wxSYS_ICONSPACING_X             = 17,
    wxSYS_ICONSPACING_Y             = 18,
    wxSYS_WINDOWMIN_X               = 19,
    wxSYS_WINDOWMIN_Y               = 20,
    wxSYS_SCREEN_X                  = 21,
    wxSYS_SCREEN_Y                  = 22,
    wxSYS_FRAMESIZE_X               = 23,
    wxSYS_FRAMESIZE_Y               = 24,
    wxSYS_SMALLICON_X               = 25,
    wxSYS_SMALLICON_Y               = 26,
    wxSYS_HSCROLL_Y                 = 27,
    wxSYS_VSCROLL_X                 = 28,
    wxSYS_VSCROLL_ARROW_X           = 29,
    wxSYS_VSCROLL_ARROW_Y           = 30,
    wxSYS_VTHUMB_Y                  = 31,
    wxSYS_CAPTION_Y                 = 32,
    wxSYS_MENU_Y                    = 33,
    wxSYS_NETWORK_PRESENT           = 34,
    wxSYS_PENWINDOWS_PRESENT        = 35,
    wxSYS_SHOW_SOUNDS               = 36,
    wxSYS_SWAP_BUTTONS              = 37

public constant
    wxLC_VRULES          = #0001,
    wxLC_HRULES          = #0002,

    wxLC_ICON            = #0004,
    wxLC_SMALL_ICON      = #0008,
    wxLC_LIST            = #0010,
    wxLC_REPORT          = #0020,

    wxLC_ALIGN_TOP       = #0040,
    wxLC_ALIGN_LEFT      = #0080,
    wxLC_AUTOARRANGE     = #0100,
    wxLC_VIRTUAL         = #0200,
    wxLC_EDIT_LABELS     = #0400,
    wxLC_NO_HEADER       = #0800,
    wxLC_NO_SORT_HEADER  = #1000,
    wxLC_SINGLE_SEL      = #2000,
    wxLC_SORT_ASCENDING  = #4000,
    wxLC_SORT_DESCENDING = #8000,

    wxLC_MASK_TYPE       = wx_or_all(wxLC_ICON & wxLC_SMALL_ICON & wxLC_LIST & wxLC_REPORT),
    wxLC_MASK_ALIGN      = wx_or_all(wxLC_ALIGN_TOP & wxLC_ALIGN_LEFT),
    wxLC_MASK_SORT       = wx_or_all(wxLC_SORT_ASCENDING & wxLC_SORT_DESCENDING),

    wxLIST_MASK_STATE           = #0001,
    wxLIST_MASK_TEXT            = #0002,
    wxLIST_MASK_IMAGE           = #0004,
    wxLIST_MASK_DATA            = #0008,
    wxLIST_SET_ITEM             = #0010,
    wxLIST_MASK_WIDTH           = #0020,
    wxLIST_MASK_FORMAT          = #0040,

-- State flags for indicating the state of an item
    wxLIST_STATE_DONTCARE       = #0000,
    wxLIST_STATE_DROPHILITED    = #0001,      -- MSW only
    wxLIST_STATE_FOCUSED        = #0002,
    wxLIST_STATE_SELECTED       = #0004,
    wxLIST_STATE_CUT            = #0008,      -- MSW only

-- Hit test flags, used in HitTest
    wxLIST_HITTEST_ABOVE            = #0001,  -- Above the client area.
    wxLIST_HITTEST_BELOW            = #0002,  -- Below the client area.
    wxLIST_HITTEST_NOWHERE          = #0004,  -- In the client area but below the last item.
    wxLIST_HITTEST_ONITEMICON       = #0020,  -- On the bitmap associated with an item.
    wxLIST_HITTEST_ONITEMLABEL      = #0080,  -- On the label (string) associated with an item.
    wxLIST_HITTEST_ONITEMRIGHT      = #0100,  -- In the area to the right of an item.
    wxLIST_HITTEST_ONITEMSTATEICON  = #0200,  -- On the state icon for a tree view item that is in a user-defined state.
    wxLIST_HITTEST_TOLEFT           = #0400,  -- To the left of the client area.
    wxLIST_HITTEST_TORIGHT          = #0800,  -- To the right of the client area.

    wxLIST_HITTEST_ONITEM = wx_or_all(
        wxLIST_HITTEST_ONITEMICON      &
        wxLIST_HITTEST_ONITEMLABEL     &
        wxLIST_HITTEST_ONITEMSTATEICON ),

-- Flags for GetNextItem (MSW only except wxLIST_NEXT_ALL)
    wxLIST_NEXT_ABOVE               = 0,    -- Searches for an item above the specified item
    wxLIST_NEXT_ALL                 = 1,    -- Searches for subsequent item by index
    wxLIST_NEXT_BELOW               = 2,    -- Searches for an item below the specified item
    wxLIST_NEXT_LEFT                = 3,    -- Searches for an item to the left of the specified item
    wxLIST_NEXT_RIGHT               = 4,    -- Searches for an item to the right of the specified item

-- Alignment flags for Arrange (MSW only except wxLIST_ALIGN_LEFT)
    wxLIST_ALIGN_DEFAULT            = 0,
    wxLIST_ALIGN_LEFT               = 1,
    wxLIST_ALIGN_TOP                = 2,
    wxLIST_ALIGN_SNAP_TO_GRID       = 3,

-- Column format (MSW only except wxLIST_FORMAT_LEFT)
    wxLIST_FORMAT_LEFT              = 0,
    wxLIST_FORMAT_RIGHT             = 1,
    wxLIST_FORMAT_CENTRE            = 2,
    wxLIST_FORMAT_CENTER            = wxLIST_FORMAT_CENTRE,

-- Autosize values for SetColumnWidth
    wxLIST_AUTOSIZE                 = -1,
    wxLIST_AUTOSIZE_USEHEADER       = -2,    -- partly supported by generic version

-- Flag values for GetItemRect
    wxLIST_RECT_BOUNDS              = 1,
    wxLIST_RECT_ICON                = 2,
    wxLIST_RECT_LABEL               = 3,

-- Flag values for FindItem (MSW only)
    wxLIST_FIND_UP                  = 0,
    wxLIST_FIND_DOWN                = 1,
    wxLIST_FIND_LEFT                = 2,
    wxLIST_FIND_RIGHT               = 3

public constant
    wxLB_SORT           =  #0010,
    wxLB_SINGLE         =  #0020,
    wxLB_MULTIPLE       =  #0040,
    wxLB_EXTENDED       =  #0080,
    wxLB_OWNERDRAW      =  #0100,   -- Windows-only
    wxLB_NEEDED_SB      =  #0200,
    wxLB_ALWAYS_SB      =  #0400,
    wxLB_HSCROLL        =  wxHSCROLL,

    wxCB_SIMPLE         =  #0004,
    wxCB_SORT           =  #0008,
    wxCB_READONLY       =  #0010,
    wxCB_DROPDOWN       =  #0020

public constant

    wxTE_NO_VSCROLL         = #0002,
    wxTE_AUTO_SCROLL        = #0008,

    wxTE_READONLY           = #0010,
    wxTE_MULTILINE          = #0020,
    wxTE_PROCESS_TAB        = #0040,

--// alignment flags
    wxTE_LEFT               = #0000,                        --// #0000
    wxTE_CENTER             = wxALIGN_CENTER_HORIZONTAL,    --// #0100
    wxTE_RIGHT              = wxALIGN_RIGHT,                --// #0200
    wxTE_CENTRE             = wxTE_CENTER,

    wxTE_RICH               = #0080,                        --// this style means to use RICHEDIT control and does something only under wxMSW
                                                            --// and Win32 and is silently ignored under all other platforms

    wxTE_PROCESS_ENTER      = #0400,
    wxTE_PASSWORD           = #0800,

    wxTE_AUTO_URL           = #1000,                        --// automatically detect the URLs and generate the events when mouse is
                                                            --// moved/clicked over an URL
                                                            --//
                                                            --// this is for Win32 richedit controls only so far

    wxTE_NOHIDESEL          = #2000,                        --// by default, the Windows text control doesn't show the selection when it
                                                            --// doesn't have focus - use this style to force it to always show it

    wxTE_DONTWRAP           = wxHSCROLL,                    --// use wxHSCROLL to not wrap text at all, wxTE_LINEWRAP to wrap it at any
    wxTE_LINEWRAP           = #4000,                        --// position and wxTE_WORDWRAP to wrap at words boundary
    wxTE_WORDWRAP           = #0000,                        --// it's just == !wxHSCROLL

    wxTE_RICH2              = #8000                         --// force using RichEdit version 2.0 or 3.0 instead of 1.0 (default) for
                                                            --// wxTE_RICH controls - can be used together with or instead of wxTE_RICH

public constant
    wxSB_HORIZONTAL         = wxHORIZONTAL,
    wxSB_VERTICAL           = wxVERTICAL

public constant
    wxODDEVEN_RULE    = 1,
    wxWINDING_RULE    = 2,

    wxCLEAR         =  0,   wxROP_BLACK         = wxCLEAR,          wxBLIT_BLACKNESS    = wxCLEAR,          --// 0
    wxXOR           =  1,   wxROP_XORPEN        = wxXOR,            wxBLIT_SRCINVERT    = wxXOR,            --// src XOR dst
    wxINVERT        =  2,   wxROP_NOT           = wxINVERT,         wxBLIT_DSTINVERT    = wxINVERT,         --// NOT dst
    wxOR_REVERSE    =  3,   wxROP_MERGEPENNOT   = wxOR_REVERSE,     wxBLIT_00DD0228     = wxOR_REVERSE,     --// src OR (NOT dst)
    wxAND_REVERSE   =  4,   wxROP_MASKPENNOT    = wxAND_REVERSE,    wxBLIT_SRCERASE     = wxAND_REVERSE,    --// src AND (NOT dst)
    wxCOPY          =  5,   wxROP_COPYPEN       = wxCOPY,           wxBLIT_SRCCOPY      = wxCOPY,           --// src
    wxAND           =  6,   wxROP_MASKPEN       = wxAND,            wxBLIT_SRCAND       = wxAND,            --// src AND dst
    wxAND_INVERT    =  7,   wxROP_MASKNOTPEN    = wxAND_INVERT,     wxBLIT_00220326     = wxAND_INVERT,     --// (NOT src) AND dst
    wxNO_OP         =  8,   wxROP_NOP           = wxNO_OP,          wxBLIT_00AA0029     = wxNO_OP,          --// dst
    wxNOR           =  9,   wxROP_NOTMERGEPEN   = wxNOR,            wxBLIT_NOTSRCERASE  = wxNOR,            --// (NOT src) AND (NOT dst)
    wxEQUIV         = 10,   wxROP_NOTXORPEN     = wxEQUIV,          wxBLIT_00990066     = wxEQUIV,          --// (NOT src) XOR dst
    wxSRC_INVERT    = 11,   wxROP_NOTCOPYPEN    = wxSRC_INVERT,     wxBLIT_NOTSCRCOPY   = wxSRC_INVERT,     --// (NOT src)
    wxOR_INVERT     = 12,   wxROP_MERGENOTPEN   = wxOR_INVERT,      wxBLIT_MERGEPAINT   = wxOR_INVERT,      --// (NOT src) OR dst
    wxNAND          = 13,   wxROP_NOTMASKPEN    = wxNAND,           wxBLIT_007700E6     = wxNAND,           --// (NOT src) OR (NOT dst)
    wxOR            = 14,   wxROP_MERGEPEN      = wxOR,             wxBLIT_SRCPAINT     = wxOR,             --// src OR dst
    wxSET           = 15,   wxROP_WHITE         = wxSET,            wxBLIT_WHITENESS    = wxSET,            --//

    wxSOLID         = 100,
    wxDOT           = 101,
    wxLONG_DASH     = 102,
    wxSHORT_DASH    = 103,
    wxDOT_DASH      = 104,
    wxUSER_DASH     = 105,
    wxTRANSPARENT   = 106,

    wxDEFAULT       = 70,
    wxDECORATIVE    = 71,
    wxROMAN         = 72,
    wxSCRIPT        = 73,
    wxSWISS         = 74,
    wxMODERN        = 75,
    wxTELETYPE      = 76,   --/* @@@@ */

-- Proportional or Fixed width fonts (not yet used)
    wxVARIABLE      = 80,
    wxFIXED         = 81,

    wxNORMAL        = 90,
    wxLIGHT         = 91,
    wxBOLD          = 92,
-- Also wxNORMAL for normal (non-italic text)
    wxITALIC        = 93,
    wxSLANT         = 94,

    wxFONTFAMILY_DEFAULT        = wxDEFAULT,
    wxFONTFAMILY_DECORATIVE     = wxDECORATIVE,
    wxFONTFAMILY_ROMAN          = wxROMAN,
    wxFONTFAMILY_SCRIPT         = wxSCRIPT,
    wxFONTFAMILY_SWISS          = wxSWISS,
    wxFONTFAMILY_MODERN         = wxMODERN,
    wxFONTFAMILY_TELETYPE       = wxTELETYPE,
    wxFONTFAMILY_MAX            = 77,

-- font styles
    wxFONTSTYLE_NORMAL          = wxNORMAL,
    wxFONTSTYLE_ITALIC          = wxITALIC,
    wxFONTSTYLE_SLANT           = wxSLANT,
    wxFONTSTYLE_MAX             = wxSLANT + 1,

-- font weights
    wxFONTWEIGHT_NORMAL         = wxNORMAL,
    wxFONTWEIGHT_LIGHT          = wxLIGHT,
    wxFONTWEIGHT_BOLD           = wxBOLD,
    wxFONTWEIGHT_MAX            = wxBOLD + 1,
    $
public enum
        -- user's default/preffered language as got from OS:
    wxLANGUAGE_DEFAULT,
    -- unknown language, if wxLocale::GetSystemLanguage fails:
    wxLANGUAGE_UNKNOWN,

    wxLANGUAGE_ABKHAZIAN,
    wxLANGUAGE_AFAR,
    wxLANGUAGE_AFRIKAANS,
    wxLANGUAGE_ALBANIAN,
    wxLANGUAGE_AMHARIC,
    wxLANGUAGE_ARABIC,
    wxLANGUAGE_ARABIC_ALGERIA,
    wxLANGUAGE_ARABIC_BAHRAIN,
    wxLANGUAGE_ARABIC_EGYPT,
    wxLANGUAGE_ARABIC_IRAQ,
    wxLANGUAGE_ARABIC_JORDAN,
    wxLANGUAGE_ARABIC_KUWAIT,
    wxLANGUAGE_ARABIC_LEBANON,
    wxLANGUAGE_ARABIC_LIBYA,
    wxLANGUAGE_ARABIC_MOROCCO,
    wxLANGUAGE_ARABIC_OMAN,
    wxLANGUAGE_ARABIC_QATAR,
    wxLANGUAGE_ARABIC_SAUDI_ARABIA,
    wxLANGUAGE_ARABIC_SUDAN,
    wxLANGUAGE_ARABIC_SYRIA,
    wxLANGUAGE_ARABIC_TUNISIA,
    wxLANGUAGE_ARABIC_UAE,
    wxLANGUAGE_ARABIC_YEMEN,
    wxLANGUAGE_ARMENIAN,
    wxLANGUAGE_ASSAMESE,
    wxLANGUAGE_AYMARA,
    wxLANGUAGE_AZERI,
    wxLANGUAGE_AZERI_CYRILLIC,
    wxLANGUAGE_AZERI_LATIN,
    wxLANGUAGE_BASHKIR,
    wxLANGUAGE_BASQUE,
    wxLANGUAGE_BELARUSIAN,
    wxLANGUAGE_BENGALI,
    wxLANGUAGE_BHUTANI,
    wxLANGUAGE_BIHARI,
    wxLANGUAGE_BISLAMA,
    wxLANGUAGE_BRETON,
    wxLANGUAGE_BULGARIAN,
    wxLANGUAGE_BURMESE,
    wxLANGUAGE_CAMBODIAN,
    wxLANGUAGE_CATALAN,
    wxLANGUAGE_CHINESE,
    wxLANGUAGE_CHINESE_SIMPLIFIED,
    wxLANGUAGE_CHINESE_TRADITIONAL,
    wxLANGUAGE_CHINESE_HONGKONG,
    wxLANGUAGE_CHINESE_MACAU,
    wxLANGUAGE_CHINESE_SINGAPORE,
    wxLANGUAGE_CHINESE_TAIWAN,
    wxLANGUAGE_CORSICAN,
    wxLANGUAGE_CROATIAN,
    wxLANGUAGE_CZECH,
    wxLANGUAGE_DANISH,
    wxLANGUAGE_DUTCH,
    wxLANGUAGE_DUTCH_BELGIAN,
    wxLANGUAGE_ENGLISH,
    wxLANGUAGE_ENGLISH_UK,
    wxLANGUAGE_ENGLISH_US,
    wxLANGUAGE_ENGLISH_AUSTRALIA,
    wxLANGUAGE_ENGLISH_BELIZE,
    wxLANGUAGE_ENGLISH_BOTSWANA,
    wxLANGUAGE_ENGLISH_CANADA,
    wxLANGUAGE_ENGLISH_CARIBBEAN,
    wxLANGUAGE_ENGLISH_DENMARK,
    wxLANGUAGE_ENGLISH_EIRE,
    wxLANGUAGE_ENGLISH_JAMAICA,
    wxLANGUAGE_ENGLISH_NEW_ZEALAND,
    wxLANGUAGE_ENGLISH_PHILIPPINES,
    wxLANGUAGE_ENGLISH_SOUTH_AFRICA,
    wxLANGUAGE_ENGLISH_TRINIDAD,
    wxLANGUAGE_ENGLISH_ZIMBABWE,
    wxLANGUAGE_ESPERANTO,
    wxLANGUAGE_ESTONIAN,
    wxLANGUAGE_FAEROESE,
    wxLANGUAGE_FARSI,
    wxLANGUAGE_FIJI,
    wxLANGUAGE_FINNISH,
    wxLANGUAGE_FRENCH,
    wxLANGUAGE_FRENCH_BELGIAN,
    wxLANGUAGE_FRENCH_CANADIAN,
    wxLANGUAGE_FRENCH_LUXEMBOURG,
    wxLANGUAGE_FRENCH_MONACO,
    wxLANGUAGE_FRENCH_SWISS,
    wxLANGUAGE_FRISIAN,
    wxLANGUAGE_GALICIAN,
    wxLANGUAGE_GEORGIAN,
    wxLANGUAGE_GERMAN,
    wxLANGUAGE_GERMAN_AUSTRIAN,
    wxLANGUAGE_GERMAN_BELGIUM,
    wxLANGUAGE_GERMAN_LIECHTENSTEIN,
    wxLANGUAGE_GERMAN_LUXEMBOURG,
    wxLANGUAGE_GERMAN_SWISS,
    wxLANGUAGE_GREEK,
    wxLANGUAGE_GREENLANDIC,
    wxLANGUAGE_GUARANI,
    wxLANGUAGE_GUJARATI,
    wxLANGUAGE_HAUSA,
    wxLANGUAGE_HEBREW,
    wxLANGUAGE_HINDI,
    wxLANGUAGE_HUNGARIAN,
    wxLANGUAGE_ICELANDIC,
    wxLANGUAGE_INDONESIAN,
    wxLANGUAGE_INTERLINGUA,
    wxLANGUAGE_INTERLINGUE,
    wxLANGUAGE_INUKTITUT,
    wxLANGUAGE_INUPIAK,
    wxLANGUAGE_IRISH,
    wxLANGUAGE_ITALIAN,
    wxLANGUAGE_ITALIAN_SWISS,
    wxLANGUAGE_JAPANESE,
    wxLANGUAGE_JAVANESE,
    wxLANGUAGE_KANNADA,
    wxLANGUAGE_KASHMIRI,
    wxLANGUAGE_KASHMIRI_INDIA,
    wxLANGUAGE_KAZAKH,
    wxLANGUAGE_KERNEWEK,
    wxLANGUAGE_KINYARWANDA,
    wxLANGUAGE_KIRGHIZ,
    wxLANGUAGE_KIRUNDI,
    wxLANGUAGE_KONKANI,
    wxLANGUAGE_KOREAN,
    wxLANGUAGE_KURDISH,
    wxLANGUAGE_LAOTHIAN,
    wxLANGUAGE_LATIN,
    wxLANGUAGE_LATVIAN,
    wxLANGUAGE_LINGALA,
    wxLANGUAGE_LITHUANIAN,
    wxLANGUAGE_MACEDONIAN,
    wxLANGUAGE_MALAGASY,
    wxLANGUAGE_MALAY,
    wxLANGUAGE_MALAYALAM,
    wxLANGUAGE_MALAY_BRUNEI_DARUSSALAM,
    wxLANGUAGE_MALAY_MALAYSIA,
    wxLANGUAGE_MALTESE,
    wxLANGUAGE_MANIPURI,
    wxLANGUAGE_MAORI,
    wxLANGUAGE_MARATHI,
    wxLANGUAGE_MOLDAVIAN,
    wxLANGUAGE_MONGOLIAN,
    wxLANGUAGE_NAURU,
    wxLANGUAGE_NEPALI,
    wxLANGUAGE_NEPALI_INDIA,
    wxLANGUAGE_NORWEGIAN_BOKMAL,
    wxLANGUAGE_NORWEGIAN_NYNORSK,
    wxLANGUAGE_OCCITAN,
    wxLANGUAGE_ORIYA,
    wxLANGUAGE_OROMO,
    wxLANGUAGE_PASHTO,
    wxLANGUAGE_POLISH,
    wxLANGUAGE_PORTUGUESE,
    wxLANGUAGE_PORTUGUESE_BRAZILIAN,
    wxLANGUAGE_PUNJABI,
    wxLANGUAGE_QUECHUA,
    wxLANGUAGE_RHAETO_ROMANCE,
    wxLANGUAGE_ROMANIAN,
    wxLANGUAGE_RUSSIAN,
    wxLANGUAGE_RUSSIAN_UKRAINE,
    wxLANGUAGE_SAMOAN,
    wxLANGUAGE_SANGHO,
    wxLANGUAGE_SANSKRIT,
    wxLANGUAGE_SCOTS_GAELIC,
    wxLANGUAGE_SERBIAN,
    wxLANGUAGE_SERBIAN_CYRILLIC,
    wxLANGUAGE_SERBIAN_LATIN,
    wxLANGUAGE_SERBO_CROATIAN,
    wxLANGUAGE_SESOTHO,
    wxLANGUAGE_SETSWANA,
    wxLANGUAGE_SHONA,
    wxLANGUAGE_SINDHI,
    wxLANGUAGE_SINHALESE,
    wxLANGUAGE_SISWATI,
    wxLANGUAGE_SLOVAK,
    wxLANGUAGE_SLOVENIAN,
    wxLANGUAGE_SOMALI,
    wxLANGUAGE_SPANISH,
    wxLANGUAGE_SPANISH_ARGENTINA,
    wxLANGUAGE_SPANISH_BOLIVIA,
    wxLANGUAGE_SPANISH_CHILE,
    wxLANGUAGE_SPANISH_COLOMBIA,
    wxLANGUAGE_SPANISH_COSTA_RICA,
    wxLANGUAGE_SPANISH_DOMINICAN_REPUBLIC,
    wxLANGUAGE_SPANISH_ECUADOR,
    wxLANGUAGE_SPANISH_EL_SALVADOR,
    wxLANGUAGE_SPANISH_GUATEMALA,
    wxLANGUAGE_SPANISH_HONDURAS,
    wxLANGUAGE_SPANISH_MEXICAN,
    wxLANGUAGE_SPANISH_MODERN,
    wxLANGUAGE_SPANISH_NICARAGUA,
    wxLANGUAGE_SPANISH_PANAMA,
    wxLANGUAGE_SPANISH_PARAGUAY,
    wxLANGUAGE_SPANISH_PERU,
    wxLANGUAGE_SPANISH_PUERTO_RICO,
    wxLANGUAGE_SPANISH_URUGUAY,
    wxLANGUAGE_SPANISH_US,
    wxLANGUAGE_SPANISH_VENEZUELA,
    wxLANGUAGE_SUNDANESE,
    wxLANGUAGE_SWAHILI,
    wxLANGUAGE_SWEDISH,
    wxLANGUAGE_SWEDISH_FINLAND,
    wxLANGUAGE_TAGALOG,
    wxLANGUAGE_TAJIK,
    wxLANGUAGE_TAMIL,
    wxLANGUAGE_TATAR,
    wxLANGUAGE_TELUGU,
    wxLANGUAGE_THAI,
    wxLANGUAGE_TIBETAN,
    wxLANGUAGE_TIGRINYA,
    wxLANGUAGE_TONGA,
    wxLANGUAGE_TSONGA,
    wxLANGUAGE_TURKISH,
    wxLANGUAGE_TURKMEN,
    wxLANGUAGE_TWI,
    wxLANGUAGE_UIGHUR,
    wxLANGUAGE_UKRAINIAN,
    wxLANGUAGE_URDU,
    wxLANGUAGE_URDU_INDIA,
    wxLANGUAGE_URDU_PAKISTAN,
    wxLANGUAGE_UZBEK,
    wxLANGUAGE_UZBEK_CYRILLIC,
    wxLANGUAGE_UZBEK_LATIN,
    wxLANGUAGE_VIETNAMESE,
    wxLANGUAGE_VOLAPUK,
    wxLANGUAGE_WELSH,
    wxLANGUAGE_WOLOF,
    wxLANGUAGE_XHOSA,
    wxLANGUAGE_YIDDISH,
    wxLANGUAGE_YORUBA,
    wxLANGUAGE_ZHUANG,
    wxLANGUAGE_ZULU,
    $

public enum
 wxFONTENCODING_SYSTEM = -1,     -- system default
    wxFONTENCODING_DEFAULT,         -- current default encoding

    -- ISO8859 standard defines a number of single-byte charsets
    wxFONTENCODING_ISO8859_1,       -- West European (Latin1)
    wxFONTENCODING_ISO8859_2,       -- Central and East European (Latin2)
    wxFONTENCODING_ISO8859_3,       -- Esperanto (Latin3)
    wxFONTENCODING_ISO8859_4,       -- Baltic (old) (Latin4)
    wxFONTENCODING_ISO8859_5,       -- Cyrillic
    wxFONTENCODING_ISO8859_6,       -- Arabic
    wxFONTENCODING_ISO8859_7,       -- Greek
    wxFONTENCODING_ISO8859_8,       -- Hebrew
    wxFONTENCODING_ISO8859_9,       -- Turkish (Latin5)
    wxFONTENCODING_ISO8859_10,      -- Variation of Latin4 (Latin6)
    wxFONTENCODING_ISO8859_11,      -- Thai
    wxFONTENCODING_ISO8859_12,      -- doesn't exist currently, but put it
                                    -- here anyhow to make all ISO8859
                                    -- consecutive numbers
    wxFONTENCODING_ISO8859_13,      -- Baltic (Latin7)
    wxFONTENCODING_ISO8859_14,      -- Latin8
    wxFONTENCODING_ISO8859_15,      -- Latin9 (a.k.a. Latin0, includes euro)
    wxFONTENCODING_ISO8859_MAX,

    -- Cyrillic charset soup (see http://czyborra.com/charsets/cyrillic.html)
    wxFONTENCODING_KOI8,            -- KOI8 Russian
    wxFONTENCODING_KOI8_U,          -- KOI8 Ukrainian
    wxFONTENCODING_ALTERNATIVE,     -- same as MS-DOS CP866
    wxFONTENCODING_BULGARIAN,       -- used under Linux in Bulgaria

    -- what would we do without Microsoft? They have their own encodings
        -- for DOS
    wxFONTENCODING_CP437,           -- original MS-DOS codepage
    wxFONTENCODING_CP850,           -- CP437 merged with Latin1
    wxFONTENCODING_CP852,           -- CP437 merged with Latin2
    wxFONTENCODING_CP855,           -- another cyrillic encoding
    wxFONTENCODING_CP866,           -- and another one
        -- and for Windows
    wxFONTENCODING_CP874,           -- WinThai
    wxFONTENCODING_CP932,           -- Japanese (shift-JIS)
    wxFONTENCODING_CP936,           -- Chinese simplified (GB)
    wxFONTENCODING_CP949,           -- Korean (Hangul charset)
    wxFONTENCODING_CP950,           -- Chinese (traditional - Big5)
    wxFONTENCODING_CP1250,          -- WinLatin2
    wxFONTENCODING_CP1251,          -- WinCyrillic
    wxFONTENCODING_CP1252,          -- WinLatin1
    wxFONTENCODING_CP1253,          -- WinGreek (8859-7)
    wxFONTENCODING_CP1254,          -- WinTurkish
    wxFONTENCODING_CP1255,          -- WinHebrew
    wxFONTENCODING_CP1256,          -- WinArabic
    wxFONTENCODING_CP1257,          -- WinBaltic (same as Latin 7)
    wxFONTENCODING_CP12_MAX,

    wxFONTENCODING_UTF7,            -- UTF-7 Unicode encoding
    wxFONTENCODING_UTF8,            -- UTF-8 Unicode encoding
    wxFONTENCODING_EUC_JP,          -- Extended Unix Codepage for Japanese
    wxFONTENCODING_UTF16BE,         -- UTF-16 Big Endian Unicode encoding
    wxFONTENCODING_UTF16LE,         -- UTF-16 Little Endian Unicode encoding
    wxFONTENCODING_UTF32BE,         -- UTF-32 Big Endian Unicode encoding
    wxFONTENCODING_UTF32LE,         -- UTF-32 Little Endian Unicode encoding

    wxFONTENCODING_MACROMAN,        -- the standard mac encodings
    wxFONTENCODING_MACJAPANESE,
    wxFONTENCODING_MACCHINESETRAD,
    wxFONTENCODING_MACKOREAN,
    wxFONTENCODING_MACARABIC,
    wxFONTENCODING_MACHEBREW,
    wxFONTENCODING_MACGREEK,
    wxFONTENCODING_MACCYRILLIC,
    wxFONTENCODING_MACDEVANAGARI,
    wxFONTENCODING_MACGURMUKHI,
    wxFONTENCODING_MACGUJARATI,
    wxFONTENCODING_MACORIYA,
    wxFONTENCODING_MACBENGALI,
    wxFONTENCODING_MACTAMIL,
    wxFONTENCODING_MACTELUGU,
    wxFONTENCODING_MACKANNADA,
    wxFONTENCODING_MACMALAJALAM,
    wxFONTENCODING_MACSINHALESE,
    wxFONTENCODING_MACBURMESE,
    wxFONTENCODING_MACKHMER,
    wxFONTENCODING_MACTHAI,
    wxFONTENCODING_MACLAOTIAN,
    wxFONTENCODING_MACGEORGIAN,
    wxFONTENCODING_MACARMENIAN,
    wxFONTENCODING_MACCHINESESIMP,
    wxFONTENCODING_MACTIBETAN,
    wxFONTENCODING_MACMONGOLIAN,
    wxFONTENCODING_MACETHIOPIC,
    wxFONTENCODING_MACCENTRALEUR,
    wxFONTENCODING_MACVIATNAMESE,
    wxFONTENCODING_MACARABICEXT,
    wxFONTENCODING_MACSYMBOL,
    wxFONTENCODING_MACDINGBATS,
    wxFONTENCODING_MACTURKISH,
    wxFONTENCODING_MACCROATIAN,
    wxFONTENCODING_MACICELANDIC,
    wxFONTENCODING_MACROMANIAN,
    wxFONTENCODING_MACCELTIC,
    wxFONTENCODING_MACGAELIC,
    wxFONTENCODING_MACKEYBOARD,

    wxFONTENCODING_MAX,             -- highest enumerated encoding value
        $

public constant

    WXK_BACK        =   8,
    WXK_TAB         =   9,
    WXK_RETURN      =  13,
    WXK_ESCAPE      =  27,
    WXK_SPACE       =  32,
    WXK_DELETE      = 127,

    WXK_START       = 300,
    WXK_LBUTTON     = 301,
    WXK_RBUTTON     = 302,
    WXK_CANCEL      = 303,
    WXK_MBUTTON     = 304,
    WXK_CLEAR       = 305,
    WXK_SHIFT       = 306,
    WXK_ALT         = 307,
    WXK_CONTROL     = 308,
    WXK_MENU        = 309,
    WXK_PAUSE       = 310,
    WXK_CAPITAL     = 311,

    WXK_END         = 312,
    WXK_HOME        = 313,
    WXK_LEFT        = 314,
    WXK_UP          = 315,
    WXK_RIGHT       = 316,
    WXK_DOWN        = 317,
    WXK_SELECT      = 318,
    WXK_PRINT       = 319,
    WXK_EXECUTE     = 320,
    WXK_SNAPSHOT    = 321,
    WXK_INSERT      = 322,
    WXK_HELP        = 323,
    WXK_NUMPAD0     = 324,
    WXK_NUMPAD1     = 325,
    WXK_NUMPAD2     = 326,
    WXK_NUMPAD3     = 327,
    WXK_NUMPAD4     = 328,
    WXK_NUMPAD5     = 329,
    WXK_NUMPAD6     = 330,
    WXK_NUMPAD7     = 331,
    WXK_NUMPAD8     = 332,
    WXK_NUMPAD9     = 333,
    WXK_MULTIPLY    = 334,
    WXK_ADD         = 335,
    WXK_SEPARATOR   = 336,
    WXK_SUBTRACT    = 337,
    WXK_DECIMAL     = 338,
    WXK_DIVIDE      = 339,
    WXK_F1          = 340,
    WXK_F2          = 341,
    WXK_F3          = 342,
    WXK_F4          = 343,
    WXK_F5          = 344,
    WXK_F6          = 345,
    WXK_F7          = 346,
    WXK_F8          = 347,
    WXK_F9          = 348,
    WXK_F10         = 349,
    WXK_F11         = 350,
    WXK_F12         = 351,
    WXK_F13         = 352,
    WXK_F14         = 353,
    WXK_F15         = 354,
    WXK_F16         = 355,
    WXK_F17         = 356,
    WXK_F18         = 357,
    WXK_F19         = 358,
    WXK_F20         = 359,
    WXK_F21         = 360,
    WXK_F22         = 361,
    WXK_F23         = 362,
    WXK_F24         = 363,
    WXK_NUMLOCK     = 364,
    WXK_SCROLL      = 365,
    WXK_PAGEUP      = 366,
    WXK_PAGEDOWN    = 367,

    WXK_NUMPAD_SPACE    = 368,
    WXK_NUMPAD_TAB      = 369,
    WXK_NUMPAD_ENTER    = 370,
    WXK_NUMPAD_F1       = 371,
    WXK_NUMPAD_F2       = 372,
    WXK_NUMPAD_F3       = 373,
    WXK_NUMPAD_F4       = 374,
    WXK_NUMPAD_HOME     = 375,
    WXK_NUMPAD_LEFT     = 376,
    WXK_NUMPAD_UP       = 377,
    WXK_NUMPAD_RIGHT    = 378,
    WXK_NUMPAD_DOWN     = 379,
    WXK_NUMPAD_PRIOR    = 380,
    WXK_NUMPAD_PAGEUP   = 381,
    WXK_NUMPAD_NEXT     = 382,
    WXK_NUMPAD_PAGEDOWN = 383,
    WXK_NUMPAD_END      = 384,
    WXK_NUMPAD_BEGIN    = 385,
    WXK_NUMPAD_INSERT   = 386,
    WXK_NUMPAD_DELETE   = 387,
    WXK_NUMPAD_EQUAL    = 388,
    WXK_NUMPAD_MULTIPLY = 389,
    WXK_NUMPAD_ADD      = 390,
    WXK_NUMPAD_SEPARATOR = 391,
    WXK_NUMPAD_SUBTRACT = 392,
    WXK_NUMPAD_DECIMAL  = 393,
    WXK_NUMPAD_DIVIDE   = 394,

    WXK_SPECIAL1    = 193,
    WXK_SPECIAL2    = 194,
    WXK_SPECIAL3    = 195,
    WXK_SPECIAL4    = 196,
    WXK_SPECIAL5    = 197,
    WXK_SPECIAL6    = 198,
    WXK_SPECIAL7    = 199,
    WXK_SPECIAL8    = 200,
    WXK_SPECIAL9    = 201,
    WXK_SPECIAL10   = 202,
    WXK_SPECIAL11   = 203,
    WXK_SPECIAL12   = 204,
    WXK_SPECIAL13   = 205,
    WXK_SPECIAL14   = 206,
    WXK_SPECIAL15   = 207,
    WXK_SPECIAL16   = 208,
    WXK_SPECIAL17   = 209,
    WXK_SPECIAL18   = 210,
    WXK_SPECIAL19   = 211,
    WXK_SPECIAL20   = 212,

    WXK_PRIOR       = WXK_PAGEUP,
    WXK_NEXT        = WXK_PAGEDOWN,

    wxOPEN              =  1,
    wxSAVE              =  2,
    wxOVERWRITE_PROMPT  =  4,
    wxHIDE_READONLY     =  8,
    wxFILE_MUST_EXIST   = 16,
    wxMULTIPLE          = 32

public constant
    wxSL_HORIZONTAL      = wxHORIZONTAL, -- 4
    wxSL_VERTICAL        = wxVERTICAL,   -- 8
    wxSL_TICKS           = #0010,
    wxSL_AUTOTICKS       = wxSL_TICKS, -- we don't support manual ticks
    wxSL_LABELS          = #0020,
    wxSL_LEFT            = #0040,
    wxSL_TOP             = #0080,
    wxSL_RIGHT           = #0100,
    wxSL_BOTTOM          = #0200,
    wxSL_BOTH            = #0400,
    wxSL_SELRANGE        = #0800

public constant
    wxSP_NOBORDER         = #0000,
    wxSP_NOSASH           = #0010,
    wxSP_BORDER           = #0020,
    wxSP_PERMIT_UNSPLIT   = #0040,
    wxSP_LIVE_UPDATE      = #0080,
    wxSP_3DSASH           = #0100,
    wxSP_3DBORDER         = #0200,
    wxSP_FULLSASH         = #0400,
    wxSP_3D               = wxSP_3DBORDER + wxSP_3DSASH,
    wxSP_SASH_AQUA        = #0800

public constant
    wxTR_NO_BUTTONS             = #0000,    -- for convenience
    wxTR_HAS_BUTTONS            = #0001,    -- generates a +/- button
    wxTR_TWIST_BUTTONS          = #0002,    -- generates a twister button
    wxTR_NO_LINES               = #0004,    -- don't generate level connectors
    wxTR_LINES_AT_ROOT          = #0008,    -- connect top-level nodes
    wxTR_MAC_BUTTONS            = wxTR_TWIST_BUTTONS, -- backward compatibility
    wxTR_AQUA_BUTTONS           = #0010,    -- used internally

    wxTR_SINGLE                 = #0000,    -- for convenience
    wxTR_MULTIPLE               = #0020,    -- can select multiple items
    wxTR_EXTENDED               = #0040,    -- TODO: allow extended selection
    wxTR_FULL_ROW_HIGHLIGHT     = #2000,    -- highlight full horizontal space

    wxTR_EDIT_LABELS            = #0200,    -- can edit item labels
    wxTR_ROW_LINES              = #0400,    -- put border around items
    wxTR_HIDE_ROOT              = #0800,    -- don't display root node
    wxTR_HAS_VARIABLE_ROW_HEIGHT= #0080     -- what it says

public constant
-- wxToolBar style flags
    wxTB_HORIZONTAL         = wxHORIZONTAL, -- = #0004
    wxTB_TOP                = wxTB_HORIZONTAL,
    wxTB_VERTICAL           = wxVERTICAL,   -- = #0008
    wxTB_LEFT               = wxTB_VERTICAL,
    wxTB_3DBUTTONS          = #0010,
    wxTB_FLAT               = #0020,        -- supported only under Win98+/GTK
    wxTB_DOCKABLE           = #0040,        -- use native docking under GTK
    wxTB_NOICONS            = #0080,        -- don't show the icons
    wxTB_TEXT               = #0100,        -- show the text
    wxTB_NODIVIDER          = #0200,        -- don't show the divider (Windows)
    wxTB_NOALIGN            = #0400,        -- no automatic alignment (Windows)
    wxTB_HORZ_LAYOUT        = #0800,        -- show the text and the icons alongside, not vertically stacked (Win32/GTK)
    wxTB_HORZ_TEXT          = or_bits( wxTB_HORZ_LAYOUT, wxTB_TEXT ),
    wxTB_NO_TOOLTIPS        = #1000,        -- don't show the toolbar short help tooltips
    wxTB_BOTTOM             = #2000,        -- lay out toolbar at the bottom of the window
    wxTB_RIGHT              = #4000         -- lay out toolbar at the right edge of the window

public constant
    wxBK_DEFAULT          = #0000,
    wxBK_TOP              = #0010,
    wxBK_BOTTOM           = #0020,
    wxBK_LEFT             = #0040,
    wxBK_RIGHT            = #0080,
    wxBK_ALIGN_MASK       = wx_or_all(wxBK_TOP & wxBK_BOTTOM & wxBK_LEFT & wxBK_RIGHT),
    wxNB_DEFAULT          = wxBK_DEFAULT,
    wxNB_TOP              = wxBK_TOP,
    wxNB_BOTTOM           = wxBK_BOTTOM,
    wxNB_LEFT             = wxBK_LEFT,
    wxNB_RIGHT            = wxBK_RIGHT,

    wxNB_FIXEDWIDTH       = #0100,
    wxNB_MULTILINE        = #0200,
    wxNB_NOPAGETHEME      = #0400,
    wxNB_FLAT             = #0800

public constant
    wxSP_HORIZONTAL       = wxHORIZONTAL, -- 4
    wxSP_VERTICAL         = wxVERTICAL,   -- 8
    wxSP_ARROW_KEYS       = #1000,
    wxSP_WRAP             = #2000

public constant
    wxST_NO_AUTORESIZE =   0001

public constant
    wxDefaultDateTime   = -1,
    wxNullFont          =  0

public constant
    wxCAL_BORDER_NONE   = 0,
    wxCAL_BORDER_SQUARE = 1,
    wxCAL_BORDER_ROUND  = 2

public constant
    wxDIR_FILES     = 0001,     --// include files
    wxDIR_DIRS      = 0002,     --// include directories
    wxDIR_HIDDEN    = 0004,     --// include hidden files
    wxDIR_DOTDOT    = 0008,     --// include '.' and '..'

    --// by default, enumerate everything except '.' and '..'
    wxDIR_DEFAULT   = wx_or_all( wxDIR_FILES & wxDIR_DIRS & wxDIR_HIDDEN )

public constant
    wxDIRCTRL_DIR_ONLY       = #0010,
    wxDIRCTRL_SELECT_FIRST   = #0020,
    wxDIRCTRL_SHOW_FILTERS   = #0040,
    wxDIRCTRL_3D_INTERNAL    = #0080,
    wxDIRCTRL_EDIT_LABELS    = #0100

public constant
    wxRA_SPECIFY_COLS   = wxHORIZONTAL,
    wxRA_SPECIFY_ROWS   = wxVERTICAL

public constant
    wxSOUND_SYNC        = 0,
    wxSOUND_ASYNC       = 1,
    wxSOUND_LOOP        = 2


public constant
    wxRE_EXTENDED = 0,
    wxRE_ADVANCED = 1,              --// use advanced RE syntax (built-in regex only)
    wxRE_BASIC    = 2,              --// use basic RE syntax
    wxRE_ICASE    = 4,              --// ignore case in match
    wxRE_NOSUB    = 8,              --// only check match, don't set back references
    wxRE_NEWLINE  = 16,             --// if not set, treat '\n' as an ordinary character, otherwise it is
                                    --// special: it is not matched by '.' and '^' and '$' always match
                                    --// after/before it regardless of the setting of wxRE_NOT[BE]OL
    wxRE_DEFAULT  = wxRE_EXTENDED,  --// default flags
    wxRE_NOTBOL = 32,               --// '^' doesn't match at the start of line
    wxRE_NOTEOL = 64                --// '$' doesn't match at the end of line


public constant
    wxMEDIASTATE_STOPPED    = 0,
    wxMEDIASTATE_PAUSED     = 1,
    wxMEDIASTATE_PLAYING    = 2,

    wxMEDIACTRLPLAYERCONTROLS_NONE      = 0,
    wxMEDIACTRLPLAYERCONTROLS_STEP      = 1,        --//Step controls like fastfoward, step one frame etc.
    wxMEDIACTRLPLAYERCONTROLS_VOLUME    = 2,        --//Volume controls like the speaker icon, volume slider, etc.
    wxMEDIACTRLPLAYERCONTROLS_DEFAULT   = wx_or_all(
        wxMEDIACTRLPLAYERCONTROLS_STEP &
        wxMEDIACTRLPLAYERCONTROLS_VOLUME )

public constant
    wxSB_NORMAL    = #0000,
    wxSB_FLAT      = #0001,
    wxSB_RAISED    = #0002

public constant
    wxUSER_ATTENTION_INFO  = 1,
    wxUSER_ATTENTION_ERROR = 2

public constant
    wxIMAGE_QUALITY_NORMAL  = 0,
    wxIMAGE_QUALITY_HIGH    = 1

public constant
    wxFD_OPEN              = 0x0001,
    wxFD_SAVE              = 0x0002,
    wxFD_OVERWRITE_PROMPT  = 0x0004,
    wxFD_FILE_MUST_EXIST   = 0x0010,
    wxFD_MULTIPLE          = 0x0020,
    wxFD_CHANGE_DIR        = 0x0080,
    wxFD_PREVIEW           = 0x0100

public constant
        wxPB_USE_TEXTCTRL             = 0x0002,


        wxDIRP_DIR_MUST_EXIST         = 0x0008,
        wxDIRP_CHANGE_DIR             = 0x0010,

        wxFNTP_FONTDESC_AS_LABEL      = 0x0008,
        wxFNTP_USEFONT_FOR_LABEL      = 0x0010, -- uses the currently selected font to draw the label of the button

        wxFLP_OPEN                    = 0x0400,
        wxFLP_SAVE                    = 0x0800,
        wxFLP_OVERWRITE_PROMPT        = 0x1000,
        wxFLP_FILE_MUST_EXIST         = 0x2000,
        wxFLP_CHANGE_DIR              = 0x4000,

        wxFLP_USE_TEXTCTRL            = wxPB_USE_TEXTCTRL,

        wxCLRP_USE_TEXTCTRL       = wxPB_USE_TEXTCTRL,
        wxCLRP_DEFAULT_STYLE      = 0,

        wxFNTP_USE_TEXTCTRL       = wxPB_USE_TEXTCTRL,
        wxFNTP_DEFAULT_STYLE      = wxFNTP_FONTDESC_AS_LABEL + wxFNTP_USEFONT_FOR_LABEL,
        $

public enum  --     wxGridSelectionModes 
  wxGridSelectCells, 
  wxGridSelectRows, 
  wxGridSelectColumns, 
  wxGridSelectRowsOrColumns, 
$

ifdef not WINDOWS then
    -- GTK apps usually don't have a textctrl next to the picker
public constant
        wxFLP_DEFAULT_STYLE       = wxFLP_OPEN + wxFLP_FILE_MUST_EXIST

elsedef
public constant
        wxFLP_DEFAULT_STYLE       = wxFLP_USE_TEXTCTRL + wxFLP_OPEN + wxFLP_FILE_MUST_EXIST
end ifdef

-- Wrappers
--------------------------------------------

--Non wxObjects
-----------------------
--/topic Non wxObjects
--/parent Classes
--/info
--These are objects that do not derive from wxObject.
-- They cannot be destroyed by calling /delete_instance().
constant
INSTANCE_OF = {
    wxRegEx
    ,wxAuiPaneInfo
    ,wxXmlNode
    ,wxXmlProperty
--@INSTANCE_OF
},
DELETE_PROC ={
    WX_DELETE_WXREGEX
    ,WX_DELETE_WXAUIPANEINFO
    ,WX_DELETE_WXXMLNODE
    ,WX_DELETE_WXXMLPROPERTY
--@DELETE_PROC
}


--/topic wxMDIParentFrame
--/proc set_mdi_toolbar( atom parent_frame, atom toolbar )
--
--Sets the window to be used as a toolbar for this MDI parent window.
-- It saves the application having to manage the positioning of the
-- toolbar MDI client window.
--
--When the frame is resized, the toolbar is resized to be the width of
-- the frame client area, and the toolbar height is kept the same.
--
--The parent of the toolbar must be this frame.
public procedure set_mdi_toolbar( atom parent_frame, atom toolbar )
    c_proc( WX_SET_MDI_TOOLBAR, {parent_frame,toolbar})
end procedure

--/topic wxMDIParentFrame
--/proc set_mdi_window_menu( atom parent_frame, atom menu )
--
--Call this to change the current Window menu. Ownership of the menu
-- object passes to the frame when you call this function.
--
--This call is available under Windows only.
public procedure set_mdi_window_menu( atom parent_frame, atom menu )
    c_proc( WX_SET_MDI_WINDOW_MENU, {parent_frame,menu})
end procedure

--/topic wxApp
--/proc set_app_name( object name )
--
--Sets the /i name of the application. The name may be used in dialogs (for example by the document/view framework).
-- A default name is set by wxWidgets.
public procedure set_app_name( object name )
    c_proc( WX_SET_APP_NAME, {name})
end procedure

--/topic wxApp
--/func get_app_name()
--
--Returns the application name.
--
--/b Remarks
--
--wxWidgets sets this to a reasonable default before calling /wxMain, but the application can reset it at will.
public function get_app_name(  )
    return c_func( WX_GET_APP_NAME, {})
end function

--/topic wxApp
--/proc set_vendor_name( object name )
--
--Sets the name of application's vendor. The name will be used in registry access. A default name is set by wxWidgets.
public procedure set_vendor_name( object name )
    c_proc( WX_SET_VENDOR_NAME, {name})
end procedure

--/topic wxApp
--/func get_vendor_name()
--
--Returns the application's vendor name.
public function get_vendor_name(  )
    return c_func( WX_GET_VENDOR_NAME, {})
end function

--/topic Miscellaneous Routines
--/proc request_user_attention( atom frame, atom flags )
--
--Use a system-dependent way to attract users attention to the window when it is in background.
--
--/i flags may have the value of either /b wxUSER_ATTENTION_INFO (default) or /b wxUSER_ATTENTION_ERROR which results in
-- a more drastic action. When in doubt, use the default value.
--
--Note that this function should normally be only used when the application is not already in foreground.
--
--This function is currently implemented for Win32 where it flashes the window icon in the taskbar, and for wxGTK with task
-- bars supporting it.
public procedure request_user_attention( atom frame, atom flags )
    c_proc( WX_REQUEST_USER_ATTENTION, {frame,flags})
end procedure

--/topic wxApp
--/func is_active()
--
--Returns true if the application is active, i.e. if one of its windows is currently in the foreground. If this function returns
-- false and you need to attract users attention to the application, you may use /request_user_attention() to do it.
public function is_active(  )
    return c_func( WX_IS_ACTIVE, {})
end function

--/topic Graphics
--/func find_color( sequence name )
--
--Finds a colour given the name.
public function find_color( sequence name )
    return c_func( WX_FIND_COLOR, {name})
end function

--/topic Graphics
--/proc draw_arc( atom dc, atom x1, atom y1, atom x2, atom y2, atom xc, atom yc )
--
public procedure draw_arc( atom dc, atom x1, atom y1, atom x2, atom y2, atom xc, atom yc )
        c_proc( WX_DRAW_ARC, {dc,x1,y1,x2,y2,xc,yc})
end procedure

--/topic Graphics
--/proc draw_check_mark( atom dc, atom x, atom y, atom width, atom height )
--
public procedure draw_check_mark( atom dc, atom x, atom y, atom width, atom height )
        c_proc( WX_DRAW_CHECK_MARK, {dc,x,y,width,height})
end procedure

--/topic Graphics
--/proc draw_circle( atom dc, atom x, atom y, atom radius )
--
public procedure draw_circle( atom dc, atom x, atom y, atom radius )
        c_proc( WX_DRAW_CIRCLE, {dc,x,y,radius})
end procedure

--/topic Graphics
--/proc draw_ellipse( atom dc, atom x, atom y, atom width, atom height )
--
public procedure draw_ellipse( atom dc, atom x, atom y, atom width, atom height )
        c_proc( WX_DRAW_ELLIPSE, {dc,x,y,width,height})
end procedure

--/topic Graphics
--/proc draw_elliptic_arc( atom dc, atom x, atom y, atom width, atom height, object angle_start, object angle_end )
--
public procedure draw_elliptic_arc( atom dc, atom x, atom y, atom width, atom height, object angle_start, object angle_end )
        c_proc( WX_DRAW_ELLIPTIC_ARC, {dc,x,y,width,height,angle_start,angle_end})
end procedure

--/topic Graphics
--/proc draw_point( atom dc, atom x, atom y )
--
public procedure draw_point( atom dc, atom x, atom y )
        c_proc( WX_DRAW_POINT, {dc,x,y})
end procedure

--/topic Graphics
--/proc draw_rectangle( atom dc, atom x, atom y, atom width, atom height )
--
public procedure draw_rectangle( atom dc, atom x, atom y, atom width, atom height )
        c_proc( WX_DRAW_RECTANGLE, {dc,x,y,width,height})
end procedure

--/topic Graphics
--/proc draw_rounded_rectangle( atom dc, atom x, atom y, atom width, atom height, object radius )
--
public procedure draw_rounded_rectangle( atom dc, atom x, atom y, atom width, atom height, atom radius )
        c_proc( WX_DRAW_ROUNDED_RECTANGLE, {dc,x,y,width,height,radius})
end procedure

--/topic Graphics
--/proc flood_fill( atom dc, atom x, atom y, atom color, atom style )
--
public procedure flood_fill( atom dc, atom x, atom y, atom color, atom style )
        c_proc( WX_FLOOD_FILL, {dc,x,y,color,style})
end procedure

--/topic Graphics
--/proc gradient_fill( atom dc, object rect, atom initialColor, atom destColor, atom nDirection )
--
public procedure gradient_fill( atom dc, object rect, atom initialColor, atom destColor, atom nDirection )
        c_proc( WX_GRADIENT_FILL, {dc,rect,initialColor,destColor,nDirection})
end procedure

--/topic wxToolBar
--/proc enable_tool( atom bar, atom id, atom enable )
--
public procedure enable_tool( atom bar, atom id, atom enable )
        c_proc( WX_ENABLE_TOOL, {bar,id,enable})
end procedure

--/topic wxHtmlWindow
--/func selection_to_text( atom htmlwin )
--
public function selection_to_text( atom htmlwin )
        return c_func( WX_SELECTION_TO_TEXT, {htmlwin})
end function

--/topic Files
--/func get_type( object filename )
--
public function get_type( object filename )
        return c_func( WX_GET_TYPE, {filename})
end function


--/topic Files
--/func wx_mkdir( object directory, object perm, atom flags )
--
public function wx_mkdir( object directory, object perm, atom flags )
        return c_func( WX_WX_MKDIR, {directory,perm,flags})
end function

--/topic wxFlexGridSizer
--/proc add_growable_col( atom sizer, atom idx, atom proportion )
--
public procedure add_growable_col( atom sizer, atom idx, atom proportion )
        c_proc( WX_ADD_GROWABLE_COL, {sizer,idx,proportion})
end procedure

--/topic wxFlexGridSizer
--/proc add_growable_row( atom sizer, atom idx, atom proportion )
--
public procedure add_growable_row( atom sizer, atom idx, atom proportion )
        c_proc( WX_ADD_GROWABLE_ROW, {sizer,idx,proportion})
end procedure

--/topic wxImageList
--/func remove_all_images( atom ilist )
--
public function remove_all_images( atom ilist )
        return c_func( WX_REMOVE_ALL_IMAGES, {ilist})
end function

--/topic wxImageList
--/func remove_image( atom ilist, atom index )
--
public function remove_image( atom ilist, atom index )
        return c_func( WX_REMOVE_IMAGE, {ilist,index})
end function

--/topic wxImageList
--/func replace_image( atom ilist, atom index, atom bmp, atom mask )
--
public function replace_image( atom ilist, atom index, atom bmp, atom mask )
        return c_func( WX_REPLACE_IMAGE, {ilist,index,bmp,mask})
end function

--wxFileDialog
-----------------------------------------
--/topic wxFileDialog
--/parent Dialogs
--/info
--
--/li /b "atom parent"
--/li sequence message [=""]
--/li sequence defaultDir [=""]
--/li sequence defaultFile [=""]
--/li sequence wildcard [="*.*"]
--/li integer style [= wxFD_DEFAULT_STYLE]
--<ul>
--/li wxFD_DEFAULT_STYLE         Equivalent to wxFD_OPEN.
--/li wxFD_OPEN  This is an open dialog; usually this means that the default button's label of the dialog is "Open". Cannot be combined with wxFD_SAVE.
--/li wxFD_SAVE  This is a save dialog; usually this means that the default button's label of the dialog is "Save". Cannot be combined with wxFD_OPEN.
--/li wxFD_OVERWRITE_PROMPT      For save dialog only: prompt for a confirmation if a file will be overwritten.
--/li wxFD_FILE_MUST_EXIST       For open dialog only: the user may only select files that actually exist.
--/li wxFD_MULTIPLE      For open dialog only: allows selecting multiple files.
--/li wxFD_CHANGE_DIR    Change the current working directory to the directory where the file(s) chosen by the user are.
--/li wxFD_PREVIEW       Show the preview of the selected files (currently only supported by wxGTK using GTK+ 2.4 or later).
--</ul>
--
--This class represents the file chooser dialog.
--
-- Pops up a file selector box. In Windows and GTK2.4+, this is the common file
-- selector dialog. In X, this is a file selector box with somewhat less
-- functionality. The path and filename are distinct elements of a full file
-- pathname. If path is "", the current directory will be used. If filename is "",
-- no default filename will be supplied. The wildcard determines what files are
-- displayed in the file selector, and file extension supplies a type extension for
-- the required filename.
--
-- Both the X and Windows versions implement a wildcard filter. Typing a filename
-- containing wildcards (*, ?) in the filename text item, and clicking on Ok, will
-- result in only those files matching the pattern being displayed. The wildcard
-- may be a specification for multiple types of file with a description for each,
-- such as:
--
--  "BMP and GIF files (*.bmp;*.gif)|*.bmp;*.gif|PNG files (*.png)|*.png"
-- It must be noted that wildcard support in the native Motif file dialog is quite
-- limited: only one alternative is supported, and it is displayed without the
-- descriptive test; "BMP files (*.bmp)|*.bmp'' is displayed as "*.bmp'', and both
-- "BMP files (*.bmp)|*.bmp|GIF files (*.gif)|*.gif'' and "Image
-- files|*.bmp;*.gif'' are errors.


--/topic wxFileDialog
--/func get_fd_filename( atom fd )
--
public function get_fd_filename( atom fd )
        return c_func( WX_GET_FD_FILENAME, {fd})
end function

--/topic wxFileDialog
--/func get_fd_filenames( atom fd )
--
public function get_fd_filenames( atom fd )
        return c_func( WX_GET_FD_FILENAMES, {fd})
end function

--/topic wxFileDialog
--/func get_fd_filter_index( atom fd )
--
public function get_fd_filter_index( atom fd )
        return c_func( WX_GET_FD_FILTER_INDEX, {fd})
end function

--/topic wxFileDialog
--/func get_fd_message( atom fd )
--
public function get_fd_message( atom fd )
        return c_func( WX_GET_FD_MESSAGE, {fd})
end function

--/topic wxFileDialog
--/func get_fd_path( atom fd )
--
public function get_fd_path( atom fd )
        return c_func( WX_GET_FD_PATH, {fd})
end function

--/topic wxFileDialog
--/func get_fd_paths( atom fd )
--
public function get_fd_paths( atom fd )
        return c_func( WX_GET_FD_PATHS, {fd})
end function

--/topic wxFileDialog
--/func get_fd_wildcard( atom fd )
--
public function get_fd_wildcard( atom fd )
        return c_func( WX_GET_FD_WILDCARD, {fd})
end function

--/topic wxFileDialog
--/proc set_fd_directory( atom fd, object directory )
--
public procedure set_fd_directory( atom fd, object directory )
        c_proc( WX_SET_FD_DIRECTORY, {fd,directory})
end procedure

--/topic wxFileDialog
--/proc set_fd_filename( atom fd, object filename )
--
public procedure set_fd_filename( atom fd, object filename )
        c_proc( WX_SET_FD_FILENAME, {fd,filename})
end procedure

--/topic wxFileDialog
--/proc set_fd_filterindex( atom fd, atom filter_index )
--
public procedure set_fd_filterindex( atom fd, atom filter_index )
        c_proc( WX_SET_FD_FILTERINDEX, {fd,filter_index})
end procedure

--/topic wxFileDialog
--/proc set_fd_message( atom fd, object message )
--
public procedure set_fd_message( atom fd, object message )
        c_proc( WX_SET_FD_MESSAGE, {fd,message})
end procedure

--/topic wxFileDialog
--/proc set_fd_path( atom fd, object path )
--
public procedure set_fd_path( atom fd, object path )
        c_proc( WX_SET_FD_PATH, {fd,path})
end procedure

--/topic wxFileDialog
--/proc set_fd_wildcard( atom fd, object wildcard )
--
public procedure set_fd_wildcard( atom fd, object wildcard )
        c_proc( WX_SET_FD_WILDCARD, {fd,wildcard})
end procedure

--/topic wxSplashScreen
--/func get_splash_window( atom splash )
--
public function get_splash_window( atom splash )
        return c_func( WX_GET_SPLASH_WINDOW, {splash})
end function

--/topic wxDC
--/func get_dc_gdk_window( atom windowdc )
--
-- Linux (or other GTK based wxWidgets) version only.  Returns
-- a pointer to the GDK window, which can be used for integration
-- with Cairo or other libraries.
public function get_dc_gdk_window( atom windowdc )
        return c_func( WX_GET_DC_GDK_WINDOW, {windowdc})
end function

--/topic Picker Controls
--/info
--Controls for picking different types of data, along with a separate text control.
--



--/topic Picker Controls
--/proc set_picker_internal_margin( atom picker, atom margin )
--
--Sets the margin (in pixel) between the picker and the text control.
-- This function can be used only when /picker_has_textctrl returns true.
public procedure set_picker_internal_margin( atom picker, atom margin )
        c_proc( WX_SET_PICKER_INTERNAL_MARGIN, {picker,margin})
end procedure

--/topic Picker Controls
--/func get_picker_internal_margin( atom picker )
--
--Returns the margin (in pixel) between the picker and the text control.
-- This function can be used only when /picker_has_textctrl returns true.
public function get_picker_internal_margin( atom picker )
        return c_func( WX_GET_PICKER_INTERNAL_MARGIN, {picker})
end function

--/topic Picker Controls
--/proc set_picker_textctrl_proportion( atom picker, atom prop )
--
--Sets the proportion value of the text control. Look at the overview of wxPickerBase for more details about this.
--This function can be used only when /picker_has_textctrl returns true.
public procedure set_picker_textctrl_proportion( atom picker, atom prop )
        c_proc( WX_SET_PICKER_TEXTCTRL_PROPORTION, {picker,prop})
end procedure

--/topic Picker Controls
--/func get_picker_textctrl_proportion( atom picker )
--
--Returns the margin (in pixel) between the picker and the text control.
-- This function can be used only when /picker_has_textctrl returns true.
public function get_picker_textctrl_proportion( atom picker )
        return c_func( WX_GET_PICKER_TEXTCTRL_PROPORTION, {picker})
end function

--/topic Picker Controls
--/proc set_pickerctrl_proportion( atom picker, atom prop )
--
--
public procedure set_pickerctrl_proportion( atom picker, atom prop )
        c_proc( WX_SET_PICKERCTRL_PROPORTION, {picker,prop})
end procedure

--/topic Picker Controls
--/func get_pickerctrl_proportion( atom picker )
--
--Returns the proportion value of the picker.
public function get_pickerctrl_proportion( atom picker )
        return c_func( WX_GET_PICKERCTRL_PROPORTION, {picker})
end function

--/topic Picker Controls
--/func picker_has_textctrl( atom picker )
--
--Returns true if this window has a valid text control (i.e. if the wxPB_USE_TEXTCTRL style was given when creating this control).
public function picker_has_textctrl( atom picker )
        return c_func( WX_PICKER_HAS_TEXTCTRL, {picker})
end function

--/topic Picker Controls
--/func get_picker_textctrl( atom picker )
--
--Returns a pointer to the text control handled by this window or NULL if the
-- wxPB_USE_TEXTCTRL style was not specified when this control was created. Very
-- important: the contents of the text control could be containing an invalid
-- representation of the entity which can be chosen through the picker (e.g. the
-- user entered an invalid colour syntax because of a typo). Thus you should never
-- parse the content of the textctrl to get the user's input; rather use the
-- derived-class getter (e.g. /get_picker_color, /get_filepicker_path, etc).
public function get_picker_textctrl( atom picker )
        return c_func( WX_GET_PICKER_TEXTCTRL, {picker})
end function

--/topic Picker Controls
--/func picker_is_textctrl_growable( atom picker )
--
--Returns true if the text control is growable. This function can be used only when /picker_has_textctrl returns true.
public function picker_is_textctrl_growable( atom picker )
        return c_func( WX_PICKER_IS_TEXTCTRL_GROWABLE, {picker})
end function

--/topic Picker Controls
--/func picker_is_pickerctrl_growable( atom picker )
--
--Returns true if the picker control is growable.
public function picker_is_pickerctrl_growable( atom picker )
        return c_func( WX_PICKER_IS_PICKERCTRL_GROWABLE, {picker})
end function

--/topic Picker Controls
--/proc set_pickerctrl_growable( atom picker, atom grow )
--
--Sets the picker control as growable when grow is true.
public procedure set_pickerctrl_growable( atom picker, atom grow )
        c_proc( WX_SET_PICKERCTRL_GROWABLE, {picker,grow})
end procedure

--/topic Picker Controls
--/proc set_picker_textctrl_growable( atom picker, atom grow )
--
--Sets the text control as growable when grow is true. This function can be used only when /picker_has_textctrl returns true.
public procedure set_picker_textctrl_growable( atom picker, atom grow )
        c_proc( WX_SET_PICKER_TEXTCTRL_GROWABLE, {picker,grow})
end procedure

--wxDirPickerCtrl
-----------------------------------------
--/topic wxDirPickerCtrl
--/parent Picker Controls
--/parent Controls
--/info
-- --This control allows the user to select a directory. The generic implementation
-- is a button which brings up a wxDirDialog when clicked. Native implementation
-- may differ but this is usually a (small) widget which give access to the
-- dir-chooser dialog. It is only available if wxUSE_DIRPICKERCTRL is set to 1 (the
-- default).
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li atom parent
--/li integer id [=-1]
--/li sequence path [=""]
--/li sequence message [="Select a folder"]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cy [=-1]
--/li integer style [=wxDIRP_DEFAULT_STYLE]
--<ul>
--/li wxDIRP_DEFAULT_STYLE       The default style: includes wxDIRP_DIR_MUST_EXIST and, under wxMSW only, wxDIRP_USE_TEXTCTRL.
--/li wxDIRP_USE_TEXTCTRL        Creates a text control to the left of the picker button which is completely managed by the wxDirPickerCtrl and which can be used by the user to specify a path (see SetPath). The text control is automatically synchronized with button's value. Use functions defined in wxPickerBase to modify the text control.
--/li wxDIRP_DIR_MUST_EXIST      Creates a picker which allows to select only existing directories. wxGTK control always adds this flag internally as it does not support its absence.
--/li wxDIRP_CHANGE_DIR  Change current working directory on each user directory selection change.
--</ul></ul>
--
--You can use the wxEVT_COMMAND_DIRPICKER_CHANGED event, along with /get_filedirpicker_event_path.

--/topic wxDirPickerCtrl
--/func get_dirpicker_path( atom picker )
--
--Returns the absolute path of the currently selected directory.
public function get_dirpicker_path( atom picker )
        return c_func( WX_GET_DIRPICKER_PATH, {picker})
end function

--/topic wxDirPickerCtrl
--/proc set_dirpicker_path( atom picker, object path )
--
--Sets the absolute path of the currently selected directory. This must be a valid directory if wxDIRP_DIR_MUST_EXIST style was given.
public procedure set_dirpicker_path( atom picker, object path )
        c_proc( WX_SET_DIRPICKER_PATH, {picker,path})
end procedure

--wxColourPickerCtrl
-----------------------------------------
--/topic wxColourPickerCtrl
--/parent Picker Controls
--/parent Controls
--/info
-- This control allows the user to select a colour. The generic implementation is
-- a button which brings up a wxColourDialog when clicked. Native implementation
-- may differ but this is usually a (small) widget which give access to the
-- colour-chooser dialog. It is only available if wxUSE_COLOURPICKERCTRL is set to
-- 1 (the default).
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li atom parent
--/li integer id [=-1]
--/li atom color [=wxBLACK]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cy [=-1]
--/li integer style [=wxCLRP_DEFAULT_STYLE]
--<ul>
--/li wxCLRP_DEFAULT_STYLE       The default style: 0.
--/li wxCLRP_USE_TEXTCTRL        Creates a text control to the left of the picker button which is completely managed by the wxColourPickerCtrl and which can be used by the user to specify a colour (see SetColour). The text control is automatically synchronized with button's value. Use functions defined in wxPickerBase to modify the text control.
--</ul></ul>
--
--You can use the wxEVT_COMMAND_COLOURPICKER_CHANGED event, along with /get_colorpickerevent_color.

--/topic wxColourPickerCtrl
--/func get_picker_color( atom picker )
--
--Returns the currently selected color.
public function get_picker_color( atom picker )
        return c_func( WX_GET_PICKER_COLOR, {picker})
end function

--/topic wxColourPickerCtrl
--/proc set_picker_color( atom picker, atom color )
--
--Sets the currently selected color.
public procedure set_picker_color( atom picker, atom color )
        c_proc( WX_SET_PICKER_COLOR, {picker,color})
end procedure

--wxDatePickerCtrl
-----------------------------------------
--/topic wxDatePickerCtrl
--/parent Picker Controls
--/parent Controls
--/info
--This control allows the user to select a date. Unlike wxCalendarCtrl, which is
-- a relatively big control, wxDatePickerCtrl is implemented as a small window
-- showing the currently selected date. The control can be edited using the
-- keyboard, and can also display a popup window for more user-friendly date
-- selection, depending on the styles used and the platform, except PalmOS where
-- date is selected using native dialog.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li atom parent
--/li integer id [=-1]
--/li sequence date_time [=wxDefaultDateTime]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cy [=-1]
--/li integer style [=wxCLRP_DEFAULT_STYLE]
--<ul>
--/li wxDP_SPIN  Creates a control without a month calendar drop down but with spin-control-like arrows to change individual date components. This style is not supported by the generic version.
--/li wxDP_DROPDOWN      Creates a control with a month calendar drop-down part from which the user can select a date.
--/li wxDP_DEFAULT       Creates a control with the style that is best supported for the current platform (currently wxDP_SPIN under Windows and wxDP_DROPDOWN elsewhere).
--/li wxDP_ALLOWNONE     With this style, the control allows the user to not enter any valid date at all. Without it - the default - the control always has some valid date.
--/li wxDP_SHOWCENTURY   Forces display of the century in the default date format. Without this style the century could be displayed, or not, depending on the default date representation in the system.
--</ul></ul>
--
--You can use the wxEVT_DATE_CHANGED event along with /get_dateevent_date.

--/topic wxDatePickerCtrl
--/func get_picker_range( atom picker )
--
--If the control had been previously limited to a range of dates using
-- /set_picker_range , returns the lower and upper bounds of this range. If no
-- range is set (or only one of the bounds is set), returns 0.
public function get_picker_range( atom picker )
        return c_func( WX_GET_PICKER_RANGE, {picker})
end function

--/topic wxDatePickerCtrl
--/func get_picker_date( atom picker )
--
--Returns the currently selected date. If there is no selection or the selection
-- is outside of the current range, an invalid object is returned.
public function get_picker_date( atom picker )
        return c_func( WX_GET_PICKER_DATE, {picker})
end function

--/topic wxDatePickerCtrl
--/proc set_picker_range( atom picker, object dt1, object dt2 )
--
--Sets the valid range for the date selection. If dt1 is valid, it becomes the
-- earliest date (inclusive) accepted by the control. If dt2 is valid, it becomes
-- the latest possible date.
--
-- Remarks
--
-- If the current value of the control is outside of the newly set range bounds, the behaviour is undefined.
public procedure set_picker_range( atom picker, object dt1, object dt2 )
        c_proc( WX_SET_PICKER_RANGE, {picker,dt1,dt2})
end procedure

--/topic wxDatePickerCtrl
--/proc set_picker_date( atom picker, object dt )
--
-- Changes the current value of the control. The date should be valid and included in the currently selected range, if any.
--
-- Calling this method does not result in a date change event.
public procedure set_picker_date( atom picker, object dt )
        c_proc( WX_SET_PICKER_DATE, {picker,dt})
end procedure

--wxFilePickerCtrl
-----------------------------------------
--/topic wxFilePickerCtrl
--/parent Picker Controls
--/parent Controls
--/info
--
--This control allows the user to select a file. The generic implementation is a
-- button which brings up a wxFileDialog when clicked. Native implementation may
-- differ but this is usually a (small) widget which give access to the
-- file-chooser dialog.
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li atom parent
--/li integer id [=-1]
--/li sequence path [=""]
--/li sequence message [="Select a file"]
--/li sequence wildcard [="*.*"]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cy [=-1]
--/li integer style [=wxDIRP_DEFAULT_STYLE]
--<ul>
--/li wxFLP_DEFAULT_STYLE        The default style: includes wxFLP_OPEN | wxFLP_FILE_MUST_EXIST and, under wxMSW only, wxFLP_USE_TEXTCTRL.
--/li wxFLP_USE_TEXTCTRL         Creates a text control to the left of the picker button which is completely managed by the wxFilePickerCtrl and which can be used by the user to specify a path (see SetPath). The text control is automatically synchronized with button's value. Use functions defined in wxPickerBase to modify the text control.
--/li wxFLP_OPEN         Creates a picker which allows the user to select a file to open.
--/li wxFLP_SAVE         Creates a picker which allows the user to select a file to save.
--/li wxFLP_OVERWRITE_PROMPT     Can be combined with wxFLP_SAVE only: ask confirmation to the user before selecting a file.
--/li wxFLP_FILE_MUST_EXIST      Can be combined with wxFLP_OPEN only: the selected file must be an existing file.
--/li wxFLP_CHANGE_DIR   Change current working directory on each user file selection change.
--</ul></ul>
--
--You can use wxEVT_COMMAND_FILEPICKER_CHANGED along with /get_filedirpicker_event_path.

--/topic wxFilePickerCtrl
--/func get_filepicker_path( atom picker )
--
-- Returns the absolute path of the currently selected file.
public function get_filepicker_path( atom picker )
        return c_func( WX_GET_FILEPICKER_PATH, {picker})
end function

--/topic wxFilePickerCtrl
--/proc set_filepicker_path( atom picker, object path )
--
--Sets the absolute path of the currently selected file. This must be a valid file if the wxFLP_FILE_MUST_EXIST style was given.
public procedure set_filepicker_path( atom picker, object path )
        c_proc( WX_SET_FILEPICKER_PATH, {picker,path})
end procedure

--wxFontPickerCtrl
-----------------------------------------
--/topic wxFontPickerCtrl
--/parent Picker Controls
--/parent Controls
--/info
--
-- This control allows the user to select a font. The generic implementation is a
-- button which brings up a wxFontDialog when clicked. Native implementation may
-- differ but this is usually a (small) widget which give access to the
-- font-chooser dialog.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li atom parent
--/li integer id [=-1]
--/li sequence date_time [=wxDefaultDateTime]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cy [=-1]
--/li integer style [=wxCLRP_DEFAULT_STYLE]
--<ul>
--/li wxFNTP_DEFAULT_STYLE       The default style: wxFNTP_FONTDESC_AS_LABEL | wxFNTP_USEFONT_FOR_LABEL.
--/li wxFNTP_USE_TEXTCTRL        Creates a text control to the left of the picker button which is completely managed by the wxFontPickerCtrl and which can be used by the user to specify a font (see SetSelectedFont). The text control is automatically synchronized with button's value. Use functions defined in wxPickerBase to modify the text control.
--/li wxFNTP_FONTDESC_AS_LABEL   Keeps the label of the button updated with the fontface name and the font size. E.g. choosing "Times New Roman bold, italic with size 10" from the fontdialog, will update the label (overwriting any previous label) with the "Times New Roman, 10" text.
--/li wxFNTP_USEFONT_FOR_LABEL   Uses the currently selected font to draw the label of the button.
--</ul></ul>
--
--You can use wxEVT_COMMAND_FONTPICKER_CHANGED along with /get_fontpicker_event_font.

--/topic wxFontPickerCtrl
--/func get_picker_font( atom picker )
--
--Returns the currently selected font.
public function get_picker_font( atom picker )
        return c_func( WX_GET_PICKER_FONT, {picker})
end function

--/topic wxFontPickerCtrl
--/proc set_picker_font( atom picker, atom font )
--
--Sets the currently selected font.
public procedure set_picker_font( atom picker, atom font )
        c_proc( WX_SET_PICKER_FONT, {picker,font})
end procedure

--/topic wxFontPickerCtrl
--/func get_picker_max_point( atom picker )
--
--Returns the maximum point size value allowed for the user-chosen font.
public function get_picker_max_point( atom picker )
        return c_func( WX_GET_PICKER_MAX_POINT, {picker})
end function

--/topic wxFontPickerCtrl
--/proc set_picker_max_point( atom picker, atom max )
--
-- Sets the maximum point size value allowed for the user-chosen font. The default
-- value is 100. Note that big fonts can require a lot of memory and CPU time both
-- for creation and for rendering; thus, specially because the user has the option
-- to specify the fontsize through a text control (see wxFNTP_USE_TEXTCTRL), it's a
-- good idea to put a limit to the maximum font size when huge fonts do not make
-- much sense.
public procedure set_picker_max_point( atom picker, atom max )
        c_proc( WX_SET_PICKER_MAX_POINT, {picker,max})
end procedure

--/topic wxColourPickerEvent
--/parent Events
--/parent wxColourPickerCtrl
--/info
--
--Use wxEVT_COMMAND_COLOURPICKER_CHANGED to handle this event.

--/topic wxColourPickerEvent
--/func get_colorpickerevent_color( atom event )
--
public function get_colorpickerevent_color( atom event )
        return c_func( WX_GET_COLORPICKEREVENT_COLOR, {event})
end function

--/topic wxDateEvent
--/parent wxDatePickerCtrl
--/parent Events
--/info


--/topic wxDateEvent
--/func get_dateevent_date( atom event )
--
public function get_dateevent_date( atom event )
        return c_func( WX_GET_DATEEVENT_DATE, {event})
end function

--/topic Picker Controls
--/func get_filedirpicker_event_path( atom event )
--
--Works for both /wxFilePickerCtrl and /wxDirPickerCtrl
public function get_filedirpicker_event_path( atom event )
        return c_func( WX_GET_FILEDIRPICKER_EVENT_PATH, {event})
end function

--/topic wxFontPickerEvent
--/parent wxFontPickerCtrl
--/parent Events
--/info
--
--Use the wxEVT_COMMAND_FONTPICKER_CHANGED to handle this event.

--/topic wxFontPickerEvent
--/func get_fontpicker_event_font( atom event )
--
public function get_fontpicker_event_font( atom event )
        return c_func( WX_GET_FONTPICKER_EVENT_FONT, {event})
end function

--wxCollapsiblePane
-----------------------------------------
--/topic wxCollapsiblePane
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence label_text [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= 0]
--<ul>
-- A collapsible pane is a container with an embedded button-like control which can
-- be used by the user to collapse or expand the pane's contents.
--
-- Once constructed you should use the /get_collpane_win function to access the pane and add
-- your controls inside it (i.e. use the GetPane's returned pointer as parent for
-- the controls which must go in the pane, NOT the wxCollapsiblePane itself!).
--
-- Note that because of its nature of control which can dynamically (and
-- drastically) change its size at run-time under user-input, when putting
-- wxCollapsiblePane inside a wxSizer you should be careful to add it with a
-- proportion value of zero; this is because otherwise all other windows with
-- non-null proportion values would automatically get resized each time the user
-- expands or collapse the pane window resulting usually in a weird, flickering
-- effect.
--
-- Use the wxEVT_COMMAND_COLLPANE_CHANGED (/wxCollapsiblePaneEvent) event to handle
-- the event when the user collapses or expands the pane.  Note that you may need
-- to handle this, and call /layout_sizer in the event handler in order to get the
-- sizers to properly resize.

--/topic wxCollapsiblePane
--/func is_collpane_collapsed( atom pane )
--
public function is_collpane_collapsed( atom pane )
        return c_func( WX_IS_COLLPANE_COLLAPSED, {pane})
end function

--/topic wxCollapsiblePane
--/func is_collpane_expanded( atom pane )
--
public function is_collpane_expanded( atom pane )
        return c_func( WX_IS_COLLPANE_EXPANDED, {pane})
end function

--/topic wxCollapsiblePane
--/proc collapse_collpane( atom pane, atom collapse )
--
public procedure collapse_collpane( atom pane, atom collapse )
        c_proc( WX_COLLAPSE_COLLPANE, {pane,collapse})
end procedure

--/topic wxCollapsiblePane
--/proc expand_collpane( atom pane )
--
public procedure expand_collpane( atom pane )
        c_proc( WX_EXPAND_COLLPANE, {pane})
end procedure

--/topic wxCollapsiblePane
--/func get_collpane_win( atom pane )
--
public function get_collpane_win( atom pane )
        return c_func( WX_GET_COLLPANE_WIN, {pane})
end function

--/topic wxCollapsiblePaneEvent
--/parent Events
--/parent wxCollapsiblePane
--/info

--/topic wxCollapsiblePaneEvent
--/func get_collpan_event_collapsed( atom event )
--
public function get_collpan_event_collapsed( atom event )
        return c_func( WX_GET_COLLPAN_EVENT_COLLAPSED, {event})
end function

--/topic Clipboard
--/proc use_primary_selection( atom primary )
--
-- On platforms supporting it (currently only GTK), selects the so called PRIMARY
-- SELECTION as the clipboard as opposed to the normal clipboard, if primary is
-- true.
public procedure use_primary_selection( atom primary )
        c_proc( WX_USE_PRIMARY_SELECTION, {primary})
end procedure

-- Internationalization
-----------------------------------------
--/topic Internationalization
--/info
-- Writing applications for multiple languages.


--wxLocale
-----------------------------------------
--/topic wxLocale
--/parent Internationalization
--/parent Classes
--/info
--

--/topic wxLocale
--/func get_language( atom locale )
--
public function get_language( atom locale )
        return c_func( WX_GET_LANGUAGE, {locale})
end function

--/topic wxLocale
--/func get_canonical_language_name( atom locale )
--
public function get_canonical_language_name( atom locale  )
        return c_func( WX_GET_CANONICAL_LANGUAGE_NAME, {locale})
end function

--/topic wxLocale
--/func get_language_name( atom lang )
--
public function get_language_name( atom lang )
        return c_func( WX_GET_LANGUAGE_NAME, {lang})
end function

--/topic wxLocale
--/func get_locale( atom locale )
--
public function get_locale( atom locale )
        return c_func( WX_GET_LOCALE, {locale})
end function

--/topic wxLocale
--/func get_locale_name( atom locale )
--
public function get_locale_name( atom locale )
        return c_func( WX_GET_LOCALE_NAME, {locale})
end function

--/topic wxLocale
--/func get_locale_sys_name( atom locale )
--
public function get_locale_sys_name( atom locale )
        return c_func( WX_GET_LOCALE_SYS_NAME, {locale})
end function

--/topic wxLocale
--/func get_system_encoding(  )
--
public function get_system_encoding(  )
        return c_func( WX_GET_SYSTEM_ENCODING, {})
end function

--/topic wxLocale
--/func get_system_encoding_name(  )
--
public function get_system_encoding_name(  )
        return c_func( WX_GET_SYSTEM_ENCODING_NAME, {})
end function

--/topic wxLocale
--/func get_system_language(  )
--
public function get_system_language(  )
        return c_func( WX_GET_SYSTEM_LANGUAGE, {})
end function

--wxCSConv
-----------------------------------------
--/topic wxCSConv
--/parent Internationalization
--/parent Classes
--/info
--
--This class converts between any character sets and Unicode.
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li object Either a sequence with the name of the character set, or the font encoding constant (see /set_default_font_encoding).
--<ul>

--/topic wxCSConv
--/func conv_ok( atom conv )
--
public function conv_ok( atom conv )
        return c_func( WX_CONV_OK, {conv})
end function

--/topic wxCSConv
--/func conv_mb2wc( atom conv, object in_seq )
--
-- Converts the sequence from a multi-byte encoding into wxEuphoria's native unicode
-- encoding.  If the sequence cannot be converted (either due to an encoding not being
-- supported on the system or an incorrectly encoded string) zero will be returned.
public function conv_mb2wc( atom conv, object in_seq )
        return c_func( WX_CONV_MB2WC, {conv,in_seq})
end function

--/topic wxCSConv
--/func conv_wc2mb( atom conv, object in_seq )
--
-- Converts the sequence from wxEuphoria's native unicode into the multi-byte
-- encoding.  If the sequence cannot be converted (either due to an encoding not being
-- supported on the system or an incorrectly encoded string) zero will be returned.
public function conv_wc2mb( atom conv, object in_seq )
        return c_func( WX_CONV_WC2MB, {conv,in_seq})
end function

--/topic wxLocale
--/func is_locale_available( atom lang )
--
-- Language constants:
--<pre style="height: 320px; width: 480px; overflow: scroll;">
--     // user's default/preffered language as got from OS:
--     wxLANGUAGE_DEFAULT,
--     // unknown language, if wxLocale::GetSystemLanguage fails:
--     wxLANGUAGE_UNKNOWN,
--
--     wxLANGUAGE_ABKHAZIAN,
--     wxLANGUAGE_AFAR,
--     wxLANGUAGE_AFRIKAANS,
--     wxLANGUAGE_ALBANIAN,
--     wxLANGUAGE_AMHARIC,
--     wxLANGUAGE_ARABIC,
--     wxLANGUAGE_ARABIC_ALGERIA,
--     wxLANGUAGE_ARABIC_BAHRAIN,
--     wxLANGUAGE_ARABIC_EGYPT,
--     wxLANGUAGE_ARABIC_IRAQ,
--     wxLANGUAGE_ARABIC_JORDAN,
--     wxLANGUAGE_ARABIC_KUWAIT,
--     wxLANGUAGE_ARABIC_LEBANON,
--     wxLANGUAGE_ARABIC_LIBYA,
--     wxLANGUAGE_ARABIC_MOROCCO,
--     wxLANGUAGE_ARABIC_OMAN,
--     wxLANGUAGE_ARABIC_QATAR,
--     wxLANGUAGE_ARABIC_SAUDI_ARABIA,
--     wxLANGUAGE_ARABIC_SUDAN,
--     wxLANGUAGE_ARABIC_SYRIA,
--     wxLANGUAGE_ARABIC_TUNISIA,
--     wxLANGUAGE_ARABIC_UAE,
--     wxLANGUAGE_ARABIC_YEMEN,
--     wxLANGUAGE_ARMENIAN,
--     wxLANGUAGE_ASSAMESE,
--     wxLANGUAGE_AYMARA,
--     wxLANGUAGE_AZERI,
--     wxLANGUAGE_AZERI_CYRILLIC,
--     wxLANGUAGE_AZERI_LATIN,
--     wxLANGUAGE_BASHKIR,
--     wxLANGUAGE_BASQUE,
--     wxLANGUAGE_BELARUSIAN,
--     wxLANGUAGE_BENGALI,
--     wxLANGUAGE_BHUTANI,
--     wxLANGUAGE_BIHARI,
--     wxLANGUAGE_BISLAMA,
--     wxLANGUAGE_BRETON,
--     wxLANGUAGE_BULGARIAN,
--     wxLANGUAGE_BURMESE,
--     wxLANGUAGE_CAMBODIAN,
--     wxLANGUAGE_CATALAN,
--     wxLANGUAGE_CHINESE,
--     wxLANGUAGE_CHINESE_SIMPLIFIED,
--     wxLANGUAGE_CHINESE_TRADITIONAL,
--     wxLANGUAGE_CHINESE_HONGKONG,
--     wxLANGUAGE_CHINESE_MACAU,
--     wxLANGUAGE_CHINESE_SINGAPORE,
--     wxLANGUAGE_CHINESE_TAIWAN,
--     wxLANGUAGE_CORSICAN,
--     wxLANGUAGE_CROATIAN,
--     wxLANGUAGE_CZECH,
--     wxLANGUAGE_DANISH,
--     wxLANGUAGE_DUTCH,
--     wxLANGUAGE_DUTCH_BELGIAN,
--     wxLANGUAGE_ENGLISH,
--     wxLANGUAGE_ENGLISH_UK,
--     wxLANGUAGE_ENGLISH_US,
--     wxLANGUAGE_ENGLISH_AUSTRALIA,
--     wxLANGUAGE_ENGLISH_BELIZE,
--     wxLANGUAGE_ENGLISH_BOTSWANA,
--     wxLANGUAGE_ENGLISH_CANADA,
--     wxLANGUAGE_ENGLISH_CARIBBEAN,
--     wxLANGUAGE_ENGLISH_DENMARK,
--     wxLANGUAGE_ENGLISH_EIRE,
--     wxLANGUAGE_ENGLISH_JAMAICA,
--     wxLANGUAGE_ENGLISH_NEW_ZEALAND,
--     wxLANGUAGE_ENGLISH_PHILIPPINES,
--     wxLANGUAGE_ENGLISH_SOUTH_AFRICA,
--     wxLANGUAGE_ENGLISH_TRINIDAD,
--     wxLANGUAGE_ENGLISH_ZIMBABWE,
--     wxLANGUAGE_ESPERANTO,
--     wxLANGUAGE_ESTONIAN,
--     wxLANGUAGE_FAEROESE,
--     wxLANGUAGE_FARSI,
--     wxLANGUAGE_FIJI,
--     wxLANGUAGE_FINNISH,
--     wxLANGUAGE_FRENCH,
--     wxLANGUAGE_FRENCH_BELGIAN,
--     wxLANGUAGE_FRENCH_CANADIAN,
--     wxLANGUAGE_FRENCH_LUXEMBOURG,
--     wxLANGUAGE_FRENCH_MONACO,
--     wxLANGUAGE_FRENCH_SWISS,
--     wxLANGUAGE_FRISIAN,
--     wxLANGUAGE_GALICIAN,
--     wxLANGUAGE_GEORGIAN,
--     wxLANGUAGE_GERMAN,
--     wxLANGUAGE_GERMAN_AUSTRIAN,
--     wxLANGUAGE_GERMAN_BELGIUM,
--     wxLANGUAGE_GERMAN_LIECHTENSTEIN,
--     wxLANGUAGE_GERMAN_LUXEMBOURG,
--     wxLANGUAGE_GERMAN_SWISS,
--     wxLANGUAGE_GREEK,
--     wxLANGUAGE_GREENLANDIC,
--     wxLANGUAGE_GUARANI,
--     wxLANGUAGE_GUJARATI,
--     wxLANGUAGE_HAUSA,
--     wxLANGUAGE_HEBREW,
--     wxLANGUAGE_HINDI,
--     wxLANGUAGE_HUNGARIAN,
--     wxLANGUAGE_ICELANDIC,
--     wxLANGUAGE_INDONESIAN,
--     wxLANGUAGE_INTERLINGUA,
--     wxLANGUAGE_INTERLINGUE,
--     wxLANGUAGE_INUKTITUT,
--     wxLANGUAGE_INUPIAK,
--     wxLANGUAGE_IRISH,
--     wxLANGUAGE_ITALIAN,
--     wxLANGUAGE_ITALIAN_SWISS,
--     wxLANGUAGE_JAPANESE,
--     wxLANGUAGE_JAVANESE,
--     wxLANGUAGE_KANNADA,
--     wxLANGUAGE_KASHMIRI,
--     wxLANGUAGE_KASHMIRI_INDIA,
--     wxLANGUAGE_KAZAKH,
--     wxLANGUAGE_KERNEWEK,
--     wxLANGUAGE_KINYARWANDA,
--     wxLANGUAGE_KIRGHIZ,
--     wxLANGUAGE_KIRUNDI,
--     wxLANGUAGE_KONKANI,
--     wxLANGUAGE_KOREAN,
--     wxLANGUAGE_KURDISH,
--     wxLANGUAGE_LAOTHIAN,
--     wxLANGUAGE_LATIN,
--     wxLANGUAGE_LATVIAN,
--     wxLANGUAGE_LINGALA,
--     wxLANGUAGE_LITHUANIAN,
--     wxLANGUAGE_MACEDONIAN,
--     wxLANGUAGE_MALAGASY,
--     wxLANGUAGE_MALAY,
--     wxLANGUAGE_MALAYALAM,
--     wxLANGUAGE_MALAY_BRUNEI_DARUSSALAM,
--     wxLANGUAGE_MALAY_MALAYSIA,
--     wxLANGUAGE_MALTESE,
--     wxLANGUAGE_MANIPURI,
--     wxLANGUAGE_MAORI,
--     wxLANGUAGE_MARATHI,
--     wxLANGUAGE_MOLDAVIAN,
--     wxLANGUAGE_MONGOLIAN,
--     wxLANGUAGE_NAURU,
--     wxLANGUAGE_NEPALI,
--     wxLANGUAGE_NEPALI_INDIA,
--     wxLANGUAGE_NORWEGIAN_BOKMAL,
--     wxLANGUAGE_NORWEGIAN_NYNORSK,
--     wxLANGUAGE_OCCITAN,
--     wxLANGUAGE_ORIYA,
--     wxLANGUAGE_OROMO,
--     wxLANGUAGE_PASHTO,
--     wxLANGUAGE_POLISH,
--     wxLANGUAGE_PORTUGUESE,
--     wxLANGUAGE_PORTUGUESE_BRAZILIAN,
--     wxLANGUAGE_PUNJABI,
--     wxLANGUAGE_QUECHUA,
--     wxLANGUAGE_RHAETO_ROMANCE,
--     wxLANGUAGE_ROMANIAN,
--     wxLANGUAGE_RUSSIAN,
--     wxLANGUAGE_RUSSIAN_UKRAINE,
--     wxLANGUAGE_SAMOAN,
--     wxLANGUAGE_SANGHO,
--     wxLANGUAGE_SANSKRIT,
--     wxLANGUAGE_SCOTS_GAELIC,
--     wxLANGUAGE_SERBIAN,
--     wxLANGUAGE_SERBIAN_CYRILLIC,
--     wxLANGUAGE_SERBIAN_LATIN,
--     wxLANGUAGE_SERBO_CROATIAN,
--     wxLANGUAGE_SESOTHO,
--     wxLANGUAGE_SETSWANA,
--     wxLANGUAGE_SHONA,
--     wxLANGUAGE_SINDHI,
--     wxLANGUAGE_SINHALESE,
--     wxLANGUAGE_SISWATI,
--     wxLANGUAGE_SLOVAK,
--     wxLANGUAGE_SLOVENIAN,
--     wxLANGUAGE_SOMALI,
--     wxLANGUAGE_SPANISH,
--     wxLANGUAGE_SPANISH_ARGENTINA,
--     wxLANGUAGE_SPANISH_BOLIVIA,
--     wxLANGUAGE_SPANISH_CHILE,
--     wxLANGUAGE_SPANISH_COLOMBIA,
--     wxLANGUAGE_SPANISH_COSTA_RICA,
--     wxLANGUAGE_SPANISH_DOMINICAN_REPUBLIC,
--     wxLANGUAGE_SPANISH_ECUADOR,
--     wxLANGUAGE_SPANISH_EL_SALVADOR,
--     wxLANGUAGE_SPANISH_GUATEMALA,
--     wxLANGUAGE_SPANISH_HONDURAS,
--     wxLANGUAGE_SPANISH_MEXICAN,
--     wxLANGUAGE_SPANISH_MODERN,
--     wxLANGUAGE_SPANISH_NICARAGUA,
--     wxLANGUAGE_SPANISH_PANAMA,
--     wxLANGUAGE_SPANISH_PARAGUAY,
--     wxLANGUAGE_SPANISH_PERU,
--     wxLANGUAGE_SPANISH_PUERTO_RICO,
--     wxLANGUAGE_SPANISH_URUGUAY,
--     wxLANGUAGE_SPANISH_US,
--     wxLANGUAGE_SPANISH_VENEZUELA,
--     wxLANGUAGE_SUNDANESE,
--     wxLANGUAGE_SWAHILI,
--     wxLANGUAGE_SWEDISH,
--     wxLANGUAGE_SWEDISH_FINLAND,
--     wxLANGUAGE_TAGALOG,
--     wxLANGUAGE_TAJIK,
--     wxLANGUAGE_TAMIL,
--     wxLANGUAGE_TATAR,
--     wxLANGUAGE_TELUGU,
--     wxLANGUAGE_THAI,
--     wxLANGUAGE_TIBETAN,
--     wxLANGUAGE_TIGRINYA,
--     wxLANGUAGE_TONGA,
--     wxLANGUAGE_TSONGA,
--     wxLANGUAGE_TURKISH,
--     wxLANGUAGE_TURKMEN,
--     wxLANGUAGE_TWI,
--     wxLANGUAGE_UIGHUR,
--     wxLANGUAGE_UKRAINIAN,
--     wxLANGUAGE_URDU,
--     wxLANGUAGE_URDU_INDIA,
--     wxLANGUAGE_URDU_PAKISTAN,
--     wxLANGUAGE_UZBEK,
--     wxLANGUAGE_UZBEK_CYRILLIC,
--     wxLANGUAGE_UZBEK_LATIN,
--     wxLANGUAGE_VIETNAMESE,
--     wxLANGUAGE_VOLAPUK,
--     wxLANGUAGE_WELSH,
--     wxLANGUAGE_WOLOF,
--     wxLANGUAGE_XHOSA,
--     wxLANGUAGE_YIDDISH,
--     wxLANGUAGE_YORUBA,
--     wxLANGUAGE_ZHUANG,
--     wxLANGUAGE_ZULU
-- </pre>
public function is_locale_available( atom lang )
        return c_func( WX_IS_LOCALE_AVAILABLE, {lang})
end function

--/topic wxFont
--/func get_default_font_encoding(  )
--
--Returns the current application's default encoding.
public function get_default_font_encoding(  )
        return c_func( WX_GET_DEFAULT_FONT_ENCODING, {})
end function

--wxAuiDockArt
-----------------------------------------
--/topic wxAuiDockArt
--/parent Advanced User Interface
--/info
--A dock art provider provides all drawing functionality to the wxAui dock manager.
-- This allows the dock manager to have a plugable look-and-feel.
--
--There are two ways to use a custom dock art manager:
--<ul>
--/li Simply override the default settings using these functions:
--<ul>
--/li /set_dockart_colour()
--/li /set_dockart_font()
--/li /set_dockart_metric()
--</ul>
--/li Override the drawing functions to draw your own graphics, using these functions:
--<ul>
--/li /dockart_drawbackground_id()
--/li /dockart_drawborder_id()
--/li /dockart_drawcaption_id()
--/li /dockart_drawgripper_id()
--/li /dockart_drawpanebutton_id()
--/li /dockart_drawsash_id()
--</ul>
--</ul>
--
--Finally, set your new art provider using /set_art_provider(). Don't forget to setup a /wxAuiTabArt
-- provider if you're using any /wxAuiNotebook controls.


--/topic wxAuiDockArt
--/proc dockart_drawbackground_id( atom dockArt, atom rid )
--Sets the /b DrawBackground routine id. To unset, pass -1 for /b rid.
--
--Your /b DrawBackground procedure should take 4 parameters:
--<ul>
--/li atom dc
--/li atom window
--/li integer orientation
--/li sequence rect (x, y, width, height)
--</ul>
public procedure dockart_drawbackground_id( atom dockArt, atom rid )
        c_proc( WX_DOCKART_DRAWBACKGROUND_ID, {dockArt,rid})
end procedure

--/topic wxAuiDockArt
--/proc dockart_drawborder_id( atom dockArt, atom rid )
--Sets the /b DrawBorder routine id. To unset, pass -1 for /b rid.
--
--Your /b DrawBorder procedure should take 4 parameters:
--<ul>
--/li atom dc
--/li atom window
--/li sequence rect (x, y, width, height)
--/li atom pane ( /wxAuiPaneInfo )
--</ul>
public procedure dockart_drawborder_id( atom dockArt, atom rid )
        c_proc( WX_DOCKART_DRAWBORDER_ID, {dockArt,rid})
end procedure

--/topic wxAuiDockArt
--/proc dockart_drawcaption_id( atom dockArt, atom rid )
--Sets the /b DrawCaption routine id. To unset, pass -1 for /b rid.
--
--Your /b DrawCaption procedure should take 5 parameters:
--<ul>
--/li atom dc
--/li atom window
--/li sequence text
--/li sequence rect (x, y, width, height)
--/li atom pane ( /wxAuiPaneInfo )
--</ul>
public procedure dockart_drawcaption_id( atom dockArt, atom rid )
        c_proc( WX_DOCKART_DRAWCAPTION_ID, {dockArt,rid})
end procedure

--/topic wxAuiDockArt
--/proc dockart_drawgripper_id( atom dockArt, atom rid )
--Sets the /b DrawGripper routine id. To unset, pass -1 for /b rid.
--
--Your /b DrawGripper procedure should take 4 parameters:
--<ul>
--/li atom dc
--/li atom window
--/li sequence rect (x, y, width, height)
--/li atom pane ( /wxAuiPaneInfo )
--</ul>
public procedure dockart_drawgripper_id( atom dockArt, atom rid )
        c_proc( WX_DOCKART_DRAWGRIPPER_ID, {dockArt,rid})
end procedure

--/topic wxAuiDockArt
--/proc dockart_drawpanebutton_id( atom dockArt, atom rid )
--Sets the /b DrawPaneButton routine id. To unset, pass -1 for /b rid.
--
--Your /b DrawPaneButton procedure should take 6 parameters:
--<ul>
--/li atom dc
--/li atom window
--/li integer button
--/li integer button_state
--/li sequence rect (x, y, width, height)
--/li atom pane ( /wxAuiPaneInfo )
--</ul>
public procedure dockart_drawpanebutton_id( atom dockArt, atom rid )
        c_proc( WX_DOCKART_DRAWPANEBUTTON_ID, {dockArt,rid})
end procedure

--/topic wxAuiDockArt
--/proc dockart_drawsash_id( atom dockArt, atom rid )
--Sets the /b DrawSash routine id. To unset, pass -1 for /b rid.
--
--Your /b DrawSash procedure should take 4 parameters:
--<ul>
--/li atom dc
--/li atom window
--/li integer orientation
--/li sequence rect (x, y, width, height)
--</ul>
public procedure dockart_drawsash_id( atom dockArt, atom rid )
        c_proc( WX_DOCKART_DRAWSASH_ID, {dockArt,rid})
end procedure

--/topic wxAuiDockArt
--/func get_dockart_colour( atom dockArt, atom id )
--Get the colour of a certain setting.
--
--/i id can be one of the following values:
--<ul>
--/li wxAUI_DOCKART_BACKGROUND_COLOUR
--/li wxAUI_DOCKART_SASH_COLOUR
--/li wxAUI_DOCKART_ACTIVE_CAPTION_COLOUR
--/li wxAUI_DOCKART_ACTIVE_CAPTION_GRADIENT_COLOUR
--/li wxAUI_DOCKART_INACTIVE_CAPTION_COLOUR
--/li wxAUI_DOCKART_INACTIVE_CAPTION_GRADIENT_COLOUR
--/li wxAUI_DOCKART_ACTIVE_CAPTION_TEXT_COLOUR
--/li wxAUI_DOCKART_INACTIVE_CAPTION_TEXT_COLOUR
--/li wxAUI_DOCKART_BORDER_COLOUR
--/li wxAUI_DOCKART_GRIPPER_COLOUR
--</ul>
public function get_dockart_colour( atom dockArt, atom id )
        return c_func( WX_GET_DOCKART_COLOUR, {dockArt,id})
end function

--/topic wxAuiDockArt
--/func get_dockart_font( atom dockArt, atom id )
--Get a font setting.
--
--/i id should always be wxAUI_DOCKART_CAPTION_FONT.
public function get_dockart_font( atom dockArt, atom id )
        return c_func( WX_GET_DOCKART_FONT, {dockArt,id})
end function

--/topic wxAuiDockArt
--/func get_dockart_metric( atom dockArt, atom id )
--Get the value of a certain setting.
--
--/i id can be one of the following values:
--<ul>
--/li wxAUI_DOCKART_SASH_SIZE
--/li wxAUI_DOCKART_CAPTION_SIZE
--/li wxAUI_DOCKART_GRIPPER_SIZE
--/li wxAUI_DOCKART_PANE_BORDER_SIZE
--/li wxAUI_DOCKART_PANE_BUTTON_SIZE
--</ul>
public function get_dockart_metric( atom dockArt, atom id )
        return c_func( WX_GET_DOCKART_METRIC, {dockArt,id})
end function

--/topic wxAuiDockArt
--/proc set_dockart_colour( atom dockArt, atom id, atom colour )
--Set a certain setting with the value /i colour.
--
--/i id can be one of the following values:
--<ul>
--/li wxAUI_DOCKART_BACKGROUND_COLOUR
--/li wxAUI_DOCKART_SASH_COLOUR
--/li wxAUI_DOCKART_ACTIVE_CAPTION_COLOUR
--/li wxAUI_DOCKART_ACTIVE_CAPTION_GRADIENT_COLOUR
--/li wxAUI_DOCKART_INACTIVE_CAPTION_COLOUR
--/li wxAUI_DOCKART_INACTIVE_CAPTION_GRADIENT_COLOUR
--/li wxAUI_DOCKART_ACTIVE_CAPTION_TEXT_COLOUR
--/li wxAUI_DOCKART_INACTIVE_CAPTION_TEXT_COLOUR
--/li wxAUI_DOCKART_BORDER_COLOUR
--/li wxAUI_DOCKART_GRIPPER_COLOUR
--</ul>
public procedure set_dockart_colour( atom dockArt, atom id, atom colour )
        c_proc( WX_SET_DOCKART_COLOUR, {dockArt,id,colour})
end procedure

--/topic wxAuiDockArt
--/proc set_dockart_font( atom dockArt, atom id, atom font )
--Set a font setting.
--
--/i id should always be wxAUI_DOCKART_CAPTION_FONT.
public procedure set_dockart_font( atom dockArt, atom id, atom font )
        c_proc( WX_SET_DOCKART_FONT, {dockArt,id,font})
end procedure

--/topic wxAuiDockArt
--/proc set_dockart_metric( atom dockArt, atom id, atom new_val )
--Set a certain setting with the value /i new_val.
--
--/i id can be one of the following values:
--<ul>
--/li wxAUI_DOCKART_SASH_SIZE
--/li wxAUI_DOCKART_CAPTION_SIZE
--/li wxAUI_DOCKART_GRIPPER_SIZE
--/li wxAUI_DOCKART_PANE_BORDER_SIZE
--/li wxAUI_DOCKART_PANE_BUTTON_SIZE
--</ul>
public procedure set_dockart_metric( atom dockArt, atom id, atom new_val )
        c_proc( WX_SET_DOCKART_METRIC, {dockArt,id,new_val})
end procedure

--wxRichTextPrinting
-----------------------------------------
--/topic wxRichTextPrinting
--/parent Classes
--/parent wxRichTextCtrl
--/info
--

--/topic wxRichTextPrinting
--/func get_buffer( atom rich )
--
public function get_buffer( atom rich )
        return c_func( WX_GET_BUFFER, {rich})
end function

--/topic wxRichTextPrinting
--/func get_richprint_footer_text( atom richprint )
--
public function get_richprint_footer_text( atom richprint )
        return c_func( WX_GET_RICHPRINT_FOOTER_TEXT, {richprint})
end function

--/topic wxRichTextPrinting
--/func get_richprint_header_footer_data( atom richprint )
--
public function get_richprint_header_footer_data( atom richprint )
        return c_func( WX_GET_RICHPRINT_HEADER_FOOTER_DATA, {richprint})
end function

--/topic wxRichTextPrinting
--/func get_richprint_header_text( atom richprint )
--
public function get_richprint_header_text( atom richprint )
        return c_func( WX_GET_RICHPRINT_HEADER_TEXT, {richprint})
end function

--/topic wxRichTextPrinting
--/func get_richprint_page_setup_data( atom richprint )
--
public function get_richprint_page_setup_data( atom richprint )
        return c_func( WX_GET_RICHPRINT_PAGE_SETUP_DATA, {richprint})
end function

--/topic wxRichTextPrinting
--/func get_richprint_parent_window( atom richprint )
--
public function get_richprint_parent_window( atom richprint )
        return c_func( WX_GET_RICHPRINT_PARENT_WINDOW, {richprint})
end function

--/topic wxRichTextPrinting
--/func get_richprint_preview_rect( atom richprint )
--
public function get_richprint_preview_rect( atom richprint )
        return c_func( WX_GET_RICHPRINT_PREVIEW_RECT, {richprint})
end function

--/topic wxRichTextPrinting
--/func get_richprint_print_data( atom richprint )
--
public function get_richprint_print_data( atom richprint )
        return c_func( WX_GET_RICHPRINT_PRINT_DATA, {richprint})
end function

--/topic wxRichTextPrinting
--/func get_richprint_title( atom richprint )
--
public function get_richprint_title( atom richprint )
        return c_func( WX_GET_RICHPRINT_TITLE, {richprint})
end function

--/topic wxRichTextPrinting
--/proc show_richprint_page_setup( atom richprint )
--
public procedure show_richprint_page_setup( atom richprint )
        c_proc( WX_SHOW_RICHPRINT_PAGE_SETUP, {richprint})
end procedure

--/topic wxRichTextPrinting
--/proc show_richprint_preview_buffer( atom richprint, atom buffer )
--
public procedure show_richprint_preview_buffer( atom richprint, atom buffer )
        c_proc( WX_SHOW_RICHPRINT_PREVIEW_BUFFER, {richprint,buffer})
end procedure

--/topic wxRichTextPrinting
--/proc show_richprint_preview_file( atom richprint, object richTextFile )
--
public procedure show_richprint_preview_file( atom richprint, object richTextFile )
        c_proc( WX_SHOW_RICHPRINT_PREVIEW_FILE, {richprint,richTextFile})
end procedure

--/topic wxRichTextPrinting
--/proc show_richprint_print_buffer( atom richprint, atom buffer )
--
public procedure show_richprint_print_buffer( atom richprint, atom buffer )
        c_proc( WX_SHOW_RICHPRINT_PRINT_BUFFER, {richprint,buffer})
end procedure

--/topic wxRichTextPrinting
--/proc show_richprint_print_file( atom richprint, object richTextFile )
--
public procedure show_richprint_print_file( atom richprint, object richTextFile )
        c_proc( WX_SHOW_RICHPRINT_PRINT_FILE, {richprint,richTextFile})
end procedure

--/topic wxRichTextPrinting
--/proc set_richprint_footer_text( atom richprint, object text, atom page, atom location )
--
public procedure set_richprint_footer_text( atom richprint, object text, atom page, atom location )
        c_proc( WX_SET_RICHPRINT_FOOTER_TEXT, {richprint,text,page,location})
end procedure

--/topic wxRichTextPrinting
--/proc set_richprint_header_footer_data( atom richprint, atom data )
--
public procedure set_richprint_header_footer_data( atom richprint, atom data )
        c_proc( WX_SET_RICHPRINT_HEADER_FOOTER_DATA, {richprint,data})
end procedure

--/topic wxRichTextPrinting
--/proc set_richprint_header_footer_font( atom richprint, atom font )
--
public procedure set_richprint_header_footer_font( atom richprint, atom font )
        c_proc( WX_SET_RICHPRINT_HEADER_FOOTER_FONT, {richprint,font})
end procedure

--/topic wxRichTextPrinting
--/proc set_richprint_header_footer_text_colour( atom richprint, atom colour )
--
public procedure set_richprint_header_footer_text_colour( atom richprint, atom colour )
        c_proc( WX_SET_RICHPRINT_HEADER_FOOTER_TEXT_COLOUR, {richprint,colour})
end procedure

--/topic wxRichTextPrinting
--/proc set_richprint_header_text( atom richprint, object text, atom page, atom location )
--
public procedure set_richprint_header_text( atom richprint, object text, atom page, atom location )
        c_proc( WX_SET_RICHPRINT_HEADER_TEXT, {richprint,text,page,location})
end procedure

--/topic wxRichTextPrinting
--/proc set_richprint_page_setup_data( atom richprint, atom pageSetupData )
--
public procedure set_richprint_page_setup_data( atom richprint, atom pageSetupData )
        c_proc( WX_SET_RICHPRINT_PAGE_SETUP_DATA, {richprint,pageSetupData})
end procedure

--/topic wxRichTextPrinting
--/proc set_richprint_parent_window( atom richprint, atom parent )
--
public procedure set_richprint_parent_window( atom richprint, atom parent )
        c_proc( WX_SET_RICHPRINT_PARENT_WINDOW, {richprint,parent})
end procedure

--/topic wxRichTextPrinting
--/proc set_richprint_preview_rect( atom richprint, object rect )
--
public procedure set_richprint_preview_rect( atom richprint, object rect )
        c_proc( WX_SET_RICHPRINT_PREVIEW_RECT, {richprint,rect})
end procedure

--/topic wxRichTextPrinting
--/proc set_richprint_print_data( atom richprint, atom printData )
--
public procedure set_richprint_print_data( atom richprint, atom printData )
        c_proc( WX_SET_RICHPRINT_PRINT_DATA, {richprint,printData})
end procedure

--/topic wxRichTextPrinting
--/proc show_richprint_on_first_page( atom richprint, atom show )
--
public procedure show_richprint_on_first_page( atom richprint, atom show )
        c_proc( WX_SHOW_RICHPRINT_ON_FIRST_PAGE, {richprint,show})
end procedure

--/topic wxRichTextPrinting
--/proc set_richprint_title( atom richprint, object title )
--
public procedure set_richprint_title( atom richprint, object title )
        c_proc( WX_SET_RICHPRINT_TITLE, {richprint,title})
end procedure

--/topic Key Events
--/func get_unicode_key( atom event )
--
public function get_unicode_key( atom event )
	return c_func( WX_GET_UNICODE_KEY, {event})
end function

--wxToolBarItem
-----------------------------------------
--/topic wxToolBarItem
--/parent wxToolBar
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li "atom parent"
--/li integer id
--/li sequence label
--/li atom /wxBitmap [= 0] the primary tool bitmap
--/li atom /wxBitmap [= 0] The bitmap used then the tool is disabled.
-- If it is equal to wxNullBitmap, the disabled bitmap is automatically generated by greing the normal one.
--/li integer kind [= wxITEM_NORMAL] May be wxITEM_NORMAL for a normal button
-- (default), wxITEM_CHECK for a checkable tool (such tool stays pressed after it
-- had been toggled) or wxITEM_RADIO for a checkable tool which makes part of a
-- radio group of tools each of which is automatically unchecked whenever another
-- button in the group is checked
--/li sequence short_help [= ""]  This string is used for the tools tooltip
--/li sequence long_help [= ""]  This string is shown in the statusbar (if any) of the parent frame when the mouse pointer is inside the tool
--</ul>
-- This is equivalent to /add_tool.

--Advanced Graphics
--/topic Advanced Graphics
--/parent Graphics
--/info
--
--These classes and routines use advanced graphics rendering provided by the native platform.
-- On Windows, GDI+, and on GTK, Cairo.

--wxGraphicsMatrix
-----------------------------------------
--/topic wxGraphicsMatrix
--/parent Advanced Graphics
--/info
-- A wxGraphicsMatrix is a native representation of an affine matrix. The
-- contents are specific and private to the respective renderer. Instances are ref
-- counted and can therefore be assigned as usual. The default creates an identity
-- matrix.
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li a [= 1.0]
--/li b [= 0.0]
--/li c [= 0.0]
--/li d [= 1.0]
--/li tx [= 0.0]
--/li ty [= 0.0]
--</ul>

--/topic wxGraphicsMatrix
--/proc graphics_matrix_concat( atom m1, atom m2 )
--Concatenates the matrix passed with the current matrix.
public procedure graphics_matrix_concat( atom m1, atom m2 )
	c_proc( WX_GRAPHICS_MATRIX_CONCAT, {m1,m2})
end procedure

--/topic wxGraphicsMatrix
--/func graphics_matrix_get( atom matrix )
--Returns the component values of the matrix as a sequence (a, b, c, d, tx, ty).
public function graphics_matrix_get( atom matrix )
	return c_func( WX_GRAPHICS_MATRIX_GET, {matrix})
end function

--/topic wxGraphicsMatrix
--/func graphics_matrix_get_native( atom matrix )
--Returns the native representation of the matrix. For CoreGraphics this is a CFAffineMatrix pointer. For GDIPlus a Matrix Pointer and for Cairo a cairo_matrix_t pointer.
public function graphics_matrix_get_native( atom matrix )
	return c_func( WX_GRAPHICS_MATRIX_GET_NATIVE, {matrix})
end function

--/topic wxGraphicsMatrix
--/proc graphics_matrix_invert( atom matrix )
--Inverts the matrix.
public procedure graphics_matrix_invert( atom matrix )
	c_proc( WX_GRAPHICS_MATRIX_INVERT, {matrix})
end procedure

--/topic wxGraphicsMatrix
--/func graphics_matrix_equal( atom matrix, atom other )
--Returns true if the elements of the transformation matrix are equal.
public function graphics_matrix_equal( atom matrix, atom other )
	return c_func( WX_GRAPHICS_MATRIX_EQUAL, {matrix,other})
end function

--/topic wxGraphicsMatrix
--/func graphics_matrix_identity( atom matrix )
--Return true if this is the identity matrix.
public function graphics_matrix_identity( atom matrix )
	return c_func( WX_GRAPHICS_MATRIX_IDENTITY, {matrix})
end function

--/topic wxGraphicsMatrix
--/proc graphics_matrix_rotate( atom matrix, atom angle )
--Rotates this matrix (radians).
public procedure graphics_matrix_rotate( atom matrix, atom angle )
	c_proc( WX_GRAPHICS_MATRIX_ROTATE, {matrix,angle})
end procedure

--/topic wxGraphicsMatrix
--/proc graphics_matrix_scale( atom matrix, atom x, atom y )
--Scales this matrix.
public procedure graphics_matrix_scale( atom matrix, atom x, atom y )
	c_proc( WX_GRAPHICS_MATRIX_SCALE, {matrix,x,y})
end procedure

--/topic wxGraphicsMatrix
--/proc graphics_matrix_translate( atom matrix, atom x, atom y )
--Translates this matrix.
public procedure graphics_matrix_translate( atom matrix, atom x, atom y )
	c_proc( WX_GRAPHICS_MATRIX_TRANSLATE, {matrix,x,y})
end procedure

--/topic wxGraphicsMatrix
--/proc graphics_matrix_set( atom matrix, atom a, atom b, atom c, atom d, atom tx, atom ty )
--Sets the matrix to the respective values
public procedure graphics_matrix_set( atom matrix, atom a, atom b, atom c, atom d, atom tx, atom ty )
	c_proc( WX_GRAPHICS_MATRIX_SET, {matrix,a,b,c,d,tx,ty})
end procedure

--/topic wxGraphicsMatrix
--/func graphics_matrix_transform_point( atom matrix, atom x, atom y )
--Applies this matrix to a point.
public function graphics_matrix_transform_point( atom matrix, atom x, atom y )
	return c_func( WX_GRAPHICS_MATRIX_TRANSFORM_POINT, {matrix,x,y})
end function

--/topic wxGraphicsMatrix
--/func graphics_matrix_transform_distance( atom matrix, atom dx, atom dy )
--Applies this matrix to a distance (ie. performs all transforms except translations)
public function graphics_matrix_transform_distance( atom matrix, atom dx, atom dy )
	return c_func( WX_GRAPHICS_MATRIX_TRANSFORM_DISTANCE, {matrix,dx,dy})
end function

--/topic wxGraphicsMatrix
--/func graphics_matrix_renderer( atom matrix )
--Returns the renderer that was used to create this instance, or NULL if it has not been initialized yet
public function graphics_matrix_renderer( atom matrix )
	return c_func( WX_GRAPHICS_MATRIX_RENDERER, {matrix})
end function

--/topic wxGraphicsMatrix
--/func graphics_matrix_isnull( atom matrix )
--Is this object valid (false) or still empty (true)?
public function graphics_matrix_isnull( atom matrix )
	return c_func( WX_GRAPHICS_MATRIX_ISNULL, {matrix})
end function

--/topic wxMenu
--/func find_menuitem_by_id( atom menu, atom id )
--
-- Returns a pointer to the /wxMenuItem specified by the id.
public function find_menuitem_by_id( atom menu, atom id )
	return c_func( WX_FIND_MENUITEM_BY_ID, {menu,id})
end function

--/topic Advanced Graphics
--/func get_graphics_object_renderer( atom graphics_object )
--
--Returns the renderer that was used to create this instance, or NULL if it has not been initialized yet
public function get_graphics_object_renderer( atom graphics_object )
	return c_func( WX_GET_GRAPHICS_OBJECT_RENDERER, {graphics_object})
end function

--/topic Advanced Graphics
--/func graphics_object_isnull( atom graphics_object )
--
-- Is this object valid (false) or still empty (true)?
public function graphics_object_isnull( atom graphics_object )
	return c_func( WX_GRAPHICS_OBJECT_ISNULL, {graphics_object})
end function

public enum
	GC_DC = 0,
	GC_WINDOW,
	GC_NATIVE_CONTEXT,
	GC_NATIVE_WINDOW

--wxGraphicsContext
-----------------------------------------
--/topic wxGraphicsContext
--/parent Advanced Graphics
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li integer from_type: one of GC_DC, GC_WINDOW, GC_NATIVE_CONTEXT, GC_NATIVE_WINDOW
--/li atom from_ptr: The /wxDC, /wxWindow, native context or native window
--</ul>
--
-- A wxGraphicsContext instance is the object that is drawn upon. It is created by
-- a renderer using the CreateContext calls.., this can be either directly using a
-- renderer instance, or indirectly using the static convenience CreateXXX
-- functions of wxGraphicsContext that always delegate the task to the default
-- renderer.


--wxGraphicsPen
-----------------------------------------
--/topic wxGraphicsPen
--/parent Advanced Graphics
--/info
--
-- A wxGraphicsPen is a native representation of a pen. It is used for stroking a
-- path on a graphics context. The contents are specific and private to the
-- respective renderer. Instances are ref counted and can therefore be assigned as
-- usual. The only way to get a valid instance is via a CreatePen call on the
-- graphics context or the renderer instance.

--wxGraphicsBrush
-----------------------------------------
--/topic wxGraphicsBrush
--/parent Auto
--/info
--
-- A wxGraphicsBrush is a native representation of a brush. It is used for filling
-- a path on a graphics context. The contents are specific an private to the
-- respective renderer. Instances are ref counted and can therefore be assigned as
-- usual. The only way to get a valid instance is via a Create...Brush call on the
-- graphics context or the renderer instance.

--/topic Advanced Graphics
--/func get_default_graphics_renderer(  )
--

public function get_default_graphics_renderer(  )
	return c_func( WX_GET_DEFAULT_GRAPHICS_RENDERER, {})
end function

--/topic Advanced Graphics
--/func create_linear_gradient_brush( atom renderer, atom x1, atom y1, atom x2, atom y2, atom c1, atom c2 )
--
-- Creates a native brush, having a radial gradient originating at (xo,yc) with
-- color oColour and ends on a circle around (xc,yc) with radius r and color
-- cColour
public function create_linear_gradient_brush( atom renderer, atom x1, atom y1, atom x2, atom y2, atom c1, atom c2 )
	return c_func( WX_CREATE_LINEAR_GRADIENT_BRUSH, {renderer,x1,y1,x2,y2,c1,c2})
end function

--/topic Advanced Graphics
--/func create_radial_gradient_brush( atom renderer, atom xo, atom yo, atom xc, atom yc, atom radius, atom o_color, atom c_color )
--
-- Creates a native brush, having a linear gradient, starting at (x1,y1) with color
-- c1 to (x2,y2) with color c2
public function create_radial_gradient_brush( atom renderer, atom xo, atom yo, atom xc, atom yc, atom radius, atom o_color, atom c_color )
	return c_func( WX_CREATE_RADIAL_GRADIENT_BRUSH, {renderer,xo,yo,xc,yc,radius,o_color,c_color})
end function

--/topic Advanced Graphics
--/func create_graphics_font( atom renderer, atom font, atom color )
--
-- Creates a native graphics font from a wxFont and a text colour.
public function create_graphics_font( atom renderer, atom font, atom color )
	return c_func( WX_CREATE_GRAPHICS_FONT, {renderer,font,color})
end function

--wxGraphicsPath
-----------------------------------------
--/topic wxGraphicsPath
--/parent Advanced Graphics
--/parent Classes
--/info
--
-- Creation Parameters:
--/li atom renderer [=default graphics renderer] The graphics renderer to use
--
-- A wxGraphicsPath is a native representation of an geometric path. The contents
-- are specific an private to the respective renderer. Instances are ref counted
-- and can therefore be assigned as usual. The only way to get a valid instance is
-- via a CreatePath call on the graphics context or the renderer instance.instance

--/topic wxGraphicsPath
--/proc path_move_to( atom path, atom x, atom y )
--
-- Begins a new subpath at (x,y)
public procedure path_move_to( atom path, atom x, atom y )
	c_proc( WX_PATH_MOVE_TO, {path,x,y})
end procedure

--/topic wxGraphicsPath
--/proc path_add_arc( atom path, atom x, atom y, atom r, atom start_angle, atom end_angle, atom clockwise )
--
-- Adds an arc of a circle centering at (x,y) with radius (r) from startAngle to endAngle.
public procedure path_add_arc( atom path, atom x, atom y, atom r, atom start_angle, atom end_angle, atom clockwise )
	c_proc( WX_PATH_ADD_ARC, {path,x,y,r,start_angle,end_angle,clockwise})
end procedure

--/topic wxGraphicsPath
--/proc path_add_arc_to_point( atom path, atom x1, atom y1, atom x2, atom y2, atom r )
--
-- Appends a an arc to two tangents connecting (current) to (x1,y1) and (x1,y1) to (x2,y2), also a straight line from (current) to (x1,y1).
public procedure path_add_arc_to_point( atom path, atom x1, atom y1, atom x2, atom y2, atom r )
	c_proc( WX_PATH_ADD_ARC_TO_POINT, {path,x1,y1,x2,y2,r})
end procedure

--/topic wxGraphicsPath
--/proc path_add_circle( atom path, atom x, atom y, atom r )
--
-- Appends a circle around (x,y) with radius r as a new closed subpath.
public procedure path_add_circle( atom path, atom x, atom y, atom r )
	c_proc( WX_PATH_ADD_CIRCLE, {path,x,y,r})
end procedure

--/topic wxGraphicsPath
--/proc path_add_curve_to_point( atom path, atom cx1, atom cy1, atom cx2, atom cy2, atom x, atom y )
--
-- Adds a cubic Bezier curve from the current point, using two control points and an end point.
public procedure path_add_curve_to_point( atom path, atom cx1, atom cy1, atom cx2, atom cy2, atom x, atom y )
	c_proc( WX_PATH_ADD_CURVE_TO_POINT, {path,cx1,cy1,cx2,cy2,x,y})
end procedure

--/topic wxGraphicsPath
--/proc path_add_ellipse( atom path, atom x, atom y, atom w, atom h )
--
-- Appends an ellipse fitting into the passed in rectangle.
public procedure path_add_ellipse( atom path, atom x, atom y, atom w, atom h )
	c_proc( WX_PATH_ADD_ELLIPSE, {path,x,y,w,h})
end procedure

--/topic wxGraphicsPath
--/proc path_add_line_to_point( atom path, atom x, atom y )
--
-- Adds a straight line from the current point to (x,y).
public procedure path_add_line_to_point( atom path, atom x, atom y )
	c_proc( WX_PATH_ADD_LINE_TO_POINT, {path,x,y})
end procedure

--/topic wxGraphicsPath
--/proc path_add( atom path, atom path_to_add )
--
-- Adds another path.
public procedure path_add( atom path, atom path_to_add )
	c_proc( WX_PATH_ADD, {path,path_to_add})
end procedure

--/topic wxGraphicsPath
--/proc path_add_quad_curve_to_point( atom path, atom cx, atom cy, atom x, atom y )
--
-- Adds a quadratic Bezier curve from the current point, using a control point and an end point.
public procedure path_add_quad_curve_to_point( atom path, atom cx, atom cy, atom x, atom y )
	c_proc( WX_PATH_ADD_QUAD_CURVE_TO_POINT, {path,cx,cy,x,y})
end procedure

--/topic wxGraphicsPath
--/proc path_add_rectangle( atom path, atom x, atom y, atom w, atom h )
--
-- Appends a rectangle as a new closed subpath.
public procedure path_add_rectangle( atom path, atom x, atom y, atom w, atom h )
	c_proc( WX_PATH_ADD_RECTANGLE, {path,x,y,w,h})
end procedure

--/topic wxGraphicsPath
--/proc path_add_rounded_rectangle( atom path, atom x, atom y, atom w, atom h, atom radius )
--
-- Appends a rounded rectangle as a new closed subpath.
public procedure path_add_rounded_rectangle( atom path, atom x, atom y, atom w, atom h, atom radius )
	c_proc( WX_PATH_ADD_ROUNDED_RECTANGLE, {path,x,y,w,h,radius})
end procedure

--/topic wxGraphicsPath
--/proc path_close_subpath( atom path )
--
-- Closes the current sub-path.
public procedure path_close_subpath( atom path )
	c_proc( WX_PATH_CLOSE_SUBPATH, {path})
end procedure

--/topic wxGraphicsPath
--/func path_contains( atom path, atom x, atom y, atom fill_style )
--
-- Returns true if the point is within the path.
public function path_contains( atom path, atom x, atom y, atom fill_style )
	return c_func( WX_PATH_CONTAINS, {path,x,y,fill_style})
end function

--/topic wxGraphicsPath
--/func path_get_box( atom path )
--
-- Gets the bounding box enclosing all points (possibly including control points).
public function path_get_box( atom path )
	return c_func( WX_PATH_GET_BOX, {path})
end function

--/topic wxGraphicsPath
--/func path_get_current_point( atom path )
--
-- Gets the last point of the current path, {0,0} if not yet set.
public function path_get_current_point( atom path )
	return c_func( WX_PATH_GET_CURRENT_POINT, {path})
end function

--/topic wxGraphicsPath
--/func path_transform( atom path, atom matrix )
--
-- Transforms each point of this path by the matrix.
public function path_transform( atom path, atom matrix )
	return c_func( WX_PATH_TRANSFORM, {path,matrix})
end function

--/topic wxGraphicsPath
--/func path_get_native( atom path )
--
-- Returns the native path (CGPathRef for Core Graphics, Path pointer for GDIPlus and a cairo_path_t pointer for cairo).
public function path_get_native( atom path )
	return c_func( WX_PATH_GET_NATIVE, {path})
end function

--/topic wxGraphicsPath
--/func path_unget_native_path( atom path, atom native )
--
-- Gives back the native path returned by GetNativePath() because there might be
-- some deallocations necessary (eg on cairo the native path returned by
-- GetNativePath is newly allocated each time).
public function path_unget_native_path( atom path, atom native )
	return c_func( WX_PATH_UNGET_NATIVE_PATH, {path,native})
end function

--/topic wxGraphicsContext
--/proc gc_clip( atom context, atom x, atom y, atom w, atom h )
--
-- Clips drawings to the rectangle.
public procedure gc_clip( atom context, atom x, atom y, atom w, atom h )
	c_proc( WX_GC_CLIP, {context,x,y,w,h})
end procedure

--/topic wxGraphicsContext
--/proc gc_reset_clip( atom context )
--
-- Resets the clipping to original shape.
public procedure gc_reset_clip( atom context )
	c_proc( WX_GC_RESET_CLIP, {context})
end procedure

--/topic wxGraphicsContext
--/proc gc_draw_bitmap( atom context, atom bmp, atom x, atom y, atom w, atom h )
--
-- Draws the bitmap. In case of a mono bitmap, this is treated as a mask and the current brushed is used for filling.
public procedure gc_draw_bitmap( atom context, atom bmp, atom x, atom y, atom w, atom h )
	c_proc( WX_GC_DRAW_BITMAP, {context,bmp,x,y,w,h})
end procedure

--/topic wxGraphicsContext
--/proc gc_draw_ellipse( atom context, atom x, atom y, atom w, atom h )
--
-- Draws an ellipse.
public procedure gc_draw_ellipse( atom context, atom x, atom y, atom w, atom h )
	c_proc( WX_GC_DRAW_ELLIPSE, {context,x,y,w,h})
end procedure

--/topic wxGraphicsContext
--/proc gc_draw_icon( atom context, atom icon, atom x, atom y, atom w, atom h )
--
-- Draws the icon.
public procedure gc_draw_icon( atom context, atom icon, atom x, atom y, atom w, atom h )
	c_proc( WX_GC_DRAW_ICON, {context,icon,x,y,w,h})
end procedure

--/topic wxGraphicsContext
--/proc gc_draw_path( atom context, atom path, atom fill_style [=wxODDEVEN_RULE] )
--
-- Draws the path by first filling and then stroking.
public procedure gc_draw_path( atom context, atom path, atom fill_style = wxODDEVEN_RULE )
	c_proc( WX_GC_DRAW_PATH, {context,path,fill_style})
end procedure

--/topic wxGraphicsContext
--/proc gc_draw_rectangle( atom context, atom x, atom y, atom w, atom h )
--
-- Draws a rectangle.
public procedure gc_draw_rectangle( atom context, atom x, atom y, atom w, atom h )
	c_proc( WX_GC_DRAW_RECTANGLE, {context,x,y,w,h})
end procedure

--/topic wxGraphicsContext
--/proc gc_draw_rounded_rectangle( atom context, atom x, atom y, atom w, atom h, atom radius )
--
-- Draws a rounded rectangle.
public procedure gc_draw_rounded_rectangle( atom context, atom x, atom y, atom w, atom h, atom radius )
	c_proc( WX_GC_DRAW_ROUNDED_RECTANGLE, {context,x,y,w,h,radius})
end procedure

--/topic wxGraphicsContext
--/proc gc_draw_text( atom context, object str, atom x, atom y, atom angle )
--
-- Draws a text at the defined position, at the given angle.
public procedure gc_draw_text( atom context, object str, atom x, atom y, atom angle )
	c_proc( WX_GC_DRAW_TEXT, {context,str,x,y,angle})
end procedure

--/topic wxGraphicsContext
--/proc gc_fill_path( atom context, atom path, atom fill_style )
--
-- Fills the path with the current brush.
public procedure gc_fill_path( atom context, atom path, atom fill_style )
	c_proc( WX_GC_FILL_PATH, {context,path,fill_style})
end procedure

--/topic wxGraphicsContext
--/proc gc_stroke_path( atom context, atom path )
--
-- Strokes along a path with the current pen.
public procedure gc_stroke_path( atom context, atom path )
	c_proc( WX_GC_STROKE_PATH, {context,path})
end procedure

--/topic wxGraphicsContext
--/func gc_get_native_context( atom context )
--
-- Returns the native context (CGContextRef for Core Graphics, Graphics pointer for GDIPlus and cairo_t pointer for cairo).
public function gc_get_native_context( atom context )
	return c_func( WX_GC_GET_NATIVE_CONTEXT, {context})
end function

--/topic wxGraphicsContext
--/func gc_get_partial_text_extents( atom context, object text )
--
-- Returns a sequence of widths with the widths from the beginning of text to the corresponding character of text.
public function gc_get_partial_text_extents( atom context, object text )
	return c_func( WX_GC_GET_PARTIAL_TEXT_EXTENTS, {context,text})
end function

--/topic wxGraphicsContext
--/func gc_get_text_extent( atom context, object text )
--
-- Gets the dimensions of the string using the currently selected font. string is
-- the text string to measure, w and h are the total width and height respectively,
-- descent is the dimension from the baseline of the font to the bottom of the
-- descender, and externalLeading is any extra vertical space added to the font by
-- the font designer (usually is zero).
--
-- Returns a sequence: { width, height, descent, external_leading }
public function gc_get_text_extent( atom context, object text )
	return c_func( WX_GC_GET_TEXT_EXTENT, {context,text})
end function

--/topic wxGraphicsContext
--/proc gc_rotate( atom context, atom angle )
--
-- Rotates the current transformation matrix (radians).
public procedure gc_rotate( atom context, atom angle )
	c_proc( WX_GC_ROTATE, {context,angle})
end procedure

--/topic wxGraphicsContext
--/proc gc_scale( atom context, atom x_scale, atom y_scale )
--
-- Scales the current transformation matrix.
public procedure gc_scale( atom context, atom x_scale, atom y_scale )
	c_proc( WX_GC_SCALE, {context,x_scale,y_scale})
end procedure

--/topic wxGraphicsContext
--/proc gc_translate( atom context, atom dx, atom dy )
--
-- Translates the current transformation matrix.
public procedure gc_translate( atom context, atom dx, atom dy )
	c_proc( WX_GC_TRANSLATE, {context,dx,dy})
end procedure

--/topic wxGraphicsContext
--/func gc_get_transform( atom context )
--
-- Gets the current transformation matrix of this context.
-- Returns a /wxGraphicsMatrix.
public function gc_get_transform( atom context )
	return c_func( WX_GC_GET_TRANSFORM, {context})
end function

--/topic wxGraphicsContext
--/proc gc_set_transform( atom context, atom matrix )
--
-- Sets the current transformation matrix of this context
public procedure gc_set_transform( atom context, atom matrix )
	c_proc( WX_GC_SET_TRANSFORM, {context,matrix})
end procedure

--/topic wxGraphicsContext
--/proc gc_concat_transform( atom context, atom matrix )
--
-- Concatenates the passed in transform with the current transform of this context
public procedure gc_concat_transform( atom context, atom matrix )
	c_proc( WX_GC_CONCAT_TRANSFORM, {context,matrix})
end procedure

--/topic wxGraphicsContext
--/proc gc_set_brush( atom context, atom brush )
--
-- Sets the brush for filling paths.  Note that brush is a /wxBrush.
public procedure gc_set_brush( atom context, atom brush )
	c_proc( WX_GC_SET_BRUSH, {context,brush})
end procedure

--/topic wxGraphicsContext
--/proc gc_set_graphics_brush( atom context, atom brush )
--
-- Sets the brush for filling paths.  Note that brush is a /wxGraphicsBrush.
public procedure gc_set_graphics_brush( atom context, atom brush )
	c_proc( WX_GC_SET_GRAPHICS_BRUSH, {context,brush})
end procedure

--/topic wxGraphicsContext
--/proc gc_set_font( atom context, atom font, atom color )
--
-- Sets the font for drawing text.  Note that font is a /wxFont.
public procedure gc_set_font( atom context, atom font, atom color )
	c_proc( WX_GC_SET_FONT, {context,font,color})
end procedure

--/topic wxGraphicsContext
--/proc gc_set_graphics_font( atom context, atom font )
--
-- Sets the font for drawing text.  Note that font is a /wxGraphicsFont.
public procedure gc_set_graphics_font( atom context, atom font )
	c_proc( WX_GC_SET_GRAPHICS_FONT, {context,font})
end procedure

--/topic wxGraphicsContext
--/proc gc_set_pen( atom context, atom pen )
--
-- Sets the pen used for stroking. Note that pen is a /wxPen.
public procedure gc_set_pen( atom context, atom pen )
	c_proc( WX_GC_SET_PEN, {context,pen})
end procedure

--/topic wxGraphicsContext
--/proc gc_set_graphics_pen( atom context, atom pen )
--
-- Sets the pen used for stroking. Note that pen is a /wxGraphicsPen.
public procedure gc_set_graphics_pen( atom context, atom pen )
	c_proc( WX_GC_SET_GRAPHICS_PEN, {context,pen})
end procedure

--/topic wxGraphicsContext
--/proc gc_stroke_line( atom context, atom x1, atom y1, atom x2, atom y2 )
--
-- Strokes a single line.
public procedure gc_stroke_line( atom context, atom x1, atom y1, atom x2, atom y2 )
	c_proc( WX_GC_STROKE_LINE, {context,x1,y1,x2,y2})
end procedure

--/topic wxGraphicsContext
--/proc gc_stroke_lines( atom context, object points )
--
-- Stroke disconnected lines from begin to end points, fastest method available for this purpose.
public procedure gc_stroke_lines( atom context, object points )
	c_proc( WX_GC_STROKE_LINES, {context,points})
end procedure

--wxArrayStringProperty
-----------------------------------------
--/topic wxArrayStringProperty
--/parent Auto
--/info
--

--wxBoolProperty
-----------------------------------------
--/topic wxBoolProperty
--/parent Auto
--/info
--

--wxColourProperty
-----------------------------------------
--/topic wxColourProperty
--/parent Auto
--/info
--

--wxCursorProperty
-----------------------------------------
--/topic wxCursorProperty
--/parent Auto
--/info
--

--wxDateProperty
-----------------------------------------
--/topic wxDateProperty
--/parent Auto
--/info
--

--wxDirProperty
-----------------------------------------
--/topic wxDirProperty
--/parent Auto
--/info
--

--wxEditEnumProperty
-----------------------------------------
--/topic wxEditEnumProperty
--/parent Auto
--/info
--

--wxEnumProperty
-----------------------------------------
--/topic wxEnumProperty
--/parent Auto
--/info
--

--wxFileProperty
-----------------------------------------
--/topic wxFileProperty
--/parent Auto
--/info
--

--wxFloatProperty
-----------------------------------------
--/topic wxFloatProperty
--/parent Auto
--/info
--

--wxFontProperty
-----------------------------------------
--/topic wxFontProperty
--/parent Auto
--/info
--

--wxImageFileProperty
-----------------------------------------
--/topic wxImageFileProperty
--/parent Auto
--/info
--

--wxIntProperty
-----------------------------------------
--/topic wxIntProperty
--/parent Auto
--/info
--

--wxLongStringProperty
-----------------------------------------
--/topic wxLongStringProperty
--/parent Auto
--/info
--

--wxMultiChoiceProperty
-----------------------------------------
--/topic wxMultiChoiceProperty
--/parent Auto
--/info
--

--wxPropertyCategory
-----------------------------------------
--/topic wxPropertyCategory
--/parent Auto
--/info
--

--wxStringProperty
-----------------------------------------
--/topic wxStringProperty
--/parent Auto
--/info
--

--wxSystemColourProperty
-----------------------------------------
--/topic wxSystemColourProperty
--/parent Auto
--/info
--

--wxUIntProperty
-----------------------------------------
--/topic wxUIntProperty
--/parent Auto
--/info
--

--wxPropertyGrid
-----------------------------------------
--/topic wxPropertyGrid
--/parent Auto
--/info
--

--/topic Auto
--/proc propgrid_append( atom pg, atom property )
--
public procedure propgrid_append( atom pg, atom property )
	c_proc( WX_PROPGRID_APPEND, {pg,property})
end procedure

--wxFlagsProperty
-----------------------------------------
--/topic wxFlagsProperty
--/parent Auto
--/info
--

--/topic Auto
--/func pgprop_add_choice( atom prop, sequence prop_label, atom value )
--
public function pgprop_add_choice( atom prop, sequence prop_label, atom value )
	return c_func( WX_PGPROP_ADD_CHOICE, {prop,prop_label,value})
end function

--/topic Auto
--/func pgprop_get_value( atom prop )
--
public function pgprop_get_value( atom prop )
	return c_func( WX_PGPROP_GET_VALUE, {prop})
end function

--/topic Controls
--/proc set_min_size( atom window, atom x, atom y )
--
-- Sets the minimum size of the window, to indicate to the sizer layout mechanism
-- that this is the minimum required size. You may need to call this if you change
-- the window size after construction and before adding to its parent sizer.
public procedure set_min_size( atom window, atom x, atom y )
	c_proc( WX_SET_MIN_SIZE, {window,x,y})
end procedure

--/topic wxTreeCtrl
--/proc set_tree_compare_rid( atom tree, atom rid )
--
-- Routine id for a sort comparison function to be called when
-- the wxTreeCtrl sorts items. The routine should take 3 parameters:
-- the tree and the two items. The routine should return -1, 0 or 1
-- based on how the two items compare.
public procedure set_tree_compare_rid( atom tree, atom rid )
	c_proc( WX_SET_TREE_COMPARE_RID, {tree,rid})
end procedure

--@routines

public sequence PATH
integer slash
PATH = command_line()
if length(PATH) then
    PATH = reverse(PATH[2])
else
    PATH = ""
end if

ifdef WINDOWS then
        slash = '\\'
elsedef
    slash = '/'
end ifdef

if find(slash, PATH) then
    PATH = stdseq:reverse( PATH[find(slash, PATH)..$] )
else
    PATH = ""
end if

function get_the_object()
    return c_func( WX_GET_THE_OBJECT, {} )
end function

include std/console.e
function callProc( atom rid )
        object o = get_the_object()
        if atom( o ) then
                o = {o}
        end if
        call_proc( rid, o )
    return 0
end function
constant callProcCB = call_back('+' & routine_id("callProc"))


function callFunc( atom rid )

    return call_func( rid, get_the_object() )
end function
constant callFuncCB = call_back('+' & routine_id("callFunc"))

procedure error_call_back( sequence message)
        error:crash( message )
end procedure


procedure initialize()
    integer build
    build = c_func( WX_INITIALIZE, {callFuncCB, callProcCB, routine_id("error_call_back")} )
    if wxNEED_BUILD > build then
        printf(2, "wxEuphoria Version %d.%d.%d requires at least library build %d, but build %d was loaded.\n",
            wxEuphoria_Version & {wxNEED_BUILD, build})
    end if
end procedure


initialize()

constant get_ptr = define_c_func( wxWidgets, "get_obj_pointer", {dll:E_OBJECT}, dll:E_OBJECT)
public function get_obj_ptr( object eu )
    atom ptr
    ptr = c_func( get_ptr, {eu})
    return ptr
end function



--Advanced User Interface
-----------------------------------------
--/topic Advanced User Interface
--/info
--This framework aims to give its user a cutting edge interface.
--
--wxAUI stands for Advanced User Interface and the wxAUI framework aims to give its user a cutting edge
--interface for use with the wxWidgets based applications. The original wxAUI sources have kindly been
--made available under the wxWindows licence by Kirix Corp. and they have since then been integrated
--into wxWidgets CVS and further improved.
--
--wxAUI attempts to encapsulate the following aspects of the user interface:
--
--Frame Management: Frame management provides the means to open, move and hide common controls that
--are needed to interact with the document, and allow these configurations to be saved into different
--perspectives and loaded at a later time.
--
--Toolbars: Toolbars are a specialized subset of the frame management system and should behave similarly
--to other docked components. However, they also require additional functionality, such as "spring-loaded"
-- rebar support, "chevron" buttons and end-user customizability.
--
--Modeless Controls: Modeless controls expose a tool palette or set of options that float above the
--application content while allowing it to be accessed. Usually accessed by the toolbar, these controls
--disappear when an option is selected, but may also be "torn off" the toolbar into a floating frame of
--their own.
--
--Look and Feel: Look and feel encompasses the way controls are drawn, both when shown statically as well
--as when they are being moved. This aspect of user interface design incorporates "special effects" such
--as transparent window dragging as well as frame animation.
--
--wxAUI adheres to the following principles:
--
--Use native floating frames to obtain a native look and feel for all platforms. Use existing wxWidgets
--code where possible, such as sizer implementation for frame management. Use classes included in wxCore
--and wxBase only. Use standard wxWidgets coding conventions.
--
--<h3>Using wxAUI with wxEuphoria</h3>
--
--Using the advanced user interface in wxEuphoria is straightforward:
--<ul>
--/li Create your frame.
--/li Create a /wxAuiManager and set the frame to be the managed window ( /set_managed_window )
--/li Create your controls
--/li Create /wxAuiPaneInfo objects for each control (if you need the added power they bring)
--/li Add your controls to the /wxAuiManager using either /add_pane_simple(), /add_pane_with_info(),
--or /add_pane_with_pos().
--</ul>



--Classes
------------------------------
--/topic Classes
--/info
--Classes are objects which are not /Controls or /Events.

--Each class should have at least two things defined, a constant identifying the class
--and a sizeof_ constant.  In order to refer to class /wxWindow, simply use its name, "wxWindow".  To
--refer to the size in bytes of an object of type wxWindow, use sizeof_wxWindow.

--/topic Classes
--/func create( integer class, object params )
--/b Returns: An object pointer
--
--This function is used to create all classes included in wxEuphoria.  You should check the
-- documentation of the class to see what the required and optional parameters are.
public function create( integer class, object params )

    object rid

    if atom( params ) then
        params = {params}
    end if

    rid = WX_CREATE_ARRAY[class]
    if atom( rid ) then
        -- wxEuphoria routine
        return c_func( rid, {params} )

    else
        -- user-defined routine
        return call_func( rid[1], {params} )

    end if

end function



--Clipboard
------------------------------
--/topic Clipboard
--/info
--Cut and paste

--/topic Clipboard
--/func get_clip_text()
--/b Returns: Current clipboard text
--
--This returns the text currently within the clipboard. If there is
-- non-text data in the clipboard ( such as a picture ), it will
-- return a sequence with the word data inside of it.
public function get_clip_text()
    return c_func( WX_GET_CLIP_TEXT, {} )
end function

--/topic Clipboard
--/proc set_clip_text( sequence text )
--
--This removes the current text in the clipboard and inserts the user-defined text in it's place.
public procedure set_clip_text( sequence text )
    c_proc( WX_SET_CLIP_TEXT, {text} )
end procedure



--Controls
------------------------------
--/topic Controls
--/info
--Common widgets available in wxEuphoria
--
--<div class="subsection">See also</div>
--<ul>
--/li /Classes
--/li /Dialogs
--/li /Sizers
--/li /Events
--/li /IDs
--</ul>

--/topic Controls
--/proc enable( object control, integer bool )
--
--If /i bool is equal to 0, then the specified /i control will be frozen
-- and changes will not be allowed to be made to it or any controls
-- made upon it. If /i bool is equal to 1 then it reverses the effect.
-- You can also use a sequence to enable or disable multiple windows
-- at once with /i control.
public procedure enable( object control, integer bool )
    c_proc( WX_ENABLE, {control, bool} )
end procedure

--/topic Controls
--/func is_visible( atom window )
--/b Returns: Visible state of the window
--
--Returns a 1 if the /i window is visible, or a zero if it is not.
-- You can use /show_window to toggle a window's visibility.
public function is_visible( atom window )
    return c_func( WX_IS_VISIBLE, {window} )
end function

--/topic Controls
--/proc destroy( atom this )
--
--You should use this instead of /delete_instance to destroy any object
-- that handles events (generally, any control or window).
public procedure destroy( atom this )
    c_proc( WX_DESTROY, {this} )
end procedure

--/topic Controls
--/func get_handle( atom this )
--/b Returns: Object handle of the window
--
--Returns the platform specific handle for objects derived from class wxWindow.
public function get_handle( atom this )
    return c_func( WX_GET_HANDLE, {this} )
end function

--/topic Controls
--/proc set_tip( atom this, sequence tip )
--
--Sets the tooltip for an object derived from class wxWindow.  To create
-- multi-line tooltips, use \n under Linux/BSD, and \r\n under Windows.
public procedure set_tip( atom this, sequence tip )
    c_proc( WX_SET_TIP, {this, tip} )
end procedure

--/topic Controls
--/func get_label( atom control )
--/b Returns: Text label of the control
--
--Returns the default label associated with the /i control.  This would be the caption for a
-- /wxFrame or the label on a /wxButton or the text of a /wxStaticText. To get the text
-- in a /wxTextCtrl, you should use /get_text_value.
public function get_label( atom control )
    return c_func( WX_GET_LABEL, {control} )
end function

--/topic Controls
--/proc set_size( atom win, integer cx, integer cy )
--
--This sets the size of the control to the designated values. /i cx is
-- the size from left to right and /i cy is the size from top to bottom.
public procedure set_size( atom win, integer cx, integer cy )
    c_proc( WX_SET_SIZE, {win, cx, cy} )
end procedure

--/topic Controls
--/proc move( atom win, integer x, integer y )
--
--Moves the window to the given position.
public procedure move( atom win, integer x, integer y )
    c_proc( WX_MOVE, {win, x, y} )
end procedure

--/topic Controls
--/proc set_focus( atom window )
--
--Sets the cursor to the specified /i window as if it had clicked on
-- it. If it is a text control, the blinking cursor will appear at
-- the beginning of the control where text can be entered. If it is
-- a control, then the top selection will be highlighted. If it is a
-- /wxTreeCtrl, then the root will be highlighted.
public procedure set_focus( atom window )
    c_proc( WX_SET_FOCUS, {window} )
end procedure

--/topic Controls
--/func get_rect( atom win )
--/b Returns: {x, y, width, height}
--
--Returns the size and position of the control.
public function get_rect( atom win )
    return c_func( WX_GET_RECT, {win} )
end function

--/topic Controls
--/proc set_cursor( atom window, atom cursor )
--
--Changes the mouse cursor for the specified /i window to be the specified /i cursor.
--
--See /wxCursor for the available cursor shapes.
public procedure set_cursor( atom window, atom cursor )
    c_proc( WX_SET_CURSOR, {window, cursor} )
end procedure

--/topic Controls
--/proc enable_tooltips( integer enable )
--
--Enables or disables tooltips globally.
--
--This procedure does not work under GTK2.12 or later.
public procedure enable_tooltips( integer enable )
    c_proc( WX_ENABLE_TOOLTIPS, {enable})
end procedure

--/topic Controls
--/proc set_tooltip_delay( integer msec )
--
--Set the delay (in milliseconds) after with tooltips display.
--
--This procedure does not work under GTK2.12 or later.
public procedure set_tooltip_delay( integer msec )
    c_proc( WX_SET_TOOLTIP_DELAY, {msec})
end procedure

--/topic Controls
--/func get_parent( atom window )
--
--Returns the parent of the /i window, or NULL if there is no parent.
public function get_parent( atom window )
    return c_func( WX_GET_PARENT, {window})
end function

--/topic Controls
--/func get_children( atom window )
--
--Returns a sequence of the /i window's children.
public function get_children( atom window )
    return c_func( WX_GET_CHILDREN, {window})
end function

--/topic Controls
--/func get_desktop_window()
--
--Returns the desktop's window handle.
public function get_desktop_window(  )
    return c_func( WX_GET_DESKTOP_WINDOW, {})
end function

--/topic Controls
--/func client_to_screen( atom window, atom x, atom y )
--
--Converts from client to screen window coordinates.
public function client_to_screen( atom window, atom x, atom y )
    return c_func( WX_CLIENT_TO_SCREEN, {window,x,y})
end function

--/topic Controls
--/func screen_to_client( atom window, atom x, atom y )
--
--Converts from screen to client window coordinates.
public function screen_to_client( atom window, atom x, atom y )
    return c_func( WX_SCREEN_TO_CLIENT, {window,x,y})
end function

--/topic Controls
--/func hit_test( atom window, atom x, atom y, atom flags )
--
--Hit-tests a /wxListCtrl, /wxTreeCtrl, or /wxTextCtrl.
--
--/b Note: /i flags is unused for /wxTextCtrl.
--
--See also:
--<ul>
--/li /list_hit_test()
--/li /tree_hit_test()
--/li /text_hit_test()
--</ul>
public function hit_test( atom window, atom x, atom y, atom flags )
    return c_func( WX_HIT_TEST, {window,x,y,flags})
end function



--List Controls
------------------------------
--/topic List Controls
--/parent Controls
--/info
--Controls which contain items

--/topic List Controls
--/proc insert_item( atom list, integer position, sequence text )
--
--Inserts the item into the /i list before /i position. Not valid for /b wxLB_SORT or /b wxCB_SORT styles, use /add_item instead.
public procedure insert_item( atom list, integer position, sequence text )
    c_proc( WX_INSERT_ITEM, {list, position, text} )
end procedure

--/topic List Controls
--/proc add_item( atom combo, sequence text )
--
--Used to add one or more strings into a /wxListBox, /wxComboBox, or /wxChoice.
--  /i text can be either a string, or a sequence of strings, in which case all of the
-- items will be  added.
--
--This is also used to enter one or more string, bitmap sequences into a /wxBitmapComboBox.
public procedure add_item( atom list, sequence text )
    c_proc( WX_ADD_ITEM, {list, text} )
end procedure

--/topic List Controls
--/func get_selection( atom control )
--
--Returns the selected index(es) of the control
--
--Used to get the index of the selected item of a /wxListBox, /wxListCtrl,
-- /wxComboBox /wxChoice, or /wxMultiChoiceDialog.  For the case of a
-- /wxListBox or /wxMultiChoiceDialog. The return is always a sequence,
-- or an empty sequence, if nothing is selected.  The other
-- types of controls return -1 is no item is selected.
public function get_selection( atom control )
    return c_func( WX_GET_SELECTION, {control} )
end function

--/topic List Controls
--/func get_list_string( atom list, integer item )
--
--Returns the text label of the list item
--
--This function works for both /wxListCtrl and /wxListBox.
public function get_list_string( atom list, integer item )
    return c_func( WX_GET_LIST_STRING, {list, item} )
end function

--/topic List Controls
--/func get_string_selection( atom list )
--
--Returns the text of selected item
--
--Used to get the text of the selected item of a /wxListBox, /wxComboBox or /wxChoice.
public function get_string_selection( atom control )
    return c_func( WX_GET_STRING_SELECTION, {control} )
end function

--/topic List Controls
--/proc clear_items( atom control )
--
--Clears all items from a /wxListBox, /wxListCtrl, /wxChoice or /wxComboBox.
public procedure clear_items( atom control )
    c_proc( WX_CLEAR_ITEMS, {control} )
end procedure

--/topic List Controls
--/func find_string( atom control, sequence string )
--
--Returns the index of specified item string
public function find_string( atom control, sequence string )
    return c_func( WX_FIND_STRING, {control, string} )
end function

--/topic List Controls
--/proc delete_item( atom control, integer item )
--
--Delete an item from a /wxListBox, /wxListCtrl, /wxComboBox or
-- /wxChoice.
public procedure delete_item( atom control, object items )
        if atom(items) then
                items = { items }
        end if
        for t=1 to length(items) do
            c_proc( WX_DELETE_ITEM, {control, items[t]} )
        end for
end procedure

--/topic List Controls
--/proc set_selection( atom control, object item )
--
--To unselect a /wxListBox item, the /b item parameter should
-- be a sequence: {item_number, -1}.  To unselect a /wxListCtrl
-- item, /b item should be {item_number, 0}.  To select items
-- for a /wxMultiChoiceDialog, /b item should be either an index number
-- starting from zero, or a sequence of indexes.
public procedure set_selection( atom control, object item )
    c_proc( WX_SET_SELECTION, {control, item} )
end procedure

--/topic List Controls
--/proc set_string( atom control, integer item, sequence string )
--
--Set the string of an item in a /wxListBox, /wxComboBox or /wxChoice.
public procedure set_string( atom control, integer item, sequence string )
    c_proc( WX_SET_STRING, {control, item, string} )
end procedure



--Cross Platform Tips
------------------------------
--/topic Cross Platform Tips
--/info
--Helpful hints for coding for multiple platforms
--
--<ul>
--/li Use portable graphics.  This means no windows icons (*.ico).
--Try to use bmp, jpg, png or xpm graphics.
--/li Be aware of line ending issues.  Under Windows, line endings are "\r\n" and
--under Linux/BSD, they are "\n".  This is important when using /set_tip().
--</ul>



--Demos
------------------------------
--/topic Demos
--/parent Getting Started
--/info
--Some examples to get you started
--
--There are a number of demos included with wxEuphoria.  They should be located in the
-- demos subdirectory of your wxEuphoria directory if you preserved the directory structure
-- when you decompressed wxEuphoria.  First, make sure that you properly installed wxEuphoria.
-- This means that the dynamic or shared libary is located in the correct directory.  Under
-- Windows, it can be in either your euphoria/bin directory, or your %WINDIR%\system32 directory.
-- Under Linux, it should be in your /usr/local/lib/ directory. Also, ensure that
-- the wxEuphoria source files are either in a directory that's included in your EUINC variable
-- (see the Euphoria docs for information about changing this), or that any program that uses
-- wxEuphoria is in the same directory as the wxEuphoria library.  Then, to run the window_basic.exw
-- demo from the command line, you can type:
--
--/code
--Windows:
--  > exw hello_world.exw
--
--Linux:
--  $ exu hello_world.exw
--/endcode
--
--In Windows, if you have the ".exw" extension associated with the Euphoria interpreter, you may
-- simply type:
--
--/code
--  > window_basic.exw
--/endcode
--
--In Linux, there is a similar feature available to you, but you will need to change the first
-- line of the file, which is sometimes called the shebang. This line tells your shell that
-- it should use the interpreter listed on the first line to execute the file.  You may also
-- need to change the permissions on the file using /b chmod to allow the file to execute.
--
--To see what the shebang should be, you could type:
--
--/code
--  > which exu
--/endcode
--
--This will give the path to your copy of the Euphoria interpreter.  It will probably look similar
-- to the current shebang in the demos. The path must be preceded by "#!" to tell the shell that
-- the line is actually the interpreter to be used:
--
--/code
--  Current shebang (based on my system):
--       #!/home/matt/euphoria/bin/exu
--
--  Yours might look like:
--       #!/home/myuserid/euphoria/bin/exu
--
--  Or perhaps:
--       #!/usr/local/euphoria/bin/exu
--/endcode
--
--You also need to make sure that the file does not have DOS style new lines (carriage returns + new line).
-- Once you have done this, you can run the demo from the command line:
--
--/code
--  > ./window_basic.exw
--/endcode
--
--Note that this assumes your current directory is the same directory where the file is located.
-- Linux doesn't include the current directory in your executable PATH environment variable (like
-- Windows does), so you need to tell Linux to look there using "./".
--
--Greg Haberek recently wrote a gateway to the wxEuphoria demos, called /i all_demos.exw. A demo
-- itself, it provides a list of all available the demos with a brief description.



--Developer's Notes
------------------------------
--/topic Developer's Notes
--/info
--Changes, Bugs and TODO List
--
--The windows DLL was built using the open source Open Watcom C++ compiler v1.7.
--



--Dialogs
------------------------------
--/topic Dialogs
--/info
--Using dialogs with wxEuphoria

--/topic Dialogs
--/func font_selector( atom parent, atom initial )
--/b Returns: Selected font, or zero
--
--This is a wrapper for /wxFontDialog.  /b parent is the
-- parent for the dialog (it may be null), and /b initial is
-- the /wxFont object that will be initially selected when
-- the dialog is opened.  If a font is selected (the user clicks
-- OK), the new /wxFont object is returned, otherwise the return
-- value is zero.
public function font_selector( atom parent, atom initial )
    return c_func( WX_FONT_SELECTOR, {parent, initial} )
end function

--/topic Dialogs
--/func color_selector( atom parent, atom initial )
--/b Returns: Selected color, or zero
--
--This is a wrapper for /wxColourDialog.  It returns a
-- /wxColour object, or zero, if the user clicked the cancel button.
public function color_selector( atom parent, atom initial )
    return c_func( WX_COLOR_SELECTOR, {parent, initial} )
end function

--/topic Dialogs
--/func file_selector( atom parent, sequence message, sequence info, atom flag )
--/b Returns: Selected file path, or an empty string
--
--Pops up a file selector box. In Windows, this is the common file selector dialog. In X,
-- this is a file selector box with the same functionality. This function is used for
-- both open and save dialogs.
--
--<ul>
--/i info is a sequence containing:
--<ul>
--/li path
--/li file
--/li extension
--/li wildcard
--</ul>
--</ul>
--
--You may supply none or any number of these, but /b file_selector() will assume they are
-- in this order, so in order to specify a wildcard selection, you must specify the others, even
-- if they are empty sequences.
--
--The path and filename are distinct elements of a full file pathname. If path is empty, the
-- current directory will be used. If filename is empty, no default filename will be supplied.
-- The wildcard determines what files are displayed in the file selector, and file extension
-- supplies a type extension for the required filename.
--
--Flags may be a combination of /b "wxOPEN, wxSAVE, wxOVERWRITE_PROMPT",
-- /b "wxHIDE_READONLY, wxFILE_MUST_EXIST, wxMULTIPLE" or 0.
--
--Both the Unix and Windows versions implement a wildcard filter. Typing a filename
-- containing wildcards (*, ?) in the filename text item, and clicking on Ok, will result
-- in only those files matching the pattern being displayed.
--
--The wildcard may be a specification for multiple types of file with a description for each, such as:
--/code
--    "BMP files (*.bmp)|*.bmp|GIF files (*.gif)|*.gif"
--/endcode
public function file_selector( atom parent, sequence message, sequence info, atom flag )
    return c_func( WX_FILE_SELECTOR, {parent, message, info, flag} )
end function



--Drag and Drop
------------------------------
--/topic Drag and Drop
--/info
--Working with files and other controls
--

function get_drop_files( atom target )
    return c_func( WX_GET_DROP_FILES, {target})
end function

function dnd_file_callback( integer rid, atom target )
    call_proc( rid, { get_drop_files( target ) } )
    return 1
end function
constant cb_dnd = call_back('+' & routine_id("dnd_file_callback"))

--/topic Drag and Drop
--/proc accept_files( atom window, integer rid )
--
--Sets a wxWindow up to accept drag and drop files.  The routine
-- specified by /i rid will be called when any files are dropped.  The
-- parameter should be a sequence, which will contain the list of
-- file names that have been dropped on the window.
public procedure accept_files( atom window, integer rid )
    c_proc( WX_ACCEPT_FILES, {window, rid, cb_dnd} )
end procedure



--Events
------------------------------
--/topic Events
--/info
--Using events with wxEuphoria
--
--Event handler procedures take 4 arguments:
--<ul>
--/li atom this
--/li atom event_type
--/li atom id
--/li atom event
--</ul>
--
--<div class="subsection">Notes</div>
--
--<h4><a name="wxEVT_PAINT">wxEVT_PAINT Event</a></h4>
--
--During a paint event, you need to create a /wxPaintDC.  This must be destroyed
--( /delete_instance ) before you return from the paint handler.  Also, any drawing
--that is done should be within calls to /begin_drawing() and /end_drawing():
--
--Example:
--/code
--
--    procedure onPaint( atom this, atom event_type, atom id, atom event )
--        atom dc
--
--        dc = create( wxPaintDC, {this} )
--        begin_drawing(dc)
--        -- ... do drawing here...
--        end_drawing(dc)
--        delete_instance(dc)
--    end procedure
--
--/endcode
--
--<h4><a name="keyevents">Key Events</a></h4>
--
--Key events under Linux should be set with an id of -1.  Also, no events will be passed until
--the window has been clicked by the user.
--
--<div class="subsection">Event Constants</div>
--
--Full listing of events as defined in /b "wxeud.e":
--<ul style="height: 320px; width: 480px; overflow: scroll;">
--/li wxEVT_ACTIVATE
--/li wxEVT_ACTIVATE_APP
--/li wxEVT_CHAR
--/li wxEVT_CHAR_HOOK
--/li wxEVT_CHILD_FOCUS
--/li wxEVT_CLOSE_WINDOW
--/li wxEVT_COMMAND_BUTTON_CLICKED
--/li wxEVT_COMMAND_CHECKBOX_CLICKED
--/li wxEVT_COMMAND_CHECKLISTBOX_TOGGLED
--/li wxEVT_COMMAND_CHOICE_SELECTED
--/li wxEVT_COMMAND_COLLPANE_CHANGED
--/li wxEVT_COMMAND_COLOURPICKER_CHANGED
--/li wxEVT_COMMAND_COMBOBOX_SELECTED
--/li wxEVT_COMMAND_DIRPICKER_CHANGED
--/li wxEVT_COMMAND_ENTER
--/li wxEVT_COMMAND_FILEPICKER_CHANGED
--/li wxEVT_COMMAND_FONTPICKER_CHANGED
--/li wxEVT_COMMAND_HTML_LINK_CLICKED
--/li wxEVT_COMMAND_KILL_FOCUS
--/li wxEVT_COMMAND_LEFT_CLICK
--/li wxEVT_COMMAND_LEFT_DCLICK
--/li wxEVT_COMMAND_LISTBOX_DOUBLECLICKED
--/li wxEVT_COMMAND_LISTBOX_SELECTED
--/li wxEVT_COMMAND_LIST_ITEM_RIGHT_CLICK
--/li wxEVT_COMMAND_LIST_ITEM_SELECTED
--/li wxEVT_COMMAND_MENU_SELECTED
--/li wxEVT_COMMAND_NOTEBOOK_PAGE_CHANGED
--/li wxEVT_COMMAND_NOTEBOOK_PAGE_CHANGING
--/li wxEVT_COMMAND_RADIOBOX_SELECTED
--/li wxEVT_COMMAND_RADIOBUTTON_SELECTED
--/li wxEVT_COMMAND_RICHTEXT_CHARACTER
--/li wxEVT_COMMAND_RICHTEXT_CONTENT_DELETED
--/li wxEVT_COMMAND_RICHTEXT_CONTENT_INSERTED
--/li wxEVT_COMMAND_RICHTEXT_DELETE
--/li wxEVT_COMMAND_RICHTEXT_RETURN
--/li wxEVT_COMMAND_RICHTEXT_STYLESHEET_CHANGING
--/li wxEVT_COMMAND_RICHTEXT_STYLESHEET_REPLACED
--/li wxEVT_COMMAND_RICHTEXT_STYLESHEET_REPLACING
--/li wxEVT_COMMAND_RICHTEXT_STYLE_CHANGED
--/li wxEVT_COMMAND_RIGHT_CLICK
--/li wxEVT_COMMAND_RIGHT_DCLICK
--/li wxEVT_COMMAND_SCROLLBAR_UPDATED
--/li wxEVT_COMMAND_SET_FOCUS
--/li wxEVT_COMMAND_SLIDER_UPDATED
--/li wxEVT_COMMAND_SPINCTRL_UPDATED
--/li wxEVT_COMMAND_SPLITTER_DOUBLECLICKED
--/li wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGED
--/li wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGING
--/li wxEVT_COMMAND_SPLITTER_UNSPLIT
--/li wxEVT_COMMAND_TEXT_ENTER
--/li wxEVT_COMMAND_TEXT_MAXLEN
--/li wxEVT_COMMAND_TEXT_UPDATED
--/li wxEVT_COMMAND_TEXT_URL
--/li wxEVT_COMMAND_TOGGLEBUTTON_CLICKED
--/li wxEVT_COMMAND_TOOL_CLICKED
--/li wxEVT_COMMAND_TOOL_ENTER
--/li wxEVT_COMMAND_TOOL_RCLICKED
--/li wxEVT_COMMAND_TREE_BEGIN_DRAG
--/li wxEVT_COMMAND_TREE_BEGIN_LABEL_EDIT
--/li wxEVT_COMMAND_TREE_BEGIN_RDRAG
--/li wxEVT_COMMAND_TREE_DELETE_ITEM
--/li wxEVT_COMMAND_TREE_END_DRAG
--/li wxEVT_COMMAND_TREE_END_LABEL_EDIT
--/li wxEVT_COMMAND_TREE_GET_INFO
--/li wxEVT_COMMAND_TREE_ITEM_ACTIVATED
--/li wxEVT_COMMAND_TREE_ITEM_COLLAPSED
--/li wxEVT_COMMAND_TREE_ITEM_COLLAPSING
--/li wxEVT_COMMAND_TREE_ITEM_EXPANDED
--/li wxEVT_COMMAND_TREE_ITEM_EXPANDING
--/li wxEVT_COMMAND_TREE_ITEM_GETTOOLTIP
--/li wxEVT_COMMAND_TREE_ITEM_MENU\
--/li wxEVT_COMMAND_TREE_ITEM_MIDDLE_CLICK
--/li wxEVT_COMMAND_TREE_ITEM_RIGHT_CLICK
--/li wxEVT_COMMAND_TREE_KEY_DOWN
--/li wxEVT_COMMAND_TREE_SEL_CHANGED
--/li wxEVT_COMMAND_TREE_SEL_CHANGED
--/li wxEVT_COMMAND_TREE_SEL_CHANGING
--/li wxEVT_COMMAND_TREE_SET_INFO
--/li wxEVT_COMMAND_TREE_STATE_IMAGE_CLICK=
--/li wxEVT_COMMAND_VLBOX_SELECTED
--/li wxEVT_COMPARE_ITEM
--/li wxEVT_CONTEXT_MENU
--/li wxEVT_CREATE
--/li wxEVT_DATE_CHANGED
--/li wxEVT_DESTROY
--/li wxEVT_DETAILED_HELP
--/li wxEVT_DISPLAY_CHANGED
--/li wxEVT_DRAW_ITEM
--/li wxEVT_DROP_FILES
--/li wxEVT_END_SESSION
--/li wxEVT_ENTER_WINDOW
--/li wxEVT_ERASE_BACKGROUND
--/li wxEVT_GRID_CELL_CHANGE
--/li wxEVT_GRID_CELL_LEFT_CLICK
--/li wxEVT_GRID_CELL_LEFT_DCLICK
--/li wxEVT_GRID_CELL_RIGHT_CLICK
--/li wxEVT_GRID_CELL_RIGHT_DCLICK
--/li wxEVT_GRID_COL_SIZE
--/li wxEVT_GRID_EDITOR_CREATED
--/li wxEVT_GRID_EDITOR_HIDDEN
--/li wxEVT_GRID_EDITOR_SHOWN
--/li wxEVT_GRID_LABEL_LEFT_CLICK
--/li wxEVT_GRID_LABEL_LEFT_DCLICK
--/li wxEVT_GRID_LABEL_RIGHT_CLICK
--/li wxEVT_GRID_LABEL_RIGHT_DCLICK
--/li wxEVT_GRID_RANGE_SELECT
--/li wxEVT_GRID_ROW_SIZE
--/li wxEVT_GRID_SELECT_CELL
--/li wxEVT_HELP
--/li wxEVT_HIBERNATE
--/li wxEVT_HOTKEY
--/li wxEVT_ICONIZE
--/li wxEVT_IDLE
--/li wxEVT_INIT_DIALOG
--/li wxEVT_JOY_BUTTON_DOWN
--/li wxEVT_JOY_BUTTON_UP
--/li wxEVT_JOY_MOVE
--/li wxEVT_JOY_ZMOVE
--/li wxEVT_KEY_DOWN
--/li wxEVT_KEY_UP
--/li wxEVT_KILL_FOCUS
--/li wxEVT_LEAVE_WINDOW
--/li wxEVT_LEFT_DCLICK
--/li wxEVT_LEFT_DOWN
--/li wxEVT_LEFT_UP
--/li wxEVT_MAXIMIZE
--/li wxEVT_MEASURE_ITEM
--/li wxEVT_MENU_CLOSE
--/li wxEVT_MENU_HIGHLIGHT
--/li wxEVT_MENU_OPEN
--/li wxEVT_MIDDLE_DCLICK
--/li wxEVT_MIDDLE_DOWN
--/li wxEVT_MIDDLE_UP
--/li wxEVT_MOTION
--/li wxEVT_MOUSEWHEEL
--/li wxEVT_MOUSE_CAPTURE_CHANGED
--/li wxEVT_MOVE
--/li wxEVT_MOVING
--/li wxEVT_NAVIGATION_KEY
--/li wxEVT_NC_ENTER_WINDOW
--/li wxEVT_NC_LEAVE_WINDOW
--/li wxEVT_NC_LEFT_DCLICK
--/li wxEVT_NC_LEFT_DOWN
--/li wxEVT_NC_LEFT_UP
--/li wxEVT_NC_MIDDLE_DCLICK
--/li wxEVT_NC_MIDDLE_DOWN
--/li wxEVT_NC_MIDDLE_UP
--/li wxEVT_NC_MOTION
--/li wxEVT_NC_PAINT
--/li wxEVT_NC_RIGHT_DCLICK
--/li wxEVT_NC_RIGHT_DOWN
--/li wxEVT_NC_RIGHT_UP
--/li wxEVT_PAINT
--/li wxEVT_PAINT_ICON
--/li wxEVT_PALETTE_CHANGED
--/li wxEVT_QUERY_END_SESSION
--/li wxEVT_QUERY_NEW_PALETTE
--/li wxEVT_RIGHT_DCLICK
--/li wxEVT_RIGHT_DOWN
--/li wxEVT_RIGHT_UP
--/li wxEVT_SCROLLWIN_BOTTOM
--/li wxEVT_SCROLLWIN_LINEDOWN
--/li wxEVT_SCROLLWIN_LINEUP
--/li wxEVT_SCROLLWIN_PAGEDOWN
--/li wxEVT_SCROLLWIN_PAGEUP
--/li wxEVT_SCROLLWIN_THUMBRELEASE
--/li wxEVT_SCROLLWIN_THUMBTRACK
--/li wxEVT_SCROLLWIN_TOP
--/li wxEVT_SCROLL_BOTTOM
--/li wxEVT_SCROLL_CHANGED
--/li wxEVT_SCROLL_LINEDOWN
--/li wxEVT_SCROLL_LINEUP
--/li wxEVT_SCROLL_PAGEDOWN
--/li wxEVT_SCROLL_PAGEUP
--/li wxEVT_SCROLL_THUMBRELEASE
--/li wxEVT_SCROLL_THUMBTRACK
--/li wxEVT_SCROLL_TOP
--/li wxEVT_SETTING_CHANGED
--/li wxEVT_SET_CURSOR
--/li wxEVT_SET_FOCUS
--/li wxEVT_SHOW
--/li wxEVT_SIZE
--/li wxEVT_SIZING
--/li wxEVT_SOCKET
--/li wxEVT_SYS_COLOUR_CHANGED
--/li wxEVT_TIMER
--/li wxEVT_UPDATE_UI
--/li wxEVT_WIZARD_CANCEL
--/li wxEVT_WIZARD_FINISHED
--/li wxEVT_WIZARD_HELP
--/li wxEVT_WIZARD_PAGE_CHANGED
--/li wxEVT_WIZARD_PAGE_CHANGING
--</ul>


--/topic Events
--/proc set_event_handler( object this, object id, object event, integer rid )
--
--Set up an event handler.
--<ul>
--/li /b this is the object that receives the event.
--/li /b id is the id of the item generating the event, such a a menu item.
--/li /b event is an event constant, and
--/li /b rid is the routine id of your event handler.
--</UL>
--
--Example:
--/code
--
--    procedure myWindow_onClose( atom this, atom event_type, atom id, atom event )
--        exit_main()
--    end procedure
--    set_event_handler( myWindow, wxID_EXIT, wxEVT_COMMAND_MENU_SELECTED, routine_id("myWindow_onClose") )
--
--/endcode
--
--See class documentation for valid event constants.
public procedure set_event_handler( object this, object id, object event, integer rid )
    if sequence(this) then
        this = this[2]
    end if
    if atom(id) then
        id = {id}
    end if
    if atom(event) then
        event = {event}
    end if

    for i = 1 to length(id) do
        for e = 1 to length(event) do
            c_proc( WX_SET_EVENT_HANDLER, {this, id[i], event[e], rid})
        end for
    end for

end procedure


--/topic Events
--/proc skip( atom event )
--
--Called by an event handler to tell the event system that the event
--handler should be skipped, and the next valid handler used instead.
public procedure skip( atom event )
    c_proc( WX_SKIP, {event} )
end procedure

function get_event_type( atom event  )
    return c_func( WX_GET_EVENT_TYPE, {event})
end function

function get_event_id( atom event )
    return c_func( WX_GET_EVENT_ID, {event})
end function

function get_event_object( atom event )
    return c_func( WX_GET_EVENT_OBJECT, {event})
end function

--/topic Events
--/func can_veto( atom event )
--/b Returns: True or false
--Used to determine if the wxEVT_CLOSE_WINDOW event can be vetoed.
public function can_veto( atom event )
    return c_func( WX_CAN_VETO, {event})
end function



--Files
------------------------------
--/topic Files
--/info
--Working with files and directories
--
--See also /file_selector()
--

--/topic Files
--/func copy_file( sequence file1, sequence file2 )
--/b Returns: An integer which will be 1 if successful, 0 if unsuccessful.
--
--File1 is the file to be copied and File2 is the file that will be made.
--If you are using Windows, use \\ instead of /.
--
--You may /i NOT:
--<ul>
--/li Create a new directory with it.
--/li Copy one directory to another.
--</ul>
--
--You may:
--<ul>
--/li Copy something and give it a name ( ex : C:\\EUPHORIA\\INCLUDE\\wxEuphoria.e,C:\\EUPHORIA\\INCLUDE\\wxUforiuh.K )
--/li Have spaces between each \\ or / 'marks C:\\Documents and Settings\\'
--/li Have incorrect caps ( 'C:\\DoCuMeNtS aNd SeTtInGs\\' will still work. )
--</ul>
public function copy_file( sequence file1, sequence file2 )
    return c_func( WX_COPY_FILE, {file1, file2} )
end function

--/topic Files
--/func remove_file( sequence file )
--/b Returns: An integer which will be 1 if successful, 0 if unsuccessful.
--
--Removes (deletes) the specified file, returns true if successful.
public function remove_file( sequence file )
    return c_func( WX_REMOVE_FILE, {file} )
end function

--/topic Files
--/func rmdir( sequence directory )
--/b Returns: An integer which will be 1 if successful, 0 if unsuccessful.
--
--Deletes the specified directory from the file system.
public function rmdir( sequence directory )
    return c_func( WX_WX_RMDIR, {directory})
end function

function wx_rmdir( sequence directory )
    return c_func( WX_WX_RMDIR, {directory})
end function


--Getting Started
------------------------------
--/topic Getting Started
--/info
--How do I use wxEuphoria?
--
--wxEuphoria is designed to be very similar to other Euphoria GUI libraries (i.e.,
-- Win32Lib).  Although I recommend downloading the wxWidgets documentation, a
-- program written in wxEuphoria will take a different form than a C++ program
-- written using wxWidgets. In wxEuphoria, you don't have to worry about
-- creating a class derived from wxApp, or overloading an OnInit() function.
--
--All you need to do is to create your controls, define and connect your event
-- handlers, and start the wxEuphoria event loop.  Below is a basic outline
-- of what a wxEuphoria program might look like:
--
--/code
--   [include statements]
--
--   [create the controls]
--
--   [event handlers and other code]
--
--   wxMain()
--/endcode
--
--Here is the classic 'Hello World!' program written in wxEuphoria (see hello_world.exw):
--
--Example:
--/code
--
--  include wxeud.e
--
--  constant
--      HelloFrame = create( wxFrame, {0, -1, "My First Message", -1, -1, 200, 100} ),
--      HelloWin   = create( wxPanel, {HelloFrame})
--
--  procedure onPaint_HelloWin( atom this, atom event, atom it, atom event_type )
--      wx_puts( this, "Hello, World!")
--  end procedure
--  set_event_handler( HelloWin, get_id(HelloWin), wxEVT_PAINT, routine_id( "onPaint_HelloWin" ))
--
--  wxMain( HelloFrame )
--
--/endcode
--
--Notice the syntax of the calls to /create().  This is the standard function
-- that should be called to create an object of almost any wxEuphoria class that
-- is wrapped.  The class documentation should indicate what the second parameter
-- (always a sequence) should be.  Many classes have /i optional parameters.  You
-- can omit these parameters if the default values are acceptable.  Note that if
-- the third and fourth parameters are optional, you can't specify the fourth and
-- not the third--i.e., you can't skip parameters in the middle of the list.
--
--Unfortunately, all functions and classes haven't been documented as consistently
-- or as thourough as they should be.  If something in particular doesn't make sense,
-- please ask (best is the <a href="http://openeuphoria.org/EUforum/">EUforum</a>),
-- and I'll try to correct the situation.  Contributions, in the
-- form of code and/or documentation are also welcomed and encouraged.



--Graphics
------------------------------
--/topic Graphics
--/info
--Drawing, Text, Fonts
--
--This section details functions and classes that are used to draw
--on windows.  This includes drawing text (when a /wxTextCtrl or
--/wxStaticText don't do what you need).

--/topic Graphics
--/func get_text_extent( atom win, sequence text, atom font )
--/b Returns: Dimensions of the specified text
--
--Gets the dimensions of /i text using the specified /i font.
public function get_text_extent( atom win, sequence text, atom font )
    return c_func( WX_GET_TEXT_EXTENT, {win, text, font} )
end function

--/topic Graphics
--/proc set_font( atom dc, atom font )
--
--Sets the font for a specified /wxDC or /wxPaintDC assiciated with a window.  To set the default
--font for a window or control, you should use /set_default_font().
public procedure set_font( atom dc, atom font )
    c_proc( WX_SET_FONT, {dc, font} )
end procedure

--/topic Graphics
--/proc set_default_font( atom win, atom font )
--
--Sets the default font for a window.
public procedure set_default_font( atom win, atom font )
    c_proc( WX_SET_DEFAULT_FONT, {win, font} )
end procedure

--/topic Graphics
--/proc wx_puts( object win, sequence text )
--
--Works similar to puts().
--
--/i win can take several forms:
--<ul>
--/li /i this pointer to a wxWindow
--/li { this, x, y }
--<br> /i x and /i y set the place in the window to start drawing text, otherwise the default of 0,0 is used
--/li { this, x, y, dc }
--<br> /i x and /i y set the place in the window to start drawing text, otherwise the default of 0,0 is used
--<br> /i dc is a /wxDC (or some derivative of /wxDC, such as /wxMemoryDC) or /wxPaintDC for the window.
--<br>(If you already have the dc, then it is faster to pass it to wx_puts() than for wx_puts() to
-- create a new dc on the fly.)
--</ul>
--/i text is a sequence containing the text to be drawn
public procedure wx_puts( object win, sequence text )
    c_proc( WX_WX_PUTS, {win, text} )
end procedure

--/topic Graphics
--/proc wx_printf( object win, sequence text, object format )
--
--Works similar to printf().
--
--/i win can take several forms:<ul>
--/li /i this pointer to a wxWindow
--/li { this, x, y }
--<br> /i x and /i y set the place in the window to start drawing text, otherwise the default of 0,0 is used
--/li { this, x, y, dc }
--<br> /i x and /i y set the place in the window to start drawing text, otherwise the default of 0,0 is used
--<br> /i dc is a /wxDC or /wxPaintDC for the window.
--<br> (If you already have the dc, then it is faster to pass it to wx_printf() than for wx_printf() to
-- create a new dc on the fly.)
--</ul>
--/i text is a sequence containing the text to be drawn
--/i format is the object(s) to be insertined into /i text
public procedure wx_printf( object win, sequence text, object format )
    text = c_func( WX_W_SPRINTF, {text,format} )
    c_proc( WX_WX_PUTS, {win, text} )
end procedure

--/topic Graphics
--/proc set_pen( atom dc, atom pen )
--
--Sets /i pen to be the current /wxPen for /dc /wxDC.
public procedure set_pen( atom dc, atom pen )
    c_proc( WX_SET_PEN, {dc, pen} )
end procedure

--/topic Graphics
--/proc draw_line( atom dc, sequence points )
--
--Draws a line using the current /wxPen.  /i points should be a 4-element sequence:<br>
--
--Example:
--/code
--  draw_line( dc, {x1, y1, x2, y2} )
--/endcode
public procedure draw_line( atom dc, sequence points )
    c_proc( WX_DRAW_LINE, {dc, points} )
end procedure

--/topic Graphics
--/proc set_text_back_color( atom dc, object color )
--
--This sets a back color to the DC. Color can either be an RGB sequence {Red,Green,Blue} or a
-- pointer to a /wxColour.
public procedure set_text_back_color( atom dc, object color )
    c_proc( WX_SET_TEXT_BACK_COLOR, {dc, color} )
end procedure

--/topic Graphics
--/proc draw_icon( atom dc, atom icon, atom x, atom y)
--
--Draws a /wxIcon to the DC /i dc.
public procedure draw_icon( atom dc, atom icon, atom x, atom y)
    c_proc( WX_DRAW_ICON, {dc, icon, x, y} )
end procedure

--/topic Graphics
--/proc draw_bitmap( atom dc, atom bitmap, integer x, integer y, integer trans )
--
--Draw a /wxBitmap onto dc /i dest.  Set /i trans to 1 if it has an associated mask, and should
-- be drawn with transparency.
public procedure draw_bitmap( atom dc, atom bitmap, integer x, integer y, integer trans )
    c_proc( WX_DRAW_BITMAP, {dc, bitmap, x, y, trans} )
end procedure

--/topic Graphics
--/proc draw_polygon( atom dc, sequence points, atom x, atom y, atom style )
--
--Draws a polygon in the specified /i dc.  /i points is a sequence containing all of the
-- points to be drawn.  It should be of the form {x1,y1,x2,y2,...,xn,yn}.  /i x and /i y are
-- offset coordinates.
--
--The last argument specifies the fill rule:
--<ul>
--/li wxODDEVEN_RULE
--/li wxWINDING_RULE
--</ul>
--
--The current pen is used for drawing the outline, and the current brush for
-- filling the shape. Using a transparent brush suppresses filling.
--
--Note that wxWidgets automatically closes the first and last points.
public procedure draw_polygon( atom dc, sequence points, atom x, atom y, atom style )
    c_proc( WX_DRAW_POLYGON, {dc, points, x, y, style} )
end procedure

--/topic Graphics
--/func get_regions( atom window )
--
--Returns a wxRegionIterator, which can be used in conjunction
-- with /get_next_region() to get all of the regions to be
-- updated in an on paint event.  This should only be called from
-- within an on paint event.
public function get_regions( atom window )
    return c_func( WX_GET_REGIONS, {window} )
end function

function get_region( atom iter )
    return c_func( WX_GET_REGION, {iter} )
end function

--/topic Graphics
--/func get_first_region( atom iter )
--
--Returns a sequence with the x, y and width and height of the first region.
public function get_first_region( atom iter )
    return c_func( WX_GET_FIRST_REGION, {iter} )
end function

--/topic Graphics
--/func get_next_region( atom iter )
--
--Returns a sequence with the x, y and width and height of the next region,
-- or an empty sequence if the last region has already been accessed.
public function get_next_region( atom iter )
    return c_func( WX_GET_NEXT_REGION, {iter} )
end function

--/topic Graphics
--/proc begin_drawing( atom dc )
--
--This must be called before you begin drawing to any window (i.e., not to
-- a /wxMemoryDC), and must be followed with a call to /end_drawing() when
-- finished.
public procedure begin_drawing( atom dc )
    c_proc( WX_BEGIN_DRAWING, {dc} )
end procedure

--/topic Graphics
--/proc end_drawing( atom dc )
--
--This stops drawing to the specified DC. You should use this and
-- /begin_drawing to specify when drawing should begin and end on a
-- /wxDC.
public procedure end_drawing( atom dc )
    c_proc( WX_END_DRAWING, {dc} )
end procedure

--/topic Graphics
--/proc set_brush( atom dc, atom brush )
--
--Sets /i brush to be the current /wxBrush for /dc /wxDC.
public procedure set_brush( atom dc, atom brush )
    c_proc( WX_SET_BRUSH, {dc, brush} )
end procedure

--/topic Graphics
--/proc set_back_brush( atom dc, atom brush )
--
--Sets background /i brush to be the current /wxBrush for /dc /wxDC.
public procedure set_back_brush( atom dc, atom brush )
    c_proc( WX_SET_BACK_BRUSH, {dc, brush} )
end procedure

--/topic Graphics
--/proc clear_dc( atom dc )
--
--Clears the image currently in the /i dc.
public procedure clear_dc( atom dc )
    c_proc( WX_CLEAR_DC, {dc} )
end procedure

--/topic Graphics
--/proc blit( atom dstDC, integer dstX, integer dstY, atom srcDC, integer srcX, integer srcY, integer height, integer width, integer operation )
--
--Copy a bitmap from one DC to another.
--<ul>
--/li /b dstDC The destination DC
--/li /b dstX X coordinate in the destination DC
--/li /b dstY Y coordinate in the destination DC
--/li /b srcDC Source DC
--/li /b srcX X coordinate of image to copy in the source
--/li /b srcY y coordinate of image to copy in the source
--/li /b width of image to copy in the source
--/li /b height of image to copy in the source
--/li /b operation Logigal operation to use (wxCOPY, wxAND, wxOR, etc)
--</ul>
public procedure blit( atom dstDC, integer dstX, integer dstY, atom srcDC, integer srcX,    integer srcY, integer height, integer width, integer operation )
    c_proc( WX_BLIT, {dstDC, dstX, dstY, srcDC, srcX, srcY, height, width, operation} )
end procedure

--/topic Graphics
--/proc set_background_mode( atom dc, atom mode )
--
--/i mode may be either /b wxSOLID (text is drawn with a background color)
-- or /b wxTRANSPARENT (text is drawn without a background color).
public procedure set_background_mode( atom dc, atom mode )
    c_proc( WX_SET_BACKGROUND_MODE, {dc, mode} )
end procedure

--/topic Graphics
--/func get_dc_text_extent( atom dc, sequence text )
--/b Returns: Dimensions of the specified text
--
--Gets the dimensions of /i text using the font in the current /i dc.
public function get_dc_text_extent( atom dc, sequence text )
    return c_func( WX_GET_DC_TEXT_EXTENT, {dc, text} )
end function

--/topic Graphics
--/func get_user_scale( atom dc, atom x, atom y )
--/b Returns: Scaling factors
--
--Returns the scaling factors for the DC.
public function get_user_scale( atom dc )
    return c_func( WX_GET_USER_SCALE, {dc} )
end function

--/topic Graphics
--/proc set_user_scale( atom dc, atom x, atom y )
--
--Sets the scaling factors for the DC.
public procedure set_user_scale( atom dc, atom x, atom y )
    c_proc( WX_SET_USER_SCALE, {dc, x, y} )
end procedure

--/topic Graphics
--/func get_dc_size( atom dc )
--/b Returns: Width and height
--
--Returns the width and height of the dc in pixels.
public function get_dc_size( atom dc )
    return c_func( WX_GET_DC_SIZE, {dc} )
end function

--/topic Graphics
--/proc set_text_color( atom dc, object color )
--
--Set the current text foreground color.  /i color can be either an RGB sequence, or a pointer to
-- a wxColour object.
public procedure set_text_color( atom dc, object color )
    if sequence( color ) then
        color = create( wxColour, color )
    end if
    c_proc( WX_SET_TEXT_COLOR, {dc, color} )
end procedure

--/topic Graphics
--/proc draw_label( atom dc, object text, atom bitmap, object rect, atom alignment )
--
--Draw optional bitmap and the text into the given rectangle and aligns it as specified by alignment parameter.
public procedure draw_label( atom dc, object text, atom bitmap, object rect, atom alignment )
    c_proc( WX_DRAW_LABEL, {dc,text,bitmap,rect,alignment})
end procedure

--/topic Graphics
--/func get_font( atom dc )
--/b Returns: Font object
--
--Gets the current font. Notice that even although each device context object has some default font after creation,
-- this method would return a /i wxNullFont initially and only after calling /set_font a valid font is returned.
public function get_font( atom dc )
    return c_func( WX_GET_FONT, {dc})
end function


--IDs
------------------------------
--/topic IDs
--/parent Controls
--/parent Events
--/info
--Some things, such as /Events, require control IDs be used.
--
--Most operations in wxEuphoria are done using /i this pointers to objects (the value returned
--by /create().  Some things, such as events, however, require that control IDs be used.  It is
--recommended that you create your controls with the default ID of -1, and allow wxWidgets to
--automatically generate a unique id for you.  IDs are generated starting at -201 and decrease
--from there.  You can use /get_id() or /get_menuitem_id() to determine IDs when necessary.<br><br>
--There are some IDs, however, that have been set aside for specific uses, and will often be
--used for menu items or command buttons.  Here is a list:
--
--/b "Standard menu IDs"
--<ul>
--/li wxID_LOWEST
--/li wxID_OPEN
--/li wxID_CLOSE
--/li wxID_NEW
--/li wxID_SAVE
--/li wxID_SAVEAS
--/li wxID_REVERT
--/li wxID_EXIT
--/li wxID_UNDO
--/li wxID_REDO
--/li wxID_HELP
--/li wxID_PRINT
--/li wxID_PRINT_SETUP
--/li wxID_PREVIEW
--/li wxID_ABOUT
--/li wxID_HELP_CONTENTS
--/li wxID_HELP_COMMANDS
--/li wxID_HELP_PROCEDURES
--/li wxID_HELP_CONTEXT
--/li wxID_CUT
--/li wxID_COPY
--/li wxID_PASTE
--/li wxID_CLEAR
--/li wxID_FIND
--/li wxID_DUPLICATE
--/li wxID_SELECTALL
--/li wxID_FILE1
--/li wxID_FILE2
--/li wxID_FILE3
--/li wxID_FILE4
--/li wxID_FILE5
--/li wxID_FILE6
--/li wxID_FILE7
--/li wxID_FILE8
--/li wxID_FILE9
--</ul>
--
--/b Standard /b button /b IDs
--<ul>
--/li wxID_OK
--/li wxID_CANCEL
--/li wxID_APPLY
--/li wxID_YES
--/li wxID_NO
--/li wxID_STATIC
--/li wxID_FORWARD
--/li wxID_BACKWARD
--/li wxID_DEFAULT
--/li wxID_MORE
--/li wxID_SETUP
--/li wxID_RESET
--</ul>
--
--/b "IDs used by generic file dialog (11 consecutive starting from this value)"
--<ul>
--/li wxID_FILEDLGG
--/li wxID_HIGHEST
--</ul>

--/topic IDs
--/func new_id()
--/b Returns: A new ID
--
--Generates an integer identifier unique to this run of the program.  This can be used
-- to generate ids for menuitems or other controls.
public function new_id()
    return c_func( WX_NEW_ID, {} )
end function



--Key Codes
------------------------------
--/topic Key Codes
--/parent Key Events
--/info
--Key Code Constants
--<ul>
--/li WXK_BACK
--/li WXK_TAB
--/li WXK_RETURN
--/li WXK_ESCAPE
--/li WXK_SPACE
--/li WXK_DELETE
--/li WXK_START
--/li WXK_LBUTTON
--/li WXK_RBUTTON
--/li WXK_CANCEL
--/li WXK_MBUTTON
--/li WXK_CLEAR
--/li WXK_SHIFT
--/li WXK_ALT
--/li WXK_CONTROL
--/li WXK_MENU
--/li WXK_PAUSE
--/li WXK_CAPITAL
--/li WXK_PRIOR /i "(same as WXK_PAGEUP)"
--/li WXK_NEXT  /i "(same as WXK_PAGEDOWN)"
--/li WXK_END
--/li WXK_HOME
--/li WXK_LEFT
--/li WXK_UP
--/li WXK_RIGHT
--/li WXK_DOWN
--/li WXK_SELECT
--/li WXK_PRINT
--/li WXK_EXECUTE
--/li WXK_SNAPSHOT
--/li WXK_INSERT
--/li WXK_HELP
--/li WXK_NUMPAD0
--/li WXK_NUMPAD1
--/li WXK_NUMPAD2
--/li WXK_NUMPAD3
--/li WXK_NUMPAD4
--/li WXK_NUMPAD5
--/li WXK_NUMPAD6
--/li WXK_NUMPAD7
--/li WXK_NUMPAD8
--/li WXK_NUMPAD9
--/li WXK_MULTIPLY
--/li WXK_ADD
--/li WXK_SEPARATOR
--/li WXK_SUBTRACT
--/li WXK_DECIMAL
--/li WXK_DIVIDE
--/li WXK_F1
--/li WXK_F2
--/li WXK_F3
--/li WXK_F4
--/li WXK_F5
--/li WXK_F6
--/li WXK_F7
--/li WXK_F8
--/li WXK_F9
--/li WXK_F10
--/li WXK_F11
--/li WXK_F12
--/li WXK_F13
--/li WXK_F14
--/li WXK_F15
--/li WXK_F16
--/li WXK_F17
--/li WXK_F18
--/li WXK_F19
--/li WXK_F20
--/li WXK_F21
--/li WXK_F22
--/li WXK_F23
--/li WXK_F24
--/li WXK_NUMLOCK
--/li WXK_SCROLL
--/li WXK_PAGEUP
--/li WXK_PAGEDOWN
--/li WXK_NUMPAD_SPACE
--/li WXK_NUMPAD_TAB
--/li WXK_NUMPAD_ENTER
--/li WXK_NUMPAD_F1
--/li WXK_NUMPAD_F2
--/li WXK_NUMPAD_F3
--/li WXK_NUMPAD_F4
--/li WXK_NUMPAD_HOME
--/li WXK_NUMPAD_LEFT
--/li WXK_NUMPAD_UP
--/li WXK_NUMPAD_RIGHT
--/li WXK_NUMPAD_DOWN
--/li WXK_NUMPAD_PRIOR
--/li WXK_NUMPAD_PAGEUP
--/li WXK_NUMPAD_NEXT
--/li WXK_NUMPAD_PAGEDOWN
--/li WXK_NUMPAD_END
--/li WXK_NUMPAD_BEGIN
--/li WXK_NUMPAD_INSERT
--/li WXK_NUMPAD_DELETE
--/li WXK_NUMPAD_EQUAL
--/li WXK_NUMPAD_MULTIPLY
--/li WXK_NUMPAD_ADD
--/li WXK_NUMPAD_SEPARATOR
--/li WXK_NUMPAD_SUBTRACT
--/li WXK_NUMPAD_DECIMAL
--/li WXK_NUMPAD_DIVIDE
--</ul>



--Key Events
------------------------------
--/topic Key Events
--/parent Events
--/info
--Key Event Types
--<ul>
--/li wxEVT_CHAR
--/li wxEVT_CHAR_HOOK
--/li wxEVT_NAVIGATION_KEY
--/li wxEVT_KEY_DOWN
--/li wxEVT_KEY_UP
--</ul>

--/topic Key Events
--/func get_key_code( atom event )
--/b Returns: Virtual key code
--
--Returns the virtual key code. ASCII events return normal ASCII values, while non-ASCII events
-- return values such as WXK_LEFT for the left cursor key.
public function get_key_code( atom event )
    return c_func( WX_GET_KEY_CODE, {event} )
end function

--/topic Key Events
--/func key_event_shiftdown( atom event )
--/b Returns: True or false
--
--This returns an atom of 1 if the Shift key is currently pressed
-- down and 0 if it is not.
public function key_event_shiftdown( atom event )
    return c_func( WX_KEY_EVENT_SHIFTDOWN, {event} )
end function

--/topic Key Events
--/func key_event_controldown( atom event )
--/b Returns: True or false
--
--This returns an atom of 1 if the Control key is currently pressed down and
-- 0 if it is not.
public function key_event_controldown( atom event )
    return c_func( WX_KEY_EVENT_CONTROLDOWN, {event} )
end function

--/topic Key Events
--/func key_event_altdown( atom event )
--/b Returns: True or false
--
--This returns an atom of 1 if the Alt key is currently pressed
-- down and 0 if it is not.
public function key_event_altdown( atom event )
    return c_func( WX_KEY_EVENT_ALTDOWN, {event} )
end function



--Known Bugs
------------------------------
--/topic Known Bugs
--/info
--Stuff that needs fixing
--
--/b Win32
--<ul>
--/li Sometimes, controls in a /wxPanel won't size themselves properly at
-- program startup if the /wxPanel is created before menus are added.  The controls
-- will properly position themselves after their parent is resized, or if the creation
-- of the /wxPanel is moved after the menus are created.
--</ul>
--
--/b Linux
--<ul>
--/li /wxMDIChildFrame windows do not show up correctly. This is being investigated.
--</ul>
--
--/b Both
--<ul>
--/li The /b wxEVT_MEDIA* events don't work yet.
--</ul>
--



--Menus
------------------------------
--/topic Menus
--/info
--Using menus
-- requires you to create at least three objects: /wxMenuBar, /wxMenu and /wxMenuItem.
--
--The /wxMenuBar is child to the /wxFrame of your application.  The /wxMenuBar can have multiple
-- /wxMenu children, and each /wxMenu can have multiple /wxMenuItem children.
--
--<div class="subsection">Menu Events</div>
--<ul>
--/li wxEVT_COMMAND_MENU_SELECTED
--/li wxEVT_MENU_CHAR
--/li wxEVT_MENU_INIT
--/li wxEVT_MENU_HIGHLIGHT
--/li wxEVT_POPUP_MENU_INIT
--</ul>
--When setting a /b wxEVT_COMMAND_MENU_SELECTED event, you should pass the pointer to the /wxMenu parent
-- of the /wxMenuItem.
--
--Use of the { /wxFrame, /wxMenu } parameter is deprecated, and should not be used on any platform.

--/topic Menus
--/func get_menuitem_id( atom item )
--/b Returns: Menu ID
--
--Returns the id for a /wxMenuItem.  Since /wxMenuItem is not derived from /wxWindow,
-- you should never use /get_id() with a /wxMenuItem.
public function get_menuitem_id( atom item )
    return c_func( WX_GET_MENUITEM_ID, {item} )
end function

--/topic Menus
--/func get_menuevent_id( atom event )
--/b Returns: MenuItem ID
--
--Returns the id of the /wxMenuItem that generated the event.  This should only be
-- called from within a wxEVT_COMMAND_MENU_SELECTED event procedure.
public function get_menuevent_id( atom event )
    return c_func( WX_GET_MENUEVENT_ID, {event} )
end function

--/topic Menus
--/func insert_menuitem( atom menu, integer position, atom id, sequence text, sequence help, integer checkable )
--/b Returns: True or false
--
--Insert a /wxMenuItem at /i position in the specified /i menu.
public function insert_menuitem( atom menu, integer position, atom id, sequence text, sequence help,    integer checkable )
    return c_func( WX_INSERT_MENUITEM, {menu, position, id, text, help, checkable} )
end function

--/topic Menus
--/proc set_menu_label( atom menu, atom id, sequence label_text )
--
--Use to change the text of a menu item.  /i id should be the id
-- of the menuitem.
public procedure set_menu_label( atom menu, atom id, sequence label_text )
    c_proc( WX_SET_MENU_LABEL, {menu, id, label_text} )
end procedure

--/topic Menus
--/func is_menuitem_checked( atom menuitem )
--/b Returns: Checked state of the MenuItem
--
--Returns 1 if the menuitem is checked, 0 otherwise.
public function is_menuitem_checked( atom menuitem )
    return c_func( WX_IS_MENUITEM_CHECKED, {menuitem} )
end function

--/topic Menus
--/proc check_menuitem( atom menuitem, integer check )
--
--If /i check is 0, then the /i menuitem is unchecked.  If /check is 1,
-- then the /i menuitem is checked.
public procedure check_menuitem( atom menuitem, integer check )
    c_proc( WX_CHECK_MENUITEM, {menuitem, check} )
end procedure

--/topic Menus
--/proc enable_menuitem( atom menuitem, integer enable )
--
--1 enables a /i menuitem, 0 disables.
public procedure enable_menuitem( atom menuitem, integer enable )
    c_proc( WX_ENABLE_MENUITEM, {menuitem, enable} )
end procedure

--/topic Menus
--/proc remove_menuitem( atom menu, atom item )
--
--Removes a /wxMenuItem from a /wxMenu and deletes the /wxMenuItem.
public procedure remove_menuitem( atom menu, atom item )
    c_proc( WX_REMOVE_MENUITEM, {menu, item} )
end procedure

--/topic Menus
--/proc set_menuitem_bitmap( atom item, atom bitmap )
--
public procedure set_menuitem_bitmap( atom item, atom bitmap )
    c_proc( WX_SET_MENUITEM_BITMAP, {item,bitmap})
end procedure



--Miscellaneous Routines
------------------------------
--/topic Miscellaneous Routines
--/info
--Random stuff
--

--/topic Miscellaneous Routines
--/func get_sys_metric( integer metric )
--/b Returns: atom
--
--Available metrics are:
--<ul>
--/li wxSYS_MOUSE_BUTTONS
--/li wxSYS_BORDER_X
--/li wxSYS_BORDER_Y
--/li wxSYS_CURSOR_X
--/li wxSYS_CURSOR_Y
--/li wxSYS_DCLICK_X
--/li wxSYS_DCLICK_Y
--/li wxSYS_DRAG_X
--/li wxSYS_DRAG_Y
--/li wxSYS_EDGE_X
--/li wxSYS_EDGE_Y
--/li wxSYS_HSCROLL_ARROW_X
--/li wxSYS_HSCROLL_ARROW_Y
--/li wxSYS_HTHUMB_X
--/li wxSYS_ICON_X
--/li wxSYS_ICON_Y
--/li wxSYS_ICONSPACING_X
--/li wxSYS_ICONSPACING_Y
--/li wxSYS_WINDOWMIN_X
--/li wxSYS_WINDOWMIN_Y
--/li wxSYS_SCREEN_X
--/li wxSYS_SCREEN_Y
--/li wxSYS_FRAMESIZE_X
--/li wxSYS_FRAMESIZE_Y
--/li wxSYS_SMALLICON_X
--/li wxSYS_SMALLICON_Y
--/li wxSYS_HSCROLL_Y
--/li wxSYS_VSCROLL_X
--/li wxSYS_VSCROLL_ARROW_X
--/li wxSYS_VSCROLL_ARROW_Y
--/li wxSYS_VTHUMB_Y
--/li wxSYS_CAPTION_Y
--/li wxSYS_MENU_Y
--/li wxSYS_NETWORK_PRESENT
--/li wxSYS_PENWINDOWS_PRESENT
--/li wxSYS_SHOW_SOUNDS
--/li wxSYS_SWAP_BUTTONS
--</ul>
public function get_sys_metric( integer metric )
    return c_func( WX_GET_SYS_METRIC, {metric} )
end function

--/topic Miscellaneous Routines
--/func message_box( sequence text, sequence title, atom ok )
--/b Returns: Selected button ID
--
--Creates a message box dialog, with style /i ok.
--
--Allowable styles are:
--<ul>
--/li wxOK
--/li wxYES_NO
--/li wxCANCEL
--/li wxYES
--/li wxNO
--/li wxNO_DEFAULT
--/li wxYES_DEFAULT
--/li wxICON_EXCLAMATION
--/li wxICON_HAND
--/li wxICON_WARNING
--/li wxICON_ERROR
--/li wxICON_QUESTION
--/li wxICON_INFORMATION
--/li wxICON_STOP
--/li wxICON_ASTERISK
--</ul>
--
--Possible return values are:
--<ul>
--/li wxOK
--/li wxCANCEL
--/li wxYES
--/li wxNO
--</ul>
public function message_box( object s1, object s2, atom ok )
    return c_func( WX_MESSAGE_BOX, {s1, s2, ok} )
end function

--/topic Miscellaneous Routines
--/proc wx_execute( sequence command, integer flag )
--
--/i flag can be a combination of:
--<ul>
--/li wxEXEC_ASYNC: execute the process asynchronously
--/li wxEXEC_SYNC: execute it synchronously, i.e. wait until it finishes
--/li wxEXEC_NOHIDE: under Windows, don't hide the child even if it's IO is redirected (this
-- is done by default)
--/li wxEXEC_MAKE_GROUP_LEADER: under Unix, if the process is the group leader then killing -pid kills
-- all children as well as pid
--</ul>
public function wx_execute( sequence command, integer flag )
    return c_func( WX_WX_EXECUTE, {command, flag} )
end function

--/topic Miscellaneous Routines
--/proc wx_shell( sequence command )
--
--Executes a /i command in the default shell
public procedure wx_shell( sequence command )
    c_proc( WX_WX_SHELL, {command} )
end procedure

--/topic Miscellaneous Routines
--/func get_key_state( integer keycode )
--/b Returns: True or false
--
--Returns 1 if specified key (see /"Key Codes" for a list of codes to use) is currently
-- pressed down.
public function get_key_state( integer keycode )
    return c_func( WX_GET_KEY_STATE, {keycode} )
end function

--/topic Miscellaneous Routines
--/proc cleanup()
--
--This cleans up internal wxWidgets memory.  /wxMain() automatically calls this before
-- it exits.  If you do not use /wxMain(), then you should manually call cleanup() before
-- your app exits.
public procedure cleanup()
    c_proc( WX_CLEANUP, {} )
end procedure

--/topic Miscellaneous Routines
--/func find_window( integer id, atom parent )
--/b Returns: Window pointer
--
--Returns the window identified by its id and parent.
public function find_window (integer id, atom parent)
    return c_func( WX_FIND_WINDOW, {id, parent} )
end function

--/topic Miscellaneous Routines
--/proc wxMain( atom topWindow )
--
--This starts the event loop and your main application.
public procedure wxMain( atom topWindow )
    c_proc( WX_WXMAIN, {topWindow} )
end procedure

--/topic Miscellaneous Routines
--/func mkdir( sequence directory, object perm, integer flags )
--
--Creates a /i directory with the (optional) given permissions.
--
--If the /i flags contain /b wxPATH_MKDIR_FULL flag, try to create each directory in
-- the path and also don't return an error if the target directory already exists.
public function mkdir( sequence directory, object perm, integer flags )
    return c_func( WX_WX_MKDIR, {directory, perm, flags} )
end function

--/topic Miscellaneous Routines
--/func w_sprintf( sequence format, object values )
--
--A unicode replacement for the builtin sprintf().
public function w_sprintf( sequence format, object values )
    return c_func( WX_W_SPRINTF, {format,values} )
end function

--/topic Miscellaneous Routines
--/func get_sys_color( atom color )
--/b Returns: pointer to /wxColour
--Available system colors are:
--<ul>
--/li wxSYS_COLOUR_SCROLLBAR
--/li wxSYS_COLOUR_BACKGROUND
--/li wxSYS_COLOUR_ACTIVECAPTION
--/li wxSYS_COLOUR_INACTIVECAPTION
--/li wxSYS_COLOUR_MENU
--/li wxSYS_COLOUR_WINDOW
--/li wxSYS_COLOUR_WINDOWFRAME
--/li wxSYS_COLOUR_MENUTEXT
--/li wxSYS_COLOUR_WINDOWTEXT
--/li wxSYS_COLOUR_CAPTIONTEXT
--/li wxSYS_COLOUR_ACTIVEBORDER
--/li wxSYS_COLOUR_INACTIVEBORDER
--/li wxSYS_COLOUR_APPWORKSPACE
--/li wxSYS_COLOUR_HIGHLIGHT
--/li wxSYS_COLOUR_HIGHLIGHTTEXT
--/li wxSYS_COLOUR_BTNFACE
--/li wxSYS_COLOUR_BTNSHADOW
--/li wxSYS_COLOUR_GRAYTEXT
--/li wxSYS_COLOUR_BTNTEXT
--/li wxSYS_COLOUR_INACTIVECAPTIONTEXT
--/li wxSYS_COLOUR_BTNHIGHLIGHT
--/li wxSYS_COLOUR_3DDKSHADOW
--/li wxSYS_COLOUR_3DLIGHT
--/li wxSYS_COLOUR_INFOTEXT
--/li wxSYS_COLOUR_INFOBK
--/li wxSYS_COLOUR_LISTBOX
--/li wxSYS_COLOUR_DESKTOP
--/li wxSYS_COLOUR_3DFACE
--/li wxSYS_COLOUR_3DSHADOW
--/li wxSYS_COLOUR_3DHIGHLIGHT
--/li wxSYS_COLOUR_3DHILIGHT
--/li wxSYS_COLOUR_BTNHILIGHT
--</ul>
public function get_sys_color( atom color )
    return c_func( WX_GET_SYS_COLOR, {color} )
end function

--/topic Miscellaneous Routines
--/proc delete_instance( object obj )
--
--Deletes an object from memory.
public procedure delete_instance( object obj )
    c_proc( WX_DELETE_INSTANCE, {obj})
end procedure

--/topic Miscellaneous Routines
--/proc launch_default_browser( sequence url )
--
--Launches the user's default browser to open the specified url.
public procedure launch_default_browser( sequence url )
    c_proc( WX_LAUNCH_DEFAULT_BROWSER, {url})
end procedure

--/topic Miscellaneous Routines
--/func to_utf8( sequence string )
--
--Returns a sequence with the specified string in UTF8 encoding.
--
--See also /from_utf8()
public function to_utf8( sequence string )
    return c_func( WX_TO_UTF8, {string})
end function

--/topic Miscellaneous Routines
--/func from_utf8( sequence utf8 )
--
--Converts a sequence with UTF8 encoded text into a unicode string that can
--be used like a normal string.
--
--See also /to_utf8()
public function from_utf8( sequence utf8 )
    return c_func( WX_FROM_UTF8, {utf8})
end function

--/topic Miscellaneous Routines
--/func get_display_size()
--
--Returns the display size in pixels.
public function get_display_size(  )
    return c_func( WX_GET_DISPLAY_SIZE, {})
end function



--Mouse Events
------------------------------
--/topic Mouse Events
--/parent Events
--/info
--Mouse Event Routines

--/topic Mouse Events
--/func mouse_event_shiftdown( atom event )
--
--Returns true if the shift key was down at the time of the event.
public function mouse_event_shiftdown( atom event )
    return c_func( WX_MOUSE_EVENT_SHIFTDOWN, {event} )
end function

--/topic Mouse Events
--/func mouse_event_position( atom event )
--
--Returns the coordinates of the mouse at the time of the event.
public function mouse_event_position( atom event )
    return c_func( WX_MOUSE_EVENT_POSITION, {event} )
end function


--Networking
------------------------------
--/topic Networking
--/info
--Doing things across networks
--

--/topic Networking
--/proc socket_write( atom socket, sequence bytes )
--
--Works for both /wxSocketServer and /wxSocketClient.
public procedure socket_write( atom stream, sequence bytes )
    c_proc( WX_SOCKET_WRITE, {stream, bytes} )
end procedure

--/topic Networking
--/proc socket_read( atom socket )
--
--Works for /wxSocketServer, /wxSocketClient, wxHTTP and wxFTP.
public function socket_read( atom socket )
    return c_func( WX_SOCKET_READ, {socket} )
end function

--/topic Networking
--/proc set_socket_notify( atom socket, object notify )
--
--Works for both /wxSocketServer and /wxSocketClient.
--Specifies which socket events are to be sent to the event handler. The
--flags parameter may be combination of flags ORed together. The following flags can be used:
--<ul>
--/li /b wxSOCKET_INPUT_FLAG  to receive wxSOCKET_INPUT
--/li /b wxSOCKET_OUTPUT_FLAG  to receive wxSOCKET_OUTPUT
--/li /b wxSOCKET_CONNECTION_FLAG  to receive wxSOCKET_CONNECTION
--/li /b wxSOCKET_LOST_FLAG  to receive wxSOCKET_LOST
--</ul>
public procedure set_socket_notify( atom socket, object notify )
    c_proc( WX_SET_SOCKET_NOTIFY, {socket, notify} )
end procedure

--/topic Networking
--/func is_connected( atom socket )
--
--Returns a 0 if the socket is not connected, 1 if it is.
public function is_connected( atom socket )
    return c_func( WX_IS_CONNECTED, {socket} )
end function

--/topic Networking
--/proc socket_event_handler( atom socket, atom evt_handler, atom id )
--
--When calling /set_event_handler() for a wxEVT_SOCKET event, for the 'this'
--parameter, you must pass a sequence consisting of the object to handle the
--events and the socket.
--
--Example:
--/code
--   socket_event_handler( socket, form, socket_id )
--   set_event_handler( {form, socket}, -1, wxEVT_SOCKET, routine_id("socket_event"))
--/endcode
public procedure socket_event_handler( atom socket, atom evt_handler, atom id )
    c_proc( WX_SOCKET_EVENT_HANDLER, {socket, evt_handler, id} )
end procedure

--/topic Networking
--/func get_socket_event( atom event )
--
--Returns the socket event type.
public function get_socket_event( atom event )
    return c_func( WX_GET_SOCKET_EVENT, {event} )
end function

--/topic Networking
--/func socket_from_event( atom event )
--
--Returns the pointer to the socket that sent the wxSocketEvent.
public function socket_from_event( atom event )
    return c_func( WX_SOCKET_FROM_EVENT, {event} )
end function

--/topic Networking
--/proc set_flags( atom socket, atom flags )
--
--Sets a sequence of flags to a specified socket. You must assign a
--socket flags after every procedure/function loop or it will not
--look for them and you could lose soemthing.
--
--Events that can be chosen are
--<ul>
--/li wxSOCKET_INPUT -When the Server or Client receives input
--/li wxSOCKET_OUTPUT - When the Server or Client outputs
--/li wxSOCKET_CONNECT - This fires when there is a new connection
--/li wxSOCKET_LOST - When a socket connection is dropped.
--</ul>
--Alternatively, you can do set_flags ( Socket, {0} ) which will
--make the socket stop waiting for events. However, the connection
--will still remain.
public procedure set_flags( atom socket, atom flags )
    c_proc( WX_SET_FLAGS, {socket, flags} )
end procedure


--/topic Non wxObjects
--/proc delete_instance_of( integer class, object obj )
--
--Use this to delete objects that aren't derived from wxObject.
--/i obj can be either an atom or a sequence of atoms.
public procedure delete_instance_of( integer clazz, object obj )
    integer ix
    ix = find( clazz, INSTANCE_OF )
    if ix then
        c_proc( DELETE_PROC[ix], {obj})
    end if
end procedure

--/topic Non wxObjects
--/topic wxRegEx
--/proc delete_wxRegEx( object obj )
--
--Deletes a /wxRegEx object.
public procedure delete_wxRegEx( object obj )
    c_proc( WX_DELETE_WXREGEX, {obj})
end procedure

--/topic Non wxObjects
--/proc delete_wxAuiPaneInfo( object obj )
--
--Deletes a /wxAuiPaneInfo object.
public procedure delete_wxAuiPaneInfo( object obj )
    c_proc( WX_DELETE_WXAUIPANEINFO, {obj})
end procedure



--Printing
------------------------------
--/topic Printing
--/info
--Using a printer
--



--Sizers
------------------------------
--/topic Sizers
--/info
--Automatically manage resizing controls
--
--Sizers take care of resizing and repositioning controls within a window whenever
--the window is resized.  Sizers may be nested within each other to create complex
--relationships between the controls in a window.
--
--The most common sizer used will be the /wxBoxSizer.

--/topic Sizers
--/proc add_window_to_sizer( atom sizer, atom window, atom option, atom flag, atom border )
--
--This adds the specified window (i.e., control) to the sizer.
--<ul>
--/li /i option can be any integer greater than or equal to 0.  For /wxBoxSizer, 0 means that
--the control cannot grow in the main direction of the sizer (either wxHORIZONTAL
--or wxVERTICAL--see /wxBoxSizer).  Greater than 0 means that the control will grow
--in proportion to the /i option values of the other controls placed in the sizer.
--/li /i flag may take on any of several values (use /wx_or_all() to combine multiple):
--<ul>
--/li wxGROW: grow to fill entire space
--/li wxSHAPED: grow proportionately
--/li wxTOP: the border may be applied at the top of the sizer
--/li wxBOTTOM: the border may be applied at the bottom of the sizer
--/li wxRIGHT: the border may be applied at the right of the sizer
--/li wxLEFT: the border may be applied at the left of the sizer
--/li wxALIGN_CENTER_HORIZONTAL
--/li wxALIGN_LEFT
--/li wxALIGN_TOP
--/li wxALIGN_RIGHT
--/li wxALIGN_BOTTOM
--/li wxALIGN_CENTER_VERTICAL
--</ul>
--/li /i border is the size of the border
--</ul>
public procedure add_window_to_sizer( atom sizer, atom window, atom option, atom flag, atom border )
    c_proc( WX_ADD_WINDOW_TO_SIZER, {sizer, window, option, flag, border} )
end procedure

--/topic Sizers
--/proc add_sizer_to_sizer( atom sizer1, atom sizer2, atom option, atom flag, atom border )
--
--This adds sizer2 to sizer1.
--<ul>
--/li /i option can be any integer greater than or equal to 0.  For /wxBoxSizer, 0 means that
--the sizer cannot grow in the main direction of the sizer (either wxHORIZONTAL
--or wxVERTICAL--see /wxBoxSizer).  Greater than 0 means that the sizer will grow
--in proportion to the /i option values of the other controls placed in the sizer.
--/li /i flag may take on any of several values (use /wx_or_all() to combine multiple):
--<ul>
--/li wxGROW: grow to fill entire space
--/li wxSHAPED: grow proportionately
--/li wxTOP: the border may be applied at the top of the sizer
--/li wxBOTTOM: the border may be applied at the bottom of the sizer
--/li wxRIGHT: the border may be applied at the right of the sizer
--/li wxLEFT: the border may be applied at the left of the sizer
--/li wxALIGN_CENTER_HORIZONTAL
--/li wxALIGN_LEFT
--/li wxALIGN_TOP
--/li wxALIGN_RIGHT
--/li wxALIGN_BOTTOM
--/li wxALIGN_CENTER_VERTICAL
--</ul>
--/li /i border is the size of the border
--</ul>
public procedure add_sizer_to_sizer( atom sizer, atom window, atom option, atom flag, atom border )
    c_proc( WX_ADD_SIZER_TO_SIZER, {sizer, window, option, flag, border} )
end procedure

--/topic Sizers
--/proc space_sizer( atom sizer, atom width, atom height, atom option, atom flag, atom border)
--
--Adds a blank region to the sizer in order to space controls within a sizer.
--<ul>
--/li /i option can be any integer greater than or equal to 0.  For /wxBoxSizer, 0 means that
--the sizer cannot grow in the main direction of the sizer (either wxHORIZONTAL
--or wxVERTICAL--see /wxBoxSizer).  Greater than 0 means that the sizer will grow
--in proportion to the /i option values of the other controls placed in the sizer.
--/li /i flag may take on any of several values (use /wx_or_all() to combine multiple):
--<ul>
--/li wxGROW: grow to fill entire space
--/li wxSHAPED: grow proportionately
--/li wxTOP: the border may be applied at the top of the sizer
--/li wxBOTTOM: the border may be applied at the bottom of the sizer
--/li wxRIGHT: the border may be applied at the right of the sizer
--/li wxLEFT: the border may be applied at the left of the sizer
--/li wxALIGN_CENTER_HORIZONTAL
--/li wxALIGN_LEFT
--/li wxALIGN_TOP
--/li wxALIGN_RIGHT
--/li wxALIGN_BOTTOM
--/li wxALIGN_CENTER_VERTICAL
--</ul>
--/li /i border is the size of the border
--</ul>
public procedure space_sizer( atom sizer, atom width, atom height, atom option, atom flag, atom border)
    c_proc( WX_SPACE_SIZER, {sizer, width, height, option, flag, border} )
end procedure

--/topic Sizers
--/proc set_sizer( atom window, atom sizer )
--
--This must be called before a /i sizer will resize its children. /i window is the parent window
--of the controls to be resized.  The sizer will resize and reposition its children based on /i window.
public procedure set_sizer( atom window, atom sizer )
    c_proc( WX_SET_SIZER, {window, sizer} )
end procedure

--/topic Sizers
--/func fit_window( atom sizer, atom window )
--
--Tell the /i sizer to resize the /i window to match the sizer's minimal size. Returns the new size.
public function fit_window( atom sizer, atom window )
    return c_func( WX_FIT_WINDOW, {sizer, window} )
end function

--/topic Sizers
--/func get_sizer_size( atom sizer )
--
--Returns the current size of the /i sizer.
public function get_sizer_size( atom sizer )
    return c_func( WX_GET_SIZER_SIZE, {sizer} )
end function

--/topic Sizers
--/func get_sizer_minsize( atom sizer )
--
--Returns the minimal size of the /i sizer. This is either the combined minimal
-- size of all the children and their borders or the minimal size set by /set_sizer_min_size(),
-- depending on which is bigger
public function get_sizer_minsize( atom sizer )
    return c_func( WX_GET_SIZER_MINSIZE, {sizer} )
end function

--/topic Sizers
--/func get_sizer_position( atom sizer )
--
--Returns the current position of the sizer.
public function get_sizer_position( atom sizer )
    return c_func( WX_GET_SIZER_POSITION, {sizer} )
end function

--/topic Sizers
--/func sizer_calc_min( atom sizer )
--
--Have the /i sizer perform the actual calculation of its children's minimal sizes.  The
-- size is returned as a sequence.
public function sizer_calc_min( atom sizer )
    return c_func( WX_SIZER_CALC_MIN, {sizer} )
end function

--/topic Sizers
--/proc detatch_sizer( atom sizer, atom detach )
--
--Detach a child from the sizer without destroying it. his method does not cause any layout or resizing to take place,
-- call /layout_sizer() to update the layout "on screen" after detaching a child from the sizer.
public procedure detatch_sizer( atom sizer, atom detach )
    c_proc( WX_DETATCH_SIZER, {sizer,detach})
end procedure

--/topic Sizers
--/func get_sizer_window( atom sizer )
--
--Returns the window this /i sizer is used in or /i NULL if none.
public function get_sizer_window( atom sizer )
    return c_func( WX_GET_SIZER_WINDOW, {sizer})
end function

--/topic Sizers
--/func sizer_item_shown( atom sizer, atom item )
--
--Returns true if the /i item is shown in the  /i sizer.
public function sizer_item_shown( atom sizer, atom item )
    return c_func( WX_SIZER_ITEM_SHOWN, {sizer,item})
end function

--/topic Sizers
--/proc layout_sizer( atom sizer )
--
--Call this to force layout of the children anew, e.g. after having added a child to or removed a child (window, other sizer or space)
-- from the sizer while keeping the current dimension.
public procedure layout_sizer( atom sizer )
    c_proc( WX_LAYOUT_SIZER, {sizer})
end procedure

--/topic Sizers
--/func replace_sizer_item( atom sizer, atom old_item, atom new_item, atom recursive )
--
--Replaces one item with another.  Both old_itm and new_item must be either sizers, or
--both must be objects derived from a wxWindow.
--
--Use parameter recursive to search the given element recursively in subsizers.
--
--You must call /layout_sizer() to have the sizer adjust the display.
public function replace_sizer_item( atom sizer, atom old_item, atom new_item, atom recursive )
    return c_func( WX_REPLACE_SIZER_ITEM, {sizer,old_item,new_item,recursive})
end function

--/topic Sizers
--/proc show_sizer_item_sizer( atom sizer, atom item, atom show, atom recursive )
--
--Shows or hides sizer /i item. To make a sizer item disappear or reappear,
--you must call layout_sizer() after this procedure.
--
--Use parameter recursive to search the given element recursively in subsizers.
public procedure show_sizer_item_sizer( atom sizer, atom item, atom show, atom recursive )
    c_proc( WX_SHOW_SIZER_ITEM_SIZER, {sizer,item,show,recursive})
end procedure

--/topic Sizers
--/proc show_sizer_item_win( atom sizer, atom item, atom show, atom recursive )
--
--Shows or hides window /i item. To make a sizer item disappear or reappear,
--you must call layout_sizer() after this procedure
--
--Use parameter recursive to search the given element recursively in subsizers.
public procedure show_sizer_item_win( atom sizer, atom item, atom show, atom recursive )
    c_proc( WX_SHOW_SIZER_ITEM_WIN, {sizer,item,show,recursive})
end procedure



--Timers
------------------------------
--/topic Timers
--/info
--Setting and using timers
--



--TODO
------------------------------
--/topic TODO
--/info
--What's next?
--This is a list of the classes I plan to wrap, although I'm not promising any
--particular class will make any particular version, or even the order in which
--these get implemented.  If you want or need something on this list (or even
--something not on this list), let me know, and I'll probably move it up on my
--priority list.  Better yet, write the wrappers yourself and send them to me!
--
--<ul><table width="90%">
--<tr>
--<td valign="top" bgcolor="#E0E0E0"> /b Controls </td>
--<td bgcolor="#D4D4D4"><ul>
--/li wxConfig classes
--/li wxArchive classes
--/li wxDragImage
--/li wxPropertySheetDialog
--/li wxAutomationObject (low priority, as this is MSW only)
--/li There are many functions not yet wrapped.  Please let me know if there is something you need.
--</ul></td></tr>
--<td valign="top" bgcolor="#E0E0E0"> /b Database </td>
--<td bgcolor="#D4D4D4">Low priority now that
--<a href="http://www.rapideuphoria.com/cgi-bin/asearch.exu?gen=on&keywords=odbc">
--odbc.e</a> is cross platform<ul>
--/li wxDB
--/li wxDBConnectInf
--/li wxDBTable
--</ul></td></tr>
--<td valign="top" bgcolor="#E0E0E0"> /b Printing </td>
--<td bgcolor="#D4D4D4">These don't work well yet:<ul>
--/li wxPrinter
--/li wxPrinterDC
--/li wxPrintPreview
--</ul></td></tr>
--<td valign="top" bgcolor="#E0E0E0"> /b Documentation </td>
--<td bgcolor="#D4D4D4">(Please contribute!)<ul>
--/li Filled in and improved just about everywhere
--/li More and better demos
--</ul></td></tr>
--<td valign="top" bgcolor="#E0E0E0"> /b General </td>
--<td bgcolor="#D4D4D4"><ul>
--/li wxPowerEvent
--</ul>
--</table></ul>



--Unicode
---------------------------------------
--/topic Unicode
--/info Strings in wxEuphoria
--All strings in wxEuphoria are Unicode.
-- It uses wide characters (which are basically 4-byte integers) to create wxStrings.
-- If your application only uses ASCII characters, you probably won't even notice a difference.
-- If you use Unicode characters, they should be stored as an integer (or atom) in a sequence,
-- using the normal U+xxxx values.
--
--Note that you may need to switch fonts (especially under windows) in order to see all of
-- the glyphs, as not all fonts have all glyphs.



--wxAboutDialogInfo
-----------------------------------------
--/topic wxAboutDialogInfo
--/parent Classes
--/info
--Contains information shown in the standard /i About dialog displayed by the /about_dlg_show() function.
--
--This function shows the standard about dialog containing the information
--specified in info. If the current platform has a native about dialog
--which is capable of showing all the fields in info, the native dialog
--is used, otherwise the function falls back to the generic wxWidgets
--version of the dialog

--/topic wxAboutDialogInfo
--/proc about_dlg_add_artist( atom info, object person )
--
--Adds an artist name to be shown in the program credits.
public procedure about_dlg_add_artist( atom info, object person )
    c_proc( WX_ABOUT_DLG_ADD_ARTIST, {info,person})
end procedure

--/topic wxAboutDialogInfo
--/proc about_dlg_add_developer( atom info, object person )
--
--Adds a developer name to be shown in the program credits.
public procedure about_dlg_add_developer( atom info, object person )
    c_proc( WX_ABOUT_DLG_ADD_DEVELOPER, {info,person})
end procedure

--/topic wxAboutDialogInfo
--/proc about_dlg_add_doc_writer( atom info, object person )
--
--Adds a documentation writer name to be shown in the program credits.
public procedure about_dlg_add_doc_writer( atom info, object person )
    c_proc( WX_ABOUT_DLG_ADD_DOC_WRITER, {info,person})
end procedure

--/topic wxAboutDialogInfo
--/proc about_dlg_add_translator( atom info, object person )
--
--Adds a translator name to be shown in the program credits. Notice that if no translator names are specified explicitely,
-- /about_dlg_show() will try to use the translation of the string translator-credits from the currently used message catalog
-- -- this can be used to show just the name of the translator of the program in the current language.
public procedure about_dlg_add_translator( atom info, object person )
    c_proc( WX_ABOUT_DLG_ADD_TRANSLATOR, {info,person})
end procedure

--/topic wxAboutDialogInfo
--/proc about_dlg_set_copyright( atom info, object copyright )
--
--Set the short string containing the program copyright information. Notice that any occurrences of "(C)" in copyright will be
-- replaced by the copyright symbol (circled C, &copy;) automatically, which means that you can avoid using this symbol in the program
-- source code which can be problematic.
public procedure about_dlg_set_copyright( atom info, object copyright )
    c_proc( WX_ABOUT_DLG_SET_COPYRIGHT, {info,copyright})
end procedure

--/topic wxAboutDialogInfo
--/proc about_dlg_set_description( atom info, object desc )
--
--Set brief, but possibly multiline, description of the program.
public procedure about_dlg_set_description( atom info, object desc )
    c_proc( WX_ABOUT_DLG_SET_DESCRIPTION, {info,desc})
end procedure

--/topic wxAboutDialogInfo
--/proc about_dlg_set_icon( atom info, atom icon )
--
--Set the icon to be shown in the dialog. By default the icon of the main frame will be shown if the native about dialog supports
-- custom icons. If it doesn't but a valid icon is specified using this method, the generic about dialog is used instead so you
-- should avoid calling this function for maximally native look and feel.
public procedure about_dlg_set_icon( atom info, atom icon )
    c_proc( WX_ABOUT_DLG_SET_ICON, {info,icon})
end procedure

--/topic wxAboutDialogInfo
--/proc about_dlg_set_license( atom info, object license )
--
--Set the long, multiline string containing the text of the program licence.
--
--Only GTK+ version supports showing the licence text in the native about dialog currently so the generic version will be used under
-- all the other platforms if this method is called. To preserve the native look and feel it is advised that you do not call this
-- method but provide a separate menu item in the "Help" menu for displaying the text of your program licence.
public procedure about_dlg_set_license( atom info, object license )
    c_proc( WX_ABOUT_DLG_SET_LICENSE, {info,license})
end procedure

--/topic wxAboutDialogInfo
--/proc about_dlg_set_name( atom info, object name )
--
--Set the name of the program. If this method is not called, the string returned by wxApp::GetAppName() will be shown in the dialog.
public procedure about_dlg_set_name( atom info, object name )
    c_proc( WX_ABOUT_DLG_SET_NAME, {info,name})
end procedure

--/topic wxAboutDialogInfo
--/proc about_dlg_set_version( atom info, object version )
--
--Set the version of the program. The version is in free format, i.e. not necessarily in the x.y.z form but it shouldn't contain the
-- "version" word.
public procedure about_dlg_set_version( atom info, object version )
    c_proc( WX_ABOUT_DLG_SET_VERSION, {info,version})
end procedure

--/topic wxAboutDialogInfo
--/proc about_dlg_set_web_site( atom info, object params )
--
--/i params should be {"url", "description"}
--
--Set the web site for the program and its description (which defaults to URL itself if empty).
--
--Please notice that only GTK+ version currently supports showing the link in the native about dialog so if this method is called,
-- the generic version will be used under all the other platforms.
public procedure about_dlg_set_web_site( atom info, object params )
    c_proc( WX_ABOUT_DLG_SET_WEB_SITE, {info,params})
end procedure

--/topic wxAboutDialogInfo
--/proc about_dlg_show( atom info )
--
--This function shows the standard about dialog containing the information specified in info. If the current platform has a native
-- about dialog which is capable of showing all the fields in info, the native dialog is used, otherwise the function falls back to
-- the generic wxWidgets version of the dialog.
--
--Example:
--/code
--
--procedure show_simple_about_dialog()
--
--    atom info
--
--    info = create( wxAboutDialogInfo, {} )
--    about_dlg_set_name( info, "My Program" )
--    about_dlg_set_version( info, "1.2.3 Beta" )
--    about_dlg_set_description( info, "This program does something great." )
--    about_dlg_set_copyright( info, "(C) 2007 Me <me@email.addre.ss>" )
--
--    about_dlg_show( info )
--
--end procedure
--
--/endcode
public procedure about_dlg_show( atom info )
    c_proc( WX_ABOUT_DLG_SHOW, {info})
end procedure



--wxApp
------------------------------
--/topic wxApp
--/parent Classes
--/info
--The wxApp class represents the application itself.
--
--You probably won't need to do much with this class.  wxEuphoria creates
--a wxApp object automatically.

--/topic wxApp
--/proc exit_main()
--
--This is the procedure that should be used to exit the program.  If you
--create an on close event for your main frame, this must be called from
--the event handler, or your application will not close.  wxEuphoria automatically
--does this for you if you do not define an on close event.
public procedure exit_main()
    c_proc( WX_EXIT_MAIN, {} )
end procedure

--procedure OnCloseWindow( atom this, atom event_type, atom id, atom event )
--  c_proc( WX_ONCLOSEWINDOW, {this, event_type, id, event} )
--end procedure

--/topic wxApp
--/func app_yield( integer onlyIfNeeded )
--
--Yields control to pending messages in the windowing system. This can be useful, for example, when a time-consuming
-- process writes to a text window. Without an occasional yield, the text window will not be updated properly, and on
-- systems with cooperative multitasking, such as Windows 3.1 other processes will not respond.
--
--Caution should be exercised, however, since yielding may allow the user to perform actions which are not compatible
-- with the current task. Disabling menu items or whole menus during processing can avoid unwanted reentrance of code.
--
--Note that Yield() will not flush the message logs. This is intentional as calling Yield() is usually done too quickly
-- update the screen and popping up a message box dialog may be undesirable.
--
--Calling Yield() recursively is normally an error and an assert failure is raised in debug build if such situation is
-- detected. However if the /i onlyIfNeeded parameter is true, the method will just silently return false instead.
public function app_yield( integer onlyIfNeeded )
    return c_func( WX_APP_YIELD, {onlyIfNeeded})
end function



--wxAuiManager
-----------------------------------------
--/topic wxAuiManager
--/parent Advanced User Interface
--/parent Classes
--/info
--wxAuiManager is the central class of the wxAUI class framework.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li atom managed_wnd [= 0]: specifies the wxFrame which should be managed
--/li atom flags [= wxAUI_MGR_DEFAULT]: specifies options which allow the frame management behavior to be modified
--<ul>
--/li wxAUI_MGR_ALLOW_FLOATING
--/li wxAUI_MGR_ALLOW_ACTIVE_PANE
--/li wxAUI_MGR_TRANSPARENT_DRAG
--/li wxAUI_MGR_TRANSPARENT_HINT
--/li wxAUI_MGR_VENETIAN_BLINDS_HINT
--/li wxAUI_MGR_RECTANGLE_HINT
--/li wxAUI_MGR_HINT_FADE
--/li wxAUI_MGR_NO_VENETIAN_BLINDS_FADE
--/li wxAUI_MGR_DEFAULT
--</ul>
--</ul>
--
--See also /"Advanced User Interface" overview.
--
--wxAuiManager manages the panes associated with it for a particular wxFrame, using a pane's
--wxAuiPaneInfo information to determine each pane's docking and floating behavior. wxAuiManager
--uses wxWidgets' sizer mechanism to plan the layout of each frame. It uses a replaceable dock
--art class (not yet implemented in wxEuphoria v0.12) to do all drawing, so all drawing is localized
--in one area, and may be customized depending on an application's specific needs.
--
--wxAuiManager works as follows: the programmer adds panes to the class, or makes changes to
--existing pane properties (dock position, floating state, show state, etc.). To apply these
--changes, wxAuiManager's Update() function is called. This batch processing can be used to
--avoid flicker, by modifying more than one pane at a time, and then "committing" all of the
--changes at once by calling Update().
--
--Panes can be added quite easily:
--/code
--constant text1 = /create( /wxTextCtrl, parent )
--constant text2 = /create( /wxTextCtrl, parent )
--
--/add_pane_simple( manager, text1, wxLEFT, "Pane Caption" )
--/add_pane_simple( manager, text1, wxBOTTOM, "Pane Caption" )
--/update_aui( manager )
--/endcode
--
--Later on, the positions can be modified easily. The following will float an existing pane in a
--tool window:
--/code
--/aui_float( /get_pane( manager, text1 ) )
--/endcode
--<h3>Layers, Rows and Directions, Positions</h3>
--
--Inside wxAUI, the docking layout is figured out by checking several pane parameters. Four of
--these are important for determining where a pane will end up:
--
--/b Direction: Each docked pane has a direction, Top, Bottom, Left, Right, or Center. This is fairly
--self-explanatory. The pane will be placed in the location specified by this variable.
--
--/b Position: More than one pane can be placed inside of a dock. Imagine two panes being docked on the
--left side of a window. One pane can be placed over another. In proportionally managed docks, the pane
--position indicates its sequential position, starting with zero. So, in our scenario with two panes
--docked on the left side, the top pane in the dock would have position 0, and the second one would
--occupy position 1.
--
--/b Row: A row can allow for two docks to be placed next to each other. One of the most common places for
--this to happen is in the toolbar. Multiple toolbar rows are allowed, the first row being row 0, and
--the second row 1. Rows can also be used on vertically docked panes.
--
--/b Layer: A layer is akin to an onion. Layer 0 is the very center of the managed pane. Thus, if a pane
--is in layer 0, it will be closest to the center window (also sometimes known as the "content window").
--Increasing layers "swallow up" all layers of a lower value. This can look very similar to multiple
--rows, but is different because all panes in a lower level yield to panes in higher levels.


--/topic wxAuiManager
--/proc add_pane_with_info( atom manager, atom win, atom pane_info )
--
--Tells the frame manager to start managing a child window with the specified /i pane_info
public procedure add_pane_with_info( atom manager, atom win, atom pane_info )
    c_proc( WX_ADD_PANE_WITH_INFO, {manager,win,pane_info})
end procedure

--/topic wxAuiManager
--/proc add_pane_simple( atom manager, atom win, atom direction, sequence caption )
--
--Tells the frame manager to start managing a child window.
public procedure add_pane_simple( atom manager, atom win, atom direction, sequence caption )
    c_proc( WX_ADD_PANE_SIMPLE, {manager,win,direction,caption})
end procedure

--/topic wxAuiManager
--/proc add_pane_with_pos( atom manager, atom win, atom pane_info, atom x, atom y )
--
--Just like /add_pane_with_info, but includes an option position for the new pane.
public procedure add_pane_with_pos( atom manager, atom win, atom pane_info, atom x, atom y )
    c_proc( WX_ADD_PANE_WITH_POS, {manager,win,pane_info,x,y})
end procedure

--/topic wxAuiManager
--/func detach_pane( atom manager, atom window )
--
--Tells the wxAuiManager to stop managing the pane specified by window.
-- The window, if in a floated frame, is reparented to the frame managed by wxAuiManager.
public function detach_pane( atom manager, atom window )
    return c_func( WX_DETACH_PANE, {manager,window})
end function

--/topic wxAuiManager
--/func get_managed_window( atom manager )
--
--Returns the frame currently being managed by wxAuiManager.
public function get_managed_window( atom manager )
    return c_func( WX_GET_MANAGED_WINDOW, {manager})
end function

--/topic wxAuiManager
--/func get_manager( atom window )
--
--Calling this method will return the wxAuiManager for a given window. The window parameter should specify
-- any child window or sub-child window of the frame or window managed by wxAuiManager. The window parameter
-- need not be managed by the manager itself, nor does it even need to be a child or sub-child of a managed
-- window. It must however be inside the window hierarchy underneath the managed window.
public function get_manager( atom window )
    return c_func( WX_GET_MANAGER, {window})
end function

--/topic wxAuiManager
--/func get_pane( atom manager, atom window )
--
--Used to lookup a wxAuiPaneInfo object either by window pointer or by pane name, which acts as a unique id
-- for a window pane. The returned wxAuiPaneInfo object may then be modified to change a pane's look, state
-- or position. After one or more modifications to wxAuiPaneInfo, /update_aui should be called to
-- commit the changes to the user interface. If the lookup failed (meaning the pane could not be found in
-- the manager), a call to the returned wxAuiPaneInfo's IsOk() method will return false.
public function get_pane( atom manager, atom window )
    return c_func( WX_GET_PANE, {manager,window})
end function

--/topic wxAuiManager
--/proc hide_hint( atom manager )
--
--Hides any docking hint that may be visible.
public procedure hide_hint( atom manager )
    c_proc( WX_HIDE_HINT, {manager})
end procedure

--/topic wxAuiManager
--/func insert_pane( atom manager, atom window, atom insert_location, atom insert_level )
--
--This routine is used to insert either a previously unmanaged pane window into the frame manager, or to
--insert a currently managed pane somewhere else. InsertPane will push all panes, rows, or docks aside and
--insert the window into the position specified by insert_location. Because /b insert_location can specify
--either a pane, dock row, or dock layer, the /b insert_level parameter is used to disambiguate this. The
--parameter insert_level can take a value of:
--<ul>
--/li /b wxAUI_INSERT_PANE
--/li /b wxAUI_INSERT_ROW
--/li /b wxAUI_INSERT_DOCK
--</ul>
--
public function insert_pane( atom manager, atom window, atom insert_location, atom insert_level )
    return c_func( WX_INSERT_PANE, {manager,window,insert_location,insert_level})
end function

--/topic wxAuiManager
--/proc load_pane_info( atom manager, sequence pane_part, atom pane )
--
--Similar to to /load_perspective, with the exception that it only loads information about a single pane. Used in combination with /save_pane_info.
public procedure load_pane_info( atom manager, sequence pane_part, atom pane )
    c_proc( WX_LOAD_PANE_INFO, {manager,pane_part,pane})
end procedure

--/topic wxAuiManager
--/func load_perspective( atom manager, sequence perspective, atom update )
--
--Loads a saved perspective. If update is true, /update_aui is automatically invoked, thus realizing the saved perspective on screen.
public function load_perspective( atom manager, sequence perspective, atom update )
    return c_func( WX_LOAD_PERSPECTIVE, {manager,perspective,update})
end function

--/topic wxAuiManager
--/func save_pane_info( atom manager, atom pane )
--Similar to /save_perspective, with the exception that it only saves information about a single pane. Used in combination with /load_pane_info.
public function save_pane_info( atom manager, atom pane )
    return c_func( WX_SAVE_PANE_INFO, {manager,pane})
end function

--/topic wxAuiManager
--/func save_perspective( atom manager )
--
--Saves the entire user interface layout into an encoded wxString, which can then be stored by the application (probably using wxConfig).
-- When a perspective is restored using /load_perspective, the entire user interface will return to the state it was when the perspective was saved.
public function save_perspective( atom manager )
    return c_func( WX_SAVE_PERSPECTIVE, {manager})
end function

--/topic wxAuiManager
--/proc set_art_provider( atom manager, atom art_provider )
--
--Instructs wxAuiManager to use art provider specified by parameter art_provider for all drawing calls. This allows plugable look-and-feel features.
-- The previous art provider object, if any, will be deleted by wxAuiManager.
public procedure set_art_provider( atom manager, atom art_provider )
    c_proc( WX_SET_ART_PROVIDER, {manager,art_provider})
end procedure

--/topic wxAuiManager
--/proc set_dock_size_constraint( atom manager, atom widthpct, atom heightpct )
--
--When a user creates a new dock by dragging a window into a docked position, often times the large size of the window will create a dock that is
-- unwieldly large. wxAuiManager by default limits the size of any new dock to 1/3 of the window size. For horizontal docks, this would be 1/3 of
-- the window height. For vertical docks, 1/3 of the width. Calling this function will adjust this constraint value. The numbers must be between
-- 0.0 and 1.0. For instance, calling SetDockSizeContraint with 0.5, 0.5 will cause new docks to be limited to half of the size of the entire
-- managed window.
public procedure set_dock_size_constraint( atom manager, atom widthpct, atom heightpct )
    c_proc( WX_SET_DOCK_SIZE_CONSTRAINT, {manager,widthpct,heightpct})
end procedure

--/topic wxAuiManager
--/proc set_aui_flags( atom manager, atom flags )
--
--Specifies wxAuiManager's settings flags. flags specifies options which allow the frame management behavior to be modified.
public procedure set_aui_flags( atom manager, atom flags )
    c_proc( WX_SET_AUI_FLAGS, {manager,flags})
end procedure

--/topic wxAuiManager
--/proc set_managed_window( atom manager, atom managed_wnd )
--
--Called to specify the frame or window which is to be managed by wxAuiManager. Frame management is not restricted to just frames. Child windows
-- or custom controls are also allowed.
public procedure set_managed_window( atom manager, atom managed_wnd )
    c_proc( WX_SET_MANAGED_WINDOW, {manager,managed_wnd})
end procedure

--/topic wxAuiManager
--/proc uninit_aui( atom manager )
--
--Uninitializes the framework and should be called before a managed frame or window is destroyed. /b uninit_aui is usually called in the managed
-- /wxFrame's destructor. It is necessary to call this function before the managed frame or window is destroyed, otherwise the manager cannot
-- remove its custom event handlers from a window.
public procedure uninit_aui( atom manager )
    c_proc( WX_UNINIT_AUI, {manager})
end procedure

--/topic wxAuiManager
--/proc update_aui( atom manager )
--
--This method is called after any number of changes are made to any of the managed panes.
--/update_aui() must be invoked after adding or inserting a pane ( /add_pane_simple(), /add_pane_with_info(),
--/add_pane_with_pos, /insert_pane() )are called in order to "realize" or "commit" the changes. In
--addition, any number of changes may be made to wxAuiPaneInfo structures (retrieved with
--/get_pane() ), but to realize the changes, update_aui() must be called. This construction allows pane
--flicker to be avoided by updating the whole layout at one time.
public procedure update_aui( atom manager )
    c_proc( WX_UPDATE_AUI, {manager})
end procedure



--wxAuiNotebook
-----------------------------------------
--/topic wxAuiNotebook
--/parent Advanced User Interface
--/parent Controls
--/info
--wxAuiNotebook is part of the wxAUI class framework. See also the /"Advanced User Interface" overview.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cy [=-1]
--/li integer style [= wxAUI_NB_DEFAULT_STYLE]
--<ul><table>
--<tr><td> /li wxAUI_NB_TAB_SPLIT <td>Allows the tab control to be split by dragging a tab.
--<tr><td> /li wxAUI_NB_TAB_MOVE <td>Allows a tab to be moved horizontally by dragging.
--<tr><td> /li wxAUI_NB_TAB_EXTERNAL_MOVE <td>Allows a tab to be moved to another tab control.
--<tr><td> /li wxAUI_NB_TAB_FIXED_WIDTH <td>With this style, all tabs have the same width.
--<tr><td> /li wxAUI_NB_SCROLL_BUTTONS <td>With this style, left and right scroll buttons are displayed.
--<tr><td> /li wxAUI_NB_WINDOWLIST_BUTTON <td>With this style, a drop-down list of windows is available.
--<tr><td> /li wxAUI_NB_CLOSE_BUTTON <td>With this style, a close button is available on the tab bar.
--<tr><td> /li wxAUI_NB_CLOSE_ON_ACTIVE_TAB <td>With this style, the close button is visible on the active tab.
--<tr><td> /li wxAUI_NB_CLOSE_ON_ALL_TABS <td>With this style, the close button is visible on all tabs.
--<tr><td valign="top"> /li wxAUI_NB_DEFAULT_STYLE <td>Defined as: wxAUI_NB_TOP | wxAUI_NB_TAB_SPLIT | wxAUI_NB_TAB_MOVE | wxAUI_NB_SCROLL_BUTTONS | wxAUI_NB_CLOSE_ON_ACTIVE_TAB .
--</table></ul>
--</ul>
--
--wxAuiNotebook is a notebook control which implements many features common in applications with
--dockable panes. Specifically, wxAuiNotebook implements functionality which allows the user to
--rearrange tab order via drag-and-drop, split the tab window into many different splitter
--configurations, and toggle through different themes to customize the control's look and feel
--(not fully implemented in wxEuphoria v0.12)
--
--An effort has been made to try to maintain an API as similar to that of wxNotebook.


--/topic wxAuiNotebook
--/func aui_notebook_add_page( atom nb, atom page, sequence caption, atom select, atom bitmap )
--
--Adds a page. If the select parameter is true, calling this will generate a page change event.
public function aui_notebook_add_page( atom nb, atom page, sequence caption, atom select, atom bitmap )
    return c_func( WX_AUI_NOTEBOOK_ADD_PAGE, {nb,page,caption,select,bitmap})
end function

--/topic wxAuiNotebook
--/func aui_notebook_delete_page( atom nb, atom page )
--
--Deletes a page at the given index. Calling this method will generate a page change event.
public function aui_notebook_delete_page( atom nb, atom page )
    return c_func( WX_AUI_NOTEBOOK_DELETE_PAGE, {nb,page})
end function

--/topic wxAuiNotebook
--/func aui_notebook_get_art_provider( atom nb )
--
--Returns the associated art provider.
public function aui_notebook_get_art_provider( atom nb )
    return c_func( WX_AUI_NOTEBOOK_GET_ART_PROVIDER, {nb})
end function

--/topic wxAuiNotebook
--/func aui_notebook_get_page( atom nb, atom page_idx )
--
--Returns the page specified by the given index.
public function aui_notebook_get_page( atom nb, atom page_idx )
    return c_func( WX_AUI_NOTEBOOK_GET_PAGE, {nb,page_idx})
end function

--/topic wxAuiNotebook
--/func aui_notebook_get_page_bitmap( atom nb, atom page )
--
--Returns the tab bitmap for the page.
public function aui_notebook_get_page_bitmap( atom nb, atom page )
    return c_func( WX_AUI_NOTEBOOK_GET_PAGE_BITMAP, {nb,page})
end function

--/topic wxAuiNotebook
--/func aui_notebook_get_page_count( atom nb )
--
--Returns the number of pages in the notebook.
public function aui_notebook_get_page_count( atom nb )
    return c_func( WX_AUI_NOTEBOOK_GET_PAGE_COUNT, {nb})
end function

--/topic wxAuiNotebook
--/func aui_notebook_get_page_index( atom nb, atom page_wnd )
--
--Returns the page index for the specified window. If the window is not found in the notebook, wxNOT_FOUND is returned.
public function aui_notebook_get_page_index( atom nb, atom page_wnd )
    return c_func( WX_AUI_NOTEBOOK_GET_PAGE_INDEX, {nb,page_wnd})
end function

--/topic wxAuiNotebook
--/func aui_get_selection( atom nb )
--
--Returns the currently selected page.
public function aui_get_selection( atom nb )
    return c_func( WX_AUI_GET_SELECTION, {nb})
end function

--/topic wxAuiNotebook
--/func aui_notebook_insert_page( atom nb, atom page_idx, atom page, sequence caption, atom select, atom bitmap )
--
--Similar to /aui_notebook_add_page, but allows the ability to specify the insert location. If the /i select parameter is true,
-- calling this will generate a page change event.
public function aui_notebook_insert_page( atom nb, atom page_idx, atom page, sequence caption, atom select, atom bitmap )
    return c_func( WX_AUI_NOTEBOOK_INSERT_PAGE, {nb,page_idx,page,caption,select,bitmap})
end function

--/topic wxAuiNotebook
--/func aui_notebook_remove_page( atom nb, atom page )
--
--Removes a page, without deleting the window pointer.
public function aui_notebook_remove_page( atom nb, atom page )
    return c_func( WX_AUI_NOTEBOOK_REMOVE_PAGE, {nb,page})
end function

--/topic wxAuiNotebook
--/proc aui_notebook_set_art_provider( atom nb, atom art )
--
--Sets the art provider to be used by the notebook.
public procedure aui_notebook_set_art_provider( atom nb, atom art )
    c_proc( WX_AUI_NOTEBOOK_SET_ART_PROVIDER, {nb,art})
end procedure

--/topic wxAuiNotebook
--/func aui_notebook_set_page_bitmap( atom nb, atom page, atom bitmap )
--
--Sets the bitmap for the page. To remove a bitmap from the tab caption, pass wxNullBitmap.
public function aui_notebook_set_page_bitmap( atom nb, atom page, atom bitmap )
    return c_func( WX_AUI_NOTEBOOK_SET_PAGE_BITMAP, {nb,page,bitmap})
end function

--/topic wxAuiNotebook
--/func aui_notebook_set_page_text( atom nb, atom page, sequence text )
--
--Sets the tab label for the page.
public function aui_notebook_set_page_text( atom nb, atom page, sequence text )
    return c_func( WX_AUI_NOTEBOOK_SET_PAGE_TEXT, {nb,page,text})
end function

--/topic wxAuiNotebook
--/func aui_notebook_set_selection( atom nb, atom new_page )
--
--Sets the page selection. Calling this method will generate a page change event.
public function aui_notebook_set_selection( atom nb, atom new_page )
    return c_func( WX_AUI_NOTEBOOK_SET_SELECTION, {nb,new_page})
end function

--/topic wxAuiNotebook
--/proc aui_notebook_set_tab_ctrl_height( atom nb, atom height )
--
--Sets the tab height. By default, the tab control height is calculated by measuring the text height and bitmap sizes on the tab captions.
-- Calling this method will override that calculation and set the tab control to the specified height parameter. A call to this method will
-- override any call to /aui_notebook_set_uniform_bitmap_size. Specifying -1 as the height will return the control to its default auto-sizing
-- behaviour.
public procedure aui_notebook_set_tab_ctrl_height( atom nb, atom height )
    c_proc( WX_AUI_NOTEBOOK_SET_TAB_CTRL_HEIGHT, {nb,height})
end procedure

--/topic wxAuiNotebook
--/proc aui_notebook_set_uniform_bitmap_size( atom nb, atom x, atom y )
--
--Ensures that all tabs will have the same height, even if some tabs don't have bitmaps. Passing wxDefaultSize to this function will instruct
-- the control to use dynamic tab height, which is the default behaviour. Under the default behaviour, when a tab with a large bitmap is added,
-- the tab control's height will automatically increase to accommodate the larger bitmap.
public procedure aui_notebook_set_uniform_bitmap_size( atom nb, atom x, atom y )
    c_proc( WX_AUI_NOTEBOOK_SET_UNIFORM_BITMAP_SIZE, {nb,x,y})
end procedure

--/topic wxAuiNotebook
--/proc aui_notebook_split( atom nb, atom page, atom direction )
--
--Split performs a split operation programmatically. The argument page indicates the page that will be split off. This page will also become
-- the active page after the split. The direction argument specifies where the pane should go, it should be one of the following: wxTOP,
-- wxBOTTOM, wxLEFT, or wxRIGHT.
public procedure aui_notebook_split( atom nb, atom page, atom direction )
    c_proc( WX_AUI_NOTEBOOK_SPLIT, {nb,page,direction})
end procedure



--wxAuiPaneInfo
-----------------------------------------
--/topic wxAuiPaneInfo
--/parent Advanced User Interface
--/parent Classes
--/info
--wxAuiPaneInfo is part of the wxAUI class framework. See also the /"Advanced User Interface" overview.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b None
--</ul>
--
--wxAuiPaneInfo specifies all the parameters for a pane. These parameters specify where the pane is on
--the screen, whether it is docked or floating, or hidden. In addition, these parameters specify the
--pane's docked position, floating position, preferred size, minimum size, caption text among many
--other parameters.

--/topic wxAuiPaneInfo
--/proc best_size_pane( atom info, atom x, atom y )
--
--Sets the ideal size for the pane. The docking manager will attempt to use this size as much as possible when docking or floating the pane.
public procedure best_size_pane( atom info, atom x, atom y )
    c_proc( WX_BEST_SIZE_PANE, {info,x,y})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_best_size( atom info, atom x, atom y )
--
--See /best_size_pane.
public procedure aui_best_size( atom info, atom x, atom y )
    c_proc( WX_AUI_BEST_SIZE, {info,x,y})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_bottom( atom info )
--
--Sets the pane dock position to the bottom side of the frame. This is the same thing as calling /aui_direction( wxAUI_DOCK_BOTTOM ).
public procedure aui_bottom( atom info )
    c_proc( WX_AUI_BOTTOM, {info})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_bottom_dockable( atom info, atom b )
--
--Indicates whether a pane can be docked at the bottom of the frame.
public procedure aui_bottom_dockable( atom info, atom b )
    c_proc( WX_AUI_BOTTOM_DOCKABLE, {info,b})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_caption( atom info, sequence c )
--
--Sets the caption of the pane.
public procedure aui_caption( atom info, sequence c )
    c_proc( WX_AUI_CAPTION, {info,c})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_caption_visible( atom info, atom visible )
--
--Indicates that a pane caption should be visible. If false, no pane caption is drawn.
public procedure aui_caption_visible( atom info, atom visible )
    c_proc( WX_AUI_CAPTION_VISIBLE, {info,visible})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_center( atom info )
--
--Sets the pane dock position to the left side of the frame. The centre pane is the space in the middle after all border panes
-- (left, top, right, bottom) are subtracted from the layout.
--
--This is the same thing as calling /aui_direction( wxAUI_DOCK_CENTRE ).
public procedure aui_center( atom info )
    c_proc( WX_AUI_CENTER, {info})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_center_pane( atom info )
--
--pecifies that the pane should adopt the default center pane settings. Centre panes usually do not have caption bars.
-- This function provides an easy way of preparing a pane to be displayed in the center dock position.
public procedure aui_center_pane( atom info )
    c_proc( WX_AUI_CENTER_PANE, {info})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_close_button( atom info, atom visible )
--
--Indicates that a close button should be drawn for the pane.
public procedure aui_close_button( atom info, atom visible )
    c_proc( WX_AUI_CLOSE_BUTTON, {info,visible})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_default_pane( atom info )
--
--Specifies that the pane should adopt the default pane settings.
public procedure aui_default_pane( atom info )
    c_proc( WX_AUI_DEFAULT_PANE, {info})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_destroy_on_close( atom info, atom b )
--
--Indicates whether a pane should be detroyed when it is closed. Normally a pane is simply hidden when the close button is clicked.
-- Setting /i b to true will cause the window to be destroyed when the user clicks the pane's close button.
public procedure aui_destroy_on_close( atom info, atom b )
    c_proc( WX_AUI_DESTROY_ON_CLOSE, {info,b})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_direction( atom info, atom direction )
--
--Determines the direction of the docked pane. It is functionally the same as calling /aui_left, /aui_right(), /aui_top or /aui_bottom,
-- except that docking direction may be specified programmatically via the parameter.
public procedure aui_direction( atom info, atom direction )
    c_proc( WX_AUI_DIRECTION, {info,direction})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_dock( atom info )
--
--Indicates that a pane should be docked. It is the opposite of /aui_float.
public procedure aui_dock( atom info )
    c_proc( WX_AUI_DOCK, {info})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_dockable( atom info, atom b )
--
--Specifies whether a frame can be docked or not. It is the same as specifying /aui_top_dockable, /aui_bottom_dockable,
-- /aui_left_dockable, /aui_right_dockable.
public procedure aui_dockable( atom info, atom b )
    c_proc( WX_AUI_DOCKABLE, {info,b})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_fixed( atom info )
--
--Forces a pane to be fixed size so that it cannot be resized. After calling /b aui_fixed, /aui_is_fixed will return true.
public procedure aui_fixed( atom info )
    c_proc( WX_AUI_FIXED, {info})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_float( atom info )
--
--Indicates that a pane should be floated. It is the opposite of /aui_dock.
public procedure aui_float( atom info )
    c_proc( WX_AUI_FLOAT, {info})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_floatable( atom info, atom b )
--
--Sets whether the user will be able to undock a pane and turn it into a floating window.
public procedure aui_floatable( atom info, atom b )
    c_proc( WX_AUI_FLOATABLE, {info,b})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_floating_position( atom info, atom x, atom y )
--
--Sets the position of the floating pane.
public procedure aui_floating_position( atom info, atom x, atom y )
    c_proc( WX_AUI_FLOATING_POSITION, {info,x,y})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_floating_size( atom info, atom x, atom y )
--
--Sets the size of the floating pane.
public procedure aui_floating_size( atom info, atom x, atom y )
    c_proc( WX_AUI_FLOATING_SIZE, {info,x,y})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_gripper( atom info, atom visible )
--
--Indicates that a gripper should be drawn for the pane.
public procedure aui_gripper( atom info, atom visible )
    c_proc( WX_AUI_GRIPPER, {info,visible})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_gripper_top( atom info, atom attop )
--
--Indicates that a gripper should be drawn at the top of the pane.
public procedure aui_gripper_top( atom info, atom attop )
    c_proc( WX_AUI_GRIPPER_TOP, {info,attop})
end procedure

--/topic wxAuiPaneInfo
--/func aui_has_border( atom info )
--
--Returns true if the pane displays a border.
public function aui_has_border( atom info )
    return c_func( WX_AUI_HAS_BORDER, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_has_caption( atom info )
--
--Returns true if the pane displays a caption.
public function aui_has_caption( atom info )
    return c_func( WX_AUI_HAS_CAPTION, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_has_close_button( atom info )
--
--Returns true if the pane displays a button to close the pane.
public function aui_has_close_button( atom info )
    return c_func( WX_AUI_HAS_CLOSE_BUTTON, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_has_flag( atom info, atom flag )
--
--Returns true if the the property specified by flag is active for the pane.
public function aui_has_flag( atom info, atom flag )
    return c_func( WX_AUI_HAS_FLAG, {info,flag})
end function

--/topic wxAuiPaneInfo
--/func aui_has_gripper( atom info )
--
--Returns true if the pane displays a gripper.
public function aui_has_gripper( atom info )
    return c_func( WX_AUI_HAS_GRIPPER, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_has_gripper_top( atom info )
--
--Returns true if the pane displays a gripper at the top.
public function aui_has_gripper_top( atom info )
    return c_func( WX_AUI_HAS_GRIPPER_TOP, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_has_maximize_button( atom info )
--
--Returns true if the pane displays a button to maximize the pane.
public function aui_has_maximize_button( atom info )
    return c_func( WX_AUI_HAS_MAXIMIZE_BUTTON, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_has_minimize_button( atom info )
--
--Returns true if the pane displays a button to minimize the pane.
public function aui_has_minimize_button( atom info )
    return c_func( WX_AUI_HAS_MINIMIZE_BUTTON, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_has_pin_button( atom info )
--
--Returns true if the pane displays a button to float the pane.
public function aui_has_pin_button( atom info )
    return c_func( WX_AUI_HAS_PIN_BUTTON, {info})
end function

--/topic wxAuiPaneInfo
--/proc aui_hide( atom info )
--
--Indicates that a pane should be hidden.
public procedure aui_hide( atom info )
    c_proc( WX_AUI_HIDE, {info})
end procedure

--/topic wxAuiPaneInfo
--/func aui_is_bottom_dockable( atom info )
--
--Returns true if the pane can be docked at the bottom of the managed frame.
public function aui_is_bottom_dockable( atom info )
    return c_func( WX_AUI_IS_BOTTOM_DOCKABLE, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_docked( atom info )
--
--Returns true if the pane is docked.
public function aui_is_docked( atom info )
    return c_func( WX_AUI_IS_DOCKED, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_fixed( atom info )
--
--Returns true if the pane cannot be resized.
public function aui_is_fixed( atom info )
    return c_func( WX_AUI_IS_FIXED, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_floatable( atom info )
--
--Returns true if the pane can be undocked and displayed as a floating window.
public function aui_is_floatable( atom info )
    return c_func( WX_AUI_IS_FLOATABLE, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_floating( atom info )
--
--Returns true if the pane is floating.
public function aui_is_floating( atom info )
    return c_func( WX_AUI_IS_FLOATING, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_left_dockable( atom info )
--
--Returns true if the pane can be docked on the left of the managed frame.
public function aui_is_left_dockable( atom info )
    return c_func( WX_AUI_IS_LEFT_DOCKABLE, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_movable( atom info )
--
--Returns true if the docked frame can be undocked or moved to another dock position.
public function aui_is_movable( atom info )
    return c_func( WX_AUI_IS_MOVABLE, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_ok( atom info )
--
--Returns true if the wxAuiPaneInfo structure is valid. A pane structure is valid if it has an associated window.
public function aui_is_ok( atom info )
    return c_func( WX_AUI_IS_OK, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_resizable( atom info )
--
--Returns true if the pane can be resized.
public function aui_is_resizable( atom info )
    return c_func( WX_AUI_IS_RESIZABLE, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_right_dockable( atom info )
--
--Returns true if the pane can be docked on the right of the managed frame.
public function aui_is_right_dockable( atom info )
    return c_func( WX_AUI_IS_RIGHT_DOCKABLE, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_shown( atom info )
--
--Returns true if the pane is currently shown.
public function aui_is_shown( atom info )
    return c_func( WX_AUI_IS_SHOWN, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_toolbar( atom info )
--
--Returns true if the pane contains a toolbar.
public function aui_is_toolbar( atom info )
    return c_func( WX_AUI_IS_TOOLBAR, {info})
end function

--/topic wxAuiPaneInfo
--/func aui_is_top_dockable( atom info )
--
--Returns true if the pane can be docked at the top of the managed frame.
public function aui_is_top_dockable( atom info )
    return c_func( WX_AUI_IS_TOP_DOCKABLE, {info})
end function

--/topic wxAuiPaneInfo
--/proc aui_layer( atom info, atom layer )
--
--Determines the layer of the docked pane. The dock layer is similar to an onion, the inner-most layer being layer 0.
-- Each shell moving in the outward direction has a higher layer number. This allows for more complex docking layout formation.
public procedure aui_layer( atom info, atom layer )
    c_proc( WX_AUI_LAYER, {info,layer})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_left( atom info )
--
--Sets the pane dock position to the left side of the frame. This is the same thing as calling /aui_direction( wxAUI_DOCK_LEFT ).
public procedure aui_left( atom info )
    c_proc( WX_AUI_LEFT, {info})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_left_dockable( atom info, atom b )
--
--Indicates whether a pane can be docked on the left of the frame.
public procedure aui_left_dockable( atom info, atom b )
    c_proc( WX_AUI_LEFT_DOCKABLE, {info,b})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_max_size( atom info, atom x, atom y )
--
--Sets the maximum size of the pane.
public procedure aui_max_size( atom info, atom x, atom y )
    c_proc( WX_AUI_MAX_SIZE, {info,x,y})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_maximize_button( atom info, atom visible )
--
--Indicates that a maximize button should be drawn for the pane.
public procedure aui_maximize_button( atom info, atom visible )
    c_proc( WX_AUI_MAXIMIZE_BUTTON, {info,visible})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_min_size( atom info, atom x, atom y )
--
--Sets the minimum size of the pane. Please note that this is only partially supported as of this writing.
public procedure aui_min_size( atom info, atom x, atom y )
    c_proc( WX_AUI_MIN_SIZE, {info,x,y})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_minimize_button( atom info, atom visible )
--
--Indicates that a minimize button should be drawn for the pane.
public procedure aui_minimize_button( atom info, atom visible )
    c_proc( WX_AUI_MINIMIZE_BUTTON, {info,visible})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_movable( atom info, atom b )
--
--Indicates whether a frame can be moved.
public procedure aui_movable( atom info, atom b )
    c_proc( WX_AUI_MOVABLE, {info,b})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_name( atom info, sequence n )
--
--Sets the name of the pane so it can be referenced in lookup functions. If a name is not specified by the user,
-- a random name is assigned to the pane when it is added to the manager.
public procedure aui_name( atom info, sequence n )
    c_proc( WX_AUI_NAME, {info,n})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_pane_border( atom info, atom visible )
--
--Indicates that a border should be drawn for the pane.
public procedure aui_pane_border( atom info, atom visible )
    c_proc( WX_AUI_PANE_BORDER, {info,visible})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_pin_button( atom info, atom visible )
--
--Indicates that a pin button should be drawn for the pane.
public procedure aui_pin_button( atom info, atom visible )
    c_proc( WX_AUI_PIN_BUTTON, {info,visible})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_position( atom info, atom pos )
--
--Determines the position of the docked pane.
public procedure aui_position( atom info, atom pos )
    c_proc( WX_AUI_POSITION, {info,pos})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_resizable( atom info, atom resizable )
--
--Allows a pane to be resized if the parameter is true, and forces it to be a fixed size if the parameter is false.
-- This is simply an antonym for /aui_fixed.
public procedure aui_resizable( atom info, atom resizable )
    c_proc( WX_AUI_RESIZABLE, {info,resizable})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_right( atom info )
--
--Sets the pane dock position to the right side of the frame.
public procedure aui_right( atom info )
    c_proc( WX_AUI_RIGHT, {info})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_right_dockable( atom info, atom b )
--
--Indicates whether a pane can be docked on the right of the frame.
public procedure aui_right_dockable( atom info, atom b )
    c_proc( WX_AUI_RIGHT_DOCKABLE, {info,b})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_row( atom info, atom row )
--
--Determines the row of the docked pane.
public procedure aui_row( atom info, atom row )
    c_proc( WX_AUI_ROW, {info,row})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_safe_set( atom info, atom source )
--
--Write the safe parts of a newly loaded PaneInfo structure /i source into /i info (used on loading perspectives etc).
public procedure aui_safe_set( atom info, atom source )
    c_proc( WX_AUI_SAFE_SET, {info,source})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_set_flag( atom info, atom flag, atom option_state )
--
--Turns the property given by flag on or off with the option_state parameter.
public procedure aui_set_flag( atom info, atom flag, atom option_state )
    c_proc( WX_AUI_SET_FLAG, {info,flag,option_state})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_show( atom info, atom show )
--
--Indicates that a pane should be shown.
public procedure aui_show( atom info, atom show )
    c_proc( WX_AUI_SHOW, {info,show})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_toolbar_pane( atom info )
--
--Specifies that the pane should adopt the default toolbar pane settings.
public procedure aui_toolbar_pane( atom info )
    c_proc( WX_AUI_TOOLBAR_PANE, {info})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_top( atom info )
--
--Sets the pane dock position to the top of the frame.
public procedure aui_top( atom info )
    c_proc( WX_AUI_TOP, {info})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_top_dockable( atom info, atom b )
--
--Indicates whether a pane can be docked at the top of the frame.
public procedure aui_top_dockable( atom info, atom b )
    c_proc( WX_AUI_TOP_DOCKABLE, {info,b})
end procedure

--/topic wxAuiPaneInfo
--/proc aui_window( atom info, atom w )
--
--Assigns the window pointer that the wxAuiPaneInfo should use. This normally does not need to be specified, as the window pointer
-- is automatically assigned to the wxAuiPaneInfo structure as soon as it is added to the manager.
public procedure aui_window( atom info, atom w )
    c_proc( WX_AUI_WINDOW, {info,w})
end procedure



--wxBitmap
------------------------------
--/topic wxBitmap
--/parent Graphics
--/parent Classes
--/info
--The first creation parameter should define the method of creation.
--<ul>
--/li /b BM_IN_MEMORY  This creates a bitmap in memory, similar to a Pixmap control
--in win32lib.  This form takes three additional parameters:
--<ul>
--/li /b width
--/li /b height
--/li depth [= -1] /i "(for the display depth of the screen)"
--</ul>
--/li /b BM_XPM
--/li /b BM_FROM_FILE  This loads a bitmap from a file or resource.  The second parameter should be the file or
--resource name, and the final parameter should be the type of file:
--<ul>
--/li    wxBITMAP_TYPE_INVALID
--/li    wxBITMAP_TYPE_BMP
--/li    wxBITMAP_TYPE_BMP_RESOURCE
--/li    wxBITMAP_TYPE_RESOURCE = wxBITMAP_TYPE_BMP_RESOURCE
--/li    wxBITMAP_TYPE_ICO
--/li    wxBITMAP_TYPE_ICO_RESOURCE
--/li    wxBITMAP_TYPE_CUR
--/li    wxBITMAP_TYPE_CUR_RESOURCE
--/li    wxBITMAP_TYPE_XBM
--/li    wxBITMAP_TYPE_XBM_DATA
--/li    wxBITMAP_TYPE_XPM
--/li    wxBITMAP_TYPE_XPM_DATA
--/li    wxBITMAP_TYPE_TIF
--/li    wxBITMAP_TYPE_TIF_RESOURCE
--/li    wxBITMAP_TYPE_GIF
--/li    wxBITMAP_TYPE_GIF_RESOURCE
--/li    wxBITMAP_TYPE_PNG
--/li    wxBITMAP_TYPE_PNG_RESOURCE
--/li    wxBITMAP_TYPE_JPEG
--/li    wxBITMAP_TYPE_JPEG_RESOURCE
--/li    wxBITMAP_TYPE_PNM
--/li    wxBITMAP_TYPE_PNM_RESOURCE
--/li    wxBITMAP_TYPE_PCX
--/li    wxBITMAP_TYPE_PCX_RESOURCE
--/li    wxBITMAP_TYPE_PICT
--/li    wxBITMAP_TYPE_PICT_RESOURCE
--/li    wxBITMAP_TYPE_ICON
--/li    wxBITMAP_TYPE_ICON_RESOURCE
--/li    wxBITMAP_TYPE_MACCURSOR
--/li    wxBITMAP_TYPE_MACCURSOR_RESOURCE
--/li    wxBITMAP_TYPE_ANY
--</ul>
--/li /b BM_FROM_IMAGE  This creates a new /wxBitmap object from the contents of the provided /wxImage object.
--<ul>
--/li /b /wxImage
--/li depth [= -1] /i "(for the display depth of the screen)"
--</ul>
--</ul>

--/topic wxBitmap
--/proc set_bitmap_size( atom bitmap, integer width, integer height )
--
--Sets the width and height of the /i bitmap.
public procedure set_bitmap_size( atom bm, integer cx, integer cy )
    c_proc( WX_SET_BITMAP_SIZE, {bm, cx, cy} )
end procedure

--/topic wxBitmap
--/func get_bitmap_size( atom bitmap )
--
--Returns sequence containing the width and height of the bitmap, respectively.
public function get_bitmap_size( atom bitmap )
    return c_func( WX_GET_BITMAP_SIZE, {bitmap} )
end function

--/topic wxBitmap
--/proc set_mask( atom bmp, atom mask )
--
--Sets the specified /i mask to be the associated /wxMask for the specified /i bmp.
public procedure set_mask( atom bmp, atom mask )
    c_proc( WX_SET_MASK, {bmp, mask} )
end procedure

--/topic wxBitmap
--/proc draw_bitmap2( object bitmap, atom dest, atom xsrc, atom ysrc, atom xdest,   atom ydest, atom width, atom height, atom logical_func, integer use_mask )
--
public procedure draw_bitmap2( object bitmap, atom dest, atom xsrc, atom ysrc, atom xdest,      atom ydest, atom width, atom height, atom logical_func, integer use_mask )
    c_proc( WX_DRAW_BITMAP2, {bitmap, dest, xsrc, ysrc, xdest, ydest, width, height, logical_func, use_mask} )
end procedure

--/topic wxBitmap
--/func convert_to_image( atom bitmap )
--
--Takes a /wxBitmap and returns a wxImage.
public function convert_to_image( atom bitmap )
    return c_func( WX_CONVERT_TO_IMAGE, {bitmap} )
end function

--/topic wxBitmap
--/func get_palette( atom bmp )
--Gets the associated palette (if any) which may have been loaded from a file or set for the bitmap.
public function get_palette( atom bmp )
    return c_func( WX_GET_PALETTE, {bmp})
end function

--/topic wxBitmap
--/func save_bitmap( atom bitmap, object name, atom type, atom palette )
--/b Returns: /i true if the operation succeeded, /i false otherwise.
--Saves a bitmap in the named file.
--
--/b Parameters
--<ul>
--/li /b bitmap: The image to save
--/li /b name: A filename. The meaning of /i name is determined by the type parameter.
--/li /b type: One of the following values:
--<ul>
--/li /b wxBITMAP_TYPE_BMP  Save a Windows bitmap file.
--/li /b wxBITMAP_TYPE_GIF  Save a GIF bitmap file.
--/li /b wxBITMAP_TYPE_XBM  Save an X bitmap file.
--/li /b wxBITMAP_TYPE_XPM  Save an XPM bitmap file.
--</ul>
--The validity of these flags depends on the platform and wxWidgets configuration.
--
--In addition, /wxBitmap can save all formats that /wxImage can ( /b wxBITMAP_TYPE_JPEG, /b wxBITMAP_TYPE_PNG ).
-- (Of course you must have /wxImage handlers loaded.)
--/li /b palette: An optional palette used for saving the bitmap.
--</ul>
--
public function save_bitmap( atom bitmap, object name, atom type_, atom palette )
    return c_func( WX_SAVE_BITMAP, {bitmap,name,type_,palette})
end function



--wxBitmapButton
------------------------------
--/topic wxBitmapButton
--/parent wxButton
--/parent Controls
--/info
--A bitmap button is a control that contains a bitmap.
-- It may be placed on a dialog box or panel, or indeed almost any other window.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li atom /wxBitmap [= 0]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= -1]
--</ul>

--/topic wxBitmapButton
--/proc set_bmpbutton( atom button, atom bitmap )
--
--Sets the displayed bitmap of the bitmap button.
public procedure set_bmpbutton( atom button, atom bitmap )
    c_proc( WX_SET_BMPBUTTON, {button, bitmap} )
end procedure

--/topic wxBitmapButton
--/proc set_bitmap_disabled( atom button, atom bitmap )
--
--Sets the bitmap for the disabled button appearance.
public procedure set_bitmap_disabled( atom button, atom bitmap )
    c_proc( WX_SET_BITMAP_DISABLED, {button,bitmap})
end procedure

--/topic wxBitmapButton
--/proc set_bitmap_focus( atom button, atom bitmap )
--
--Sets the bitmap for the button appearance when it has the keyboard focus.
public procedure set_bitmap_focus( atom button, atom bitmap )
    c_proc( WX_SET_BITMAP_FOCUS, {button,bitmap})
end procedure

--/topic wxBitmapButton
--/proc set_bitmap_hover( atom button, atom bitmap )
--Sets the bitmap to be shown when the mouse is over the button.
--
--This function is new since wxWidgets version 2.7.0 and the hover bitmap is currently only supported in wxMSW.
public procedure set_bitmap_hover( atom button, atom bitmap )
    c_proc( WX_SET_BITMAP_HOVER, {button,bitmap})
end procedure

--/topic wxBitmapButton
--/proc set_bitmap_label( atom button, atom bitmap )
--
--Sets the bitmap label for the button.
public procedure set_bitmap_label( atom button, atom bitmap )
    c_proc( WX_SET_BITMAP_LABEL, {button,bitmap})
end procedure

--/topic wxBitmapButton
--/proc set_bitmap_selected( atom button, atom bitmap )
--
--Sets the bitmap for the selected (depressed) button appearance.
public procedure set_bitmap_selected( atom button, atom bitmap )
    c_proc( WX_SET_BITMAP_SELECTED, {button,bitmap})
end procedure



--wxBitmapComboBox
-----------------------------------------
--/topic wxBitmapComboBox
--/parent List Controls
--/parent wxComboBox
--/parent Controls
--/info
--A combobox that displays bitmap in front of the list items.
-- It currently only allows using bitmaps of one size, and resizes itself
-- so that a bitmap can be shown next to the text field.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence value [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li sequence choices [= {}]
--/li integer style [= 0]
--</ul>
--
--See /wxComboBox for additional functions that are available.
--

--/topic wxBitmapComboBox
--/func get_combo_bitmap_size( atom combo )
--Returns size of bitmaps used in the list.
public function get_combo_bitmap_size( atom combo )
    return c_func( WX_GET_COMBO_BITMAP_SIZE, {combo})
end function

--/topic wxBitmapComboBox
--/func get_combo_item_bitmap( atom combo, atom index )
--Returns the bitmap of the item with the given index.
public function get_combo_item_bitmap( atom combo, atom index )
    return c_func( WX_GET_COMBO_ITEM_BITMAP, {combo,index})
end function

--/topic wxBitmapComboBox
--/proc set_combo_item_bitmap( atom combo, atom index, atom bitmap )
--Sets the bitmap for the given item.
public procedure set_combo_item_bitmap( atom combo, atom index, atom bitmap )
    c_proc( WX_SET_COMBO_ITEM_BITMAP, {combo,index,bitmap})
end procedure



--Notebook Controls
-----------------------------------------
--/topic Notebook Controls
--/parent Controls
--/info
--Controls with a multi-page interface

--/topic Notebook Controls
--/proc add_page( atom book, atom page, sequence text, integer select, atom image )
--
--Adds the /wxWindow /i page to the notebook, with /i text on the its tab.  /i select
--indicates whether the page should be selected, and /i image is the image index into
--an imagelist.
public procedure add_page( atom book, atom page, sequence text, integer select, atom image )
    c_proc( WX_ADD_PAGE, {book, page, text, select, image} )
end procedure

--/topic Notebook Controls
--/proc show_page( atom book, integer page )
--
--Sets the wxNotebook to show a page specified by integer page. The first page is 1 instead of being 0.
public procedure show_page( atom book, integer page )
    c_proc( WX_SHOW_PAGE, {book, page} )
end procedure

--/topic Notebook Controls
--/proc set_page_text( atom book, integer page, sequence text )
--
--Sets the text of the wxNotebook's tab for that page to text. The first page is 0, not 1.
public procedure set_page_text( atom book, integer page, sequence text )
    c_proc( WX_SET_PAGE_TEXT, {book, page, text} )
end procedure

--/topic Notebook Controls
--/proc notebook_imagelist( atom book, atom ilist )
--
--This sets the wxImageList to the wxNotebook, allowing the wxNotebook's children to use it's icons.
public procedure notebook_imagelist( atom book, atom ilist )
    c_proc( WX_NOTEBOOK_IMAGELIST, {book, ilist} )
end procedure

--/topic Notebook Controls
--/proc delete_all_pages ( atom book )
--
--This deletes all pages in a wxNotebook.
public procedure delete_all_pages ( atom book )
    c_proc( WX_DELETE_ALL_PAGES, {book} )
end procedure

--/topic Notebook Controls
--/proc insert_page ( integer position, atom book, atom page, sequence text, integer select, atom image )
--
--Sets the selected page to the wxNotebook at a certain position, rather than adding it to the end.
--The sequence text will appear on it's tab, select can be 1 to make the wxNotebook focus on the
--selected page or 0 to continue working with the same page. Image will be a pointer to an image if
--an imagelist exists, or 0 for no image. This works with a 0-based index, so position should be 0
--instead of 1 for the first page.
public procedure insert_page ( integer position, atom book, atom page, sequence text, integer select, atom image )
    c_proc( WX_INSERT_PAGE, {position, book, page, text, select, image} )
end procedure

--/topic Notebook Controls
--/func get_current_page ( atom book )
--
--This returns the current selected page of a wxNotebook. The index is 0 based, so the first page will be 0, instead of 1.
public function get_current_page ( atom book )
    return c_func( WX_GET_CURRENT_PAGE, {book} )
end function

--/topic Notebook Controls
--/func get_next_page ( atom book, integer show )
--
--This returns the next page in the book if show is 1, or shows the previous page if show is 0. If previous is chosen on the first page, the last page will be returned, and vice versa. Also, keep in mind of the wxNotebook's 0-based index.
public function get_next_page ( atom book, integer show )
    return c_func( WX_GET_NEXT_PAGE, {book, show} )
end function

--/topic Notebook Controls
--/func get_page_count ( atom book )
--
--This returns the number of pages in a wxNotebook.
public function get_page_count ( atom book )
    return c_func( WX_GET_PAGE_COUNT, {book} )
end function

--/topic Notebook Controls
--/proc get_page_image ( atom book, integer page )
--
--This returns the number of the sequence that the page gets it's image from if an
--imagelist is set. If it gets the image from the second pointer in a sequence, it will
--return a 1. Keep in mind that the wxNotebook works from a 0-based index as well.
public function get_page_image ( atom book, integer page )
    return c_func( WX_GET_PAGE_IMAGE, {book, page} )
end function

--/topic Notebook Controls
--/proc remove_page ( atom book, integer page )
--
--Removes the selected page from the wxNotebook. The index is 0 based, so the first page will be 0, instead of 1.
public procedure remove_page ( atom book, integer page )
    c_proc( WX_REMOVE_PAGE, {book, page} )
end procedure

--/topic Notebook Controls
--/proc set_page_image ( atom book, integer page, atom image )
--
--This sets the selected page to have the selected image, so long as an imagelist
--has been set.
public procedure set_page_image ( atom book, integer page, atom image )
    c_proc( WX_SET_PAGE_IMAGE, {book, page, image} )
end procedure

--/topic Notebook Controls
--/proc set_page_selection ( atom book, integer page )
--
--This sets the wxNotebook to show the specified page. The index is 0 based, so
--the first page will be 0, instead of 1.
public procedure set_page_selection ( atom book, integer page )
    c_proc( WX_SET_PAGE_SELECTION, {book, page} )
end procedure



--wxBoxSizer
------------------------------
--/topic wxBoxSizer
--/parent Sizers
--/parent Classes
--/info
--This is the most common sizer you should use.
-- It is possible to create very complex sizing relationships between
-- controls by nesting many wxBoxSizers together.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "integer orient"
--<ul>
--/li wxVERTICAL
--/li wxHORIZONTAL
--</ul>
--</ul>



--wxBrush
------------------------------
--/topic wxBrush
--/parent Graphics
--/parent Classes
--/info
--A brush is a drawing tool for filling in areas.
-- It is used for painting the background of rectangles,
-- ellipses, etc. It has a colour and a style.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom /wxColour"
--/li integer style [= wxSOLID]
--<ul>
--/li wxSOLID
--/li wxTRANSPARENT
--</ul>
--</ul>

-- wxButton
------------------------------
--/topic wxButton
--/parent Controls
--/info
--A button is a control that contains a text string, and is one of the most common elements of a GUI.
-- It may be placed on a dialog box or panel, or indeed almost any other window.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence label_text [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= 0]
--<ul><table>
--<tr><td> /li wxBU_LEFT <td>Left-justifies the label. /i "WIN32 only."
--<tr><td> /li wxBU_TOP <td>Aligns the label to the top of the button. /i "WIN32 only."
--<tr><td> /li wxBU_RIGHT <td>Right-justifies the bitmap label. /i "WIN32 only."
--<tr><td> /li wxBU_BOTTOM <td>Aligns the label to the bottom of the button. /i "WIN32 only."
--</table></ul>
--</ul>
--
--To trap a button click, use the wxEVT_COMMAND_BUTTON_CLICKED event:
--/code
--ex:
--
--  procedure on_click( atom this, atom event_type, atom id, atom event )
--    -- do stuff here
--  end procedure
--
--  set_event_handler( my_button, get_id(my_button), wxEVT_COMMAND_BUTTON_CLICKED,
--      routine_id("on_click"))
--/endcode

--/topic wxButton
--/proc set_button_label( atom button, sequence label_text )
--
--Changes the label on a button.
public procedure set_button_label( atom button, sequence label_text )
    c_proc( WX_SET_BUTTON_LABEL, {button, label_text} )
end procedure



--wxCalendarCtrl
------------------------------
--/topic wxCalendarCtrl
--/parent Controls
--/info
--The calendar control allows the user to pick a date.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li atom date [= wxDefaultDateTime]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= wxCAL_SHOW_HOLIDAYS]
--<ul><table>
--<tr><td> /li wxCAL_SUNDAY_FIRST               <td>Show Sunday as the first day in the week
--<tr><td> /li wxCAL_MONDAY_FIRST               <td>Show Monday as the first day in the week
--<tr><td> /li wxCAL_SHOW_HOLIDAYS              <td>Highlight holidays in the calendar
--<tr><td> /li wxCAL_NO_YEAR_CHANGE             <td>Disable the year changing
--<tr><td> /li wxCAL_NO_MONTH_CHANGE            <td>Disable the month (and, implicitly, the year) changing
--<tr><td> /li wxCAL_SHOW_SURROUNDING_WEEKS     <td>Show the neighbouring weeks in the previous and next months
--<tr><td> /li wxCAL_SEQUENTIAL_MONTH_SELECTION <td>Use alternative, more compact, style for the month and year selection controls.
--</table></ul>
--</ul>

--/topic wxCalendarCtrl
--/func get_date( atom cal )
--
--Returns the selected date.  See /read_datetime() for format information.
public function get_date( atom cal )
    return c_func( WX_GET_DATE, {cal} )
end function

--/topic wxCalendarCtrl
--/func get_date_time( atom cal )
--
--Returns a /wxDateTime object for the selected date on the calendar.
public function get_date_time( atom cal )
    return c_func( WX_GET_DATE_TIME, {cal} )
end function

--/topic wxCalendarCtrl
--/func get_cal_days_in_month( atom cal )
--
--Returns the number of days in the currently selected month.
public function get_cal_days_in_month( atom cal )
    return c_func( WX_GET_CAL_DAYS_IN_MONTH, {cal} )
end function

--/topic wxCalendarCtrl
--/proc set_date( atom cal, sequence Date )
--
--See /wxDateTime for the proper format of /i Date.
public procedure set_date( atom cal, sequence Date )
    c_proc( WX_SET_DATE, {cal, Date} )
end procedure

--/topic wxCalendarCtrl
--/proc set_date_text_color( atom cal, integer day, atom color )
--
--/i day can be either a single day (integer) or a sequence containing multiple days.
--It should be in the range of 1..31
public procedure set_date_text_color( atom cal, object day, atom color )
    c_proc( WX_SET_DATE_TEXT_COLOR, {cal, day, color} )
end procedure

--/topic wxCalendarCtrl
--/proc set_date_border( atom cal, integer day, integer border )
--
--/i day can be either a single day (integer) or a sequence containing multiple days.
--It should be in the range of 1..31
--
--/i border:
--<ul>
--/li wxCAL_BORDER_NONE
--/li wxCAL_BORDER_SQUARE
--/li wxCAL_BORDER_ROUND
--</ul>
public procedure set_date_border( atom cal, object day, integer border )
    c_proc( WX_SET_DATE_BORDER, {cal, day, border} )
end procedure

--/topic wxCalendarCtrl
--/proc set_date_back_color( atom cal, integer day, atom color )
--
--/i day can be either a single day (integer) or a sequence containing multiple days.
--It should be in the range of 1..31.
public procedure set_date_back_color( atom cal, object day, atom color )
    c_proc( WX_SET_DATE_BACK_COLOR, {cal, day, color} )
end procedure

--/topic wxCalendarCtrl
--/proc set_date_border_color( atom cal, integer day, atom color )
--
--/i day can be either a single day (integer) or a sequence containing multiple days.
--It should be in the range of 1..31
public procedure set_date_border_color( atom cal, object day, atom color )
    c_proc( WX_SET_DATE_BORDER_COLOR, {cal, day, color} )
end procedure

--/topic wxCalendarCtrl
--/proc set_date_font( atom cal, integer day, atom font )
--
--/i day can be either a single day (integer) or a sequence containing multiple days.
--It should be in the range of 1..31
public procedure set_date_font( atom cal, object day, atom font )
    c_proc( WX_SET_DATE_FONT, {cal, day, font} )
end procedure

--/topic wxCalendarCtrl
--/proc set_date_holiday( atom cal, integer day, atom color )
--
--/i day can be either a single day (integer) or a sequence containing multiple days.
--It should be in the range of 1..31.
--
--/i holiday should be 1 if the day is a holiday, or 0 if not a holiday.
public procedure set_date_holiday( atom cal, object day, integer holiday )
    c_proc( WX_SET_DATE_HOLIDAY, {cal, day, holiday} )
end procedure



--wxCaret
-----------------------------------------
--/topic wxCaret
--/parent Graphics
--/parent Classes
--/info
--A caret is a blinking cursor showing the position where the typed text will appear.
--The text controls usually have a caret but wxCaret class also allows to use a caret in other windows.
--
--Currently, the caret appears as a rectangle of the given size. In the future, it will be possible to
-- specify a bitmap to be used for the caret shape.
--
--A caret is always associated with a window and the current caret can be retrieved using /get_caret_window().
-- The same caret can't be reused in two different windows.

--/topic wxCaret
--/func get_caret_blink_time( atom caret )
--Returns the blink time which is measured in milliseconds and is the time elapsed between 2 inversions
-- of the caret (blink time of the caret is the same for all carets, so this functions is static).
public function get_caret_blink_time( atom caret )
    return c_func( WX_GET_CARET_BLINK_TIME, {caret})
end function

--/topic wxCaret
--/func get_caret_position( atom caret )
--Get the caret position (in pixels).
public function get_caret_position( atom caret )
    return c_func( WX_GET_CARET_POSITION, {caret})
end function

--/topic wxCaret
--/func get_caret_size( atom caret )
--Get the caret size.
public function get_caret_size( atom caret )
    return c_func( WX_GET_CARET_SIZE, {caret})
end function

--/topic wxCaret
--/func get_caret_window( atom caret )
--Get the window the caret is associated with.
public function get_caret_window( atom caret )
    return c_func( WX_GET_CARET_WINDOW, {caret})
end function

--/topic wxCaret
--/proc hide_caret( atom caret )
--Hides the caret. Notice that if the caret was hidden N times, it must be shown N times as well to
-- reappear on the screen.
public procedure hide_caret( atom caret )
    c_proc( WX_HIDE_CARET, {caret})
end procedure

--/topic wxCaret
--/func caret_isok( atom caret )
--Returns true if the caret was created successfully.
public function caret_isok( atom caret )
    return c_func( WX_CARET_ISOK, {caret})
end function

--/topic wxCaret
--/func caret_isvisible( atom caret )
--Returns true if the caret is visible and false if it is permanently hidden (if it is is blinking and
-- not shown currently but will be after the next blink, this method still returns true).
public function caret_isvisible( atom caret )
    return c_func( WX_CARET_ISVISIBLE, {caret})
end function

--/topic wxCaret
--/proc move_caret( atom caret, atom x, atom y )
--Move the caret to given position (in logical coordinates).
public procedure move_caret( atom caret, atom x, atom y )
    c_proc( WX_MOVE_CARET, {caret,x,y})
end procedure

--/topic wxCaret
--/proc set_caret_blink_time( atom caret, atom milliseconds )
--Sets the blink time for all the carets.
--
--/b Remarks
--
--Under Windows, this function will change the blink time for all carets permanently (until the next time
-- it is called), even for the carets in other applications.
public procedure set_caret_blink_time( atom caret, atom milliseconds )
    c_proc( WX_SET_CARET_BLINK_TIME, {caret,milliseconds})
end procedure

--/topic wxCaret
--/proc set_caret_size( atom caret, atom width, atom height )
--Changes the size of the caret.
public procedure set_caret_size( atom caret, atom width, atom height )
    c_proc( WX_SET_CARET_SIZE, {caret,width,height})
end procedure

--/topic wxCaret
--/proc show_caret( atom caret )
--Shows the caret. Notice that if the caret was hidden N times, it must be shown N times as well to
-- reappear on the screen.
public procedure show_caret( atom caret )
    c_proc( WX_SHOW_CARET, {caret})
end procedure

--/topic wxCaret
--/proc kill_caret_focus( atom caret )
--Performs internal maintenance on the caret when focus is lost. Call this from your /b wxEVT_KILL_FOCUS event handler.
public procedure kill_caret_focus( atom caret )
    c_proc( WX_CARET_KILL_FOCUS, {caret} )
end procedure

--/topic wxCaret
--/proc set_caret_focus( atom caret )
--Performs internal maintenance on the caret when focus is gained. Call this from your /b wxEVT_SET_FOCUS event handler.
public procedure set_caret_focus( atom caret )
    c_proc( WX_CARET_SET_FOCUS, {caret} )
end procedure

--/topic wxCaret
--/proc caret_kill_focus( atom caret )
--Performs internal maintenance on the caret. Call this when your window loses focus.
public procedure caret_kill_focus( atom caret )
    c_proc( WX_CARET_KILL_FOCUS, {caret})
end procedure

--/topic wxCaret
--/proc caret_set_focus( atom caret )
--Performs internal maintenance on the caret. Call this when your window gains focus.
public procedure caret_set_focus( atom caret )
    c_proc( WX_CARET_SET_FOCUS, {caret})
end procedure



--wxCheckBox
------------------------------
--/topic wxCheckBox
--/parent Controls
--/info
--A checkbox is a labeled box which by default is either on (checkmark is visible) or off (no checkmark).
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li seqeunce label_text [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= 0]
--</ul>

--/topic wxCheckBox
--/func get_checked( atom this )
--
--Returns 1 if checked, 0 otherwise.
--This function works for /wxCheckBox and /wxToggleButton.
public function get_checked( atom this )
    return c_func( WX_GET_CHECKED, {this} )
end function

--/topic wxCheckBox
--/proc set_checked( atom this, integer check )
--
--This procedure works for /wxCheckBox and /wxToggleButton.
public procedure set_checked( atom this, integer check )
    c_proc( WX_SET_CHECKED, {this, check} )
end procedure



--wxChoice
------------------------------
--/topic wxChoice
--/parent List Controls
--/parent Controls
--/info
--A choice item is used to select one of a list of strings. Unlike a /wxListBox, only the selection
-- is visible until the user pulls down the menu of choices.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li sequence choices [= {}]
--/li integer style [= 0]
--</ul>

--/topic wxChoice
--/func choice_count( atom choice )
--
--Used to count the number of items in a /wxChoice.
public function choice_count( atom choice )
    return c_func( WX_CHOICE_COUNT, {choice} )
end function

--/topic wxChoice
--/proc clear_choice( atom choice )
--
--Clears the data inside of a wxChoice specified by choice.
public procedure clear_choice( atom choice )
    c_proc( WX_CLEAR_CHOICE, {choice} )
end procedure

--/topic wxChoice
--/proc set_choice( atom choice, sequence text )
--
--Selects the item with the specified string in the control. This doesn't cause any command events being emitted.
public procedure set_choice( atom choice, sequence text )
    c_proc( WX_SET_CHOICE, {choice, text} )
end procedure

--/topic wxChoice
--/func find_choice_string( atom choice, sequence string )
--
--Returns the item number of the item in the list with the
--specified string.
public function find_choice_string( atom choice, sequence string )
    return c_func( WX_FIND_CHOICE_STRING, {choice, string} )
end function



--wxChoicebook
-----------------------------------------
--/topic wxChoicebook
--/parent Notebook Controls
--/parent Controls
--/info
--wxChoicebook is a class similar to /wxNotebook but which uses a /wxChoice to show the labels instead of the tabs.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cy [=-1]
--/li integer style [= 0]
--</ul>


--wxClientDC
---------------------------------
--/topic wxClientDC
--/parent wxDC
--/parent Graphics
--/parent Classes
--/info
--/b wxClientDCs are used to draw to /wxWindows outside of /b wxEVT_PAINT events.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom /wxWindow"
--</ul>
--
--/b Note: you must delete the DC with /delete_instance() when you are finished.



--wxColour
---------------------------------
--/topic wxColour
--/parent Graphics
--/parent Classes
--/info
--A /b wxColour is an object representing a combination of Red, Green, and Blue (RGB) intensity values, and is used to determine drawing colours.
--
--<div class="subsection">Creation Parameters</div>
--There are two methods:
--<ul>
--/li a sequence of {red, green, blue} values
--/li a named colour in /b wxTheColourDatabase
--<ul style="height: 320px; width: 240px; overflow: scroll;">
--/li "AQUAMARINE"
--/li "BLACK"
--/li "BLUE"
--/li "BLUE VIOLET"
--/li "BROWN"
--/li "CADET BLUE"
--/li "CORAL"
--/li "CORNFLOWER BLUE"
--/li "CYAN"
--/li "DARK GREY"
--/li "DARK GREEN"
--/li "DARK OLIVE GREEN"
--/li "DARK ORCHID"
--/li "DARK SLATE BLUE"
--/li "DARK SLATE GREY"
--/li "DARK TURQUOISE"
--/li "DIM GREY"
--/li "FIREBRICK"
--/li "FOREST GREEN"
--/li "GOLD"
--/li "GOLDENROD"
--/li "GREY"
--/li "GREEN"
--/li "GREEN YELLOW"
--/li "INDIAN RED"
--/li "KHAKI"
--/li "LIGHT BLUE"
--/li "LIGHT GREY"
--/li "LIGHT STEEL BLUE"
--/li "LIME GREEN"
--/li "MAGENTA"
--/li "MAROON"
--/li "MEDIUM AQUAMARINE"
--/li "MEDIUM BLUE"
--/li "MEDIUM FOREST GREEN"
--/li "MEDIUM GOLDENROD"
--/li "MEDIUM ORCHID"
--/li "MEDIUM SEA GREEN"
--/li "MEDIUM SLATE BLUE"
--/li "MEDIUM SPRING GREEN"
--/li "MEDIUM TURQUOISE"
--/li "MEDIUM VIOLET RED"
--/li "MIDNIGHT BLUE"
--/li "NAVY"
--/li "ORANGE"
--/li "ORANGE RED"
--/li "ORCHID"
--/li "PALE GREEN"
--/li "PINK"
--/li "PLUM"
--/li "PURPLE"
--/li "RED"
--/li "SALMON"
--/li "SEA GREEN"
--/li "SIENNA"
--/li "SKY BLUE"
--/li "SLATE BLUE"
--/li "SPRING GREEN"
--/li "STEEL BLUE"
--/li "TAN"
--/li "THISTLE"
--/li "TURQUOISE"
--/li "VIOLET"
--/li "VIOLET RED"
--/li "WHEAT"
--/li "WHITE"
--/li "YELLOW"
--/li "YELLOW GREEN"
--</ul>
--</ul>
--
--/b Note: valid RGB values are in the range 0 to 255.
--
--<div class="subsection">See also</div>
--<ul>
--/li /wxPen
--/li /wxBrush
--/li /wxColourDialog
--/li /get_sys_color
--</ul>

--/topic wxColour
--/topic wxImage
--/func get_red( object params )
--
--Returns the red intensity of a /wxColour or { /wxImage, x, y }.
public function get_red( object color )
    return c_func( WX_GET_RED, {color} )
end function

--/topic wxColour
--/topic wxImage
--/func get_green( atom color )
--
--Returns the green intensity of a /wxColour or { /wxImage, x, y }.
public function get_green( object color )
    return c_func( WX_GET_GREEN, {color} )
end function

--/topic wxColour
--/topic wxImage
--/func get_blue( atom color )
--
--Returns the blue intensity of a /wxColour or { /wxImage, x, y }.
public function get_blue( object color )
    return c_func( WX_GET_BLUE, {color} )
end function



--wxColourData
------------------------------
--/topic wxColourData
--/parent wxColourDialog
--/parent Classes
--/info
--This class holds a variety of information related to colour dialogs.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b None
--</ul>

--/topic wxColourData
--/func get_color( atom colordata )
--
--Gets the current colour associated with the colour dialog.
--
--The default colour is black.
public function get_color( atom colordata )
    return c_func( WX_GET_COLOR, {colordata} )
end function



--wxColourDialog
------------------------------
--/topic wxColourDialog
--/parent Dialogs
--/parent Classes
--/info
--This class represents the colour chooser dialog.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li atom parent [= 0]
--/li atom /wxColourData [= 0]
--</ul>

--/topic wxColourDialog
--/func get_color_data( atom dialog )
--
--Returns the colour data associated with the colour dialog.
public function get_color_data( atom dialog )
    return c_func( WX_GET_COLOR_DATA, {dialog} )
end function



--wxComboBox
------------------------------
--/topic wxComboBox
--/parent List Controls
--/parent wxChoice
--/parent Controls
--/info
--A combobox is like a combination of an edit control and a listbox.
--
--It can be displayed as static list with editable or read-only text field;
-- or a drop-down list with text field; or a drop-down list without a text field.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence value [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li sequence choices [= {}]
--/li integer style [= 0]
--</ul>

--/topic wxComboBox
--/func combobox_count( atom combo )
--
--Used to count the number of items in a /wxChoice.
public function combobox_count( atom combobox )
    return c_func( WX_COMBOBOX_COUNT, {combobox} )
end function

--/topic wxComboBox
--/func get_combo_string( atom combo )
--
--Returns the string in the edit box of the combo box.
public function get_combo_string( atom combo )
    return c_func( WX_GET_COMBO_STRING, {combo} )
end function

--/topic wxComboBox
--/func get_combo_item_string( atom combo, integer item )
--
--Returns the string for /i item.
public function get_combo_item_string( atom combo, integer item )
    return c_func( WX_GET_COMBO_ITEM_STRING, {combo, item} )
end function

--/topic wxComboBox
--/proc clear_combo( atom combo )
--
--Clears the data inside of a wxComboBox specified by combo.
public procedure clear_combo( atom combo )
    c_proc( WX_CLEAR_COMBO, {combo} )
end procedure

--/topic wxComboBox
--/proc set_combo_value( atom combo, sequence text )
--
--This sets the text of the box in a wxComboBox to text.
public procedure set_combo_value( atom combo, sequence text )
    c_proc( WX_SET_COMBO_VALUE, {combo, text} )
end procedure

--/topic wxComboBox
--/func find_combo_string( atom combo, sequence string )
--
--Returns the item number of the item in the ComboBox with the
--specified string. The top item is 0 instead of 1, so be sure to
--take that into account.
public function find_combo_string( atom combo, sequence string )
    return c_func( WX_FIND_COMBO_STRING, {combo, string} )
end function

--/topic wxComboBox
--/proc delete_combo_item( atom combo, integer item )
--
--This goes to a ComboBox specified by combo and deletes the number
--in the box corresponding to item. It starts at the top with 0
--being the first on the drop-down list.
public procedure delete_combo_item( atom combo, integer item )
    c_proc( WX_DELETE_COMBO_ITEM, {combo, item} )
end procedure



--wxConfig
-----------------------------------------
--/topic wxConfig
--/parent Classes
--/info
--A basic configuration class.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "sequence appName"
--/li sequence vendorName [= ""]
--/li sequence localFilename [= ""]
--/li sequence globalFilename [= ""]
--/li atom style [= 0]
--</ul>
--
--wxConfig allows you to write the same code regardless of whether you're working with the registry
-- under Win32 or text-based config files under Unix. To make writing the portable code even easier,
-- wxEuphoria automatically maps to the native wxConfig implementation on the given platform: i.e.
-- wxRegConfig under Win32 and wxFileConfig otherwise.

--/topic wxConfig
--/func config_read( atom config, object key, object defaultValue )
--
--Read an object from the /i key. The /i defaultValue is returned if the key was not found.
--
--/b Note: The return object is expected to be the same type as the /i defaultValue object.
public function config_read( atom config, object key, object defaultValue )
    return c_func( WX_CONFIG_READ, {config,key,defaultValue})
end function

--/topic wxConfig
--/proc config_write( atom config, object key, object value )
--
--Write the specified value to the config file.
public procedure config_write( atom config, object key, object value )
    c_proc( WX_CONFIG_WRITE, {config,key,value})
end procedure

--/topic wxConfig
--/proc config_delete_all( atom config )
--
--Delete the whole underlying object (disk file, registry key, ...). Primarly for use by uninstallation routine.
public procedure config_delete_all( atom config )
    c_proc( WX_CONFIG_DELETE_ALL, {config})
end procedure

--/topic wxConfig
--/proc config_delete_entry( atom config, object key, atom delete_group_if_empty )
--
--Deletes the specified entry and the group it belongs to if it was the last key in it and the second parameter is true.
public procedure config_delete_entry( atom config, object key, atom delete_group_if_empty )
    c_proc( WX_CONFIG_DELETE_ENTRY, {config,key,delete_group_if_empty})
end procedure

--/topic wxConfig
--/proc config_delete_group( atom config, object key )
--
--Delete the group (with all subgroups). If the current path is under the group being deleted it is changed to its deepest still
-- existing component. E.g. if the current path is /A/B/C/D and the group C is deleted the path becomes /A/B.
public procedure config_delete_group( atom config, object key )
    c_proc( WX_CONFIG_DELETE_GROUP, {config,key})
end procedure

--/topic wxConfig
--/proc config_exists( atom config, object key )
--
--Returns true if either a group or an entry with a given name exists.
public procedure config_exists( atom config, object key )
    c_proc( WX_CONFIG_EXISTS, {config,key})
end procedure

--/topic wxConfig
--/proc config_flush( atom config, atom current_only )
--
--Permanently writes all changes (otherwise, they're only written when the object is destroyed).
public procedure config_flush( atom config, atom current_only )
    c_proc( WX_CONFIG_FLUSH, {config,current_only})
end procedure

--/topic wxConfig
--/func get_config_app_name( atom config )
--
--Returns the application name.
public function get_config_app_name( atom config )
    return c_func( WX_GET_CONFIG_APP_NAME, {config})
end function

--/topic wxConfig
--/func get_first_config_group( atom config, atom index )
--
--Gets the first group.
public function get_first_config_group( atom config, atom index )
    return c_func( WX_GET_FIRST_CONFIG_GROUP, {config,index})
end function

--/topic wxConfig
--/func get_first_config_entry( atom config, atom index )
--
--Gets the first entry.
public function get_first_config_entry( atom config, atom index )
    return c_func( WX_GET_FIRST_CONFIG_ENTRY, {config,index})
end function

--/topic wxConfig
--/func get_next_config_group( atom config, atom index )
--
--Gets the next group.
public function get_next_config_group( atom config, atom index )
    return c_func( WX_GET_NEXT_CONFIG_GROUP, {config,index})
end function

--/topic wxConfig
--/func get_next_config_entry( atom config, atom index )
--
--Gets the next entry.
public function get_next_config_entry( atom config, atom index )
    return c_func( WX_GET_NEXT_CONFIG_ENTRY, {config,index})
end function

--/topic wxConfig
--/proc delete_config( object obj )
--
--Deletes the wxConfig object.
public procedure delete_config( object obj )
    c_proc( WX_DELETE_CONFIG, {obj})
end procedure



--wxCursor
------------------------------
--/topic wxCursor
--/parent Classes
--/info
--A cursor is a small bitmap usually used for denoting where the mouse pointer is, with
-- a picture that might indicate the interpretation of a mouse click.
--
--<div class="subsection">Creation Parameters</div>
--There are two methods:
--<ul>
--/li atom /wxImage (create a /wxBitmap and use /convert_to_image)
--/li Cursor id
--<ul>
--/li wxCURSOR_NONE
--/li wxCURSOR_ARROW
--/li wxCURSOR_RIGHT_ARROW
--/li wxCURSOR_BULLSEYE
--/li wxCURSOR_CHAR
--/li wxCURSOR_CROSS
--/li wxCURSOR_HAND
--/li wxCURSOR_IBEAM
--/li wxCURSOR_LEFT_BUTTON
--/li wxCURSOR_MAGNIFIER
--/li wxCURSOR_MIDDLE_BUTTON
--/li wxCURSOR_NO_ENTRY
--/li wxCURSOR_PAINT_BRUSH
--/li wxCURSOR_PENCIL
--/li wxCURSOR_POINT_LEFT
--/li wxCURSOR_POINT_RIGHT
--/li wxCURSOR_QUESTION_ARROW
--/li wxCURSOR_RIGHT_BUTTON
--/li wxCURSOR_SIZENESW
--/li wxCURSOR_SIZENS
--/li wxCURSOR_SIZENWSE
--/li wxCURSOR_SIZEWE
--/li wxCURSOR_SIZING
--/li wxCURSOR_SPRAYCAN
--/li wxCURSOR_WAIT
--/li wxCURSOR_WATCH
--/li wxCURSOR_BLANK
--</ul>
--</ul>



--wxDataFormat
------------------------------
--/topic wxDataFormat
--/parent Classes
--/parent Clipboard
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "integer data_format"
--<ul>
--/li wxDF_INVALID
--/li wxDF_TEXT
--/li wxDF_BITMAP
--/li wxDF_METAFILE
--/li wxDF_SYLK
--/li wxDF_DIF
--/li wxDF_TIFF
--/li wxDF_OEMTEXT
--/li wxDF_DIB
--/li wxDF_PALETTE
--/li wxDF_PENDATA
--/li wxDF_RIFF
--/li wxDF_WAVE
--/li wxDF_UNICODETEXT
--/li wxDF_ENHMETAFILE
--/li wxDF_FILENAME
--/li wxDF_LOCALE
--/li wxDF_PRIVATE
--</ul>
--</ul>



--wxDateTime
------------------------------
--/topic wxDateTime
--/parent wxCalendarCtrl
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b None
--</ul>
--
--Dates will be sequences with values:
--<ul>
--/li Year
--/li Month
--/li Day of Month
--/li Hour
--/li Minute
--/li Second
--/li Millisecond
--/li Day of Week
--/li Day of Year
--/li Timezone (0 = local)
--</ul>

--/topic wxDateTime
--/func get_days_in_month( atom dt )
--
--Returns the number of days in the specified month and year.
public function get_days_in_month( integer month, integer year )
    return c_func( WX_GET_DAYS_IN_MONTH, {month, year} )
end function



--wxDC
---------------------------------
--/topic wxDC
--/parent Graphics
--/parent Classes

--/topic wxDC
--/proc set_pen_size( atom dc, integer width )
--
--Sets the pen width.
public procedure set_pen_size( atom dc, integer width )
    c_proc( WX_SET_PEN_SIZE, {dc, width} )
end procedure

--/topic wxDC
--/proc set_pen_color( atom dc, atom color )
--
--Sets the pen color.
public procedure set_pen_color( atom dc, atom color )
    c_proc( WX_SET_PEN_COLOR, {dc, color} )
end procedure



--wxDefButton
------------------------------
--/topic wxDefButton
--/parent Classes
--/parent wxButton
--/info
--
--/b wxDefButton is not a true wxWidgets class.  It is actually identical to a /wxButton, except
--that /set_default_item() is called.  The parent of a wxDefButton /i must be a /wxPanel.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li See /wxButton
--</ul>


--wxDialog
------------------------------
--/topic wxDialog
--/parent Dialogs
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li /b "integer id"
--/li /b "sequence title"
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= wxDEFAULT_DIALOG_STYLE] (see /wxFrame)
--</ul>

--/topic wxDialog
--/func show_modal( atom dialog )
--
--Opens a modal dialog, which grabs the focus, and does not release
--it until it is closed.  This is the way to use a dialog to get
--user input.  See also /end_modal.
public function show_modal( atom dialog )
    return c_func( WX_SHOW_MODAL, {dialog} )
end function

--/topic wxDialog
--/proc end_modal( atom dialog, atom ret )
--
--Closes a modal dialog.  /b ret is the value returned
--by /show_modal().
public procedure end_modal( atom dialog, atom ret )
    c_proc( WX_END_MODAL, {dialog, ret} )
end procedure

--/topic wxDialog
--/func create_button_sizer( atom dlg, atom flags )
--Creates a sizer with standard buttons. /i flags is a bit list of the following flags:
-- wxOK, wxCANCEL, wxYES, wxNO, wxHELP, wxNO_DEFAULT.
public function create_button_sizer( atom dlg, atom flags )
    return c_func( WX_CREATE_BUTTON_SIZER, {dlg,flags})
end function



--wxDir
------------------------------
--/topic wxDir
--/parent Files
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li sequence directory [= ""]
--</ul>

--/topic wxDir
--/func is_opened( atom dir )
--
--Returns true if the directory was successfully opened by a previous call to /open_dir.
public function is_opened( atom dir )
    return c_func( WX_IS_OPENED, {dir} )
end function

--/topic wxDir
--/func get_all_files( object dirname, sequence filespec, integer flags, atom traverser )
--
--Walks the specified directory, returning all the files and/or directories as specified by
--/b flags.
--<ul>
--/li /b dirname
--<ul>
--/li sequence: directory name (a /wxDir will be created and destroyed)
--/li atom: instance of /wxDir to be used
--</ul>
--/li /b filespec: wildcard pattern to be matched against files
--/li /b flags: mixture of wxDIR_FILES, wxDIR_HIDDEN, wxDIR_DIRS, wxDIR_DOTDOT
--/li /b traverser
--<ul>
--/li 0: Use the default /wxDirTraverser which always returns wxDIR_CONTINUE
--/li /wxDirTraverser: Use the supplied /wxDirTraverser
--</ul>
--</ul>
public function get_all_files( object dirname, sequence filespec, integer flags, atom traverser )
    return c_func( WX_GET_ALL_FILES, {dirname, filespec, flags, traverser} )
end function

--/topic wxDir
--/func has_files( atom dir, sequence spec )
--
--Checks to see if any files in it's current directory contain
--files specified by spec. This uses wildcards to find things,
--so you can do something like "*.e" for spec to find files
--ending with '.e'. However, this returns an integer pointer if
--it finds anything, and a 0 if it finds nothing.
public function has_files( atom dir, sequence spec )
    return c_func( WX_HAS_FILES, {dir, spec} )
end function

--/topic wxDir
--/func has_subdirs( atom dir, sequence spec )
--
--This is similar to has_files, except that has_subdirs checks
--the subdirectories instead of just the current directory for
--the specified files.
public function has_subdirs( atom dir, sequence spec )
    return c_func( WX_HAS_SUBDIRS, {dir, spec} )
end function

--/topic wxDir
--/func open_dir( atom dir, sequence dname )
--
--Open the directory for enumerating, returns true on success or false if an error occurred.
public function open_dir( atom dir, sequence dname )
    return c_func( WX_OPEN_DIR, {dir, dname} )
end function

--/topic wxDir
--/func dir_exists( sequence directory )
--
--Returns 1 if the directory exists, 0 if it does not.
public function dir_exists( sequence directory )
    return c_func( WX_DIR_EXISTS, {directory} )
end function



--wxDirCtrl
------------------------------
--/topic wxDirCtrl
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence default_path [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= 0]
--<ul><table>
--<tr><td> /li wxDIRCTRL_DIR_ONLY     <td>Only show directories, and not files.
--<tr><td> /li wxDIRCTRL_3D_INTERNAL  <td>Use 3D borders for internal controls.
--<tr><td> /li wxDIRCTRL_SELECT_FIRST <td>When setting the default path, select the first file in the directory.
--<tr><td> /li wxDIRCTRL_SHOW_FILTERS <td>Show the drop-down filter list.
--<tr><td> /li wxDIRCTRL_EDIT_LABELS  <td>Allow the folder and file labels to be editable.
--</table></ul>
--</ul>

--/topic wxDirCtrl
--/proc collapse_dirctrl_tree( atom ctrl )
--Collapses the entire tree.
public procedure collapse_dirctrl_tree( atom ctrl )
    c_proc( WX_COLLAPSE_DIRCTRL_TREE, {ctrl})
end procedure

--/topic wxDirCtrl
--/func expand_dirctrl_path( atom ctrl, object path )
--Tries to expand as much of the given path as possible, so that the filename or directory is visible in the tree control.
public function expand_dirctrl_path( atom ctrl, object path )
    return c_func( WX_EXPAND_DIRCTRL_PATH, {ctrl,path})
end function

--/topic wxDirCtrl
--/func collapse_dirctrl_path( atom ctrl, object path )
--Collapse the given path.
public function collapse_dirctrl_path( atom ctrl, object path )
    return c_func( WX_COLLAPSE_DIRCTRL_PATH, {ctrl,path})
end function

--/topic wxDirCtrl
--/func get_dirctrl_default_path( atom ctrl )
--Gets the default path.
public function get_dirctrl_default_path( atom ctrl )
    return c_func( WX_GET_DIRCTRL_DEFAULT_PATH, {ctrl})
end function

--/topic wxDirCtrl
--/func get_dirctrl_path( atom ctrl )
--Gets the currently-selected directory or filename.
public function get_dirctrl_path( atom ctrl )
    return c_func( WX_GET_DIRCTRL_PATH, {ctrl})
end function

--/topic wxDirCtrl
--/func get_dirctrl_file_path( atom ctrl )
--Gets selected filename path only (else empty string).
--
--This function doesn't count a directory as a selection.
public function get_dirctrl_file_path( atom ctrl )
    return c_func( WX_GET_DIRCTRL_FILE_PATH, {ctrl})
end function

--/topic wxDirCtrl
--/func get_dirctrl_filter( atom ctrl )
--Returns the filter string.
public function get_dirctrl_filter( atom ctrl )
    return c_func( WX_GET_DIRCTRL_FILTER, {ctrl})
end function

--/topic wxDirCtrl
--/func get_dirctrl_filter_index( atom ctrl )
--Returns the current filter index (zero-based).
public function get_dirctrl_filter_index( atom ctrl )
    return c_func( WX_GET_DIRCTRL_FILTER_INDEX, {ctrl})
end function

--/topic wxDirCtrl
--/func get_dirctrl_filter_listctrl( atom ctrl )
--Returns a pointer to the filter list control (if present).
public function get_dirctrl_filter_listctrl( atom ctrl )
    return c_func( WX_GET_DIRCTRL_FILTER_LISTCTRL, {ctrl})
end function

--/topic wxDirCtrl
--/func get_dirctrl_root_id( atom ctrl )
--Returns the root id for the tree control.
public function get_dirctrl_root_id( atom ctrl )
    return c_func( WX_GET_DIRCTRL_ROOT_ID, {ctrl})
end function

--/topic wxDirCtrl
--/func get_dirctrl_treectrl( atom ctrl )
--Returns a pointer to the tree control.
public function get_dirctrl_treectrl( atom ctrl )
    return c_func( WX_GET_DIRCTRL_TREECTRL, {ctrl})
end function

--/topic wxDirCtrl
--/proc recreate_dirctrl_tree( atom ctrl )
--Collapse and expand the tree, thus re-creating it from scratch. May be used to update the displayed directory content.
public procedure recreate_dirctrl_tree( atom ctrl )
    c_proc( WX_RECREATE_DIRCTRL_TREE, {ctrl})
end procedure

--/topic wxDirCtrl
--/proc set_dirctrl_default_path( atom ctrl, object path )
--Sets the default path.
public procedure set_dirctrl_default_path( atom ctrl, object path )
    c_proc( WX_SET_DIRCTRL_DEFAULT_PATH, {ctrl,path})
end procedure

--/topic wxDirCtrl
--/proc set_dirctrl_filter( atom ctrl, object filter )
--Sets the filter string.
public procedure set_dirctrl_filter( atom ctrl, object filter )
    c_proc( WX_SET_DIRCTRL_FILTER, {ctrl,filter})
end procedure

--/topic wxDirCtrl
--/proc set_dirctrl_filter_index( atom ctrl, atom index )
--Sets the current filter index (zero-based).
public procedure set_dirctrl_filter_index( atom ctrl, atom index )
    c_proc( WX_SET_DIRCTRL_FILTER_INDEX, {ctrl,index})
end procedure

--/topic wxDirCtrl
--/proc set_dirctrl_path( atom ctrl, object path )
--Sets the current path.
public procedure set_dirctrl_path( atom ctrl, object path )
    c_proc( WX_SET_DIRCTRL_PATH, {ctrl,path})
end procedure

--/topic wxDirCtrl
--/proc show_dirctrl_hidden_files( atom ctrl, atom show )
--If /i show is true, hidden folders and files will be displayed by the control. If false, they will not be displayed.
public procedure show_dirctrl_hidden_files( atom ctrl, atom show )
    c_proc( WX_SHOW_DIRCTRL_HIDDEN_FILES, {ctrl,show})
end procedure



--wxDirDialog
------------------------------
--/topic wxDirDialog
--/parent Dialogs
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent "
--/li /b "sequence message" [= "Choose a directory"]
--/li integer style [= 0]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li sequence defaultPath [= ""]
--</ul>

--/topic wxDirDialog
--/func get_dir_path( atom dirdlg )
--
--Returns the path of the selected wxDirDialog.
public function get_dir_path( atom dirdlg )
    return c_func( WX_GET_DIR_PATH, {dirdlg} )
end function



--wxDirTraverser
------------------------------
--/topic wxDirTraverser
--/parent wxDir
--/parent Files
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /wxArrayString
--/li function OnFile( /wxDirTraverser this, /wxString fname) [= default callback]
--/li function OnDir( /wxDirTraverser this, /wxString dirname) [= default callback]
--</ul>
--
--The first parameter is required, and is a /wxArrayString.  The list of files
--will be put into this array.  OnFile and OnDir are optional, and are cdecl callbacks
--used by the traverser:
--<ul>
--/li /b OnFile:  This function is called for each file. It may return /b wxDIR_STOP
--to abort traversing (for example, if the file being searched is found) or /b wxDIR_CONTINUE to proceed.
--/li /b OnDir: This function is called for each directory. It may return /b wxDIR_STOP to abort
--traversing completely, /b wxDIR_IGNORE to skip this directory but continue with others or
--/b wxDIR_CONTINUE to enumerate all files and subdirectories in this directory.
--</ul>
--The defaults are files that always return wxDIR_CONTINUE.  OnFile adds each file name
--to the /wxArrayString.



--wxDisplay
-----------------------------------------
--/topic wxDisplay
--/parent Graphics
--/parent Classes
--/info
--Determines the sizes and locations of displays connected to the system.
--
--Displays are zero based.

--/topic wxDisplay
--/func display_change_mode( atom display, object params )
--
--/i params should be a sequence of integers: {width, height, depth, refresh rate}
public function display_change_mode( atom display, object params )
    return c_func( WX_DISPLAY_CHANGE_MODE, {display,params})
end function

--/topic wxDisplay
--/func display_get_client_area( atom display )
--
--The client area is the part of the display available for the normal (non full screen) windows,
-- usually it is the same as GetGeometry but it could be less if there is a taskbar (or equivalent)
--on this display.
--
--Returns the client area of the display: {x, y, width, height}
public function display_get_client_area( atom display )
    return c_func( WX_DISPLAY_GET_CLIENT_AREA, {display})
end function

--/topic wxDisplay
--/func display_get_count( atom display )
--
--Returns the number of available displays.
public function display_get_count( atom display )
    return c_func( WX_DISPLAY_GET_COUNT, {display})
end function

--/topic wxDisplay
--/func display_get_current_mode( atom display )
--
--Returns a wxVideoMode sequence: {width, height, depth, refresh rate}
public function display_get_current_mode( atom display )
    return c_func( WX_DISPLAY_GET_CURRENT_MODE, {display})
end function

--/topic wxDisplay
--/func display_get_from_point( atom display, object params )
--
--Returns the zero-based display number.
public function display_get_from_point( atom display, object params )
    return c_func( WX_DISPLAY_GET_FROM_POINT, {display,params})
end function

--/topic wxDisplay
--/func display_get_from_window( atom display, atom window )
--
--Returns the zero-based display number.
public function display_get_from_window( atom display, atom window )
    return c_func( WX_DISPLAY_GET_FROM_WINDOW, {display,window})
end function

--/topic wxDisplay
--/func display_get_geometry( atom display )
--
--Returns the dimensions of the display.
public function display_get_geometry( atom display )
    return c_func( WX_DISPLAY_GET_GEOMETRY, {display})
end function

--/topic wxDisplay
--/func display_get_modes( atom display )
--
--Returns the available display modes.
public function display_get_modes( atom display )
    return c_func( WX_DISPLAY_GET_MODES, {display})
end function

--/topic wxDisplay
--/func display_get_name( atom display )
--
--Returns the name of the display.
--
--/b Note: not supported on all platforms.
public function display_get_name( atom display )
    return c_func( WX_DISPLAY_GET_NAME, {display})
end function

--/topic wxDisplay
--/func display_is_primary( atom display )
--
--Returns true if the primary display, which is always display 0.
public function display_is_primary( atom display )
    return c_func( WX_DISPLAY_IS_PRIMARY, {display})
end function



--wxEuphoria Developers
------------------------------
--/topic wxEuphoria Developers
--/info
--Information about extending or debugging the library
--
--<h3>Contributing to the library</h3>
--
--I welcome and encourage others to extend the library, improving the wrappers that
--I've already written and wrapping new classes.  If you take a look at the wxWidgets
--documentation, you'll see that there are a LOT of classes, and pretty much all of them
--are available to wxEuphoria, once the proper wrapper is generated.
--
--If you do want to contribute something, please send the source to me. Also, please
--document your wrappers using the same format as the rest of the library, so that it
--can be easily included in the wxEuphoria documentation, which is automatically
--generated from the source files.
--
--<h3>Wrapping HOWTO</h3>
--
--Wrapping additional functions is usually pretty easy.  Take a look at both wxeud.e and
--the *.cpp source files for some examples.  Most of the wxEuphoria C++ source code is
--made up of one-liners that call the appropriate wxWidgets routine.  Here are a few tips:
--<ul>
--/li Declare any pointers to wxWidgets objects as 'int's in the C++ code.  To use these as
--the appropriate C++ object, you usually just have to do a simple cast:
--/code
--((wxObject*)the_object)->SomeRoutine();
--/endcode
--/li Make sure that you put /b WXEUAPI in front of the routine definition (look at the code
--for examples).
--/li If you wrap a new object, the constructor should be wrapped in a function with a name
--of the form "new_wxNewClassName".
--/li If you're returning a pointer (or any integer that could be bigger than a Euphoria integer)
--be sure to use the /b BOX_INT() macro when returning the value.  It will convert the value
--to a Euphoria atom if necessary.
--/li Use the helper functions in wxeu.cpp:
--<ul>
--/li /b get_int( seq, index ): Basic subscript operation ( seq[index] ), and returns
--the value as a C++ int.  It's used for extracting values from a sequence that's passed to
--a C++ routine from Euphoria, and handles any conversion from an atom.
--/li /b get_int( x ): Converts a Euphoria atom to a C++ int.
--/li /b get_string( seq ): Converts a sequence to a wxString.
--/li /b get_string( seq, index): Converts a subscript of a sequence to a wxString.
--/li /b get_sequence( string ): Converts a wxString into a Euphoria sequence
--</ul>
--/li Literal strings in your C++ code should be surrounded with the wxT() macro, which
--is used by wxWidgets to assist with Unicode portability.
--/li Declare any Euphoria objects (i.e., when the wx_define_c_* uses dll:E_OBJECT, dll:E_SEQUENCE, dll:E_ATOM)
--passed to C++ as "object" in the C++ code.
--/li Make sure you dereference the Euphoria objects that are passed to C++ code using the
--wxEuphoria macros defined in wxeu.h.  Not doing so will cause memory leaks, since their
--reference counts will never go to zero.
--<ul>
--/li /b wxDeRef( obj ): Checks to see if the object is a sequence or an atom stored as a double,
--and dereferences it if either condition is true.
--/li /b wxDeRefDS( obj ): Dereferences the object.  Use this only if you know that a variable
--is either a double or a sequence.
--</ul>
--/li Use the wrap.exw utility to update wxeud.e.  It will generate a file named wxeud2.e with
--your additions.  It adds the wx_define_c_* statements, updates the create constants if you
--added any new objects, adds the euphoria wrappers and some simple documentation.  The documentation
--will all be attributed to "topic Auto".  You'll need to replace the "Auto" with whatever the
--correct topic should be.
--</ul>
--
--<h3>Extending wxWidgets with other C++ libararies</h3>
--
--The library can be extended by building additional shared libraries to wrap other features
--or other contributions by the wxWidgets community (such as the wxSTC text editor widget).
--You'll probably want to take a look at how the library is put together, using be.c, wxeu.h
--and wxeu.cpp.  These files provide some useful routines for dealing with euphoria and
--wxWidgets C++ data, such as converting sequences to wxStrings.
--
--You'll need to call /add_create_func() to integrate your wrapper with the wxEuphoria library.

--/topic wxEuphoria Developers
--/func add_create_func( object rid )
--
--This function can be used to extend the library by adding a constructor that can
--be used in euphoria code.  The only parameter /em rid, is the result of a call
--to define_c_func or routine_id. (If passing a routine_id, it must be enclosed as a
--sequence.) The function should take a sequence as a parameter, and should return an
-- atom containing a pointer to the created object.
--
--The return value from add_create_func should be used in calls to /create to create
--instances of your object.
public function add_create_func( object rid )
    integer ix

    ix = find( rid, WX_CREATE_ARRAY )
    if ix then
        return ix
    end if

    WX_CREATE_ARRAY = append( WX_CREATE_ARRAY, rid )
    return length(WX_CREATE_ARRAY)
end function



--wxFileName
------------------------------
--/topic wxFileName
--/parent Files
--/parent Classes
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "sequence path"
--/li pathformat [= wxPATH_NATIVE]
--<ul>
--/li wxPATH_NATIVE
--/li wxPATH_UNIX
--/li wxPATH_MAC
--/li wxPATH_DOS
--/li wxPATH_VMS
--/li wxPATH_BEOS
--/li wxPATH_WIN
--/li wxPATH_OS2
--</ul>
--</ul>
--
--Any of the functions below can take either an existing /b wxFileName object, or a sequence containing a filename.

--/topic wxFileName
--/func get_ext( object filename )
--Returns the file name extension.
public function get_ext( object filename )
    return c_func( WX_GET_EXT, {filename})
end function

--/topic wxFileName
--/func get_forbidden_chars( object filename )
--Returns the characters that can't be used in filenames and directory names for the specified format.
public function get_forbidden_chars( object filename )
    return c_func( WX_GET_FORBIDDEN_CHARS, {filename})
end function

--/topic wxFileName
--/func get_full_name( object filename )
--Returns the full name (including extension but excluding directories).
public function get_full_name( object filename )
    return c_func( WX_GET_FULL_NAME, {filename})
end function

--/topic wxFileName
--/func get_full_path( object filename )
--Returns the full path with name and extension.
public function get_full_path( object filename )
    return c_func( WX_GET_FULL_PATH, {filename})
end function

--/topic wxFileName
--/func get_home_dir( object filename )
--Returns the home directory.
public function get_home_dir( object filename )
    return c_func( WX_GET_HOME_DIR, {filename})
end function

--/topic wxFileName
--/func get_long_path( object filename )
--Return the long form of the path (returns identity on non-Windows platforms)
public function get_long_path( object filename )
    return c_func( WX_GET_LONG_PATH, {filename})
end function

--/topic wxFileName
--/func get_name( object filename )
--Returns the name part of the filename (without extension).
--See also:
--/get_full_name()
public function get_name( object filename )
    return c_func( WX_GET_NAME, {filename})
end function

--/topic wxFileName
--/func get_short_path( object filename )
--Return the short form of the path (returns identity on non-Windows platforms).
public function get_short_path( object filename )
    return c_func( WX_GET_SHORT_PATH, {filename})
end function

--/topic wxFileName
--/func get_modified_time( object filename )
--Returns the last time the file was last modified.
public function get_modified_time( object filename )
    return c_func( WX_GET_MODIFIED_TIME, {filename})
end function

--/topic wxFileName
--/func get_size( object filename )
--Returns the size of this file (wxFileSize object) or the size of the given file (sequence).
-- If the file does not exist or its size could not be read (because e.g. the file is locked
-- by another process) the returned value is /b wxInvalidSize.
public function get_size( object filename )
    return c_func( WX_GET_SIZE, {filename})
end function



--wxFlexGridSizer
------------------------------
--/topic wxFlexGridSizer
--/parent Sizers
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "integer rows"
--/li /b "integer cols"
--/li integer vgap [= 0]
--/li integer hgap [= 0]
--</ul>
--
--A flex grid sizer is a sizer which lays out its children in a two-dimensional table with all table
-- fields in one row having the same height and all fields in one column having the same width, but
-- all rows or all columns are not necessarily the same height or width as in the /wxGridSizer.



--wxFont
------------------------------
--/topic wxFont
--/parent Graphics
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom pointSize"
--/li /b "atom family" (see /set_font_family)
--/li /b "atom style" (see /set_font_style)
--/li /b "atom weight" (see /set_font_weight)
--/li integer underline [= 0]
--/li sequence faceName [= ""]
--/li atom encoding [= wxFONTENCODING_DEFAULT] (see /set_default_font_encoding)
--</ul>

--/topic wxFont
--/func get_font_encoding( atom font )
--
--Returns the encoding of the specified font.
public function get_font_encoding( atom font )
    return c_func( WX_GET_FONT_ENCODING, {font} )
end function

--/topic wxFont
--/func get_font_name( atom font )
--
--Returns the face name of the specified font.
public function get_font_name( atom font )
    return c_func( WX_GET_FONT_NAME, {font} )
end function

--/topic wxFont
--/func get_font_underlined( atom font )
--
--Returns 1 if the specified font is underlined, 0 otherwise.
public function get_font_underlined( atom font )
    return c_func( WX_GET_FONT_UNDERLINED, {font} )
end function

--/topic wxFont
--/func get_font_weight( atom font )
--
--Returns the weight of the specified font.
public function get_font_weight( atom font )
    return c_func( WX_GET_FONT_WEIGHT, {font} )
end function

--/topic wxFont
--/func get_font_style( atom font )
--
--Returns the style of the specified font.
public function get_font_style( atom font )
    return c_func( WX_GET_FONT_STYLE, {font} )
end function

--/topic wxFont
--/func get_font_pointsize( atom font )
--
--Returns the point size for the specified font.
public function get_font_pointsize( atom font )
    return c_func( WX_GET_FONT_POINTSIZE, {font} )
end function

--/topic wxFont
--/func get_font_family( atom font )
--
--Returns the font family for the specified font.
public function get_font_family( atom font )
    return c_func( WX_GET_FONT_FAMILY, {font} )
end function

--/topic wxFont
--/func get_font_members( atom font )
--
--Returns the font data in a sequence that can be used to create
--a new font:
--/code
-- { pointSize, family, style, weight, underlined, faceName, encoding }
--/endcode
public function get_font_members( atom font )
    return c_func( WX_GET_FONT_MEMBERS, {font} )
end function

--/topic wxFont
--/proc set_point_size( atom font, integer points )
--
--This sets the point size of the designated font to the value specified by points.
public procedure set_point_size( atom font, integer points )
    c_proc( WX_SET_POINT_SIZE, {font, points} )
end procedure

--/topic wxFont
--/func get_point_size( atom font )
--
--Gets the point size.
public function get_point_size( atom font )
    return c_func( WX_GET_POINT_SIZE, {font} )
end function

--/topic wxFont
--/proc set_default_font_encoding( atom font, atom encoding )
--
--Sets the default font encoding, which can be one of:
--<ul><table>
--<tr><td> /li wxFONTENCODING_SYSTEM <td> system default
--<tr><td> /li wxFONTENCODING_DEFAULT <td> current default encoding
--<tr><td> /li wxFONTENCODING_ISO8859_1 <td> West European (Latin1)
--<tr><td> /li wxFONTENCODING_ISO8859_2 <td> Central and East European (Latin2)
--<tr><td> /li wxFONTENCODING_ISO8859_3 <td> Esperanto (Latin3)
--<tr><td> /li wxFONTENCODING_ISO8859_4 <td> Baltic (old) (Latin4)
--<tr><td> /li wxFONTENCODING_ISO8859_5 <td> Cyrillic
--<tr><td> /li wxFONTENCODING_ISO8859_6 <td> Arabic
--<tr><td> /li wxFONTENCODING_ISO8859_7 <td> Greek
--<tr><td> /li wxFONTENCODING_ISO8859_8 <td> Hebrew
--<tr><td> /li wxFONTENCODING_ISO8859_9 <td> Turkish (Latin5)
--<tr><td> /li wxFONTENCODING_ISO8859_10 <td> Variation of Latin4 (Latin6)
--<tr><td> /li wxFONTENCODING_ISO8859_11 <td> Thai
--<tr><td> /li wxFONTENCODING_ISO8859_12 <td> doesn't exist currently but put it
--<tr><td> /li wxFONTENCODING_ISO8859_13 <td> Baltic (Latin7)
--<tr><td> /li wxFONTENCODING_ISO8859_14 <td> Latin8
--<tr><td> /li wxFONTENCODING_ISO8859_15 <td> Latin9 (a.k.a. Latin0 includes euro)
--<tr><td> /li wxFONTENCODING_ISO8859_MAX <td>
--<tr><td> /li wxFONTENCODING_KOI8 <td> we don't support any of KOI8 variants
--<tr><td> /li wxFONTENCODING_ALTERNATIVE <td> same as MS-DOS CP866
--<tr><td> /li wxFONTENCODING_BULGARIAN <td> used under Linux in Bulgaria
--<tr><td> /li wxFONTENCODING_CP437 <td> original MS-DOS codepage
--<tr><td> /li wxFONTENCODING_CP850 <td> CP437 merged with Latin1
--<tr><td> /li wxFONTENCODING_CP852 <td> CP437 merged with Latin2
--<tr><td> /li wxFONTENCODING_CP855 <td> another cyrillic encoding
--<tr><td> /li wxFONTENCODING_CP866 <td> and another one
--<tr><td> /li wxFONTENCODING_CP874 <td> WinThai
--<tr><td> /li wxFONTENCODING_CP932 <td> Japanese (shift-JIS)
--<tr><td> /li wxFONTENCODING_CP936 <td> Chinese simplified (GB)
--<tr><td> /li wxFONTENCODING_CP949 <td> Korean (Hangul charset)
--<tr><td> /li wxFONTENCODING_CP950 <td> Chinese (traditional - Big5)
--<tr><td> /li wxFONTENCODING_CP1250 <td> WinLatin2
--<tr><td> /li wxFONTENCODING_CP1251 <td> WinCyrillic
--<tr><td> /li wxFONTENCODING_CP1252 <td> WinLatin1
--<tr><td> /li wxFONTENCODING_CP1253 <td> WinGreek (8859-7)
--<tr><td> /li wxFONTENCODING_CP1254 <td> WinTurkish
--<tr><td> /li wxFONTENCODING_CP1255 <td> WinHebrew
--<tr><td> /li wxFONTENCODING_CP1256 <td> WinArabic
--<tr><td> /li wxFONTENCODING_CP1257 <td> WinBaltic (same as Latin 7)
--<tr><td> /li wxFONTENCODING_CP12_MAX <td>
--<tr><td> /li wxFONTENCODING_UTF7 <td> UTF-7 Unicode encoding
--<tr><td> /li wxFONTENCODING_UTF8 <td> UTF-8 Unicode encoding
--<tr><td> /li wxFONTENCODING_GB2312 <td> Simplified Chinese
--<tr><td> /li wxFONTENCODING_BIG5 <td> Traditional Chinese
--<tr><td> /li wxFONTENCODING_SHIFT_JIS  <td> Shift JIS
--<tr><td> /li wxFONTENCODING_EUC_JP  <td> Extended Unix Codepage for Japanese
--<tr><td> /li wxFONTENCODING_UNICODE<td> Unicode
--</table></ul>
public procedure set_default_font_encoding( atom font, atom encoding )
    c_proc( WX_SET_DEFAULT_FONT_ENCODING, {font,encoding})
end procedure

--/topic wxFont
--/proc set_font_face_name( atom font, sequence face_name )
--
--Sets the facename for the font.
--
--To avoid portability problems, don't rely on a specific face, but specify the
--font family instead or as well. A suitable font will be found on the end-user's
--system. If both the family and the facename are specified, wxWidgets will first
--search for the specific face, and then for a font belonging to the same family.
public procedure set_font_face_name( atom font, sequence face_name )
    c_proc( WX_SET_FONT_FACE_NAME, {font,face_name})
end procedure

--/topic wxFont
--/proc set_font_familty( atom font, atom family )
--
--Sets the font family, which must be one of:
--<ul><table>
--<tr><td> /li wxFONTFAMILY_DEFAULT <td>Chooses a default font.
--<tr><td> /li wxFONTFAMILY_DECORATIVE <td>A decorative font.
--<tr><td> /li wxFONTFAMILY_ROMAN <td>A formal, serif font.
--<tr><td> /li wxFONTFAMILY_SCRIPT <td>A handwriting font.
--<tr><td> /li wxFONTFAMILY_SWISS <td>A sans-serif font.
--<tr><td> /li wxFONTFAMILY_MODERN <td>A fixed pitch font.
--<tr><td> /li wxFONTFAMILY_TELETYPE <td>A teletype font.
--</table></ul>
public procedure set_font_familty( atom font, atom family )
    c_proc( WX_SET_FONT_FAMILTY, {font,family})
end procedure

--/topic wxFont
--/func set_native_font_info( atom font, sequence info )
--
--Creates the font corresponding to the given native font description string and returns
--true if the creation was successful. which must have been previously returned by
--/get_native_font_info(). If the string is invalid, font is unchanged. This function is
--typically used for de-serializing a wxFont object previously saved in a string-form.
public function set_native_font_info( atom font, sequence info )
    return c_func( WX_SET_NATIVE_FONT_INFO, {font,info})
end function

--/topic wxFont
--/func get_native_font_info( atom font )
--
--Returns the platform-dependent string completely describing this font. Returned
--string is always non-empty. Note that the returned string is not meant to be shown
--or edited by the user: a typical use of this function is for serializing in
--string-form a wxFont object.
public function get_native_font_info( atom font )
    return c_func( WX_GET_NATIVE_FONT_INFO, {font})
end function

--/topic wxFont
--/func get_native_font_info_user_desc( atom font )
--
--Returns a user-friendly string for this font object. Returned string is always non-empty. Some
--examples of the formats of returned strings (which are platform-dependent) are in
--/set_native_font_info_user_desc().
public function get_native_font_info_user_desc( atom font )
    return c_func( WX_GET_NATIVE_FONT_INFO_USER_DESC, {font})
end function

--/topic wxFont
--/func set_native_font_info_user_desc( atom font, sequence info )
--
--Creates the font corresponding to the given native font description string and returns
--true if the creation was successful. Unlike SetNativeFontInfo, this function accepts
--strings which are user-friendly. Examples of accepted string formats are:
--<ul>
--/li on /b wxGTK2: [FACE-NAME] [bold] [oblique|italic] [POINTSIZE]  Monospace bold 10
--/li on /b wxMSW: [light|bold] [italic] [FACE-NAME] [POINTSIZE] [ENCODING]  Tahoma 10 WINDOWS-1252
--</ul>
public function set_native_font_info_user_desc( atom font, sequence info )
    return c_func( WX_SET_NATIVE_FONT_INFO_USER_DESC, {font,info})
end function

--/topic wxFont
--/proc set_font_style( atom font, atom style )
--
--Sets the font style, which must be one of:
--<ul>
--/li wxFONTSTYLE_NORMAL
--/li wxFONTSTYLE_SLANT
--/li wxFONTSTYLE_ITALIC.
--</ul>
public procedure set_font_style( atom font, atom style )
    c_proc( WX_SET_FONT_STYLE, {font,style})
end procedure

--/topic wxFont
--/proc set_font_underlined( atom font, atom underlined )
--
--Sets underlining.
public procedure set_font_underlined( atom font, atom underlined )
    c_proc( WX_SET_FONT_UNDERLINED, {font,underlined})
end procedure

--/topic wxFont
--/proc set_font_weight( atom font, atom weight )
--
--Sets the font weight, which must be one of:
--<ul>
--/li /b wxFONTWEIGHT_NORMAL  Normal font.
--/li /b wxFONTWEIGHT_LIGHT  Light font.
--/li /b wxFONTWEIGHT_BOLD  Bold font.
--</ul>
public procedure set_font_weight( atom font, atom weight )
    c_proc( WX_SET_FONT_WEIGHT, {font,weight})
end procedure



--wxFontData
------------------------------
--/topic wxFontData
--/parent wxFontDialog
--/parent Fonts
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li atom /wxFont [= 0]
--</ul>
--
--If a /wxFont is passed, /set_initial_font() is called for
-- you automatically. If not, you need to call it manually.


--/topic wxFontData
--/proc set_initial_font( atom fdata, atom font )
--
--Sets the font that will be initially used by the font dialog.
public procedure set_initial_font( atom fdata, atom font )
    c_proc( WX_SET_INITIAL_FONT, {fdata, font} )
end procedure

--/topic wxFontData
--/func get_chosen_font( atom fdata )
--
--Gets the font chosen by the user if the user pressed OK ( /font_selector returned non-zero ).
public function get_chosen_font( atom fdata )
    return c_func( WX_GET_CHOSEN_FONT, {fdata} )
end function



--wxFontDialog
------------------------------
--/topic wxFontDialog
--/parent Dialogs
--/parent Fonts
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li /b "atom initial" (a /wxFontData object)
--</ul>

--/topic wxFontDialog
--/func get_font_data( atom dialog )
--
--Returns the font data associated with the font dialog.
public function get_font_data( atom dialog )
    return c_func( WX_GET_FONT_DATA, {dialog} )
end function



--wxFontEnumerator
------------------------------
--/topic wxFontEnumerator
--/parent wxFont
--/parent Fonts
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b None
--</ul>
--
--/b wxFontEnumerator enumerates either all available fonts on the
--system or only the ones with given attributes - either only fixed-width
--(suited for use in programs such as terminal emulators and the like) or the
--fonts available in the given encoding.
--
--/b Note: the routines for /b wxFontEnumerator create and destroy an enumeration object
--automatically, so there is no need to create your own.

--/topic wxFontEnumerator
--/func get_facenames( sequence params )
--
--Returns a sequence of font face names available.  If /i params is empty,
--then all available fonts are returned.  Otherwise, /i params should
--be a one or two element sequence:
--<ul>
--/li /b encoding [=wxFONTENCODING_SYSTEM]
--/li /b Fixed_Width_Only [=wxFalse]
--</ul>
public function get_facenames( sequence params )
    return c_func( WX_GET_FACENAMES, {params} )
end function

--/topic wxFontEnumerator
--/func get_encodings( sequence fontname )
--
--Returns a sequence of available encodings for the specified font face.
--Pass an empty sequence for /i fontname to get all available encodings.
public function get_encodings( sequence fontname )
    return c_func( WX_GET_ENCODINGS, {fontname} )
end function

--/topic wxFontEnumerator
--/func get_facesizes( object params )
--
--Returns a sequence containing a sequence of font face names and a corresponding sequence
--holding the sizes.  Note that under GTK, the sizes will be empty unless the font is
--a bitmap font.
public function get_facesizes( object params )
    return c_func( WX_GET_FACESIZES, {params})
end function



--wxFrame
------------------------------
--/topic wxFrame
--/parent Controls
--/parent wxWindow
--/info
--/b wxFrame is a class derived from /wxWindow.  You must use a /b wxFrame as your top level window.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li /b "integer id"
--/li /b "sequence title"
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= wxDEFAULT_FRAME_STYLE]
--<ul><table>
--<tr><td> /li wxSTAY_ON_TOP           <td>Keep the window on top
--<tr><td> /li wxICONIZE               <td>&nbsp;
--<tr><td> /li wxMINIMIZE              <td>&nbsp;
--<tr><td> /li wxMAXIMIZE              <td>&nbsp;
--<tr><td> /li wxSYSTEM_MENU           <td>Show the system menu
--<tr><td> /li wxMINIMIZE_BOX          <td>Show the minimize box
--<tr><td> /li wxMAXIMIZE_BOX          <td>Show the maximize box
--<tr><td> /li wxTINY_CAPTION_HORIZ    <td>&nbsp;
--<tr><td> /li wxTINY_CAPTION_VERT     <td>&nbsp;
--<tr><td> /li wxRESIZE_BORDER         <td>Allow the window to be resized
--<tr><td> /li wxDIALOG_NO_PARENT      <td>Don't make owned by apps top window
--<tr><td> /li wxFRAME_NO_TASKBAR      <td>No taskbar button (MSW only)
--<tr><td> /li wxFRAME_TOOL_WINDOW     <td>No taskbar button, no system menu
--<tr><td> /li wxFRAME_FLOAT_ON_PARENT <td>Always above its parent
--<tr><td> /li wxFRAME_SHAPED          <td>Create a window that is able to be shaped
--</table></ul>
--</ul>
--
--All items with a default value are optional, although you must supply both x and y, or both
--cx and cy if you supply either one.
--
--Example:
--/code
--  constant myFrame = create( wxFrame, {0,         -- has no parent
--                                      -1,         -- wxWidgets will create an id
--                                      "myFrame",  -- will be dispayed as the caption
--                                      -1,         -- start at default x position
--                                      -1,         -- start at default y position
--                                      300,        -- 300 wide
--                                      200})       -- 200 tall
--                                                  -- no style supplied means will be default style
--/endcode
--
--If there is only one child of a frame, that child will be automatically resized
-- (by wxWidgets) to the client rect of the frame.

--/topic wxFrame
--/proc show_fullscreen( atom frame, integer show, integer flags )
--
--<ul>
--/li /b show 1 to make the frame full screen, 0 to restore to normal
--/li /b flags should be one or more of the following /wx_or_all()'d together:
--<ul>
--/li    wxFULLSCREEN_NOMENUBAR
--/li    wxFULLSCREEN_NOTOOLBAR
--/li    wxFULLSCREEN_NOSTATUSBAR
--/li    wxFULLSCREEN_NOBORDER
--/li    wxFULLSCREEN_NOCAPTION
--/li    wxFULLSCREEN_ALL (all of the above)
--</ul>
--</ul>
--The wxWidgets documentation warns that this function hasn't been tested on
--/wxMDIParentFrame s, but it seems to work fine under Windows.
public procedure show_fullscreen( atom frame, integer show, integer flags )
    c_proc( WX_SHOW_FULLSCREEN, {frame, show, flags} )
end procedure

--/topic wxFrame
--/proc show_window( atom window, integer show )
--
--Sets a frame invisible if show is 0. If show is 1, then it makes the window visible if it is not,
--otherwise it does nothing.
public procedure show_window( atom window, integer show )
    c_proc( WX_SHOW_WINDOW, {window, show} )
end procedure

--/topic wxFrame
--/proc raise_window( atom window )
--
--Bring a window to the front of the screen.
public procedure raise_window( atom window )
    c_proc( WX_RAISE_WINDOW, {window} )
end procedure

--/topic wxFrame
--/proc set_icon( atom frame, atom icon )
--
--This sets the icon beside of the text in a wxFrame. The image type
--cannot be a BM_XPM and it is suggested that a wxIcon be used for
--the image. The image should be 16 by 16.
public procedure set_icon( atom frame, atom icon )
    c_proc( WX_SET_ICON, {frame, icon} )
end procedure

--/topic wxFrame
--/func get_menu_bar( atom frame )
--
public function get_menu_bar( atom frame )
    return c_func( WX_GET_MENU_BAR, {frame})
end function

--/topic wxFrame
--/func get_status_bar( atom frame )
--
public function get_status_bar( atom frame )
    return c_func( WX_GET_STATUS_BAR, {frame})
end function

--/topic wxFrame
--/func get_status_bar_pane( atom frame )
--
public function get_status_bar_pane( atom frame )
    return c_func( WX_GET_STATUS_BAR_PANE, {frame})
end function

--/topic wxFrame
--/func get_tool_bar( atom frame )
--
public function get_tool_bar( atom frame )
    return c_func( WX_GET_TOOL_BAR, {frame})
end function



--wxFTP
------------------------------
--/topic wxFTP
--/parent Networking
--/parent Classes
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b None
--</ul>

--/topic wxFTP
--/func ftp_abort( atom ftp )
--
--Aborts the download currently in process, returns true if ok, false if an error occurred.
public function ftp_abort( atom ftp )
    return c_func( WX_FTP_ABORT, {ftp})
end function

--/topic wxFTP
--/proc ftp_user( atom ftp, sequence user )
--
--Sets the user name to be sent to the FTP server to be allowed to log in.
--
--The default value of the user name is "anonymous".
public procedure ftp_user( atom ftp, sequence user )
    c_proc( WX_FTP_USER, {ftp, user} )
end procedure

--/topic wxFTP
--/proc ftp_password( atom ftp, sequence pwd )
--
--Sets the password to be sent to the FTP server to be allowed to log in.
--
--The default value of the user name is your email address. For example, it could be "username@userhost.domain".
-- This password is built by getting the current user name and the host name of the local machine from the system.
public procedure ftp_password( atom ftp, sequence pwd )
    c_proc( WX_FTP_PASSWORD, {ftp, pwd} )
end procedure

--/topic wxFTP
--/proc ftp_mode( atom ftp, integer mode )
--
--Sets the transfer mode.  /i mode can be either /b ASCII or /b BINARY.
public procedure ftp_mode( atom ftp, integer mode )
    c_proc( WX_FTP_MODE, {ftp, mode} )
end procedure

--/topic wxFTP
--/func ftp_input_stream( atom ftp, sequence path )
--
--Creates a new input stream on the specified path. You can use all but the seek functionality of wxStream.
-- Seek isn't available on all streams. For example, HTTP or FTP streams do not deal with it. Other functions
-- like Tell are not available for this sort of stream, at present. You will be notified when the EOF
-- is reached by an error.
--
--Returns NULL if an error occurred (it could be a network failure or the fact that the file doesn't exist).
--
--Returns the initialized stream. You will have to delete it yourself when you don't need it anymore. The
-- destructor closes the DATA stream connection but will leave the COMMAND stream connection opened. It means
-- that you can still send new commands without reconnecting.
public function ftp_input_stream( atom ftp, sequence path )
    return c_func( WX_FTP_INPUT_STREAM, {ftp, path} )
end function

--/topic wxFTP
--/func ftp_output_stream( atom ftp, sequence path )
--
--Initializes an output stream to the specified file. The returned stream has all but the seek functionality
-- of wxStream. When the user finishes writing data, he has to delete the stream to close it.
public function ftp_output_stream( atom ftp, sequence path )
    return c_func( WX_FTP_OUTPUT_STREAM, {ftp, path} )
end function

--/topic wxFTP
--/func ftp_rm( atom ftp, sequence path )
--
--Delete the file specified by path. Returns TRUE if successful.
public function ftp_rm( atom ftp, sequence path )
    return c_func( WX_FTP_RM, {ftp, path} )
end function

--/topic wxFTP
--/func ftp_chdir( atom ftp, sequence path )
--
--Change the current FTP working directory. Returns true if successful.
public function ftp_chdir( atom ftp, sequence path )
    return c_func( WX_FTP_CHDIR, {ftp, path} )
end function

--/topic wxFTP
--/func ftp_mkdir( atom ftp, sequence path )
--
public function ftp_mkdir( atom ftp, sequence path )
    return c_func( WX_FTP_MKDIR, {ftp, path} )
end function

--/topic wxFTP
--/func ftp_rmdir( atom ftp, sequence path )
--
--Create the specified directory in the current FTP working directory. Returns true if successful.
public function ftp_rmdir( atom ftp, sequence path )
    return c_func( WX_FTP_RMDIR, {ftp, path} )
end function

--/topic wxFTP
--/func ftp_rename( atom ftp, sequence source, sequence dest )
--
--Rename the specified src element to dst. Returns true if successful.
public function ftp_rename( atom ftp, sequence source, sequence dest )
    return c_func( WX_FTP_RENAME, {ftp, source, dest} )
end function

--/topic wxFTP
--/proc ftp_pwd( atom ftp )
--
--Returns the current FTP working directory.
public function ftp_pwd( atom ftp )
    return c_func( WX_FTP_PWD, {ftp} )
end function

--/topic wxFTP
--/func ftp_connect( atom ftp, object host )
--Connects to a server using the specified address /i host.
public function ftp_connect( atom ftp, object host )
    return c_func( WX_FTP_CONNECT, {ftp,host})
end function

--/topic wxFTP
--/func ftp_close( atom ftp )
--Shuts down the connection, disabling further transmission and reception of data.
public function ftp_close( atom ftp )
    return c_func( WX_FTP_CLOSE, {ftp})
end function

--/topic wxFTP
--/proc ftp_passive( atom ftp, atom passive )
--If /i passive is true, passive connection to the FTP server is used.
-- This is the default as it works with practically all firewalls.
--If the server doesn't support passive move, you may call this function
-- with false argument to use active connection.
public procedure ftp_passive( atom ftp, atom passive )
    c_proc( WX_FTP_PASSIVE, {ftp,passive})
end procedure

--/topic wxFTP
--/func ftp_file_exists( atom ftp, object filename )
--Returns true if the given remote file exists, false otherwise.
public function ftp_file_exists( atom ftp, object filename )
    return c_func( WX_FTP_FILE_EXISTS, {ftp,filename})
end function

--/topic wxFTP
--/func ftp_file_size( atom ftp, object filename )
--Returns the file size in bytes or -1 if the file doesn't exist or the size couldn't be determined.
-- Notice that this size can be approximative size only and shouldn't be used for allocating the buffer
-- in which the remote file is copied, for example.
public function ftp_file_size( atom ftp, object filename )
    return c_func( WX_FTP_FILE_SIZE, {ftp,filename})
end function

--/topic wxFTP
--/func ftp_files( atom ftp, sequence wildcard )
--This function returns the computer-parsable list of the files in the current directory
-- (optionally only of the files matching the /i wildcard, all files by default). This list
-- always has the same format and contains one full (including the directory path) file name per line.
public function ftp_files( atom ftp, sequence wildcard )
    return c_func( WX_FTP_FILES, {ftp, wildcard} )
end function

--/topic wxFTP
--/func ftp_dir( atom ftp, sequence wildcard )
--The GetList function is quite low-level. It returns the list of the files in the current directory.
-- The list can be filtered using the wildcard string. If /i wildcard is empty (default), it will return all files in directory.
--
--The form of the list can change from one peer system to another. For example, for a UNIX peer system, it will look like this:
--
--/code
---r--r--r--     1 guilhem  lavaux      12738 Jan 16 20:17 cmndata.cpp
---r--r--r--     1 guilhem  lavaux      10866 Jan 24 16:41 config.cpp
---rw-rw-rw-     1 guilhem  lavaux      29967 Dec 21 19:17 cwlex_yy.c
---rw-rw-rw-     1 guilhem  lavaux      14342 Jan 22 19:51 cwy_tab.c
---r--r--r--     1 guilhem  lavaux      13890 Jan 29 19:18 date.cpp
---r--r--r--     1 guilhem  lavaux       3989 Feb  8 19:18 datstrm.cpp
--/endcode
--
--But on Windows system, it will look like this:
--
--/code
--winamp~1 exe  520196 02-25-1999  19:28  winamp204.exe
--      1 file(s)           520 196 bytes
--/endcode
public function ftp_dir( atom ftp, sequence wildcard )
    return c_func( WX_FTP_DIR, {ftp, wildcard} )
end function

--/topic wxFTP
--/func ftp_last_result( atom ftp )
--Returns the last command result, i.e. the full server reply for the last command.
public function ftp_last_result( atom ftp )
    return c_func( WX_FTP_LAST_RESULT, {ftp} )
end function

--/topic wxFTP
--/func ftp_send_command( atom ftp, sequence command )
--Send the specified /i command to the FTP server and return the first character of the return code.
public function ftp_send_command( atom ftp, sequence command )
    return c_func( WX_FTP_SEND_COMMAND, {ftp, command} )
end function

--/topic wxFTP
--/func ftp_check_command( atom ftp, sequence command, atom ret )
--Send the specified /i command to the FTP server. /i ret specifies the expected result.
public function ftp_check_command( atom ftp, sequence command, atom ret )
    return c_func( WX_FTP_SEND_COMMAND, {ftp, command, ret} )
end function



--wxGauge
-----------------------------------------
--/topic wxGauge
--/parent Controls
--/info
--A gauge is a horizontal or vertical bar which shows a quantity (often time).
--
--/b wxGauge supports two working modes: determinate and indeterminate progress.
--
--The first is the usual working mode (see /set_value and /set_range) while the second can be used when
-- the program is doing some processing but you don't know how much progress is being done. In this case,
-- you can periodically call the Pulse function to make the progress bar switch to indeterminate mode
-- (graphically it's usually a set of blocks which move or bounce in the bar control).
--
--/b wxGauge supports dynamic switch between these two work modes.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li /b "integer id"
--/li /b "atom range"
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= -1]
--/li sequence name [= "gauge"]
--</ul>

--/topic wxGauge
--/func get_bezel_face( atom gauge )
--Returns the width of the 3D bezel face.
--
--/b Remarks
--
--This method is not implemented (returns 0) for most platforms.
public function get_bezel_face( atom gauge )
    return c_func( WX_GET_BEZEL_FACE, {gauge})
end function

--/topic wxGauge
--/func get_range( atom gauge )
--Returns the maximum position of the gauge.
public function get_range( atom gauge )
    return c_func( WX_GET_RANGE, {gauge})
end function

--/topic wxGauge
--/func get_shadow_width( atom gauge )
--Returns the 3D shadow margin width.
--
--/b Remarks
--
--This method is not implemented (returns 0) for most platforms.
public function get_shadow_width( atom gauge )
    return c_func( WX_GET_SHADOW_WIDTH, {gauge})
end function

--/topic wxGauge
--/func get_value( atom gauge )
--Returns the current position of the gauge.
public function get_value( atom gauge )
    return c_func( WX_GET_VALUE, {gauge})
end function

--/topic wxGauge
--/func is_gauge_vertical( atom gauge )
--Returns true if the gauge is vertical (has wxGA_VERTICAL style) and false otherwise.
public function is_gauge_vertical( atom gauge )
    return c_func( WX_IS_GAUGE_VERTICAL, {gauge})
end function

--/topic wxGauge
--/proc set_bezel_face( atom gauge, atom width )
--Sets the 3D bezel face width.
--
--/b Remarks
--
--This method is not implemented (doesn't do anything) for most platforms.
public procedure set_bezel_face( atom gauge, atom width )
    c_proc( WX_SET_BEZEL_FACE, {gauge,width})
end procedure

--/topic wxGauge
--/proc set_range( atom gauge, atom range )
--Sets the range (maximum value) of the gauge.
--
--This function makes the gauge switch to determinate mode, if it's not already.
public procedure set_range( atom gauge, atom range )
    c_proc( WX_SET_RANGE, {gauge,range})
end procedure

--/topic wxGauge
--/proc set_shadow_width( atom gauge, atom width )
--Sets the 3D shadow width.
--
--/b Remarks
--
--This method is not implemented (doesn't do anything) for most platforms.
public procedure set_shadow_width( atom gauge, atom width )
    c_proc( WX_SET_SHADOW_WIDTH, {gauge,width})
end procedure

--/topic wxGauge
--/proc set_value( atom gauge, atom pos )
--Sets the position of the gauge. This function makes the gauge switch to determinate mode, if it's not already.
--
--/b Parameters
--
--/li ipos
--
--Position for the gauge level.
public procedure set_value( atom gauge, atom pos )
    c_proc( WX_SET_VALUE, {gauge,pos})
end procedure

--/topic wxGauge
--/proc pulse( atom gauge )
--Switch the gauge to indeterminate mode (if required) and makes the gauge move a bit to indicate the
-- user that some progress has been made.
--
--Note that after calling this function the value returned by GetValue is undefined and thus you need
-- to explicitely call SetValue if you want to restore the determinate mode.
public procedure pulse( atom gauge )
    c_proc( WX_PULSE, {gauge})
end procedure



--wxGrid
------------------------------
--/topic wxGrid
--/parent Controls
--/info
--This is a grid control that can be used to display tabular data.  It also allows
--editing in place by users, similar to a spreadsheet.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= 0]
--/li integer rows [= 10]
--/li integer cols [= 10]
--</ul>
--It is important to remember that all wxGrid functions and procedures use zero-based
--references to all cells (0, 1, 2,...)

--/topic wxGrid
--/func get_cell_value( atom grid, integer row, integer col )
--/b Returns: string containing the value
--
--If the cell contains numeric data, you will need to use value(), since all values
--returned by get_cell_value() are strings (this is how /wxGrid stores everything).
public function get_cell_value( atom grid, atom row, atom col )
    return c_func( WX_GET_CELL_VALUE, {grid, row, col} )
end function

--/topic wxGrid
--/proc append_cols( atom grid, integer cols )
--
--Appends /i cols columns to the grid.
public procedure append_cols( atom grid, integer cols )
    c_proc( WX_APPEND_COLS, {grid, cols} )
end procedure

--/topic wxGrid
--/proc append_rows( atom grid, integer rows )
--
--Appends /i rows rows to the grid.
public procedure append_rows( atom grid, integer rows )
    c_proc( WX_APPEND_ROWS, {grid, rows} )
end procedure

--/topic wxGrid
--/proc set_cell_value( atom grid, object val, integer row, integer col )
--
--Sets the contents of a cell.  If /i val is an atom, then the contents of
--the cell will be sprint(val).
public procedure set_cell_value( atom grid, object val, integer row, integer col )
    if atom(val) then
        val = text:sprint( val )
    end if
    c_proc( WX_SET_CELL_VALUE, {grid, val, row, col} )
end procedure

--/topic wxGrid
--/proc delete_rows( atom grid, integer pos, integer n, integer relabel )
--
--Deletes /i n rows from the grid at position /i pos. If /i relabel is TRUE, the function
--OnChangeLabels is called to give the application the opportunity to relabel.
public procedure delete_rows( atom grid, integer pos, integer n, integer relabel )
    c_proc( WX_DELETE_ROWS, {grid, pos, n, relabel} )
end procedure

--/topic wxGrid
--/proc delete_cols( atom grid, integer pos, integer n, integer relabel )
--
--Deletes /i n columns from the grid at position /i pos. If /i relabel is TRUE, the function
--OnChangeLabels is called to give the application the opportunity to relabel.
public procedure delete_cols( atom grid, integer pos, integer n, integer relabel )
    c_proc( WX_DELETE_COLS, {grid, pos, n, relabel} )
end procedure

--/topic wxGrid
--/func grid_size( atom grid )
--/b Returns: Sequence {rows,cols} indicating the current size of the grid.
public function grid_size( atom grid )
    return c_func( WX_GRID_SIZE, {grid} )
end function

--/topic wxGrid
--/proc set_col_label( atom grid, integer col, sequence text )
--
--Changes the label of a column.
public procedure set_col_label( atom grid, integer pos, sequence text )
    c_proc( WX_SET_COL_LABEL, {grid, pos, text} )
end procedure

--/topic wxGrid
--/proc set_row_label( atom grid, integer row, sequence text )
--
--Changes the label of a row.
public procedure set_row_label( atom grid, integer pos, sequence text )
    c_proc( WX_SET_ROW_LABEL, {grid, pos, text} )
end procedure

--/topic wxGrid
--/proc set_row_label_size( atom grid, integer width )
--
--This sets the labels of rows in a wxGrid to be 'width' pixels high.
public procedure set_row_label_size( atom grid, integer size )
    c_proc( WX_SET_ROW_LABEL_SIZE, {grid, size})
end procedure

--/topic wxGrid
--/proc begin_batch( atom grid )
--
--Stops updates of the grid until /end_batch() is called.  This is useful for
--when you need to modify a large number of cells.
public procedure begin_batch( atom grid )
    c_proc( WX_BEGIN_BATCH, {grid} )
end procedure

--/topic wxGrid
--/proc end_batch( atom grid )
--
--Resumes updating of the grid after a call to /begin_batch().  This is useful for
--when you need to modify a large number of cells.
public procedure end_batch( atom grid )
    c_proc( WX_END_BATCH, {grid} )
end procedure

--/topic wxGrid
--/func batch_count( atom grid )
--/b Returns: number of unmatched calls to /begin_batch()
public function batch_count( atom grid )
    return c_func( WX_BATCH_COUNT, {grid} )
end function

--/topic wxGrid
--/func last_cell_visible( atom grid )
--/b Returns: last visible cell in the /i grid
public function last_cell_visible( atom grid )
    return c_func( WX_LAST_CELL_VISIBLE, {grid} )
end function

--/topic wxGrid
--/proc set_grid_selection_mode( atom grid, integer mode )
--
--Set the selection behaviour of the grid.
--
--/i mode can be one of the following:
--<ul>
--/li /i wxGridSelectCells The default mode where individual cells are selected.
--/li /i wxGridSelectRows Selections will consist of whole rows.
--/li /i wxGridSelectColumns Selections will consist of whole columns.
--</ul>
public procedure set_grid_selection_mode( atom grid, integer mode )
    c_proc( WX_SET_GRID_SELECTION_MODE, {grid, mode})
end procedure

--/topic wxGrid
--/proc enable_drag_grid_size( atom grid, integer enable )
--
--Enables or disables row and column resizing by dragging gridlines with the mouse.
public procedure enable_drag_grid_size( atom grid, integer enable )
    c_proc( WX_ENABLE_DRAG_GRID_SIZE, {grid, enable})
end procedure

--/topic wxGrid
--/proc select_grid_block( atom grid, integer top_row, integer left_col, integer bottom_row, integer right_col, integer add_to_selected )
--Selects a rectangular block of cells. If /i add_to_selected is false then any existing selection will be deselected; if true the column will
-- be added to the existing selection.
public procedure select_grid_block( atom grid, integer top_row, integer left_col, integer bottom_row, integer right_col, integer add_to_selected )
    c_proc( WX_SELECT_GRID_BLOCK, {grid, top_row, left_col, bottom_row, right_col, add_to_selected })
end procedure

--/topic wxGrid
--/proc select_grid_row( atom grid, integer row, integer add_to_selected )
--
--Selects the specified row.
public procedure select_grid_row( atom grid, integer row, integer add_to_selected )
    c_proc( WX_SELECT_GRID_ROW, {grid, row, add_to_selected} )
end procedure

--/topic wxGrid
--/proc select_grid_col( atom grid, integer col, integer add_to_selected )
--
--Selects the specified column.
public procedure select_grid_col( atom grid, integer col, integer add_to_selected )
    c_proc( WX_SELECT_GRID_COL, {grid, col, add_to_selected})
end procedure

--/topic wxGrid
--/proc select_grid_all( atom grid )
--
--Selects all cells in the grid.
public procedure select_grid_all( atom grid )
    c_proc( WX_SELECT_GRID_ALL, {grid})
end procedure

--/topic wxGrid
--/func is_in_grid_selection( atom grid, integer row, integer col )
--
--Is this cell currently selected.
public function is_in_grid_selection( atom grid, integer row, integer col )
    return c_func( WX_IS_IN_GRID_SELECTION, {grid, row, col})
end function

--/topic wxGrid
--/proc clear_grid( atom grid )
--
--Deletes all data in the grid.
public procedure clear_grid( atom grid )
    c_proc( WX_CLEAR_GRID, {grid} )
end procedure

--/topic wxGrid
--/func grid_event_cell( atom event )
--/b Returns: {row, col}
--
--Call this inside of /wxGrid event handlers to get the row and column where the event
--is taking place.
public function grid_event_cell( atom event )
    return c_func( WX_GRID_EVENT_CELL, {event} )
end function

--/topic wxGrid
--/func is_cell_visible( atom grid, integer row, integer col, integer wholly_visible )
--
--Returns true if a cell is either wholly visible (the default) or at least partially visible in the grid window.
public function is_cell_visible( atom grid, integer row, integer col, integer wholly_visible )
    return c_func( WX_IS_CELL_VISIBLE, {grid, row, col, wholly_visible } )
end function

--/topic wxGrid
--/proc select_grid_cell( atom grid, atom row, atom col )
--
--Moves the cursor and selection to the cell specified.
public procedure select_grid_cell( atom grid, atom row, atom col )
    c_proc( WX_SELECT_GRID_CELL, {grid, row, col} )
end procedure

--/topic wxGrid
--/func grid_pos( atom grid )
--/b Returns: { cursor_row, cursor_col }
public function grid_pos( atom grid )
    return c_func( WX_GRID_POS, {grid} )
end function

--/topic wxGrid
--/proc set_cell( atom grid, atom row, atom col )
--
--Moves the cursor to the current cell being edited.  Does not
--alter the selection (see /select_grid_cell)
public procedure set_cell( atom grid, atom row, atom col )
    c_proc( WX_SET_CELL, {grid, row, col} )
end procedure

--/topic wxGrid
--/proc make_cell_visible( atom grid, atom row, atom col )
--
--Scrolls the grid if necessary to make the cell visible.
public procedure make_cell_visible( atom grid, atom row, atom col )
    c_proc( WX_MAKE_CELL_VISIBLE, {grid, row, col} )
end procedure

--/topic wxGrid
--/proc set_grid_size( atom grid, atom rows, atom cols )
--
--Adds or deletes columns or rows as required.
public procedure set_grid_size( atom grid, atom rows, atom cols )
    c_proc( WX_SET_GRID_SIZE, {grid, rows, cols} )
end procedure

--/topic wxGrid
--/proc set_col_editor( atom grid, atom col, integer ed_type, object data )
--Allows entries other than the standard text based editor to be used.
--
--The only editors currently supported are:
--<ul>
--/li wxGridCellBoolEditor: /i data is ignored
--/li wxGridCellChoiceEditor: /i data should be a sequence of strings
--/li wxGridCellAutoWrapStringEditor: /i data is ignored
--</ul>
public procedure set_col_editor( atom grid, atom col, integer ed_type, object data )
    c_proc( WX_SET_COL_EDITOR, {grid, col, ed_type, data} )
end procedure

--/topic wxGrid
--/func get_cell_editor( atom grid, sequence cell )
--
--Returns the /wxGridCellEditor for the cell
public function get_cell_editor( atom grid, sequence cell )
    return c_func( WX_GET_CELL_EDITOR, {grid, cell} )
end function

--/topic wxGrid
--/func get_cell_editor_control( atom grid, object editor )
--Returns the Control belonging to the /wxGridCellEditor
--
--If /b editor is a sequence, then it specifies the { row, column } of
--the /wxGridCellEditor to be fetched.  Otherwise, it is a /wxGridCellEditor
--(see /get_cell_editor).
public function get_cell_editor_control( atom grid, object editor )
    return c_func( WX_GET_CELL_EDITOR_CONTROL, {grid, editor} )
end function

--/topic wxGrid
--/proc set_editor_choices( atom grid, object cell, sequence choices )
--
--If /i cell is an integer, then it refers to an entire column, and the
--choices will be set for the column editor.  If /i cell is a sequence,
--it refers to the cell by {row, column}, and that cell's editor will
--be set.
public procedure set_editor_choices( atom grid, object cell, sequence choices )
    c_proc( WX_SET_EDITOR_CHOICES, {grid, cell, choices} )
end procedure

--/topic wxGrid
--/proc set_cell_editor( atom grid, atom row, atom col, integer ed_type, object data )
--Allows entries other than the standard text based editor to be used.
--
--The only editors currently supported are:
--<ul>
--/li wxGridCellBoolEditor: /i data is ignored
--/li wxGridCellChoiceEditor: /i data should be a 2-element sequence:
--<ol>
--/li sequence of strings to be added to the editor
--/li flag to indicate whether items other than the list should be allowed
--</ol>
--</ul>
public procedure set_cell_editor( atom grid, atom row, atom col, integer ed_type, object data )
    c_proc( WX_SET_CELL_EDITOR, {grid, row, col, ed_type, data} )
end procedure

--/topic wxGrid
--/proc set_default_grid_editor( atom grid, atom editor )
--
--Sets the default cell editor control for the grid.
public procedure set_default_grid_editor( atom grid, atom editor )
    c_proc( WX_SET_DEFAULT_GRID_EDITOR, {grid, editor} )
end procedure

--/topic wxGrid
--/proc save_grid_edit_control_value( atom grid )
--
--Sets the value of the current grid cell to the current in-place edit control value. This is called automatically
-- when the grid cursor moves from the current cell to a new cell. It is also a good idea to call this function when
-- closing a grid since any edits to the final cell location will not be saved otherwise.
public procedure save_grid_edit_control_value( atom grid )
    c_proc( WX_SAVE_GRID_EDIT_CONTROL_VALUE, {grid} )
end procedure

--/topic wxGrid
--/proc set_cell_readonly( atom grid, atom row, atom col, integer readonly )
--
--Makes the cell read only if /i readonly is 1, or allows editing if /i readonly is 0.
public procedure set_cell_readonly( atom grid, atom row, atom col, integer readonly )
    c_proc( WX_SET_CELL_READONLY, {grid, row, col, readonly} )
end procedure

--/topic wxGrid
--/proc set_grid_editable( atom grid, integer editable )
--
--Makes the entire grid uneditable by the user is /i editable is 0, or allows editing
--if /i editable is 1.
public procedure set_grid_editable( atom grid, integer editable )
    c_proc( WX_SET_GRID_EDITABLE, {grid, editable} )
end procedure

--/topic wxGrid
--/proc autosize_columns( atom grid )
--
--Resizes each column to the minimum size required to display all the data in each column.
public procedure autosize_columns( atom grid )
    c_proc( WX_AUTOSIZE_COLUMNS, {grid} )
end procedure

--/topic wxGrid
--/proc set_col_label_align( atom grid, atom horiz, atom vert )
--
--Use wxLEFT, wxRIGHT and wxTOP, wxBOTTOM (or zero for center)
public procedure set_col_label_align( atom grid, atom horiz, atom vert )
    c_proc( WX_SET_COL_LABEL_ALIGN, {grid, horiz, vert} )
end procedure

--/topic wxGrid
--/proc set_row_label_align( atom grid, atom horiz, atom vert )
--
--Use wxLEFT, wxRIGHT and wxTOP, wxBOTTOM (or zero for center)
public procedure set_row_label_align( atom grid, atom horiz, atom vert )
    c_proc( WX_SET_ROW_LABEL_ALIGN, {grid, horiz, vert} )
end procedure

--/topic wxGrid
--/proc set_col_label_size( atom grid, integer height )
--If height equals to /b wxGRID_AUTOSIZE then height is calculated automatically so that no label is truncated.
-- Note that this could be slow for a large table. This flag is new since wxWidgets version 2.8.8.
public procedure set_col_label_size( atom grid, integer height )
    c_proc( WX_SET_COL_LABEL_SIZE, {grid, height} )
end procedure

--/topic wxGrid
--/func get_col_label( atom grid, integer col )
--
--Returns a sequence which holds the information at the top of
--the column specified by 'column'. The first column is 0.  The
--column index increases as it goes to the right.
public function get_col_label( atom grid, integer col )
    return c_func( WX_GET_COL_LABEL, {grid, col} )
end function

--/topic wxGrid
--/proc set_grid_line_color( atom grid, atom colour )
--
--Change the color of the lines in the grid.
public procedure set_grid_line_color( atom grid, atom colour )
    c_proc( WX_SET_GRID_LINE_COLOR, {grid, colour} )
end procedure

--/topic wxGrid
--/proc force_refresh_grid( atom grid )
--
--Causes immediate repainting of the grid. Use this instead of the usual /refresh_window.
public procedure force_refresh_grid( atom grid )
    c_proc( WX_FORCE_REFRESH_GRID, {grid} )
end procedure

--/topic wxGrid
--/func set_col_size( atom grid, atom col, atom size )
--
--Change the size of a grid column.  This function does not refresh the grid.
--If you are calling it outside of a BeginBatch / EndBatch block you can use /force_refresh_grid
--to see the changes
public procedure set_col_size( atom grid, atom col, atom size )
    c_proc( WX_SET_COL_SIZE, {grid, col, size} )
end procedure

--/topic wxGrid
--/proc set_cell_alignment( atom grid, atom row, atom col, atom horz, atom vert )
--
--Set the vertical of horizontal alignment of a cell:
--<ul>
--/li wxLEFT
--/li wxRIGHT
--/li wxCENTER
--/li wxTOP
--/li wxBOTTOM
--</ul>
public procedure set_cell_alignment( atom grid, atom row, atom col, atom horz, atom vert )
    c_proc( WX_SET_CELL_ALIGNMENT, {grid, row, col, horz, vert} )
end procedure

--/topic wxGrid
--/proc set_col_alignment( atom grid, atom col, atom horz, atom vert )
--
--Set the vertical of horizontal alignment of a column:
--<ul>
--/li wxLEFT
--/li wxRIGHT
--/li wxCENTER
--/li wxTOP
--/li wxBOTTOM
--</ul>
public procedure set_col_alignment( atom grid, atom col, atom horz, atom vert )
    c_proc( WX_SET_COL_ALIGNMENT, {grid, col, horz, vert} )
end procedure

--/topic wxGrid
--/proc set_col_readonly( atom grid, atom col, integer readonly )
--
--Sets the specified /i col in the specified /i grid to be read only if
--/i readonly is 1.  Editing will be allowed if /i readonly is 0.
public procedure set_col_readonly( atom grid, atom col, integer readonly )
    c_proc( WX_SET_COL_READONLY, {grid, col, readonly} )
end procedure

--/topic wxGrid
--/func get_row_label_size( atom grid )
--
--Returns an integer which represents the width of the rows in pixels.
public function get_row_label_size( atom grid )
    return c_func( WX_GET_ROW_LABEL_SIZE, {grid} )
end function

--/topic wxGrid
--/func get_col_label_size( atom grid )
--
--Returns an integer which represents the height of the columns in pixels.
public function get_col_label_size( atom grid )
    return c_func( WX_GET_COL_LABEL_SIZE, {grid} )
end function

--/topic wxGrid
--/func get_col_size( atom grid, integer col)
--
--Returns the width of the specified column.
public function get_col_size( atom grid, integer col)
    return c_func( WX_GET_COL_SIZE, {grid, col} )
end function

--/topic wxGrid
--/func get_row_size( atom grid, integer row)
--
--Returns the height of the specified row.
public function get_row_size( atom grid, integer row)
    return c_func( WX_GET_ROW_SIZE, {grid, row} )
end function

--/topic wxGrid
--/proc autosize_grid( atom grid )
--
--Automatically sets the height and width of all rows and columns to fit their contents.
--
--/b Note: wxGrid sets up arrays to store individual row and column sizes when non-default sizes are used. The memory requirements for this
-- could become prohibitive if your grid is very large.
public procedure autosize_grid( atom grid )
    c_proc( WX_AUTOSIZE_GRID, {grid})
end procedure

--/topic wxGrid
--/proc autosize_rows( atom grid )
--
--Automatically sizes all rows to fit their contents. If setAsMin is true the calculated heights will also be set as the minimal heights for the rows.
--
--/b Note: wxGrid sets up arrays to store individual row and column sizes when non-default sizes are used. The memory requirements for this
-- could become prohibitive if your grid is very large.
public procedure autosize_rows( atom grid )
    c_proc( WX_AUTOSIZE_ROWS, {grid})
end procedure

--/topic wxGrid
--/proc auotosize_row( atom grid, atom row )
--
--Automatically sizes the row to fit its contents. If setAsMin is true the calculated height will also be set as the minimal height for the row.
--
--/b Note: wxGrid sets up arrays to store individual row and column sizes when non-default sizes are used. The memory requirements for this
-- could become prohibitive if your grid is very large.
public procedure auotosize_row( atom grid, atom row )
    c_proc( WX_AUOTOSIZE_ROW, {grid,row})
end procedure

--/topic wxGrid
--/func can_drag_grid_col_move( atom grid )
--
--Returns true if columns can be moved by dragging with the mouse. Columns can be moved by dragging on their labels.
public function can_drag_grid_col_move( atom grid )
    return c_func( WX_CAN_DRAG_GRID_COL_MOVE, {grid})
end function

--/topic wxGrid
--/func can_drag_grid_col_size( atom grid, integer col = 0 )
--
--Returns true if columns can be resized by dragging with the mouse. Columns can be resized by dragging the edges of their labels. If grid line
-- dragging is enabled they can also be resized by dragging the right edge of the column in the grid cell area (see /enable_drag_grid_col_size).
public function can_drag_grid_col_size( atom grid, integer col = 0 )
    return c_func( WX_CAN_DRAG_GRID_COL_SIZE, {grid, col })
end function

--/topic wxGrid
--/func can_drag_grid_row_size( atom grid, integer row = 0 )
--
--Returns true if rows can be resized by dragging with the mouse. Rows can be resized by dragging the edges of their labels. If grid line
-- dragging is enabled they can also be resized by dragging the lower edge of the row in the grid cell area (see /enable_drag_grid_row_size).
public function can_drag_grid_row_size( atom grid, integer row = 0 )
    return c_func( WX_CAN_DRAG_GRID_ROW_SIZE, {grid, row })
end function

--/topic wxGrid
--/func can_drag_grid_size( atom grid )
--
--Return true if the dragging of grid lines to resize rows and columns is enabled or false otherwise.
public function can_drag_grid_size( atom grid )
    return c_func( WX_CAN_DRAG_GRID_SIZE, {grid})
end function

--/topic wxGrid
--/func can_enable_cell_control( atom grid )
--
--Returns true if the in-place edit control for the current grid cell can be used and false otherwise (e.g. if the current cell is read-only).
public function can_enable_cell_control( atom grid )
    return c_func( WX_CAN_ENABLE_CELL_CONTROL, {grid})
end function

--/topic wxGrid
--/func grid_cell_to_rect( atom grid, atom row, atom col )
--
--Return the rectangle corresponding to the grid cell's size and position in logical coordinates.
public function grid_cell_to_rect( atom grid, atom row, atom col )
    return c_func( WX_GRID_CELL_TO_RECT, {grid,row,col})
end function

--/topic wxGrid
--/proc clear_grid_selection( atom grid )
--
--Deselects all cells that are currently selected.
public procedure clear_grid_selection( atom grid )
    c_proc( WX_CLEAR_GRID_SELECTION, {grid})
end procedure

--/topic wxGrid
--/proc enable_grid_cell_edit_control( atom grid, atom enable )
--
--Enables or disables in-place editing of grid cell data. The grid will issue either a wxEVT_GRID_EDITOR_SHOWN or wxEVT_GRID_EDITOR_HIDDEN event.
public procedure enable_grid_cell_edit_control( atom grid, atom enable )
    c_proc( WX_ENABLE_GRID_CELL_EDIT_CONTROL, {grid,enable})
end procedure

--/topic wxGrid
--/proc enable_drag_grid_col_move( atom grid, atom enable )
--
--Enables or disables column moving by dragging with the mouse.
public procedure enable_drag_grid_col_move( atom grid, atom enable )
    c_proc( WX_ENABLE_DRAG_GRID_COL_MOVE, {grid,enable})
end procedure

--/topic wxGrid
--/proc enable_drag_grid_col_size( atom grid, atom enable )
--
--Enables or disables column sizing by dragging with the mouse.
public procedure enable_drag_grid_col_size( atom grid, atom enable )
    c_proc( WX_ENABLE_DRAG_GRID_COL_SIZE, {grid,enable})
end procedure

--/topic wxGrid
--/proc enable_drag_grid_row_size( atom grid, atom enable )
--
--Enables or disables row sizing by dragging with the mouse.
public procedure enable_drag_grid_row_size( atom grid, atom enable )
    c_proc( WX_ENABLE_DRAG_GRID_ROW_SIZE, {grid,enable})
end procedure

--/topic wxGrid
--/proc enable_grid_editing( atom grid, atom edit )
--
--If the edit argument is false this function sets the whole grid as read-only. If the argument is true the grid is set to the default state
-- where cells may be editable. In the default state you can set single grid cells and whole rows and columns to be editable or read-only via
-- wxGridCellAttribute::SetReadOnly. For single cells you can also use the shortcut function wxGrid::SetReadOnly.
--
--For more information about controlling grid cell attributes see the wxGridCellAttr cell attribute class and the wxGrid classes overview.
public procedure enable_grid_editing( atom grid, atom edit )
    c_proc( WX_ENABLE_GRID_EDITING, {grid,edit})
end procedure

--/topic wxGrid
--/proc enable_grid_lines( atom grid, atom enable )
--
--Turns the drawing of grid lines on or off.
public procedure enable_grid_lines( atom grid, atom enable )
    c_proc( WX_ENABLE_GRID_LINES, {grid,enable})
end procedure

--/topic wxGrid
--/func get_grid_batch_count( atom grid )
--
--Returns the number of times that /begin_batch has been called without (yet) matching calls to /end_batch. While the grid's batch count is
-- greater than zero the display will not be updated.
public function get_grid_batch_count( atom grid )
    return c_func( WX_GET_GRID_BATCH_COUNT, {grid})
end function

--/topic wxGrid
--/func get_cell_alignment( atom grid, atom row, atom col )
--
--Sets the arguments to the horizontal and vertical text alignment values for the grid cell at the specified location.
--
--Horizontal alignment will be one of wxALIGN_LEFT, wxALIGN_CENTRE or wxALIGN_RIGHT.
--Vertical alignment will be one of wxALIGN_TOP, wxALIGN_CENTRE or wxALIGN_BOTTOM.
public function get_cell_alignment( atom grid, atom row, atom col )
    return c_func( WX_GET_CELL_ALIGNMENT, {grid,row,col})
end function

--/topic wxGrid
--/func get_cell_background_color( atom grid, atom row, atom col )
--
--Returns the background colour of the cell at the specified location.
public function get_cell_background_color( atom grid, atom row, atom col )
    return c_func( WX_GET_CELL_BACKGROUND_COLOR, {grid,row,col})
end function

--/topic wxGrid
--/func get_cell_font( atom grid, atom row, atom col )
--
--Returns the font for text in the grid cell at the specified location.
public function get_cell_font( atom grid, atom row, atom col )
    return c_func( WX_GET_CELL_FONT, {grid,row,col})
end function

--/topic wxGrid
--/func get_cell_text_color( atom grid, atom row, atom col )
--
--Returns the text colour for the grid cell at the specified location.
public function get_cell_text_color( atom grid, atom row, atom col )
    return c_func( WX_GET_CELL_TEXT_COLOR, {grid,row,col})
end function

--/topic wxGrid
--/func get_default_cell_background_color( atom grid )
--
--Returns the current default background colour for grid cells.
public function get_default_cell_background_color( atom grid )
    return c_func( WX_GET_DEFAULT_CELL_BACKGROUND_COLOR, {grid})
end function

--/topic wxGrid
--/func get_default_cell_font( atom grid )
--
--Returns the current default font for grid cell text.
public function get_default_cell_font( atom grid )
    return c_func( WX_GET_DEFAULT_CELL_FONT, {grid})
end function

--/topic wxGrid
--/func get_default_cell_text_colour( atom grid )
--
--Returns the current default colour for grid cell text.
public function get_default_cell_text_colour( atom grid )
    return c_func( WX_GET_DEFAULT_CELL_TEXT_COLOUR, {grid})
end function

--/topic wxGrid
--/func get_default_col_label_size( atom grid )
--
--Returns the default height for column labels.
public function get_default_col_label_size( atom grid )
    return c_func( WX_GET_DEFAULT_COL_LABEL_SIZE, {grid})
end function

--/topic wxGrid
--/func get_default_col_size( atom grid )
--
--Returns the current default width for grid columns.
public function get_default_col_size( atom grid )
    return c_func( WX_GET_DEFAULT_COL_SIZE, {grid})
end function

--/topic wxGrid
--/func get_default_grid_editor( atom grid )
--
--Returns a pointer to the current default grid cell editor.
--
--See /wxGridCellEditor and the /wxGrid overview for more information about cell editors and renderers.
public function get_default_grid_editor( atom grid )
    return c_func( WX_GET_DEFAULT_GRID_EDITOR, {grid})
end function

--/topic wxGrid
--/func get_default_editor_for_cell( atom grid, atom row, atom col )
--
public function get_default_editor_for_cell( atom grid, atom row, atom col )
    return c_func( WX_GET_DEFAULT_EDITOR_FOR_CELL, {grid,row,col})
end function

--/topic wxGrid
--/func get_default_row_label_size( atom grid )
--
--Returns the default width for the row labels.
public function get_default_row_label_size( atom grid )
    return c_func( WX_GET_DEFAULT_ROW_LABEL_SIZE, {grid})
end function

--/topic wxGrid
--/func get_default_row_size( atom grid )
--
--Returns the current default height for grid rows.
public function get_default_row_size( atom grid )
    return c_func( WX_GET_DEFAULT_ROW_SIZE, {grid})
end function

--/topic wxGrid
--/func get_grid_line_color( atom grid )
--
--Returns the colour used for grid lines.
public function get_grid_line_color( atom grid )
    return c_func( WX_GET_GRID_LINE_COLOR, {grid})
end function

--/topic wxGrid
--/func get_grid_lines_enabled( atom grid )
--
--Returns true if drawing of grid lines is turned on, false otherwise.
public function get_grid_lines_enabled( atom grid )
    return c_func( WX_GET_GRID_LINES_ENABLED, {grid})
end function

--/topic wxGrid
--/func get_grid_label_background_color( atom grid )
--
--Returns the colour used for the background of row and column labels.
public function get_grid_label_background_color( atom grid )
    return c_func( WX_GET_GRID_LABEL_BACKGROUND_COLOR, {grid})
end function

--/topic wxGrid
--/func get_grid_label_font( atom grid )
--
--Returns the font used for row and column labels.
public function get_grid_label_font( atom grid )
    return c_func( WX_GET_GRID_LABEL_FONT, {grid})
end function

--/topic wxGrid
--/func get_grid_label_text_color( atom grid )
--
--Returns the colour used for row and column label text.
public function get_grid_label_text_color( atom grid )
    return c_func( WX_GET_GRID_LABEL_TEXT_COLOR, {grid})
end function

--/topic wxGrid
--/func get_grid_row_label_align( atom grid )
--
--Sets the arguments to the current row label alignment values.
--
--Horizontal alignment will be one of wxLEFT, wxCENTRE or wxRIGHT.
--Vertical alignment will be one of wxTOP, wxCENTRE or wxBOTTOM.
public function get_grid_row_label_align( atom grid )
    return c_func( WX_GET_GRID_ROW_LABEL_ALIGN, {grid})
end function

--/topic wxGrid
--/func get_grid_scroll_line( atom grid )
--
--Returns the number of pixels per scroll increment. The default is 15.
public function get_grid_scroll_line( atom grid )
    return c_func( WX_GET_GRID_SCROLL_LINE, {grid})
end function

--/topic wxGrid
--/func get_grid_selection_background( atom grid )
--
--Access or update the selection fore/back colours
public function get_grid_selection_background( atom grid )
    return c_func( WX_GET_GRID_SELECTION_BACKGROUND, {grid})
end function

--/topic wxGrid
--/func get_grid_selection_foreground( atom grid )
--
--See /get_grid_selection_background.
public function get_grid_selection_foreground( atom grid )
    return c_func( WX_GET_GRID_SELECTION_FOREGROUND, {grid})
end function

--/topic wxGrid
--/func get_grid_view_width( atom grid )
--
--Returned number of whole cols visible.
public function get_grid_view_width( atom grid )
    return c_func( WX_GET_GRID_VIEW_WIDTH, {grid})
end function

--/topic wxGrid
--/proc hide_cell_edit_control( atom grid )
--
--Hides the in-place cell edit control.
public procedure hide_cell_edit_control( atom grid )
    c_proc( WX_HIDE_CELL_EDIT_CONTROL, {grid})
end procedure

--/topic wxGrid
--/func insert_grid_cols( atom grid, atom pos, atom num_cols )
--
--Inserts one or more new columns into a grid with the first new column at the specified position and returns true if successful.
public function insert_grid_cols( atom grid, atom pos, atom num_cols )
    return c_func( WX_INSERT_GRID_COLS, {grid,pos,num_cols})
end function

--/topic wxGrid
--/func insert_grid_rows( atom grid, atom pos, atom num_rows )
--
--Inserts one or more new rows into a grid with the first new row at the specified position and returns true if successful.
public function insert_grid_rows( atom grid, atom pos, atom num_rows )
    return c_func( WX_INSERT_GRID_ROWS, {grid,pos,num_rows})
end function

--/topic wxGrid
--/func is_cell_edit_control_enabled( atom grid )
--
--Returns true if the in-place edit control is currently enabled.
public function is_cell_edit_control_enabled( atom grid )
    return c_func( WX_IS_CELL_EDIT_CONTROL_ENABLED, {grid})
end function

--/topic wxGrid
--/func is_current_cell_read_only( atom grid )
--
--Returns true if the current cell has been set to read-only.
public function is_current_cell_read_only( atom grid )
    return c_func( WX_IS_CURRENT_CELL_READ_ONLY, {grid})
end function

--/topic wxGrid
--/func is_grid_editable( atom grid )
--
--Returns false if the whole grid has been set as read-only or true otherwise.
public function is_grid_editable( atom grid )
    return c_func( WX_IS_GRID_EDITABLE, {grid})
end function

function is_cell_read_only( atom grid, atom row, atom col )
    return c_func( WX_IS_CELL_READ_ONLY, {grid,row,col})
end function

--/topic wxGrid
--/func is_selection_for_grid( atom grid )
--
--Returns true if there are currently rows, columns or blocks of cells selected.
public function is_selection_for_grid( atom grid )
    return c_func( WX_IS_SELECTION_FOR_GRID, {grid})
end function

--/topic wxGrid
--/func move_grid_cursor_down( atom grid, atom expand_selection )
--
--Moves the grid cursor down by one row. If a block of cells was previously selected it will expand if the argument is true or be cleared
-- if /i expand_selection is false.
--
--/b Keyboard
--
--This function is called for Down cursor key presses or Shift+Down to expand a selection.
public function move_grid_cursor_down( atom grid, atom expand_selection )
    return c_func( WX_MOVE_GRID_CURSOR_DOWN, {grid,expand_selection})
end function

--/topic wxGrid
--/func move_grid_cursor_left( atom grid, atom expand_selection )
--
--Moves the grid cursor left by one column. If a block of cells was previously selected it will expand if the argument is true or be cleared
-- if /i expand_selection is false.
--
--/b Keyboard
--
--This function is called for Left cursor key presses or Shift+Left to expand a selection.
public function move_grid_cursor_left( atom grid, atom expand_selection )
    return c_func( WX_MOVE_GRID_CURSOR_LEFT, {grid,expand_selection})
end function

--/topic wxGrid
--/func move_grid_cursor_right( atom grid, atom expand_selection )
--
--Moves the grid cursor right by one column. If a block of cells was previously selected it will expand if the argument is true or be cleared
-- if /i expand_selection is false.
--
--/b Keyboard
--
--This function is called for Right cursor key presses or Shift+Right to expand a selection.
public function move_grid_cursor_right( atom grid, atom expand_selection )
    return c_func( WX_MOVE_GRID_CURSOR_RIGHT, {grid,expand_selection})
end function

--/topic wxGrid
--/func move_grid_cursor_up( atom grid, atom expand_selection )
--
--Moves the grid cursor up by one row. If a block of cells was previously selected it will expand if the argument is true or be cleared
-- if /i expand_selection is false.
--
--/b Keyboard
--
--This function is called for Up cursor key presses or Shift+Up to expand a selection.
public function move_grid_cursor_up( atom grid, atom expand_selection )
    return c_func( WX_MOVE_GRID_CURSOR_UP, {grid,expand_selection})
end function

--/topic wxGrid
--/func move_grid_cursor_down_block( atom grid, atom expand_selection )
--
--Moves the grid cursor down in the current column such that it skips to the beginning or end of a block of non-empty cells. If a block of
-- cells was previously selected it will expand if /i expand_selection is true or be cleared if the argument is false.
--
--/b Keyboard
--
--This function is called for the Ctrl+Down key combination. Shift+Ctrl+Down expands a selection.
public function move_grid_cursor_down_block( atom grid, atom expand_selection )
    return c_func( WX_MOVE_GRID_CURSOR_DOWN_BLOCK, {grid,expand_selection})
end function

--/topic wxGrid
--/func move_grid_cursor_left_block( atom grid, atom expand_selection )
--
--Moves the grid cursor left in the current row such that it skips to the beginning or end of a block of non-empty cells. If a block of
-- cells was previously selected it will expand if /i expand_selection is true or be cleared if the argument is false.
--
--/b Keyboard
--
--This function is called for the Ctrl+Left key combination. Shift+Ctrl+left expands a selection.
public function move_grid_cursor_left_block( atom grid, atom expand_selection )
    return c_func( WX_MOVE_GRID_CURSOR_LEFT_BLOCK, {grid,expand_selection})
end function

--/topic wxGrid
--/func move_grid_cursor_right_block( atom grid, atom expand_selection )
--
--Moves the grid cursor right in the current row such that it skips to the beginning or end of a block of non-empty cells. If a block of
-- cells was previously selected it will expand if /i expand_selection is true or be cleared if the argument is false.
--
--/b Keyboard
--
--This function is called for the Ctrl+Right key combination. Shift+Ctrl+Right expands a selection.
public function move_grid_cursor_right_block( atom grid, atom expand_selection )
    return c_func( WX_MOVE_GRID_CURSOR_RIGHT_BLOCK, {grid,expand_selection})
end function

--/topic wxGrid
--/func move_grid_cursor_up_block( atom grid, atom expand_selection )
--
--Moves the grid cursor up in the current column such that it skips to the beginning or end of a block of non-empty cells. If a block of
-- cells was previously selected it will expand if /i expand_selection is true or be cleared if the argument is false.
--
--/b Keyboard
--
--This function is called for the Ctrl+Up key combination. Shift+Ctrl+Up expands a selection.
public function move_grid_cursor_up_block( atom grid, atom expand_selection )
    return c_func( WX_MOVE_GRID_CURSOR_UP_BLOCK, {grid,expand_selection})
end function

--/topic wxGrid
--/func move_grid_page_down( atom grid )
--
--Moves the grid cursor down by some number of rows so that the previous bottom visible row becomes the top visible row.
--
--/b Keyboard
--
--This function is called for PgDn keypresses.
public function move_grid_page_down( atom grid )
    return c_func( WX_MOVE_GRID_PAGE_DOWN, {grid})
end function

--/topic wxGrid
--/func move_grid_page_up( atom grid )
--
--Moves the grid cursor up by some number of rows so that the previous top visible row becomes the bottom visible row.
--
--/b Keyboard
--
--This function is called for PgUp keypresses.
public function move_grid_page_up( atom grid )
    return c_func( WX_MOVE_GRID_PAGE_UP, {grid})
end function

--/topic wxGrid
--/proc set_cell_background_color( atom grid, atom row, atom col, atom color )
--
public procedure set_cell_background_color( atom grid, atom row, atom col, atom color )
    c_proc( WX_SET_CELL_BACKGROUND_COLOR, {grid,row,col,color})
end procedure

--/topic wxGrid
--/proc set_cell_font( atom grid, atom row, atom col, atom font )
--
--Sets the font for text in the grid cell at the specified location.
public procedure set_cell_font( atom grid, atom row, atom col, atom font )
    c_proc( WX_SET_CELL_FONT, {grid,row,col,font})
end procedure

--/topic wxGrid
--/proc set_cell_text_color( atom grid, atom row, atom col, atom color )
--
--Sets the text colour for the grid cell at the specified location.
public procedure set_cell_text_color( atom grid, atom row, atom col, atom color )
    c_proc( WX_SET_CELL_TEXT_COLOR, {grid,row,col,color})
end procedure

--/topic wxGrid
--/proc set_col_format_bool( atom grid, atom col )
--
--Sets the specified column to display boolean values. wxGrid displays boolean values with a checkbox.
public procedure set_col_format_bool( atom grid, atom col )
    c_proc( WX_SET_COL_FORMAT_BOOL, {grid,col})
end procedure

--/topic wxGrid
--/proc set_col_format_number( atom grid, atom col )
--
--Sets the specified column to display integer values.
public procedure set_col_format_number( atom grid, atom col )
    c_proc( WX_SET_COL_FORMAT_NUMBER, {grid,col})
end procedure

--/topic wxGrid
--/proc set_col_format_float( atom grid, atom col, atom width, atom precision )
--
--Sets the specified column to display floating point values with the given width and precision.
public procedure set_col_format_float( atom grid, atom col, atom width, atom precision )
    c_proc( WX_SET_COL_FORMAT_FLOAT, {grid,col,width,precision})
end procedure

--/topic wxGrid
--/proc set_col_pos( atom grid, atom col_id, atom new_pos )
--
--Sets the position of the specified column.
public procedure set_col_pos( atom grid, atom col_id, atom new_pos )
    c_proc( WX_SET_COL_POS, {grid,col_id,new_pos})
end procedure

--/topic wxGrid
--/proc set_default_cell_align( atom grid, atom horiz, atom vert )
--
--Sets the default horizontal and vertical alignment for grid cell text.
--
--Horizontal alignment should be one of wxALIGN_LEFT, wxALIGN_CENTRE or wxALIGN_RIGHT.
--Vertical alignment should be one of wxALIGN_TOP, wxALIGN_CENTRE or wxALIGN_BOTTOM.
public procedure set_default_cell_align( atom grid, atom horiz, atom vert )
    c_proc( WX_SET_DEFAULT_CELL_ALIGN, {grid,horiz,vert})
end procedure

--/topic wxGrid
--/proc set_default_cell_background_color( atom grid, atom color )
--
--Sets the default background colour for grid cells.
public procedure set_default_cell_background_color( atom grid, atom color )
    c_proc( WX_SET_DEFAULT_CELL_BACKGROUND_COLOR, {grid,color})
end procedure

--/topic wxGrid
--/proc set_default_cell_font( atom grid, atom font )
--
--Sets the default font to be used for grid cell text.
public procedure set_default_cell_font( atom grid, atom font )
    c_proc( WX_SET_DEFAULT_CELL_FONT, {grid,font})
end procedure

--/topic wxGrid
--/proc set_default_cell_text_color( atom grid, atom color )
--
--Sets the current default colour for grid cell text.
public procedure set_default_cell_text_color( atom grid, atom color )
    c_proc( WX_SET_DEFAULT_CELL_TEXT_COLOR, {grid,color})
end procedure

--/topic wxGrid
--/proc set_default_col_size( atom grid, atom width, atom resize_existing )
--
--Sets the default width for columns in the grid. This will only affect columns subsequently added to the grid unless /i resize_existing is true.
public procedure set_default_col_size( atom grid, atom width, atom resize_existing )
    c_proc( WX_SET_DEFAULT_COL_SIZE, {grid,width,resize_existing})
end procedure

--/topic wxGrid
--/proc set_default_row_size( atom grid, atom width, atom resize_existing )
--
--Sets the default height for rows in the grid. This will only affect rows subsequently added to the grid unless /i resize_existing is true.
public procedure set_default_row_size( atom grid, atom width, atom resize_existing )
    c_proc( WX_SET_DEFAULT_ROW_SIZE, {grid,width,resize_existing})
end procedure

--/topic wxGrid
--/proc set_grid_label_background_color( atom grid, atom color )
--
--Sets the background colour for row and column labels.
public procedure set_grid_label_background_color( atom grid, atom color )
    c_proc( WX_SET_GRID_LABEL_BACKGROUND_COLOR, {grid,color})
end procedure

--/topic wxGrid
--/proc set_grid_label_font( atom grid, atom font )
--
--Sets the font for row and column labels.
public procedure set_grid_label_font( atom grid, atom font )
    c_proc( WX_SET_GRID_LABEL_FONT, {grid,font})
end procedure

--/topic wxGrid
--/proc set_grid_label_text_color( atom grid, atom color )
--
--Sets the colour for row and column label text.
public procedure set_grid_label_text_color( atom grid, atom color )
    c_proc( WX_SET_GRID_LABEL_TEXT_COLOR, {grid,color})
end procedure

--/topic wxGrid
--/proc set_grid_margins( atom grid, atom extra_width, atom extra_height )
--
--A grid may occupy more space than needed for its rows/columns. This function allows to set how big this extra space is.
public procedure set_grid_margins( atom grid, atom extra_width, atom extra_height )
    c_proc( WX_SET_GRID_MARGINS, {grid,extra_width,extra_height})
end procedure

--/topic wxGrid
--/proc set_row_size( atom grid, atom row, atom height )
--
--Sets the height of the specified row.
--
--This function does not refresh the grid. If you are calling it outside of a /begin_batch / /end_batch block you can use /refresh_grid to see the changes.
--
--/b Note: wxGrid sets up arrays to store individual row and column sizes when non-default sizes are used. The memory requirements for this
-- could become prohibitive if your grid is very large.
public procedure set_row_size( atom grid, atom row, atom height )
    c_proc( WX_SET_ROW_SIZE, {grid,row,height})
end procedure

--/topic wxGrid
--/proc set_grid_scroll_line( atom grid, sequence xy )
--
--Sets the number of pixels per scroll increment. The default is 15. Sometimes wxGrid has trouble setting the scrollbars correctly due to
-- rounding errors: setting this to 1 can help.
public procedure set_grid_scroll_line( atom grid, sequence xy )
    c_proc( WX_SET_GRID_SCROLL_LINE, {grid,xy})
end procedure

--/topic wxGrid
--/proc set_grid_selection_background( atom grid, atom color )
--
public procedure set_grid_selection_background( atom grid, atom color )
    c_proc( WX_SET_GRID_SELECTION_BACKGROUND, {grid,color})
end procedure

--/topic wxGrid
--/proc set_grid_selection_foreground( atom grid, atom color )
--
public procedure set_grid_selection_foreground( atom grid, atom color )
    c_proc( WX_SET_GRID_SELECTION_FOREGROUND, {grid,color})
end procedure

--/topic wxGrid
--/proc show_cell_edit_control( atom grid )
--
--Displays the in-place cell edit control for the current cell.
public procedure show_cell_edit_control( atom grid )
    c_proc( WX_SHOW_CELL_EDIT_CONTROL, {grid})
end procedure

--/topic wxGrid
--/func get_number_cols( atom grid )
--Returns the total number of grid columns (actually the number of columns in the underlying grid table).
public function get_number_cols( atom grid )
    return c_func( WX_GET_NUMBER_COLS, {grid})
end function

--/topic wxGrid
--/func get_number_rows( atom grid )
--Returns the total number of grid rows (actually the number of rows in the underlying grid table).
public function get_number_rows( atom grid )
    return c_func( WX_GET_NUMBER_ROWS, {grid})
end function

--/topic wxGrid
--/func get_selected_cols( atom grid )
--
public function get_selected_cols( atom grid )
    return c_func( WX_GET_SELECTED_COLS, {grid})
end function

--/topic wxGrid
--/func get_selected_rows( atom grid )
--
public function get_selected_rows( atom grid )
    return c_func( WX_GET_SELECTED_ROWS, {grid})
end function

--/topic wxGrid
--/func get_selected_cells( atom grid )
--
public function get_selected_cells( atom grid )
    return c_func( WX_GET_SELECTED_CELLS, {grid})
end function



--wxGridCellAttr
------------------------------
--/topic wxGridCellAttr
--/parent wxGrid
--/info
--
--<div class="subsection">Creation Parameters</div>
--(two options)
--<ul>
--/li /b None: create a new wxGridCellAttr
--/li /b sequence: get or create a wxGridCellAttr for a specific cell
--<ul>
--/li atom /wxGrid
--/li integer row
--/li integer column
--</ul>
--</ul>



--wxGridCellAutoWrapStringEditor
-----------------------------------------
--/topic wxGridCellAutoWrapStringEditor
--/parent wxGrid
--/info
--



--wxGridCellEditor
------------------------------
--/topic wxGridCellEditor
--/parent wxGrid
--/parent wxGridCellAttr
--/info
--
--You shouldn't create an object of this class directly.  Instead, use
--the /set_col_editor or /set_cell_editor procedures.
--
--See also: /get_cell_editor, /get_cell_editor_control
--



--wxGridSizer
------------------------------
--/topic wxGridSizer
--/parent Sizers
--/parent Classes
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "integer columns"
--/li /b "integer rows"
--/li /b "integer vgap"
--/li /b "integer hgap"
--</ul>
--/b rows and /b columns determine the number of columns and rows in the sizer - if either of
--the parameters is zero, it will be calculated to form the total number of children in
--the sizer, thus making the sizer grow dynamically. vgap and hgap define extra space
--between all children.



--wxHtmlEasyPrinting
-----------------------------------------
--/topic wxHtmlEasyPrinting
--/parent Printing
--/parent Classes
--/info
--

--/topic wxHtmlEasyPrinting
--/func get_html_easy_printing_parent( atom easy )
--
--Note that this function was introduced in wxWidgets 2.8.5.  If you are using an earlier
--version, then this will return 0.
public function get_html_easy_printing_parent( atom easy )
    return c_func( WX_GET_HTML_EASY_PRINTING_PARENT, {easy})
end function

--/topic wxHtmlEasyPrinting
--/func get_html_easy_print_data( atom easy )
--
--Returns pointer to /wxPrintData instance used by this class. You can set its parameters (via set_xxxx methods).
public function get_html_easy_print_data( atom easy )
    return c_func( WX_GET_HTML_EASY_PRINT_DATA, {easy})
end function

--/topic wxHtmlEasyPrinting
--/func get_html_easy_print_setup( atom easy )
--
--Returns a pointer to /wxPageSetupDialogData instance used by this class. You can set its parameters (via set_xxxx methods).
public function get_html_easy_print_setup( atom easy )
    return c_func( WX_GET_HTML_EASY_PRINT_SETUP, {easy})
end function

--/topic wxHtmlEasyPrinting
--/func get_html_easy_print_preview_file( atom easy, sequence html_file )
--
--Preview HTML file. Returns false in case of error.
public function get_html_easy_print_preview_file( atom easy, sequence html_file )
    return c_func( WX_GET_HTML_EASY_PRINT_PREVIEW_FILE, {easy,html_file})
end function

--/topic wxHtmlEasyPrinting
--/func get_html_easy_print_preview_text( atom easy, object html_text, object base_path )
--
--Preview HTML text (/i not file!). Returns false in case of error.
--
--/b Parameters
--<ul>
--/li /i html_text HTML text.
--/li /i base_path base directory (html string would be stored there if it was in file). It is used to determine path for loading images,
-- for example.
--</ul>
public function get_html_easy_print_preview_text( atom easy, sequence html_text, sequence base_path )
    return c_func( WX_GET_HTML_EASY_PRINT_PREVIEW_TEXT, {easy,html_text,base_path})
end function

--/topic wxHtmlEasyPrinting
--/func print_html_file( atom easy, sequence html_file )
--
--Print HTML file. Returns false in case of error.
public function print_html_file( atom easy, sequence html_file )
    return c_func( WX_PRINT_HTML_FILE, {easy,html_file})
end function

--/topic wxHtmlEasyPrinting
--/func print_html_text( atom easy, sequence html_text, sequence base_path )
--
--Print HTML text (not file!). Returns false in case of error.
--
--/b Parameters
--<ul>
--/li /i html_text HTML text.
--/li /i base_path base directory (html string would be stored there if it was in file). It is used to determine path for loading images,
-- for example.
--</ul>
public function print_html_text( atom easy, sequence html_text, sequence base_path )
    return c_func( WX_PRINT_HTML_TEXT, {easy,html_text,base_path})
end function

--/topic wxHtmlEasyPrinting
--/proc html_easy_page_setup( atom easy )
--
--Display page setup dialog and allows the user to modify settings.
public procedure html_easy_page_setup( atom easy )
    c_proc( WX_HTML_EASY_PAGE_SETUP, {easy})
end procedure

--/topic wxHtmlEasyPrinting
--/proc html_easy_fonts( atom easy, sequence normal_face, sequence fixed_face, sequence font_size )
--
--Sets fonts.
public procedure html_easy_fonts( atom easy, sequence normal_face, sequence fixed_face, sequence font_size )
    c_proc( WX_HTML_EASY_FONTS, {easy,normal_face,fixed_face,font_size})
end procedure

--/topic wxHtmlEasyPrinting
--/proc html_easy_header( atom easy, sequence header, atom pg )
--
--Set page header. The following macros can be used inside it:
--<ul>
--/li @DATE@ is replaced by the current date in default format
--/li @PAGENUM@ is replaced by page number
--/li @PAGESCNT@ is replaced by total number of pages
--/li @TIME@ is replaced by the current time in default format
--/li @TITLE@ is replaced with the title of the document
--</ul>
--
--/b Parameters
--<ul>
--/li /i header HTML text to be used as header.
--/li /i pg one of wxPAGE_ODD, wxPAGE_EVEN and wxPAGE_ALL constants.
--</ul>
public procedure html_easy_header( atom easy, sequence header, atom pg )
    c_proc( WX_HTML_EASY_HEADER, {easy,header,pg})
end procedure

--/topic wxHtmlEasyPrinting
--/proc html_easy_footer( atom easy, sequence footer, atom pg )
--
public procedure html_easy_footer( atom easy, sequence footer, atom pg )
    c_proc( WX_HTML_EASY_FOOTER, {easy,footer,pg})
end procedure

--/topic wxHtmlEasyPrinting
--/proc html_easy_parent( atom easy, atom window )
--
--Note that this routine was introduced in wxWidgets 2.8.5.  If you have an earlier version
--of wxWidgets, calling this procedure will have no effect.
public procedure html_easy_parent( atom easy, atom window )
    c_proc( WX_HTML_EASY_PARENT, {easy,window})
end procedure



--wxHtmlLinkEvent
---------------------------------
--/topic wxHtmlLinkEvent
--/parent wxHtmlWindow
--/parent Events
--

--/topic wxHtmlLinkEvent
--/func get_link_event_info( atom event )
--
--Returns the /wxHtmlLinkInfo object which contains info about the cell clicked and the hyperlink it contains.
public function get_link_event_info( atom event )
    return c_func( WX_GET_LINK_EVENT_INFO, {event})
end function



--wxHtmlLinkInfo
---------------------------------
--/topic wxHtmlLinkInfo
--/parent wxHtmlLinkEvent
--/parent wxHtmlWindow
--/parent Classes

--/topic wxHtmlLinkInfo
--/func get_link_href( atom link )
--
--Return /i HREF value of the &lt;A&gt; tag.
public function get_link_href( atom link )
    return c_func( WX_GET_LINK_HREF, {link} )
end function

--/topic wxHtmlLinkInfo
--/func get_link_target( atom link )
--
--Return /i TARGET value of the &lt;A&gt; tag (this value is used to specify in which frame should be the page pointed by /i HREF opened).
public function get_link_target( atom link )
    return c_func( WX_GET_LINK_TARGET, {link} )
end function

function on_link_event( atom htmlwin, sequence href, sequence target, integer rid )
    call_proc( rid, {htmlwin, href, target} )
    return 0
end function
constant on_link_event_cb = call_back( '+' & routine_id("on_link_event") )

--/topic wxHtmlWindow
--/proc set_link_event( atom htmlwin, integer rid )
--
--/b Note: This method is deprecated; intercept the event instead: /i wxEVT_COMMAND_HTML_LINK_CLICKED.
--
--Set a callback for when the user clicks on a link in a wxHtmlWindow.  The
--callback must be a procedure that takes 3 parameters:
--<ul>
--/li /b htmlwin: the wxHtmlWindow
--/li /b href: a sequence that will be the href for the link
--/li /b target: a sequence that contains the target for the link
--</ul>
--If you trap this event, you must call /load_html_page() in order for the
--link to execute.
public procedure set_link_event( atom htmlwin, integer rid )
    c_proc( WX_SET_LINK_EVENT, {htmlwin, rid, on_link_event_cb} )
end procedure

--/topic wxHtmlWindow
--/proc set_related_frame( atom html_win, atom frame, sequence format )
--
--Sets the frame in which page title will be displayed. format is format of frame
--title, e.g. "HtmlHelp : %s". It must contain exactly one %s. This %s is substituted with HTML page title.
--/i status_slot is the slot in the status bar where messages will be displayed (-1 means no messages).
public procedure set_related_frame( atom html_win, atom frame, sequence format, integer status_slot )
    if not length(format) then
        format = "%s"
    end if
    c_proc( WX_SET_RELATED_FRAME, {html_win, frame, format, status_slot} )
end procedure

--/topic wxHtmlWindow
--/proc get_html_page_title( atom html_win )
--Returns the title of the currently opened page.
public function get_html_page_title( atom html_win )
    return c_func( WX_GET_HTML_PAGE_TITLE, {html_win} )
end function

--/topic wxHtmlWindow
--/proc clear_html_history( atom html_win )
--Clears the history cache.
public procedure clear_html_history( atom html_win )
    c_proc( WX_CLEAR_HTML_HISTORY, {html_win} )
end procedure

--/topic wxHtmlWindow
--/func get_html_anchor( atom html_win )
--Returns anchor within currently opened page (see /get_html_page()).
--If no page is opened or if the displayed page wasn't produced by
--call to LoadPage, empty string is returned.
public function get_html_anchor( atom html_win )
    return c_func( WX_GET_HTML_ANCHOR, {html_win} )
end function

--/topic wxHtmlWindow
--/func get_html_page( atom html_win )
--Returns full location of the opened page. If no page is opened or if
--the displayed page wasn't produced by call to LoadPage, empty string is returned.
public function get_html_page( atom html_win )
    return c_func( WX_GET_HTML_PAGE, {html_win} )
end function



--wxHtmlWindow
------------------------------
--/topic wxHtmlWindow
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= wxHW_SCROLLBAR_AUTO]
--<ul>
--/li wxHW_SCROLLBAR_NEVER
--/li wxHW_SCROLLBAR_AUTO
--</ul>
--</ul>
--
--This class can be used to display HTML documents.  It does not always handle badly formed
--html very well.  Make sure that any lists are surrounded by &lt;ul&gt; &lt;/ul&gt; or
--&lt;ol&gt; &lt;/ol&gt; tags.

--/topic wxHtmlWindow
--/proc load_html_file( atom htmlwin, sequence file)
--
--Loads the specified file.
public procedure load_html_file( atom htmlwin, sequence file)
    c_proc( WX_LOAD_HTML_FILE, {htmlwin, file} )
end procedure

--/topic wxHtmlWindow
--/proc load_html_page( atom htmlwin, sequence page )
--
--This opens a specified html page into the designated wxHtmlWindow.
--If you want it to follow HTML links, you will need this the second
--part to be ahref.
public procedure load_html_page( atom htmlwin, sequence page )
    c_proc( WX_LOAD_HTML_PAGE, {htmlwin, page} )
end procedure

--/topic wxHtmlWindow
--/proc set_html_page( atom htmlwin, sequence source )
--
--This sets the source of the displayed page to /i source.
public procedure set_html_page( atom htmlwin, sequence source )
    c_proc( WX_SET_HTML_PAGE, {htmlwin, source} )
end procedure

--/topic wxHtmlWindow
--/proc html_back( atom htmlwin )
--
--Goes backward in the list of pages in an wxHtmlWindow's cache
--of viewed pages. If there is no page before the current page,
--then nothing will happen.
public procedure html_back( atom htmlwin )
    c_proc( WX_HTML_BACK, {htmlwin} )
end procedure

--/topic wxHtmlWindow
--/proc html_forward( atom htmlwin )
--
--Goes forward in the list of pages in an wxHtmlWindow's cache of
--viewed pages. If there is no page after the current page, then
--nothing will happen.
public procedure html_forward( atom htmlwin )
    c_proc( WX_HTML_FORWARD, {htmlwin} )
end procedure

--/topic wxHtmlWindow
--/proc html_append( atom htmlwin, sequence source)
--
--This appends the HTML Source to the Control
public procedure html_append(atom htmlwin, sequence source)
    c_proc( WX_HTML_APPEND, {htmlwin, source} )
end procedure

function on_open_url_event(atom html_win, atom url_type, integer rid)
    return call_func( rid, {html_win, url_type, get_the_object() } )
end function
constant on_open_url_event_cb = call_back( '+' & routine_id("on_open_url_event"))

--/topic wxHtmlWindow
--/proc set_open_url_event( atom htmlwin, integer rid )
--
--Set a call_back to be called when an URL is being opened (either when the user
--clicks on a link or an image is loaded). The URL will be opened only if
--the call back returns wxHTML_OPEN. This method is called by wxHtmlParser::OpenURL.
--You can use the call back to selectively block some URLs (e.g. for security reasons) or
--to redirect them elsewhere. Default behaviour is to always return wxHTML_OPEN.
--
--Parameters
--<ul>
--/li /b type: Indicates type of the resource. Is one of wxHTML_URL_PAGE  Opening a HTML page.
--<ul>
--/li wxHTML_URL_IMAGE  Opening an image.
--/li wxHTML_URL_OTHER  Opening a resource that doesn't fall into any other category.
--</ul>
--/li /b url: URL being opened.
--</ul>
--
--Call back return value:
--<ul>
--/li /b wxHTML_OPEN  Open the URL.
--/li /b wxHTML_BLOCK  Deny access to the URL, wxHtmlParser::OpenURL will return NULL.
--/li /b "{wxHTML_REDIRECT, redirect_url }"  Don't open url, redirect to another URL.
--</ul>
public procedure set_open_url_event( atom htmlwin, integer rid )
    c_proc( WX_SET_OPEN_URL_EVENT, {htmlwin, rid, on_open_url_event_cb} )
end procedure



--wxHTTP
------------------------------
--/topic wxHTTP
--/parent Networking
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li sequence URL [=""]
--</ul>
--See the /b "http_demo.exw" program for a simple example.
--These are the steps needed in order to use this class.
--<ol>
--/li Call /set_http_header and set the Content-type.
--/li Call /http_connect, passing in the host, ex: "www.example.com".
--/li Call /get_http_input, passing in the relative address, use "\" for the root.
--/li Use /stream_read to retrieve the data.
--</ol>
--

--/topic wxHTTP
--/func get_http_input( atom http )
--
--Returns an input stream that can be used with /stream_read() to fetch
--the file from the url.
public function get_http_input( atom http, sequence path )
    return c_func( WX_GET_HTTP_INPUT, {http, path} )
end function

--/topic wxHTTP
--/func http_connect( atom http, object host, atom port )
--Connects to a server using the specified address /i host:port.
public function http_connect( atom http, object host, atom port )
    return c_func( WX_HTTP_CONNECT, {http,host,port})
end function

--/topic wxHTTP
--/func get_http_content_type( atom http )
--Returns the type of the content of the last opened stream. It is a mime-type.
public function get_http_content_type( atom http )
    return c_func( WX_GET_HTTP_CONTENT_TYPE, {http})
end function

--/topic wxHTTP
--/func get_http_header( atom http, object header )
--Returns the data attached with a field whose name is specified by header.
-- If the field doesn't exist, it will return an empty string and not a NULL string.
--
--/b Note
--
--The header is not case-sensitive, i.e. "CONTENT-TYPE" and "content-type" represent the same header.
public function get_http_header( atom http, object header )
    return c_func( WX_GET_HTTP_HEADER, {http,header})
end function

--/topic wxHTTP
--/proc set_http_header( atom http, object header, object data )
--Sets data of a field to be sent during the next request to the HTTP server.
-- The field name is specified by /i header and the content by /i data.
-- This is a low level function and it assumes that you know what you are doing.
public procedure set_http_header( atom http, object header, object data )
    c_proc( WX_SET_HTTP_HEADER, {http,header,data})
end procedure

--/topic wxHTTP
--/proc set_http_post_buffer( atom http, object buffer )
--
public procedure set_http_post_buffer( atom http, object buffer )
    c_proc( WX_SET_HTTP_POST_BUFFER, {http,buffer})
end procedure

--/topic wxHTTP
--/proc set_http_proxy_mode( atom http, atom on )
--
public procedure set_http_proxy_mode( atom http, atom on )
    c_proc( WX_SET_HTTP_PROXY_MODE, {http,on})
end procedure



--wxIcon
------------------------------
--/topic wxIcon
--/parent Graphics
--/parent Classes
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "sequence filename"
--/li /b "atom type" (see /wxBitmap)
--/li atom desired_height [=-1]
--/li atom desired_width  [=-1]
--</ul>
--This is a true Icon under Windows, but is really just a /wxBitmap under
--Linux.

--/topic wxIcon
--/func load_icon( atom icon, object file, atom icon_type, integer height, atom width )
--
--Loads an icon from a file or resource.
public function load_icon( atom icon, object file, atom icon_type, integer height, atom width )
    return c_func( WX_LOAD_ICON, {icon, file, icon_type, height, width} )
end function



--wxImage
-----------------------------------------
--/topic wxImage
--/parent Graphics
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "integer width"
--/li /b "integer height"
--/li atom data [=0]
--/li atom flag [=0]
--</ul>
--Creates an image from given data with the given width and height.
--If /b data is set and /b flag is true, then wxImage will not delete
--the actual image data in its destructor, otherwise it will free it
--by calling /i free(). If /b data is not set and /b flag is true,
--the new image will be initialized to black. Otherwise, the image
--data will be uninitialized.

--/topic wxImage
--/func get_alpha( object params )
--
--Returns the alpha intensity of { /wxImage, x, y }.
public function get_alpha( object params )
    return c_func( WX_GET_ALPHA, {params})
end function

--/topic wxImage
--/func convert_to_greyscale( atom image )
--
--Returns a greyscale version of the image.
-- The returned image uses the luminance component of the original to calculate the greyscale.
public function convert_to_greyscale( atom image )
    return c_func( WX_CONVERT_TO_GREYSCALE, {image})
end function

--/topic wxImage
--/func convert_to_mono( atom image, object rgb )
--
--Returns monochromatic version of the image.
-- The returned image has white colour where the original has (r,g,b) colour and black colour everywhere else.
public function convert_to_mono( atom image, object rgb )
    return c_func( WX_CONVERT_TO_MONO, {image,rgb})
end function

--/topic wxImage
--/func blur( atom image, atom blurRadius )
--
--Blurs the image in both horizontal and vertical directions by the specified pixel /i blurRadius.
-- This should not be used when using a single mask colour for transparency.
public function blur( atom image, atom blurRadius )
    return c_func( WX_BLUR, {image,blurRadius})
end function

--/topic wxImage
--/func blur_horizontal( atom image, atom blurRadius )
--
--Blurs the image in the horizontal direction only.
-- This should not be used when using a single mask colour for transparency.
public function blur_horizontal( atom image, atom blurRadius )
    return c_func( WX_BLUR_HORIZONTAL, {image,blurRadius})
end function

--/topic wxImage
--/func blur_vertical( atom image, atom blurRadius )
--
--Blurs the image in the vertical direction only.
-- This should not be used when using a single mask colour for transparency.
public function blur_vertical( atom image, atom blurRadius )
    return c_func( WX_BLUR_VERTICAL, {image,blurRadius})
end function

--/topic wxImage
--/func mirror( atom image, atom horizontal )
--
--Returns a mirrored copy of the image.
-- The parameter /i horizontally indicates the orientation.
public function mirror( atom image, atom horizontal )
    return c_func( WX_MIRROR, {image,horizontal})
end function

--/topic wxImage
--/func rescale( atom image, atom width, atom height, atom quality )
--
--Changes the size of the image in-place by scaling it:
-- after a call to this function, the image will have the given width and height.
--
--For a description of the quality parameter, see the /scale function.
--
--Returns the (modified) image itself.
public function rescale( atom image, atom width, atom height, atom quality )
    return c_func( WX_RESCALE, {image,width,height,quality})
end function

--/topic wxImage
--/func resize( atom image, object size, object pos )
--
--Changes the size of the image in-place without scaling it by adding either a border with the given colour or cropping as necessary.
-- The image is pasted into a new image with the given size and background colour at the position pos relative to the upper left of
-- the new image.
public function resize( atom image, object size, object pos )
    return c_func( WX_RESIZE, {image,size,pos})
end function

--/topic wxImage
--/func rotate( atom image, object angle, object center, atom interpolating, object offset )
--
--Rotates the image about the given point, by /i angle radians.
-- Passing true to /i interpolating results in better image quality, but is slower.
-- If the image has a mask, then the mask colour is used for the uncovered pixels in the rotated image background.
-- Else, black (rgb 0, 0, 0) will be used.
--
--Returns the rotated image, leaving this image intact.
public function rotate( atom image, object angle, object center, atom interpolating, object offset )
    return c_func( WX_ROTATE, {image,angle,center,interpolating,offset})
end function

--/topic wxImage
--/func rotate_90( atom image, atom clockwise )
--
--Returns a copy of the image rotated 90 degrees in the direction indicated by /i clockwise.
public function rotate_90( atom image, atom clockwise )
    return c_func( WX_ROTATE_90, {image,clockwise})
end function

--/topic wxImage
--/proc set_option( atom image, object name, object value )
--
--Sets a user-defined option. The function is case-insensitive to name.
--
--For example, when saving as a JPEG file, the option /b quality is used,
-- which is a number between 0 and 100 (0 is terrible, 100 is very good).
public procedure set_option( atom image, object name, object value )
    c_proc( WX_SET_OPTION, {image,name,value})
end procedure

--/topic wxImage
--/func get_option( atom image, object name )
--
--Gets a user-defined option. The function is case-insensitive to name.
--
--For example, when saving as a JPEG file, the option /b quality is used,
-- which is a number between 0 and 100 (0 is terrible, 100 is very good).
public function get_option( atom image, object name )
    return c_func( WX_GET_OPTION, {image,name})
end function

--/topic wxImage
--/func get_option_int( atom image, object name )
--
--Gets a user-defined option as an integer. The function is case-insensitive to name.
--
--If the given option is not present, the function returns 0. Use /has_option as 0
-- is a possibly valid value for the option.
public function get_option_int( atom image, object name )
    return c_func( WX_GET_OPTION_INT, {image,name})
end function

--/topic wxImage
--/func has_option( atom image, object name )
--
--Returns true if the given option is present. The function is case-insensitive to name.
public function has_option( atom image, object name )
    return c_func( WX_HAS_OPTION, {image,name})
end function



--wxImageList
------------------------------
--/topic wxImageList
--/parent Graphics
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "integer width" Width of the images in the list
--/li /b "integer height" Height of the images in the list
--/li integer mask [= 1] wxTrue if masks should be created for all images
--/li integer initial_count [= 1] The initial size of the list
--</ul>

--/topic wxImageList
--/func add_image( atom ilist, atom image, object mask )
--Add an image to an image list
--<ul>
--/li /b ilist image list
--/li /b image wxBitmap or wxIcon (Windows only)
--/li /b mask
--<ul> /i sequence
--<ul>
--/li If empty sequence, then bmp is a wxIcon
--/li Three element {R,G,B} specifying color to use as a mask
--</ul>
--/li /i atom bitmap to use as a mask (or 0 to use no mask)
--</ul>
--</ul>
public function add_image( atom ilist, atom image, object mask = 0 )
    return c_func( WX_ADD_IMAGE, {ilist, image, mask} )
end function



--wxInputStream
----------------------------------
--/topic wxInputStream
--/parent Files
--/parent Classes
--

--/topic wxInputStream
--/func stream_read( atom stream )
--
--Reads data from the input queue and returns it in a sequence. The data is read until an error is raised (such as EOF).
public function stream_read( atom stream )
    return c_func( WX_STREAM_READ, {stream} )
end function



--wxIPV4address
------------------------------
--/topic wxIPV4address
--/parent Networking
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b None
--</ul>

--/topic wxIPV4address
--/func set_ip_host( atom ip, sequence name )
--
--Set the address to hostname, which can be a host name or an IP-style address in dot notation (a.b.c.d)
public function set_ip_host( atom ip, sequence name )
    return c_func( WX_SET_IP_HOST, {ip, name} )
end function

--/topic wxIPV4address
--/func set_ip_service( atom ip, object service )
--
--Sets the port according to the specified service.  If /i service is a sequence,
--then the port is set to the corresponding service, otherwise, the port is set
--to /i service.
public function set_ip_service( atom ip, object service )
    return c_func( WX_SET_IP_SERVICE, {ip, service} )
end function

--/topic wxIPV4address
--/proc any_address( atom addr )
--
--Returns a local address.  This function should be used instead of
--using 127.0.0.1 or localhost.
public procedure any_address( atom addr )
    c_proc( WX_ANY_ADDRESS, {addr} )
end procedure



--wxListbook
-----------------------------------------
--/topic wxListbook
--/parent Notebook Controls
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= 0]
--</ul>


--wxListBox
------------------------------
--/topic wxListBox
--/parent List Controls
--/parent Controls
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li sequence choices [= {}]
--/li integer style [= 0]
--</ul>
--<div class="subsection">Event handling</div>
--<ul>
--/li wxEVT_COMMAND_LISTBOX_SELECTED
--/li wxEVT_COMMAND_LISTBOX_DOUBLECLICKED
--</ul>

--/topic wxListBox
--/func list_count( atom list )
--
--Used to count the number of items in a /wxListBox.
public function list_count( atom list )
    return c_func( WX_LIST_COUNT, {list} )
end function

--/topic wxListBox
--/proc clear_list( atom list )
--
--Clears the data inside a wxListBox or wxListCtrl specified by list.
public procedure clear_list( atom list )
    c_proc( WX_CLEAR_LIST, {list} )
end procedure

--/topic wxListBox
--/func find_list_string( atom list, sequence string )
--
--Returns the item number of the item in the list with the
--specified string.
public function find_list_string( atom list, sequence string )
    return c_func( WX_FIND_LIST_STRING, {list, string} )
end function

--/topic wxListBox
--/proc set_list_selection( atom list, object select )
--
--If /b select is a sequence, the list selection is set to the
--item with that string, otherwise, /b select denotes the
--item index.
public procedure set_list_selection( atom list, object select )
    c_proc( WX_SET_LIST_SELECTION, {list, select} )
end procedure

--/topic wxListBox
--/proc delete_list_item( atom list, object item )
--
--This goes to a ListBox or ListCtrl specified by list and deletes
--the number in the list corresponding to item. The first item is 0
--and it increases as the list goes down.
public procedure delete_list_item( atom list, object item )
    c_proc( WX_DELETE_LIST_ITEM, {list, item} )
end procedure

--/topic wxListBox
--/proc list_ensure_visible( atom list, integer item )
--
--Ensures this item is visible.
public procedure list_ensure_visible( atom list, integer item )
    c_proc( WX_LIST_ENSURE_VISIBLE, {list, item} )
end procedure

--/topic wxListBox
--/proc set_first_list_item( atom list, object item )
--
--Set the specified item to be the first visible item.
public procedure set_first_list_item( atom list, object item )
    c_proc( WX_SET_FIRST_LIST_ITEM, {list, item} )
end procedure



--wxListCtrl
------------------------------
--/topic wxListCtrl
--/parent List Controls
--/parent Controls
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= wxLC_REPORT]
--<ul>
--/li wxLC_VRULES
--/li wxLC_HRULES
--/li wxLC_ICON
--/li wxLC_SMALL_ICON
--/li wxLC_LIST
--/li wxLC_REPORT
--/li wxLC_ALIGN_TOP
--/li wxLC_ALIGN_LEFT
--/li wxLC_AUTOARRANGE
--/li wxLC_VIRTUAL
--/li wxLC_EDIT_LABELS
--/li wxLC_NO_HEADER
--/li wxLC_NO_SORT_HEADER
--/li wxLC_SINGLE_SEL
--/li wxLC_SORT_ASCENDING
--/li wxLC_SORT_DESCENDING
--/li wxLC_MASK_TYPE
--/li wxLC_MASK_ALIGN
--/li wxLC_MASK_SORT
--</ul>
--</ul>

--/topic wxListCtrl
--/proc set_list_item( atom list, atom index, atom col, object label_val, atom image )
--
--Sets a list item's information. Use this to set sub-column labels.
public procedure set_list_item( atom list, atom index, atom col, object label_val, atom image )
    c_proc( WX_SET_LIST_ITEM, {list,index,col,label_val,image})
end procedure

--/topic wxListCtrl
--/func get_list_event_column( atom event )
--
--The column position: it is only used with COL events. For the column dragging events, it
--is the column to the left of the divider being dragged, for the column click events
--it may be -1 if the user clicked in the list control header outside any column.
public function get_list_event_column( atom event )
    return c_func( WX_GET_LIST_EVENT_COLUMN, {event})
end function

--/topic wxListCtrl
--/func insert_listctrl_item( atom list, atom index, sequence text, integer image )
--
--Inserts an item, returning the index of the new item if successful, -1 otherwise.
public function insert_listctrl_item( atom list, atom index, sequence text, integer image )
    return c_func( WX_INSERT_LISTCTRL_ITEM, {list, index, text, image} )
end function

--/topic wxListCtrl
--/func list_hit_test( atom list, atom x, atom y, atom flags )
--
--Determines which item (if any) is at the specified point, giving details in flags. Returns index of the item or wxNOT_FOUND
-- if no item is at the specified point. /i flags will be a combination of the following flags:
--<ul>
--/li /i wxLIST_HITTEST_ABOVE Above the client area.
--/li /i wxLIST_HITTEST_BELOW Below the client area.
--/li /i wxLIST_HITTEST_NOWHERE In the client area but below the last item.
--/li /i wxLIST_HITTEST_ONITEMICON On the bitmap associated with an item.
--/li /i wxLIST_HITTEST_ONITEMLABEL On the label (string) associated with an item.
--/li /i wxLIST_HITTEST_ONITEMRIGHT In the area to the right of an item.
--/li /i wxLIST_HITTEST_ONITEMSTATEICON On the state icon for a tree view item that is in a user-defined state.
--/li /i wxLIST_HITTEST_TOLEFT  To the right of the client area.
--/li /i wxLIST_HITTEST_TORIGHT To the left of the client area.
--/li /i wxLIST_HITTEST_ONITEM  Combination of wxLIST_HITTEST_ONITEMICON, wxLIST_HITTEST_ONITEMLABEL, wxLIST_HITTEST_ONITEMSTATEICON.
--</ul>
public function list_hit_test( atom list, atom x, atom y, atom flags )
    return c_func( WX_LIST_HIT_TEST, {list,x,y,flags})
end function

--/topic wxListCtrl
--/proc list_imagelist( atom list, atom il, integer which )
--
--<ul>
--/li /b list the /wxListCtrl
--/li /b il the /wxImageList
--/li /b which
--<ul>
--/li wxIMAGE_LIST_NORMAL (used in wxLC_ICON mode)
--/li wxIMAGE_LIST_SMALL (used in wxLC_REPORT, wxLC_LIST modes)
--</ul>
--</ul>
public procedure list_imagelist( atom list, atom il, integer which )
    c_proc( WX_LIST_IMAGELIST, {list, il, which} )
end procedure

--/topic wxListCtrl
--/func insert_list_column( atom list, atom col, sequence text, integer format, integer width )
--
--For report view mode (only), inserts a column.
public function insert_list_column( atom list, atom col, sequence text, integer format, integer width )
    return c_func( WX_INSERT_LIST_COLUMN, {list, col, text, format, width} )
end function

--/topic wxListCtrl
--/proc set_list_item_image( atom list, integer item, integer image )
--
--Changes the image displayed for the item.  You must have previously
--assigned a /wxImageList to the list using /list_imagelist().
public procedure set_list_item_image( atom list, integer item, integer image )
    c_proc( WX_SET_LIST_ITEM_IMAGE, {list, item, image} )
end procedure

--/topic wxListCtrl
--/proc set_column_label( atom list, integer column, sequence text )
--
--Changes the label of the specified column (0-based).
public procedure set_column_label( atom list, integer column, sequence text )
    c_proc( WX_SET_COLUMN_LABEL, {list, column, text} )
end procedure

--/topic wxListCtrl
--/proc set_list_item_back_color( atom list, atom item, atom color )
--
--Sets the background color of a /wxListCtrl item.
--<ul>
--/li /b list the parent /wxListCtrl of the item
--/li /b item the index of the item (this is the number returned by /insert_listctrl_item())
--/li /b color a /wxColour object
--</ul>
public procedure set_list_item_back_color( atom list, atom item, atom color )
    c_proc( WX_SET_LIST_ITEM_BACK_COLOR, {list, item, color} )
end procedure



--wxMask
------------------------------
--/topic wxMask
--/parent wxBitmap
--/parent wxGraphics
--/parent wxClasses
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom /wxBitmap" A single color image that specifies the mask.
--/li /b "atom /wxColour" The color that specified the transparent areas of the mask.
--</ul>



--wxMDIChildFrame
------------------------------
--/topic wxMDIChildFrame
--/parent wxMDIParentFrame
--/parent Controls
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li /b "integer id"
--/li /b "sequence title"
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li atom style [= wxDEFAULT_FRAME_STYLE] (see /wxFrame)
--</ul>
--You should create a /wxMenuBar for the frame in order for everything to work properly.
--When the /wxMenuBar is created, a Window menu will be created automatically.

--/topic wxMDIChildFrame
--/proc restore_child( atom child )
--
--Restores (unmaximizes a child frame)
public procedure restore_child( atom child )
    c_proc( WX_RESTORE_CHILD, {child} )
end procedure

--/topic wxMDIChildFrame
--/proc activate_child( atom child )
--
--Activates a child frame.
public procedure activate_child( atom child )
    c_proc( WX_ACTIVATE_CHILD, {child} )
end procedure

--/topic wxMDIChildFrame
--/proc maximize_child( atom child, integer bool )
--
--Maximizes a child if bool = 1, or restores child is bool = 0.
public procedure maximize_child( atom child, integer bool )
    c_proc( WX_MAXIMIZE_CHILD, {child, bool} )
end procedure

--/topic wxMDIChildFrame
--/proc set_mdi_child_title ( atom child_frame, sequence title )
--
--This sets a title to an wxMDIChildFrame. This should be used in place of set_title,
--since wxMDIChildFrames will ignore that in Linux.
public procedure set_mdi_child_title ( atom child_frame, sequence title )
    c_proc( WX_SET_MDI_CHILD_TITLE, {child_frame, title} )
end procedure



--wxMDIClientWindow
------------------------------
--/topic wxMDIClientWindow
--/parent wxMDIParentFrame
--/parent Controls
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer style [= 0]
--</ul>



--wxMDIParentFrame
------------------------------
--/topic wxMDIParentFrame
--/parent wxFrame
--/parent Controls
--/info
--An MDI (Multiple Document Interface) parent frame is a window which can contain MDI
--child frames in its own 'desktop'. It is a convenient way to avoid window clutter,
--and is used in many popular Windows applications, such as Microsoft Word(TM).
--
--There may be multiple MDI parent frames in a single application, but this probably only makes
--sense within programming development environments.
--
--Child frames may be of class wxMDIChildFrame (contained within the parent frame) or
--/wxFrame (shown as a top-level frame).
--
--An MDI parent frame always has a wxMDIClientWindow associated with it, which is the parent
--for MDI client frames. This client window may be resized to accommodate non-MDI windows, as
--seen in Microsoft Visual C++ (TM) and Microsoft Publisher (TM), where a documentation window
--is placed to one side of the workspace.
--
--MDI remains popular despite dire warnings from Microsoft itself that MDI is an obsolete user
--interface style.
--
--The implementation is native in Windows, and simulated under Motif. Under Motif, the child window
--frames will often have a different appearance from other frames because the window decorations
--are simulated.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li /b "integer id"
--/li /b "sequence title"
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li atom style [= wxDEFAULT_FRAME_STYLE] (see /wxFrame)
--</ul>

--/topic wxMDIParentFrame
--/proc arrange_mdi_icons( atom parent )
--Arranges any iconized (minimized) MDI child windows
public procedure arrange_mdi_icons( atom parent )
    c_proc( WX_ARRANGE_MDI_ICONS, {parent} )
end procedure

--/topic wxMDIParentFrame
--/proc cascade_mdi( atom parent )
--Arranges the MDI child windows in a cascade.
public procedure cascade_mdi( atom parent )
    c_proc( WX_CASCADE_MDI, {parent} )
end procedure

--/topic wxMDIParentFrame
--/proc tile_mdi( atom parent, atom orient )
--Arranges the MDI child windows in a tile pattern.
--/i orient may be /b wxHORIZONTAL or /b wxVERTICAL
public procedure tile_mdi( atom parent, atom orient )
    c_proc( WX_TILE_MDI, {parent, orient} )
end procedure

--/topic wxMDIParentFrame
--/proc activate_next_mdi( atom parent_frame )
--
--Activates the MDI child following the currently active one.
public procedure activate_next_mdi( atom parent_frame )
    c_proc( WX_ACTIVATE_NEXT_MDI, {parent_frame})
end procedure

--/topic wxMDIParentFrame
--/proc activate_prev_mdi( atom parent_frame )
--
--Activates the MDI child preceding the currently active one.
public procedure activate_prev_mdi( atom parent_frame )
    c_proc( WX_ACTIVATE_PREV_MDI, {parent_frame})
end procedure

--/topic wxMDIParentFrame
--/func get_active_mdi_child( atom parent_frame )
--
--Returns a pointer to the active MDI child, if there is one.
public function get_active_mdi_child( atom parent_frame )
    return c_func( WX_GET_ACTIVE_MDI_CHILD, {parent_frame})
end function

--/topic wxMDIParentFrame
--/func get_client_window( atom parent_frame )
--
--Returns a pointer to the client window.
public function get_client_window( atom parent_frame )
    return c_func( WX_GET_CLIENT_WINDOW, {parent_frame})
end function

--/topic wxMDIParentFrame
--/func get_mdi_window_menu( atom parent_frame )
--
--Returns the current Window menu (added by wxWidgets to the menubar). This function is available under Windows only.
public function get_mdi_window_menu( atom parent_frame )
    return c_func( WX_GET_MDI_WINDOW_MENU, {parent_frame})
end function



--wxMediaCtrl
-------------------------
--/topic wxMediaCtrl
--/parent Controls
--/info
--
--wxMediaCtrl is a class for displaying types of media, such as videos, audio files,
--natively through native codecs.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence caption [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li atom style [= 0]
--</ul>

--/topic wxMediaCtrl
--/func load_media( atom media, sequence fileName )
--Loads the file that fileName refers to. Returns false if loading fails.
public function load_media( atom media, sequence file )
    return c_func( WX_LOAD_MEDIA, {media, file})
end function

--/topic wxMediaCtrl
--/func load_media_uri( atom media, object uri )
--
--Loads the location that /i uri refers to. Note that this is very implementation-dependant, although HTTP URI/URLs
-- are generally supported, for example. Returns false if loading fails.
public function load_media_uri( atom media, object uri )
    return c_func( WX_LOAD_MEDIA_URI, {media,uri})
end function

--/topic wxMediaCtrl
--/func play_media( atom media )
--
--Resumes playback of the movie.
public function play_media( atom media )
    return c_func( WX_PLAY_MEDIA, {media})
end function

--/topic wxMediaCtrl
--/func pause_media( atom media )
--
--Pauses playback of the movie.
public function pause_media( atom media )
    return c_func( WX_PAUSE_MEDIA, {media})
end function

--/topic wxMediaCtrl
--/func stop_media( atom media )
--
--Stops the media.
public function stop_media( atom media )
    return c_func( WX_STOP_MEDIA, {media})
end function

--/topic wxMediaCtrl
--/func show_player_controls( atom media, integer flags )
--
--A special feature to wxMediaCtrl. Applications using native toolkits such as QuickTime usually have a scrollbar,
-- play button, and more provided to them by the toolkit. By default wxMediaCtrl does not do this. However, on the
-- directshow and quicktime backends you can show or hide the native controls provided by the underlying toolkit at
-- will using ShowPlayerControls. Simply calling the function with default parameters tells wxMediaCtrl to use the
-- default controls provided by the toolkit. The function takes a wxMediaCtrlPlayerControls enumeration as follows:
--
--<ul>
--/li /i wxMEDIACTRLPLAYERCONTROLS_NONE     No controls. return wxMediaCtrl to it's default state.
--/li /i wxMEDIACTRLPLAYERCONTROLS_STEP     Step controls like fastfoward, step one frame etc.
--/li /i wxMEDIACTRLPLAYERCONTROLS_VOLUME   Volume controls like the speaker icon, volume slider, etc.
--/li /i wxMEDIACTRLPLAYERCONTROLS_DEFAULT  Default controls for the toolkit. Currently a typedef for wxMEDIACTRLPLAYERCONTROLS_STEP
-- and wxMEDIACTRLPLAYERCONTROLS_VOLUME.
--</ul>
public function show_player_controls( atom media, integer flags )
    return c_func( WX_SHOW_PLAYER_CONTROLS, {media, flags})
end function

--/topic wxMediaCtrl
--/func set_media_volume( atom media, atom volume )
--
--Sets the volume of the media from a 0.0 to 1.0 range to that referred by dVolume. 1.0 represents full volume,
-- while 0.5 represents half (50 percent) volume, for example. Note that this may not be exact due to conversion
-- and rounding errors, although setting the volume to full or none is always exact. Returns true if successful.
public function set_media_volume( atom media, atom volume )
    return c_func( WX_SET_MEDIA_VOLUME, {media, volume})
end function

--/topic wxMediaCtrl
--/func get_media_state( atom media )
--
--Obtains the state the playback of the media is in -
--<ul>
--/li /i wxMEDIASTATE_STOPPED   The movie has stopped.
--/li /i wxMEDIASTATE_PAUSED    The movie is paused.
--/li /i wxMEDIASTATE_PLAYING   The movie is currently playing.
--</ul>
public function get_media_state( atom media )
    return c_func( WX_GET_MEDIA_STATE, {media})
end function

--/topic wxMediaCtrl
--/func get_media_volume( atom media )
--
--Gets the volume of the media from a 0.0 to 1.0 range. Note that due to rounding and other errors
-- this may not be the exact value sent to /set_media_volume.
public function get_media_volume( atom media )
    return c_func( WX_GET_MEDIA_VOLUME, {media})
end function

--/topic wxMediaCtrl
--/func get_media_length( atom media )
--
--Obtains the length - the total amount of time the movie has in milliseconds.
public function get_media_length( atom media )
    return c_func( WX_GET_MEDIA_LENGTH, {media})
end function

--/topic wxMediaCtrl
--/func get_media_position( atom media )
--
--Obtains the current position in time within the movie in milliseconds.
public function get_media_position( atom media )
    return c_func( WX_GET_MEDIA_POSITION, {media})
end function

--/topic wxMediaCtrl
--/proc seek_media( atom media, atom where, atom mode )
--
--Seeks to a position within the movie.
public procedure seek_media( atom media, atom where, atom mode )
    c_proc( WX_SEEK_MEDIA, {media,where,mode})
end procedure



--wxMemoryDC
------------------------------
--/topic wxMemoryDC
--/parent wxDC
--/parent Graphics
--/parent Classes
--/info
--wxMemoryDCs are used to draw to /wxBitmaps in memory.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li atom /wxBitmap [= 0]
--</ul>
--
--You must either delete the /b wxMemoryDC (with /delete_instance) or call /select_object(
-- wxNullBitmap ) before selecting the bitmap into another /b wxMemoryDC.

--/topic wxMemoryDC
--/proc select_object( atom dc, atom obj )
--
--Selects a bitmap into the wxMemoryDC.  If /i obj is /b wxNullBitmap,
--then the bitmap currently selected into the DC is deselected.
public procedure select_object( atom dc, atom obj )
    c_proc( WX_SELECT_OBJECT, {dc, obj} )
end procedure



--wxMenu
------------------------------
--/topic wxMenu
--/parent Menus
--/parent Controls
--/info
--A menu is the top level menu item, such as "File".
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li To add a menu to a /wxMenuBar
--<ul>
--/li /b "atom parent"
--/li /b "seqeunce text"
--/li integer style [= 0]
--/li integer index [= -1]
--<ul>
--If omitted, the menu is appended to the existing menus,
-- or a zero-based index where the menu should be inserted
--</ul>
--</ul>
--/li To create a popup- or sub-menu
--<ul>
--/li /b "atom parent" (must be 0)
--</ul>
--</ul>
--
--To create a submenu, first create a /wxMenu object with no parent,
--then create a new /wxMenuItem with the menu as the /b subMenu parameter.

--/topic wxMenu
--/proc append_separator( atom menu )
--
--Adds a separator to the end of the menu.
public procedure append_separator( atom menu )
    c_proc( WX_APPEND_SEPARATOR, {menu})
end procedure

--/topic wxMenu
--/proc insert_separator( atom menu, integer pos )
--
--Inserts a separator at the given position.
public procedure insert_separator( atom menu, integer pos )
    c_proc( WX_INSERT_SEPARATOR, {menu, pos})
end procedure

--/topic wxMenu
--/func get_menuitem_count( atom menu )
--
--Returns the number of items in the menu.
public function get_menuitem_count( atom menu )
    return c_func( WX_GET_MENUITEM_COUNT, {menu} )
end function

procedure insert_menu( atom menubar, atom menu, sequence title, integer index )
    c_proc( WX_INSERT_MENU, {menubar, menu, title, index} )
end procedure

procedure popup_menu( atom window, atom menu, atom x, atom y )
    c_proc( WX_POPUP_MENU, {window,menu,x,y})
end procedure

--/topic wxMenu
--/func find_menuitem_id( atom menu, object label_text )
--
public function find_menuitem_id( atom menu, object label_text )
    return c_func( WX_FIND_MENUITEM_ID, {menu,label_text})
end function

--/topic wxMenu
--/func find_menuitem_by_position( atom menu, atom position )
--
public function find_menuitem_by_position( atom menu, atom position )
    return c_func( WX_FIND_MENUITEM_BY_POSITION, {menu,position})
end function



--wxMenuBar
------------------------------
--/topic wxMenuBar
--/parent Menus
--/parent Controls
--/info
--wxMenuBars must have a /wxFrame parent.  The only parameter to create a wxMenuBar is the pointer to a /wxFrame.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--</ul>

--/topic wxMenuBar
--/proc add_menu( atom menubar, atom menu, sequence title )
--
--Adds the item to the end of the menu bar.
public procedure add_menu( atom menubar, atom menu, sequence title )
    c_proc( WX_ADD_MENU, {menubar, menu, title} )
end procedure

--/topic wxMenuBar
--/func get_menu_by_index( atom menubar, atom pos )
--
public function get_menu_by_index( atom menubar, atom pos )
    return c_func( WX_GET_MENU_BY_INDEX, {menubar,pos})
end function

--/topic wxMenuBar
--/func get_menu_by_label( atom menubar, object label_text )
--
public function get_menu_by_label( atom menubar, object label_text )
    return c_func( WX_GET_MENU_BY_LABEL, {menubar,label_text})
end function



--wxMenuItem
------------------------------
--/topic wxMenuItem
--/parent Menus
--/parent Controls
--/info
--wxMenuItems must have a /wxMenu parent.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li /b "integer id"
--/li /b "sequence text"
--/li sequence helpString [= ""]
--/li integer kind [= wxITEM_NORMAL] (wxITEM_SEPARATOR, wxITEM_NORMAL, wxITEM_CHECK or wxITEM_RADIO)
--/li atom subMenu [= 0]
--/li atom bitmap [= 0]
--/li integer width [= -1]
--/li integer height [= -1]
--/li integer quality [= wxIMAGE_QUALITY_NORMAL]
--<ul>
--/li wxIMAGE_QUALITY_NORMAL
--/li wxIMAGE_QUALITY_HIGH
--</ul>
--</ul>
--
--If specifying a /i bitmap, you can also specify a /i width and /i height to
-- rescale the image. You may also specify a /i quality to use during the rescaling.
--
--To create a submenu, you first need to create a /wxMenu with no parent, and
-- pass the result in the /b subMenu parameter.
--
--To set up event handlers for menus, use the wxEVT_COMMAND_MENU_SELECTED event type.
-- /wxMenu controls now derive from wxEvtHandler, so you should set events with the
-- menu now, not a parent control.
--
--Example:
--/code
--    procedure on_open_menu( atom this, atom event, atom id, atom event_type )
--        -- do stuff
--    end procedure
--
--    -- if you used a specific wxID_ constant when you created the wxMenuItem, you could
--    -- use that instead of get_menuitem_id: wxID_OPEN, wxID_NEW, etc
--
--    set_event_handler( the_open_menu_parent, get_menuitem_id( the_open_menu ),
--        wxEVT_COMMAND_MENU_SELECTED, routine_id("on_open_menu"))
--/endcode



--wxMouseEvent
----------------------------------
--/topic wxMouseEvent
--/parent Events

--/topic wxMouseEvent
--/func get_wheel_rotation( atom event )
--
--Get wheel rotation, positive or negative indicates direction of rotation. Current devices all send an event when
-- rotation is at least +/-WheelDelta, but finer resolution devices can be created in the future. Because of this
-- you shouldn't assume that one event is equal to 1 line, but you should be able to either do partial line scrolling
-- or wait until several events accumulate before scrolling.
public function get_wheel_rotation( atom event )
    return c_func( WX_GET_WHEEL_ROTATION, {event})
end function



--wxMultiChoiceDialog
-----------------------------------------
--/topic wxMultiChoiceDialog
--/parent Dialogs
--/info
--
--This class represents a dialog that shows a list of strings, and allows
--the user to select one or more.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li /b "sequence message"
--/li /b "sequence caption"
--/li /b "sequence choices "
--/li atom style [= wxCHOICEDLG_STYLE]
--/li integer pos x [= -1]
--/li integer pos y [= -1]
--</ul>
--pos x,y not supported in Win32.
--
--Pass a sequence of zero based index to /set_selection().
--
--/get_selection() will return a sequence of indexes, or an empty sequence.



--wxNotebook
------------------------------
--/topic wxNotebook
--/parent Notebook Controls
--/parent Controls
--/info
--This class represents a notebook control, which manages multiple windows with associated tabs.
--
--To use the class, create a wxNotebook object and call /add_page, passing a window to be
--used as the page.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cy [=-1]
--/li integer style [=0]
--<ul>
--/li wxNB_FIXEDWIDTH  (Windows only)All tabs will have same width.
--/li wxNB_LEFT  Place tabs on the left side.
--/li wxNB_RIGHT  Place tabs on the right side.
--/li wxNB_BOTTOM  Place tabs under instead of above the notebook pages.
--</ul></ul>
--Note that sometimes you can reduce flicker by passing the wxCLIP_CHILDREN window style.

--/topic wxNotebook
--/proc set_padding ( atom book, sequence size )
--
--Sequence is divided into X and Y amounts. X is the amount to add to to the wideness of
--the wxNotebook's tab's and Y is the amount that the text on the tabs will move down. If
--the values are negative, then X will subtract from the wideness of the tabs, and Y will
--make the text move up.
public procedure set_padding ( atom book, sequence size )
    c_proc( WX_SET_PADDING, {book, size} )
end procedure

--/topic wxNotebook
--/func notebook_event_selection( atom event )
--
--Returns the currently selected page for /i event, or -1 if none was selected.
public function notebook_event_selection( atom event )
    return c_func( WX_NOTEBOOK_EVENT_SELECTION, {event} )
end function

--/topic wxNotebook
--/proc set_tab_size ( atom book, sequence size )
--
--This works much like set_padding. However instead of increasing or decreasing the size,
-- the tabs are set to have the width specified by the first value in size, and the length
-- of the second value in size. However, unlike set_padding, a negative value is not to
-- be used.
public procedure set_tab_size ( atom book, sequence size )
    c_proc( WX_SET_TAB_SIZE, {book, size} )
end procedure

--/topic wxNotebook
--/proc advance_selection( atom book, integer forward )
--
--Cycles through the tabs.
public procedure advance_selection( atom book, integer forward )
    c_proc( WX_ADVANCE_SELECTION, {book, forward} )
end procedure

--/topic wxNotebook
--/proc delete_page( atom book, integer page )
--
--Deletes the specified page, and the associated window.
public procedure delete_page( atom book, integer page )
    c_proc( WX_DELETE_PAGE, {book, page} )
end procedure

--/topic wxNotebook
--/func get_page_text( atom book, atom page )
--
--Returns the string for the given page.
public function get_page_text( atom book, atom page )
    return c_func( WX_GET_PAGE_TEXT, {book,page})
end function


--wxOutputStream
------------------------------
--/topic wxOutputStream
--/parent Files
--/parent Classes

--/topic wxOutputStream
--/proc stream_write( atom stream, object bytes )
--
--Writes data to the output queue.
public procedure stream_write( atom stream, object bytes )
    c_proc( WX_STREAM_WRITE, {stream,bytes})
end procedure


--wxPageSetupDialog
------------------------------
--/topic wxPageSetupDialog
--/parent Printing
--/parent Dialogs
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li atom parent [= 0]
--/li atom /wxPageSetupDialogData [= 0]
--</ul>

--/topic wxPageSetupDialog
--/func get_pagesetup_data( atom ps )
--
--Returns the page setup data associated with the dialog.
public function get_pagesetup_data( atom ps )
    return c_func( WX_GET_PAGESETUP_DATA, {ps} )
end function



--wxPageSetupDialogData
------------------------------
--/topic wxPageSetupDialogData
--/parent wxPageSetupDialog
--/parent Printing
--/parent Classes
--/info
--



--wxPaintDC
------------------------------
--/topic wxPaintDC
--/parent wxDC
--/parent Graphics
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom /wxWindow" The window that is being painted.
--</ul>
--
--Device context that is used to draw to a window during a paint event.
-- It should be destroyed before you exit the paint handler.  Under windows,
-- you /b must create a wxPaintDC during a paint event. If you make a call to
-- /wx_puts or /wx_printf, wxEuphoria may create one for you (see the documentation
-- for those routines for more information).
--
--To draw on a /wxWindow outside of a paint event, see /wxClientDC.



--wxPanel
------------------------------
--/topic wxPanel
--/parent wxWindow
--/parent Controls
--/info
--If you create a /wxFrame, you should create a /b wxPanel to be the child of the main frame, and place
--all controls into the panel (except for menus and /wxStatusBar 's). The only creation
--parameter required is a parent.  The panel will automatically grow or shrink whenever
--its parent does, and automatically handles tabbing between controls.
--

--/topic wxPanel
--/proc set_default_item( atom panel, atom button )
--
--Changes the default /i button for the /i panel.
public procedure set_default_item( atom panel, atom button )
    c_proc( WX_SET_DEFAULT_ITEM, {panel, button} )
end procedure



--wxPasswordEntryDialog
-----------------------------------------
--/topic wxPasswordEntryDialog
--/parent Dialogs
--/info
--
--Implementation of wxGenericDialog that prompts for a password.
--Use /show_modal to display and /get_text_value to retrieve the entered value.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li /b "sequence message"
--/li sequence caption [= "Enter Password"]
--/li sequence default [= ""]
--/li atom style [= wxOK | wxCANCEL | wxCENTRE]
--/li integer pos_x [= -1]
--/li integer pos_y [= -1]
--</ul>
--
--pos x,y not supported on Win32



--wxPen
------------------------------
--/topic wxPen
--/parent Graphics
--/parent Classes
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom /wxColor"
--/li /b "integer width"
--/li /b "integer style"
--<ul>
--/li wxSOLID
--/li wxDOT
--/li wxLONG_DASH
--/li wxSHORT_DASH
--/li wxDOT_DASH
--/li wxUSER_DASH
--/li wxTRANSPARENT
--</ul></ul>



--wxPlatformInfo
---------------------------------------
--/topic wxPlatformInfo
--/parent Classes
--/info
--
--This class holds informations about the operating system and the toolkit that the application is running under and some basic
-- architecture info of the machine where it's running.
--
--The following are the operating systems which are recognized by wxWidgets and whose version can be detected at run-time. The
-- values of the constants are chosen so that they can be combined as flags; this allows to check for operating system /i families like
-- e.g. wxOS_MAC and wxOS_UNIX.
--
--<ul><table>
--<tr><td> /li  wxOS_UNKNOWN            <td> returned on error
--<tr><td> /li  wxOS_MAC_OS             <td> Apple Mac OS 8/9/X with Mac paths
--<tr><td> /li  wxOS_MAC_OSX_DARWIN     <td> Apple Mac OS X with Unix paths
--<tr><td> /li  wxOS_MAC                <td> wxOS_MAC_OS | wxOS_MAC_OSX_DARWIN,
--<tr><td> /li  wxOS_WINDOWS_9X         <td> Windows 9x family (95/98/ME)
--<tr><td> /li  wxOS_WINDOWS_NT         <td> Windows NT family (NT/2000/XP)
--<tr><td> /li  wxOS_WINDOWS_MICRO      <td> MicroWindows
--<tr><td> /li  wxOS_WINDOWS_CE         <td> Windows CE (Window Mobile)
--<tr><td> /li  wxOS_WINDOWS            <td>wxOS_WINDOWS_9X | wxOS_WINDOWS_NT | wxOS_WINDOWS_MICRO | wxOS_WINDOWS_CE
--<tr><td> /li  wxOS_UNIX_LINUX         <td> Linux
--<tr><td> /li  wxOS_UNIX_FREEBSD       <td> FreeBSD
--<tr><td> /li  wxOS_UNIX_OPENBSD       <td> OpenBSD
--<tr><td> /li  wxOS_UNIX_NETBSD        <td> NetBSD
--<tr><td> /li  wxOS_UNIX_SOLARIS       <td> SunOS
--<tr><td> /li  wxOS_UNIX_AIX           <td> AIX
--<tr><td> /li  wxOS_UNIX_HPUX          <td> HP/UX
--<tr><td valign="top"> /li wxOS_UNIX   <td> wxOS_UNIX_LINUX | wxOS_UNIX_FREEBSD | wxOS_UNIX_OPENBSD | wxOS_UNIX_NETBSD
-- | wxOS_UNIX_SOLARIS | wxOS_UNIX_AIX | wxOS_UNIX_HPUX
--<tr><td> /li  wxOS_DOS                <td> Microsoft DOS
--<tr><td> /li  wxOS_OS2                <td> OS/2
--</table></ul>
--
--The list of wxWidgets ports. Some of them can be used with more than a single (native) toolkit; e.g. wxWinCE port sources can be
-- used with smartphones, pocket PCs and handheld devices SDKs.
--
--<ul><table>
--<tr><td> /li  wxPORT_UNKNOWN  <td> returned on error
--<tr><td> /li  wxPORT_BASE     <td> wxBase, no native toolkit used
--<tr><td> /li  wxPORT_MSW      <td> wxMSW, native toolkit is Windows API
--<tr><td> /li  wxPORT_MOTIF    <td> wxMotif, using [Open]Motif or Lesstif
--<tr><td> /li  wxPORT_GTK      <td> wxGTK, using GTK+ 1.x, 2.x, GPE or Maemo
--<tr><td> /li  wxPORT_MGL      <td> wxMGL, using wxUniversal
--<tr><td> /li  wxPORT_X11      <td> wxX11, using wxUniversal
--<tr><td> /li  wxPORT_OS2      <td> wxOS2, using OS/2 Presentation Manager
--<tr><td> /li  wxPORT_MAC      <td> wxMac, using Carbon or Classic Mac API
--<tr><td> /li  wxPORT_COCOA    <td> wxCocoa, using Cocoa NextStep/Mac API
--<tr><td> /li  wxPORT_WINCE    <td> wxWinCE, toolkit is WinCE SDK API
--<tr><td> /li  wxPORT_PALMOS   <td> wxPalmOS, toolkit is PalmOS API
--<tr><td> /li  wxPORT_DFB      <td> wxDFB, using wxUniversal
--</table></ul>
--
--The architecture of the operating system (regardless of the build environment of wxWidgets library - see /is_platform_64bit
-- documentation for more info).
--
--<ul><table>
--<tr><td> /li wxARCH_INVALID   <td>returned on error
--<tr><td> /li wxARCH_32        <td> 32 bit
--<tr><td> /li wxARCH_64        <td> 64 bit
--<tr><td> /li wxARCH_MAX       <td> &nbsp;
--</table></ul>
--
--The endian-ness of the machine.
--
--<ul><table>
--<tr><td> /li wxENDIAN_INVALID     <td> returned on error
--<tr><td> /li wxENDIAN_BIG         <td> 4321
--<tr><td> /li wxENDIAN_LITTLE      <td> 1234
--<tr><td> /li wxENDIAN_PDP         <td> 3412
--<tr><td> /li wxENDIAN_MAX         <td> &nbsp;
--</table></ul>

--/topic wxPlatformInfo
--/func check_os_version( atom major, atom minor )
--
--Returns true if the OS version is at least major.minor.
public function check_os_version( atom major, atom minor )
    return c_func( WX_CHECK_OS_VERSION, {major,minor})
end function

--/topic wxPlatformInfo
--/func check_toolkit_version( atom major, atom minor )
--
--Returns true if the toolkit version is at least major.minor.
public function check_toolkit_version( atom major, atom minor )
    return c_func( WX_CHECK_TOOLKIT_VERSION, {major,minor})
end function

--/topic wxPlatformInfo
--/func get_arch_from_name( object arch )
--
--Converts the given string to a wxArchitecture enum value or to wxARCH_INVALID if the given string is not a valid architecture
-- string (i.e. does not contain nor 32 nor 64 strings).
public function get_arch_from_name( sequence arch )
    return c_func( WX_GET_ARCH_FROM_NAME, {arch})
end function

--/topic wxPlatformInfo
--/func get_arch_name()
--
--Returns the name for the architecture of this /b wxPlatformInfo instance.
public function get_arch_name( )
    return c_func( WX_GET_ARCH_NAME, {})
end function

--/topic wxPlatformInfo
--/func get_arch()
--
--Converts the given string to a wxArchitecture enum value or to wxARCH_INVALID if the given string is not a valid architecture
-- string (i.e. does not contain nor 32 nor 64 strings).
public function get_arch(  )
    return c_func( WX_GET_ARCH, {})
end function

--/topic wxPlatformInfo
--/func get_endianness_from_name( object endian )
--
--Converts the given string to a wxEndianness enum value or to wxENDIAN_INVALID if the given string is not a valid endianness
-- string (i.e. does not contain nor little nor big strings).
public function get_endianness_from_name( sequence endian )
    return c_func( WX_GET_ENDIANNESS_FROM_NAME, {endian})
end function

--/topic wxPlatformInfo
--/func get_endianness()
--
--Returns the endianness ID of this /b wxPlatformInfo instance.
public function get_endianness(  )
    return c_func( WX_GET_ENDIANNESS, {})
end function

--/topic wxPlatformInfo
--/func get_endianness_name_from_enum( atom endian )
--
--Returns name for the given wxEndianness enumeration value.
public function get_endianness_name_from_enum( atom endian )
    return c_func( WX_GET_ENDIANNESS_NAME_FROM_ENUM, {endian})
end function

--/topic wxPlatformInfo
--/func get_endianness_name()
--
--Returns the name for the endianness of this /b wxPlatformInfo instance.
public function get_endianness_name(  )
    return c_func( WX_GET_ENDIANNESS_NAME, {})
end function

--/topic wxPlatformInfo
--/func get_os_major_version()
--
--Returns the run-time major version of the OS associated with this /b wxPlatformInfo instance.
public function get_os_major_version(  )
    return c_func( WX_GET_OS_MAJOR_VERSION, {})
end function

--/topic wxPlatformInfo
--/func get_os_minor_version()
--
--Returns the run-time minor version of the OS associated with this /b wxPlatformInfo instance.
public function get_os_minor_version(  )
    return c_func( WX_GET_OS_MINOR_VERSION, {})
end function

--/topic wxPlatformInfo
--/func get_os_family_name_from_enum( atom os )
--
--Returns the operating system family name for the given wxOperatingSystemId enumeration value: Unix for wxOS_UNIX,
-- Macintosh for wxOS_MAC, Windows for wxOS_WINDOWS, DOS for wxOS_DOS, OS/2 for wxOS_OS2.
public function get_os_family_name_from_enum( atom os )
    return c_func( WX_GET_OS_FAMILY_NAME_FROM_ENUM, {os})
end function

--/topic wxPlatformInfo
--/func get_os_family_name()
--
--Returns the operating system family name of the OS associated with this /b wxPlatformInfo instance.
public function get_os_family_name()
    return c_func( WX_GET_OS_FAMILY_NAME, {})
end function

--/topic wxPlatformInfo
--/func get_os_id_from_name( object os )
--
--Converts the given string to a wxOperatingSystemId enum value or to wxOS_UNKNOWN if the given string is not a valid operating
-- system name.
public function get_os_id_from_name( sequence os )
    return c_func( WX_GET_OS_ID_FROM_NAME, {os})
end function

--/topic wxPlatformInfo
--/func get_os_id()
--
--Returns the operating system ID of this /b wxPlatformInfo instance.
public function get_os_id(  )
    return c_func( WX_GET_OS_ID, {})
end function

--/topic wxPlatformInfo
--/func get_os_id_name_from_enum( atom os )
--
--Returns the name for the given operating system ID value. This can be a long name (e.g. /b "Microsoft Windows NT"); use
-- /get_os_family_name to retrieve a short, generic name.
public function get_os_id_name_from_enum( atom os )
    return c_func( WX_GET_OS_ID_NAME_FROM_ENUM, {os})
end function

--/topic wxPlatformInfo
--/func get_os_id_name()
--
--Returns the operating system name of the OS associated with this /b wxPlatformInfo instance.
public function get_os_id_name(  )
    return c_func( WX_GET_OS_ID_NAME, {})
end function

--/topic wxPlatformInfo
--/func get_port_id_from_name( object portname )
--
--Converts the given string to a wxWidgets port ID value or to wxPORT_UNKNOWN if the given string does not match any of the
-- wxWidgets canonical name ports ("wxGTK", "wxMSW", etc) nor any of the short wxWidgets name ports ("gtk", "msw", etc).
public function get_port_id_from_name( sequence portname )
    return c_func( WX_GET_PORT_ID_FROM_NAME, {portname})
end function

--/topic wxPlatformInfo
--/func get_port_id()
--
--Returns the wxWidgets port ID associated with this /b wxPlatformInfo instance.
public function get_port_id(  )
    return c_func( WX_GET_PORT_ID, {})
end function

--/topic wxPlatformInfo
--/func get_port_id_name_from_enum( atom port, atom using_universal )
--
--Returns the name of the given wxWidgets port ID value. The /i using_universal argument specifies whether the port is in its
-- native or wxUniversal variant.
--
--The returned string always starts with the "wx" prefix and is a mixed-case string.
public function get_port_id_name_from_enum( atom port, atom using_universal )
    return c_func( WX_GET_PORT_ID_NAME_FROM_ENUM, {port,using_universal})
end function

--/topic wxPlatformInfo
--/func get_port_id_name()
--
--Returns the name of the wxWidgets port ID associated with this /b wxPlatformInfo instance.
public function get_port_id_name(  )
    return c_func( WX_GET_PORT_ID_NAME, {})
end function

--/topic wxPlatformInfo
--/func get_port_id_short_name_from_enum( atom port, atom using_universal )
--
--Returns the short name of the given wxWidgets port ID value. The /i using_universal argument specifies whether the port is in its
-- native or wxUniversal variant.
--
--The returned string does not start with the "wx" prefix and is always lower case.
public function get_port_id_short_name_from_enum( atom port, atom using_universal )
    return c_func( WX_GET_PORT_ID_SHORT_NAME_FROM_ENUM, {port,using_universal})
end function

--/topic wxPlatformInfo
--/func get_port_id_short_name()
--
--Returns the short name of the wxWidgets port ID associated with this /b wxPlatformInfo instance.
public function get_port_id_short_name(  )
    return c_func( WX_GET_PORT_ID_SHORT_NAME, {})
end function

--/topic wxPlatformInfo
--/func get_toolkit_major_version()
--
--Returns the run-time major version of the toolkit associated with this /b wxPlatformInfo instance. Note that if GetPortId returns
-- wxPORT_BASE, then this value is zero; that is, no native toolkit is in use.
public function get_toolkit_major_version(  )
    return c_func( WX_GET_TOOLKIT_MAJOR_VERSION, {})
end function

--/topic wxPlatformInfo
--/func get_toolkit_minor_version()
--
--Returns the run-time minor version of the toolkit associated with this /b wxPlatformInfo instance. Note that if GetPortId returns
-- wxPORT_BASE, then this value is zero; that is, no native toolkit is in use.
public function get_toolkit_minor_version(  )
    return c_func( WX_GET_TOOLKIT_MINOR_VERSION, {})
end function

--/topic wxPlatformInfo
--/func platform_info_ok()
--
--Returns true if this instance is fully initialized with valid values.
public function platform_info_ok(  )
    return c_func( WX_PLATFORM_INFO_OK, {})
end function

--/topic wxPlatformInfo
--/func using_universal_widgets()
--
--Returns true if this /b wxPlatformInfo describes /b wxUniversal build.
public function using_universal_widgets(  )
    return c_func( WX_USING_UNIVERSAL_WIDGETS, {})
end function

--/topic wxPlatformInfo
--/func get_arch_name_from_enum( atom arch )
--
--Returns the name for the given /b wxArchitecture enumeration value.
public function get_arch_name_from_enum( atom arch )
    return c_func( WX_GET_ARCH_NAME_FROM_ENUM, {arch})
end function



--wxPopupWindow
-----------------------------------------
--/topic wxPopupWindow
--/parent Controls
--/info
--
--A special kind of top level window used for popup menus, combobox popups and such.

--/topic wxPopupWindow
--/proc position_popup( atom popup, object origin, object size )
--
--Move the popup window to the right position, i.e. such that it is entirely visible.
--
--The popup is positioned at /i origin + /i size if it opens below and to the right (default), at /i origin - /i size
-- if it opens above and to the left etc.
--
--/b Parameters:
--<ul>
--/li /i origin Must be given in screen coordinates!
--/li /i size The size of the popup window
--</ul>
public procedure position_popup( atom popup, object origin, object size )
    c_proc( WX_POSITION_POPUP, {popup,origin,size})
end procedure



--wxPopupTransientWindow
-----------------------------------------
--/topic wxPopupTransientWindow
--/parent wxPopupWindow
--/info
--
--A special kind of top level window used for popup menus, combobox popups and such, which automatically
-- dismisses itself when it loses focus.



--wxPrintDialog
------------------------------
--/topic wxPrintDialog
--/parent Printing
--/parent Dialogs
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b parent [= 0]
--/li /b /wxPageSetupDialogData [= 0]
--</ul>

--/topic wxPrintDialog
--/func get_print_data( atom prdlg )
--
--Returns the print_data from a wxPrintDialog.
public function get_print_data( atom prdlg )
    return c_func( WX_GET_PRINT_DATA, {prdlg} )
end function



--wxPrinter
------------------------------
--/topic wxPrinter
--/parent Printing
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom data" (a /wxPrintDialogData object)
--</ul>

--/topic wxPrinter
--/func start_print( atom printer, atom parent atom printout, atom prompt )
--
--Starts the printing process. Provide a parent window, a user-defined
--wxPrintout object which controls the printing of a document, and whether
--the print dialog should be invoked first.
--
--Print could return FALSE if there was a problem initializing the printer
--device context (current printer not set, for example) or the user cancelled
--printing. Call wxPrinter::GetLastError to get detailed information about the
--kind of the error.
public function start_print( atom printer, atom parent, atom printout, atom prompt )
    return c_func( WX_START_PRINT, {printer, parent, printout, prompt} )
end function



--wxPrintout
------------------------------
--/topic wxPrintout
--/parent Printing
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "sequence title"
--/li /b "integer printpage_rid"
--/li /b "integer print_prepare_rid"
--/li sequence page_range [= {1,1}]
--</ul>
--
--The /i printpage_rid parameter is the routine_id for a function that
--will be called when a page needs to be printed. The function should take
--three arguments: a /i this object, /i page_num integer and a /wxDC. You
-- should draw the page into the supplied dc.  If you want to cancel the
-- job return a value of zero. Otherwise, you should return a value of 1.
--
--The /i print_prepare_rid parameter is the routine_id for a procedure
--that will be called when a page needs to be printed.  It should take two
--arguments, a /wxPrintout object and a /wxDC.  You can use this routine to
--determine the number of pages required, if you need the DC in order to
--do so.  You should then call /set_page_info() so that the correct number
--of pages will be used.  If you don't need this, you can set this parameter
--to -1, which is the default value.

--/topic wxPrintout
--/func get_printout_dc( atom printout )
--
--Returns the device context associated with the printout (given to the printout at start of printing or previewing).
-- This will be a wxPrinterDC if printing under Windows or Mac, a wxPostScriptDC if printing on other platforms,
-- and a wxMemoryDC if previewing.
public function get_printout_dc( atom printout )
    return c_func( WX_GET_PRINTOUT_DC, {printout} )
end function

--/topic wxPrintout
--/proc set_page_info( atom printout, atom min, atom max )
--
--This routine sets the max and min pages for a printout object.
--If you need to have the DC of the printer in order to determine
--the number of pages,
public procedure set_page_info( atom printout, atom min, atom max )
    c_proc( WX_SET_PAGE_INFO, {printout, min, max} )
end procedure

--/topic wxPrintout
--/func get_ppi_printer( atom printout )
--
--Returns the number of pixels per logical inch of the printer device
--context. Dividing the printer PPI by the screen PPI can give a suitable
--scaling factor for drawing text onto the printer. Remember to multiply
--this by a scaling factor to take the preview DC size into account.
public function get_ppi_printer( atom printout )
    return c_func( WX_GET_PPI_PRINTER, {printout} )
end function

--/topic wxPrintout
--/func get_ppi_screen( atom printout )
--
--Returns the number of pixels per logical inch of the screen device
--context. Dividing the printer PPI by the screen PPI can give a suitable
--scaling factor for drawing text onto the printer. Remember to multiply
--this by a scaling factor to take the preview DC size into account.
public function get_ppi_screen( atom printout )
    return c_func( WX_GET_PPI_SCREEN, {printout} )
end function

--/topic wxPrintout
--/func get_page_pixels( atom printout )
--
--Returns the size of the printer page in pixels. These may not be the
--same as the values returned from wxDC::GetSize if the printout is being
--used for previewing, since in this case, a memory device context is used,
--using a bitmap size reflecting the current preview zoom. The application
--must take this discrepancy into account if previewing is to be supported.
public function get_page_pixels( atom printout )
    return c_func( WX_GET_PAGE_PIXELS, {printout} )
end function

--/topic wxPrintout
--/func get_page_mm( atom printout )
--
--Returns the size of the printer page in millimeters.
public function get_page_mm( atom printout )
    return c_func( WX_GET_PAGE_MM, {printout} )
end function

--/topic wxPrintout
--/func is_preview( atom printout )
--
--Returns 1 if the printout is currently being used for previewing.
public function is_preview( atom printout )
    return c_func( WX_IS_PREVIEW, {printout} )
end function

--/topic wxPrintout
--/func get_scaling_factor( atom printout )
--
--Returns the ratio of screen to printer PPI (Pixels Per Inch).
-- This provides a suitable scaling factor for drawing text onto the printer.
public function get_scaling_factor( atom printout )
    sequence printer, screen
    printer = get_ppi_printer( printout )
    screen = get_ppi_screen( printout )

    return screen / printer
end function



--wxPrintPreview
------------------------------
--/topic wxPrintPreview
--/parent Printing
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom /wxPrintout"
--/li atom /wxPreviewFrame [= 0]
--</ul>
--
--If /i wxPrintout is non-null, then a Print...
-- button will be placed on the preview frame so that the
-- user can print directly from the preview interface
--
--If /i wxPreviewFrame is not supplied, one will be created, which
-- can be accessed through /get_preview_frame]
--
--Previewing is started by showing the preview frame.
--
--Properly scaling the print preview is fairly simple.  There are three pieces
--of data that are required:
--<ol>
--/li /get_dc_size()
--/li /get_page_pixels()
--/li /get_ppi_printer()
--</ol>
--
--You can get the scaling factor required by dividing the results of /get_dc_size() by
-- /get_page_pixels().  Pass the result to /set_user_scale().  Now, you can use coordinates
-- in inches multiplied by the printer ppi (relative to the margins):
--
--Example:
--/code
--
--    procedure onPrintPage( atom this, atom page, atom dc )
--        sequence ppi, factor
--
--        ppi = /get_ppi_printer( this )
--        if /is_preview( this ) then
--            factor = /get_dc_size( dc ) // /get_page_pixels( this )
--            /set_user_scale( dc, factor[1], factor[2] )
--         else
--            /set_user_scale( dc, 1, 1)
--         end if
--
--         -- print "Hello World" at 1", 1"
--         /wx_puts( 0 & ppi * {1,1} & dc, "Hello World" )
--    end procedure
--
--/endcode

--/topic wxPrintPreview
--/func get_preview_frame( atom preview )
--
--This returns an integer which serves as a pointer to the wxPrintPreview.
public function get_preview_frame( atom preview )
    return c_func( WX_GET_PREVIEW_FRAME, {preview} )
end function

--/topic wxPrintPreview
--/proc set_preview_frame( atom preview, atom frame )
--
--This sets the selected wxPrintPreview to be a print previewer for the specified wxFrame.
public procedure set_preview_frame( atom preview, atom frame )
    c_proc( WX_SET_PREVIEW_FRAME, {preview, frame} )
end procedure

--/topic wxPrintPreview
--/func get_zoom( atom preview )
--
--Gets the percentage preview zoom.
public function get_zoom( atom preview )
    return c_func( WX_GET_ZOOM, {preview} )
end function



--wxProcess
-----------------------------------------
--/topic wxProcess
--/parent Classes
--/info
--A class to launch, monitor, and control child processes.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer callback [= -1]
--</ul>
--
--To be notified of a process termination, set /i callback to the routine_id() of your /b OnTerminate() routine.
--
--Example:
--/code
--    atom ps, ps_pid
--
--    procedure OnTerminate( integer pid, integer status )
--
--        if pid = ps_pid then
--            -- our process has terminated!
--        end if
--
--    end procedure
--
--    ps = create( wxProcess, {parent, new_id(), routine_id("OnTerminate")} )
--
--    pid = start_process( ps, "command", wxEXEC_ASYNC )
--    if pid = 0 then
--        -- process failed to start
--    end if
--/endcode

--/topic wxProcess
--/proc close_process_output( atom id )
--
--Closes the output stream (the one connected to the stdin of the child process).
-- This function can be used to indicate to the child process that there is no more data to be
-- read - usually, a filter program will only terminate when the input stream is closed.
public procedure close_process_output( atom id )
    c_proc( WX_CLOSE_PROCESS_OUTPUT, {id})
end procedure

--/topic wxProcess
--/proc detach_process( atom id )
--
--Normally, a /b wxProcess object is deleted by its parent when it receives the notification about
-- the process termination. However, it might happen that the parent object is destroyed before
-- the external process is terminated (e.g. a window from which this external process was launched
-- is closed by the user) and in this case it should not delete the wxProcess object, but should
-- call /b detach_process() instead. After the /b wxProcess object is detached from its parent, no
-- notification events will be sent to the parent and the object will delete itself upon reception
-- of the process termination notification.
public procedure detach_process( atom id )
    c_proc( WX_DETACH_PROCESS, {id})
end procedure

--/topic wxProcess
--/func process_error_stream( atom id )
--
--Returns an input stream which corresponds to the standard error output (stderr) of the child process.
public function process_error_stream( atom id )
    return c_func( WX_PROCESS_ERROR_STREAM, {id})
end function

--/topic wxProcess
--/func process_input_stream( atom id )
--
--Returns an input stream corresponding to the standard output stream of the subprocess. If it is NULL,
-- you have not turned on the redirection. See /process_redirect().
public function process_input_stream( atom id )
    return c_func( WX_PROCESS_INPUT_STREAM, {id})
end function

--/topic wxProcess
--/func process_output_stream( atom id )
--
--Returns an output stream correspoding to the input stream of the subprocess. If it is NULL,
-- you have not turned on the redirection. See /redirect_process().
public function process_output_stream( atom id )
    return c_func( WX_PROCESS_OUTPUT_STREAM, {id})
end function

--/topic wxProcess
--/func process_error_available( atom id )
--
--Returns true if there is data to be read on the child process standard error stream.
-- See /process_input_available()
public function process_error_available( atom id )
    return c_func( WX_PROCESS_ERROR_AVAILABLE, {id})
end function

--/topic wxProcess
--/func process_input_available( atom id )
--
--Returns true if there is data to be read on the child process standard output stream.
-- This allows to write simple (and extremely inefficient) polling-based code waiting for a
-- better mechanism in future wxWidgets versions.
public function process_input_available( atom id )
    return c_func( WX_PROCESS_INPUT_AVAILABLE, {id})
end function

--/topic wxProcess
--/func process_input_opened( atom id )
--
--Returns true if the child process standard output stream is opened.
public function process_input_opened( atom id )
    return c_func( WX_PROCESS_INPUT_OPENED, {id})
end function

--/topic wxProcess
--/func kill_process( atom id, atom signal, atom flags )
--
--Send the specified /b signal to the given process. Possible /b signal values are:
--
--/code
--enum wxSignal
--{
--  wxSIGNONE = 0,  // verify if the process exists under Unix
--  wxSIGHUP,
--  wxSIGINT,
--  wxSIGQUIT,
--  wxSIGILL,
--  wxSIGTRAP,
--  wxSIGABRT,
--  wxSIGEMT,
--  wxSIGFPE,
--  wxSIGKILL,      // forcefully kill, dangerous!
--  wxSIGBUS,
--  wxSIGSEGV,
--  wxSIGSYS,
--  wxSIGPIPE,
--  wxSIGALRM,
--  wxSIGTERM       // terminate the process gently
--};
--/endcode
--
--wxSIGNONE, wxSIGKILL and wxSIGTERM have the same meaning under both Unix and Windows but all
-- the other signals are equivalent to wxSIGTERM under Windows.
--
--The /b flags parameter can be wxKILL_NOCHILDREN (the default), or wxKILL_CHILDREN, in which case
-- the child processes of this process will be killed too. Note that under Unix, for wxKILL_CHILDREN
-- to work you should have created the process passing wxEXEC_MAKE_GROUP_LEADER.
--
--Returns the element of wxKillError enum:
--/code
--enum wxKillError
--{
--  wxKILL_OK,              // no error
--  wxKILL_BAD_SIGNAL,      // no such signal
--  wxKILL_ACCESS_DENIED,   // permission denied
--  wxKILL_NO_PROCESS,      // no such process
--  wxKILL_ERROR            // another, unspecified error
--};
--/endcode
public function kill_process( atom id, atom signal, atom flags )
    return c_func( WX_KILL_PROCESS, {id,signal,flags})
end function

--/topic wxProcess
--/func process_exists( atom id )
--
--Returns true if the given process exists in the system.
public function process_exists( atom id )
    return c_func( WX_PROCESS_EXISTS, {id})
end function

--/topic wxProcess
--/func get_pid( atom id )
--
--Returns the process ID of the process launched by /process_open().
public function get_pid( atom id )
    return c_func( WX_GET_PID, {id})
end function

--/topic wxProcess
--/proc process_redirect( atom id )
--
--Turns on redirection. wxExecute will try to open a couple of pipes to catch the subprocess stdio.
-- The caught input stream is returned by /process_output_stream() as a non-seekable stream. The caught
-- output stream is returned by /process_input_stream() as a non-seekable stream.
public procedure process_redirect( atom id )
    c_proc( WX_PROCESS_REDIRECT, {id})
end procedure

--/topic wxProcess
--/func start_process( atom id, object cmd, atom flag )
--
--Executes a /i command using a /b wxProcess /i id. See /wx_execute() for /i flag values.
public function start_process( atom id, object cmd, atom flag )
    return c_func( WX_START_PROCESS, {id,cmd,flag})
end function



--wxProcessEvent
-----------------------------------------
--/topic wxProcessEvent
--/parent Events

--/topic wxProcessEvent
--/func process_event_pid( atom event )
--
--Returns the process id.
public function process_event_pid( atom event )
    return c_func( WX_PROCESS_EVENT_PID, {event})
end function

--/topic wxProcessEvent
--/func process_event_exitcode( atom event )
--
--Returns the exist status.
public function process_event_exitcode( atom event )
    return c_func( WX_PROCESS_EVENT_EXITCODE, {event})
end function



--wxProgressDialog
------------------------------
--/topic wxProgressDialog
--/parent Dialogs
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "sequence title"
--/li /b "sequence message"
--/li integer maximum [= 100]
--/li atom parent [= 0]
--/li integer style [= wxPD_AUTO_HIDE | wxPD_APP_MODAL]
--</ul>

--/topic wxProgressDialog
--/func progress_update( atom dialog, integer val, sequence message )
--
--Updates the dialog, setting the progress bar to the new value and,
--if given changes the message above it. Returns true unless the Cancel
--button has been pressed.
--
--If FALSE is returned, the application can either immediately destroy the
--dialog or ask the user for the confirmation and if the abort is not confirmed
--the dialog may be resumed with Resume function.
public function progress_update( atom dialog, integer val, sequence message )
    return c_func( WX_PROGRESS_UPDATE, {dialog, val, message} )
end function

--/topic wxProgressDialog
--/proc progress_resume( atom dialog )
--
--Can be used to continue with the dialog, after the user had chosen ABORT.
public procedure progress_resume( atom dialog )
    c_proc( WX_PROGRESS_RESUME, {dialog} )
end procedure



--wxRadioBox
------------------------------
--/topic wxRadioBox
--/parent wxRadioButton
--/parent Controls
--/info
--A radio box item is used to select one of number of mutually exclusive choices.
--It is displayed as a vertical column or horizontal row of labelled buttons.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence label_text [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li sequence choices [= {}]
--/li integer majorDimension [= 0]
--<ul>
--Specifies the maximum number of rows (if style contains /b wxRA_SPECIFY_ROWS)
-- or columns (if style contains /b wxRA_SPECIFY_COLS) for a two-dimensional radiobox.
--</ul>
--/li atom style [= wxRA_SPECIFY_COLS]
--<ul><table>
--<tr><td> /li wxRA_SPECIFY_ROWS <td>The major dimension parameter refers to the maximum number of rows.
--<tr><td> /li wxRA_SPECIFY_COLS <td>The major dimension parameter refers to the maximum number of columns.
--</table></ul>
--</ul>

--/topic wxRadioBox
--/func get_radio_selection( atom this )
--
--Return the 0-based index of the selected radio item in the wxRadioBox.
public function get_radio_selection( atom this )
    return c_func( WX_GET_RADIO_SELECTION, {this} )
end function

--/topic wxRadioBox
--/proc set_radiobox_selection( atom box, integer selection )
--
--This sets the Radiobox to being selected if selection is one, unselected if selection is zero.
public procedure set_radiobox_selection( atom box, integer selection )
    c_proc( WX_SET_RADIOBOX_SELECTION, {box, selection} )
end procedure



--wxRadioButton
------------------------------
--/topic wxRadioButton
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence text_label [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= 0]
--</ul>



--wxRegEx
-------------------------
--/topic wxRegEx
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li sequence regex [= ""]
--/li integer flags
--<ul><table>
--<tr><td> /li wxRE_EXTENDED    <td>use extended RE syntax
--<tr><td> /li wxRE_ADVANCED    <td>use advanced RE syntax
--<tr><td> /li wxRE_BASIC       <td>use basic RE syntax
--<tr><td> /li wxRE_ICASE       <td>ignore case in match
--<tr><td> /li wxRE_NOSUB       <td>only check match, don't set back references
--<tr><td valign="top"> /li wxRE_NEWLINE <td>if not set, treat '\n' as an
-- ordinary character, otherwise it is special: it is not matched by '.' and
-- '^' and '$' always match after/before it regardless of the setting of wxRE_NOT[BE]OL
--<tr><td> /li wxRE_DEFAULT     <td>wxRE_EXTENDED
--<tr><td> /li wxRE_NOTBOL      <td>'^' doesn't match at the start of line
--<tr><td> /li wxRE_NOTEOL      <td>'$' doesn't match at the end of line
--</table></ul>
--</ul>
--
--For more information, see the on-line wxWidgets documentation:
--<ul>
--/li <a href="http://www.wxwidgets.org/manuals/stable/wx_wxregex.html#wxregex">wxWidgets class documentation for wxRegEx</a>
--/li <a href="http://www.wxwidgets.org/manuals/stable/wx_wxresyn.html#wxresyn">Syntax of the wxWidgets regular expression library</a>
--</ul>

--/topic wxRegEx
--/func compile_regex(atom regex, sequence pattern, integer flags)
--
--Create and compile the regular expression, use /is_valid_regex to test for compilation errors.
public function compile_regex(atom regex, sequence pattern, integer flags)
    return c_func( WX_COMPILE_REGEX, {regex, pattern, flags})
end function

--/topic wxRegEx
--/func regex_matches( atom regex, sequence text, integer flags )
--
--Matches the precompiled regular expression against the string text, returns true if matches and false otherwise.
--
--Flags may be combination of wxRE_NOTBOL and wxRE_NOTEOL.
--
--May only be called after successful call to /compile_regex.
public function regex_matches( atom regex, sequence text, integer flags )
    return c_func( WX_REGEX_MATCHES, {regex, text, flags})
end function

--/topic wxRegEx
--/func get_match_count( atom regex )
--
--Returns the size of the array of matches, i.e. the number of bracketed subexpressions plus one for the expression itself, or 0 on error.
--
--May only be called after successful call to Compile(). and only if wxRE_NOSUB was not used.
public function get_match_count( atom regex )
    return c_func( WX_GET_MATCH_COUNT, {regex})
end function

--/topic wxRegEx
--/func get_match( atom regex, sequence text, integer index )
--
--Get the start index and the length of the match of the expression (if index is 0) or a bracketed subexpression (index different from 0).
--
--May only be called after successful call to Matches() and only if wxRE_NOSUB was not used in /compile_regex.
--
--Returns false if no match or if an error occurred.
public function get_match( atom regex, sequence text, integer index )
    return c_func( WX_GET_MATCH, {regex, text, index})
end function

--/topic wxRegEx
--/func is_valid_regex( atom regex )
--
--Return 1 if this is a valid compiled regular expression, 0 otherwise.
public function is_valid_regex( atom regex )
    return c_func( WX_IS_VALID_REGEX, {regex})
end function

--/topic wxRegEx
--/func replace_regex( atom regex, sequence text, sequence replacement, integer max_matches, integer in_place )
--
--Replaces the current regular expression in the string pointed to by text, with the text in replacement.  If
--in_place is 1, then only the modified text is returned.  If in_place is 1, then the return value is a
--sequence, with the first element the number of replaces made, and the second element the modified text.
--
--The replacement text may contain back references \number which will be replaced with the value of the
--corresponding subexpression in the pattern match. \0 corresponds to the entire match and & is a synonym
--for it. Backslash may be used to quote itself or & character.
--
--max_matches may be used to limit the number of replacements made, setting it to 1, for example, will only
--replace first occurrence (if any) of the pattern in the text while default value of 0 means replace all.
public function replace_regex( atom regex, sequence text, sequence replacement, integer max_matches, integer in_place )
    return c_func( WX_REPLACE_REGEX, {regex, text, replacement, max_matches, in_place})
end function



--wxRichTextAttr
-----------------------------------------
--/topic wxRichTextAttr
--/parent wxRichTextCtrl
--/parent Classes
--/info
--

--/topic wxRichTextAttr
--/func rich_to_textex_attr( atom attr )
--
--Makes a wxTextAttrEx object from /attr.
public function rich_to_textex_attr( atom attr )
    return c_func( WX_RICH_TO_TEXTEX_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func create_font_from_rich_attr( atom attr )
--
--Creates a font from the font attributes.
public function create_font_from_rich_attr( atom attr )
    return c_func( WX_CREATE_FONT_FROM_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_font_attributes( atom attr, atom font )
--
--Sets the font attributes from the given font.
public function get_rich_attr_font_attributes( atom attr, atom font )
    return c_func( WX_GET_RICH_ATTR_FONT_ATTRIBUTES, {attr,font})
end function

--/topic wxRichTextAttr
--/proc set_rich_attr_text_color( atom attr, atom color )
--
--Sets the text foreground colout.
public procedure set_rich_attr_text_color( atom attr, atom color )
    c_proc( WX_SET_RICH_ATTR_TEXT_COLOR, {attr,color})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_back_color( atom attr, atom color )
--
--Sets the background colour.
public procedure set_rich_attr_back_color( atom attr, atom color )
    c_proc( WX_SET_RICH_ATTR_BACK_COLOR, {attr,color})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_alignment( atom attr, atom align )
--
--Sets the paragraph alignment.
--
--These are the possible values for alignment:
--<ul>
--/li wxTEXT_ALIGNMENT_DEFAULT
--/li wxTEXT_ALIGNMENT_LEFT
--/li wxTEXT_ALIGNMENT_CENTRE
--/li wxTEXT_ALIGNMENT_CENTER
--/li wxTEXT_ALIGNMENT_RIGHT
--/li wxTEXT_ALIGNMENT_JUSTIFIED
--</ul>
--
--Of these, wxTEXT_ALIGNMENT_JUSTIFIED is unimplemented. In future justification may be supported when printing or previewing, only.
public procedure set_rich_attr_alignment( atom attr, atom align )
    c_proc( WX_SET_RICH_ATTR_ALIGNMENT, {attr,align})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_tabs( atom attr, object tabs )
--
--Sets the tab stops, expressed in tenths of a millimetre. Each stop is measured from the left margin and therefore each value must be
-- larger than the last.
public procedure set_rich_attr_tabs( atom attr, object tabs )
    c_proc( WX_SET_RICH_ATTR_TABS, {attr,tabs})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_left_indent( atom attr, atom indent, atom sub_indent )
--
--Sets the left indent and left subindent in tenths of a millimetre.
--
--The sub-indent is an offset from the left of the paragraph, and is used for all but the first line in a paragraph. A positive value will
-- cause the first line to appear to the left of the subsequent lines, and a negative value will cause the first line to be indented relative
-- to the subsequent lines.
--
--wxRichTextBuffer uses indentation to render a bulleted item. The left indent is the distance between the margin and the bullet. The content
-- of the paragraph, including the first line, starts at leftMargin + leftSubIndent. So the distance between the left edge of the bullet and
-- the left of the actual paragraph is leftSubIndent.
public procedure set_rich_attr_left_indent( atom attr, atom indent, atom sub_indent )
    c_proc( WX_SET_RICH_ATTR_LEFT_INDENT, {attr,indent,sub_indent})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_right_indent( atom attr, atom indent )
--
--Sets the right indent in tenths of a millimetre.
public procedure set_rich_attr_right_indent( atom attr, atom indent )
    c_proc( WX_SET_RICH_ATTR_RIGHT_INDENT, {attr,indent})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_font_size( atom attr, atom points )
--
--Sets the font size in points.
public procedure set_rich_attr_font_size( atom attr, atom points )
    c_proc( WX_SET_RICH_ATTR_FONT_SIZE, {attr,points})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_font_style( atom attr, atom style )
--
--Sets the font style (normal, italic or slanted).
public procedure set_rich_attr_font_style( atom attr, atom style )
    c_proc( WX_SET_RICH_ATTR_FONT_STYLE, {attr,style})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_font_name( atom attr, object name )
--
--Sets the font face name.
public procedure set_rich_attr_font_name( atom attr, object name )
    c_proc( WX_SET_RICH_ATTR_FONT_NAME, {attr,name})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_underlined( atom attr, atom underline )
--
public procedure set_rich_attr_underlined( atom attr, atom underline )
    c_proc( WX_SET_RICH_ATTR_UNDERLINED, {attr,underline})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_flags( atom attr, atom flags )
--
--Sets the flags determining which styles are being specified. The following flags can be passed in a bitlist:
--
--Standard wxTextAttr constants:
--<ul>
--/li wxTEXT_ATTR_TEXT_COLOUR
--/li wxTEXT_ATTR_BACKGROUND_COLOUR
--/li wxTEXT_ATTR_FONT_FACE
--/li wxTEXT_ATTR_FONT_SIZE
--/li wxTEXT_ATTR_FONT_WEIGHT
--/li wxTEXT_ATTR_FONT_ITALIC
--/li wxTEXT_ATTR_FONT_UNDERLINE
--/li wxTEXT_ATTR_FONT
--/li wxTEXT_ATTR_ALIGNMENT
--/li wxTEXT_ATTR_LEFT_INDENT
--/li wxTEXT_ATTR_RIGHT_INDENT
--/li wxTEXT_ATTR_TABS
--</ul>
--
--Extra formatting flags not in wxTextAttr
--<ul>
--/li wxTEXT_ATTR_PARA_SPACING_AFTER
--/li wxTEXT_ATTR_PARA_SPACING_BEFORE
--/li wxTEXT_ATTR_LINE_SPACING
--/li wxTEXT_ATTR_CHARACTER_STYLE_NAME
--/li wxTEXT_ATTR_PARAGRAPH_STYLE_NAME
--/li wxTEXT_ATTR_LIST_STYLE_NAME
--/li wxTEXT_ATTR_BULLET_STYLE
--/li wxTEXT_ATTR_BULLET_NUMBER
--/li wxTEXT_ATTR_BULLET_TEXT
--/li wxTEXT_ATTR_BULLET_NAME
--/li wxTEXT_ATTR_URL
--/li wxTEXT_ATTR_PAGE_BREAK
--/li wxTEXT_ATTR_EFFECTS
--/li wxTEXT_ATTR_OUTLINE_LEVEL
--</ul>
public procedure set_rich_attr_flags( atom attr, atom flags )
    c_proc( WX_SET_RICH_ATTR_FLAGS, {attr,flags})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_char_style( atom attr, object style )
--
--Sets the character style name.
public procedure set_rich_attr_char_style( atom attr, object style )
    c_proc( WX_SET_RICH_ATTR_CHAR_STYLE, {attr,style})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_para_style( atom attr, object style )
--
--Sets the name of the paragraph style.
public procedure set_rich_attr_para_style( atom attr, object style )
    c_proc( WX_SET_RICH_ATTR_PARA_STYLE, {attr,style})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_list_style( atom attr, object style )
--
--Sets the list style name.
public procedure set_rich_attr_list_style( atom attr, object style )
    c_proc( WX_SET_RICH_ATTR_LIST_STYLE, {attr,style})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_para_spacing_after( atom attr, atom space )
--
--Sets the spacing after a paragraph, in tenths of a millimetre.
public procedure set_rich_attr_para_spacing_after( atom attr, atom space )
    c_proc( WX_SET_RICH_ATTR_PARA_SPACING_AFTER, {attr,space})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_para_spacing_before( atom attr, atom space )
--
--Sets the spacing before a paragraph, in tenths of a millimetre.
public procedure set_rich_attr_para_spacing_before( atom attr, atom space )
    c_proc( WX_SET_RICH_ATTR_PARA_SPACING_BEFORE, {attr,space})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_line_spacing( atom attr, atom space )
--
--Sets the line spacing. spacing is a multiple, where 10 means single-spacing, 15 means 1.5 spacing, and 20 means double spacing.
-- The following constants are defined for convenience:
--
--<ul>
--/li wxTEXT_ATTR_LINE_SPACING_NORMAL (= 10)
--/li wxTEXT_ATTR_LINE_SPACING_HALF   (= 15)
--/li wxTEXT_ATTR_LINE_SPACING_TWICE  (= 20)
--</ul>
public procedure set_rich_attr_line_spacing( atom attr, atom space )
    c_proc( WX_SET_RICH_ATTR_LINE_SPACING, {attr,space})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_bullet_style( atom attr, atom style )
--
--Sets the bullet style. The following styles can be passed:
--
--<ul>
--/li wxTEXT_ATTR_BULLET_STYLE_NONE
--/li wxTEXT_ATTR_BULLET_STYLE_ARABIC
--/li wxTEXT_ATTR_BULLET_STYLE_LETTERS_UPPER
--/li wxTEXT_ATTR_BULLET_STYLE_LETTERS_LOWER
--/li wxTEXT_ATTR_BULLET_STYLE_ROMAN_UPPER
--/li wxTEXT_ATTR_BULLET_STYLE_ROMAN_LOWER
--/li wxTEXT_ATTR_BULLET_STYLE_SYMBOL
--/li wxTEXT_ATTR_BULLET_STYLE_BITMAP
--/li wxTEXT_ATTR_BULLET_STYLE_PARENTHESES
--/li wxTEXT_ATTR_BULLET_STYLE_PERIOD
--/li wxTEXT_ATTR_BULLET_STYLE_STANDARD
--/li wxTEXT_ATTR_BULLET_STYLE_RIGHT_PARENTHESIS
--/li wxTEXT_ATTR_BULLET_STYLE_OUTLINE
--/li wxTEXT_ATTR_BULLET_STYLE_ALIGN_LEFT
--/li wxTEXT_ATTR_BULLET_STYLE_ALIGN_RIGHT
--/li wxTEXT_ATTR_BULLET_STYLE_ALIGN_CENTRE
--</ul>
--
--Currently wxTEXT_ATTR_BULLET_STYLE_BITMAP is not supported.
public procedure set_rich_attr_bullet_style( atom attr, atom style )
    c_proc( WX_SET_RICH_ATTR_BULLET_STYLE, {attr,style})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_bullet_number( atom attr, atom number )
--
--Sets the bullet number.
public procedure set_rich_attr_bullet_number( atom attr, atom number )
    c_proc( WX_SET_RICH_ATTR_BULLET_NUMBER, {attr,number})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_bullet_font( atom attr, object font_name )
--
--Sets the name of the font associated with the bullet symbol. Only valid for attributes with wxTEXT_ATTR_BULLET_SYMBOL.
public procedure set_rich_attr_bullet_font( atom attr, object font_name )
    c_proc( WX_SET_RICH_ATTR_BULLET_FONT, {attr,font_name})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_bullet_name( atom attr, object name )
--
--Sets the standard bullet name, applicable if the bullet style is wxTEXT_ATTR_BULLET_STYLE_STANDARD. See /get_rich_attr_bullet_name for a
-- list of supported names, and how to expand the range of supported types.
public procedure set_rich_attr_bullet_name( atom attr, object name )
    c_proc( WX_SET_RICH_ATTR_BULLET_NAME, {attr,name})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_url( atom attr, object url )
--
--Sets the URL for the content. Sets the wxTEXT_ATTR_URL style; content with this style causes wxRichTextCtrl to show a hand cursor over it,
-- and wxRichTextCtrl generates a wxTextUrlEvent when the content is clicked.
public procedure set_rich_attr_url( atom attr, object url )
    c_proc( WX_SET_RICH_ATTR_URL, {attr,url})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_page_break( atom attr, atom page_break )
--
--Specifies a page break before this paragraph.
public procedure set_rich_attr_page_break( atom attr, atom page_break )
    c_proc( WX_SET_RICH_ATTR_PAGE_BREAK, {attr,page_break})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_effects( atom attr, atom effects )
--
--Sets the text effects, a bit list of styles. The following styles can be passed:
--
--<ul>
--/li wxTEXT_ATTR_EFFECT_NONE
--/li wxTEXT_ATTR_EFFECT_CAPITALS
--/li wxTEXT_ATTR_EFFECT_SMALL_CAPITALS
--/li wxTEXT_ATTR_EFFECT_STRIKETHROUGH
--/li wxTEXT_ATTR_EFFECT_DOUBLE_STRIKETHROUGH
--/li wxTEXT_ATTR_EFFECT_SHADOW
--/li wxTEXT_ATTR_EFFECT_EMBOSS
--/li wxTEXT_ATTR_EFFECT_OUTLINE
--/li wxTEXT_ATTR_EFFECT_ENGRAVE
--/li wxTEXT_ATTR_EFFECT_SUPERSCRIPT
--/li wxTEXT_ATTR_EFFECT_SUBSCRIPT
--</ul>
--
--Of these, only wxTEXT_ATTR_EFFECT_CAPITALS and wxTEXT_ATTR_EFFECT_STRIKETHROUGH are implemented. wxTEXT_ATTR_EFFECT_CAPITALS capitalises text
-- when displayed (leaving the case of the actual buffer text unchanged), and wxTEXT_ATTR_EFFECT_STRIKETHROUGH draws a line through text.
--
--To set effects, you should also pass wxTEXT_ATTR_EFFECTS to /set_rich_attr_flags, and call wxRichTextAttr::SetTextEffectFlags with the
-- styles (taken from the above set) that you are interested in setting.
public procedure set_rich_attr_effects( atom attr, atom effects )
    c_proc( WX_SET_RICH_ATTR_EFFECTS, {attr,effects})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_effect_flags( atom attr, atom flags )
--
--Sets the text effect bits of interest. You should also pass wxTEXT_ATTR_EFFECTS to /set_rich_attr_flags. See /set_rich_attr_flags
-- for further information.
public procedure set_rich_attr_effect_flags( atom attr, atom flags )
    c_proc( WX_SET_RICH_ATTR_EFFECT_FLAGS, {attr,flags})
end procedure

--/topic wxRichTextAttr
--/proc set_rich_attr_outline_level( atom attr, atom level )
--
--Specifies the outline level. Zero represents normal text. At present, the outline level is not used, but may be used in future for determining
-- list levels and for applications that need to store document structure information.
public procedure set_rich_attr_outline_level( atom attr, atom level )
    c_proc( WX_SET_RICH_ATTR_OUTLINE_LEVEL, {attr,level})
end procedure

--/topic wxRichTextAttr
--/func get_rich_attr_color( atom attr )
--
--Returns the text foreground colour.
public function get_rich_attr_color( atom attr )
    return c_func( WX_GET_RICH_ATTR_COLOR, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_back_color( atom attr )
--
--Returns the background colour.
public function get_rich_attr_back_color( atom attr )
    return c_func( WX_GET_RICH_ATTR_BACK_COLOR, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_alignment( atom attr )
--
--Returns the alignment flags. See /set_rich_attr_alignment for a list of available styles.
public function get_rich_attr_alignment( atom attr )
    return c_func( WX_GET_RICH_ATTR_ALIGNMENT, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_tabs( atom attr )
--
--Returns a sequence of tab stops, each expressed in tenths of a millimeter. Each stop is measured from the left margin and therefore each
-- value must be larger than the last.
public function get_rich_attr_tabs( atom attr )
    return c_func( WX_GET_RICH_ATTR_TABS, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_left_indent( atom attr )
--
--Returns the left indent in tenths of a millimetre.
public function get_rich_attr_left_indent( atom attr )
    return c_func( WX_GET_RICH_ATTR_LEFT_INDENT, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_left_sub_indent( atom attr )
--
--Returns the left sub-indent in tenths of a millimetre.
public function get_rich_attr_left_sub_indent( atom attr )
    return c_func( WX_GET_RICH_ATTR_LEFT_SUB_INDENT, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_right_indent( atom attr )
--
--Returns the right indent in tenths of a millimeter.
public function get_rich_attr_right_indent( atom attr )
    return c_func( WX_GET_RICH_ATTR_RIGHT_INDENT, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_flags( atom attr )
--
--Returns flags indicating which attributes are applicable. See /set_rich_attr_flags for a list of available flags.
public function get_rich_attr_flags( atom attr )
    return c_func( WX_GET_RICH_ATTR_FLAGS, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_font_size( atom attr )
--
--Returns the font size in points.
public function get_rich_attr_font_size( atom attr )
    return c_func( WX_GET_RICH_ATTR_FONT_SIZE, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_font_style( atom attr )
--
--Returns the font style.
public function get_rich_attr_font_style( atom attr )
    return c_func( WX_GET_RICH_ATTR_FONT_STYLE, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_font_weight( atom attr )
--
--Returns the font weight
public function get_rich_attr_font_weight( atom attr )
    return c_func( WX_GET_RICH_ATTR_FONT_WEIGHT, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_underlined( atom attr )
--
public function get_rich_attr_underlined( atom attr )
    return c_func( WX_GET_RICH_ATTR_UNDERLINED, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_font_name( atom attr )
--
--Returns the font face name.
public function get_rich_attr_font_name( atom attr )
    return c_func( WX_GET_RICH_ATTR_FONT_NAME, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_char_style( atom attr )
--
--Returns the name of the character style.
public function get_rich_attr_char_style( atom attr )
    return c_func( WX_GET_RICH_ATTR_CHAR_STYLE, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_para_style( atom attr )
--
--Returns the name of the paragraph style.
public function get_rich_attr_para_style( atom attr )
    return c_func( WX_GET_RICH_ATTR_PARA_STYLE, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_list_style( atom attr )
--
--Returns the name of the list style.
public function get_rich_attr_list_style( atom attr )
    return c_func( WX_GET_RICH_ATTR_LIST_STYLE, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_para_spacing_after( atom attr )
--
--Returns the space in tenths of a millimeter after the paragraph.
public function get_rich_attr_para_spacing_after( atom attr )
    return c_func( WX_GET_RICH_ATTR_PARA_SPACING_AFTER, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_para_spacing_before( atom attr )
--
--Returns the space in tenths of a millimeter before the paragraph.
public function get_rich_attr_para_spacing_before( atom attr )
    return c_func( WX_GET_RICH_ATTR_PARA_SPACING_BEFORE, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_line_spacing( atom attr )
--
--Returns the line spacing value, one of wxTEXT_ATTR_LINE_SPACING_NORMAL, wxTEXT_ATTR_LINE_SPACING_HALF, and wxTEXT_ATTR_LINE_SPACING_TWICE.
public function get_rich_attr_line_spacing( atom attr )
    return c_func( WX_GET_RICH_ATTR_LINE_SPACING, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_bullet_style( atom attr )
--
--Returns the bullet style. See /set_rich_attr_bullet_style for a list of available styles.
public function get_rich_attr_bullet_style( atom attr )
    return c_func( WX_GET_RICH_ATTR_BULLET_STYLE, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_bullet_number( atom attr )
--
--Returns the bullet number.
public function get_rich_attr_bullet_number( atom attr )
    return c_func( WX_GET_RICH_ATTR_BULLET_NUMBER, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_bullet_text( atom attr )
--
--Returns the bullet text, which could be a symbol, or (for example) cached outline text.
public function get_rich_attr_bullet_text( atom attr )
    return c_func( WX_GET_RICH_ATTR_BULLET_TEXT, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_bullet_font( atom attr )
--
--Returns a string containing the name of the font associated with the bullet symbol. Only valid for attributes with wxTEXT_ATTR_BULLET_SYMBOL.
public function get_rich_attr_bullet_font( atom attr )
    return c_func( WX_GET_RICH_ATTR_BULLET_FONT, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_bullet_name( atom attr )
--
--Returns the standard bullet name, applicable if the bullet style is wxTEXT_ATTR_BULLET_STYLE_STANDARD.
-- Currently the following standard bullet names are supported:
--<ul>
--/li standard/circle
--/li standard/square
--/li standard/diamond
--/li standard/triangle
--</ul>
--If you wish your application to support further bullet graphics, you can derive a class from wxRichTextRenderer or
-- wxRichTextStdRenderer, override DrawStandardBullet and EnumerateStandardBulletNames, and set an instance of the class
-- using /set_rich_text_renderer.
public function get_rich_attr_bullet_name( atom attr )
    return c_func( WX_GET_RICH_ATTR_BULLET_NAME, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_url( atom attr )
--
--Returns the URL for the content. Content with wxTEXT_ATTR_URL style causes wxRichTextCtrl to show a hand cursor over it, and wxRichTextCtrl
-- generates a wxTextUrlEvent when the content is clicked.
public function get_rich_attr_url( atom attr )
    return c_func( WX_GET_RICH_ATTR_URL, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_effects( atom attr )
--
--Returns the text effects, a bit list of styles. See /set_rich_attr_effects for details.
public function get_rich_attr_effects( atom attr )
    return c_func( WX_GET_RICH_ATTR_EFFECTS, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_effect_flags( atom attr )
--
--Returns the text effect bits of interest. See /set_rich_attr_flags for further information.
public function get_rich_attr_effect_flags( atom attr )
    return c_func( WX_GET_RICH_ATTR_EFFECT_FLAGS, {attr})
end function

--/topic wxRichTextAttr
--/func get_rich_attr_outline_level( atom attr )
--
--Returns the outline level.
public function get_rich_attr_outline_level( atom attr )
    return c_func( WX_GET_RICH_ATTR_OUTLINE_LEVEL, {attr})
end function

--/topic wxRichTextAttr
--/func has_text_color_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a text foreground colour.
public function has_text_color_rich_attr( atom attr )
    return c_func( WX_HAS_TEXT_COLOR_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_back_color_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a background colour.
public function has_back_color_rich_attr( atom attr )
    return c_func( WX_HAS_BACK_COLOR_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_alignment_rich_attr( atom attr )
--
--Returns true if the attribute object specifies alignment.
public function has_alignment_rich_attr( atom attr )
    return c_func( WX_HAS_ALIGNMENT_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_tabs_rich_attr( atom attr )
--
--Returns true if the attribute object specifies tab stops.
public function has_tabs_rich_attr( atom attr )
    return c_func( WX_HAS_TABS_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_left_indent_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a left indent.
public function has_left_indent_rich_attr( atom attr )
    return c_func( WX_HAS_LEFT_INDENT_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_right_indent_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a right indent.
public function has_right_indent_rich_attr( atom attr )
    return c_func( WX_HAS_RIGHT_INDENT_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_font_weight_rich_attr( atom attr )
--
--Returns true if the attribute object specifies font weight (bold, light or normal).
public function has_font_weight_rich_attr( atom attr )
    return c_func( WX_HAS_FONT_WEIGHT_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_font_size_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a font point size.
public function has_font_size_rich_attr( atom attr )
    return c_func( WX_HAS_FONT_SIZE_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_italic_rich_attr( atom attr )
--
--Returns true if the attribute object specifies italic style.
public function has_italic_rich_attr( atom attr )
    return c_func( WX_HAS_ITALIC_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_underlined_rich_attr( atom attr )
--
--Returns true if the attribute object specifies either underlining or no underlining.
public function has_underlined_rich_attr( atom attr )
    return c_func( WX_HAS_UNDERLINED_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_font_name_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a font face name.
public function has_font_name_rich_attr( atom attr )
    return c_func( WX_HAS_FONT_NAME_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_font_rich_attr( atom attr )
--
--Returns true if the attribute object specifies any font attributes.
public function has_font_rich_attr( atom attr )
    return c_func( WX_HAS_FONT_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_para_spacing_after_rich_attr( atom attr )
--
--Returns true if the attribute object specifies spacing after a paragraph.
public function has_para_spacing_after_rich_attr( atom attr )
    return c_func( WX_HAS_PARA_SPACING_AFTER_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_para_spacing_before_rich_attr( atom attr )
--
--Returns true if the attribute object specifies spacing before a paragraph.
public function has_para_spacing_before_rich_attr( atom attr )
    return c_func( WX_HAS_PARA_SPACING_BEFORE_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_line_spacing_rich_attr( atom attr )
--
--Returns true if the attribute object specifies line spacing.
public function has_line_spacing_rich_attr( atom attr )
    return c_func( WX_HAS_LINE_SPACING_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_char_style_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a character style name.
public function has_char_style_rich_attr( atom attr )
    return c_func( WX_HAS_CHAR_STYLE_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_para_style_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a paragraph style name.
public function has_para_style_rich_attr( atom attr )
    return c_func( WX_HAS_PARA_STYLE_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_list_style_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a list style name.
public function has_list_style_rich_attr( atom attr )
    return c_func( WX_HAS_LIST_STYLE_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_bullet_style_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a bullet style.
public function has_bullet_style_rich_attr( atom attr )
    return c_func( WX_HAS_BULLET_STYLE_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_bullet_number_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a bullet number.
public function has_bullet_number_rich_attr( atom attr )
    return c_func( WX_HAS_BULLET_NUMBER_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_bullet_text_rich_attr( atom attr )
--
--Returns true if the attribute object specifies bullet text (usually specifying a symbol).
public function has_bullet_text_rich_attr( atom attr )
    return c_func( WX_HAS_BULLET_TEXT_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_bullet_name_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a standard bullet name.
public function has_bullet_name_rich_attr( atom attr )
    return c_func( WX_HAS_BULLET_NAME_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_url_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a URL.
public function has_url_rich_attr( atom attr )
    return c_func( WX_HAS_URL_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_page_break_rich_attr( atom attr )
--
--Returns true if the attribute object specifies a page break before this paragraph.
public function has_page_break_rich_attr( atom attr )
    return c_func( WX_HAS_PAGE_BREAK_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_effects_rich_attr( atom attr )
--
--Returns true if the attribute object specifies text effects.
public function has_effects_rich_attr( atom attr )
    return c_func( WX_HAS_EFFECTS_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_effect_rich_attr( atom attr, atom effect )
--
--Returns true if the attribute object specifies a text effect.
public function has_effect_rich_attr( atom attr, atom effect )
    return c_func( WX_HAS_EFFECT_RICH_ATTR, {attr,effect})
end function

--/topic wxRichTextAttr
--/func has_outline_level_rich_attr( atom attr )
--
--Returns true if the attribute object specifies an outline level.
public function has_outline_level_rich_attr( atom attr )
    return c_func( WX_HAS_OUTLINE_LEVEL_RICH_ATTR, {attr})
end function

--/topic wxRichTextAttr
--/func has_flag_rich_attr( atom attr, atom flag )
--
--Returns true if the flag is present in the attribute object's flag bitlist.
public function has_flag_rich_attr( atom attr, atom flag )
    return c_func( WX_HAS_FLAG_RICH_ATTR, {attr,flag})
end function

--/topic wxRichTextAttr
--/func is_rich_attr_char_style( atom attr )
--
--Returns true if the object represents a character style, that is, the flags specify a font or a text background or foreground colour.
public function is_rich_attr_char_style( atom attr )
    return c_func( WX_IS_RICH_ATTR_CHAR_STYLE, {attr})
end function

--/topic wxRichTextAttr
--/func is_rich_attr_para_style( atom attr )
--
--Returns true if the object represents a paragraph style, that is, the flags specify alignment, indentation, tabs, paragraph spacing,
-- or bullet style.
public function is_rich_attr_para_style( atom attr )
    return c_func( WX_IS_RICH_ATTR_PARA_STYLE, {attr})
end function

--/topic wxRichTextAttr
--/func is_rich_attr_default( atom attr )
--
--Returns false if we have any attributes set, true otherwise.
public function is_rich_attr_default( atom attr )
    return c_func( WX_IS_RICH_ATTR_DEFAULT, {attr})
end function

--/topic wxRichTextAttr
--/proc apply_rich_attr( atom attr, atom compare_with )
--
--Applies the attributes in style to the original object, but not those attributes from style that are the same as those in compareWith (if passed).
--
--See also /combine_rich_attr for a function that does almost the same but returns a new object instead of modifying the original object.
public procedure apply_rich_attr( atom attr, atom compare_with )
    c_proc( WX_APPLY_RICH_ATTR, {attr,compare_with})
end procedure



--wxRichTextCtrl
----------------------------------
--/topic wxRichTextCtrl
--/parent wxTextCtrl
--/parent Controls
--/info
--
--wxRichTextCtrl provides a generic implementation of a rich text editor that can
--handle different character styles, paragraph formatting, and images. It's aimed at editing
--'natural' language text - if you need an editor that supports code editing, wxStyledTextCtrl
-- is a better choice.
--
--Despite its name, it cannot currently read or write RTF (rich text format) files. Instead,
--it uses its own XML format, and can also read and write plain text. In future we expect to
--provide RTF file capabilities. Custom file formats can be supported by creating additional
--file handlers and registering them with the control.
--
--wxRichTextCtrl is largely compatible with the /wxTextCtrl API, but extends it where necessary.
--The control can be used where the native rich text capabilities of /wxTextCtrl are not adequate
--(this is particularly true on Windows) and where more direct access to the content representation
--is required. It is difficult and inefficient to read the style information in a wxTextCtrl,
--whereas this information is readily available in wxRichTextCtrl. Since it's written in pure
--wxWidgets, any customizations you make to wxRichTextCtrl will be reflected on all platforms.
--
--wxRichTextCtrl supports basic printing via the easy-to-use wxRichTextPrinting class. Creating
--applications with simple word processing features is simplified with the inclusion of
--wxRichTextFormattingDialog, a tabbed dialog allowing interactive tailoring of paragraph and
--character styling. Also provided is the multi-purpose dialog /wxRichTextStyleOrganiserDialog
--that can be used for managing style definitions, browsing styles and applying them, or selecting
--list styles with a renumber option.
--
--There are a few disadvantages to using wxRichTextCtrl. It is not native, so does not behave
--exactly as a native /wxTextCtrl, although common editing conventions are followed. Users
--may miss the built-in spelling correction on Mac OS X, or any special character input that
--may be provided by the native control. It would also be a poor choice if intended users rely
--on screen readers that would be not work well with non-native text input implementation. You
--might mitigate this by providing the choice between /wxTextCtrl and wxRichTextCtrl, with fewer
--features in the former case.

--/topic wxRichTextCtrl
--/func write_image( atom rich, atom image, atom bitmap_type )
--
--Write a bitmap or image at the current insertion point. Supply an optional type (see /wxBitmap) to use for
--internal and file storage of the raw data.
public function write_image( atom rich, atom image, atom bitmap_type )
    return c_func( WX_WRITE_IMAGE, {rich,image,bitmap_type})
end function

--/topic wxRichTextCtrl
--/func write_image_from_file( atom rich, sequence filename, atom bitmap_type )
--
--Write a bitmap or image from a file at the current insertion point. Supply an optional type
--(see /wxBitmap) to use for internal and file storage of the raw data.
public function write_image_from_file( atom rich, sequence filename, atom bitmap_type )
    return c_func( WX_WRITE_IMAGE_FROM_FILE, {rich,filename,bitmap_type})
end function

--/topic wxRichTextCtrl
--/func write_image_block( atom rich, atom image_block )
--
--Writes an image block at the current insertion point.
public function write_image_block( atom rich, atom image_block )
    return c_func( WX_WRITE_IMAGE_BLOCK, {rich,image_block})
end function

--/topic wxRichTextCtrl
--/proc line_break( atom rich )
--
--Inserts a line break at the current insertion point. A line break forces wrapping within a
--paragraph, and can be introduced by using this function, or by typing Shift-Return.
public procedure line_break( atom rich )
    c_proc( WX_LINE_BREAK, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc set_basic_rich_style( atom rich, atom style )
--
--Sets the basic (overall) style. This is the style of the whole buffer before further styles are applied,
--unlike the default style, which only affects the style currently being applied (for example, setting the
--default style to bold will cause subsequently inserted text to be bold).
--
--/b style is a /wxTextAttrEx
public procedure set_basic_rich_style( atom rich, atom style )
    c_proc( WX_SET_BASIC_RICH_STYLE, {rich,style})
end procedure

--/topic wxRichTextCtrl
--/proc set_basic_rich_text_style( atom rich, atom style )
--
--Sets the basic (overall) style. This is the style of the whole buffer before further styles are applied,
--unlike the default style, which only affects the style currently being applied (for example, setting the
--default style to bold will cause subsequently inserted text to be bold).
--
--/b style is a /wxRichAttrEx
public procedure set_basic_rich_text_style( atom rich, atom style )
    c_proc( WX_SET_BASIC_RICH_TEXT_STYLE, {rich,style})
end procedure

--/topic wxRichTextCtrl
--/func get_basic_rich_style( atom rich )
--
--Gets the basic (overall) style. This is the style of the whole buffer before further styles are applied, unlike the default style, which only
-- affects the style currently being applied (for example, setting the default style to bold will cause subsequently inserted text to be bold).
public function get_basic_rich_style( atom rich )
    return c_func( WX_GET_BASIC_RICH_STYLE, {rich})
end function

--/topic wxRichTextCtrl
--/proc begin_rich_style( atom rich, atom style )
--
--Begins applying a style.
public procedure begin_rich_style( atom rich, atom style )
    c_proc( WX_BEGIN_RICH_STYLE, {rich,style})
end procedure

--/topic wxRichTextCtrl
--/proc end_rich_style( atom rich )
--
--Ends the current style.
public procedure end_rich_style( atom rich )
    c_proc( WX_END_RICH_STYLE, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc end_all_rich_styles( atom rich )
--
--Ends application of all styles in the current style stack.
public procedure end_all_rich_styles( atom rich )
    c_proc( WX_END_ALL_RICH_STYLES, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_bold( atom rich )
--
--Begins using bold.
public procedure begin_bold( atom rich )
    c_proc( WX_BEGIN_BOLD, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc end_bold( atom rich )
--
--Ends using bold.
public procedure end_bold( atom rich )
    c_proc( WX_END_BOLD, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_italic( atom rich )
--
--Begins using italic.
public procedure begin_italic( atom rich )
    c_proc( WX_BEGIN_ITALIC, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc end_italic( atom rich )
--
--Ends using italic.
public procedure end_italic( atom rich )
    c_proc( WX_END_ITALIC, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_underline( atom rich )
--
--Begins using underlining.
public procedure begin_underline( atom rich )
    c_proc( WX_BEGIN_UNDERLINE, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc end_underline( atom rich )
--
--End applying underlining.
public procedure end_underline( atom rich )
    c_proc( WX_END_UNDERLINE, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_font_size( atom rich, atom size )
--
--Begins using the given point size.
public procedure begin_font_size( atom rich, atom size )
    c_proc( WX_BEGIN_FONT_SIZE, {rich,size})
end procedure

--/topic wxRichTextCtrl
--/proc end_font_size( atom rich )
--
--Ends using a point size.
public procedure end_font_size( atom rich )
    c_proc( WX_END_FONT_SIZE, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_font( atom rich, atom font )
--
--Begins using this font.
public procedure begin_font( atom rich, atom font )
    c_proc( WX_BEGIN_FONT, {rich,font})
end procedure

--/topic wxRichTextCtrl
--/proc end_font( atom rich )
--
--Ends using a font.
public procedure end_font( atom rich )
    c_proc( WX_END_FONT, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_text_color( atom rich, atom color )
--
--Begins using this colour.
public procedure begin_text_color( atom rich, atom color )
    c_proc( WX_BEGIN_TEXT_COLOR, {rich,color})
end procedure

--/topic wxRichTextCtrl
--/proc end_text_color( atom rich )
--
--Ends applying a text colour.
public procedure end_text_color( atom rich )
    c_proc( WX_END_TEXT_COLOR, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_alignment( atom rich, atom align )
--
--Begins using alignment
--
--For alignment values, see /wxTextAttrEx.
public procedure begin_alignment( atom rich, atom align )
    c_proc( WX_BEGIN_ALIGNMENT, {rich,align})
end procedure

--/topic wxRichTextCtrl
--/proc end_alignment( atom rich )
--
--Ends alignment.
public procedure end_alignment( atom rich )
    c_proc( WX_END_ALIGNMENT, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_left_indent( atom rich, atom indent, atom sub_indent )
--
--Begins applying a left indent and subindent in tenths of a millimetre.
--
--The sub-indent is an offset from the left of the paragraph, and is used for all but the first line in a paragraph. A positive value will cause
-- the first line to appear to the left of the subsequent lines, and a negative value will cause the first line to be indented relative to the
-- subsequent lines.
--
--wxRichTextBuffer uses indentation to render a bulleted item. The left indent is the distance between the margin and the bullet. The content of
-- the paragraph, including the first line, starts at leftMargin + leftSubIndent. So the distance between the left edge of the bullet and the
-- left of the actual paragraph is leftSubIndent.
public procedure begin_left_indent( atom rich, atom indent, atom sub_indent )
    c_proc( WX_BEGIN_LEFT_INDENT, {rich,indent,sub_indent})
end procedure

--/topic wxRichTextCtrl
--/proc end_left_indent( atom rich )
--
--Ends left indent.
public procedure end_left_indent( atom rich )
    c_proc( WX_END_LEFT_INDENT, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_right_indent( atom rich, atom indent )
--
--Begins a right indent, specified in tenths of a millimetre.
public procedure begin_right_indent( atom rich, atom indent )
    c_proc( WX_BEGIN_RIGHT_INDENT, {rich,indent})
end procedure

--/topic wxRichTextCtrl
--/proc end_right_indent( atom rich )
--
--Ends right indent.
public procedure end_right_indent( atom rich )
    c_proc( WX_END_RIGHT_INDENT, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_paragraph_spacing( atom rich, atom before, atom after )
--
--Begins paragraph spacing; pass the before-paragraph and after-paragraph spacing in tenths of a millimetre.
public procedure begin_paragraph_spacing( atom rich, atom before, atom after )
    c_proc( WX_BEGIN_PARAGRAPH_SPACING, {rich,before,after})
end procedure

--/topic wxRichTextCtrl
--/proc end_paragraph_spacing( atom rich )
--
--Ends paragraph spacing.
public procedure end_paragraph_spacing( atom rich )
    c_proc( WX_END_PARAGRAPH_SPACING, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_line_spacing( atom rich, atom spacing )
--
--Begins appling line spacing. spacing is a multiple, where 10 means single-spacing, 15 means 1.5 spacing, and 20 means double spacing.
-- The following constants are defined for convenience:
--
--<ul>
--/li wxTEXT_ATTR_LINE_SPACING_NORMAL (= 10)
--/li wxTEXT_ATTR_LINE_SPACING_HALF   (= 15)
--/li wxTEXT_ATTR_LINE_SPACING_TWICE  (= 20)
--</ul>
public procedure begin_line_spacing( atom rich, atom spacing )
    c_proc( WX_BEGIN_LINE_SPACING, {rich,spacing})
end procedure

--/topic wxRichTextCtrl
--/proc end_line_spacing( atom rich )
--
--Ends line spacing.
public procedure end_line_spacing( atom rich )
    c_proc( WX_END_LINE_SPACING, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_numbered_bullet( atom rich, atom number, atom indent, atom sub_indent, atom style )
--
--Begins a numbered bullet. This call will be needed for each item in the list, and the application should take care of incrementing the numbering.
--
--<ul>
--/li /i number is a number, usually starting with 1.
--/li /i indent and /i sub_indent are values in tenths of a millimetre.
--/li /i style is a bitlist of the following values:
--<ul>
--/li wxTEXT_ATTR_BULLET_STYLE_NONE
--/li wxTEXT_ATTR_BULLET_STYLE_ARABIC
--/li wxTEXT_ATTR_BULLET_STYLE_LETTERS_UPPER
--/li wxTEXT_ATTR_BULLET_STYLE_LETTERS_LOWER
--/li wxTEXT_ATTR_BULLET_STYLE_ROMAN_UPPER
--/li wxTEXT_ATTR_BULLET_STYLE_ROMAN_LOWER
--/li wxTEXT_ATTR_BULLET_STYLE_SYMBOL
--/li wxTEXT_ATTR_BULLET_STYLE_BITMAP
--/li wxTEXT_ATTR_BULLET_STYLE_PARENTHESES
--/li wxTEXT_ATTR_BULLET_STYLE_PERIOD
--/li wxTEXT_ATTR_BULLET_STYLE_STANDARD
--/li wxTEXT_ATTR_BULLET_STYLE_RIGHT_PARENTHESIS
--/li wxTEXT_ATTR_BULLET_STYLE_OUTLINE
--/li wxTEXT_ATTR_BULLET_STYLE_ALIGN_LEFT
--/li wxTEXT_ATTR_BULLET_STYLE_ALIGN_RIGHT
--/li wxTEXT_ATTR_BULLET_STYLE_ALIGN_CENTRE
--</ul>
--</ul>
--
--wxRichTextBuffer uses indentation to render a bulleted item. The left indent is the distance between the margin and the bullet. The content of
-- the paragraph, including the first line, starts at leftMargin + /i sub_indent. So the distance between the left edge of the bullet and the left
-- of the actual paragraph is /i sub_indent.
public procedure begin_numbered_bullet( atom rich, atom number, atom indent, atom sub_indent, atom style )
    c_proc( WX_BEGIN_NUMBERED_BULLET, {rich,number,indent,sub_indent,style})
end procedure

--/topic wxRichTextCtrl
--/proc end_numbered_bullet( atom rich )
--
--Ends application of a numbered bullet.
public procedure end_numbered_bullet( atom rich )
    c_proc( WX_END_NUMBERED_BULLET, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_symbol_bullet( atom rich, sequence symbol, atom indent, atom sub_indent, atom style )
--
--Begins applying a symbol bullet, using a character from the current font. See /begin_numbered_bullet for an explanation of how indentation is used
-- to render the bulleted paragraph.
public procedure begin_symbol_bullet( atom rich, sequence symbol, atom indent, atom sub_indent, atom style )
    c_proc( WX_BEGIN_SYMBOL_BULLET, {rich,symbol,indent,sub_indent,style})
end procedure

--/topic wxRichTextCtrl
--/proc end_symbol_bullet( atom rich )
--
--Ends applying a symbol bullet.
public procedure end_symbol_bullet( atom rich )
    c_proc( WX_END_SYMBOL_BULLET, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_standard_bullet( atom rich, sequence bullet_name, atom indent, atom sub_indent, atom style )
--
--Begins applying a standard bullet. See /begin_numbered_bullet for an explanation of how indentation is used to render the bulleted paragraph.
public procedure begin_standard_bullet( atom rich, sequence bullet_name, atom indent, atom sub_indent, atom style )
    c_proc( WX_BEGIN_STANDARD_BULLET, {rich,bullet_name,indent,sub_indent,style})
end procedure

--/topic wxRichTextCtrl
--/proc end_standard_bullet( atom rich )
--
--Ends applying a standard bullet.
public procedure end_standard_bullet( atom rich )
    c_proc( WX_END_STANDARD_BULLET, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_char_style( atom rich, sequence named_style )
--
--Begins using the named character style.
public procedure begin_char_style( atom rich, sequence named_style )
    c_proc( WX_BEGIN_CHAR_STYLE, {rich,named_style})
end procedure

--/topic wxRichTextCtrl
--/proc end_char_style( atom rich )
--
--Ends application of a named character style.
public procedure end_char_style( atom rich )
    c_proc( WX_END_CHAR_STYLE, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_paragraph_style( atom rich, sequence named_style )
--
--Begins applying the named paragraph style.
public procedure begin_paragraph_style( atom rich, sequence named_style )
    c_proc( WX_BEGIN_PARAGRAPH_STYLE, {rich,named_style})
end procedure

--/topic wxRichTextCtrl
--/proc end_paragraph_style( atom rich )
--
--Ends application of a named paragraph style.
public procedure end_paragraph_style( atom rich )
    c_proc( WX_END_PARAGRAPH_STYLE, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_list_style( atom rich, sequence named_style, atom level, atom number )
--
--Begins using a specified list style. Optionally, you can also pass a level and a number.
public procedure begin_list_style( atom rich, sequence named_style, atom level, atom number )
    c_proc( WX_BEGIN_LIST_STYLE, {rich,named_style,level,number})
end procedure

--/topic wxRichTextCtrl
--/proc end_list_style( atom rich )
--
--Ends using a specified list style.
public procedure end_list_style( atom rich )
    c_proc( WX_END_LIST_STYLE, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc begin_url( atom rich, sequence url, sequence style )
--
--Begins applying wxTEXT_ATTR_URL to the content. Pass a URL and optionally, a character style to apply, since it is common to mark a URL with
-- a familiar style such as blue text with underlining.
public procedure begin_url( atom rich, sequence url, sequence style )
    c_proc( WX_BEGIN_URL, {rich,url,style})
end procedure

--/topic wxRichTextCtrl
--/proc end_url( atom rich )
--
--Ends applying a URL.
public procedure end_url( atom rich )
    c_proc( WX_END_URL, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc set_default_style_to_cursor_style( atom rich )
--
--Sets the default style to the style under the cursor.
public procedure set_default_style_to_cursor_style( atom rich )
    c_proc( WX_SET_DEFAULT_STYLE_TO_CURSOR_STYLE, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc select_none( atom rich )
--
--Cancels any selection.
public procedure select_none( atom rich )
    c_proc( WX_SELECT_NONE, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc select_word( atom rich, atom pos )
--
--Selects the word at /i pos.
public procedure select_word( atom rich, atom pos )
    c_proc( WX_SELECT_WORD, {rich,pos})
end procedure

--/topic wxRichTextCtrl
--/func get_rich_selection_range( atom rich )
--
--Returns the selection range in character positions. {-1, -1} means no selection.
public function get_rich_selection_range( atom rich )
    return c_func( WX_GET_RICH_SELECTION_RANGE, {rich})
end function

--/topic wxRichTextCtrl
--/proc set_rich_selection_range( atom rich, atom start, atom finish )
--
--Sets the selection to the given range.
--
--The end point of range is specified as the last character position of the span of text, plus one. So, for example, to set the selection for
-- a character at position 5, use the range (5,6).
public procedure set_rich_selection_range( atom rich, atom start, atom finish )
    c_proc( WX_SET_RICH_SELECTION_RANGE, {rich,start,finish})
end procedure

--/topic wxRichTextCtrl
--/func get_rich_internal_selection_range( atom rich )
--
public function get_rich_internal_selection_range( atom rich )
    return c_func( WX_GET_RICH_INTERNAL_SELECTION_RANGE, {rich})
end function

--/topic wxRichTextCtrl
--/proc set_rich_internal_selection_range( atom rich, atom start, atom finish )
--
public procedure set_rich_internal_selection_range( atom rich, atom start, atom finish )
    c_proc( WX_SET_RICH_INTERNAL_SELECTION_RANGE, {rich,start,start})
end procedure

--/topic wxRichTextCtrl
--/func add_paragraph( atom rich, sequence text )
--
--Adds a new paragraph of text to the end of the buffer.
public function add_paragraph( atom rich, sequence text )
    return c_func( WX_ADD_PARAGRAPH, {rich,text})
end function

--/topic wxRichTextCtrl
--/func add_rich_image( atom rich, atom image )
--
--Adds an image to the control's buffer.
public function add_rich_image( atom rich, atom image )
    return c_func( WX_ADD_RICH_IMAGE, {rich,image})
end function

--/topic wxRichTextCtrl
--/proc layout_rich_content( atom rich, atom only_visible )
--
--Lays out the buffer, which must be done before certain operations, such as setting the caret position.
-- This function should not normally be required by the application.
public procedure layout_rich_content( atom rich, atom only_visible )
    c_proc( WX_LAYOUT_RICH_CONTENT, {rich,only_visible})
end procedure

--/topic wxRichTextCtrl
--/proc move_rich_caret( atom rich, atom pos, atom at_start )
--
--Move the caret to the given character position.
public procedure move_rich_caret( atom rich, atom pos, atom at_start )
    c_proc( WX_MOVE_RICH_CARET, {rich,pos,at_start})
end procedure

--/topic wxRichTextCtrl
--/proc move_right( atom rich, atom n_pos, atom flags )
--
--Moves right.
public procedure move_right( atom rich, atom n_pos, atom flags )
    c_proc( WX_MOVE_RIGHT, {rich,n_pos,flags})
end procedure

--/topic wxRichTextCtrl
--/proc move_left( atom rich, atom n_pos, atom flags )
--
--Moves left.
public procedure move_left( atom rich, atom n_pos, atom flags )
    c_proc( WX_MOVE_LEFT, {rich,n_pos,flags})
end procedure

--/topic wxRichTextCtrl
--/proc move_up( atom rich, atom n_pos, atom flags )
--
--Moves up.
public procedure move_up( atom rich, atom n_pos, atom flags )
    c_proc( WX_MOVE_UP, {rich,n_pos,flags})
end procedure

--/topic wxRichTextCtrl
--/proc move_down( atom rich, atom n_pos, atom flags )
--
--Moves down.
public procedure move_down( atom rich, atom n_pos, atom flags )
    c_proc( WX_MOVE_DOWN, {rich,n_pos,flags})
end procedure

--/topic wxRichTextCtrl
--/proc move_to_line_end( atom rich, atom flags )
--
--Moves to the end of the line.
public procedure move_to_line_end( atom rich, atom flags )
    c_proc( WX_MOVE_TO_LINE_END, {rich,flags})
end procedure

--/topic wxRichTextCtrl
--/proc move_to_line_start( atom rich, atom flags )
--
--Moves to the start of the line.
public procedure move_to_line_start( atom rich, atom flags )
    c_proc( WX_MOVE_TO_LINE_START, {rich,flags})
end procedure

--/topic wxRichTextCtrl
--/proc move_to_paragraph_end( atom rich, atom flags )
--
--Moves to the end of the paragraph.
public procedure move_to_paragraph_end( atom rich, atom flags )
    c_proc( WX_MOVE_TO_PARAGRAPH_END, {rich,flags})
end procedure

--/topic wxRichTextCtrl
--/proc move_to_paragraph_start( atom rich, atom flags )
--
--Moves to the start of the paragraph.
public procedure move_to_paragraph_start( atom rich, atom flags )
    c_proc( WX_MOVE_TO_PARAGRAPH_START, {rich,flags})
end procedure

--/topic wxRichTextCtrl
--/proc move_home( atom rich, atom flags )
--
--Moves to the start of the buffer.
public procedure move_home( atom rich, atom flags )
    c_proc( WX_MOVE_HOME, {rich,flags})
end procedure

--/topic wxRichTextCtrl
--/proc move_end( atom rich, atom flags )
--
--Moves to the end of the buffer.
public procedure move_end( atom rich, atom flags )
    c_proc( WX_MOVE_END, {rich,flags})
end procedure

--/topic wxRichTextCtrl
--/proc page_up( atom rich, atom pages, atom flags )
--
--Moves one or more pages up.
public procedure page_up( atom rich, atom pages, atom flags )
    c_proc( WX_PAGE_UP, {rich,pages,flags})
end procedure

--/topic wxRichTextCtrl
--/proc page_down( atom rich, atom pages, atom flags )
--
--Moves one or more pages down.
public procedure page_down( atom rich, atom pages, atom flags )
    c_proc( WX_PAGE_DOWN, {rich,pages,flags})
end procedure

--/topic wxRichTextCtrl
--/proc word_left( atom rich, atom words, atom flags )
--
--Moves a number of words to the left.
public procedure word_left( atom rich, atom words, atom flags )
    c_proc( WX_WORD_LEFT, {rich,words,flags})
end procedure

--/topic wxRichTextCtrl
--/proc word_right( atom rich, atom words, atom flags )
--
--Moves a number of words to the right.
public procedure word_right( atom rich, atom words, atom flags )
    c_proc( WX_WORD_RIGHT, {rich,words,flags})
end procedure

--/topic wxRichTextCtrl
--/proc begin_batch_undo( atom rich, sequence cmd_name )
--
--Starts batching undo history for commands.
public procedure begin_batch_undo( atom rich, sequence cmd_name )
    c_proc( WX_BEGIN_BATCH_UNDO, {rich,cmd_name})
end procedure

--/topic wxRichTextCtrl
--/proc end_batch_undo( atom rich )
--
--Ends batching undo command history.
public procedure end_batch_undo( atom rich )
    c_proc( WX_END_BATCH_UNDO, {rich})
end procedure

--/topic wxRichTextCtrl
--/func batching_undo( atom rich )
--
--Returns true if undo commands are being batched.
public function batching_undo( atom rich )
    return c_func( WX_BATCHING_UNDO, {rich})
end function

--/topic wxRichTextCtrl
--/proc begin_suppress_undo( atom rich )
--
--Starts suppressing undo history for commands.
public procedure begin_suppress_undo( atom rich )
    c_proc( WX_BEGIN_SUPPRESS_UNDO, {rich})
end procedure

--/topic wxRichTextCtrl
--/func suppressing_undo( atom rich )
--
--Returns true if undo history suppression is on.
public function suppressing_undo( atom rich )
    return c_func( WX_SUPPRESSING_UNDO, {rich})
end function

--/topic wxRichTextCtrl
--/proc end_suppress_undo( atom rich )
--
--Ends suppressing undo command history.
public procedure end_suppress_undo( atom rich )
    c_proc( WX_END_SUPPRESS_UNDO, {rich})
end procedure

--/topic wxRichTextCtrl
--/func has_char_text_attr( atom rich, atom first, atom last, atom text_attr_ex )
--
--Test if this whole range has character attributes of the specified kind. If any of the attributes are different within the range, the
-- test fails. You can use this to implement, for example, bold button updating. style must have flags indicating which attributes are
-- of interest.
public function has_char_text_attr( atom rich, atom first, atom last, atom text_attr_ex )
    return c_func( WX_HAS_CHAR_TEXT_ATTR, {rich,first,last,text_attr_ex})
end function

--/topic wxRichTextCtrl
--/func has_char_rich_attr( atom rich, atom first, atom last, atom rich_attr )
--
--Test if this whole range has character attributes of the specified kind. If any of the attributes are different within the range, the
-- test fails. You can use this to implement, for example, bold button updating. style must have flags indicating which attributes are
-- of interest.
public function has_char_rich_attr( atom rich, atom first, atom last, atom rich_attr )
    return c_func( WX_HAS_CHAR_RICH_ATTR, {rich,first,last,rich_attr})
end function

--/topic wxRichTextCtrl
--/func has_para_text_attr( atom rich, atom first, atom last, atom text_attr_ex )
--
--Test if this whole range has paragraph attributes of the specified kind. If any of the attributes are different within the range, the
-- test fails. You can use this to implement, for example, centering button updating. style must have flags indicating which attributes
-- are of interest.
public function has_para_text_attr( atom rich, atom first, atom last, atom text_attr_ex )
    return c_func( WX_HAS_PARA_TEXT_ATTR, {rich,first,last,text_attr_ex})
end function

--/topic wxRichTextCtrl
--/func has_para_char_attr( atom rich, atom first, atom last, atom rich_attr )
--
--Test if this whole range has paragraph attributes of the specified kind. If any of the attributes are different within the range, the
-- test fails. You can use this to implement, for example, centering button updating. style must have flags indicating which attributes
-- are of interest.
public function has_para_char_attr( atom rich, atom first, atom last, atom rich_attr )
    return c_func( WX_HAS_PARA_CHAR_ATTR, {rich,first,last,rich_attr})
end function

--/topic wxRichTextCtrl
--/func is_selection_bold( atom rich )
--
--Returns true if all of the selection is bold.
public function is_selection_bold( atom rich )
    return c_func( WX_IS_SELECTION_BOLD, {rich})
end function

--/topic wxRichTextCtrl
--/func is_selection_italics( atom rich )
--
--Returns true if all of the selection is italic.
public function is_selection_italics( atom rich )
    return c_func( WX_IS_SELECTION_ITALICS, {rich})
end function

--/topic wxRichTextCtrl
--/func is_selection_underlined( atom rich )
--
--Returns true if all of the selection is underlined.
public function is_selection_underlined( atom rich )
    return c_func( WX_IS_SELECTION_UNDERLINED, {rich})
end function

--/topic wxRichTextCtrl
--/func is_selection_aligned( atom rich, atom alignment )
--
--Returns true if all of the selection is aligned according to the specified flag.
public function is_selection_aligned( atom rich, atom alignment )
    return c_func( WX_IS_SELECTION_ALIGNED, {rich,alignment})
end function

--/topic wxRichTextCtrl
--/proc apply_bold_to_selection( atom rich )
--
--Apples bold to the selection (undoable).
public procedure apply_bold_to_selection( atom rich )
    c_proc( WX_APPLY_BOLD_TO_SELECTION, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc apply_italic_to_selection( atom rich )
--
--Applies italic to the selection (undoable).
public procedure apply_italic_to_selection( atom rich )
    c_proc( WX_APPLY_ITALIC_TO_SELECTION, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc apply_underline_to_selection( atom rich )
--
--Applies underline to the selection (undoable).
public procedure apply_underline_to_selection( atom rich )
    c_proc( WX_APPLY_UNDERLINE_TO_SELECTION, {rich})
end procedure

--/topic wxRichTextCtrl
--/proc apply_alignment_to_selection( atom rich, atom alignment )
--
--Applies the given alignment to the selection (undoable).
--
--For alignment values, see /wxTextAttrEx.
public procedure apply_alignment_to_selection( atom rich, atom alignment )
    c_proc( WX_APPLY_ALIGNMENT_TO_SELECTION, {rich,alignment})
end procedure

--/topic wxRichTextCtrl
--/proc apply_rich_style( atom rich, atom style_def )
--
--Applies the given style to the selection.
public procedure apply_rich_style( atom rich, atom style_def )
    c_proc( WX_APPLY_RICH_STYLE, {rich,style_def})
end procedure

--/topic wxRichTextCtrl
--/proc set_rich_style_sheet( atom rich, atom style_sheet )
--
--Sets the style sheet associated with the control. A style sheet allows named character and paragraph styles to be applied.
public procedure set_rich_style_sheet( atom rich, atom style_sheet )
    c_proc( WX_SET_RICH_STYLE_SHEET, {rich,style_sheet})
end procedure

--/topic wxRichTextCtrl
--/func get_rich_style_sheet( atom rich )
--
--Returns the style sheet associated with the control, if any. A style sheet allows named character and paragraph styles to be applied.
public function get_rich_style_sheet( atom rich )
    return c_func( WX_GET_RICH_STYLE_SHEET, {rich})
end function

--/topic wxRichTextCtrl
--/proc push_rich_style_sheet( atom rich, atom style_sheet )
--
public procedure push_rich_style_sheet( atom rich, atom style_sheet )
    c_proc( WX_PUSH_RICH_STYLE_SHEET, {rich,style_sheet})
end procedure

--/topic wxRichTextCtrl
--/func pop_rich_style_sheet( atom rich )
--
public function pop_rich_style_sheet( atom rich )
    return c_func( WX_POP_RICH_STYLE_SHEET, {rich})
end function

--/topic wxRichTextCtrl
--/proc apply_rich_style_sheet( atom rich, atom style_sheet )
--
--Applies the style sheet to the buffer, matching paragraph styles in the sheet against named styles in the buffer.
-- This might be useful if the styles have changed. If sheet is NULL, the sheet set with /set_rich_style_sheet is used.
--
--Currently this applies paragraph styles only.
public procedure apply_rich_style_sheet( atom rich, atom style_sheet )
    c_proc( WX_APPLY_RICH_STYLE_SHEET, {rich,style_sheet})
end procedure



--wxRichTextEvent
----------------------------------
--/topic wxRichTextEvent
--/parent wxRichTextCtrl
--/parent Events
--/info
--
--<ul>
--/li /b wxEVT_COMMAND_RICHTEXT_CHARACTER event, generated when the user presses a character key. Valid event functions: GetFlags, GetPosition,
-- GetCharacter.
--/li /b wxEVT_COMMAND_RICHTEXT_DELETE event, generated when the user presses the backspace or delete key. Valid event functions: GetFlags,
-- GetPosition.
--/li /b wxEVT_COMMAND_RICHTEXT_RETURN event, generated when the user presses the return key. Valid event functions: GetFlags, GetPosition.
--/li /b wxEVT_COMMAND_RICHTEXT_STYLE_CHANGED event, generated when styling has been applied to the control. Valid event functions: GetPosition,
-- GetRange.
--/li /b wxEVT_COMMAND_RICHTEXT_STYLESHEET_CHANGING event, generated when the control's stylesheet has changed, for example the user added, edited
-- or deleted a style. Valid event functions: GetRange, GetPosition.
--/li /b wxEVT_COMMAND_RICHTEXT_STYLESHEET_REPLACING event, generated when the control's stylesheet is about to be replaced, for example when a file
-- is loaded into the control. Valid event functions: Veto, GetOldStyleSheet, GetNewStyleSheet.
--/li /b wxEVT_COMMAND_RICHTEXT_STYLESHEET_REPLACED event, generated when the control's stylesheet has been replaced, for example when a file is
-- loaded into the control. Valid event functions: GetOldStyleSheet, GetNewStyleSheet.
--/li /b wxEVT_COMMAND_RICHTEXT_CONTENT_INSERTED event, generated when content has been inserted into the control. Valid event functions: GetPosition,
-- GetRange.
--/li /b wxEVT_COMMAND_RICHTEXT_CONTENT_DELETED event, generated when content has been deleted from the control. Valid event functions: GetPosition,
-- GetRange.
--</ul>

--/topic wxRichTextEvent
--/func get_position_rich_event( atom event )
--
--Returns the buffer position at which the event occured.
public function get_position_rich_event( atom event )
    return c_func( WX_GET_POSITION_RICH_EVENT, {event})
end function

--/topic wxRichTextEvent
--/func get_rich_event_flags( atom event )
--
--Returns flags indicating modifier keys pressed. Possible values are wxRICHTEXT_CTRL_DOWN, wxRICHTEXT_SHIFT_DOWN, and wxRICHTEXT_ALT_DOWN.
public function get_rich_event_flags( atom event )
    return c_func( WX_GET_RICH_EVENT_FLAGS, {event})
end function

--/topic wxRichTextEvent
--/func get_rich_event_old_style_sheet( atom event )
--
--Returns the old style sheet. Can be used in a wxEVT_COMMAND_RICHTEXT_STYLESHEET_CHANGING or wxEVT_COMMAND_RICHTEXT_STYLESHEET_CHANGED event handler.
public function get_rich_event_old_style_sheet( atom event )
    return c_func( WX_GET_RICH_EVENT_OLD_STYLE_SHEET, {event})
end function

--/topic wxRichTextEvent
--/func get_rich_event_range( atom event )
--
--Gets the range for the current operation.
public function get_rich_event_range( atom event )
    return c_func( WX_GET_RICH_EVENT_RANGE, {event})
end function

--/topic wxRichTextEvent
--/func get_rich_event_char( atom event )
--
--Returns the character pressed, within a wxEVT_COMMAND_RICHTEXT_CHARACTER event.
public function get_rich_event_char( atom event )
    return c_func( WX_GET_RICH_EVENT_CHAR, {event})
end function



--wxRichTextFormattingDialog
-----------------------------------------
--/topic wxRichTextFormattingDialog
--/parent wxRichTextCtrl
--/info
--

--/topic wxRichTextFormattingDialog
--/proc apply_rich_style_from_dialog( atom dialog, atom rich, atom start_range, atom end_range, atom flags )
--
public procedure apply_rich_style_from_dialog( atom dialog, atom rich, atom start_range, atom end_range, atom flags )
    c_proc( WX_APPLY_RICH_STYLE_FROM_DIALOG, {dialog,rich,start_range,end_range,flags})
end procedure

--/topic wxRichTextFormattingDialog
--/func get_rich_dialog_attributes( atom dialog )
--
public function get_rich_dialog_attributes( atom dialog )
    return c_func( WX_GET_RICH_DIALOG_ATTRIBUTES, {dialog})
end function

--/topic wxRichTextFormattingDialog
--/func get_rich_dialog_image_list( atom dialog )
--
public function get_rich_dialog_image_list( atom dialog )
    return c_func( WX_GET_RICH_DIALOG_IMAGE_LIST, {dialog})
end function

--/topic wxRichTextFormattingDialog
--/proc get_rich_dialog_style( atom dialog, atom rich, atom start_range, atom end_range )
--
public procedure get_rich_dialog_style( atom dialog, atom rich, atom start_range, atom end_range )
    c_proc( WX_GET_RICH_DIALOG_STYLE, {dialog,rich,start_range,end_range})
end procedure



--wxScrollBar
------------------------------
--/topic wxScrollBar
--/parent Controls
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li atom style [= wxSB_HORIZONTAL]
--<ul>
--/li wxSB_HORIZONTAL
--/li wxSB_VERTICAL
--</ul>
--</ul>

--/topic wxScrollBar
--/proc set_scrollbar( atom scroll, atom position, atom thumbSize, atom range, atom pageSize, atom refresh )
--
--Sets the scrollbar properties.
--<ul>
--/li /i position The position of the scrollbar in scroll units.
--/li /i thumbSize The size of the thumb, or visible portion of the scrollbar, in scroll units.
--/li /i range The maximum position of the scrollbar.
--/li /i pageSize The size of the page size in scroll units. This is the number of units the scrollbar will scroll when it is paged up or down.
-- Often it is the same as the thumb size.
--/li /i refresh true to redraw the scrollbar, false otherwise.
--</ul>
public procedure set_scrollbar( atom scroll, atom position, atom thumbSize, atom range,     atom pageSize, atom refresh )
    c_proc( WX_SET_SCROLLBAR, {scroll, position, thumbSize, range, pageSize, refresh} )
end procedure

--/topic wxScrollBar
--/func get_scroll_position( atom scroll )
--
--Returns the current position of the scrollbar.
public function get_scroll_position( atom scroll )
    return c_func( WX_GET_SCROLL_POSITION, {scroll} )
end function

--/topic wxScrollBar
--/proc set_scroll_position( atom scroll, atom pos )
--
--Sets the position of the scrollbar.
public procedure set_scroll_position( atom scroll, atom pos )
    c_proc( WX_SET_SCROLL_POSITION, {scroll, pos} )
end procedure

--/topic wxScrollBar
--/func get_scroll_range( atom scroll )
--
--Returns the length of the scrollbar.
public function get_scroll_range( atom scroll )
    return c_func( WX_GET_SCROLL_RANGE, {scroll})
end function

--/topic wxScrollBar
--/func get_scroll_page_size( atom scroll )
--
--Returns the page size of the scrollbar. This is the number of scroll units that will be scrolled when the user pages up or down.
-- Often it is the same as the thumb size.
public function get_scroll_page_size( atom scroll )
    return c_func( WX_GET_SCROLL_PAGE_SIZE, {scroll})
end function

--/topic wxScrollBar
--/func get_scroll_thumb_size( atom scroll )
--
--Returns the thumb or 'view' size.
public function get_scroll_thumb_size( atom scroll )
    return c_func( WX_GET_SCROLL_THUMB_SIZE, {scroll})
end function



--wxScrolledWindow
------------------------------
--/topic wxScrolledWindow
--/parent wxWindow
--/parent Controls
--/info
--The /b wxScrolledWindow class manages scrolling for its client area, transforming the coordinates according to the scrollbar positions,
-- and setting the scroll positions, thumb sizes and ranges according to the area in view.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li atom style [= wxHSCROL+wxVSCROLL]
--<ul>
--/li wxHSCROLL
--/li wxVSCROLL
--</ul>
--</ul>
--
--Starting from version 2.4 of wxWidgets, there are several ways to use a /b wxScrolledWindow. In particular, there are now three ways to
-- set the size of the scrolling area:
--<ul>
--/li One way is to set the scrollbars directly using a call to /set_scrollbars. This is the way it used to be in any previous
-- version of wxWidgets and it will be kept for backwards compatibility.
--/li An additional method of manual control, which requires a little less computation of your own, is to set the total size of the scrolling
-- area by calling either /set_virtual_size, or /fit_inside, and setting the scrolling increments for it by calling
-- /set_scroll_rate. Scrolling in some orientation is enabled by setting a non-zero increment for it.
--/li The most automatic and newest way is to simply let sizers determine the scrolling area. This is now the default when you set an interior
-- sizer into a /b wxScrolledWindow with /set_sizer. The scrolling area will be set to the size requested by the sizer and the scrollbars
-- will be assigned for each orientation according to the need for them and the scrolling increment set by wxScrolledWindow::SetScrollRate. As
-- above, scrolling is only enabled in orientations with a non-zero increment. You can influence the minimum size of the scrolled area controlled
-- by a sizer by calling /set_virtual_size_hints. (calling /set_scrollbars has analogous effects in wxWidgets 2.4 &em; in
-- later versions it may not continue to override the sizer)
--</ul>
--
--/b Note: if Maximum size hints are still supported by SetVirtualSizeHints, use them at your own dire risk. They may or may not have been removed
-- for 2.4, but it really only makes sense to set minimum size hints here. We should probably replace /set_virtual_size_hints with
-- /set_min_virtual_size or similar and remove it entirely in future.
--
--As with all windows, an application can draw onto a /b wxScrolledWindow using a device context.
--
--If you don't wish to calculate your own scrolling, you must call /prepare_dc when not drawing from within OnDraw, to set the device origin for the
-- device context according to the current scroll position.
--
--A /b wxScrolledWindow will normally scroll itself and therefore its child windows as well. It might however be desired to scroll a different window
-- than itself: e.g. when designing a spreadsheet, you will normally only have to scroll the (usually white) cell area, whereas the (usually grey)
-- label area will scroll very differently. For this special purpose, you can call /set_target_window which means that pressing the scrollbars will
-- scroll a different window.
--
--Note that the underlying system knows nothing about scrolling coordinates, so that all system functions (mouse events, expose events, refresh calls
-- etc) as well as the position of subwindows are relative to the "physical" origin of the scrolled window. If the user insert a child window at
-- position (10,10) and scrolls the window down 100 pixels (moving the child window out of the visible area), the child window will report a position
-- of (10,-90).

--/topic wxScrolledWindow
--/func calc_scrolled_position( atom scrollwin, atom x, atom y )
--/b Returns: Sequence: {xx, yy}
--
--Translates the logical coordinates to the device ones. For example, if a window is scrolled 10 pixels to the bottom, the device coordinates of the
-- origin are (0, 0) (as always), but the logical coordinates are (0, 10) and so the call to calc_scrolled_position(0, 10) will return 0 in /i yy.
public function calc_scrolled_position( atom scrollwin, atom x, atom y )
    return c_func( WX_CALC_SCROLLED_POSITION, {scrollwin,x,y})
end function

--/topic wxScrolledWindow
--/func calc_unscrolled_position( atom scrollwin, atom x, atom y )
--/b Returns: {xx, yy}
--
--Translates the device coordinates to the logical ones. For example, if a window is scrolled 10 pixels to the bottom, the device coordinates of the
-- origin are (0, 0) (as always), but the logical coordinates are (0, 10) and so the call to CalcUnscrolledPosition(0, 0, &xx, &yy) will return 10 in
-- /i yy.
public function calc_unscrolled_position( atom scrollwin, atom x, atom y )
    return c_func( WX_CALC_UNSCROLLED_POSITION, {scrollwin,x,y})
end function

--/topic wxScrolledWindow
--/proc set_scrollbars( atom scrollwin, atom pixelsPerUnitX, atom pixelsPerUnitY, atom noUnitsX, atom noUnitsY, atom xPos, atom yPos, atom noRefresh )
--
--Sets up vertical and/or horizontal scrollbars.
--
--/b Parameters
--
--<ul>
--/li /i pixelsPerUnitX Pixels per scroll unit in the horizontal direction.
--/li /i pixelsPerUnitY Pixels per scroll unit in the vertical direction.
--/li /i noUnitsX Number of units in the horizontal direction.
--/li /i noUnitsY Number of units in the vertical direction.
--/li /i xPos Position to initialize the scrollbars in the horizontal direction, in scroll units.
--/li /i yPos Position to initialize the scrollbars in the vertical direction, in scroll units.
--/li /i noRefresh Will not refresh window if true.
--</ul>
--
--/b Remarks
--
--The first pair of parameters give the number of pixels per 'scroll step', i.e. amount moved when the up or down scroll arrows are pressed.
-- The second pair gives the length of scrollbar in scroll steps, which sets the size of the virtual window.
--
--xPos and yPos optionally specify a position to scroll to immediately.
--
--For example, the following gives a window horizontal and vertical scrollbars with 20 pixels per scroll step, and a size of 50 steps
-- (1000 pixels) in each direction.
--
--/code
--  set_scrollbars( window, 20, 20, 50, 50, 0, 0, 0 );
--/endcode
--
--/wxScrolledWindow manages the page size itself, using the current client window size as the page size.
--
--Note that for more sophisticated scrolling applications, for example where scroll steps may be variable according to the position in the document,
-- it will be necessary to derive a new class from wxWindow, overriding OnSize and adjusting the scrollbars appropriately.
public procedure set_scrollbars( atom scrollwin, atom pixelsPerUnitX, atom pixelsPerUnitY, atom noUnitsX, atom noUnitsY, atom xPos, atom yPos, atom noRefresh )
    c_proc( WX_SET_SCROLLBARS, {scrollwin,pixelsPerUnitX,pixelsPerUnitY,noUnitsX,noUnitsY,xPos,yPos,noRefresh})
end procedure

--/topic wxScrolledWindow
--/proc set_virtual_size( atom scrollwin, atom width, atom height )
--
--Sets the virtual size of the window in pixels.
public procedure set_virtual_size( atom scrollwin, atom width, atom height )
    c_proc( WX_SET_VIRTUAL_SIZE, {scrollwin,width,height})
end procedure

--/topic wxScrolledWindow
--/proc fit_inside( atom scrollwin )
--
--Similar to /fit_window, but sizes the interior (virtual) size of a window. Mainly useful with scrolled windows to reset scrollbars after sizing
-- changes that do not trigger a size event, and/or scrolled windows without an interior sizer. This function similarly won't do anything if there
-- are no subwindows.
public procedure fit_inside( atom scrollwin )
    c_proc( WX_FIT_INSIDE, {scrollwin})
end procedure

--/topic wxScrolledWindow
--/proc enable_scrolling( atom scrollwin, atom xScrolling, atom yScrolling )
--
--Enable or disable physical scrolling in the given direction. Physical scrolling is the physical transfer of bits up or down the screen when a
-- scroll event occurs. If the application scrolls by a variable amount (e.g. if there are different font sizes) then physical scrolling will not
-- work, and you should switch it off. Note that you will have to reposition child windows yourself, if physical scrolling is disabled.
--
--/b Parameters
--
--<ul>
--/li /i xScrolling If true, enables physical scrolling in the x direction.
--/li /i yScrolling If true, enables physical scrolling in the y direction.
--</ul>
--
--/b Remarks
--
--Physical scrolling may not be available on all platforms. Where it is available, it is enabled by default.
public procedure enable_scrolling( atom scrollwin, atom xScrolling, atom yScrolling )
    c_proc( WX_ENABLE_SCROLLING, {scrollwin,xScrolling,yScrolling})
end procedure

--/topic wxScrolledWindow
--/func get_scroll_pixels_per_unit( atom scrollwin )
--
--Get the number of pixels per scroll unit (line), in each direction, as set by /set_scrollbars.
-- A value of zero indicates no scrolling in that direction.
public function get_scroll_pixels_per_unit( atom scrollwin )
    return c_func( WX_GET_SCROLL_PIXELS_PER_UNIT, {scrollwin})
end function

--/topic wxScrolledWindow
--/func get_view_start( atom scrollwin )
--/b Returns: {x, y}
--
--Get the position at which the visible portion of the window starts.
--
--/b Remarks
--
--If either of the scrollbars is not at the home position, x and/or y will be greater than zero. Combined with /get_client_size, the application
-- can use this function to efficiently redraw only the visible portion of the window. The positions are in logical scroll units, not pixels, so
-- to convert to pixels you will have to multiply by the number of pixels per scroll increment.
public function get_view_start( atom scrollwin )
    return c_func( WX_GET_VIEW_START, {scrollwin})
end function

--/topic wxScrolledWindow
--/func get_virtual_size( atom scrollwin )
--
--Gets the size in device units of the scrollable window area (as opposed to the client size, which is the area of the window currently visible).
public function get_virtual_size( atom scrollwin )
    return c_func( WX_GET_VIRTUAL_SIZE, {scrollwin})
end function

--/topic wxScrolledWindow
--/func is_retained( atom scrollwin )
--
--Motif only: true if the window has a backing bitmap.
public function is_retained( atom scrollwin )
    return c_func( WX_IS_RETAINED, {scrollwin})
end function

--/topic wxScrolledWindow
--/proc prepare_dc( atom scrollwin, atom dc )
--
--Call this function to prepare the device context for drawing a scrolled image. It sets the device origin according to the current scroll position.
--
--If you wish to implement OnPaint yourself, you must call this function yourself. For example:
--
--/code
--  procedure onPaint( atom this, atom event_type, atom id, atom event )
--
--    atom dc
--    sequence size
--
--    dc = create( wxPaintDC, {this} )
--    prepare_dc( this, dc )
--
--    size = get_dc_size( dc )
--
--    set_pen( dc, wxBLACK_PEN )
--    draw_line( dc, {0, 0} & size )
--
--    delete_instance( dc )
--
--  end procedure
--/endcode
public procedure prepare_dc( atom scrollwin, atom dc )
    c_proc( WX_PREPARE_DC, {scrollwin,dc})
end procedure

--/topic wxScrolledWindow
--/proc scroll_window( atom scrollwin, atom x, atom y )
--
--Scrolls a window so the view start is at the given point.
--
--/b Parameters
--
--<ul>
--/li /i x The x position to scroll to, in scroll units.
--/li /i y The y position to scroll to, in scroll units.
--</ul>
public procedure scroll_window( atom scrollwin, atom x, atom y )
    c_proc( WX_SCROLL_WINDOW, {scrollwin,x,y})
end procedure

--/topic wxScrolledWindow
--/proc set_scroll_rate( atom scrollwin, atom xstep, atom ystep )
--
--Set the horizontal and vertical scrolling increment only. See the pixelsPerUnit parameter in /set_scrollbars.
public procedure set_scroll_rate( atom scrollwin, atom xstep, atom ystep )
    c_proc( WX_SET_SCROLL_RATE, {scrollwin,xstep,ystep})
end procedure

--/topic wxScrolledWindow
--/proc set_target_window( atom scrollwin, atom target )
--
--Call this function to tell /b wxScrolledWindow to perform the actual scrolling on a different window (and not on itself).
public procedure set_target_window( atom scrollwin, atom target )
    c_proc( WX_SET_TARGET_WINDOW, {scrollwin,target})
end procedure



--wxSingleChoiceDialog
-----------------------------------------
--/topic wxSingleChoiceDialog
--/parent Dialogs
--/info
--This class represents a dialog that shows a list of strings, and allows the user to select one.
--
--Double-clicking on a list item is equivalent to single-clicking and then pressing OK.
--
--<div class="subsection">Creation parameters</div>
--<ul>
--/li /b parent
--/li /b message
--/li /b caption
--/li /b choices
--/li /b style [= wxCHOICEDLG_STYLE]
--/li /b x [= -1]
--/li /b y [= -1]
--</ul>



--wxSlider
------------------------------
--/topic wxSlider
--/parent Controls
--/parent wxScrollBar
--/info
--
--<div class="subsection">Creation parameters</div>
--<ul>
--/li /b parent
--/li /b id
--/li /b initial_value
--/li /b minValue
--/li /b maxValue
--/li /b x [= -1]
--/li /b y [= -1]
--/li /b cx [= -1]
--/li /b cy [= -1]
--/li /b style [= wxSL_HORIZONTAL]
--<ul>
--/li wxSL_HORIZONTAL      = wxHORIZONTAL, -- 4
--/li wxSL_VERTICAL        = wxVERTICAL,   -- 8
--/li wxSL_TICKS           = #0010,
--/li wxSL_AUTOTICKS       = wxSL_TICKS,
--/li wxSL_LABELS          = #0020,
--/li wxSL_LEFT            = #0040,
--/li wxSL_TOP             = #0080,
--/li wxSL_RIGHT           = #0100,
--/li wxSL_BOTTOM          = #0200,
--/li wxSL_BOTH            = #0400,
--/li wxSL_SELRANGE        = #0800
--</ul>
--</ul>

--/topic wxSlider
--/func get_slide_value( atom slider )
--
--Returns an integer representing the current position of the slider.
public function get_slide_value( atom slider )
    return c_func( WX_GET_SLIDE_VALUE, {slider} )
end function

--/topic wxSlider
--/proc set_slide_value( atom slider, atom val )
--
--Sets the specified slide to the position noted by val.
public procedure set_slide_value( atom slider, atom val )
    c_proc( WX_SET_SLIDE_VALUE, {slider, val} )
end procedure



--wxSocketClient
------------------------------
--/topic wxSocketClient
--/parent Networking
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "integer flags"
--<ul><table>
--<tr><td> /li wxSOCKET_NONE    <td>Normal functionality.
--<tr><td> /li wxSOCKET_NOWAIT  <td>Read/write as much data as possible and return immediately.
--<tr><td> /li wxSOCKET_WAITALL <td>Wait for all required data to be read/written unless an error occurs.
--<tr><td> /li wxSOCKET_BLOCK   <td>Block the GUI (do not yield) while reading/writing data.
--</table></ul>
--</ul>

--/topic wxSocketClient
--/func wait_on_connect( atom client, integer seconds, integer milliseconds )
--
--Wait until a connection request completes, or until the specified timeout elapses.
-- Use this function after issuing a call to /socket_connect with /i wait set to false.
public function wait_on_connect( atom client, integer seconds, integer milliseconds )
    return c_func( WX_WAIT_ON_CONNECT, {client, seconds, milliseconds})
end function

--/topic wxSocketClient
--/func socket_connect( atom client, atom address, integer wait )
--Connects to a server using the specified address.
--
--If wait is TRUE, Connect will wait until the connection completes. Warning: This will block the GUI.
--
--If wait is FALSE, Connect will try to establish the connection and return
--immediately, without blocking the GUI. When used this way, even if Connect
--returns FALSE, the connection request can be completed later. To detect this,
--use WaitOnConnect, or catch wxSOCKET_CONNECTION events (for successful establishment)
--and wxSOCKET_LOST events (for connection failure).
public function socket_connect( atom client, atom address, integer wait )
    return c_func( WX_SOCKET_CONNECT, {client, address, wait} )
end function



--wxSocketServer
------------------------------
--/topic wxSocketServer
--/parent Networking
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom /wxIPV4address"
--/li integer flags [= wxSOCKET_NONE] (see /wxSocketClient)
--</ul>
--
--In order to receive data through sockets, you first need to create a wxSocketServer object
--to wait for new connections.  Then, in an event handler, you need to create a new wxSocketServer
--object that will handle the connection, and set up an event handler for that socket, as well.
--Alternatively, you could call /socket_accept() with /i wait = 1, although this will not return
--until a connection has been made.
--
--Example:
--/code
--
--     server = create( wxSocketServer, address )
--     server_id = get_id( server )
--     set_socket_notify( server, {wxSOCKET_CONNECTION_FLAG, wxSOCKET_LOST_FLAG})
--     socket_event_handler( server, main_form, server_id )
--
--     procedure main_socket_event( atom this, atom event_type, atom id, atom event )
--         sequence bytes
--         atom socket
--
--         event_type = get_socket_event( event )
--
--         if id = server_id and event_type = wxSOCKET_CONNECTION then
--             connection = socket_accept( server, 0 )
--             connection_id = get_id( connection )
--             set_socket_notify( connection, {wxSOCKET_INPUT_FLAG, wxSOCKET_LOST_FLAG})
--             socket_event_handler( connection_id, main_form, connection_id )
--             set_event_handler( main_form, connection_id, wxEVT_SOCKET, routine_id("main_socket_event"))
--
--         elsif event_type = wxSOCKET_LOST then
--             delete_instance( socket_from_event( event ) )
--
--         elsif event_type = wxSOCKE_INPUT then
--             bytes = socket_read( socket_from_event( event ) )
--
--         end if
--     end procedure
--     set_event_handler( main_form, server_id, wxEVT_SOCKET, routine_id("main_socket_event")
--
--/endcode

--/topic wxSocketServer
--/func socket_accept( atom socket, integer wait )
--
--Accepts an incoming connection request, and creates a new wxSocketBase object which
--represents the server-side of the connection.
--
--If wait is TRUE and there are no pending connections to be accepted, it will wait for
--the next incoming connection to arrive. Warning: This will block the GUI.
--
--If wait is FALSE, it will try to accept a pending connection if there is one, but it
--will always return immediately without blocking the GUI. If you want to use Accept in
--this way, you can either check for incoming connections with WaitForAccept or catch
--wxSOCKET_CONNECTION events, then call Accept once you know that there is an incoming
--connection waiting to be accepted.
public function socket_accept( atom socket, integer wait )
    return c_func( WX_SOCKET_ACCEPT, {socket, wait} )
end function



--wxSound
-----------------------------------------
--/topic wxSound
--/parent Classes
--/info
--
--This class represents a short sound (loaded from Windows WAV file),
--that can be stored in memory and played. Currently this class is
--implemented on Windows and Unix (and uses either Open Sound System or Simple DirectMedia Layer).
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "sequence filename" (or resource file)
--/li integer isResource [= wxFalse]
--</ul>

--/topic wxSound
--/func sound_create( atom sound, object file, atom isResource )
--Constructs a wave object from a file or resource.
public function sound_create( atom sound, object file, atom isResource )
    return c_func( WX_SOUND_CREATE, {sound,file,isResource})
end function

--/topic wxSound
--/func sound_is_ok( atom sound )
--Returns true if the object contains a successfully loaded file or resource, false otherwise.
public function sound_is_ok( atom sound )
    return c_func( WX_SOUND_IS_OK, {sound})
end function

--/topic wxSound
--/func sound_play( atom sound, atom flags )
--Plays the sound file. If another sound is playing, it will be interrupted.
-- Returns true on success, false otherwise. Note that in general it is possible to delete the
-- object which is being asynchronously played any time after calling this function and the sound
-- would continue playing, however this currently doesn't work under Windows for sound objects
-- loaded from memory data.
--
--The possible values for flags are:<ul>
--/li /b wxSOUND_SYNC Play will block and wait until the sound is replayed.
--/li /b wxSOUND_ASYNC  Sound is played asynchronously, Play returns immediately
--/li /b wxSOUND_ASYNC | wxSOUND_LOOP  Sound is played asynchronously and loops until
-- another sound is played, wxSound::Stop is called or the program terminates.
--</ul>
public function sound_play( atom sound, atom flags )
    return c_func( WX_SOUND_PLAY, {sound,flags})
end function

--/topic wxSound
--/proc sound_stop(  )
--If a sound is played, this function stops it.
public procedure sound_stop(  )
    c_proc( WX_SOUND_STOP, {})
end procedure



--wxSpinCtrl
------------------------------
--/topic wxSpinCtrl
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence text
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= wxSP_ARROW_KEYS]
--<ul>
--/li wxSP_HORIZONTAL
--/li wxSP_VERTICAL
--/li wxSP_ARROW_KEYS
--/li wxSP_WRAP
--</ul>
--/li integer min [= 0]
--/li integer max [= 100]
--/li integer initial [= 0]
--</ul>

--/topic wxSpinCtrl
--/func get_spin_value( atom spin )
--
--Returns the current Value of a wxSpinCtrl
public function get_spin_value( atom spin )
    return c_func( WX_GET_SPIN_VALUE, {spin} )
end function

--/topic wxSpinCtrl
--/proc set_spin_value( atom spin, atom val )
--
--Set's the current Value of a wxSpinCtrl to val.
public procedure set_spin_value( atom spin, atom val )
    c_proc( WX_SET_SPIN_VALUE, {spin, val} )
end procedure



--wxSplashScreen
------------------------------
--/topic wxSplashScreen
--/parent Controls
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom bitmap"
--/li /b "integer style"
--<ul>
--/li wxSPLASH_CENTRE_ON_PARENT
--/li wxSPLASH_CENTRE_ON_SCREEN
--/li wxSPLASH_NO_CENTRE
--/li wxSPLASH_TIMEOUT
--/li wxSPLASH_NO_TIMEOUT
--</ul>
--/li /b "integer milliseconds"
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= wxSIMPLE_BORDER|wxFRAME_NO_TASKBAR|wxSTAY_ON_TOP] (see /wxFrame)
--</ul>



--wxSplitterEvent
---------------------
--/topic wxSplitterEvent
--/parent Events
--/Classes
--/info
--
--This class represents the events generated by a splitter control. Also there is only one event class, the data associated to the different
-- events is not the same and so not all accessor functions may be called for each event. The documentation mentions the kind of event(s) for
-- which the given accessor function makes sense: calling it for other types of events will result in assert failure (in debug mode) and will
-- return meaningless results.

--/topic wxSplitterEvent
--/proc split_event_set_sash_pos( atom event, atom pos )
--
--In the case of /b wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGED events, sets the new sash position. In the case of
-- /b wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGING events, sets the new tracking bar position so visual feedback
-- during dragging will represent that change that will actually take place. Set to -1 from the event handler
-- code to prevent repositioning.
--
--May only be called while processing /b wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGING and
-- /b wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGED events.
public procedure split_event_set_sash_pos( atom event, atom pos )
    c_proc( WX_SPLIT_EVENT_SET_SASH_POS, {event,pos})
end procedure

--/topic wxSplitterEvent
--/func split_event_x( atom event )
--
--Returns the x coordinate of the double-click point.
--
--May only be called while processing /b wxEVT_COMMAND_SPLITTER_DOUBLECLICKED events.
public function split_event_x( atom event )
    return c_func( WX_SPLIT_EVENT_X, {event})
end function

--/topic wxSplitterEvent
--/func split_event_y( atom event )
--
--Returns the y coordinate of the double-click point.
--
--May only be called while processing /b wxEVT_COMMAND_SPLITTER_DOUBLECLICKED events.
public function split_event_y( atom event )
    return c_func( WX_SPLIT_EVENT_Y, {event})
end function

--/topic wxSplitterEvent
--/func split_event_win_removed( atom event )
--
--Returns a pointer to the /wxWindow being removed when a splitter window is unsplit.
--
--May only be called while processing /b wxEVT_COMMAND_SPLITTER_UNSPLIT events.
public function split_event_win_removed( atom event )
    return c_func( WX_SPLIT_EVENT_WIN_REMOVED, {event})
end function

--/topic wxSplitterEvent
--/func split_event_get_sash_pos( atom event )
--
--Returns the new sash position.
--
--May only be called while processing /b wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGING and
-- /b wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGED events.
public function split_event_get_sash_pos( atom event )
    return c_func( WX_SPLIT_EVENT_GET_SASH_POS, {event})
end function



--wxSplitterWindow
------------------------------
--/topic wxSplitterWindow
--/parent Controls
--/parent Sizers
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li atom style [= wxSP_BORDER | wxSP_FULLSASH]
--</ul>

--/topic wxSplitterWindow
--/proc split_window( atom split, atom win1, atom win2, integer orient, integer pos )
--
--<ul>
--/li /b split: /wxSplitterWindow to split
--/li /b win1: First window (placed on left or top, depending on /i orient)
--/li /b win2: Second window (placed on right or bottom, depending on /i orient)
--/li /b orient: wxHORIZONTAL or wxVERTICAL
--/li /b pos: Initial position, in pixels, for the sash
--</ul>
--Splits the window, using /i win1 and /i win2, which must have /i split as their
--parent.  /i orient
public procedure split_window( atom split, atom win1, atom win2, integer orient, integer pos )
    c_proc( WX_SPLIT_WINDOW, {split, win1, win2, orient, pos} )
end procedure

--/topic wxSplitterWindow
--/proc set_min_pane_size( atom split, integer size )
--
--The default minimum pane size is zero, which means that either pane
--can be reduced to zero by dragging the sash, thus removing one of the panes.
--To prevent this behaviour (and veto out-of-range sash dragging), set a
--minimum size, for example 20 pixels. If the wxSP_PERMIT_UNSPLIT style
--is used when a splitter window is created, the window may be unsplit
--even if minimum size is non-zero
public procedure set_min_pane_size( atom split, integer size )
    c_proc( WX_SET_MIN_PANE_SIZE, {split, size} )
end procedure

--/topic wxSplitterWindow
--/proc set_sash_gravity( atom split, atom gravity )
--
--Sets the sash gravity.
--
--<ul>
--/li /i gravity The sash gravity. Value between 0.0 and 1.0.
--</ul>
--
--Gravity is real factor which controls position of sash while resizing wxSplitterWindow.
-- Gravity tells wxSplitterWindow how much will left/top window grow while resizing.
--
--Example values:
--<ul>
--/li 0.0 - only the bottom/right window is automatically resized
--/li 0.5 - both windows grow by equal size
--/li 1.0 - only left/top window grows
--</ul>
--
--Gravity should be a real value between 0.0 and 1.0.
--
--Default value of sash gravity is 0.0. That value is compatible with previous (before gravity was introduced) behaviour of wxSplitterWindow.
public procedure set_sash_gravity( atom split, atom gravity )
    c_proc( WX_SET_SASH_GRAVITY, {split, gravity})
end procedure

--/topic wxSplitterWindow
--/func get_sash_gravity( atom split )
--
--Returns the current sash gravity.
public function get_sash_gravity( atom split )
    return c_func( WX_GET_SASH_GRAVITY, {split})
end function

--/topic wxSplitterWindow
--/proc set_splitter_sash_position( atom split, integer pos, integer redraw )
--
--Sets the sash position of the /wxSplitterWindow at /i pos pixels.  The windows
--will not be redrawn unless /i redraw is /b 1.
public procedure set_splitter_sash_position( atom split, integer pos, integer redraw )
    c_proc( WX_SET_SPLITTER_SASH_POSITION, {split, pos, redraw} )
end procedure

--/topic wxSplitterWindow
--/func get_splitter_sash_position( atom split )
--
--Returns the position of the sash in the specified /wxSplitterWindow in pixels.
public function get_splitter_sash_position( atom split )
    return c_func( WX_GET_SPLITTER_SASH_POSITION, {split} )
end function

--/topic wxSplitterWindow
--/proc unsplit_window( atom split, atom window )
public procedure unsplit_window( atom split, atom window )
    c_proc( WX_UNSPLIT_WINDOW, {split, window} )
end procedure

--/topic wxSplitterWindow
--/proc initialize_split( atom split, atom window )
public procedure initialize_split( atom split, atom window )
    c_proc( WX_INITIALIZE_SPLIT, {split, window} )
end procedure



--wxStaticBitmap
------------------------------
--/topic wxStaticBitmap
--/parent Graphics
--/parent Controls
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li atom /wxBitmap [= 0]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cy [=-1]
--/li atom style [=-1]
--</ul>

--/topic wxStaticBitmap
--/func get_staticbitmap( atom staticbitmap )
--
--Returns the asociated bitmap
public function get_staticbitmap ( atom staticbitmap )
    return c_func( WX_GET_STATICBITMAP, {staticbitmap} )
end function

-- Orphan
--/topic wxStaticBitmap
--/proc set_staticbitmap( atom staticbitmap, atom bitmap )
--
--Sets the displayed bitmap of the static.
public procedure set_staticbitmap ( atom staticbitmap, atom bitmap)
    c_proc( WX_SET_STATICBITMAP, {staticbitmap, bitmap} )
end procedure



--wxStaticBox
------------------------------
--/topic wxStaticBox
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence label_text [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= 0]
--</ul>
--A static box is a rectangle drawn around other panel items to denote a logical grouping of items.
--
-- Please note that a static box should not be used as the parent for the controls it contains, instead
-- they should be siblings of each other. Although using a static box as a parent might work in some
-- versions of wxWidgets, it results in a crash under, for example, wxGTK.
--
-- Also, please note that because of this, the order in which you create new controls is important.
-- Create your wxStaticBox control before any siblings that are to appear inside the wxStaticBox in
-- order to preserve the correct Z-Order of controls.



--wxStaticBoxSizer
------------------------------
--/topic wxStaticBoxSizer
--/parent Sizers
--/info
--wxStaticBoxSizer is a sizer derived from wxBoxSizer but adds a static box around the sizer.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li To use an existing /wxStaticBox:
--<ul>
--/li /b "atom /wxStaticBox"
--/li /b "integer orientation"
--<ul>
--/li wxVERTICAL
--/li wxHORIZONTAL
--</ul>
--</ul>
--/li To have the /wxStaticBox created automatically:
--<ul>
--/li /b "integer orienataion"
--<ul>
--/li wxVERTICAL
--/li wxHORIZONTAL
--</ul>
--/li /b "atom parent"
--/li /b "sequence label_text"
--</ul>
--</ul>



--wxStaticLine
------------------------------
--/topic wxStaticLine
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [=-1]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cy [=-1]
--/li integer style [= wxHORIZONTAL] (Note that wxVERTICAL doesn't seem to work under Windows)
--</ul>



--wxStaticText
------------------------------
--/topic wxStaticText
--/parent Controls
--/info
--
--This control is used for creating a static label.  The text in the
--label will be automatically refreshed (as opposed to using /wx_puts).
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence text [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li atom style [= wxST_NO_AUTORESIZE]
--<ul>
--/li wxALIGN_LEFT
--/li wxALIGN_RIGHT
--/li wxALIGN_CENTRE
--/li wxST_NO_AUTORESIZE
--
--/b Note: by default, the control will adjust its size to exactly fit to the size of the text
-- when SetLabel is called. If this style flag is given, the control will not change its size
-- (this style is especially useful with controls which also have wxALIGN_RIGHT or CENTER style
-- because otherwise they won't make sense any longer after a call to /set_label)
--</ul>
--</ul>

--/topic wxStaticText
--/proc set_label( atom static_label, sequence text )
--
--Sets the static text label and updates the controls size to exactly fit the label unless the control has wxST_NO_AUTORESIZE flag.
public procedure set_label( atom static_label, sequence text )
    c_proc( WX_SET_LABEL, {static_label, text} )
end procedure

--/topic wxStaticText
--/proc wrap( atom text, atom width )
--This function wraps the controls label so that each of its lines becomes at most /i width pixels wide if possible
-- (the lines are broken at words boundaries so it might not be the case if words are too long). If /i width is negative,
-- no wrapping is done.
public procedure wrap( atom text, atom width )
    c_proc( WX_WRAP, {text,width})
end procedure



--wxStatusBar
------------------------------
--/topic wxStatusBar
--/parent Controls
--/info
--The status bar should be a child of a /wxFrame.  <div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer fields [= 1]
--/li integer style [= 0]
--</ul>

--/topic wxStatusBar
--/func get_field_rect( atom bar, atom field )
--
--Returns the size and position of a field's internal bounding rectangle.
public function get_field_rect( atom bar, atom field )
    return c_func( WX_GET_FIELD_RECT, {bar,field})
end function

--/topic wxStatusBar
--/proc set_status_fields( atom bar, sequence widths )
--
--/i widths:  An array of n integers, each of which is a status field width in pixels.
--A value of -1 indicates that the field is variable width; at least one field must be -1.
public procedure set_status_fields( atom bar, sequence widths )
    c_proc( WX_SET_STATUS_FIELDS, {bar, widths} )
end procedure

--/topic wxStatusBar
--/proc set_status_text( atom bar, sequence text, integer field )
--
--Set the status text of field /i field (zero based).  This routine does
--nothing if /i field refers to a non-existant status field.
public procedure set_status_text( atom bar, sequence text, integer field )
    c_proc( WX_SET_STATUS_TEXT, {bar, text, field} )
end procedure

--/topic wxStatusBar
--/func get_status_fields_count( atom bar )
--
--Returns the number of fields in the status bar.
public function get_status_fields_count( atom bar )
    return c_func( WX_GET_STATUS_FIELDS_COUNT, {bar})
end function

--/topic wxStatusBar
--/func get_status_text( atom bar, atom field )
--
--Returns the string associated with a status bar /i field.
public function get_status_text( atom bar, atom field )
    return c_func( WX_GET_STATUS_TEXT, {bar,field})
end function

--/topic wxStatusBar
--/proc pop_status_text( atom bar, atom field )
--
--Sets the /i field text to the top of the stack, and pops the stack of saved strings.
public procedure pop_status_text( atom bar, atom field )
    c_proc( WX_POP_STATUS_TEXT, {bar,field})
end procedure

--/topic wxStatusBar
--/proc push_status_text( atom bar, sequence text, atom field )
--
--Saves the current /i field text in a per field stack, and sets the /i field text to the string passed as argument.
public procedure push_status_text( atom bar, sequence text, atom field )
    c_proc( WX_PUSH_STATUS_TEXT, {bar,text,field})
end procedure

--/topic wxStatusBar
--/proc set_min_status_height( atom bar, atom height )
--
--Sets the minimal possible height for the status bar. The real height may be bigger than the height specified here depending on the size
-- of the font used by the status bar.
public procedure set_min_status_height( atom bar, atom height )
    c_proc( WX_SET_MIN_STATUS_HEIGHT, {bar,height})
end procedure

--/topic wxStatusBar
--/proc set_status_styles( atom bar, object styles )
--
--Sets the styles of the fields in the status line which can make fields appear flat or raised instead of the standard sunken 3D border.
--
--/b Parameters
--<ul>
--/li /i n The number of fields in the status bar. Must be equal to the number passed to SetFieldsCount the last time it was called.
--/li /i styles Contains an array of n integers with the styles for each field. There are three possible styles:<ul>
--/li /i wxSB_NORMAL (default) The field appears sunken with a standard 3D border.
--/li /i wxSB_FLAT No border is painted around the field so that it appears flat.
--/li /i wxSB_RAISED A raised 3D border is painted around the field.</ul>
--</ul>
public procedure set_status_styles( atom bar, object styles )
    c_proc( WX_SET_STATUS_STYLES, {bar,styles})
end procedure



--wxStdDialogButtonSizer
-----------------------------------------
--/topic wxStdDialogButtonSizer
--/parent wxSizers
--/parent wxDialog
--/info
--This class creates button layouts which conform to the standard button spacing and ordering defined by the platform
-- or toolkit's user interface guidelines (if such things exist). By using this class, you can ensure that all your
-- standard dialogs look correct on all major platforms. Currently it conforms to the Windows, GTK+ and Mac OS X human
-- interface guidelines.
--
--When there aren't interface guidelines defined for a particular platform or toolkit, wxStdDialogButtonSizer reverts to
-- the Windows implementation.
--
--To use this class, first add buttons to the sizer by calling /add_button() and then call Realize in order to create
-- the actual button layout used. Other than these special operations, this sizer works like any other sizer.

--/topic wxStdDialogButtonSizer
--/proc add_button( atom dbs, atom button )
--Adds a button to the wxStdDialogButtonSizer /i dbs.
--
--The button must have one of the following identifiers:
--/li wxID_OK
--/li wxID_YES
--/li wxID_SAVE
--/li wxID_APPLY
--/li wxID_NO
--/li wxID_CANCEL
--/li wxID_HELP
--/li wxID_CONTEXT_HELP
public procedure add_button( atom dbs, atom button )
    c_proc( WX_ADD_BUTTON, {dbs,button})
end procedure

--/topic wxStdDialogButtonSizer
--/proc realize( atom dbs )
--Rearranges the buttons and applies proper spacing between buttons to make them match the platform or toolkit's interface guidelines.
public procedure realize( atom dbs )
    c_proc( WX_REALIZE, {dbs})
end procedure



--wxTaskBarIcon
-----------------------------------------
--/topic wxTaskBarIcon
--/parent Controls
--/info
--

--/topic wxTaskBarIcon
--/func get_taskbar_menu( atom taskbar )
--/b Returns: A wxMenu object
--
--Gets the wxMenu associated with the taskbar icon.
public function get_taskbar_menu( atom taskbar )
    return c_func( WX_GET_TASKBAR_MENU, {taskbar})
end function

--/topic wxTaskBarIcon
--/proc set_taskbar_icon( atom taskbar, atom icon, sequence tooltip )
--
--Sets the icon and tooltip text for the taskbar icon.
public procedure set_taskbar_icon( atom taskbar, atom icon, sequence tooltip )
    c_proc( WX_SET_TASKBAR_ICON, {taskbar,icon,tooltip})
end procedure

--/topic wxTaskBarIcon
--/func is_icon_installed( atom taskbar )
--
--Returns true if /set_taskbar_icon was called with no subsequent /remove_taskbar_icon.
public function is_icon_installed( atom taskbar )
    return c_func( WX_IS_ICON_INSTALLED, {taskbar})
end function

--/topic wxTaskBarIcon
--/proc remove_taskbar_icon( atom taskbar )
--
--Removes the icon previously set with set_taskbar_icon.
public procedure remove_taskbar_icon( atom taskbar )
    c_proc( WX_REMOVE_TASKBAR_ICON, {taskbar})
end procedure



--wxTextAttr
------------------------------
--/topic wxTextAttr
--/parent wxRichTextCtrl
--/parent Classes
--/info
--

--/topic wxTextAttr
--/func get_text_attr_color( atom attr )
--
--Returns the Color Attribute of a wxTextAttr
public function get_text_attr_color( atom attr )
    return c_func( WX_GET_TEXT_ATTR_COLOR, {attr} )
end function

--/topic wxTextAttr
--/func get_text_attr_back_color( atom attr )
--
--Returns the Background color of a wxTextAttr
public function get_text_attr_back_color( atom attr )
    return c_func( WX_GET_TEXT_ATTR_BACK_COLOR, {attr} )
end function

--/topic wxTextAttr
--/func get_text_attr_font( atom attr )
--
--Returns the Font stored in a wxTextAttr
public function get_text_attr_font( atom attr )
    return c_func( WX_GET_TEXT_ATTR_FONT, {attr} )
end function

--/topic wxTextAttr
--/proc set_text_attr_color( atom attr, atom color )
--
--Stores the Text color in a wxTextAttr
public procedure set_text_attr_color( atom attr, atom color )
    c_proc( WX_SET_TEXT_ATTR_COLOR, {attr, color} )
end procedure

--/topic wxTextAttr
--/proc set_text_attr_back_color( atom attr, atom color )
--
--Stores the Background color in a wxTextAttr
public procedure set_text_attr_back_color( atom attr, atom color )
    c_proc( WX_SET_TEXT_ATTR_BACK_COLOR, {attr, color} )
end procedure

--/topic wxTextAttr
--/proc set_text_attr_font( atom attr, atom font )
--
--Stores the Font in a wxTextAttr
public procedure set_text_attr_font( atom attr, atom font )
    c_proc( WX_SET_TEXT_ATTR_FONT, {attr, font} )
end procedure



--wxTextAttrEx
-----------------------------------------
--/topic wxTextAttrEx
--/parent wxTextAttr
--/parent wxRichTextCtrl
--/parent Classes
--/info
--

--/topic wxTextAttrEx
--/proc set_textattrex_char_style_name( atom attr, object name )
--
--Sets the character style name.
public procedure set_textattrex_char_style_name( atom attr, object name )
    c_proc( WX_SET_TEXTATTREX_CHAR_STYLE_NAME, {attr,name})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_para_style_name( atom attr, object name )
--
--Sets the name of the paragraph style.
public procedure set_textattrex_para_style_name( atom attr, object name )
    c_proc( WX_SET_TEXTATTREX_PARA_STYLE_NAME, {attr,name})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_list_style_name( atom attr, object name )
--
--Sets the list style name.
public procedure set_textattrex_list_style_name( atom attr, object name )
    c_proc( WX_SET_TEXTATTREX_LIST_STYLE_NAME, {attr,name})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_para_spacing_after( atom attr, atom after )
--
--Sets the spacing after a paragraph, in tenths of a millimetre.
public procedure set_textattrex_para_spacing_after( atom attr, atom after )
    c_proc( WX_SET_TEXTATTREX_PARA_SPACING_AFTER, {attr,after})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_para_spacing_before( atom attr, atom before )
--
--Sets the spacing before a paragraph, in tenths of a millimetre.
public procedure set_textattrex_para_spacing_before( atom attr, atom before )
    c_proc( WX_SET_TEXTATTREX_PARA_SPACING_BEFORE, {attr,before})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_line_spacing( atom attr, atom spacing )
--
--Sets the line spacing. spacing is a multiple, where 10 means single-spacing, 15 means 1.5 spacing, and 20 means double spacing.
--
-- The following constants are defined for convenience:
--/code
--#define wxTEXT_ATTR_LINE_SPACING_NORMAL       10
--#define wxTEXT_ATTR_LINE_SPACING_HALF         15
--#define wxTEXT_ATTR_LINE_SPACING_TWICE        20
--/endcode
public procedure set_textattrex_line_spacing( atom attr, atom spacing )
    c_proc( WX_SET_TEXTATTREX_LINE_SPACING, {attr,spacing})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_bullet_style( atom attr, atom style )
--
--Sets the bullet style. The following styles can be passed:
--
--/code
--#define wxTEXT_ATTR_BULLET_STYLE_NONE             0x00000000
--#define wxTEXT_ATTR_BULLET_STYLE_ARABIC           0x00000001
--#define wxTEXT_ATTR_BULLET_STYLE_LETTERS_UPPER        0x00000002
--#define wxTEXT_ATTR_BULLET_STYLE_LETTERS_LOWER        0x00000004
--#define wxTEXT_ATTR_BULLET_STYLE_ROMAN_UPPER      0x00000008
--#define wxTEXT_ATTR_BULLET_STYLE_ROMAN_LOWER      0x00000010
--#define wxTEXT_ATTR_BULLET_STYLE_SYMBOL           0x00000020
--#define wxTEXT_ATTR_BULLET_STYLE_BITMAP           0x00000040
--#define wxTEXT_ATTR_BULLET_STYLE_PARENTHESES      0x00000080
--#define wxTEXT_ATTR_BULLET_STYLE_PERIOD           0x00000100
--#define wxTEXT_ATTR_BULLET_STYLE_STANDARD         0x00000200
--#define wxTEXT_ATTR_BULLET_STYLE_RIGHT_PARENTHESIS    0x00000400
--#define wxTEXT_ATTR_BULLET_STYLE_OUTLINE          0x00000800
--#define wxTEXT_ATTR_BULLET_STYLE_ALIGN_LEFT       0x00000000
--#define wxTEXT_ATTR_BULLET_STYLE_ALIGN_RIGHT      0x00001000
--#define wxTEXT_ATTR_BULLET_STYLE_ALIGN_CENTRE     0x00002000
--/endcode
--
--Currently wxTEXT_ATTR_BULLET_STYLE_BITMAP is not supported.
public procedure set_textattrex_bullet_style( atom attr, atom style )
    c_proc( WX_SET_TEXTATTREX_BULLET_STYLE, {attr,style})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_bullet_number( atom attr, atom num )
--
--Sets the bullet number.
public procedure set_textattrex_bullet_number( atom attr, atom num )
    c_proc( WX_SET_TEXTATTREX_BULLET_NUMBER, {attr,num})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_bullet_text( atom attr, object text )
--
--Sets the bullet text, which could be a symbol, or (for example) cached outline text.
public procedure set_textattrex_bullet_text( atom attr, object text )
    c_proc( WX_SET_TEXTATTREX_BULLET_TEXT, {attr,text})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_bullet_name( atom attr, object name )
--
--Sets the standard bullet name, applicable if the bullet style is wxTEXT_ATTR_BULLET_STYLE_STANDARD. See /get_textattrex_bullet_name for a
-- list of supported names, and how to expand the range of supported types.
public procedure set_textattrex_bullet_name( atom attr, object name )
    c_proc( WX_SET_TEXTATTREX_BULLET_NAME, {attr,name})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_url_name( atom attr, object url )
--
--Sets the URL for the content. Sets the wxTEXT_ATTR_URL style; content with this style causes wxRichTextCtrl to show a hand cursor over it,
-- and wxRichTextCtrl generates a wxTextUrlEvent when the content is clicked.
public procedure set_textattrex_url_name( atom attr, object url )
    c_proc( WX_SET_TEXTATTREX_URL_NAME, {attr,url})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_page_break( atom attr, atom page_break )
--
--Specifies a page break before this paragraph.
public procedure set_textattrex_page_break( atom attr, atom page_break )
    c_proc( WX_SET_TEXTATTREX_PAGE_BREAK, {attr,page_break})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_bullet_font( atom attr, object font_name )
--
--Sets the name of the font associated with the bullet symbol. Only valid for attributes with wxTEXT_ATTR_BULLET_SYMBOL.
public procedure set_textattrex_bullet_font( atom attr, object font_name )
    c_proc( WX_SET_TEXTATTREX_BULLET_FONT, {attr,font_name})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_effects( atom attr, atom effects )
--
--Sets the text effects, a bit list of styles.
--
--The following styles can be passed:
--/code
--#define wxTEXT_ATTR_EFFECT_NONE                   0x00000000
--#define wxTEXT_ATTR_EFFECT_CAPITALS               0x00000001
--#define wxTEXT_ATTR_EFFECT_SMALL_CAPITALS         0x00000002
--#define wxTEXT_ATTR_EFFECT_STRIKETHROUGH          0x00000004
--#define wxTEXT_ATTR_EFFECT_DOUBLE_STRIKETHROUGH   0x00000008
--#define wxTEXT_ATTR_EFFECT_SHADOW                 0x00000010
--#define wxTEXT_ATTR_EFFECT_EMBOSS                 0x00000020
--#define wxTEXT_ATTR_EFFECT_OUTLINE                    0x00000040
--#define wxTEXT_ATTR_EFFECT_ENGRAVE                    0x00000080
--#define wxTEXT_ATTR_EFFECT_SUPERSCRIPT                0x00000100
--#define wxTEXT_ATTR_EFFECT_SUBSCRIPT              0x00000200
--/endcode
--
--Of these, only wxTEXT_ATTR_EFFECT_CAPITALS and wxTEXT_ATTR_EFFECT_STRIKETHROUGH are implemented. wxTEXT_ATTR_EFFECT_CAPITALS capitalises
-- text when displayed (leaving the case of the actual buffer text unchanged), and wxTEXT_ATTR_EFFECT_STRIKETHROUGH draws a line through text.
--
--To set effects, you should also pass wxTEXT_ATTR_EFFECTS to wxTextAttr::SetFlags, and call wxTextAttrEx::SetTextEffectFlags with the styles
-- (taken from the above set) that you are interested in setting.
public procedure set_textattrex_effects( atom attr, atom effects )
    c_proc( WX_SET_TEXTATTREX_EFFECTS, {attr,effects})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_effect_flags( atom attr, atom flags )
--
--Sets the text effect bits of interest. You should also pass wxTEXT_ATTR_EFFECTS to /set_textattrex_flags.
public procedure set_textattrex_effect_flags( atom attr, atom flags )
    c_proc( WX_SET_TEXTATTREX_EFFECT_FLAGS, {attr,flags})
end procedure

--/topic wxTextAttrEx
--/proc set_textattrex_outline_level( atom attr, atom level )
--
--Specifies the outline level. Zero represents normal text. At present, the outline level is not used, but may be used in future for
-- determining list levels and for applications that need to store document structure information.
public procedure set_textattrex_outline_level( atom attr, atom level )
    c_proc( WX_SET_TEXTATTREX_OUTLINE_LEVEL, {attr,level})
end procedure

--/topic wxTextAttrEx
--/func get_textattrex_char_style( atom attr )
--
--Returns the name of the character style.
public function get_textattrex_char_style( atom attr )
    return c_func( WX_GET_TEXTATTREX_CHAR_STYLE, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_para_style( atom attr )
--
--Returns the name of the paragraph style.
public function get_textattrex_para_style( atom attr )
    return c_func( WX_GET_TEXTATTREX_PARA_STYLE, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_list_style( atom attr )
--
--Returns the name of the list style.
public function get_textattrex_list_style( atom attr )
    return c_func( WX_GET_TEXTATTREX_LIST_STYLE, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_para_spacing_before( atom attr )
--
--Returns the space in tenths of a millimeter before the paragraph.
public function get_textattrex_para_spacing_before( atom attr )
    return c_func( WX_GET_TEXTATTREX_PARA_SPACING_BEFORE, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_para_spacing_after( atom attr )
--
--Returns the space in tenths of a millimeter after the paragraph.
public function get_textattrex_para_spacing_after( atom attr )
    return c_func( WX_GET_TEXTATTREX_PARA_SPACING_AFTER, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_line_spacing( atom attr )
--
--Returns the line spacing value, one of wxTEXT_ATTR_LINE_SPACING_NORMAL, wxTEXT_ATTR_LINE_SPACING_HALF, and wxTEXT_ATTR_LINE_SPACING_TWICE.
public function get_textattrex_line_spacing( atom attr )
    return c_func( WX_GET_TEXTATTREX_LINE_SPACING, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_bullet_spacing( atom attr )
--
public function get_textattrex_bullet_spacing( atom attr )
    return c_func( WX_GET_TEXTATTREX_BULLET_SPACING, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_bullet_number( atom attr )
--
--Returns the bullet number.
public function get_textattrex_bullet_number( atom attr )
    return c_func( WX_GET_TEXTATTREX_BULLET_NUMBER, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_bullet_text( atom attr )
--
--Returns the bullet text, which could be a symbol, or (for example) cached outline text.
public function get_textattrex_bullet_text( atom attr )
    return c_func( WX_GET_TEXTATTREX_BULLET_TEXT, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_bullet_name( atom attr )
--
--Returns the standard bullet name, applicable if the bullet style is wxTEXT_ATTR_BULLET_STYLE_STANDARD.
--
--Currently the following standard bullet names are supported:
--<ul>
--/li standard/circle
--/li standard/square
--/li standard/diamond
--/li standard/triangle
--</ul>
public function get_textattrex_bullet_name( atom attr )
    return c_func( WX_GET_TEXTATTREX_BULLET_NAME, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_bullet_font( atom attr )
--
--Returns a string containing the name of the font associated with the bullet symbol. Only valid for attributes with wxTEXT_ATTR_BULLET_SYMBOL.
public function get_textattrex_bullet_font( atom attr )
    return c_func( WX_GET_TEXTATTREX_BULLET_FONT, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_url( atom attr )
--
--Returns the URL for the content. Content with wxTEXT_ATTR_URL style causes wxRichTextCtrl to show a hand cursor over it, and wxRichTextCtrl generates a wxTextUrlEvent when the content is clicked.
public function get_textattrex_url( atom attr )
    return c_func( WX_GET_TEXTATTREX_URL, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_effects( atom attr )
--
--Returns the text effects, a bit list of styles. See /set_textattrex_effects for details.
public function get_textattrex_effects( atom attr )
    return c_func( WX_GET_TEXTATTREX_EFFECTS, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_effect_flags( atom attr )
--
--Returns the text effect bits of interest. See /set_textattrex_flags for further information.
public function get_textattrex_effect_flags( atom attr )
    return c_func( WX_GET_TEXTATTREX_EFFECT_FLAGS, {attr})
end function

--/topic wxTextAttrEx
--/func get_textattrex_outline_level( atom attr )
--
--Returns the outline level.
public function get_textattrex_outline_level( atom attr )
    return c_func( WX_GET_TEXTATTREX_OUTLINE_LEVEL, {attr})
end function

--/topic wxTextAttrEx
--/func has_font_weight_textattrex( atom attr )
--
public function has_font_weight_textattrex( atom attr )
    return c_func( WX_HAS_FONT_WEIGHT_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_font_size_textattrex( atom attr )
--
public function has_font_size_textattrex( atom attr )
    return c_func( WX_HAS_FONT_SIZE_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_font_italic_textattrex( atom attr )
--
public function has_font_italic_textattrex( atom attr )
    return c_func( WX_HAS_FONT_ITALIC_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_font_underlined_textattrex( atom attr )
--
public function has_font_underlined_textattrex( atom attr )
    return c_func( WX_HAS_FONT_UNDERLINED_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_font_name_textattrex( atom attr )
--
public function has_font_name_textattrex( atom attr )
    return c_func( WX_HAS_FONT_NAME_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_para_spacing_after_textattrex( atom attr )
--
--Returns true if the attribute object specifies spacing after a paragraph.
public function has_para_spacing_after_textattrex( atom attr )
    return c_func( WX_HAS_PARA_SPACING_AFTER_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_para_spacing_before_textattrex( atom attr )
--
--Returns true if the attribute object specifies spacing after a paragraph.
public function has_para_spacing_before_textattrex( atom attr )
    return c_func( WX_HAS_PARA_SPACING_BEFORE_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_line_spacing_textattrex( atom attr )
--
--Returns true if the attribute object specifies line spacing.
public function has_line_spacing_textattrex( atom attr )
    return c_func( WX_HAS_LINE_SPACING_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_char_style_textattrex( atom attr )
--
--Returns true if the attribute object specifies a character style name.
public function has_char_style_textattrex( atom attr )
    return c_func( WX_HAS_CHAR_STYLE_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_para_style_textattrex( atom attr )
--
--Returns true if the attribute object specifies a paragraph style name.
public function has_para_style_textattrex( atom attr )
    return c_func( WX_HAS_PARA_STYLE_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_list_style_textattrex( atom attr )
--
--Returns true if the attribute object specifies a list style name.
public function has_list_style_textattrex( atom attr )
    return c_func( WX_HAS_LIST_STYLE_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_bullet_style_textattrex( atom attr )
--
--Returns true if the attribute object specifies a bullet style.
public function has_bullet_style_textattrex( atom attr )
    return c_func( WX_HAS_BULLET_STYLE_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_bullet_number_textattrex( atom attr )
--
--Returns true if the attribute object specifies a bullet number.
public function has_bullet_number_textattrex( atom attr )
    return c_func( WX_HAS_BULLET_NUMBER_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_bullet_text_textattrex( atom attr )
--
--Returns true if the attribute object specifies bullet text (usually containing a symbol).
public function has_bullet_text_textattrex( atom attr )
    return c_func( WX_HAS_BULLET_TEXT_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_bullet_name_textattrex( atom attr )
--
--Returns true if the attribute object specifies a standard bullet name.
public function has_bullet_name_textattrex( atom attr )
    return c_func( WX_HAS_BULLET_NAME_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_url_textattrex( atom attr )
--
--Returns true if the attribute object specifies a URL.
public function has_url_textattrex( atom attr )
    return c_func( WX_HAS_URL_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_page_break_textattrex( atom attr )
--
--Returns true if the attribute object specifies a page break before this paragraph.
public function has_page_break_textattrex( atom attr )
    return c_func( WX_HAS_PAGE_BREAK_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_text_effects_textattrex( atom attr )
--
--Returns true if the attribute object specifies text effects.
public function has_text_effects_textattrex( atom attr )
    return c_func( WX_HAS_TEXT_EFFECTS_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func has_text_effect_textattrex( atom attr, atom effect )
--
public function has_text_effect_textattrex( atom attr, atom effect )
    return c_func( WX_HAS_TEXT_EFFECT_TEXTATTREX, {attr,effect})
end function

--/topic wxTextAttrEx
--/func has_outline_level_textattrex( atom attr )
--
--Returns true if the attribute object specifies an outline level.
public function has_outline_level_textattrex( atom attr )
    return c_func( WX_HAS_OUTLINE_LEVEL_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func is_char_style_textattrex( atom attr )
--
--Returns true if the object represents a character style, that is, the flags specify a font or a text background or foreground colour.
public function is_char_style_textattrex( atom attr )
    return c_func( WX_IS_CHAR_STYLE_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func is_para_style_textattrex( atom attr )
--
--Returns true if the object represents a paragraph style, that is, the flags specify alignment, indentation, tabs, paragraph spacing, or bullet style.
public function is_para_style_textattrex( atom attr )
    return c_func( WX_IS_PARA_STYLE_TEXTATTREX, {attr})
end function

--/topic wxTextAttrEx
--/func is_default_textattrex( atom attr )
--
--Returns false if we have any attributes set, true otherwise.
public function is_default_textattrex( atom attr )
    return c_func( WX_IS_DEFAULT_TEXTATTREX, {attr})
end function



--wxTextCtrl
------------------------------
--/topic wxTextCtrl
--/parent Controls
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence text [= ""]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li atom style  [= 0]
--<ul><table>
--<tr><td valign="top"> /li wxTE_PROCESS_ENTER <td>The control will generate the message wxEVENT_TYPE_TEXT_ENTER_COMMAND
-- (otherwise pressing <Enter> is either processed internally by the control or used for navigation
-- between dialog controls).
--<tr><td valign="top"> /li wxTE_PROCESS_TAB   <td>The control will receive EVT_CHAR messages for TAB pressed - normally,
-- TAB is used for passing to the next control in a dialog instead. For the control created with this
-- style, you can still use Ctrl-Enter to pass to the next control from the keyboard.
--<tr><td> /li wxTE_MULTILINE <td>The text control allows multiple lines.
--<tr><td> /li wxTE_PASSWORD  <td>The text will be echoed as asterisks.
--<tr><td> /li wxTE_READONLY  <td>The text will not be user-editable.
--<tr><td valign="top"> /li wxTE_RICH      <td>Use rich text control under Win32, this allows to have more than 64Kb of text in
-- the control even under Win9x. This style is ignored under other platforms.
--<tr><td valign="top"> /li wxTE_RICH2     <td>Use rich text control version 2.0 or 3.0 under Win32, this style is ignored
-- under other platforms. You must use this style if you want to set the background color (see /wxTextAttr).
--<tr><td valign="top"> /li wxTE_AUTO_URL  <td>Highlight the URLs and generate the wxTextUrlEvents when mouse events
-- occur over them. This style is supported under Win32 only and requires wxTE_RICH.
--<tr><td valign="top"> /li wxTE_NOHIDESEL <td>By default, the Windows text control doesn't show the selection when
-- it doesn't have focus - use this style to force it to always show it. It doesn't do anything
-- under other platforms.
--<tr><td> /li wxHSCROLL      <td>A horizontal scrollbar will be created. No effect under GTK+.
--<tr><td> /li wxTE_LEFT      <td>The text control will be left-justified (default).
--<tr><td> /li wxTE_CENTRE    <td>The text control will be centre-justified.
--<tr><td> /li wxTE_RIGHT     <td>The text control will be right-justified.
--<tr><td> /li wxTE_DONTWRAP  <td>Same as wxHSCROLL style.
--</table></ul>
--</ul>
--Events associated with wxTextCtrl:
--<ul><table>
--<tr><td valign="top"> /li wxEVT_COMMAND_TEXT_UPDATED <td>Generated when the text changes. Notice that this event will always be sent when
-- the text controls contents changes - whether this is due to user input or comes from the program itself (for example, if
-- /set_text_value() is called)
--<tr><td> /li wxEVT_COMMAND_TEXT_ENTER <td>Generated when enter is pressed in a single-line text control.
--<tr><td> /li wxEVT_TEXT_URL           <td>A mouse event occured over an URL in the text control (Win32 only)
--<tr><td> /li wxEVT_TEXT_MAXLEN        <td>User tried to enter more text into the control than the limit set by SetMaxLength.
--</table></ul>

--/topic wxTextCtrl
--/func xy_to_position( atom ctrl, atom x, atom y )
--
--Converts the given zero based column (x) and line number (y) to a position.
public function xy_to_position( atom ctrl, atom x, atom y )
    return c_func( WX_XY_TO_POSITION, {ctrl,x,y})
end function

--/topic wxTextCtrl
--/func position_to_xy( atom ctrl, atom position )
--
--Converts given position to a zero-based column, line number pair.
public function position_to_xy( atom ctrl, atom position )
    return c_func( WX_POSITION_TO_XY, {ctrl,position})
end function

--/topic wxTextCtrl
--/func text_hit_test( atom ctrl, atom x, atom y )
--
--This function finds the character at the specified position expressed in pixels.
public function text_hit_test( atom ctrl, atom x, atom y )
    return c_func( WX_TEXT_HIT_TEST, {ctrl,x,y})
end function

--/topic wxTextCtrl
--/func get_text_value( atom text )
--
--This retrives the text stored in a wxTextCtrl buffer.  It will return the actual string, if you want
--the actual value, such as a numeric value, use get_text_number() instead.
public function get_text_value( atom text )
    return c_func( WX_GET_TEXT_VALUE, { text } )
end function

--/topic wxTextCtrl
--/func get_text_number( atom textctrl )
--
--Returns the value of the text control after passing it through Euphoria's value() function.
public function get_text_number( atom textctrl )
    sequence val
    val = stdget:value( c_func( WX_GET_TEXT_VALUE, {textctrl} ))
    return val[2]
end function

--/topic wxTextCtrl
--/proc write_text( atom ctrl, sequence text )
--
--Writes the text at the current insertion point.
public procedure write_text( atom ctrl, sequence text )
    c_proc( WX_WRITE_TEXT, {ctrl, text} )
end procedure

--/topic wxTextCtrl
--/proc set_text( atom text, sequence val )
--
--Replaces the text in the control with /i val.
public procedure set_text( atom text, sequence val )
    c_proc( WX_SET_TEXT, {text, val} )
end procedure

--/topic wxTextCtrl
--/proc set_text_selection( atom text, atom first, atom last )
--
--Set the highlight area in a wxTextCtrl
public procedure set_text_selection( atom text, atom first, atom last )
    c_proc( WX_SET_TEXT_SELECTION, {text, first, last} )
end procedure

--/topic wxTextCtrl
--/func get_text_length( atom text )
--
--Returns the length of the text buffer currently in the Control.
public function get_text_length( atom text )
    return c_func( WX_GET_TEXT_LENGTH, {text} )
end function

--/topic wxTextCtrl
--/proc select_all_text( atom text )
--
--Set's the highlight area for the wxTextCtrl to the entire buffer.
--Similarly, you can do: set_text_selection(myEdit,0,-1) or set_text_selection(myEdit,0,get_text_length(myEdit))
public procedure select_all_text( atom text )
    c_proc( WX_SELECT_ALL_TEXT, {text} )
end procedure

--/topic wxTextCtrl
--/proc append_text( atom text, sequence val )
--
--This appends text to the end of the wxTextCtrl buffer.
public procedure append_text(atom text, sequence val)
    c_proc( WX_APPEND_TEXT, {text, val} )
end procedure

--/topic wxTextCtrl
--/proc set_text_style( atom text, atom first, atom last, atom attr )
--<ul>
--/li /b text: The /wxTextCtrl
--/li /b first: The start of selection to change.
--/li /b last: The end of selection to change.
--/li /b attr: The /wxTextAttr which holds the new style for the selection
--</ul>
public procedure set_text_style( atom text, atom first, atom last, atom attr )
    c_proc( WX_SET_TEXT_STYLE, {text, first, last, attr} )
end procedure

--/topic wxTextCtrl
--/proc set_default_text_style( atom text, atom attr )
--
--Changes the default style to use for the new text which is going to be added to the
--control using /write_text().
public procedure set_default_text_style( atom text, atom attr )
    c_proc( WX_SET_DEFAULT_TEXT_STYLE, {text, attr} )
end procedure

--/topic wxTextCtrl
--/proc copy_text( atom text )
--
--Copies the selected text to the clipboard.
public procedure copy_text( atom text )
    c_proc( WX_COPY_TEXT, {text} )
end procedure

--/topic wxTextCtrl
--/proc cut_text( atom text )
--
--Copies the selected text to the clipboard and removes the selection.
public procedure cut_text( atom text )
    c_proc( WX_CUT_TEXT, {text} )
end procedure

--/topic wxTextCtrl
--/proc paste_text( atom text )
--
--Pastes the text from the clipboard to the text item.
public procedure paste_text( atom text )
    c_proc( WX_PASTE_TEXT, {text} )
end procedure

--/topic wxTextCtrl
--/proc undo_text( atom text )
--
--Undo last operation in the text item.
public procedure undo_text( atom text )
    c_proc( WX_UNDO_TEXT, {text} )
end procedure

--/topic wxTextCtrl
--/proc redo_text( atom text )
--
--Redo last operation in the text item.
public procedure redo_text( atom text )
    c_proc( WX_REDO_TEXT, {text} )
end procedure

--/topic wxTextCtrl
--/proc clear_text( atom text )
--
--Clear text in the text item.
public procedure clear_text( atom text )
    c_proc( WX_CLEAR_TEXT, {text} )
end procedure

--/topic wxTextCtrl
--/proc set_maximum_text_length( atom text, atom max_text )
--
--Set maximum number of lines in text item.
public procedure set_maximum_text_length( atom text, atom max_text )
    c_proc( WX_SET_MAXIMUM_TEXT_LENGTH, {text, max_text} )
end procedure

--/topic wxTextCtrl
--/func can_copy_text( atom text )
--
--Returns true if the selection can be copied to the clipboard.
public function can_copy_text( atom text )
    return c_func( WX_CAN_COPY_TEXT, {text})
end function

--/topic wxTextCtrl
--/func can_cut_text( atom text )
--
--Returns true if the selection can be cut to the clipboard.
public function can_cut_text( atom text )
    return c_func( WX_CAN_CUT_TEXT, {text})
end function

--/topic wxTextCtrl
--/func can_paste_text( atom text )
--
--Returns true if the contents of the clipboard can be pasted into the text control. On some platforms (Motif, GTK) this is an approximation
-- and returns true if the control is editable, false otherwise.
public function can_paste_text( atom text )
    return c_func( WX_CAN_PASTE_TEXT, {text})
end function

--/topic wxTextCtrl
--/func can_redo_text( atom text )
--
--Returns true if there is a redo facility available and the last operation can be redone.
public function can_redo_text( atom text )
    return c_func( WX_CAN_REDO_TEXT, {text})
end function

--/topic wxTextCtrl
--/func can_undo_text( atom text )
--
--Returns true if there is an undo facility available and the last operation can be undone.
public function can_undo_text( atom text )
    return c_func( WX_CAN_UNDO_TEXT, {text})
end function

--/topic wxTextCtrl
--/proc discard_text_edits( atom text )
--
--Resets the internal 'modified' flag as if the current edits had been saved.
public procedure discard_text_edits( atom text )
    c_proc( WX_DISCARD_TEXT_EDITS, {text})
end procedure

--/topic wxTextCtrl
--/func get_default_text_style( atom text )
--
--Returns the style currently used for the new text.
public function get_default_text_style( atom text )
    return c_func( WX_GET_DEFAULT_TEXT_STYLE, {text})
end function

--/topic wxTextCtrl
--/func get_text_insertion_point( atom text )
--Returns the insertion point. This is defined as the zero based index of the character position to the right of the insertion point.
-- For example, if the insertion point is at the end of the text control, it is equal to /b length( /get_text_value( /i text ) ).
public function get_text_insertion_point( atom text )
    return c_func( WX_GET_TEXT_INSERTION_POINT, {text})
end function

--/topic wxTextCtrl
--/func get_text_line_length( atom text, atom line )
--
--Gets the length of the specified line, not including any trailing newline character(s).
public function get_text_line_length( atom text, atom line )
    return c_func( WX_GET_TEXT_LINE_LENGTH, {text,line})
end function

--/topic wxTextCtrl
--/func get_text_lines( atom text )
--
--Returns the number of lines in the text control buffer.
public function get_text_lines( atom text )
    return c_func( WX_GET_TEXT_LINES, {text})
end function

--/topic wxTextCtrl
--/func get_text_line( atom text, atom line )
--
--Returns the contents of a given line in the text control, not including any trailing newline character(s).
public function get_text_line( atom text, atom line )
    return c_func( WX_GET_TEXT_LINE, {text,line})
end function

--/topic wxTextCtrl
--/func get_text_range( atom text, atom text_from, atom text_to )
--
--Returns the string containing the text starting in the positions from and up to to in the control. The positions must have been
-- returned by another wxTextCtrl method.
--
--Please note that the positions in a multiline wxTextCtrl do not correspond to the indices in the string returned by /get_text_value
-- because of the different new line representations (CR or CR LF) and so this method should be used to obtain the correct results
-- instead of extracting parts of the entire value. It may also be more efficient, especially if the control contains a lot of data.
public function get_text_range( atom text, atom text_from, atom text_to )
    return c_func( WX_GET_TEXT_RANGE, {text,text_from,text_to})
end function

--/topic wxTextCtrl
--/func get_text_selection_range( atom text )
--
--Gets the current selection span. If the returned values are equal, there was no selection.
--
--Please note that the indices returned may be used with the other wxTextCtrl methods but don't necessarily represent the correct indices
-- into the string returned by GetValue() for multiline controls under Windows (at least,) you should use /get_text_selection to get the
-- selected text.
public function get_text_selection_range( atom text )
    return c_func( WX_GET_TEXT_SELECTION_RANGE, {text})
end function

--/topic wxTextCtrl
--/func get_text_selection( atom text )
--
--Gets the text currently selected in the control. If there is no selection, the returned string is empty.
public function get_text_selection( atom text )
    return c_func( WX_GET_TEXT_SELECTION, {text})
end function

--/topic wxTextCtrl
--/func get_text_style( atom text, atom position )
--
--Returns the style at this position in the text control. Not all platforms support this function.
public function get_text_style( atom text, atom position )
    return c_func( WX_GET_TEXT_STYLE, {text,position})
end function

--/topic wxTextCtrl
--/func is_text_editable( atom text )
--
--Returns true if the controls contents may be edited by user (note that it always can be changed by the program), i.e. if the control
-- hasn't been put in read-only mode by a previous call to /set_text_editable.
public function is_text_editable( atom text )
    return c_func( WX_IS_TEXT_EDITABLE, {text})
end function

--/topic wxTextCtrl
--/func is_text_empty( atom text )
--
--Returns true if the control is currently empty. This is the same as /b length( /get_text_value( /i text ) ) but can be much more efficient
-- for the multiline controls containing big amounts of text.
--
--This function is new since wxWidgets version 2.7.1
public function is_text_empty( atom text )
    return c_func( WX_IS_TEXT_EMPTY, {text})
end function

--/topic wxTextCtrl
--/func is_text_modified( atom text )
--
--Returns true if the text has been modified by user. Note that calling /set_text doesn't make the control modified.
public function is_text_modified( atom text )
    return c_func( WX_IS_TEXT_MODIFIED, {text})
end function

--/topic wxTextCtrl
--/func is_text_multi_line( atom text )
--
--Returns true if this is a multi line edit control and false otherwise.
public function is_text_multi_line( atom text )
    return c_func( WX_IS_TEXT_MULTI_LINE, {text})
end function

--/topic wxTextCtrl
--/func is_text_single_line( atom text )
--
--Returns true if this is a single line edit control and false otherwise.
public function is_text_single_line( atom text )
    return c_func( WX_IS_TEXT_SINGLE_LINE, {text})
end function

--/topic wxTextCtrl
--/func load_text_file( atom text, sequence file )
--
--Loads and displays the named file, if it exists.
public function load_text_file( atom text, sequence file )
    return c_func( WX_LOAD_TEXT_FILE, {text,file})
end function

--/topic wxTextCtrl
--/proc mark_text_dirty( atom text )
--
--Mark text as modified (dirty).
public procedure mark_text_dirty( atom text )
    c_proc( WX_MARK_TEXT_DIRTY, {text})
end procedure

--/topic wxTextCtrl
--/proc replace_text( atom text, atom pos_from, atom pos_to, sequence str )
--
--Replaces the text starting at the first position up to (but not including) the character at the last position with the given text.
public procedure replace_text( atom text, atom pos_from, atom pos_to, sequence str )
    c_proc( WX_REPLACE_TEXT, {text,pos_from,pos_to,str})
end procedure

--/topic wxTextCtrl
--/func save_text_file( atom text, sequence file )
--
--Saves the contents of the control in a text file.
public function save_text_file( atom text, sequence file )
    return c_func( WX_SAVE_TEXT_FILE, {text,file})
end function

--/topic wxTextCtrl
--/proc set_text_editable( atom text, atom editable )
--
--Makes the text item editable or read-only, overriding the wxTE_READONLY flag.
public procedure set_text_editable( atom text, atom editable )
    c_proc( WX_SET_TEXT_EDITABLE, {text,editable})
end procedure

--/topic wxTextCtrl
--/proc set_text_insertion_point( atom text, atom pos )
--
--Sets the insertion point at the given position.
public procedure set_text_insertion_point( atom text, atom pos )
    c_proc( WX_SET_TEXT_INSERTION_POINT, {text,pos})
end procedure

--/topic wxTextCtrl
--/proc set_text_insertion_point_end( atom text )
--
--Sets the insertion point at the end of the text control.
public procedure set_text_insertion_point_end( atom text )
    c_proc( WX_SET_TEXT_INSERTION_POINT_END, {text})
end procedure

--/topic wxTextCtrl
--/proc set_text_modified( atom text, atom modified )
--
--Marks the control as being modified by the user or not.
public procedure set_text_modified( atom text, atom modified )
    c_proc( WX_SET_TEXT_MODIFIED, {text,modified})
end procedure

--/topic wxTextCtrl
--/proc change_text_value( atom text, sequence str )
--
--Sets the text value and marks the control as not-modified (which means that /is_text_modified would return false).
--
--Note that this function will /i not generate the wxEVT_COMMAND_TEXT_UPDATED event.
--
--This function is new since wxWidgets version 2.7.1
public procedure change_text_value( atom text, sequence str )
    c_proc( WX_CHANGE_TEXT_VALUE, {text,str})
end procedure

--/topic wxTextCtrl
--/proc show_text_position( atom text, atom pos )
--
--Makes the line containing the given position visible.
public procedure show_text_position( atom text, atom pos )
    c_proc( WX_SHOW_TEXT_POSITION, {text,pos})
end procedure



--wxTextDataObject
------------------------------
--/topic wxTextDataObject
--/parent Clipboard
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li sequence text [= ""]
--</ul>



--wxTextEntryDialog
-----------------------------------------
--/topic wxTextEntryDialog
--/parent Dialogs
--/info
--
--Implementation of wxGenericDialog that prompts for a text string.
--Use /show_modal to display and /get_text_value to retrieve the entered value
-- and /set_text to set the text.
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li /b "sequence message"
--/li sequencecaption [= "Please enter text"]
--/li sequencedefault [= ""]
--/li atom style [= wxOK | wxCANCEL | wxCENTRE]
--/li integer pos_x [= -1]
--/li integer pos_y [= -1]
--</ul>



--wxTimer
------------------------------
--/topic wxTimer
--/parent Timers
--/parent Events
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--</ul>
--The /i parent is important, as this is the only way to receive wxTimer events.
--
--The /i id is also important, especially if you plan to have more than one wxTimer,
--because timer events can only be distinguished by their id.
--
--To set a timer event handler, use wxEVT_TIMER:
--
--Example:
--/code
--
--    constant
--    main = create( wxFrame, {0,-1,"Timer"} ),
--    timer_id = new_id(),
--    timer = create( wxTimer, {main, timer_id} )
--
--    procedure on_timer( atom this, atom event_type, atom id, atom event )
--        puts(1,"Timer\n")
--    end procedure
--
--    -- NOTE: using the parent and the id to identify the timer.
--    set_event_handler( main, timer_id, wxEVT_TIMER, routine_id("on_timer"))
--
--    -- set the timer to fire every 1 second, repeatedly
--    start_timer( timer, 1000, 0 )
--
--    wxMain( main )
--
--/endcode

--/topic wxTimer
--/proc start_timer( atom timer, atom msec, integer oneshot )
--
--Set the timer to fire every /i msec millisecond ( 1000msec = 1sec ).  /i oneshot
--indicates whether the timer should stop firing after the first time, or fire
--repeatedly until /stop_timer() is called.
public procedure start_timer( atom timer, atom msec, integer oneshot )
    c_proc( WX_START_TIMER, {timer, msec, oneshot} )
end procedure

--/topic wxTimer
--/proc stop_timer( atom timer )
--
--Stops a timer from firing until reset using /start_timer().
public procedure stop_timer( atom timer )
    c_proc( WX_STOP_TIMER, {timer} )
end procedure



--wxToggleButton
------------------------------
--/topic wxToggleButton
--/parent wxButton
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence label_text [= ""]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cy [=-1]
--/li integer style [= 0]
--</ul>



--wxToolBar
------------------------------
--/topic wxToolBar
--/parent Controls
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [=-1]
--/li integer y [=-1]
--/li integer cx [=-1]
--/li integer cx [=-1]
--/li integer style [= wxTB_HORIZONTAL | wxNO_BORDER]
--</ul>

--/topic wxToolBar
--/proc set_tool_bitmap_size( atom bar, integer x, integer y )
--
--Sets the size of the images to be loaded into the toolbar. The default is 32 x 32.
--
--/b Note: this does not seem to have an effect under Linux (our investigation of wxWidgets source
-- doesn't ever show these properties being used by wxGTK).
public procedure set_tool_bitmap_size( atom bar, integer x, integer y )
    c_proc( WX_SET_TOOL_BITMAP_SIZE, {bar, x, y} )
end procedure

--/topic wxToolBar
--/proc set_tool_margins( atom toolbar, integer x, integer y )
--
--Set the values to be used as margins for the toolbar.
public procedure set_tool_margins( atom toolbar, integer x, integer y )
    c_proc( WX_SET_TOOL_MARGINS, {toolbar, x, y} )
end procedure

--/topic wxToolBar
--/func add_tool( atom bar, integer id, atom b1, atom b2, sequence label_text, sequence s1, sequence s2, integer kind )
--
--<ul>
--/li bar: toolbar to add to
--/li id: the tool's id
--/li b1: bitmap
--/li b2: bitmap
--/li s1: short help string
--/li s2: long help string
--/li kind: type of item (wxITEM_NORMAL, wxITEM_CHECK, wxITEM_RADIO)
--</ul>
public function add_tool( atom bar, integer id, atom b1, atom b2, sequence label_text, sequence s1, sequence s2, integer kind )
    return c_func( WX_ADD_TOOL, {bar, id, b1, b2, label_text, s1, s2, kind} )
end function

--/topic wxToolBar
--/proc show_toolbar( atom bar )
--
--Must be called after tools are added.
public procedure show_toolbar( atom bar )
    c_proc( WX_SHOW_TOOLBAR, {bar} )
end procedure

--/topic wxToolBar
--/proc add_separator( atom bar )
--
--This adds a line from the top to the bottom of the toolbar. It must be
--declared between the two tools that you want it to go between.
public procedure add_separator( atom bar )
    c_proc( WX_ADD_SEPARATOR, {bar} )
end procedure

--/topic wxToolBar
--/proc toggle_tool( atom bar, atom id, integer toggle )
--
--Toggles a tool on or off. This does not cause any event to get emitted.
public procedure toggle_tool( atom bar, atom id, integer toggle )
    c_proc( WX_TOGGLE_TOOL, {bar, id, toggle} )
end procedure

--/topic wxToolBar
--/proc set_tool_bitmap( atom bar, atom id, atom b1, atom b2 )
--Sets the bitmap to be used by the tool with the given ID.
-- This can only be used on Button tools, not controls.
public procedure set_tool_bitmap( atom bar, atom id, atom b1, atom b2 )
    c_proc( WX_SET_TOOL_BITMAP, {bar,id,b1,b2})
end procedure

--/topic wxToolBar
--/func get_tool_state( atom bar, atom id )
--Gets the on/off state of a toggle tool.
public function get_tool_state( atom bar, atom id )
    return c_func( WX_GET_TOOL_STATE, {bar,id})
end function

--/topic wxToolBar
--/proc set_tool_label( atom bar, atom id, sequence label_val )
--Sets the label of a toolbar button
public procedure set_tool_label( atom bar, atom id, sequence label_val )
    c_proc( WX_SET_TOOL_LABEL, {bar,id,label_val})
end procedure

--/topic wxToolBar
--/proc set_tool_help( atom bar, atom id, sequence short_help, sequence long_help )
--Sets the help text of a toolbar button
public procedure set_tool_help( atom bar, atom id, sequence short_help, sequence long_help )
    c_proc( WX_SET_TOOL_HELP, {bar,id,short_help,long_help})
end procedure

--/topic wxToolBar
--/func get_tool_pos( atom bar, atom id )
--
public function get_tool_pos( atom bar, atom id )
    return c_func( WX_GET_TOOL_POS, {bar,id})
end function

--/topic wxToolBar
--/func get_tool_size( atom bar, atom id )
--
public function get_tool_size( atom bar )
    return c_func( WX_GET_TOOL_SIZE, {bar})
end function



--wxToolbook
-----------------------------------------
--/topic wxToolbook
--/parent Notebook Controls
--/parent Controls
--/info
--



--wxTreebook
-----------------------------------------
--/topic wxTreebook
--/parent Notebook Controls
--/parent Controls
--/info
--



--wxTreeCtrl
------------------------------
--/topic wxTreeCtrl
--/parent Controls
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li integer style [= wxTR_HAS_BUTTONS + wxTR_LINES_AT_ROOT ]
--<ul><table>
--<tr><td> /li wxTR_NO_BUTTONS              <tr> for convenience
--<tr><td> /li wxTR_HAS_BUTTONS             <tr> generates a +/- button
--<tr><td> /li wxTR_TWIST_BUTTONS           <tr> generates a twister button
--<tr><td> /li wxTR_NO_LINES                <tr> don't generate level connectors
--<tr><td> /li wxTR_LINES_AT_ROOT           <tr> connect top-level nodes
--<tr><td> /li wxTR_MAC_BUTTONS             <tr> backward compatibility
--<tr><td> /li wxTR_AQUA_BUTTONS            <tr> used internally
--<tr><td> /li wxTR_SINGLE                  <tr> for convenience
--<tr><td> /li wxTR_MULTIPLE                <tr> can select multiple items
--<tr><td> /li wxTR_EXTENDED                <tr> TODO: allow extended selection
--<tr><td> /li wxTR_FULL_ROW_HIGHLIGHT      <tr> highlight full horizontal space
--<tr><td> /li wxTR_EDIT_LABELS             <tr> can edit item labels
--<tr><td> /li wxTR_ROW_LINES               <tr> put border around items
--<tr><td> /li wxTR_HIDE_ROOT               <tr> don't display root node
--<tr><td> /li wxTR_HAS_VARIABLE_ROW_HEIGHT <tr> what it says
--</table></ul>
--</ul>
--
--When working with tree items, the id values returned are actually pointers to
--wxTreeItem pointers.  They are not unique ids.  The value returned is a new wxTreeItemId,
--which is a simple wxWidgets object.  It is just a pointer to either a windows handle or a custom
--data pointer.  It is the duty of the application to free the wxTreeItemId.

--/topic wxTreeCtrl
--/func add_tree_root( atom tree, sequence text, atom image, atom selimage )
--
--<ul>
--/li /i Tree : The wxTreeCtrl in which the item will be added.
--/li /i Text : The text to appear on the item.
--/li /i Image : The image that will show when one of it's children is clicked.
--/li /i SelImage : The image that will show when one of it's children is NOT currently chosen.
--</ul>
--You must declare tree_imagelist before you can assign any images to image
--and selimage. If you do not have a tree_imagelist declared, use -1 for image and selimage.
--
--If you declare tree_imagelist, -1 will become it's first image and all
--items will have an image beside them.
public function add_tree_root( atom tree, sequence text, atom image, atom selimage )
    return c_func( WX_ADD_TREE_ROOT, {tree, text, image, selimage} )
end function

--/topic wxTreeCtrl
--/func add_tree_item( atom tree, atom parent, sequence text, atom image, atom selimage )
--
--<ul>
--/li /i tree : The wxTreeCtrl in which the item will be added.
--/li /i parent : What the item will appear under. This can be an existing item, or the root.
--/li /i text : The text that will appear on the item.
--/li /i image : The image that will show when one of it's children is currently chosen.
--/li /i selImage : The image that will show when one of it's children is NOT currently chosen.
--</ul>
--You must declare tree_imagelist before you can assign any images to image and
--selimage. If you do not have one declared, use -1 for Image and SelImage
public function add_tree_item( atom tree, atom parent, sequence text, atom image, atom selimage )
    return c_func( WX_ADD_TREE_ITEM, {tree, parent, text, image, selimage} )
end function

--/topic wxTreeCtrl
--/proc delete_tree_item( atom tree, object item )
--Deletes one or all items from a /wxTreeCtrl.<br>
--If /i item is an empty sequence, then all items are deleted, otherwise, /i item can be either
--an integer or a sequence of item ids to be deleted.
public procedure delete_tree_item( atom tree, object item )
    c_proc( WX_DELETE_TREE_ITEM, {tree, item} )
end procedure

--/topic wxTreeCtrl
--/proc tree_imagelist( atom tree, atom il )
--
--This sets the wxImageList to the wxTreeCtrl, allowing the wxTreeCtrl's children to use it's icons.
public procedure tree_imagelist( atom tree, atom il )
    c_proc( WX_TREE_IMAGELIST, {tree, il} )
end procedure

--/topic wxTreeCtrl
--/func get_tree_selection( atom tree )
--
--This returns an integer which serves as a stand_in for the
--actual item. You can set things to the pointer as if it were
--the actual wxTreeCtrl item.
public function get_tree_selection( atom tree )
    return c_func( WX_GET_TREE_SELECTION, {tree} )
end function

--/topic wxTreeCtrl
--/func get_tree_text( atom tree, atom id )
--
--This goes to the specified tree and grabs the value of the
--item, specified by it's ID.
public function get_tree_text( atom tree, atom id )
    return c_func( WX_GET_TREE_TEXT, {tree, id} )
end function

--/topic wxTreeCtrl
--/func get_tree_selected_text( atom tree )
--
--This gets the text of whatever item is currently highlighted in the chosen wxTreeCtrl.
public function get_tree_selected_text( atom tree )
    return c_func( WX_GET_TREE_SELECTED_TEXT, {tree} )
end function

--/topic wxTreeCtrl
--/proc expand_tree_item( atom tree, atom item )
--
public procedure expand_tree_item( atom tree, atom item )
    c_proc( WX_EXPAND_TREE_ITEM, {tree, item} )
end procedure

--/topic wxTreeCtrl
--/proc set_tree_item_text( atom tree, atom item, sequence text )
--
--Goes to the specified item inside of the specified /wxTreeCtrl and sets the text in it to text.
public procedure set_tree_item_text( atom tree, atom item, sequence text )
    c_proc( WX_SET_TREE_ITEM_TEXT, {tree, item, text} )
end procedure

--/topic wxTreeCtrl
--/func get_tree_parent( atom tree, atom item )
--
--This returns an integer which serves as a stand_in for the item's
--actual parent. You can set things to the pointer as if it were the
--actual wxTreeCtrl item.
public function get_tree_parent( atom tree, atom item )
    return c_func( WX_GET_TREE_PARENT, {tree, item} )
end function

--/topic wxTreeCtrl
--/func get_tree_root( atom tree )
--
--This returns an integer which serves as a stand_in for the
--actual tree root. You can set things to the pointer as if it
--were the actual wxTreeCtrl item.
public function get_tree_root( atom tree )
    return c_func( WX_GET_TREE_ROOT, {tree} )
end function

--/topic wxTreeCtrl
--/func valid_tree_item( atom item )
--
--Returns true if this instance is referencing a valid tree item.
public function valid_tree_item( atom item )
    return c_func( WX_VALID_TREE_ITEM, {item} )
end function

--/topic wxTreeCtrl
--/proc select_tree_item( atom tree, atom treeitem )
--
--Sets the highlight to the specified tree item by using it's
--actual id or an integer pointer that goes to it.
public procedure select_tree_item( atom tree, atom treeitem )
    c_proc( WX_SELECT_TREE_ITEM, {tree, treeitem} )
end procedure

--/topic wxTreeCtrl
--/proc delete_tree_children( atom tree, atom treeitem )
--
--This goes to a TreeCtrl specified by tree and deletes all children
--under the specifed item.
public procedure delete_tree_children( atom tree, atom treeitem )
    c_proc( WX_DELETE_TREE_CHILDREN, {tree, treeitem} )
end procedure

--/topic wxTreeCtrl
--/proc collapse_tree_item( atom tree, atom item )
--
--Collapses the given item.
public procedure collapse_tree_item( atom tree, atom item )
    c_proc( WX_COLLAPSE_TREE_ITEM, {tree,item})
end procedure

--/topic wxTreeCtrl
--/proc collapse_all_tree_items( atom tree )
--
--Collapses the root item.
public procedure collapse_all_tree_items( atom tree )
    c_proc( WX_COLLAPSE_ALL_TREE_ITEMS, {tree})
end procedure

--/topic wxTreeCtrl
--/proc collapse_all_tree_children( atom tree, atom item )
--
--Collapses this item and all of its children, recursively.
public procedure collapse_all_tree_children( atom tree, atom item )
    c_proc( WX_COLLAPSE_ALL_TREE_CHILDREN, {tree,item})
end procedure

--/topic wxTreeCtrl
--/proc collapse_and_reset_tree_item( atom tree, atom item )
--
--Collapses the given item and removes all children.
public procedure collapse_and_reset_tree_item( atom tree, atom item )
    c_proc( WX_COLLAPSE_AND_RESET_TREE_ITEM, {tree,item})
end procedure

--/topic wxTreeCtrl
--/proc edit_tree_item( atom tree, atom item )
--
--Starts editing the label of the given item. This function generates a EVT_TREE_BEGIN_LABEL_EDIT event which can be vetoed so that no text
-- control will appear for in-place editing.
--
--If the user changed the label (i.e. s/he does not press ESC or leave the text control without changes, a EVT_TREE_END_LABEL_EDIT event will
-- be sent which can be vetoed as well.
public procedure edit_tree_item( atom tree, atom item )
    c_proc( WX_EDIT_TREE_ITEM, {tree,item})
end procedure

--/topic wxTreeCtrl
--/proc ensure_tree_item_visual( atom tree, atom item )
--
--Scrolls and/or expands items to ensure that the given item is visible.
public procedure ensure_tree_item_visual( atom tree, atom item )
    c_proc( WX_ENSURE_TREE_ITEM_VISUAL, {tree,item})
end procedure

--/topic wxTreeCtrl
--/proc expand_tree( atom tree )
--
--Expands all items in the tree.
public procedure expand_tree( atom tree )
    c_proc( WX_EXPAND_TREE, {tree})
end procedure

--/topic wxTreeCtrl
--/proc expand_all_tree_children( atom tree, atom item )
--
--Expands the given item and all its children recursively.
public procedure expand_all_tree_children( atom tree, atom item )
    c_proc( WX_EXPAND_ALL_TREE_CHILDREN, {tree,item})
end procedure

--/topic wxTreeCtrl
--/func get_tree_item_bounding_rect( atom tree, atom item, atom text_only )
--
--Retrieves the rectangle bounding the item. If /i text_only is true, only the rectangle around the item's label will be returned, otherwise the
-- item's image is also taken into account.
--
--The return value a sequence of { ok, rect }. /i ok is true if the rectangle was successfully retrieved or false if it was not (in this case rect
-- is not changed) for example, if the item is currently invisible (in which case /i rect will be all zeros).
--
--Notice that the rectangle coordinates are logical, not physical ones. So, for example, the x coordinate may be negative if the tree has a
-- horizontal scrollbar and its position is not 0.
public function get_tree_item_bounding_rect( atom tree, atom item, atom text_only )
    return c_func( WX_GET_TREE_ITEM_BOUNDING_RECT, {tree,item,text_only})
end function

--/topic wxTreeCtrl
--/func get_tree_item_children_count( atom tree, atom item, atom recursively )
--
--Returns the number of items in the branch. If /i recursively is true, returns the total number of descendants, otherwise only one level of
-- children is counted.
public function get_tree_item_children_count( atom tree, atom item, atom recursively )
    return c_func( WX_GET_TREE_ITEM_CHILDREN_COUNT, {tree,item,recursively})
end function

--/topic wxTreeCtrl
--/func get_tree_item_count( atom tree )
--
--Returns the number of items in the control.
public function get_tree_item_count( atom tree )
    return c_func( WX_GET_TREE_ITEM_COUNT, {tree})
end function

--/topic wxTreeCtrl
--/func get_tree_edit_control( atom tree )
--
--Returns the edit control being currently used to edit a label. Returns NULL if no label is being edited.
--
--/b Note: this is currently only implemented for wxMSW.
public function get_tree_edit_control( atom tree )
    return c_func( WX_GET_TREE_EDIT_CONTROL, {tree})
end function

--/topic wxTreeCtrl
--/func get_first_visible_tree_item( atom tree )
--
--Returns the first visible item.
public function get_first_visible_tree_item( atom tree )
    return c_func( WX_GET_FIRST_VISIBLE_TREE_ITEM, {tree})
end function

--/topic wxTreeCtrl
--/func get_tree_imagelist( atom tree )
--
--Returns the normal image list.
public function get_tree_imagelist( atom tree )
    return c_func( WX_GET_TREE_IMAGELIST, {tree})
end function

--/topic wxTreeCtrl
--/func get_tree_indent( atom tree )
--
--Returns the current tree control indentation.
public function get_tree_indent( atom tree )
    return c_func( WX_GET_TREE_INDENT, {tree})
end function

--/topic wxTreeCtrl
--/func get_tree_item_background_color( atom tree, atom item )
--
--Returns the background colour of the item.
public function get_tree_item_background_color( atom tree, atom item )
    return c_func( WX_GET_TREE_ITEM_BACKGROUND_COLOR, {tree,item})
end function

--/topic wxTreeCtrl
--/func get_tree_item_font( atom tree, atom item )
--
--Returns the font of the item label.
public function get_tree_item_font( atom tree, atom item )
    return c_func( WX_GET_TREE_ITEM_FONT, {tree,item})
end function

--/topic wxTreeCtrl
--/func get_tree_item_image( atom tree, atom item, atom which )
--
--Gets the specified item image. The value of which may be:
--<ul>
--/li /b wxTreeItemIcon_Normal to get the normal item image
--/li /b wxTreeItemIcon_Selected to get the selected item image (i.e. the image which is shown when the item is currently selected)
--/li /b wxTreeItemIcon_Expanded to get the expanded image (this only makes sense for items which have children - then this image is shown when the item is expanded and the normal image is shown when it is collapsed)
--/li /b wxTreeItemIcon_SelectedExpanded to get the selected expanded image (which is shown when an expanded item is currently selected)
--</ul>
public function get_tree_item_image( atom tree, atom item, atom which )
    return c_func( WX_GET_TREE_ITEM_IMAGE, {tree,item,which})
end function

--/topic wxTreeCtrl
--/func get_tree_item_text_color( atom tree, atom item )
--
--Returns the colour of the item label.
public function get_tree_item_text_color( atom tree, atom item )
    return c_func( WX_GET_TREE_ITEM_TEXT_COLOR, {tree,item})
end function

--/topic wxTreeCtrl
--/func get_first_tree_child( atom tree, atom item )
--
--Returns a sequence containing the first child and an opaque handle; call /get_next_tree_child for the next child.
--
-- The second element of the returned sequence is the 'cookie' parameter which is opaque for the application but is necessary for the
-- library to make these functions reentrant (i.e. allow more than one enumeration on one and the same object simultaneously). The
-- cookie passed to subsequent calls of /get_next_tree_child should be the same variable.
--
--Returns an invalid tree item (i.e. IsOk() returns false) if there are no further children.
public function get_first_tree_child( atom tree, atom item )
    return c_func( WX_GET_FIRST_TREE_CHILD, {tree,item})
end function

--/topic wxTreeCtrl
--/func get_next_tree_child( atom tree, atom item, atom cookie )
--
--Returns the next child; call /get_first_tree_child for the first child.
--
--For this enumeration function you must pass in a 'cookie' parameter which is opaque for the application but is necessary for the
-- library to make these functions reentrant (i.e. allow more than one enumeration on one and the same object simultaneously). The
-- cookie passed to /get_first_tree_child and /get_next_tree_child should be the same variable.
--
--Returns an invalid tree item if there are no further children.
public function get_next_tree_child( atom tree, atom item, atom cookie )
    return c_func( WX_GET_NEXT_TREE_CHILD, {tree,item,cookie})
end function

--/topic wxTreeCtrl
--/func get_next_tree_sibling( atom tree, atom item )
--
--Returns the next sibling of the specified item; call /get_prev_tree_sibling for the previous sibling.
--
--Returns an invalid tree item if there are no further siblings.
public function get_next_tree_sibling( atom tree, atom item )
    return c_func( WX_GET_NEXT_TREE_SIBLING, {tree,item})
end function

--/topic wxTreeCtrl
--/func get_prev_tree_sibling( atom tree, atom item )
--
--Returns the previous sibling of the specified item; call /get_next_tree_sibling for the next sibling.
--
--Returns an invalid tree item if there are no further children.
public function get_prev_tree_sibling( atom tree, atom item )
    return c_func( WX_GET_PREV_TREE_SIBLING, {tree,item})
end function

--/topic wxTreeCtrl
--/func get_next_visible_tree_item( atom tree, atom item )
--
--Returns the next visible item.
public function get_next_visible_tree_item( atom tree, atom item )
    return c_func( WX_GET_NEXT_VISIBLE_TREE_ITEM, {tree,item})
end function

--/topic wxTreeCtrl
--/func get_prev_visible_tree_item( atom tree, atom item )
--
--Returns the previous visible item.
public function get_prev_visible_tree_item( atom tree, atom item )
    return c_func( WX_GET_PREV_VISIBLE_TREE_ITEM, {tree,item})
end function

--/topic wxTreeCtrl
--/func get_tree_selections( atom tree )
--
--Returns a sequence containing the currently selected items. This function can be called only if the control has the wxTR_MULTIPLE style.
public function get_tree_selections( atom tree )
    return c_func( WX_GET_TREE_SELECTIONS, {tree})
end function

public constant
        wxTREE_HITTEST_ABOVE            = 0x0001,
        wxTREE_HITTEST_BELOW            = 0x0002,
        wxTREE_HITTEST_NOWHERE          = 0x0004,
        wxTREE_HITTEST_ONITEMBUTTON     = 0x0008,
        wxTREE_HITTEST_ONITEMICON       = 0x0010,
        wxTREE_HITTEST_ONITEMINDENT     = 0x0020,
        wxTREE_HITTEST_ONITEMLABEL      = 0x0040,
        wxTREE_HITTEST_ONITEMRIGHT      = 0x0080,
        wxTREE_HITTEST_ONITEMSTATEICON  = 0x0100,
        wxTREE_HITTEST_TOLEFT           = 0x0200,
        wxTREE_HITTEST_TORIGHT          = 0x0400,
        wxTREE_HITTEST_ONITEMUPPERPART  = 0x0800,
        wxTREE_HITTEST_ONITEMLOWERPART  = 0x1000,
        wxTREE_HITTEST_ONITEM  = or_bits( wxTREE_HITTEST_ONITEMICON, wxTREE_HITTEST_ONITEMLABEL ),
        $
--/topic wxTreeCtrl
--/func tree_hit_test( atom tree, atom x, atom y, atom flags )
--
--Calculates which (if any) item is under the given point, returning the tree item id at this point plus extra information flags.
--
--/i flags is a bitlist of the following:
--<ul>
--/li /i wxTREE_HITTEST_ABOVE Above the client area.
--/li /i wxTREE_HITTEST_BELOW Below the client area.
--/li /i wxTREE_HITTEST_NOWHERE In the client area but below the last item.
--/li /i wxTREE_HITTEST_ONITEMBUTTON On the button associated with an item.
--/li /i wxTREE_HITTEST_ONITEMICON On the bitmap associated with an item.
--/li /i wxTREE_HITTEST_ONITEMINDENT In the indentation associated with an item.
--/li /i wxTREE_HITTEST_ONITEMLABEL On the label (string) associated with an item.
--/li /i wxTREE_HITTEST_ONITEMRIGHT In the area to the right of an item.
--/li /i wxTREE_HITTEST_ONITEMSTATEICON On the state icon for a tree view item that is in a user-defined state.
--/li /i wxTREE_HITTEST_TOLEFT To the right of the client area.
--/li /i wxTREE_HITTEST_TORIGHT To the left of the client area.
--/li /i wxTREE_HITTEST_ONITEM or_bits( wxTREE_HITTEST_ONITEMICON, wxTREE_HITTEST_ONITEMLABEL )
--</ul>
public function tree_hit_test( atom tree, atom x, atom y, atom flags )
    return c_func( WX_TREE_HIT_TEST, {tree,x,y,flags})
end function

--/topic wxTreeCtrl
--/func insert_tree_item_after( atom tree, atom parent, atom previous, sequence text, atom image, atom sel_image )
--
--Inserts an item after a given one ( /i previous ).
--
--The /i image and /i sel_image parameters are an index within the normal image list specifying the image to use for unselected and selected items,
-- respectively. If /i image > -1 and /i selimage is -1, the same image is used for both selected and unselected items.
public function insert_tree_item_after( atom tree, atom parent, atom previous, sequence text, atom image, atom sel_image )
    return c_func( WX_INSERT_TREE_ITEM_AFTER, {tree,parent,previous,text,image,sel_image})
end function

--/topic wxTreeCtrl
--/func insert_tree_item_before( atom tree, atom parent, atom before, sequence text, atom image, atom sel_image )
--
--Inserts an item after a given one ( /i before ).
--
--The /i image and /i sel_image parameters are an index within the normal image list specifying the image to use for unselected and selected items,
-- respectively. If /i image > -1 and /i selimage is -1, the same image is used for both selected and unselected items.
public function insert_tree_item_before( atom tree, atom parent, atom before, sequence text, atom image, atom sel_image )
    return c_func( WX_INSERT_TREE_ITEM_BEFORE, {tree,parent,before,text,image,sel_image})
end function

--/topic wxTreeCtrl
--/func is_tree_item_bold( atom tree, atom item )
--
--Returns true if the given item is in bold state.
public function is_tree_item_bold( atom tree, atom item )
    return c_func( WX_IS_TREE_ITEM_BOLD, {tree,item})
end function

--/topic wxTreeCtrl
--/func is_tree_empty( atom tree )
--
--Returns true if the control is empty (i.e. has no items, even no root one).
public function is_tree_empty( atom tree )
    return c_func( WX_IS_TREE_EMPTY, {tree})
end function

--/topic wxTreeCtrl
--/func is_tree_item_expanded( atom tree, atom item )
--
--Returns true if the item is expanded (only makes sense if it has children).
public function is_tree_item_expanded( atom tree, atom item )
    return c_func( WX_IS_TREE_ITEM_EXPANDED, {tree,item})
end function

--/topic wxTreeCtrl
--/func is_tree_item_selected( atom tree, atom item )
--
--Returns true if the item is selected.
public function is_tree_item_selected( atom tree, atom item )
    return c_func( WX_IS_TREE_ITEM_SELECTED, {tree,item})
end function

--/topic wxTreeCtrl
--/func is_tree_item_visible( atom tree, atom item )
--
--Returns true if the item is visible (it might be outside the view, or not expanded).
public function is_tree_item_visible( atom tree, atom item )
    return c_func( WX_IS_TREE_ITEM_VISIBLE, {tree,item})
end function

--/topic wxTreeCtrl
--/func tree_item_has_children( atom tree, atom item )
--
--Returns true if the item has children.
public function tree_item_has_children( atom tree, atom item )
    return c_func( WX_TREE_ITEM_HAS_CHILDREN, {tree,item})
end function

--/topic wxTreeCtrl
--/func prepend_tree_item( atom tree, atom parent, sequence text, atom image, atom sel_image )
--
--Appends an item as the first child of parent, return a new item id.
--
--The /i image and /i sel_image parameters are an index within the normal image list specifying the image to use for unselected and selected items,
-- respectively. If /i image > -1 and /i selimage is -1, the same image is used for both selected and unselected items.
public function prepend_tree_item( atom tree, atom parent, sequence text, atom image, atom sel_image )
    return c_func( WX_PREPEND_TREE_ITEM, {tree,parent,text,image,sel_image})
end function

--/topic wxTreeCtrl
--/proc scroll_to_tree_item( atom tree, atom item )
--
--Scrolls the specified item into view.
public procedure scroll_to_tree_item( atom tree, atom item )
    c_proc( WX_SCROLL_TO_TREE_ITEM, {tree,item})
end procedure

--/topic wxTreeCtrl
--/proc set_tree_indent( atom tree, atom indent )
--
--Sets the indentation for the tree control.
public procedure set_tree_indent( atom tree, atom indent )
    c_proc( WX_SET_TREE_INDENT, {tree,indent})
end procedure

--/topic wxTreeCtrl
--/proc set_tree_item_background( atom tree, atom item, atom color )
--
--Sets the colour of the item's background.
public procedure set_tree_item_background( atom tree, atom item, atom color )
    c_proc( WX_SET_TREE_ITEM_BACKGROUND, {tree,item,color})
end procedure

--/topic wxTreeCtrl
--/proc set_tree_item_bold( atom tree, atom item, atom bold )
--
--Makes item appear in bold font if bold parameter is true or resets it to the normal state.
public procedure set_tree_item_bold( atom tree, atom item, atom bold )
    c_proc( WX_SET_TREE_ITEM_BOLD, {tree,item,bold})
end procedure

--/topic wxTreeCtrl
--/proc set_tree_item_drop_highlight( atom tree, atom item, atom highlight )
--
--Gives the item the visual feedback for Drag'n'Drop actions, which is useful if something is dragged from the outside onto the tree control
-- (as opposed to a DnD operation within the tree control, which already is implemented internally).
public procedure set_tree_item_drop_highlight( atom tree, atom item, atom highlight )
    c_proc( WX_SET_TREE_ITEM_DROP_HIGHLIGHT, {tree,item,highlight})
end procedure

--/topic wxTreeCtrl
--/proc set_tree_item_font( atom tree, atom item, atom font )
--
--Sets the item's font. All items in the tree should have the same height to avoid text clipping, so the fonts height should be the same for
-- all of them, although font attributes may vary.
public procedure set_tree_item_font( atom tree, atom item, atom font )
    c_proc( WX_SET_TREE_ITEM_FONT, {tree,item,font})
end procedure

--/topic wxTreeCtrl
--/proc set_tree_item_has_children( atom tree, atom item, atom has_children )
--
--Force appearance of the button next to the item. This is useful to allow the user to expand the items which don't have any children now,
-- but instead adding them only when needed, thus minimizing memory usage and loading time.
public procedure set_tree_item_has_children( atom tree, atom item, atom has_children )
    c_proc( WX_SET_TREE_ITEM_HAS_CHILDREN, {tree,item,has_children})
end procedure

--/topic wxTreeCtrl
--/proc set_tree_item_image( atom tree, atom item, atom image, atom which )
--
--Sets the specified item image. See /get_tree_item_image for the description of the which parameter.
public procedure set_tree_item_image( atom tree, atom item, atom image, atom which )
    c_proc( WX_SET_TREE_ITEM_IMAGE, {tree,item,image,which})
end procedure

--/topic wxTreeCtrl
--/proc set_tree_item_text_color( atom tree, atom item, atom color )
--
--Sets the colour of the item's text.
public procedure set_tree_item_text_color( atom tree, atom item, atom color )
    c_proc( WX_SET_TREE_ITEM_TEXT_COLOR, {tree,item,color})
end procedure

--/topic wxTreeCtrl
--/proc sort_tree_children( atom tree, atom item )
--
--Sorts the children of the given item.
public procedure sort_tree_children( atom tree, atom item )
    c_proc( WX_SORT_TREE_CHILDREN, {tree,item})
end procedure

--/topic wxTreeCtrl
--/proc toggle_tree_item( atom tree, atom item )
--
--Toggles the given item between collapsed and expanded states.
public procedure toggle_tree_item( atom tree, atom item )
    c_proc( WX_TOGGLE_TREE_ITEM, {tree,item})
end procedure

--/topic wxTreeCtrl
--/proc toggle_tree_item_selection( atom tree, atom item )
--
--Toggles the given item between selected and unselected states. For multiselection controls only.
public procedure toggle_tree_item_selection( atom tree, atom item )
    c_proc( WX_TOGGLE_TREE_ITEM_SELECTION, {tree,item})
end procedure

--/topic wxTreeCtrl
--/proc unselect_tree_item( atom tree, object items )
--
--Removes the selection from the currently selected item (if any).
public procedure unselect_tree_item( atom tree, object items )
    c_proc( WX_UNSELECT_TREE_ITEM, {tree,items})
end procedure



--wxTreeEvent
------------------------------
--/topic wxTreeEvent
--/parent wxTreeCtrl
--/parent Events
--/info
--
--This is the class of the event that is passed to an event handler for a wxTree type event.

--/topic wxTreeEvent
--/func get_tree_event_item( atom event )
--
--Returns the item (valid for all events).
public function get_tree_event_item( atom event )
    return c_func( WX_GET_TREE_EVENT_ITEM, {event} )
end function



--wxUpdateUIEvent
------------------------------------
--/topic wxUpdateUIEvent
--/parent Events

--/topic wxUpdateUIEvent
--/func can_update( atom window )
--
--Returns true if it is appropriate to update (send UI update events to) this window.
--
--This function looks at the mode used, the wxWS_EX_PROCESS_UPDATE_EVENTS flag in window, the time update events were last sent in idle time,
-- and the update interval, to determine whether events should be sent to this window now. By default this will always return true because the
-- update mode is initially wxUPDATE_UI_PROCESS_ALL and the interval is set to 0; so update events will be sent as often as possible. You can
-- reduce the frequency that events are sent by changing the mode and/or setting an update interval.
public function can_update( atom window )
    return c_func( WX_CAN_UPDATE, {window})
end function

--/topic wxUpdateUIEvent
--/proc check_ui( atom update_event, atom check )
--
--Check or uncheck the UI element.
public procedure check_ui( atom update_event, atom check )
    c_proc( WX_CHECK_UI, {update_event,check})
end procedure

--/topic wxUpdateUIEvent
--/proc enable_ui( atom update_event, atom enable )
--
--Enable or disable the UI element.
public procedure enable_ui( atom update_event, atom enable )
    c_proc( WX_ENABLE_UI, {update_event,enable})
end procedure

--/topic wxUpdateUIEvent
--/proc show_ui( atom update_event, atom show )
--
--Show or hide the UI element.
public procedure show_ui( atom update_event, atom show )
    c_proc( WX_SHOW_UI, {update_event,show})
end procedure

--/topic wxUpdateUIEvent
--/func get_checked_ui( atom update_event )
--
--Returns true if the UI element should be checked.
public function get_checked_ui( atom update_event )
    return c_func( WX_GET_CHECKED_UI, {update_event})
end function

--/topic wxUpdateUIEvent
--/func get_enabled_ui( atom update_event )
--
--Returns true if the UI element should be enabled.
public function get_enabled_ui( atom update_event )
    return c_func( WX_GET_ENABLED_UI, {update_event})
end function

--/topic wxUpdateUIEvent
--/func get_shown_ui( atom update_event )
--
--Returns true if the UI element should be shown.
public function get_shown_ui( atom update_event )
    return c_func( WX_GET_SHOWN_UI, {update_event})
end function

--/topic wxUpdateUIEvent
--/func get_text_ui( atom update_event )
--
public function get_text_ui( atom update_event )
    return c_func( WX_GET_TEXT_UI, {update_event})
end function

--/topic wxUpdateUIEvent
--/func get_mode_ui( atom update_event )
--
--Static function returning a value specifying how wxWidgets will send update events: to all windows, or only to those which specify
-- that they will process the events.
public function get_mode_ui( atom update_event )
    return c_func( WX_GET_MODE_UI, {update_event})
end function

--/topic wxUpdateUIEvent
--/func get_update_interval_ui(  )
--
--Returns the current interval between updates in milliseconds. -1 disables updates, 0 updates as frequently as possible.
public function get_update_interval_ui(  )
    return c_func( WX_GET_UPDATE_INTERVAL_UI, {})
end function

--/topic wxUpdateUIEvent
--/proc set_ui_update_mode( atom mode )
--
--Specify how wxWidgets will send update events: to all windows, or only to those which specify that they will process the events.
--
--/i mode may be one of the following values. The default is wxUPDATE_UI_PROCESS_ALL.
--<ul>
--/li /i wxUPDATE_UI_PROCESS_ALL Send UI update events to all windows
--/li /i wxUPDATE_UI_PROCESS_SPECIFIED Send UI update events to windows that have the wxWS_EX_PROCESS_UI_UPDATES flag specified
--</ul>
public procedure set_ui_update_mode( atom mode )
    c_proc( WX_SET_UI_UPDATE_MODE, {mode})
end procedure

--/topic wxUpdateUIEvent
--/proc set_text_ui( atom update_event, sequence text )
--
--Sets the text for this UI element.
public procedure set_text_ui( atom update_event, sequence text )
    c_proc( WX_SET_TEXT_UI, {update_event,text})
end procedure

--/topic wxUpdateUIEvent
--/proc set_ui_update_interval( atom updateInterval )
--
--Sets the interval between updates in milliseconds. Set to -1 to disable updates, or to 0 to update as frequently as possible. The default is 0.
--
--Use this to reduce the overhead of UI update events if your application has a lot of windows. If you set the value to -1 or greater than 0,
-- you may also need to call wxWindow::UpdateWindowUI at appropriate points in your application, such as when a dialog is about to be shown.
public procedure set_ui_update_interval( atom updateInterval )
    c_proc( WX_SET_UI_UPDATE_INTERVAL, {updateInterval})
end procedure



--wxURL
------------------------------
--/topic wxURL
--/parent Networking
--/parent Classes
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "sequence url"
--</ul>

--/topic wxURL
--/func get_url_input( atom url )
--
--Returns an input stream that can be used with /stream_read() to fetch
--the file from the url.
public function get_url_input( atom url )
    return c_func( WX_GET_URL_INPUT, {url} )
end function



--wxWindow
------------------------------
--/topic wxWindow
--/parent Controls
--/info
--/b wxWindow is a very well used class.  Most controls are derived from /b wxWindow, so you should
--remember that anything that can be done to a window can be done to an object derived from wxWindow.
--The parameters for creating a /b wxWindow are:
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li integer x [= -1]
--/li integer y [= -1]
--/li integer cx [= -1]
--/li integer cy [= -1]
--/li atom style [= 0] (see /wxFrame)
--</ul>
--
--An object of class /b wxWindow /b must have a /i parent.  Therefore, you cannot create an object of
-- class /b wxWindow to be your top level window. Your top level window should be either of type
-- /wxFrame or /wxDialog.  In fact, you should probably rarely create an object using /b wxWindow.
-- See /wxPanel.

--/topic wxWindow
--/func get_window_scroll_position( atom window, atom orient )
--
--Returns the built-in scrollbar position.
--
--/i orient is either wxVERTICAL or wxHORIZONTAL
public function get_window_scroll_position( atom window, atom orient )
    return c_func( WX_GET_WINDOW_SCROLL_POSITION, {window,orient})
end function

--/topic wxWindow
--/proc set_back_color( atom window, atom color )
--
--<ul>
--/li /b window: the window whose color is to be set
--/li /b color: a /wxColour object
--</ul>
public procedure set_back_color( atom window, atom color )
    c_proc( WX_SET_BACK_COLOR, {window, color} )
end procedure

--/topic wxWindow
--/proc set_fore_color( atom window, atom color )
--
--<ul>
--/li /b window: the window whose color is to be set
--/li /b color: a /wxColour object
--</ul>
public procedure set_fore_color( atom window, atom color )
    c_proc( WX_SET_FORE_COLOR, {window, color} )
end procedure


--/topic wxWindow
--/func get_id( atom this )
--
--Returns the id for objects that are derived from class wxWindow.  This includes
--frames, windows and most controls, but does not include menu items.  You should
--use /get_menuitem_id() for that.
public function get_id( atom this )
    return c_func( WX_GET_ID, {this} )
end function

--/topic wxWindow
--/proc release_mouse( atom window )
--
--This releases the mouse position after capture_mouse has been called.
public procedure release_mouse( atom window )
    c_proc( WX_RELEASE_MOUSE, {window} )
end procedure

--/topic wxWindow
--/proc capture_mouse( atom window )
--
--This forces a mouse to keep the position it had when it left a window.
--If you call this while the cursor is in the 'double-arrow stance, it
--will keep that. However, if it has the normal arrow position when it is
--called, it will make the cursor turn into a wait stance while it is over
--the specified window.
public procedure capture_mouse( atom window )
    c_proc( WX_CAPTURE_MOUSE, {window} )
end procedure

--/topic wxWindow
--/proc set_title( atom window, sequence title )
--
--This resets the title of a wxFrame and allows you to put whatever
--you want in it's place. You can use /set_icon to change the icon
--beside of it.
public procedure set_title( atom window, sequence title )
    c_proc( WX_SET_TITLE, {window, title} )
end procedure

--/topic wxWindow
--/proc refresh_window( object params )
--
--Forces a wxWindow to repaint itself.  params can take one of several forms:
--<ul>
--/li /b atom: /i this pointer to wxWindow to be refreshed.  /i erase_background is set to 1
--/li /b sequence:
--<ul>
--/li { /i this, /i erase_background}
--/li { /i this, /i erase_background, /i x1, /i y1, /i x2, /i y2}
--/li If /i erase_background is 0, then the background is not erased before repainting.
--/li /i "{ x1, y1, x2, y2}" defines a rectangle to repaint.  If no rectangle is specified,
--then the entire wxWindow will be refreshed.
--</ul></ul>
public procedure refresh_window( object params )
    c_proc( WX_REFRESH_WINDOW, {params} )
end procedure

--/topic wxWindow
--/func get_client_size( atom this )
--
--This returns the size of this in a sequence of {x,y}. This can
--be any wx class that is graphical ( i.e., not /wxFontData or
--/wxArrayString. )
public function get_client_size( atom this )
    return c_func( WX_GET_CLIENT_SIZE, {this} )
end function

--/topic wxWindow
--/func get_top_level_parent( atom window )
--
--Returns the first top level parent of the given window, or in other words, the frame or dialog containing it, or NULL.
public function get_top_level_parent( atom window )
    return c_func( WX_GET_TOP_LEVEL_PARENT, {window})
end function

--/topic wxWindow
--/proc show_popup_menu( atom win, atom menu, integer x, integer y )
--
--Pops up the given menu at the specified coordinates, relative to this window, and returns control when the
-- user has dismissed the menu. If a menu item is selected, the corresponding menu event is generated and will
-- be processed as usually. If the coordinates are not specified, the current mouse cursor position is used.
public procedure show_popup_menu( atom win, atom menu, integer x, integer y )
    c_proc( WX_POPUP_MENU, {win, menu, x, y } )
end procedure

--/topic wxWindow
--/proc set_window_style( atom win, atom style )
--
--Set the style for the specified window.
public procedure set_window_style( atom win, atom style )
    c_proc( WX_SET_WINDOW_STYLE, {win, style})
end procedure

--/topic wxWindow
--/proc center_on_parent( atom win, integer direction )
--
--Centers the window on its parent.  direction can be wxHORIZONTAL, wxVERTICAL or wxBOTH.
public procedure center_on_parent( atom win, integer direction )
    c_proc( WX_CENTER_ON_PARENT, {win, direction})
end procedure

--/topic wxWindow
--/func get_window_style( atom win )
--
--Returns the style flag for the specified window.
public function get_window_style( atom win )
    return c_func( WX_GET_WINDOW_STYLE, {win} )
end function

--/topic wxWindow
--/proc close_window( atom win, atom force )
--
--Calls the close handler for the window, providing an opportunity for the window to
--choose whether to destroy the window. Usually it is only used with the top level
--windows (wxFrame and wxDialog classes) as the others are not supposed to have any
--special OnClose() logic.
--The close handler should check whether the window is being deleted forcibly, using
-- /can_veto, in which case it should /destroy the window.
--
--Note that calling Close does not guarantee that the window will be destroyed; but it
--provides a way to simulate a manual close of a window, which may or may not be
--implemented by destroying the window. The default implementation of
-- wxDialog::OnCloseWindow does not necessarily delete the dialog, since it will
-- simply simulate an wxID_CANCEL event which is handled by the appropriate button
-- event handler and may do anything at all.
--
--To guarantee that the window will be destroyed, call /destroy instead
public procedure close_window( atom win, atom force )
    c_proc( WX_CLOSE_WINDOW, {win,force})
end procedure

--/topic wxWindow
--/func get_window_sizer( atom wnd )
--Return the sizer associated with the window by a previous call to /set_sizer() or /b NULL.
public function get_window_sizer( atom wnd )
    return c_func( WX_GET_WINDOW_SIZER, {wnd})
end function



--wxWizard
------------------------------
--/topic wxWizard
--/parent Controls
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li integer id [= -1]
--/li sequence title [= ""]
--/li atom /wxBitmap [= 0]
--/li integer x [= -1]
--/li integer y [= -1]
--</ul>

--/topic wxWizard
--/func wizard_event_page( atom event )
--
--Returns the pointer to the /wxWizardPage that the event refers to.
public function wizard_event_page( atom event )
    return c_func( WX_WIZARD_EVENT_PAGE, {event} )
end function

--/topic wxWizard
--/func wizard_event_direction( atom event )
--
--Returns the direction of the page change.  For a forward change,
--the return value is 1, and the value is 0 for a backward change.
public function wizard_event_direction( atom event )
    return c_func( WX_WIZARD_EVENT_DIRECTION, {event} )
end function

--/topic wxWizard
--/func run_wizard( atom wizard, atom first_page )
--
--Executes the wizard starting from the given page, returning true if it was successfully finished
-- or false if user cancelled it. The firstPage can not be NULL.
public function run_wizard( atom wizard, atom first_page )
    return c_func( WX_RUN_WIZARD, {wizard, first_page} )
end function



--wxWizardPage
------------------------------
--/topic wxWizardPage
--/parent wxWizard
--/parent Controls
--/info
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li atom /wxBitmap [= 0]
--/li atom resource [= 0]
--</ul>
--
--A wxWizardPage is actually a panel that will be displayed in the /wxWizard.  You must call
-- /set_next_wizard_page() and /set_prev_wizard_page() for each page to set the next and previous
-- pages.  The first and last pages will have previous and next pages of 0, respectively.

--/topic wxWizardPage
--/proc set_next_wizard_page( atom page, atom next )
--
--This makes the next button on page go to the page specified by
--preview.
--/code
--  -- Ex.
--  -- P1 = First page
--  -- P2 = Second page
--
--  set_next_wizard page ( p1, p2 )
--/endcode
--would make it so that clicking the next button on p1 would make the
--wizard show p2.
public procedure set_next_wizard_page( atom page, atom next )
    c_proc( WX_SET_NEXT_WIZARD_PAGE, {page, next} )
end procedure

--/topic wxWizardPage
--/proc set_prev_wizard_page( atom page, atom prev )
--
--This makes the prev button on page go to the page specified by preview.
--/code
--  -- Ex.
--  -- P1 = First page
--  -- P2 = Second page
--
--  set_prev_wizard page ( p2, p1 )
--/endcode
--would make it so that clicking the prev button on p2 would make
--the wizard show p1.
public procedure set_prev_wizard_page( atom page, atom prev )
    c_proc( WX_SET_PREV_WIZARD_PAGE, {page, prev} )
end procedure



--wxXmlDocument
-----------------------------------------
--/topic wxXmlDocument
--/parent Files
--/parent Non wxObjects
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li sequence filename [= ""]
--/li sequence encoding [= "UTF-8"]
--</ul>

--/topic wxXmlDocument
--/func detatch_xml_root( atom xmldoc )
--
--Detaches the document root node and returns it. The document root node will be set to NULL and thus /xml_is_ok will return false after calling
-- this function.
--
--Note that the caller is reponsible for deleting the returned node in order to avoid memory leaks.
public function detatch_xml_root( atom xmldoc )
    return c_func( WX_DETATCH_XML_ROOT, {xmldoc})
end function

--/topic wxXmlDocument
--/func get_xml_encoding( atom xmldoc )
--
--Returns encoding of in-memory representation of the document (same as passed to Load or constructor, defaults to UTF-8).
--
--/b Note: this is meaningless in Unicode build where data are stored as /i wchar_t*.
public function get_xml_encoding( atom xmldoc )
    return c_func( WX_GET_XML_ENCODING, {xmldoc})
end function

--/topic wxXmlDocument
--/func get_xml_file_encoding( atom xmldoc )
--
--Returns encoding of document (may be empty).
--
--/b Note: this is the encoding original file was saved in, /b not the encoding of in-memory representation!
public function get_xml_file_encoding( atom xmldoc )
    return c_func( WX_GET_XML_FILE_ENCODING, {xmldoc})
end function

--/topic wxXmlDocument
--/func get_xml_root( atom xmldoc )
--
--Returns the root node of the document.
public function get_xml_root( atom xmldoc )
    return c_func( WX_GET_XML_ROOT, {xmldoc})
end function

--/topic wxXmlDocument
--/func get_xml_version( atom xmldoc )
--
--Returns the version of document. This is the value in the &lt;?xml version="1.0"?&gt; header of the XML document. If the version property
-- was not explicitely given in the header, this function returns an empty string.
public function get_xml_version( atom xmldoc )
    return c_func( WX_GET_XML_VERSION, {xmldoc})
end function

--/topic wxXmlDocument
--/func xml_is_ok( atom xmldoc )
--
--Returns true if the document has been loaded successfully.
public function xml_is_ok( atom xmldoc )
    return c_func( WX_XML_IS_OK, {xmldoc})
end function

--/topic wxXmlDocument
--/func load_xml( atom xmldoc, sequence filename, object options )
--
--Parses filename as an xml document and loads its data.
--
--If /i options does not contain wxXMLDOC_KEEP_WHITESPACE_NODES, then, while loading, all nodes of type wxXML_TEXT_NODE (see /wxXmlNode) are
-- automatically skipped if they contain whitespaces only. The removal of these nodes makes the load process slightly faster and requires less
-- memory however makes impossible to recreate exactly the loaded text with a Save call later. Read the initial description of this class for
-- more info.
--
--Returns true on success, false otherwise.
public function load_xml( atom xmldoc, sequence filename, object options )
    return c_func( WX_LOAD_XML, {xmldoc,filename,options})
end function

--/topic wxXmlDocument
--/func save_xml( atom xmldoc, sequence filename, atom indentstep )
--
--Saves XML tree creating a file named with given string.
--
--If /i indentstep is greater than or equal to zero, then, while saving, an automatic indentation is added with steps composed by indentstep spaces.
-- If /i indentstep is wxXML_NO_INDENTATION, then, automatic indentation is turned off.
public function save_xml( atom xmldoc, sequence filename, atom indentstep )
    return c_func( WX_SAVE_XML, {xmldoc,filename,indentstep})
end function

--/topic wxXmlDocument
--/proc set_xml_encoding( atom xmldoc, sequence enc )
--
--Sets the enconding of the document.
public procedure set_xml_encoding( atom xmldoc, sequence enc )
    c_proc( WX_SET_XML_ENCODING, {xmldoc,enc})
end procedure

--/topic wxXmlDocument
--/proc set_xml_file_encoding( atom xmldoc, sequence enc )
--
--Sets the enconding of the file which will be used to save the document.
public procedure set_xml_file_encoding( atom xmldoc, sequence enc )
    c_proc( WX_SET_XML_FILE_ENCODING, {xmldoc,enc})
end procedure

--/topic wxXmlDocument
--/proc set_xml_root( atom xmldoc, atom node )
--
--Sets the root node of this document. Deletes previous root node. Use DetachRoot and then SetRoot if you want to replace the root node without
-- deleting the old document tree.
public procedure set_xml_root( atom xmldoc, atom node )
    c_proc( WX_SET_XML_ROOT, {xmldoc,node})
end procedure

--/topic wxXmlDocument
--/proc set_xml_version( atom xmldoc, object version )
--
--Sets the version of the XML file which will be used to save the document.
public procedure set_xml_version( atom xmldoc, object version )
    c_proc( WX_SET_XML_VERSION, {xmldoc,version})
end procedure



--wxXmlNode
-----------------------------------------
--/topic wxXmlNode
--/parent wxXmlDocument
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "atom parent"
--/li /b "atom node_type"
--<ul>
--/li wxXML_ELEMENT_NODE
--/li wxXML_ATTRIBUTE_NODE
--/li wxXML_TEXT_NODE
--/li wxXML_CDATA_SECTION_NODE
--/li wxXML_ENTITY_REF_NODE
--/li wxXML_ENTITY_NODE
--/li wxXML_PI_NODE
--/li wxXML_COMMENT_NODE
--/li wxXML_DOCUMENT_NODE
--/li wxXML_DOCUMENT_TYPE_NODE
--/li wxXML_DOCUMENT_FRAG_NODE
--/li wxXML_NOTATION_NODE
--/li wxXML_HTML_DOCUMENT_NODE
--</ul>
--/li /b "sequence name"
--/li sequence content [= ""]
--/li integer properties [= 0]
--/li integer next [= 0]
--</ul>
--
--/i parent can be NULL

--/topic wxXmlNode
--/proc delete_wxXmlNode( object obj )
--
--Deletes a wxXmlNode object (use this instead of /delete_instance or /destroy)
public procedure delete_wxXmlNode( object obj )
    c_proc( WX_DELETE_WXXMLNODE, {obj})
end procedure

--/topic wxXmlNode
--/proc add_xml_node_child( atom node, atom child )
--
--Adds the given node as child of this node. To attach a second children to this node, use the /set_next_xml_node function of the child node.
public procedure add_xml_node_child( atom node, atom child )
    c_proc( WX_ADD_XML_NODE_CHILD, {node,child})
end procedure

--/topic wxXmlNode
--/proc add_xml_node_property( atom node, sequence name, sequence val )
--
--Appends a property with given name and value to the list of properties for this node.
public procedure add_xml_node_property( atom node, object name, object val )
    c_proc( WX_ADD_XML_NODE_PROPERTY, {node,name,val})
end procedure

--/topic wxXmlNode
--/func delete_xml_node_property( atom node, sequence name )
--
--Removes the first properties which has the given name from the list of properties for this node.
public function delete_xml_node_property( atom node, sequence name )
    return c_func( WX_DELETE_XML_NODE_PROPERTY, {node,name})
end function

--/topic wxXmlNode
--/func get_xml_node_children( atom node )
--
--Returns the first child of this node. To get a pointer to the second child of this node (if it does exist), use the /get_next_xml_node function
-- on the returned value.
public function get_xml_node_children( atom node )
    return c_func( WX_GET_XML_NODE_CHILDREN, {node})
end function

--/topic wxXmlNode
--/func get_xml_content( atom node )
--
--Returns the content of this node. Can be an empty string. Be aware that for nodes of type wxXML_ELEMENT_NODE (the most used node type) the content
-- is an empty string. See /get_xml_node_content for more details.
public function get_xml_content( atom node )
    return c_func( WX_GET_XML_CONTENT, {node})
end function

--/topic wxXmlNode
--/func get_xml_node_depth( atom node, atom grandparent )
--
--Returns the number of nodes which separe this node from grandparent.
--
--This function searches only the parents of this node until it finds grandparent or the NULL node (which is the parent of non-linked nodes or the
-- parent of a wxXmlDocument's root node).
public function get_xml_node_depth( atom node, atom grandparent )
    return c_func( WX_GET_XML_NODE_DEPTH, {node,grandparent})
end function

--/topic wxXmlNode
--/func get_xml_node_content( atom node )
--
--Returns the content of the first child node of type wxXML_TEXT_NODE or wxXML_CDATA_SECTION_NODE. This function is very useful since the XML
-- snippet "&lt;tagname&gt;tagcontent&lt;/tagname&gt;" is represented by expat with the following tag tree:
--/code
--wxXML_ENTITY_NODE name="tagname", content=""
--|-- wxXML_TEXT_NODE name="", content="tagcontent"
--/endcode
--or eventually:
--/code
--wxXML_ENTITY_NODE name="tagname", content=""
--|-- wxXML_CDATA_SECTION_NODE name="", content="tagcontent"
--/endcode
--An empty string is returned if the node has no children of type wxXML_TEXT_NODE or wxXML_CDATA_SECTION_NODE, or if the content of the first child
-- of such types is empty.
public function get_xml_node_content( atom node )
    return c_func( WX_GET_XML_NODE_CONTENT, {node})
end function

--/topic wxXmlNode
--/func get_xml_node_name( atom node )
--
--Returns the name of this node. Can be an empty string (e.g. for nodes of type wxXML_TEXT_NODE or wxXML_CDATA_SECTION_NODE).
public function get_xml_node_name( atom node )
    return c_func( WX_GET_XML_NODE_NAME, {node})
end function

--/topic wxXmlNode
--/func get_next_xml_node( atom node )
--
--Returns a pointer to the sibling of this node or NULL if there are no siblings.
public function get_next_xml_node( atom node )
    return c_func( WX_GET_NEXT_XML_NODE, {node})
end function

--/topic wxXmlNode
--/func get_parent_xml_node( atom node )
--
--Returns a pointer to the parent of this node or NULL if this node has no parent.
public function get_parent_xml_node( atom node )
    return c_func( WX_GET_PARENT_XML_NODE, {node})
end function

--/topic wxXmlNode
--/func get_xml_node_prop_val( atom node, sequence prop_name, sequence default )
--
--Returns the value of the property named /i prop_name if it does exist. If it does not exist, the /i default is returned.
public function get_xml_node_prop_val( atom node, sequence prop_name, sequence default )
    return c_func( WX_GET_XML_NODE_PROP_VAL, {node,prop_name,default})
end function

--/topic wxXmlNode
--/func get_xml_node_properties( atom node )
--
--Return a pointer to the first property of this node.
public function get_xml_node_properties( atom node )
    return c_func( WX_GET_XML_NODE_PROPERTIES, {node})
end function

--/topic wxXmlNode
--/func get_xml_node_type( atom node )
--
--Returns the type of this node.
public function get_xml_node_type( atom node )
    return c_func( WX_GET_XML_NODE_TYPE, {node})
end function

--/topic wxXmlNode
--/func xml_node_has_prop( atom node, sequence prop_name )
--
--Returns true if this node has a property named /i prop_name.
public function xml_node_has_prop( atom node, object prop_name )
    return c_func( WX_XML_NODE_HAS_PROP, {node,prop_name})
end function

--/topic wxXmlNode
--/func insert_xml_child( atom node, atom child, atom before_node )
--
--Inserts the /i child node immediately before /i before_node in the children list. If /i before_node is NULL, then /i child is prepended to the
-- list of children and becomes the first child of this node. Returns true if /i before_node has been found and the /i child node has been inserted.
public function insert_xml_child( atom node, atom child, atom before_node )
    return c_func( WX_INSERT_XML_CHILD, {node,child,before_node})
end function

--/topic wxXmlNode
--/func is_xml_whitespace_only( atom node )
--
--Returns true if the content of this node is a string containing only whitespaces (spaces, tabs, new lines, etc). Note that this function is
-- locale-independent since the parsing of XML documents must always produce the exact same tree regardless of the locale it runs under.
public function is_xml_whitespace_only( atom node )
    return c_func( WX_IS_XML_WHITESPACE_ONLY, {node})
end function

--/topic wxXmlNode
--/func remove_xml_node_child( atom node, atom child )
--
--Removes the given node from the children list. Returns true if the node was found and removed or false if the node could not be found.
--
--Note that the caller is reponsible for deleting the removed node in order to avoid memory leaks.
public function remove_xml_node_child( atom node, atom child )
    return c_func( WX_REMOVE_XML_NODE_CHILD, {node,child})
end function

--/topic wxXmlNode
--/proc set_xml_node_children( atom node, atom child )
--
--Sets as first child the given node. The caller is responsible to delete any previously present children node.
public procedure set_xml_node_children( atom node, atom child )
    c_proc( WX_SET_XML_NODE_CHILDREN, {node,child})
end procedure

--/topic wxXmlNode
--/proc set_xml_node_content( atom node, sequence con )
--
--Sets the content of this node.
public procedure set_xml_node_content( atom node, object con )
    c_proc( WX_SET_XML_NODE_CONTENT, {node,con})
end procedure

--/topic wxXmlNode
--/proc set_xml_node_name( atom node, sequence name )
--
--Sets the name of this node.
public procedure set_xml_node_name( atom node, object name )
    c_proc( WX_SET_XML_NODE_NAME, {node,name})
end procedure

--/topic wxXmlNode
--/proc set_next_xml_node( atom node, atom next )
--
--Sets as sibling the given node. The caller is responsible to delete any previously present sibling node.
public procedure set_next_xml_node( atom node, atom next )
    c_proc( WX_SET_NEXT_XML_NODE, {node,next})
end procedure

--/topic wxXmlNode
--/proc set_parent_xml_mode( atom node, atom parent )
--
--Sets as parent the given node. The caller is responsible to delete any previously present parent node.
public procedure set_parent_xml_mode( atom node, atom parent )
    c_proc( WX_SET_PARENT_XML_MODE, {node,parent})
end procedure

--/topic wxXmlNode
--/proc set_xml_node_properties( atom node, atom prop )
--
--Sets as first property the given wxXmlProperty object. The caller is responsible to delete any previously present properties attached to this node.
public procedure set_xml_node_properties( atom node, atom prop )
    c_proc( WX_SET_XML_NODE_PROPERTIES, {node,prop})
end procedure

--/topic wxXmlNode
--/proc set_xml_node_type( atom node, atom xml_type )
--
--Sets the type of this node.
public procedure set_xml_node_type( atom node, atom xml_type )
    c_proc( WX_SET_XML_NODE_TYPE, {node,xml_type})
end procedure



--wxXmlProperty
-----------------------------------------
--/topic wxXmlProperty
--/parent wxXmlNode
--/parent wxXmlDocument
--/parent Classes
--/info
--
--<div class="subsection">Creation Parameters</div>
--<ul>
--/li /b "sequence name"
--/li /b "sequence value"
--/li integer next [= 0]
--</ul>
--Represents a node property.
--
--Example: in &lt;img src="hello.gif" id="3"/&gt;, "src" is property with value "hello.gif" and "id" is a property with value "3".

--/topic wxXmlProperty
--/topic Non wxObjects
--/proc delete_wxXmlProperty( atom obj )
--
--Deletes a wxXmlProperty (use this instead of /delete_instance or /destroy)
public procedure delete_wxXmlProperty( atom obj )
    c_proc( WX_DELETE_WXXMLPROPERTY, {obj})
end procedure

--/topic wxXmlProperty
--/func get_xml_prop_name( atom prop )
--
--Returns the name of this property.
public function get_xml_prop_name( atom prop )
    return c_func( WX_GET_XML_PROP_NAME, {prop})
end function

--/topic wxXmlProperty
--/func get_next_xml_prop( atom prop )
--
--Returns the sibling of this property or NULL if there are no siblings.
public function get_next_xml_prop( atom prop )
    return c_func( WX_GET_NEXT_XML_PROP, {prop})
end function

--/topic wxXmlProperty
--/func get_xml_prop_value( atom prop )
--
--Returns the value of this property.
public function get_xml_prop_value( atom prop )
    return c_func( WX_GET_XML_PROP_VALUE, {prop})
end function

--/topic wxXmlProperty
--/proc set_xml_prop_name( atom prop, sequence name )
--
--Sets the name of this property.
public procedure set_xml_prop_name( atom prop, sequence name )
    c_proc( WX_SET_XML_PROP_NAME, {prop,name})
end procedure

--/topic wxXmlProperty
--/proc set_next_xml_prop( atom prop, atom next )
--
--Sets the sibling of this property.
public procedure set_next_xml_prop( atom prop, atom next )
    c_proc( WX_SET_NEXT_XML_PROP, {prop,next})
end procedure

--/topic wxXmlProperty
--/proc set_xml_prop_value( atom prop, sequence val )
--
--Sets the value of this property.
public procedure set_xml_prop_value( atom prop, object val )
    c_proc( WX_SET_XML_PROP_VALUE, {prop,val})
end procedure



--wxXmlResource
----------------------------------
--/topic wxXmlResource
--/parent Classes
--/info
--
--wxXmlResource is the class used to load XML-based resource files.
--
--The XML-based resource system, known as XRC, allows user interface elements such as dialogs,
-- menu bars and toolbars, to be stored in text files and loaded into the application at run-time.
-- XRC files can also be zipped into binary XRS files.
--
--Once an XRC file is loaded, call either /xrc_load_dialog or /xrc_load_frame.  This will load the
-- form and all controls.  Call /xrc_get_control to create references to any controls execpt menus
-- and tool bars.
--
--For menus, use /get_menu_bar to reference the menu bar and /get_menu_by_label to reference a
-- menu, /find_menuitem_by_position to reference a menuitem, and either /find_menuitem_id or
-- or /find_menuitem_by_position to retrieve a menuitem id.  One of the issues with using menus is
-- that IDs are not saved in the XRC file (even though some tools may lead you to think otherwise).

--/topic wxXmlResource
--/func xrc_get_control( atom resource, atom parent, object name )
--
--Returns a reference to the given control name.  Does not work for menus or
--toolbars.  Use /xrc_load_menu_bar, /xrc_load_menu or /xrc_load_tool_bar respectively.
public function xrc_get_control( atom resource, atom parent, object name )
    return c_func( WX_XRC_GET_CONTROL, {resource,parent,name})
end function

--/topic wxXmlResource
--/func xrc_get_xrcid( atom resource, object name )
--
--Returns wxNOT_FOUND if not found.
public function xrc_get_xrcid( atom resource, object name )
    return c_func( WX_XRC_GET_XRCID, {resource,name})
end function

--/topic wxXmlResource
--/func xrc_load_dialog( atom resource, atom parent, object name )
--
--Loads the dialog form.  /i parent can be 0.
public function xrc_load_dialog( atom resource, atom parent, object name )
    return c_func( WX_XRC_LOAD_DIALOG, {resource,parent,name})
end function

--/topic wxXmlResource
--/func xrc_load_panel( atom resource, atom parent, object name )
--
--Loads a panel. /i parent points to the parent window (if any).
-- This form is used to finish creation of an already existing instance.
public function xrc_load_panel( atom resource, atom parent, object name )
    return c_func( WX_XRC_LOAD_PANEL, {resource,parent,name})
end function

--/topic wxXmlResource
--/func xrc_load_frame( atom resource, atom parent, object name )
--
--Loads the frame.  /i parent should be 0 for main frame.
public function xrc_load_frame( atom resource, atom parent, object name )
    return c_func( WX_XRC_LOAD_FRAME, {resource,parent,name})
end function

--/topic wxXmlResource
--/func xrc_load_bitmap( atom resource, object name )
--
--Loads a bitmap resource from a file.
public function xrc_load_bitmap( atom resource, object name )
    return c_func( WX_XRC_LOAD_BITMAP, {resource,name})
end function

--/topic wxXmlResource
--/func xrc_load_icon( atom resource, object name )
--
public function xrc_load_icon( atom resource, object name )
    return c_func( WX_XRC_LOAD_ICON, {resource,name})
end function

--/topic wxXmlResource
--/func xrc_load_menu( atom resource, object name )
--
--Loads menu from resource. Returns NULL on failure.
public function xrc_load_menu( atom resource, object name )
    return c_func( WX_XRC_LOAD_MENU, {resource,name})
end function

--/topic wxXmlResource
--/func xrc_load_menu_bar( atom resource, atom parent, object name )
--
--Loads a menubar from resource. Returns NULL on failure.
public function xrc_load_menu_bar( atom resource, atom parent, object name )
    return c_func( WX_XRC_LOAD_MENU_BAR, {resource,parent,name})
end function

--/topic wxXmlResource
--/func xrc_load_tool_bar( atom resource, atom parent, object name )
--
--Loads a toolbar.
public function xrc_load_tool_bar( atom resource, atom parent, object name )
    return c_func( WX_XRC_LOAD_TOOL_BAR, {resource,parent,name})
end function


--Change History
------------------------------
--/topic Change History
--/info
--Enhancements and bug fixes
--<dl>
--
-- <dt> /b"v0.17.0"
--<dd><ul>
--/li Wrappers for /wxGraphicsMatrix, /wxGraphicsContext, /wxGraphicsPath
--/li Properly wrapped /set_selection() for wxSingleChoiceDialog
--/li added /find_menuitem_by_id()
--/li added /set_min_size()
--</ul></dd>
-- <dt> /b"v0.16.1"
--<dd><ul>
--/li Fixed wrappers for /select_grid_col and /seelct_grid_row
--</ul></dd>
-- <dt> /b"v0.16.0"
--<dd><ul>
--/li Updated /wxMenuItem to use wxItemKind instead of boolean for checkable
--/li Added Picker controls: /wxDirPickerCtrl, /wxFilePickerCtrl, /wxColourPickerCtrl, /wxFontPickerCtrl, /wxDatePickerCtrl
--/li Added /wxCollapsiblePane
--/li Added /use_primary_selection, /get_default_font_encoding
--/li Added /wxLocale
--/li Added /wxCSConv
--/li Added /wxRichTextPrinting
--/li Added /get_unicode_key
--/li Added convenience class /wxToolBarItem
--/li Added ifdef versioning
--</ul></dd>
-- <dt> /b"v0.15.0"
--<dd><ul>
--/li Added wxTreeCtrl hit test constants
--/li Added wxTreeEvents
--/li Better error reporting from errors inside C++ code
--/li Added /get_dc_gdk_window
--</ul>
-- <dt> /b"v0.14.0"
-- <dt> /b"v0.13.0"
-- <dd><ul>
--/li Added /b "mdi_editor.exw" demo to showcase MDI routines
--/li Added /wxSingleChoiceDialog, /wxTaskBarIcon
--/li Added bitmap parameter to /wxMenuItem
--/li Added /b BM_FROM_IMAGE to /wxBitmap
--/li Added /find_color()
--/li Added /get_display_size()
--/li More documentation updates
--/li Updated sequence/double creation to use euphoria v4.0
--compatible structures.
--/li Updated makefile.wat to work with OpenWatcom v1.8
--</ul>
--
-- <dt> /b"v0.12.2"
-- <dd><ul>
-- /li Updated to avoid conflicts with Euphoria v4.0 keywords
-- /li Fixed exported symbols on Windows
-- /li /wxXmlResource, /wxDisplay now built using GNU makefile
-- /li Requires library build #13
-- </ul>
--
--<dt> /b"v0.12.1"
--<dd><ul>
--/li Updated /wxHtmlWindow to support /wxHtmlLinkEvent
--/li Updated /wxDisplay and /wxXmlResource routines
--/li Added /b "search.exw" to demos (a nifty code-search utility)
--/li Added some /wxTextCtrl, /wxListCtrl and /wxStatusBar routines
--/li Added /hit_test, /get_window_scroll_position and /text_hit_test.
--/li Fixed bug in /show_text_position
--/li Fixed bug in /wxConfig routines
--/li Fixed wx_define routines to use debugging output
--/li Updated documentation
--</ul>
--
--<dt> /b"v0.12.0"
--<dd><ul>
--/li /wxBookCtrl: /wxChoiceBook, /wxListBook, /wxToolBook, /wxTreeBook
--/li /wxPlatformInfo, /wxImage
--/li Added /w_sprintf, /from_utf8, /to_utf8 to properly format unicode strings.
--/li Fixed /wx_printf to use w_sprintf instead of built-in sprintf
--/li Fixed /set_event_handler() for wxGTK and /wxMenu event handling to not
--  require the parent window of the menu to be passed
--/li Added /enable_tooltips, /set_tooltip_delay
--/li Added wrappers for /Sizers
--/li Added wrappers for /wxSplitterEvent
--/li Added /close_window and /can_veto
--/li Added wxGridCellAutoWrapStringEditor
--/li Added /get_facesizes
--/li Restored /wxStaticBoxSizer
--/li Bug fix to /add_create_func (now accepts a routine_id)
--/li Added /wxCaret, /wxDirCtrl, and /wxProcess
--/li Added all_demos.exw
--/li Updated /wxFileName routines
--/li Added /get_parent, /get_children, /get_desktop_window
--/li Added /client_to_screen, /screen_to_client
--/li Updated /wxMediaCtrl routines
--/li Added missing /b wxDIALOG_NO_PARENT constant
--/li Re-organized library:<ul>
--/li All same-class routines are in the same sections
--/li Moved /b Introduction to top of library (like Win32Lib)
--/li Moved /b "Change History" to bottom of library
--</ul>
--/li Updated documentation:<ul>
--/li Updated with new classes
--/li Fixed incorrect or out-dated information
--/li Added more descriptions for vague routines
--</ul></ul>
--
--<dt> /b"v0.11.0"
--<dd><ul>
--/li Now supports /Unicode
--/li Added /rmdir
--/li Added /select_grid_col, /select_grid_row, /select_grid_all, /is_in_grid_selection,
-- /select_grid_block, /enable_drag_grid_size, /set_grid_selection_mode,
-- /wait_on_connect, /get_list_event_column, /set_window_style, /get_window_style
--/li Added /b wxEVT_COMMAND_LIST_ITEM_RIGHT_CLICK
--/li Added /wxAuiManager, /wxAuiPaneInfo, /wxAuiNotebook
--/li More wrappers for /wxTextCtrl, /wxTreeCtrl, /wxGrid, /wxStatusBar, /wxMDIParentFrame
--/li Added ways to destroy /"Non wxObjects"
--/li Added /wxXmlDocument, /wxXmlNode, /wxXmlProperty
--/li Filled in wrappers for /wxFont
--/li /Printing (minimally) works
--/li Added /wxHtmlEasyPrinting
--/li Added some error handling to give library ability to report errors to Euphoria
--</ul>
--
--<dt> /b"v0.10.0"
--<dd><ul>
--/li Support for FreeBSD
--/li Reengineered the structure of the library, both Euphoria and C++
--/li Additions from Josef Jindra: /find_window, /get_col_size, /get_row_size,
--/li /listctrl_count, /listctrl_getitem, /wxStaticBitmap
--/li Documentation additions from Jesse Adkins
--/li Added many /wxNotebook wrappers (Jesse Adkins)
--/li Added /dir_exists, /mkdir.
--/li Fixed /wxToolBar constructor for linux (Jesse Adkins)
--/li Added /set_child_frame_title (Jesse Adkins)
--/li Fixed /wxStaticLine (bad constructor) (Jesse Adkins)
--/li /launch_default_browser(), /wxRegEx
--/li /wxMenu derives from wxEvtHandler, so event creation is easier
--</ul>
--
--<dt> /b"v0.9.1"
--<dd><ul>
--/li Removed some code that was left over from debugging (thanks to Jason Gade)
--/li Added /move() to allow changing the position of windows
--/li Added several /wxTextCtrl routines (thanks to Gbadebo Oladosu)
--/li Added /get_red, /get_green and /get_blue to get the color components of a /wxColour
--</ul>
--
--<dt> /b"v0.9.0"
--<dd><ul>
--/li Made class_table local in cplusplus.e to avoid conflicts with embedded ooeu
--/li Documented /set_cursor()
--/li Added /convert_to_image()
--/li Requires library build #9
--</ul>
--
--<dt> /b"v0.8.1"
--<dd><ul>
--/li Fixed bug in /wxWizardPage -- vtable offset was changed
--/li Added /wxSplashScreen
--/li Added /get_days_in_month() and /get_cal_days_in_month()
--/li Fixed bug that could prevent wxEVT_CLOSE_WINDOW events from being handled correctly
--/li Fixed bug that would prevent translation to a dll, when wxEuphoria looks at the command_line()
--</ul>
--
--<dt> /b"v0.8.0"
--<dd><ul>
--/li Fixed bug in /wxTreeCtrl creation for detecting optional parameters
--/li New classes: /wxFileName, /wxHtmlWindow
--/li New wrappers: /select_tree_item(), /delete_tree_children(), /is_visible()
--/li Both versions use build #8.
--/li Added /call_member_str() and /call_cdecl_str() too allow easy wrapping of
-- functions that return a wxString by value.
--/li Added /init_image_handlers() and fixed constructor for wxGIFHandler for Win32
--/li /set_column_label() didn't work properly under Windows.
--/li Added /set_bmpbutton() to change the displayed bitmap of a /wxBitmapButton
--/li wxeu.so now belongs in "/usr/lib"
--/li Added new accessor functions for /wxFont.
--/li Added /choice_count(), /combobox_count()
--/li Global sequence PATH contains the path of the startup directory, if possible
--</ul>
--
--<dt> /b"v0.7.2"
--<dd><ul>
--/li Documentation was wrong for /set_editor_choices()
--/li Fixed bug in /ftp_rename() (Thanks to Mario Steele)
--/li /set_ip_host() was using the wrong wxWidgets call under Linux
--          (Thanks to Mario Steele)
--</ul>
--
--<dt> /b"v0.7.1"
--<dd><ul>
--/li Fixed /get_list_string() to work for /wxListCtrl on Linux
--/li Fixed constructor for wxURL on Linux
--/li Linux uses library build #7.
--</ul>
--
--<dt> /b"v0.7.0"
--<dd><ul>
--/li Fixed /set_brush() and /set_back_brush() to work with /wxMemoryDC under GTK
--/li New classes: /wxMDIParentFrame, /wxMDIClientWindow, /wxMDIChildFrame, /wxFontEnumerator, /wxCursor
--/li Added /set_choice() for /wxChoice
--/li Can now insert a /wxMenu at specified location in the /wxMenuBar upon creation
--/li Added /get_label(), which should work for all controls.
--/li Fixed /get_list_string() under GTK.
--/li Uses library build #6.
--/li Several bugfixes for /wxTreeCtrl under GTK.
--/li Added /"Resource Management" routines (thanks to Mario Steele)
--</ul>
--
--<dt> /b"v0.6.0"
--<dd><ul>
--/li Docs now in multiple file format
--/li CK Lester helped updating demos
--/li Added /set_bitmap_size() and made /get_bitmap_size() work under Linux
--/li /set_checked() didn't work for /wxToggleButton (CK Lester)
--/li Added /get_text_number()
--/li New classes: /wxMask, /wxFlexGridSizer, /wxStaticBoxSizer
--/li Added /set_mask() to use a /wxMask with a /wxBitmap
--/li Added public constant wxTrue
--/li Uses build 5
--/li Added several new /wxSizer functions.
--/li Added /get_key_state() (from wxMSW 2.5.5, and implemented in Euphoria for wxGTK,
-- since the wxGTK implementation only works for shift, ctrl, alt keys)
--/li Fixed bug in Linux where wxWidgets frees image handlers and that memory
-- is reused, cplusplus.e still thought the object was of the old class' type
-- (thanks to Eumario)
--/li Added /font_ok() to check for a valid /wxFont
--</ul>
--
--<dt> /b"v0.5.0"
--<dd><ul>
--/li New classes: /wxTextAttr, /wxStaticLine, /wxStaticBox, /wxCalendarCtrl
--/li Fixed constructor for /wxTextCtrl so that default parameters can be used for
-- id, text, size and position.
--/li wxEuphoria will display an error message and abort if open_dll( wxWin_name ) fails
--/li Cleaned up warnings
--/li Size of wxColour was wrong under Windows
--/li Can now /destroy() controls
--/li Added version constant wxEuphoria_Version = {0,5,0}
--/li Can set the column labels in a /wxListCtrl: /set_column_label()
--/li Fixed bug in constructor for /wxToggleButton
--/li Uses library build 4.   The file is now named wxeu.[dll|so]
--/li Allows you to change the location or name of the wxEuphoria library by
-- declaring a public function /wxEu_Lib that returns the path to the
-- wxEuphoria library.
--/li /set_text_back_color() works correctly in Linux
--/li Change instance memory handling to cooperate with wxWidgets under Win32
-- so that memory is freed in the correct place
--/li /cleanup() added.  This should be manually called before your app exits
-- if you don't use /wxMain() ( /wxMain() calls /cleanup() automatically before exit)
--/li /set_status_text() now checks to make sure that you don't set a status field that doesn't exist
-- (thanks to CK Lester)
--/li Global constant wxFalse is now 0 (was a pointer to the actual value) (thanks to CK Lester)
--/li Added /set_button_label()
--</ul>
--
--<dt> /b"v0.4.0"
--<dd><ul>
--/li New classes:
-- /wxDir, /wxDirTraverser, /wxWizard, /wxWizardPage, /wxSocketClient,
-- /wxSocketServer, /wxURL, /wxInputStream, /wxIPV4Address, /wxHTTP, /wxFTP
--/li Fixed destructor for /wxArrayString
--</ul>
--
--<dt> /b"v0.3.0"
--<dd><ul>
--/li New classes:<ul>
--/li /Controls: /wxTreeCtrl, /wxSplitterWindow, /wxSpinCtrl,
-- /wxSlider, /wxListCtrl
--/li /Dialogs: /wxDirDialog, /wxFontDialog (see also /font_selector),
-- /wxPrintDialog, /wxProgressDialog, /wxColourDialog
-- (see also /color_selector)
--/li /Printing: /wxPreviewFrame, /wxPrintout, /wxPrintPreview
--/li /wxGrid: /wxGridCellAttr, /wxGridCellEditor
--</ul>
--/li /wxImageList
--/li /wxDialog has been moved from wxEuphoria.e to the new file wxDialog.e
--/li Support for popup- and sub- menus (See /Menus)
--/li /create(), /call_member(), /call_member_s(), /call_cdecl(), /call_cdecl_s()
-- will now take an atom if only one parameter is passed
--/li A /wxString can be created without wrapping the string in
-- an extra set of braces:
--/code
-- str = create( wxString, "A String" )
--/endcode
--<br/>
--/li Uses Libarary build 3
--/li Checks for Library build >= Required build (used to be =)
--/li platform.pl ported to Euphoria (platform.exw)
--/li /blit() parameters were incorrect (src and dst coords were swapped)
--/li /wx_execute() is now a function, and takes flags
--</ul>
--
--<dt> /b"v0.2.0"
--<dd><ul>
--/li New classes:<ul>
--/li /Sizers: /wxBoxSizer, /wxGridSizer, /wxNotebookSizer
--/li /wxNotebook
--/li /wxGrid
--/li /wxToolBar
--/li /wxBitmapButton
--/li Clipboard, Drag'n'Drop: /wxClipboard, /wxDataObject, /wxTextDataObject,
-- /wxFileDataObject, /wxDropTarget, /wxFileDropTarget
--</ul>
--/li New functions:<ul>
--/li /file_selector()
--/li /get_sys_metric()
--/li /get_sys_color()
--</ul>
--/li Uses Library build 2<ul>
--/li Both<ul>
--/li Added string access routines ( /wxChoice, /wxComboBox, /wxGrid, /wxClipboard)
--/li Moved implementation of some routines from grid.h to grid.cpp to
-- allow them to be exported</ul>
--/li Win32:<ul>
--/li Added call to IDataObject::AddRef when adding data to the clipboard
--/li GetColour added for use in /get_sys_colour </ul>
--/li Dumped fptr.e and added some code in src\common\utilscmn.cpp for calling
-- functions.  Random crashes have disapeared.  Still seem to need string
-- accessors, though.
--</ul></ul>
--
--<dt> /b"v0.1.0"
--<dd><ul>
--/li Using wxWidgets 2.4.2<ul>
--/li Win32: Borland C++ v5.5 (free version)
--/li Linux gcc v3.2
--/li Library build 1 for both platforms (WX_DLL_BUILD)
--</ul>
--/li Overhauled and reengineered library development<ul>
--/li All files are *.e with platform specific code wrapped with
-- 'preprocessor' directives (see platform.pl)
--/li Library exports managed through EuSQL databases
--/li generate.exw, wxeu.exw, EDB forms included to assist db and
-- file maintenance
--</ul></ul>
--
--<dt> /b"7/3/2003"
--<dd><ul>
--/li Got rid of cross-platform + platform-specific includes.  Combined
-- into platform-specific format.
--</ul></dl>
